/**(c)

  Copyright (C) 2006, 2007, 2008, 2009 Michael Stilkerich, Johannes Bauer, Christian Wawersich 

  This file is part of the JOSEK operating system.

	JOSEK is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

	JOSEK is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with JOSEK. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
//? false
/* JOSEK BCC Scheduler
 *
 * Common Datastructures
 *   - current_task: task identifier of the running task, INVALID_TASK
 *       if no task (i.e. the idle task) is currently running.
 *   - __JOSEK_launch_functions[]: Array of launch function pointers,
 *       indexed by task identifiers.
 *   - readyplist: Bitmap for Ready Priorities. Contains one bit per
 *       priority level. A set bit means that at least one task of the
 *       priority level is ready or running. Bit 0 corresponds to the
 *       highest priority level. The priority bit for the special
 *       priority of the Idle Task (PRIO_IDLE_TASK) must always remain
 *       set.
 *   - JOSEK_sched_enabled: boolean value, indicating if the scheduler
 *       is currently enabled or disabled. FIXME purpose ISR2??
 *
 * Variant 1: Single Task Activation, Single Task per Priority
 *   - no additional data structures are required
 *   - ready tasks are marked by setting the task priority ready,
 *     suspended tasks have the priority-ready bit cleared. 
 * 
 * Variant 2: Multiple Task Activations, Single Task per Priority
 *   - Additional Datastructures
 *     * __JOSEK_task_actcounter[]: Array of activation counters,
 *       indexed with task identifiers. Counting is performed
 *       downwards, i.e. the max. number of activations is reached as
 *       the counter becomes 0.
 *     * __JOSEK_task_maxact[]: constant array with the maximum number
 *       of task activations for each task, indexed with task ids.
 *
 *   - ready, suspended and running tasks are detected like in Var. 1
 *   - on task activation and termination, the activation counters
 *     need to be updated and checked
 *
 * Variant 3: Single Task Activation, Multiple Tasks per Priority
 *   - Additional Datastructures
 *     * readytlist: Bitmap with one bit corresponding to one task
 *       identifier. A set bit means that the task is ready.
 *     * readyqueue: Array of ready tasks' identifiers. Subareas of
 *       the array correspond to the different priority levels.
 *     * bbqstate[]: Array indexed by priority levels maintaining
 *       the state of the priority's readyqueue: Start and End index of
 *       the prioritiy's portion in the readyqueue, index of the head
 *       and the tail element, and the fill level.
 *   - Ready and Suspended tasks can be relatively cheaply determined
 *     by checking the readytlist. Otherwise the readyqueue of the
 *     corresponding priority would have to be traversed.
 *   - The readyqueue contains on slot per task of the priority level,
 *     plus one slot for ceiling priority levels.
 *
 * Variant 4: Multiple Task Activations, Multiple Tasks per Priority
 *  - Additional Datastructures
 *    * Activation Counters of Variant 2
 *    * readyqueue, readytlist and bbqstate of Variant 3
 *  - The readyqueue contains one slot for each possible activation of
 *    each task in the priority level, plus one slot for ceiling prios
 *  - We do not need max activations, because we can now determine a
 *    non-ready priority when its readyqueue is empty
 *  - Ready and Suspended tasks are determined by checking the readytlist
 */
//?

/* Non-preemptive Scheduler for 1 Task/Priority, Single Activation */
extern void ClearStackPart(void) __attribute__((noreturn));
extern void StartTask(void (*) (void));

StatusType GetTaskState(TaskType TaskID, TaskStateRefType State) {
	register MACHINEWORD rpl;
	register TaskType ct;

//? have_extstatus
	if(TaskID >= INVALID_TASK)
		return E_OS_ID;
//?
	rpl=readyplist;
	ct=current_task;

	if(ct==TaskID) {
		*State = RUNNING;
//? have_multi_act && !have_readyqueues
	} else if(__JOSEK_task_actcounter[TaskID] < __JOSEK_task_maxact[TaskID]) {
//! have_readyqueues
	} else if(readytlist & (1<<TaskID)) {
//!
	} else if(rpl & (1<<TaskID)) {
//?
		*State = READY;
	} else {
		*State = SUSPENDED;
	}
	return E_OK;	
}

/* TODO hooks */
static StatusType activatetask(TaskType);

void schedule(uint8_t term) {
	volatile TaskType ptask;
	register MACHINEWORD rdytask;

	DisableAllInterrupts();
//? have_readyqueues
	if(term != 0) {
//!
	if(term == 1) {
//?
//? have_multi_act
		__JOSEK_task_actcounter[current_task]++;
//?

//? have_readyqueues
		if(remove_element(BasePriority(current_task))==0)
//! have_multi_act
		if(__JOSEK_task_actcounter[current_task]==__JOSEK_task_maxact[current_task])
//?
		/* Mark current priority as not ready */
		readyplist &= ~(1<< BasePriority(current_task));

//? have_readyqueues
		/* Mark task as not ready */
		readytlist &= ~(1 << current_task);
//?

//? have_readyqueues
		if(term == 2) { /* Chain Self */
			activatetask(current_task);
		}
//?
	}

	/* Which is the task priority that is runnable? */
	bitsearch(readyplist, rdytask);

//? have_readyqueues
	if(rdytask == PRIO_IDLE_TASK) {
		rdytask = INVALID_TASK;
	} else {
		rdytask = head_element(rdytask); 
	}
//?

	if(term==0) {
		if(rdytask==current_task) {
			EnableAllInterrupts();
			return;
		}
//? have_resources
		/* If the ceiling priority is higher or equal than that of the chosen task,
		 * we return and let the task in the critical section finish.
		 */
		if(JOSEK_used_resources_stack[JOSEK_used_resources_index] <= BasePriority(rdytask)) {
			EnableAllInterrupts();
			return;
		}
//?
		ptask = current_task;
//? have_posttask_hook
		if(current_task != INVALID_TASK) {
			PostTaskHook();
		}
//?
		current_task = rdytask;
//? have_pretask_hook
		if(current_task != INVALID_TASK) {
			PreTaskHook();
		}
//?
		do {
			StartTask(__JOSEK_launch_functions[current_task]);
			/* if a task returns without calling TerminateTask() or
			 * ChainTask(), it will remain activated. This conforms
			 * to the OSEK OS spec, which requires tasks to call one
			 * of these services
			 */
		} while(current_task != ptask);
	} else {
//? have_posttask_hook
		if(current_task != INVALID_TASK) {
			PostTaskHook();
		}
//?
		current_task = rdytask;
//? have_pretask_hook
		if(current_task != INVALID_TASK) {
			PreTaskHook();
		}
//?
		/* Clear the Tasks stack partition */
		ClearStackPart();
		/* no return to here, ClearStackPart() will return from StartTask() above */
	}

	/* else fall through to the preceeding task, eventually the idle task */
	EnableAllInterrupts();
}

