/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.classfile;

import java.io.*;
import keso.classfile.constantpool.*;

/**
	All data about the bytecode of a method, that can be found in a
	class file (code attribute of method). This is a rather passive class.
	It only reads the data from the class file and stores it.
	*/
final public class CodeData {

	int maxStack;
	int maxLocals;

	byte[] codeBytes;
	int numHandlers;
	private ExceptionHandlerData[] exceptionHandler;

	private LineAttributeData[] lineNumberTable;
	private LocalVariableData[] localVariableTable;

	//needed by jx.verifier.wcet.WCETAnalysis.createETCheckedMethod
	public void increaseMaxLocals() {maxLocals++;}
	public void increaseMaxStack() {maxStack++;}

	public void setCodeBytes(byte[] newCode) {codeBytes = newCode;}

	public String methodName;
	public CodeData(MethodData m) { methodName=m.getMethodName(); }
	public CodeData(DataInput input, ConstantPool cPool)
		throws IOException {
		readFromClassFile(input, cPool);
	}

	//Constructor for copy
	private CodeData(int maxStack,
			int maxLocals,
			byte[] codeBytes,
			ExceptionHandlerData[] exceptionHandler,
			LineAttributeData []lineNumberTable) {

		this.maxStack = maxStack;
		this.maxLocals = maxLocals;
		this.codeBytes = new byte[codeBytes.length];

		for (int i = 0; i < codeBytes.length; i++)
			this.codeBytes[i] = codeBytes[i];

		numHandlers = exceptionHandler.length;
		this.exceptionHandler = new ExceptionHandlerData[numHandlers];
		for (int i = 0; i < numHandlers; i++)
			this.exceptionHandler[i] = exceptionHandler[i].copy();
		this.lineNumberTable = new LineAttributeData[lineNumberTable.length];
		for (int i = 0; i < lineNumberTable.length; i++)
			this.lineNumberTable[i] = lineNumberTable[i].copy();
	}

	public CodeData copy() {
		return new CodeData(maxStack, maxLocals, codeBytes, exceptionHandler,
				lineNumberTable);
	}

	public int getNumStackSlots() {return maxStack;}

	public int getNumLocalVariables() {return maxLocals;}

	public byte[] getBytecode() {return codeBytes; }

	public int getBytecodeLength() {return codeBytes.length;}

	public LineAttributeData[] getLineNumberTable() { return lineNumberTable; }
	public LocalVariableData[] getLocalVariableTable() { return localVariableTable; }

	public ExceptionHandlerData[] getExceptionHandlers() {
		return exceptionHandler;
	}

	/**
	 * read class file code attribute
	 * @see <a href=http://java.sun.com/docs/books/jvms/second_edition/html/ClassFile.doc.html#1546>
	 *      JVM Spec 4.7.3 The Code Attribute</a>
	 */
	public void readFromClassFile(DataInput input, ConstantPool cPool)
		throws IOException {
		// bytes_count already read
		maxStack = input.readUnsignedShort();
		maxLocals = input.readUnsignedShort();

		int numCodeBytes = input.readInt();
		codeBytes = new byte[numCodeBytes];
		input.readFully(codeBytes, 0, numCodeBytes);

		// System.out.println(getDescription(cPool));

		numHandlers = input.readUnsignedShort();
		exceptionHandler = new ExceptionHandlerData[numHandlers];
		for(int i=0; i<numHandlers; i++) {
			exceptionHandler[i] = new ExceptionHandlerData();
			exceptionHandler[i].readFromClassFile(input, cPool);
		}

		int numAttributes = input.readUnsignedShort();
		for(int i=0; i<numAttributes; i++)
			readAttribute(input, cPool);
	}

	private void readAttribute(DataInput input, ConstantPool cPool)
		throws IOException {
		int attrNameCPIndex = input.readUnsignedShort();
		int numBytes = input.readInt();

		String attrName = cPool.getUTF8StringAt(attrNameCPIndex);
		if (attrName.equals("LineNumberTable")) {
			int numLines = input.readShort();
			lineNumberTable = new LineAttributeData[numLines];
			for(int i=0; i<numLines; i++) {
				lineNumberTable[i] = new LineAttributeData(input.readShort(), input.readShort());
			}
		} else if( attrName.equals("LocalVariableTable")) {
			int num_lv = input.readShort();
			localVariableTable = new LocalVariableData[num_lv];
			for(int i=0; i<num_lv; i++) {
				localVariableTable[i] =
					new LocalVariableData(cPool, input.readShort(), input.readShort(),
							input.readShort(),input.readShort(),input.readShort());
			}
		} else {
			input.skipBytes(numBytes);
		}
	}

	public String toString() {
		return super.toString() + "\n" +
			"MaxStack     : " + maxStack + "\n" +
			"maxLocals    : " + maxLocals + "\n" +
			"numCodeBytes : " + codeBytes.length + "\n";
	}

	public String getDescription(ConstantPool cPool) {
		return toString();
	}

}
