/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.classfile.constantpool;

import java.io.*;
import keso.util.*;

/**
    This class represents a constant pool of a class.
    It stores a array of entries, each of which contains the
    data of a constant pool entry.
    It can be initialized either with a java class file, or with
    a instance of MetaxaConstantPool, or with a ClassInterface.
    The last way is actually deprecated, and not used anymore.

    When initialized with a MetaxaConstantPool, the entries are
    loaded on demand, i.e. an entry is loaded when it is referenced
    using one of the entryAt()-Methods. Only entry types that are actually
    referenced by bytecode can be loaded through MetaxaConstantPool
    (e.g. NameAndTypeCPEntry's can not).
*/
public class ConstantPool {

	// the number of entries in the constant pool + 1
	private int numEntries;

	/** the array of cp_info structures, numEntries in size.
	 * 0 is not a valid index into the CP (see JVM Spec).
	 * The first entry in this array is therefore not valid.
	 * A dummy entry will be placed in there.
	 */
	private ConstantPoolEntry[] entries;

	/**
	  necessary to access string entries of this constantpool through
	  the interface in bytecode.execenv.ExecEnvironment
	  */
	private String className, superClassName;

	/**
	  constructors for reading from classFile
	  (call readFromClassFile() after using this constructor)
	  */
	public ConstantPool() {
		//    throw new RuntimeException(); // this constructor is unsafe, does not create the entry array!
	}

	/**
	  constructors for reading from classFile
	  */
	public ConstantPool(DataInput input) throws IOException {
		readFromClassFile(input);
	}

	/**
	  set the index of the entry that contains the name of the
	  constantpools class.
	  */
	public void setThisClassCPIndex(int thisClassCPIndex) {
		className = ((ClassCPEntry)entryAt(thisClassCPIndex)).getClassName();
	}

	public void setSuperClassCPIndex(int cpIndex) {
		superClassName = ((ClassCPEntry)entryAt(cpIndex)).getClassName();
	}

	/**
	  this method is necessay to access the address of
	  strings in the constant pool through the interface int
	  bytecode.execenv.ExecEnvironment
	  */
	public String getClassName() {
		return className;
	}

	public String getSuperClassName() {
		return superClassName;
	}

	/**
	  read all entries from a class file
	  */
	public void readFromClassFile(DataInput input) throws IOException {
		numEntries = input.readUnsignedShort();
		entries = new ConstantPoolEntry[numEntries];
		entries[0] = new DummyCPEntry();
		int i = 1;

		/* Read the constant pool. The number of entries is numEntries-1,
		 * valid indices range from 1 to numEntries.
		 */
		while (i < numEntries) {
			ConstantPoolEntry newEntry=null;
			int tag = input.readUnsignedByte();

			switch (tag) {
				case ConstantPoolEntry.CONSTANT_UTF8:
					newEntry = new UTF8CPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_INTEGER:
				case ConstantPoolEntry.CONSTANT_FLOAT:
				case ConstantPoolEntry.CONSTANT_LONG:
				case ConstantPoolEntry.CONSTANT_DOUBLE:
					newEntry = new NumericCPEntry(tag);
					break;
				case ConstantPoolEntry.CONSTANT_CLASS:
					newEntry = new ClassCPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_STRING:
					newEntry = new StringCPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_FIELDREF:
					newEntry = new FieldRefCPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_METHODREF:
					newEntry = new MethodRefCPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_INTERFACEMETHODREF:
					newEntry = new InterfaceMethodRefCPEntry();
					break;
				case ConstantPoolEntry.CONSTANT_NAMEANDTYPE:
					newEntry = new NameAndTypeCPEntry();
					break;
				default:
					Debug.ASSERT(false, "This must not happen "+tag);
					break;
			}

			newEntry.readFromClassFile(input);
			newEntry.setCPIndex(i);

			this.entries[i++] = newEntry;

			if (tag == ConstantPoolEntry.CONSTANT_LONG ||
					tag == ConstantPoolEntry.CONSTANT_DOUBLE)
				this.entries[i++] = new DummyCPEntry();
		}

		// link the entries
		for(int j=0;j<numEntries;j++)
			entries[j].linkCPEntries(this);
	}


	// ***** Methods to access the entries *****

	public ConstantPoolEntry constantEntryAt(int index) {
		if (entries[index]==null) {
			Debug.throwError();
		}
		return entries[index];
	}


	public ClassCPEntry classEntryAt(int cpIndex) {
		if (entries[cpIndex]==null) {
			Debug.throwError();
		}
		if (!(entries[cpIndex] instanceof ClassCPEntry)) {
			Debug.out.println("cpIndex="+cpIndex+", entry="+getEntryStringAt(cpIndex));
			Debug.throwError("wrong CP entry type (expect ClassCPEntry)");
		}
		return (ClassCPEntry)entries[cpIndex];
	}

	public MethodRefCPEntry methodRefEntryAt(int index) {
		if (entries[index]==null)  {
			Debug.throwError();
		}
		return (MethodRefCPEntry)entries[index];
	}

	public FieldRefCPEntry fieldRefEntryAt(int index) {
		if (entries[index]==null)  {
			Debug.throwError();
		}
		return (FieldRefCPEntry)entries[index];
	}

	public InterfaceMethodRefCPEntry InterfaceMethodRefEntryAt(int index) {
		if (entries[index]==null)  {
			Debug.throwError();
		}
		return (InterfaceMethodRefCPEntry)entries[index];
	}

	/**
	  only if the entries are loaded from a class file,
	  you can use this method to access them. Otherwise, this
	  class will throw an error.
	  It is better to use the methods above for this task.
	  */
	public ConstantPoolEntry entryAt(int index) {
		if (index == -1) {
			throw new Error("Attempt to access a dynamically generated CPEntry via an index.");
		}
		if (entries[index]==null)
			Debug.throwError("Accessing of entries not allowed");

		return entries[index];
	}

	public String toString() {
		String string = "";
		for(int i=0;i<numEntries;i++) {
			string += ""+i+" : "+entries[i].getTypeString() + ": " +
				entries[i].getDescription(this, true) + "\n";
		}
		return string;
	}

	public int getNumberOfEntries() { return numEntries; }

	public String getEntryStringAt(int cpIndex) {
		return "(" + cpIndex + ") " + entries[cpIndex].getDescription(this, true);
	}

	public String getUTF8StringAt(int cpIndex) {
		if (! (entries[cpIndex] instanceof UTF8CPEntry)) {
			Debug.out.println("cpIndex="+cpIndex+", entry="+getEntryStringAt(cpIndex));
			Debug.throwError("wrong CP entry type");
		}
		return ((UTF8CPEntry)entries[cpIndex]).value();
	}

	/** @return cp index of added entry */
	public int addEntry(ConstantPoolEntry newEntry) {
		int index = numEntries;
		newEntry.setCPIndex(index);
		if (entries.length == numEntries) {
			ConstantPoolEntry[] e = new ConstantPoolEntry[numEntries + 10];
			System.arraycopy(entries, 0, e, 0, numEntries);
			entries = e;
		}
		entries[index] = newEntry;
		numEntries++;
		return index;
	}

}
