/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.util.zip.*;
import keso.compiler.analysis.*;
import keso.compiler.config.*;
import keso.compiler.dump.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.*;
import keso.util.*;

final public class Builder {

	private final BuilderOptions opts;
	private ClassStore repository;
	private Thread stdoutthread;
	private Thread stderrthread;

	public Builder(BuilderOptions opts) {
		this.opts=opts;
	}

	private Map<String,? extends Collection<String>> findFilesToPrecompile(String[] modules) {
		// collect existing precompiled files
		Map<String,File> delete_files = new HashMap<String,File>();
		String classpath = opts.getClassDir();
		File classpathfile = new File(classpath);
		if(classpathfile.exists()) {
			findFiles(classpathfile, classpathfile.getPath().length()+1, delete_files, ".class");
		}

		// collect source files of all modules
		final HashSet<String> seen_files = new HashSet<String>();

		Map<String,Vector<String>> result = new HashMap<String, Vector<String>>();

		for(int i=modules.length-1; i>=0; i--) {
			Vector<String> mresult = new Vector<String>();
			String modpath = opts.getAppSourcePath() +modules[i];
			File moduleDir = new File(modpath);

			// only process regular modules here
			if(!moduleDir.isDirectory())
				continue;

			Map<String,File> source_files = new HashMap<String,File>();
			findFiles(moduleDir, modpath.length()+1, source_files, ".java");

			for(String jfile: source_files.keySet()) {

				// file overridden by a higher-priority module?
				if(seen_files.add(jfile)) {
					File sourcefile = source_files.get(jfile);

					// remove from the delete list if existing
					File existing = delete_files.remove(jfile);
					if(existing == null ||
							existing.lastModified() < sourcefile.lastModified()) {

						// add to the todo list if the dest file does not exist
						// or if the source file was recently modified
						mresult.add(sourcefile.getPath());
					}
				}
			}

			result.put(modules[i], mresult);
		}

		for(File delfile: delete_files.values()) {
			// delete inner classes only if the main file is deleted, too
			int dindex = delfile.getName().indexOf('$');
			if(dindex >= 0) {
				String p = delfile.getPath();
				int index_to_end = delfile.getName().length() - dindex;
				String outerclass = p.substring(0,p.length()-index_to_end) + ".class";
				if(!delete_files.containsKey(outerclass))
					continue;
			}

			delfile.delete();
		}

		return result;
	}

	private void findFiles(File f, int stripprefix, Map<String,File> files, String suffix) {
		if (f.isDirectory()) {
			for(File e: f.listFiles()) {
				if (!e.getName().startsWith(".")) {
					findFiles(e, stripprefix, files, suffix);
				}
			}
		} else if (f.getName().endsWith(suffix)) {
			String p = f.getPath();
			files.put(p.substring(stripprefix,p.length()-suffix.length()), f);
		}
		// else: ignore all non-Java files
	}

	/**
	 * Recursively collects all Java files in the given module.
	 *
	 * @param module File object denoting the root path of the module
	 * @param work_todo Collection of Strings that the pathnames of the module's Java files are appended to
	 */
	private void collectWorkToDo(File module, Collection<String> work_todo) {
		if (module.isDirectory()) {
			for(String entry: module.list()) {
				if (entry.startsWith(".")) continue;
				collectWorkToDo(new File(module.toString()+"/"+entry), work_todo);
			}
		} else {
			if (module.getName().endsWith(".java")) {
				String todo = module.toString();
				work_todo.add(todo);
			}
			// else: ignore all non-Java files
		}
	}

	private void copy(String src, String dst) {
		int pos = dst.lastIndexOf(File.separatorChar);
		File dfile = new File(dst);
		File ddir = new File(dst.substring(0,pos));
		byte buffer[] = new byte[4096];
		int len;

		try {
			ddir.mkdirs();
			dfile.createNewFile();
			FileInputStream in = new FileInputStream(src);
			FileOutputStream out = new FileOutputStream(dst);
			while ((len=in.read(buffer))>-1) { out.write(buffer,0,len); }
			in.close();
			out.close();
		} catch (IOException ex) {
			opts.warn("copy "+src+" fails!");
			opts.verbose(ex);
			System.exit(1);
		}
	}

	private int javac(Collection<String> args) {
		String[] nargs = new String[args.size()];
		StringBuffer out = new StringBuffer("javac");
		int i=0;

		for(String s: args) {
			nargs[i++] = s;
			out.append(" ");
			out.append(s);
		}
		opts.verbose(out);

		return javac(nargs);
	}

	/**
	 * execute the java compiler from sun.
	 */
	private int javac(String[] args) {
		try {
			java.lang.Class<?> c = java.lang.Class.forName("com.sun.tools.javac.Main");
			Class<?>[] parameterTypes = new Class[1];
			parameterTypes[0] = args.getClass();
			java.lang.reflect.Method method = c.getMethod("main", parameterTypes);
			Object pars[] = new Object[1];
			pars[0] = args;
			try {
				method.invoke(null, pars);
			} catch (java.lang.reflect.InvocationTargetException e) {
				throw e.getTargetException();
			}
		} catch (NoExitException ex) {
			return ex.status;
		} catch (Throwable ex) {
			Debug.out.println(ex);
		}

		return 0;
	}

	public void startOutputThreads(final BufferedReader stdout, final BufferedReader stderr) {
		stdoutthread = new Thread() {
			public void run() {
				try {
					String text;
					while ((text = stdout.readLine()) != null) {
						System.out.println(text);
					}
				} catch (Exception e) {

				}
			}
		};
		stderrthread = new Thread() {
			public void run() {
				try {
					String text;
					while ((text = stderr.readLine()) != null) {
						System.err.println(text);
					}
				} catch (Exception e) {

				}
			}
		};

		stdoutthread.start();
		stderrthread.start();
	}

	public void waitForEndOfOutput() {
		try { stdoutthread.join(); } catch (InterruptedException ex) {}
		try { stderrthread.join(); } catch (InterruptedException ex) {}
	}

	public void callMake() throws IOException {
		Runtime runtime = Runtime.getRuntime();
		Process make = runtime.exec("sh "+opts.getFirstLevelOutputPath()+"/make.sh");
		startOutputThreads(new BufferedReader(new InputStreamReader(make.getInputStream())),
				new BufferedReader(new InputStreamReader(make.getErrorStream())));
		waitForEndOfOutput();
	}

	public void writeFirstLevelMakefile() throws IOException {

		String opath = opts.getFirstLevelOutputPath();
		File opathfile = new File(opath);
		opathfile.mkdirs();

		String doit = opath+"/make.sh";
		PrintStream doit_out = new PrintStream(new FileOutputStream(doit));
		doit_out.println("# THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT\n");
		doit_out.println("cd "+opath);
		doit_out.println("make");
		doit_out.close();

		String makefile = opath+"/Makefile";
		PrintStream out = new PrintStream(new FileOutputStream(makefile));
		out.println("# THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT\n");
		out.println(".PHONY: all clear\n");

		WorldDefinition worldDef = opts.getWorldDefinition();

		out.println("all:");

		for(SystemDefinition sys: worldDef.getSystems()) {
			out.print("\t@cd ");
			out.print(sys.getIdentifier());
			out.print("\t; make 2>&1");
			if (opts.hasOption("log_backend"))
				out.print(" > last-compile.log ");
			out.println("");
		}

		out.println("\nclean:");
		for(SystemDefinition sys: worldDef.getSystems()) {
			out.print("\t@cd ");
			out.print(sys.getIdentifier());
			out.println("\t; make clean");
		}
		out.close();
	}

	/**
	 * Deletes a file or directory tree, omitting files that do not match a
	 * pattern.
	 *
	 * The root of the tree itself is not deleted.
	 *
	 * @param path    File or directory to be deleted.
	 * @param pattern A regular expression. Only files matching this pattern
	 *                will be deleted.
	 */
	private static void deleteTree(File path, String... patterns) {
		if (path.exists() == false)
			return;

		for (File f: path.listFiles()) {
			boolean doDelete = false;

			if (f.isDirectory())
				deleteTree(f, patterns);

			for(String pattern: patterns) {
				if (Pattern.matches(pattern, f.getName())) {
					doDelete=true;
					break;
				}
			}

			if(doDelete || patterns.length==0) f.delete();
		}
	}

	/**
	 * Reads information on the previous build, and stores information for the
	 * current build.
	 *
	 * @return true if the previous build used the same configuration
	 */
	private int _sameBuild;
	private boolean getBuildInfo() throws IOException {
		// TODO record origin module of all source files so that subsequent build can detect
		//   when an identical Java class is used from a different package

		// 0: not checked
		// 1: same build
		// 2: other build
		if(_sameBuild == 0) {
			_sameBuild = 2;

			File buildinfo = new File(opts.getClassDir()+"/"+".buildinfo");
			String rcfile = opts.getWorldDefinition().rcFile();
			if(buildinfo.exists()) {
				BufferedReader bird = new BufferedReader(new FileReader(buildinfo));
				String lastrc = bird.readLine();
				bird.close();

				if(lastrc != null && lastrc.trim().equals(rcfile)) {
					_sameBuild = 1;
				}
			}

			Writer biwr = new FileWriter(buildinfo);
			biwr.write(rcfile);
			biwr.close();
		}

		return (_sameBuild==1);
	}

	/**
	 *  compile source files into classes
	 */
	public void preCompileModules() throws IOException, Exception {
		// prepare output directories
		// target directory for the precompiled class files
		String classpath = opts.getClassDir();
		File classpathfile = new File(classpath);
		classpathfile.mkdirs();

		// read / update build info
		getBuildInfo();

		/// gather all Java modules required by the application
		String[] modules  = opts.getAllModulePaths(); // from the RC file
		modules = getRequiredModules(modules); // depended modules (from META files)

		// call source generators of the needed modules
		generateModuleClasses(modules);

		String cp = opts.getJDKClassPath();
		if (cp!=null && !cp.equals("")) { classpath= classpath+":"+cp; }

		opts.phase("javac precompilation (.java -> .class)");

		// clear output directories
		deleteTree(new File(opts.getOutputPath()));

		// TODO record the RC file used for the last translation, so that the class
		// file generators don't check the wrong modification time!

		// for all regular modules, collect the source files that need to
		// be compiled
		// maps module name => source files for the module that need compilation
		Map<String,? extends Collection<String>> files = findFilesToPrecompile(modules);

		// we compile each module separately, so that precompiled archives can replace
		// files from source modules according to the module priority ordering
		for(String module: modules) {
			String module_path = opts.getAppSourcePath()+module;
			File moduleDir = new File(module_path);

			if(moduleDir.isDirectory()) {
				Collection<String> worklist = files.get(module);
				opts.verbose("pre-compile module " + module + ": " +  worklist.size() + " files");

				boolean coreclasses = module.equals("core");

				if(compile_set(classpath,modules,worklist,coreclasses) != 0)
					throw new CompileException("javac failed");
			} else if (moduleDir.isFile()) {
				if (module.endsWith(".zip") || module.endsWith(".jar")) {
					addArchive(classpath, moduleDir);
				} else {
					opts.warn("Module "+module+" has an unsupported file format!");
					throw new CompileException(moduleDir.toString()+" has an unsupported file format!");
				}
			} else {
				opts.warn("Module "+module+" not found!");
				opts.warn("Please check modules and source path statement in the config file.");
				throw new IOException(moduleDir.toString()+" does not not exist!");
			}
		}
	}

	/**
	 * Parse META files and extend the module list.
	 *
	 * An output list is created that contains depended-on modules first and
	 * eventually the initial set in the specified order, unless already
	 * contained in the dependencies.
	 *
	 * @param modules initial set of needed modules (from KESORC)
	 * @return ordered module list
	 */
	public String[] getRequiredModules(String[] modules) throws IOException {
		HashSet<String> contains = new HashSet<String>();
		List<String> required = new ArrayList<String>();

		// process dependencies
		for(String module: modules) {
			getMetaModulesFor(module, contains, required);
		}

		// add explicitly needed modules to the end of the list, if not
		// already contained in the dependencies
		for(String module: modules) {
			if (contains.add(module)) {
				required.add(module);
			}
		}

		return required.toArray(new String[required.size()]);
	}

	/**
	 * Recursively retrieves the module dependencies of a module listed in its
	 * META file.
	 *
	 * @param mname    module whose dependencies are to be determined
	 * @param contains HashSet to track already included modules, intially empty
	 * @param required ordered output list of the required modules
	 */
	private void getMetaModulesFor(String mname, HashSet<String> contains,
			Collection<String> required) throws IOException {
		String module = opts.getAppSourcePath() +mname;
		File modul_dir = new File(module);

		if (modul_dir.isFile()) return;

		if (!modul_dir.exists()) {
			Debug.out.println("Module "+mname+" not found!");
			Debug.out.println("Please check modules and source path statement in the kcl file.");
			return;
		}

		File meta_file = new File(module+"/META");
		if (!meta_file.exists()) {
			opts.warn("Module "+mname+" is not valid!");
			opts.warn("META file is missing!");
			return;
		}

		BufferedReader meta_reader = new BufferedReader(new FileReader(meta_file));
		String line;
		while (null!=(line=meta_reader.readLine())) {
			line = line.trim();
			if (line.startsWith("#")) continue;
			String[] val = line.split("=");
			if (val.length<2) continue;
			val[0] = val[0].trim().toUpperCase();
			val[1] = val[1].trim();

			// recursively add depended modules
			if (val[0].equals("REQUIRE")) {
				if (contains.add(val[1])) {
					// recursive deps first
					getMetaModulesFor(val[1], contains, required);
					// then the module itself
					required.add(val[1]);
				}
			}
		}
		meta_reader.close();
	}

	/**
	 *  Executes the generator templates for the given modules.
	 *
	 *  @param modules Array of module names whose generator templates are executed.
	 */
	private void generateModuleClasses(String[] modules) throws IOException, Exception {
		for(String mname : modules) {
			String module = opts.getAppSourcePath() +mname;
			File modul_dir = new File(module);

			if (modul_dir.isFile()) return;

			if (!modul_dir.exists()) {
				Debug.out.println("Module "+mname+" not found!");
				Debug.out.println("Please check modules and source path statement in the kcl file.");
				return;
			}

			// read META file, process GENERATE statements
			File meta_file = new File(module+"/META");
			if (!meta_file.exists()) {
				opts.warn("Module "+mname+" is not valid!");
				opts.warn("META file is missing!");
				return;
			}

			BufferedReader meta_reader = new BufferedReader(new FileReader(meta_file));
			String line;
			while (null!=(line=meta_reader.readLine())) {
				line = line.trim();
				if (line.startsWith("#")) continue;
				String[] val = line.split("=");
				if (val.length<2) continue;
				val[0] = val[0].trim().toUpperCase();
				val[1] = val[1].trim();
				if (val[0].equals("GENERATE")) {
					GenerateKesoClasses.generateModuleClass(opts.getWorldDefinition(), opts.getKesoSourcePath(), val[1], "templates/", "libs/", mname + "/", getBuildInfo());
				}
			}
			meta_reader.close();
		}
	}

	private int addArchive(String classpath, File module) throws IOException {
		ZipFile archive = new ZipFile(module);
		byte buffer[] = new byte[4096];

		Enumeration<? extends ZipEntry> ee = archive.entries();
		while (ee.hasMoreElements()) {
			ZipEntry entry = ee.nextElement();

			String dst = classpath+"/"+entry.getName();
			int pos = dst.lastIndexOf(File.separatorChar);
			File dfile = new File(dst);
			File ddir = new File(dst.substring(0,pos));
			ddir.mkdirs();
			dfile.createNewFile();

			InputStream in = archive.getInputStream(entry);
			FileOutputStream out = new FileOutputStream(dst);

			int len;
			while ((len=in.read(buffer))>-1) { out.write(buffer,0,len); }

			out.close();
		}

		archive.close();

		return 0;
	}

	private int compile_set(String classpath, String[] modules,
			Collection<String> work_todo, boolean coreclasses) throws IOException {

		if(work_todo.size() <= 0)
			return 0;

		Collection<String> args = new ArrayList<String>();
		//args.addElement("-verbose");
		args.add("-O");
		if (!opts.hasOption("no_javac_debug")) args.add("-g");
		//args.addElement("-target 1.4");
		//args.addElement("-nowarn");

		args.add("-d");
		args.add(classpath);

		if(!coreclasses && !opts.hasOption("jdk_bootclasses")) {
			args.add("-bootclasspath");
			args.add(classpath);
		}

		args.add("-classpath");
		args.add(classpath);

		args.add("-sourcepath");
		StringBuilder sp = new StringBuilder(opts.getAppSourcePath()+"/"+modules[0]);
		for(int i=1; i<modules.length;i++) {
			sp.append(":");
			sp.append(opts.getAppSourcePath());
			sp.append("/");
			sp.append(modules[i]);
		}
		args.add(sp.toString());

		args.addAll(work_todo);

		return javac(args);
	}

	/**
	 * This method copies the java source files into the output directory.
	 *
	 * @param dest Pathname of the destination directory
	 * @param modules Array of module names that are copied to the destination directory
	 */
	public void copySourceFiles(String dest, String[] modules) throws IOException {
		for(String module: modules) {

			Collection<String> work_todo = new ArrayList<String>();
			String module_path = opts.getAppSourcePath() +module;
			opts.verbose("copy "+module);

			File module_dir = new File(module_path);
			collectWorkToDo(module_dir, work_todo);

			for (String current : work_todo) {
				copy(current, dest + "/"+current.substring(module_path.length()+1));
			}
		}
	}

	public void computeServices() {
		try {
			repository = new ClassStore(opts, opts.getSourcePaths());
			opts.registerWeavelets(repository);
			ServiceManager.instance.computeServiceMethods(repository);
			repository.destroy();
			repository = null;
		} catch (Exception ex) {
			ex.printStackTrace(Debug.out);
			Debug.out.println(ex);
			System.exit(-1);
		}
	}

	/**
	 * loadModules is loading all needed modules
	 */
	public void loadModules() {
		try {
			if (opts.pruneClassTree()) {
				ClassStore disk = new ClassStore(opts,opts.getSourcePaths());
				this.repository = disk.getPrunedClassStore();
				disk.destroy();
			} else {
				this.repository = new ClassStore(opts,opts.getSourcePaths());
				opts.registerWeavelets(repository);
			}
		} catch (Exception ex) {
			ex.printStackTrace(Debug.out);
			Debug.out.println(ex);
			System.exit(-1);
		}
	}

	/**
	 * loadClasses simulates the class loading ahead of time.
	 *
	 * The repository only includes the required classes and
	 * methods so far! (because this method is called after
	 * loadModules)
	 */
	public void loadClasses() {
		try {
			opts.phase("load classes");
			// This call ensures that all domains
			// and heaps are initialized.
			opts.getDomains();

			IMClass imclasses[] = repository.getAllClasses();
			for (int i=0;i<imclasses.length;i++) imclasses[i].load();
		} catch (Exception ex) {
			Debug.out.println(ex.toString());
			ex.printStackTrace();
			System.exit(-1);
		}
	}

	private void writeRTCheckStat_line(PrintStream out,
			int nce, int ncaotok, int ncaotfail,
			int bcf, int bcki, int bcc, int bco,
			int cce, int ccf, int ccif
			) throws IOException {
		// null ref checks
		out.print(" ");
		out.print(nce);
		out.print(" ");
		out.print(ncaotok);
		out.print(" ");
		out.print(ncaotfail);

		// array bound checks
		out.print(" ");
		out.print(bcf);
		out.print(" ");
		out.print(bcki);
		out.print(" ");
		out.print(bcc);
		out.print(" ");
		out.print(bco);

		// cast checks
		out.print(" ");
		out.print(cce);
		out.print(" ");
		out.print(ccf);
		out.print(" ");
		out.print(ccif);

		out.println();
	}

	public void writeRTCheckStats() throws CompileException {
		if (!opts.hasOption("rtcheckstat")) return;
		try {
			String filename = opts.getOutputPath()+"/rtCheckStat.lst";
			PrintStream out = new PrintStream(new FileOutputStream(filename));
			out.println("symbol NCE/NullChecksEmitted NCAS/NullChecksAOTSuccess NCAF/NullChecksAOTFail BCF/BoundsCheckFull BCKI/BoundsCheckKnownIndex BCC/BoundsCheckConstant BCO/BoundsCheckOmitted CCE/CastCheckEmitted CCF/CastCheckFast CCIF/CastCheckInterface");

			int nce=0,ncaotok=0,ncaotfail=0;
			int bcf=0,bcki=0,bcc=0,bco=0;
			int cce=0,ccf=0,ccif=0;
			for(IMClass imclass: repository.getAllClasses()) {
				IMMethod[] methods = imclass.getMethods();
				if(methods==null) continue;
				for(IMMethod mth: methods) {
					out.print(mth.getAlias());

					writeRTCheckStat_line(out,
							mth.nullCheckEmitted,mth.nullCheckAOTSuccess,mth.nullCheckAOTFail,
							mth.boundsCheckFull,mth.boundsCheckKnownIndex,mth.boundsCheckConstant,mth.boundsCheckOmitted,
							mth.castCheckEmitted,mth.castCheckFast,mth.castCheckIFace);

					// null ref checks
					nce += mth.nullCheckEmitted;
					ncaotok += mth.nullCheckAOTSuccess;
					ncaotfail += mth.nullCheckAOTFail;

					// array bound checks
					bcf += mth.boundsCheckFull;
					bcki += mth.boundsCheckKnownIndex;
					bcc += mth.boundsCheckConstant;
					bco += mth.boundsCheckOmitted;

					// cast checks
					cce += mth.castCheckEmitted;
					ccf += mth.castCheckFast;
					ccif += mth.castCheckIFace;
				}
			}

			out.print("TOTAL");
			writeRTCheckStat_line(out,nce,ncaotok,ncaotfail,bcf,bcki,bcc,bco,cce,ccf,ccif);
			out.flush();
			out.close();
		} catch (IOException ex) {
			Debug.out.println(ex.toString());
			ex.printStackTrace();
		}
	}

	public void runPasses() throws CompileException {

		// Run transformation and analysis passes
		PassManager trafoPm
		= new PassManager(opts, "Transformation pass manager");
		trafoPm.addPass(new CallGraphAnalysis());
		trafoPm.addPass(new CFGSimplification());
		trafoPm.addPass(new CheckInsertion());
		trafoPm.addPass(new ClassTypeInfoComputation());
		trafoPm.addPass(new ClinitAnalysis());
		trafoPm.addPass(new ConstantFolding());
		trafoPm.addPass(new ConstantParameterPropagation());
		trafoPm.addPass(new DomTreeComputation());
		trafoPm.addPass(new EscapeAnalysis());
		trafoPm.addPass(new ExtendedEscapeAnalysis());
		trafoPm.addPass(new FieldAccessReset());
		trafoPm.addPass(new FlowAnalysis());
		trafoPm.addPass(new FlowBasedOptimization());
		trafoPm.addPass(new InterPassByteCodeDump());
		trafoPm.addPass(new LivenessAnalysis());
		trafoPm.addPass(new MethodInlining());
		trafoPm.addPass(new SimpleReachabilityAnalysis());
		trafoPm.addPass(new ReachabilityDump());
		trafoPm.addPass(new SSAConstruction());
		trafoPm.addPass(new SSADeconstruction());
		trafoPm.addPass(new StackAllocation());
		trafoPm.addPass(new UnusedMethodRemoval());
		trafoPm.addPass(new VTableComputation());
		trafoPm.addPass(new VTableReset());
		trafoPm.runOnRepository(repository);

		// Run dump passes
		PassManager dumpPm
		= new PassManager(opts, "Dump pass manager", trafoPm);
		dumpPm.addPass(new ByteCodeDump());
		dumpPm.addPass(new CallGraphWriter());
		dumpPm.addPass(new ClassListDump());
		dumpPm.addPass(new ClassStatisticsDump());
		dumpPm.addPass(new MethodStatisticsDump());
		dumpPm.addPass(new CFGWriter());
		dumpPm.addPass(new ClassTreeWriter());
		dumpPm.addPass(new DomTreeWriter());
		dumpPm.addPass(new ReachabilityGraphWriter());
		dumpPm.addPass(new StatisticsDump());
		dumpPm.addPass(new UMLDiagramWriter());
		dumpPm.runOnRepository(repository);

		// Run translation passes
		PassManager translatePm
		= new PassManager(opts, "Translation pass manager", dumpPm);
		translatePm.addPass(new Translation());
		translatePm.runOnRepository(repository);
	}
}
