/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import java.util.*;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import keso.classfile.datatypes.BCBasicDatatype;
import keso.compiler.kni.JoinPointChecker;

import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.*;
import keso.compiler.config.parser.ConfigReader;
import keso.compiler.imcode.IMDomain;
import keso.compiler.imcode.IMGlobalHeap;
import keso.util.Debug;
import keso.util.DecoratedNames;

final public class BuilderOptions {

	private final static class SysObjComparator<D extends SystemObject<?>>
	implements Comparator<D> {

		@Override
		public int compare(D o1, D o2) {
			return o1.getIdentifier().compareTo(o2.getIdentifier());
		}

	}

	public static final int DEFAULT_TARGET = -1;
	public static final int GENERIC_TARGET = 0;
	public static final int TRICORE_PROOSEK_TARGET = 1;
	public static final int AVR_JOSEK_TARGET = 2;
	public static final int CIAO_TARGET = 3;
	public static final int TRICORE_RTAOSEK_TARGET = 4;
	public static final int X86LINUX_JOSEK_TARGET = 5;
	public static final int POSIX_TRAMPOLINE_TARGET = 6;
	public static final int AVR_TRAMPOLINE_TARGET = 7;

	private static final String[] target_names = {
		"GENERIC",
		"TRICORE_PROOSEK",
		"AVR_JOSEK",
		"TRICORE_CIAO",
		"TRICORE_RTAOSEK",
		"X86LINUX_JOSEK",
		"POSIX_TRAMPOLINE",
		"AVR_TRAMPOLINE",
	};

	private static final String[] type_size_index = {
		"void",
		"jbyte", "jchar",
		"jshort",
		"jint", "jfloat",
		"jlong", "jdouble",
		"object_t*",
	};

	private static final int[] type_size_value = new int [] {
		0,
		1, 1,
		2,
		4, 4,
		8, 8,
		-1,
	};

	public static final String OBJ_T     = "object_t*";
	public static final String VOID_T    = "void";
	public static final String BOOLEAN_T = "jbyte";
	public static final String BYTE_T    = "jbyte";
	public static final String CHAR_T    = "jchar";
	public static String SHORT_T         = "jshort";
	public static String INT_T           = "jint";
	public static String LONG_T          = "jlong";
	public static String FLOAT_T         = "jfloat";
	public static String DOUBLE_T        = "jdouble";

	private String sourcepath   = ".";
	private String outputpath   = "./tmp";
	private String[] modulepaths = null;
	private String kesosrcpath  = ".";
	private String classtree    = null;
	private String jdk_classpath = "";
	private String override_cc  = null;
	private String[] allModules;

	private WorldDefinition  worldDef = null;
	private SystemDefinition systemDef = null;
	private IMGlobalHeap	 global_heap = null;
	private List<AlarmDefinition> alarms;
	private List<ISRDefinition> isrs;
	private List<JavaTask> java_tasks;

	private int target_arch_opt = GENERIC_TARGET;
	private int target_arch = GENERIC_TARGET;

	private boolean quiet	  = false;
	private boolean verbose   = false;
	private boolean very_verbose  = false;
	private boolean debug	  = false;
	private boolean dump_bc	  = false;
	private boolean prune_CT  = true;
	private boolean buildimage= true;
	private boolean flMakefile= true;
	private boolean dbg_syms  = true;
	private boolean create_CFG = false;
	private boolean fast_method_alias = false;
	private boolean precompile = false;
	private boolean translate = false;

	/*
	 * default option to tune the method in-line operation
	 */
	private boolean inline_methods      = true;
	private boolean inline_rvalue       = true;
	private int	inline_costs 	    = 80;
	private int	inline_costs_single = 1000;

	private final HashSet<String> extras = new HashSet<String>();
	private HashSet<String> verbose_classes = null;
	private HashSet<String> force_inline = null;
	private String current_class = "xxx";
	private String current_phase = null;
	private int current_class_id = 0;

	private final ArrayList<String> m_linkerscript_paths;

	private String bootmodules = null;

	private JoinPointChecker joinPoints;

	public BuilderOptions(String argv[]) throws CompileException {
		m_linkerscript_paths = new ArrayList<String>();

		BuilderOptions.gopts = this;
		parseArguments(argv);
		if (this.hasOption("int8")) {
			SHORT_T   = "jbyte";
			INT_T   	= "jbyte";
			LONG_T  	= "jshort";
		} else if (this.hasOption("int16") || is8BitController()) {
			INT_T   	= "jshort";
			LONG_T  	= "jint";
		}
		if (useFixPointMath()) {
			FLOAT_T 	= "jshort";
			if (this.hasOption("no_double")) {
				DOUBLE_T	= "jshort";
			} else {
				DOUBLE_T  	= "jint";
			}
		} else {
			if (this.hasOption("no_double")) {
				DOUBLE_T  	= "jfloat";
			}
		}
		joinPoints = new JoinPointChecker(this, null);

		setExtraOption("gcc_no_const");
		setExtraOption("gcc_no_pure");
	}

	/**
	 * This Method is the right place to register weavelets. It is called
	 * while building the class store.
	 */
	public void registerWeavelets(ClassStore repository) {
		joinPoints.reset();
		joinPoints.registerWeavelet(new keso.compiler.kni.Alarm(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.AlarmService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.CMath(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.CiAOIPStack(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Config(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.DebugOutWeavelet(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.DoubleWeavelet(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.EventService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.InterruptService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Memory(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.MemoryService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.MemoryTypes(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.MessagePorts(this, repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.OSService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Object(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.PortalServiceWeavelet(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Resource(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.ResourceService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Runtime(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.SharedMemory(this, repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.SystemTimeWeavelet(this, repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Task(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.TaskService(this,repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.TimeStampWeavelet(this, repository));
		joinPoints.registerWeavelet(new keso.compiler.kni.Tricore(this,repository));
	}

	/**
	 * Returns the directory name where we can find the source classes.
	 */
	public String getClassDir() {
		return "classes";
	}

	/**
	 * Returns the directory name where we can find the JDK classes.
	 */
	public String getJDKClassPath() {
		return jdk_classpath;
	}

	/**
	 * Returns the directory name where we can find the Java source files.
	 */
	public String[] getModulePaths() {
		return modulepaths;
	}


	/**
	 * Returns the directory name where we can find the Java source files.
	 */
	public String[] getAllModulePaths() {
		return allModules;
	}

	/**
	 * Returns true if we use a linked list of local object references on
	 * the method stack.
	 */
	public boolean hasLinkedListOfLocalReferences() {
		return (global_heap.needStackScan() && hasBlockingTasks());
	}

	public boolean needFinalizers() {
		return global_heap.needGCInfo();
	}

	private int fp_frac_float = 8;
	private int fp_frac_double = 12;

	public boolean useFixPointMath() throws CompileException {

		if (systemDef==null) return false;

		ComplexAttribute<?> fp
		= (ComplexAttribute<?>) systemDef.getAttribute("UseFixPointMath");
		if (fp==null || !fp.enabled()) return false;

		IntegerAttr<?> ff = (IntegerAttr<?>) fp.getAttribute("FloatFraction");
		if (ff!=null) fp_frac_float = ff.valueInt();

		IntegerAttr<?> fd = (IntegerAttr<?>) fp.getAttribute("DoubleFraction");
		if (fd!=null) fp_frac_double = fd.valueInt();

		return true;
	}

	public int fractionBitsFloat() {
		return fp_frac_float;
	}

	public int fractionBitsDouble() {
		return fp_frac_double;
	}

	/**
	 * Returns true if we want to prune the class tree.
	 */
	public boolean pruneClassTree() {
		return prune_CT;
	}

	final public boolean doBuildImage() {
		return buildimage;
	}

	final public boolean doFirstLevelMakefile() {
		return flMakefile;
	}

	final public void doFirstLevelMakefile(boolean setting) {
		flMakefile = setting;
	}

	private boolean verbose_class() {
		if (verbose_classes==null) return false;
		if (verbose_classes.contains(current_class)) return true;
		if (verbose_classes.contains("c"+current_class_id)) return true;
		if (current_phase==null) return false;
		if (verbose_classes.contains("c"+current_class_id+current_phase))
			return true;
		return false;
	}

	public boolean vvverbose() {
		if (verbose_class()) return true;
		return false;
	}

	public void vvverbose(Object msg) {
		if (verbose_class()) Debug.message(msg.toString());
	}

	public void vverbose(Object msg) {
		if (very_verbose || verbose_class()) Debug.message(msg.toString());
	}

	public void verbose_hr() {
		verbose("===============================================");
	}

	public void verbose(Object msg) {
		if (verbose || verbose_class()) Debug.message(msg.toString());
	}

	public void xverbose(Object msg) {
		Debug.message(msg.toString());
	}

	public void phase(Object msg) {
		if (quiet) return;
		Debug.message("=========== "+msg.toString()+" ===========");
	}

	public void info(Object msg) {
		Debug.message("info: "+msg.toString());
	}

	/**
	 * Shows a compiler warning, only in very verbose mode.
	 */
	public void vvwarn(Object msg) {
		if (very_verbose) warn(msg);
	}

	/**
	 * Shows a compiler warning, only in verbose mode.
	 */
	public void vwarn(Object msg) {
		if (verbose) warn(msg);
	}

	/**
	 * Shows a compiler warning.
	 */
	public void warn(Object msg) {
		Debug.message("warning: "+msg.toString());
	}

	public void todo(Object msg) {
		if (this.hasOption("todo")) {
			Debug.message("TODO: "+msg.toString());
			return;
		}
		vwarn("TODO: "+msg.toString());
	}

	public void stackTrace() {
		try {
			throw new CompileException();
		} catch (CompileException ex) {
			ex.printStackTrace();
		}
	}

	public static void fatal(Object msg) throws CompileException {
		throw new CompileException(msg.toString());
	}

	/**
	 * In normal a cases critical() throws an exception, but
	 * it is possible to override this behaviour with -X:ignore_critical.
	 */
	public void critical(Object msg) throws CompileException {
		if (this.hasOption("ignore_critical") || !prune_CT) {
			warn(msg);
		} else {
			throw new CompileException(msg.toString());
		}
	}

	public void setCurrentClass(String name, int class_id) {
		current_class = name;
		current_class_id = class_id;
		current_phase = null;
	}

	public void setCurrentClass(String name, int class_id, String phase) {
		current_class = name;
		current_class_id = class_id;
		current_phase = phase;
	}

	public boolean fastMethodAlias() { return fast_method_alias; }

	public boolean hasVerbose() { return verbose; }

	public boolean hasDebugCore() { return debug; }

	public boolean hasDbgSymbols() { return dbg_syms; }

	public boolean hasDumpBC() { return dump_bc; }

	public boolean getPreCompile() {
		return this.precompile || !this.translate;
	}

	public boolean getTranslate() {
		return this.translate || !this.precompile;
	}

	public boolean doInlineCalls() { return inline_methods; }

	public boolean forceInline(String name) {
		if (force_inline==null) return false;
		if (force_inline.contains(name)) return true;
		String part = name.substring(0, name.indexOf("("));
		if (force_inline.contains(part)) return true;
		return false;
	}

	public int maxInlineCosts() { return inline_costs; }

	public int maxInlineCostsSingle() {
		return inline_costs_single;
	}

	public boolean doInlineRValues() { return inline_rvalue; }

	public boolean useGCCExtensions() {
		if (target_arch==TRICORE_RTAOSEK_TARGET) return false;
		if (this.hasOption("no_gcc_extensions")) return false;
		return true;
	}

	public boolean createCFG() { return create_CFG; }

	public boolean createDomTree() { return this.hasOption("dump_dominator_tree"); }

	public boolean createCallGraph() {
		return hasOption("callgraph");
	}

	public boolean createReachabilityGraph() {
		return hasOption("reachabilitygraph");
	}

	public boolean createUMLDiagram() { return this.hasOption("uml"); }

	public String declareConst() {
		return " KESO_CONST ";
	}

	public boolean hasOption(String name) {
		if (name.equals("_X_dbg_astack") && !hasOption("ssa_astack")) return false;
		if (name.startsWith("ssa_") && !hasOption("ssa")) return false;
		if (extras.contains(name) == false) return false;
		return true;
	}

	private void setTarget(int target) {
		if (target==DEFAULT_TARGET) {
			target_arch = target_arch_opt;
		} else {
			target_arch = target;
		}
	}

	public static int getTargetByName(String target_name) {
		String tname = target_name.toUpperCase();

		for (int i=0;i<target_names.length;i++) {
			if (tname.equals(target_names[i])) return i;
		}

		//return DEFAULT_TARGET;
		throw new RuntimeException("Unknown target \""+target_name+"\" in configuration file!\n");
	}

	// FIXME we now have non-Java domains; this function will return true
	// if there is a single java domain but multiple other domains in the
	// system. This may break some locations...
	public boolean isSingleDomainSystem() throws CompileException {
		return systemDef.getJavaDomains().size()==1;
	}

	public boolean useExceptionStrings() {
		switch (target_arch) {
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return false;
		}
		return true;
	}

	/**
	 *
	 * 8 bit micro controller have more restrictions.
	 *    - max. object size is 256 byte.
	 */
	public boolean is8BitController() {
		switch (target_arch) {
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return true;
		}
		return false;
	}

	/**
	 * True if writing a memory word - in particular an object reference - is
	 * non atomic on the target architecture.
	 *
	 * The memory access will then be guarded by disabling the interrupts. This
	 * is only needed for globally visible memory accesses such as static field
	 * accesses or object field accesses.
	 *
	 * We do not need to guard the access if
	 *
	 *  - the domain has a single task or ISR only and does not export services
	 *  - there are no async task activations (by ISR/alarm) in the system
	 */
	public boolean needGuardedMemoryAccess() {
		if (is8BitController()) return true;
		return false;
	}

	/**
	 * Returns whether there are blocking tasks in the system.
	 *
	 * We are only interested in blocking Java tasks, since our interest lies
	 * in the need for stack scans of the garbage collector. Blocking native
	 * tasks are not an issue.
	 */
	public boolean hasBlockingTasks() {
		for(TaskDefinition td: getSysJavaTasks())
			if(td.getEvents().size() > 0)
				return true;
		return false;
	}

	/**
	 * Returns whether local portal calls are made in the current system.
	 */
	public boolean hasLocalPortalUse() {
		return (0 < ServiceManager.instance.getImportsOnCurrentSystem().size());
	}

	public boolean useLLRefMarker2() {
		if (this.hasOption("_X_llref_mark2")) return true;
		if (is8BitController()) return true;
		return false;
	}

	public boolean hasRDTSC() {
		if (this.hasOption("no_rdtsc")) return false;
		return false;
	}

	public static int getTypeSize(String cType) {
		int i=0;

		for (;i<type_size_index.length;i++) {
			if (type_size_index[i].equals(cType)) break;
		}

		if (i==type_size_index.length) throw new RuntimeException("unknown type");

		int size = type_size_value[i];

		if (size==-1) return getOpts().targetAddrSize();

		return size;
	}

	public String targetAddrType() {
		return "void *";
	}

	public int targetAddrSize() {
		switch (target_arch) {
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return 2;
		}
		return 4;
	}

	public int targetAddrSizeShift() {
		switch (target_arch) {
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return 1; /* 2 Byte */
		}
		return 2; /* 4 Byte */
	}

	public String getCC() {
		if (override_cc!=null) return override_cc;
		switch (target_arch) {
			case TRICORE_PROOSEK_TARGET:
			case CIAO_TARGET:
				return "tricore-gcc";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-gcc";
		}
		return "gcc";
	}

	public String getObjVarName() {
		return "OBJS_KESO";
	}

	public String getCFlagsVarName() {
		switch (target_arch) {
			case TRICORE_PROOSEK_TARGET:
				return "CC_OPT_USER";
		}
		return "CFLAGS";
	}

	public String getStrip() {
		switch (target_arch) {
			case CIAO_TARGET:
			case TRICORE_PROOSEK_TARGET:
				return "tricore-strip";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-strip";
		}
		return "strip";
	}

	public String getObjSize() {
		switch (target_arch) {
			case CIAO_TARGET:
			case TRICORE_PROOSEK_TARGET:
				return "tricore-size";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-size";
		}
		return "size";
	}

	public String getObjCopy() {
		switch (target_arch) {
			case CIAO_TARGET:
			case TRICORE_PROOSEK_TARGET:
				return "tricore-objcopy";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-objcopy";
		}
		return "objcopy";
	}

	public String getObjDump() {
		switch (target_arch) {
			case CIAO_TARGET:
			case TRICORE_PROOSEK_TARGET:
				return "tricore-objdump";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-objdump";
		}
		return "objdump";
	}

	public String getAR() {
		switch (target_arch) {
			case CIAO_TARGET:
			case TRICORE_PROOSEK_TARGET:
				return "tricore-ar";
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "avr-ar";
		}
		return "ar";
	}

	public String getDebugFormat() {
		switch (target_arch) {
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
				return "stabs";
			default:
				return "";
		}
	}

	public String getProcessorType() {
		if (systemDef==null || systemDef.getProcessorType()==null) {
			warn("missing processor type! use default.\n");
			if (isAVR()) return "atmega8535";
			return "tc1796";
		} else {
			return systemDef.getProcessorType();
		}
	}

	public String getASFlags() {
		return "-x assembler-with-cpp -c ";
	}

	public String getCFlags() {
		String flags="";

		flags = "-DTARGET_"+target_names[target_arch];

		if (target_arch != CIAO_TARGET) {

			// We have to use gnu99 instead of c99 because C99 does not define
			// clock_gettime() et al.
			flags += " -std=gnu99 -g" + getDebugFormat() + " ";

			// use dynamic linking on hosted platforms (e.g., to support valgrind usage)
			if (!isMacOS() && target_arch!=POSIX_TRAMPOLINE_TARGET && target_arch!=X86LINUX_JOSEK_TARGET) {
				flags += "-static ";
			}

			flags += "-Os -freorder-blocks ";

			if (this.hasOption("gcc-short-double")) {
				flags += "-fshort-double ";
			}

			if (this.hasOption("gcc-fnct-section")) {
				flags+="-ffunction-sections -fdata-sections ";
			}

			if (this.hasOption("gcc-no-stdlibs")) {
				flags += "-nodefaultlibs -nostdlib ";
			}
		}

		switch (target_arch) {
		case CIAO_TARGET:
			return flags;
		case TRICORE_PROOSEK_TARGET:
			return flags+"-nostartfiles -Wall -mcpu=" + getProcessorType();

		case AVR_JOSEK_TARGET:
		case AVR_TRAMPOLINE_TARGET:
			flags += " -DAVR -I/usr/avr/include -ffreestanding -fshort-enums -fpack-struct -mmcu=" + getProcessorType();
			break;
		case X86LINUX_JOSEK_TARGET:
			if (!this.hasOption("gcc-no-whole-program"))
				flags += " -combine -fwhole-program ";
			flags += " -m32 ";
			break;
		case POSIX_TRAMPOLINE_TARGET:
			if (target_arch != AVR_JOSEK_TARGET && target_arch != AVR_TRAMPOLINE_TARGET)
				flags += " -D_XOPEN_SOURCE=500 ";
			return flags + " -Wall -funsigned-bitfields -fomit-frame-pointer -m32";
		}
		return flags + " -Wall ";
	}

	public String getLinkerOptions() {
		String linker = "";
		if (hasOption("gcc-fnct-section"))
			linker += "--gc-sections ";
		if (linker.length() == 0)
			return null;
		return linker;
	}

	public String getLDFlags() {
		String ldFlags = "";

		if (hasOption("use_mathlib"))
			ldFlags += "-lm ";

		switch (target_arch) {
		case TRICORE_PROOSEK_TARGET:
			ldFlags += "-mcpu=" + getProcessorType() + " ";
			break;
		case AVR_JOSEK_TARGET:
		case AVR_TRAMPOLINE_TARGET:
			ldFlags += "-mmcu=" + getProcessorType() + " ";
			break;
		case X86LINUX_JOSEK_TARGET:
		case POSIX_TRAMPOLINE_TARGET:
			ldFlags += "-m32";
			if (isMacOS())
				ldFlags += " -read_only_relocs suppress";
		}
		return ldFlags;
	}

	public String getCoreLibName() {
		switch (target_arch) {
			case CIAO_TARGET:
			case AVR_JOSEK_TARGET:
			case AVR_TRAMPOLINE_TARGET:
			case POSIX_TRAMPOLINE_TARGET:
				/* core lib is not used */
				return null;
			case X86LINUX_JOSEK_TARGET:
				if (isMacOS())
					return null;
				return "rt"; // Quick fix for undefined ref. to clock_gettime()
			case TRICORE_PROOSEK_TARGET:
				return "KESO-tricore-osek";
		}
		return "FIXME-UNKNOWN-LIB";
	}

	public String getCoreLib() {
		switch (target_arch) {
			case TRICORE_PROOSEK_TARGET:
				return "libKESO-tricore-osek.a";
		}
		return null;
	}

	public static boolean smartFileUpdate(File file, String new_value) {
		try {
			if (file==null || !file.canRead()) return true;
			int size = (int)file.length();
			char[] buf = new char[size];
			FileReader old_file = new FileReader(file);
			old_file.read(buf);
			String old_value = new String(buf);
			if (old_value.equals(new_value)) return false;
		} catch (IOException ex) { }
		return true;
	}

	public int getTarget() {
		return target_arch;
	}

	public boolean isTarget(int target) {
		return (target_arch==target);
	}

	public boolean isAVR() {
		return target_arch==AVR_JOSEK_TARGET || target_arch==AVR_TRAMPOLINE_TARGET;
	}

	public boolean isTriCore() {
		return (target_arch==TRICORE_PROOSEK_TARGET || target_arch==CIAO_TARGET);
	}

	public String getOutputClassTree() {
		return classtree;
	}

	public String[] getSourcePaths() {
		return split(sourcepath,':');
	}

	public String getFirstLevelOutputPath() {
		return outputpath;
	}

	public String getOutputPath() {
		return outputpath + "/" + getSystemName();
	}

	public String getSystemName() {
		if (systemDef==null) return "unknown";
		return systemDef.getIdentifier();
	}

	public int getCurrentSystemID() {
		if (systemDef == null) {
			return -1;
		}
		return systemDef.getSystemID();
	}

	public String getAppSourcePath() {
		return kesosrcpath + "/libs/";
	}

	public String getCorePath() {
		return kesosrcpath + "/kore/";
	}

	public String getKesoSourcePath() { return kesosrcpath; }

	public JoinPointChecker getJoinPoints() {
		return joinPoints;
	}

	public boolean hasLittleEndianMemory() {
		return hasLittleEndianMemory(target_arch);
	}

	public static boolean hasLittleEndianMemory(int target_arch) {
		switch (target_arch) {
			case TRICORE_PROOSEK_TARGET:
			case CIAO_TARGET:
			default:
				return true;
		}
	}

	private void usage() {
		System.err.println("\nusage:\tjino [-[no]build] [-v[v]] [-v:class] [-cfg] [-ct/classtree <file>] [-OF]");
		System.err.println("\t[-[no]prune] [-[no]inline] [-inline_costs:value:value] [-X:option[:option]]");
		System.err.println("\t[-forceinline:...]");
		System.err.println("\t[-sp source-path] [-kesosrcpath path] [-def/f config-file]");
	}

	private void parseArguments(String[] argv) {
		// allow GenerateKesoClasses to circumvent BuilderOptions
		if (argv==null) return;

		String systemDefFile = "kesorc";

		for (int i=0;i<argv.length;i++) {
			String arg = argv[i];

			if (arg.equals("-help")||arg.equals("--help")) {
				usage();
				System.exit(0);
			}

			if ((arg.equals("-def")||arg.equals("-f")) && (i+1)<argv.length ) {
				systemDefFile = argv[++i];
				File test = new File(systemDefFile);
				if (!test.exists()) {
					warn("config file = \""+systemDefFile+"\" not found!");
					systemDefFile="kesorc";
				}
				continue;
			}

			if ((arg.equals("-sourcepath")||arg.equals("-sp")) && (i+1)<argv.length ) {
				sourcepath = argv[++i];
				if (verbose) Debug.out.println("source path = "+sourcepath);
				continue;
			}

			if (arg.equals("-jdk") && (i+1)<argv.length ) {
				jdk_classpath = argv[++i];
				if (verbose) Debug.out.println("jdk path = "+jdk_classpath);
				continue;
			}

			if (arg.equals("-bootmodules") && (i+1)<argv.length ) {
				bootmodules = argv[++i];
				if (verbose) Debug.out.println("bootmodules = "+bootmodules);
				continue;
			}

			if ((arg.equals("-output")||arg.equals("-op")) && (i+1)<argv.length ) {
				outputpath = argv[++i];
				if (verbose) Debug.out.println("output path = "+outputpath);
				continue;
			}

			if (arg.equals("-kesosrcpath") && (i+1)<argv.length ) {
				kesosrcpath = argv[++i];
				if (verbose) Debug.out.println("kesosrcpath = "+kesosrcpath);
				continue;
			}

			if (arg.equals("-build")) {
				buildimage=true;
				continue;
			}

			if (arg.equals("-nobuild")) {
				buildimage=false;
				continue;
			}

			if (arg.equals("-nomakefile")) {
				doFirstLevelMakefile(false);
				continue;
			}

			if (arg.equals("-prune")) {
				prune_CT = true;
				continue;
			}

			if (arg.equals("-noprune")) {
				prune_CT = false;
				continue;
			}

			if (arg.equals("-q")) {
				quiet = true;
				continue;
			}

			if (arg.equals("-v")) {
				verbose = true;
				continue;
			}

			if (arg.equals("-vv")) {
				very_verbose = true;
				verbose = true;
				continue;
			}

			if (arg.startsWith("-v:")) {
				if (verbose_classes==null)
					verbose_classes = new HashSet<String>();
				String [] ext = split(arg,':');
				for (int n=1;n<ext.length;n++) verbose_classes.add(ext[n]);
				Debug.out.println("v-options = "+arg);
				continue;
			}

			if (arg.startsWith("-cc:")) {
				String [] ext = split(arg,':');
				if (ext.length>1) override_cc = ext[1];
				if (verbose) Debug.out.println("compiler = "+override_cc);
				continue;
			}

			if (arg.equals("-inline")) {
				inline_methods = true;
				continue;
			}

			if (arg.equals("-noinline")) {
				inline_methods = false;
				continue;
			}

			if (arg.startsWith("-forceinline:")) {
				if (force_inline==null)
					force_inline = new HashSet<String>();
				String [] ext = split(arg,':');
				for (int n=1;n<ext.length;n++) {
					this.warn("inline: "+ext[n]);
					force_inline.add(ext[n]);
				}
				continue;
			}

			if (arg.equals("-dump")) {
				dump_bc = true;
				continue;
			}

			if (arg.equals("-cfg")) {
				create_CFG = true;
				continue;
			}

			if ((arg.equals("-classtree") || arg.equals("-ct")) && (i+1)<argv.length ) {
				classtree = argv[++i];
				if (verbose) Debug.out.println("output class tree = "+classtree);
				continue;
			}

			if (arg.equals("-debug") || arg.equals("-g")) {
				debug = true;
				dbg_syms = true;
				continue;
			}

			if (arg.equals("-OF")) {
				dbg_syms = false;
				continue;
			}

			if (arg.equals("-tricore")) {
				target_arch_opt = TRICORE_PROOSEK_TARGET;
				continue;
			}

			if (arg.startsWith("-inline_costs:")) {
				String [] ext = split(arg,':');
				inline_costs = Integer.parseInt(ext[1]);
				if (ext.length>2) inline_costs_single = Integer.parseInt(ext[2]);
				if (verbose) Debug.out.println("inline costs = "+inline_costs+":"+inline_costs_single);
				inline_methods = true;
				continue;
			}

			if (arg.startsWith("-inline_costs=")) {
				String [] ext = split(arg,'=');
				inline_costs = Integer.parseInt(ext[1]);
				if (verbose) Debug.out.println("inline costs = "+inline_costs);
				inline_methods = true;
				continue;
			}

			if (arg.startsWith("-X")) {
				String [] ext = split(arg,':');
				for (int n=1;n<ext.length;n++) {
					setExtraOption(ext[n]);
				}
				if (this.hasOption("inline__all_")) {
					setExtraOption("inline_clinit");
					setExtraOption("inline_task");
					setExtraOption("inline_isr");
				}
				if (is8BitController()) {
					setExtraOption("stack8");
				}
				if (verbose) Debug.out.println("options = "+arg);
				continue;
			}

			if (arg.equals("-precompile")) {
				this.precompile = true;
				continue;
			}

			if (arg.equals("-translate")) {
				this.translate = true;
				continue;
			}

			if ( arg.equals("-linkerscript-dir") && (i+1)<argv.length ) {
				final String to_add = argv[++i];
				m_linkerscript_paths.add( to_add );
				if (verbose) Debug.out.println("added linkerscript path : "+to_add);
				continue;
			}

			Debug.out.println("Unknown option: \""+arg+"\"");
			System.exit(-1);
		}

		m_linkerscript_paths.add( "" );

		if (verbose) Debug.out.println("definition = "+systemDefFile);
		try {
			worldDef = ConfigReader.parseDefinition(systemDefFile);

			try {
				Attribute<?> xoptions = worldDef.getAttribute("ExtraOptions");
				for(String xopt: xoptions.valueString().split(":")) {
					setExtraOption(xopt);
				}
			} catch (NullPointerException ex) { }

			parseServices();
			parseFirstSystemDefinition();
		} catch (Exception e) {
			System.err.print("Failed to parse config file: ");
			System.err.println(e.toString());
			e.printStackTrace();
			usage();
			System.exit(-1);
		}

		if (systemDef==null) {
			Debug.out.println("System definition missing!");
			Debug.out.println("please set KESORC to the right definition");
			System.exit(-1);
		}
	}

	private void parseServices() throws CompileException {
		if (worldDef == null) {
			throw new CompileException("No world definition");
		}

		List<String> modules = new ArrayList<String>();

		if (bootmodules != null) {
			for(String s: bootmodules.split(":"))
				modules.add(s);
		}

		// Search for all networks, services and imports
		for(NetworkDefinition ndef: worldDef.getNetworks()) {
				ServiceManager.instance.addNetwork(ndef);
		}
		for(SystemDefinition sys: worldDef.getSystems()) {
			// Modules
			StringAttr<?> mods = (StringAttr<?>) sys.getAttribute("Modules");

			if (mods != null) {
				String src = mods.valueString();
				if (bootmodules!=null) src = bootmodules+":"+src;
				String[] modNames = src.split(":");

				for (int j = 0; j < modNames.length; ++j) {
					if (!modules.contains(modNames[j])) {
						modules.add(modNames[j]);
					}
				}
			}

			for(JavaDomain dom: sys.getJavaDomains()) {
				ServiceManager.instance.addServices(dom.getServices());
				ServiceManager.instance.addImports(dom.getImports());
			}
		}

		allModules = modules.toArray(new String[modules.size()]);

		ServiceManager.instance.buildDependencyTree();
	}

	public WorldDefinition getWorldDefinition() {
		return worldDef;
	}

	public boolean parseFirstSystemDefinition() throws CompileException {
		worldDef.rewindSystemDef();
		return parseNextSystemDefinition();
	}

	public boolean parseNextSystemDefinition() throws CompileException {
		systemDef = worldDef.nextSystemDef();
		if (systemDef==null) return false;

		DecoratedNames.reset();
		global_heap=null;

		Attribute<?> target_attr = systemDef.getAttribute("Target");
		if (target_attr==null) setTarget(DEFAULT_TARGET);
		else setTarget(getTargetByName(target_attr.valueString()));

		alarms = new ArrayList<AlarmDefinition>();
		isrs = new ArrayList<ISRDefinition>();
		java_tasks = new ArrayList<JavaTask>();

		for(JavaDomain ddef: systemDef.getJavaDomains()) {
			java_tasks.addAll(ddef.getTasks());
			isrs.addAll(ddef.getISRs());
		}

		for (DomainDefinition<?> ddef : systemDef.getAppDomains()) {
			alarms.addAll(ddef.getAlarms());
		}

		/* Modules */
		Attribute<?> src_path = systemDef.getAttribute("Modules");
		if (src_path!=null) {
			String src = src_path.valueString();
			if (bootmodules!=null) src = bootmodules+":"+src;
			verbose(getAppSourcePath()+src);
			modulepaths=src.split(":");
		} else {
			if (bootmodules!=null) modulepaths=bootmodules.split(":");
			warn("Source modules missing!");
			System.exit(-1);
		}

		return true;
	}

	public static String typeToString(int bctype) {
		switch (bctype) {
			case BCBasicDatatype.INT:
				return INT_T;
			case BCBasicDatatype.LONG:
				return LONG_T;
			case BCBasicDatatype.FLOAT:
				return FLOAT_T;
			case BCBasicDatatype.DOUBLE:
				return DOUBLE_T;
			case BCBasicDatatype.REFERENCE:
				return OBJ_T;
			case BCBasicDatatype.BYTE:
				return BYTE_T;
			case BCBasicDatatype.CHAR:
				return CHAR_T;
			case BCBasicDatatype.SHORT:
				return SHORT_T;
			case BCBasicDatatype.BOOLEAN:
				return BOOLEAN_T;
			case BCBasicDatatype.RETURN_ADDRESS:
				return "void*";
			default:
				return "void";
		}
	}

	private static BuilderOptions gopts = null;

	public static void setGlobalOptions(BuilderOptions opts) { gopts = opts; }
	public static BuilderOptions getOpts() { return gopts; }

	private static String[] split(String src, char delim) {
		List<String> v = new ArrayList<String>();

		if (src==null) return null;

		int s=0;
		for (int p=0;p<src.length();p++) {
			if (src.charAt(p)==delim) {
				v.add(src.substring(s,p));
				s=p+1;
			}
		}
		v.add(src.substring(s,src.length()));

		return v.toArray(new String[v.size()]);
	}

	public IMDomain[] getDomains() throws CompileException {
		return IMDomain.createDomains(getSystemName(), this);
	}

	public SystemDefinition getSysDef() { return systemDef; }

	public void setupGlobalHeap() throws CompileException {
		global_heap = new IMGlobalHeap(this);
	}

	public IMGlobalHeap getGlobalHeap() {
		if (global_heap==null) {
			warn("Heap not defined yet!");
			throw new Error();
		}
		return global_heap;
	}

	public List<ResourceDefinition> getSysJavaReachableResources() {
		SortedSet<ResourceDefinition> allUnique = new
		TreeSet<ResourceDefinition>(new SysObjComparator<ResourceDefinition>());
		SystemDefinition sysDef = getSysDef();
		for (JavaDomain dom : sysDef.getJavaDomains()) {
			for (ResourceDefinition td: dom.getResources()) {
				allUnique.add(td);
			}
			for (String resName : dom.getResourceImports()) {
				allUnique.add(sysDef.getSysResource(resName));
			}
		}
		return new ArrayList<ResourceDefinition>(allUnique);
	}

	public List<ResourceDefinition> getSysAllResources() {
		ArrayList<ResourceDefinition> all_resources = new ArrayList<ResourceDefinition>();
		for(DomainDefinition<?> dd: getSysDef().getAllSysDomains()) {
			all_resources.addAll(dd.getResources());
		}
		return all_resources;
	}

	public List<AlarmDefinition> getSysAllAlarms() {
		return alarms;
	}

	public List<AlarmDefinition> getSysJavaReachableAlarms() {
		HashSet<AlarmDefinition> all_unique = new HashSet<AlarmDefinition>();

		List<JavaDomain> domains = getSysDef().getJavaDomains();
		for(JavaDomain dom: domains) {
			for(AlarmDefinition td: dom.getAlarms()) {
				all_unique.add(td);
			}
			for(String resName: dom.getAlarmImports()) {
				all_unique.add(getSysDef().getSysAlarm(resName));
			}
		}
		return alarms; // Hmm, is this intended?
	}

	public List<ISRDefinition> getJavaISRs() { return isrs; }

	public List<JavaTask> getSysJavaTasks() { return java_tasks; }

	/**
	 * Returns a List of all tasks in the system, including tasks of the KESO
	 * runtime such as the garbage collector task.
	 */
	public List<TaskDefinition> getSysAllTasks() {
		ArrayList<TaskDefinition> all_tasks = new ArrayList<TaskDefinition>();
		for(DomainDefinition<?> dd: getSysDef().getAllSysDomains()) {
			all_tasks.addAll(dd.getTasks());
		}
		return all_tasks;
	}

	/**
	 * Returns the list of tasks actually reachable from the Java code.
	 *
	 * These are
	 *
	 *  a) all Java tasks
	 *  b) imported native tasks
	 */
	public List<TaskDefinition> getSysJavaReachableTasks() {
		SortedSet<TaskDefinition> allTasksUnique
		= new TreeSet<TaskDefinition>(new SysObjComparator<TaskDefinition>());
		SystemDefinition sysDef = getSysDef();
		for (JavaDomain dom : sysDef.getJavaDomains()) {
			for (TaskDefinition td : dom.getTasks()) {
				allTasksUnique.add(td);
			}
			for (String resName : dom.getTaskImports()) {
				allTasksUnique.add(sysDef.getSysTask(resName));
			}
		}
		return new ArrayList<TaskDefinition>(allTasksUnique);
	}

	public void setExtraOption(String opt) {
		extras.add(opt);
	}

	private static boolean isMacOS() {
		return System.getProperty("os.name").toLowerCase().contains("mac");
	}


	public List<String> getLinkerscriptPaths() {
		return m_linkerscript_paths;
	}

}
