/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

/**
    A simple input stream to read the bytecode.
    It is similar to java.io.ByteArrayInputStream,
    but it has some additional methods needed in this case.
*/

public final class BytecodeInputStream {

	private byte[] code;
	private int numBytes;
	private int pos;       // stream pointer

	/**
	  @param code bytecode array
	 */
	public BytecodeInputStream(byte[] code) {
		this.numBytes = code.length;
		this.code = code;
		this.pos = 0;
	}

	/**
	  @param code bytecode array
	  @param numBytes size of the bytecode array
	 */
	public BytecodeInputStream(byte[] code, int numBytes) {
		this.numBytes = numBytes;
		this.code = code;
		this.pos = 0;
	}

	/**
	  @return the current position of the stream
	 */
	public int getCurrentPosition() { return pos;}

	/**
	  not really necessary, only used for exp.FastTranslator,
	 */
	public void setCurrentPosition(int newPos) {pos = newPos;}

	/**
	  @return true, if end of stream is reached
	 */
	public boolean endOfStream() {return pos >= numBytes;}

	/**
	  @return true
	 */
	public boolean hasMore() {return pos<numBytes;}

	/**
	  @return number of bytes in this BytecodeInputStream
	 */
	public int getNumBytes() {return numBytes;}

	public byte[] getByteArray() {return code;}

	/**
	  get the current byte without modifying the stream state
	 */
	public final byte getCurrentByte() {
		return code[pos];
	}

	/**
	  get an unsigned byte that is 'offset' bytes after
	  the current position (result >= 0, always)
	 */
	private final int getByteUnsigned(int offset) {
		return ((int)code[pos+offset]) & 0xff;
	}

	/**
	  get an unsigned byte at current position
	  (result >= 0, always)
	 */
	public final int getCurrentByteUnsigned() {
		return ((int)code[pos]) & 0xff;
	}

	/**
	  read byte and advance stream pointer
	 */
	public final byte readByte() {
		return code[pos++];
	}

	/**
	  read unsigned byte and advance stream pointer
	 */
	public final int readUnsignedByte() {
		return ((int)code[pos++]) & 0xff;
	}

	/**
	  read signed short and advance stream pointer
	 */
	public final short readShort() {
		int val = getByteUnsigned(1) +
			(getByteUnsigned(0) << 8);
		pos+=2;
		return (short) val;
	}

	/**
	  read unsigned short and advance stream pointer
	 */
	public final int readUnsignedShort() {
		return (int)readShort() & 0xffff;
	}

	/**
	  read integer, advance stream pointer
	 */
	public final int readInt() {
		int val = getByteUnsigned(3) +
			(getByteUnsigned(2) << 8) +
			(getByteUnsigned(1) << 16) +
			(getByteUnsigned(0) << 24);
		pos+=4;
		return val;
	}
}
