/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;


import keso.compiler.config.*;
import keso.compiler.imcode.*;
import keso.compiler.kni.JoinPointChecker;
import keso.compiler.kni.Weavelet;
import keso.classfile.*;

import keso.compiler.backend.*;

import java.io.*;

import java.util.*;

import keso.util.*;

final class FieldAccessEntry {

	private int hCode;
	protected String class_name;
	protected String field_name;
	protected boolean read = false;
	protected boolean write = false;
	protected IMNode forward = null;

	FieldAccessEntry(IMField field) {
		class_name=field.getType().getClassName();
		field_name=field.getAlias();
		hCode=field_name.hashCode();
	}

	public void mark_read() { read = true; }
	public void mark_write() { write = true; }

	public boolean omit() {
		return !( read && write );
	}

	public void forward(IMNode fwd) { forward=fwd; }

	public IMNode forwardObj() throws CompileException {
		return forward;
	}

	public String toString() {
		StringBuffer msg = new StringBuffer(class_name);
		msg.append(" ");
		msg.append(field_name);
		msg.append(" ");
		if (read) { msg.append("r");} else { msg.append("-"); }
		if (write) { msg.append("w ");} else { msg.append("- "); }
		if (forward!=null) msg.append(forward.toReadableString());
		return msg.toString();
	}

	public boolean equals(Object obj) {
		if (obj == null) return false;

		if (!(obj instanceof FieldAccessEntry)) return false;

		FieldAccessEntry entry = (FieldAccessEntry)obj;

		if (!entry.class_name.equals(this.class_name)) return false;

		return entry.field_name.equals(this.field_name);
	}

	public int hashCode() { return hCode; }
}

/**
 * This class contains all source classes.
 */
final public class ClassStore {

	private BuilderOptions opts = null;
	private JoinPointChecker jps = null;
	private Map<String, IMClass> classes = new HashMap<String, IMClass>();
	private Map<String, Map<String, IMClass>> sub_classes
	= new HashMap<String, Map<String, IMClass>>();
	private Map<String, Map<String, IMClass>> impl_ifaces
	= new HashMap<String, Map<String, IMClass>>();
	private IMClass   rootClass;

	private SortedSet<IMField> statics_to_index = new TreeSet<IMField>();
	private Map<FieldAccessEntry, FieldAccessEntry> statics_to_access
	= new HashMap<FieldAccessEntry, FieldAccessEntry>();
	private Map<FieldAccessEntry, FieldAccessEntry> fields_to_access
	= new HashMap<FieldAccessEntry, FieldAccessEntry>();
	private Map<String, TranslationCallback> translation_callback
	= new HashMap<String, TranslationCallback>();
	private IntegerHashtable<ClassTypeInfo> class_type_info;

	private int num_static_refs = 0;

	//private static int uniqInterfaceID=0;
	//There is only one ClassStore!
	private int uniqInterfaceID=0;

	private int sharedMemorySize = 0;

	private ClassStore current_src=null;

	private HashSet<IMClass> need_check;
	private boolean need_class_checkcast = false;
	private boolean need_iface_checkcast = false;
	private boolean need_class_instanceof = false;
	private boolean need_iface_instanceof = false;
	private boolean need_obj_dup = false;

	private IMPortalService[] systemServices = null;
	private IMPortalClass[] portalProxies = null;

	/**
	 * Creates an empty ClassStore
	 */
	private ClassStore(BuilderOptions opts, ClassStore source) throws CompileException {
		this.opts=opts;
		this.jps=opts.getJoinPoints();
		this.current_src = source;
		opts.registerWeavelets(this);
	}

	/**
	 * Create the ClassStore and populate with standard Arrayclasses
	 * and the classes discovered in the sourcepath
	 */
	public ClassStore(BuilderOptions opts, String[] sourcePaths) throws IOException, CompileException {
		this.opts=opts;
		this.jps=opts.getJoinPoints();

		// Add standard Arrayclasses
		addPredefinedClass(new IMObjectArrayClass(opts,this,"[Ljava/lang/Object;","object_array","object_t*",true));
		addPredefinedClass(new IMArrayClass(opts,this,"[B","byte_array",BuilderOptions.BYTE_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[C","char_array",BuilderOptions.CHAR_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[S","short_array",BuilderOptions.SHORT_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[I","int_array",BuilderOptions.INT_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[J","long_array",BuilderOptions.LONG_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[F","float_array",BuilderOptions.FLOAT_T,true));
		addPredefinedClass(new IMArrayClass(opts,this,"[D","double_array",BuilderOptions.DOUBLE_T,true));

		// Source Path parameter contains multiple paths/classfiles separated by ':'
		// Add all Classfiles discoverable within the sourcepath to the repository
		for (int n=0; n<sourcePaths.length;n++) {
			opts.verbose("source path: "+sourcePaths[n]);
			addClassFiles(sourcePaths[n]);
		}

		// all class files have been loaded now
		registerInterfaces();

		registerSystemServices(this);
		createPortalClasses();
	}

	public BuilderOptions getOptions() {
		return opts;
	}

	public void resetCallGraph() {
		need_check = null;
		need_iface_instanceof = false;
		need_class_instanceof = false;
		need_iface_checkcast = false;
		need_class_checkcast = false;
	}

	public void needObjDUP() {
		need_obj_dup = true;
	}

	public void needInstanceOf(IMClass clazz) {
		if (need_check==null) need_check = new HashSet<IMClass>();
		need_check.add(clazz);
		if (clazz.isInterface()) {
			need_iface_instanceof = true;
		} else {
			need_class_instanceof = true;
		}
	}

	public void needCheckCast(IMClass clazz) {
		if (need_check==null) need_check = new HashSet<IMClass>();
		need_check.add(clazz);
		if (clazz.isInterface()) {
			need_iface_checkcast = true;
		} else {
			need_class_checkcast = true;
		}
	}

	/**
	 * is true if the type is used for checkcast or instanceof
	 */
	public boolean typeInUse(IMClass clazz) {
		if (need_check==null) return false;
		if (need_check.contains(clazz)) return true;
		return false;
	}

	private class StringEntry {
		String alias;
		String emit;
		String value;
	}

	private static int str_counter=0;
	private Map<String, StringEntry> string_index
	        = new HashMap<String, StringEntry>(64);
	private IntegerHashtable<StringEntry> struct_index
	        = new IntegerHashtable<StringEntry>(64);
	private IMClass str_class = null;
	public String allocString(Coder coder, IMNode node, String str) throws CompileException {

		StringEntry e = string_index.get(str);
		if (e==null) {
			if (str_class==null) str_class=getClass("java/lang/String");
			e = new StringEntry();
			e.alias = "str"+str_counter++;
			//e.emit = "((object_t*)&"+e.alias+")";
			e.emit = coder.add_accessStaticObject(str_class, e.alias);
			e.value = str;
			string_index.put(str,e);
			struct_index.put(str.length(), e);
		}

		return e.emit;
	}

	private void emit_Strings(Coder coder) throws CompileException {

		int[] keys = struct_index.sortedKeys();
		if (keys.length==0) return;

		coder.global_add("/* global strings */\n");

		IMClass str_clazz = getClass("java/lang/String");
		coder.global_header_add_class(str_clazz);

		IMClass ca_clazz = getClass("[C");
		coder.global_header_add_class(ca_clazz);

		for (int i=0;i<keys.length;i++) {
			String static_struct  = "str"+keys[i]+"_s";
			coder.global_add("struct ");
			coder.global_add(static_struct);
			coder.global_add(" {\nARRAY_HEADER\njchar data[");
			coder.global_add(keys[i]+1);
			coder.global_add("];\n};\n");
		}

		for (StringEntry e : string_index.values()) {
			String static_string  = e.alias;
			String static_chararr = e.alias+"_arr";
			String static_struct  = "str"+e.value.length()+"_s";
			String str = e.value;

			coder.global_header_add("extern ");
			coder.global_header_add(str_clazz.getClassTypeString());
			coder.global_header_add(" ");
			coder.global_header_add(static_string);
			coder.global_header_add(";\n");

			coder.global_add("static struct ");
			coder.global_add(static_struct);
			coder.global_add(" ");
			coder.global_add(static_chararr);
			coder.global_add(" ALIGN4 ={\n");
			coder.global_add(coder.add_initArray(ca_clazz, str.length(),true));

			coder.global_add("{");
			for (int i=0;i<str.length();i++) {
				char c = str.charAt(i);
				if (!opts.hasOption("wide_char") && c>255) {
					if(!wide_char_warned) {
						opts.warn("Warning: application uses wide characters. Consider using -X:wide_char or your code might not work");
						wide_char_warned = true;
					}
					coder.global_add("'?'");
				} else {
					coder.global_add((int)c);
				}
				coder.global_add(",");
			}
			coder.global_add("0} /* data */,\n};\n");

			coder.global_add(str_clazz.getClassTypeString());
			coder.global_add(" ALIGN4 ");
			coder.global_add(e.alias);
			coder.global_add("={\n");
			coder.global_add("(object_t*)&");
			coder.global_add(static_chararr);
			coder.global_add(" /* value */,");
			coder.global_add(coder.add_initObjectHeader(str_clazz,true));
			coder.global_add("\n};\n");
		}
	}

	// init global variables with parity bit
	private void emit_initGlobalParity(Coder coder) throws CompileException{
		StringBuffer p_function = new StringBuffer();
		int[] keys = struct_index.sortedKeys();
		//p_function.add("void initGlobalParity() {\n");

	
	}

	private static boolean wide_char_warned = false;

	public void registerStaticRefField(IMField sfield) {
		if (statics_to_index.contains(sfield)==false) {
			if (sfield.isReference()) sfield.setStaticIndex(num_static_refs++);
			statics_to_index.add(sfield);
		}
	}

	public void registerGlobalTranslationCallback(String key, TranslationCallback callback) {
		translation_callback.put(key, callback);
	}

	public void registerIMGetStatic(IMField field) {
		FieldAccessEntry oentry;
		FieldAccessEntry entry = new FieldAccessEntry(field);

		if ((oentry=statics_to_access.get(entry))==null) {
			oentry=entry;
			statics_to_access.put(entry, entry);
		}

		oentry.mark_read();
		opts.vverbose("registerIMGetStatic "+oentry.toString());
	}

	public void registerIMPutStatic(IMField field, IMNode forward) {
		FieldAccessEntry oentry;
		FieldAccessEntry entry = new FieldAccessEntry(field);

		if ((oentry=statics_to_access.get(entry))==null) {
			oentry=entry;
			statics_to_access.put(entry, entry);
		}

		oentry.mark_write();
		oentry.forward(forward);
		opts.vverbose("registerIMPutStatic "+oentry.toString());
	}

	public void registerIMGetField(IMField field) {
		FieldAccessEntry oentry=null;
		FieldAccessEntry entry = new FieldAccessEntry(field);

		if ((oentry=fields_to_access.get(entry))==null) {
			oentry=entry;
			fields_to_access.put(entry, entry);
		}

		oentry.mark_read();
		opts.vverbose("++ omit registerIMGetField "+oentry.toString());
	}

	public void registerIMPutField(IMField field) {
		FieldAccessEntry oentry=null;
		FieldAccessEntry entry = new FieldAccessEntry(field);

		if ((oentry=fields_to_access.get(entry))==null) {
			oentry=entry;
			fields_to_access.put(entry, entry);
		}

		oentry.mark_write();
		opts.vverbose("++ omit registerIMPutField "+oentry.toString());
	}

	public boolean omitField(IMField field) {
		if (!opts.hasOption("omit_fields")) return false;

		String classname = field.getType().getClassName();
		String fieldname = field.getFieldName();

		if (jps.checkFieldAttribute(classname,fieldname,Weavelet.ATTR_DO_NOT_OMIT)==Weavelet.TRUE)
			return false;

		FieldAccessEntry entry = new FieldAccessEntry(field);
		FieldAccessEntry fld = fields_to_access.get(entry);

		if (field.isReference() && !field.isArray()) {
			String type = field.getRefClassName();
			IMClass clazz = getClass(type);
			if (clazz.isMemoryType()) {
				/* all memory types are protected */
				return false;
			}
		}

		if (fld == null) {
			if (opts.hasVerbose())
				opts.warn(entry + " unused!");
			return true;
		}

		if (fld.omit()) opts.verbose("++ omit "+fld);
		return fld.omit();
	}

	public boolean omitStaticField(IMField field) {
		if (!opts.hasOption("omit_fields")) return false;

		String classname = field.getType().getClassName();
		String fieldname = field.getFieldName();

		if (jps.checkFieldAttribute(classname,fieldname,Weavelet.ATTR_DO_NOT_OMIT)==Weavelet.TRUE)
			return false;

		FieldAccessEntry entry = new FieldAccessEntry(field);
		FieldAccessEntry fld = statics_to_access.get(entry);

		if (field.isReference() && !field.isArray()) {
			String type = field.getRefClassName();
			if (getClass(type).isMemoryType()) {
				/* all memory types are protected */
				//opts.verbose(fld+" memory type!");
				return false;
			}
		}

		if (fld == null) {
			if (opts.hasVerbose())
				opts.warn(entry + " unused!");
			return true;
		}

		if (fld.omit()) opts.verbose("++ omit "+fld);
		return fld.omit();
	}

	public IMNode forwardStaticField(IMMethod dstMethod, IMField imfield) throws CompileException {
		if (opts.hasOption("no_forward_fields")) return null;
		FieldAccessEntry entry = new FieldAccessEntry(imfield);
		FieldAccessEntry fld = statics_to_access.get(entry);
		if (fld==null) return null;
		if (fld.forwardObj()==null) return null;
		return fld.forwardObj().copy(new IMReplaceVisitor(dstMethod, null, null));
	}

	public int getMaxInterfaceID() {
		if (opts.hasOption("iface_dbg128") && uniqInterfaceID<129) return 128;
		if (uniqInterfaceID==0) return 0;
		return uniqInterfaceID-1;
	}

	public int getNewInterfaceID() {
		if (classes==null) throw new Error();
		return uniqInterfaceID++;
	}

	public IMClass getRootClass() {
		if (rootClass==null) rootClass = getClass("java/lang/Object");
		return rootClass;
	}

	public IMClass getClass(String classname) {
		return classes.get(classname);
	}

	public boolean hasSubClasses(String classname) {
		if (sub_classes.get(classname)==null) return false;
		return true;
	}

	public IMClass[] getSubClasses(String classname) {

		Map<String, IMClass> subs = sub_classes.get(classname);
		if (subs==null) return null;

		IMClass[] all = new IMClass[subs.size()];

		if (false) {
			int i = 0;
			for (Object e : subs.values()) {
				all[i++] = (IMClass) e;
			}
		} else {
			/* all Classes first */
			int p=0;
			for (Object e : subs.values()) {
				IMClass cls = (IMClass) e;
				if (cls.isInterface()) continue;
				all[p++]=cls;
			}

			/* interfaces */
			for (Object e : subs.values()) {
				IMClass cls = (IMClass) e;
				if (!cls.isInterface()) continue;
				all[p++]=cls;
			}
		}

		return all;
	}

	/**
	 * Find all classes implementing the given interface.
	 */
	public IMClass[] findImplements(String iface_name) {
		Map<String, IMClass> impl = impl_ifaces.get(iface_name);
		if (impl==null) return null;

		IMClass[] all = new IMClass[impl.size()];
		int i = 0;

		for (IMClass e : impl.values()) {
			all[i++] = e;
		}

		return all;
	}

	public IMClass[] getAllClasses() {
		IMClass[] all = new IMClass[classes.size()];
		int i = 0;
		for (IMClass c : classes.values()) {
			all[i++] = c;
		}

		return all;
	}

	public IntegerHashtable<ClassTypeInfo> getAllClassTypeInfo() {
		return class_type_info;
	}

	public String toString() {
		return classes.toString();
	}

	/**
	 * Add all classfiles in the given directory and its subdirs to the repository
	 */
	private void addClassFiles(String path) throws IOException {
		try {
			File dir = new File(path);
			if (dir.isDirectory()) {
				String dlist[] = dir.list();
				for (int e=0;e<dlist.length;e++) addClassFiles(path+"/"+dlist[e]);
			} else {
				if (path.endsWith(".class")) addClassFile(dir);
			}
		} catch (FileNotFoundException ex) {
			Debug.out.println(ex.toString());
			throw new IOException (ex.toString());
		}
	}

	/**
	 * Read the supplied classfile and add it to the repository.
	 */
	private void addClassFile(File file) throws IOException, FileNotFoundException {
		RandomAccessFile clazz_file = new RandomAccessFile(file, "r");

		ClassData clazz = new ClassData(clazz_file);
		IMClass imclass = new IMClass(opts,this,clazz);
		addClass(imclass, clazz.getSuperClassName(), clazz.getInterfaceNames());

		clazz_file.close();
	}

	private void addPredefinedClass(IMClass imclass) {
		addClass(imclass,"java/lang/Object",null);
	}

	private void addClass(IMClass imclass, String superclass, String[] ifaces) {
		String name = imclass.getClassName();

		if (classes.get(name)!=null) {
			try {
				throw new Error();
			} catch (Error err) {
				System.err.println(name+" loaded a second time!");
				err.printStackTrace();
			}
		}

		classes.put(name, imclass);

		// register subclass at super class
		opts.vverbose("add class: "+name+" -> "+superclass);
		if (superclass==null) return;

		Map<String, IMClass> subs = sub_classes.get(superclass);
		if (subs==null) {
			subs = new HashMap<String, IMClass>(5);
			sub_classes.put(superclass,subs);
		}
		subs.put(name,imclass);
	}

	private void registerInterfacesRecursive(java.util.Set<String> impl_ifaces_all, IMClass clazz) {
		if(clazz == null) return;

		String[] ifaces = clazz.getInterfaceNames();
		if(ifaces != null) {
			for (String iface: ifaces) {
				impl_ifaces_all.add(iface);
				registerInterfacesRecursive(impl_ifaces_all, getClass(iface));
			}
		}

		registerInterfacesRecursive(impl_ifaces_all, clazz.getSuperClass());
	}

	private void registerInterfaces() {
		for(String name: classes.keySet()) {
			IMClass imclass = classes.get(name);

			// gather all interface names implemented by the current class
			// (directly or indirectly)
			HashSet<String> impl_ifaces_all = new HashSet<String>();
			registerInterfacesRecursive(impl_ifaces_all, imclass);

			for (String iface: impl_ifaces_all) {
				Map<String, IMClass> impl = impl_ifaces.get(iface);
				if (impl==null) {
					impl = new HashMap<String, IMClass>(5);
					impl_ifaces.put(iface,impl);
				}
				impl.put(name,imclass);
			}
		}
	}

	private boolean requireSharedMemory_once = true;

	public void requireSharedMemory() {
		if(!requireSharedMemory_once) return;
		requireSharedMemory_once = false;

		Attribute<?> shmSize = opts.getSysDef().getAttribute("DynMemSize");

		if(shmSize == null) {
			opts.warn("Dynamic Memory used but DynMemSize Attribute of system block not specified. Assuming 512 bytes");
			sharedMemorySize = 512;
		} else {
			sharedMemorySize = ((IntegerAttr<?>) shmSize).valueInt();
		}
	}

	private Map<String, Object> used_methods = new HashMap<String, Object>();
	private Map<String, Vector<IMClass>> method_to_class
	        = new HashMap<String, Vector<IMClass>>();

	public IMClass requireClass(String className) {
		// TODO: domain support
		return requireClass(0,className);
	}

	/**
	 * Mark a class as required. This is used for class
	 * elimination at the beginning.
	 *
	 * Die benoetigte Klasse und rekursive auch all Ihre Vorgaenger- und
	 * Interface-Klassen werden in den ClassStore uebernommen.
	 *
	 * Alle Methoden werden registriert und benutzte Methoden
	 * geladen und untersucht.
	 */
	public IMClass requireClass(int domainID, String className) {
		assert_prune();

		IMClass req_class = getClass(className);

		if (current_src==null && req_class==null) {
			// Falls wir kein Source-Repository haben, kann es sein,
			// dass wir eine Array-Klasse noch automatisch erzeugen muessen.
			if (className.charAt(0)=='[') {
				opts.verbose("## auto class "+className);
				String array_type = className.substring(1);
				IMClass elem_class = requireClass(domainID,array_type);
				addPredefinedClass(new IMObjectArrayClass(opts,this,className,
							elem_class,"object_t*",true));
				return getClass(className);
			}

			// ODER: Der Klassenname ist falsch, dann wollen wir mal nicht
			// so sein. L...; -> ...
			if (className.charAt(0)=='L' && className.endsWith(";")) {
				className=className.substring(1,className.length()-1);
				req_class=getClass(className);
			}
		}

		try {
			if (req_class==null) {
				opts.verbose("#### require "+className);

				if (current_src==null) throw new Error("class "+className+" not found! (1)");


				IMClass src_class = current_src.getClass(className);
				if (src_class==null) throw new Error("class "+className+" not found! (2)");

				req_class = src_class.cloneClass(this);
				addClass(req_class, req_class.getSuperClassName(), req_class.getInterfaceNames());

				req_class.require(domainID);

				// Alle aus der Klassendatei bekannten Methoden der Klasse werden registriert
				// und Methoden die in Verwendung sind werden uebernommen.
				// isUsedMethod -> preParseMethodData
				req_class.collectMethodNames(domainID, method_to_class);
			}
		} catch (CompileException ex) {
			ex.printStackTrace();
			throw new Error("CompileException");
		}

		// require finalizer if we have a GC
		if(opts.needFinalizers()) {
			requireMethodInternal(domainID, req_class, "finalize()V");
		}

		return req_class;
	}

	/**
	 * Mark a method as required. This is used for method
	 * elimination at the beginning.
	 *
	 * Die Klasse der angegebenen Methode wird als benoetigt markiert
	 * und damit den moeglichen Typen im ClassStore hinzugefuegt.
	 *
	 * Anschliessend werden alle Methoden mit dem angebenen
	 * Namen und Argumenten als moeglicherweise erreichbar angesehen.
	 *
	 * Der Bytecode all dieser Methoden wird "geladen" und
	 * nach (erreichbaren) Methodenaufrufen und benoetigten Klassen
	 * untersucht. Wodurch sich der Vorgang rekursive wiederholt.
	 *
	 * Ein aehnlicher Vorgang reduziert in der Optimierungs-Phase
	 * die Anzahl der Methoden noch einmal.
	 */
	public void requireMethod(int domainID, String className, String methodName) {

		// Hier wird die Klasse als erreichbar markiert und falls
		// noetig aus dem Source-Repository kopiert.
		//
		// Alle Methoden der Klasse werden dabei registriert und
		// bereits erreichbare Methoden werden geladen und untersucht.
		IMClass req_class = requireClass(domainID,className);

		// Die Klasse, welche die Methode implementiert, kann auf Grund von
		// Vererbung auch eine der Oberklassen sein. Benoetigt wird die Klasse
		// in der die Methode implementiert wird.
		while (!req_class.sourceClassImplementsMethod(methodName)) {
			req_class = requireClass(domainID, req_class.getSuperClass().getClassName());
		}

		jps.require(domainID, className, methodName);

		requireMethodInternal(domainID, req_class, methodName);
	}

	private void requireMethodInternal(int domainID, IMClass req_class, String methodName) {
		// Nur wenn der Methodentyp noch nicht als erreichbar markiert ist
		// muessen wir die Methode explizit als erreichbar markieren.
		if (!isUsedMethod(domainID,req_class,methodName)) {

			// Die Methode als erreichbar markieren.
			registerUsedMethod(domainID, req_class, methodName);

			// Alle Methoden von diesem Typ werden nun als erreichbar
			// betrachtet und muessen auch uebernommen werden.
			callbackUsedMethod(domainID, req_class, methodName);
		}
	}

	@SuppressWarnings("unchecked")
	public void registerUsedMethod(int domainID, IMClass clazz, String methodName) {
		assert_prune();

		if (opts.hasOption("fast_prune")) {
			used_methods.put(methodName, methodName);
		} else {
			Vector<IMClass> types=(Vector<IMClass>)used_methods.get(methodName);
			if (types==null) {
				types = new Vector<IMClass>();
				used_methods.put(methodName, types);
			}
			types.add(clazz);
		}
	}

	@SuppressWarnings("unchecked")
	public boolean isUsedMethod(int domainID, IMClass method_clazz, String methodName) {
		if (opts.hasOption("fast_prune")) {
			return used_methods.get(methodName)!=null;
		} else {

			Vector<IMClass> types=(Vector<IMClass>)used_methods.get(methodName);
			if (types==null) return false;

			for (IMClass used_clazz : types) {

				/* we have to require the super class or typeOf won't work */
				if (method_clazz.getSuperClassName()!=null)
					requireClass(domainID,method_clazz.getSuperClassName());

				if (!method_clazz.typeOf(used_clazz)) continue;

				return true;
			}

			return false;
		}
	}

	//=====================================================
	// Hier werden die erreichbaren Methoden nun "geladen".
	//
	final public void callbackUsedMethod(int domainID, IMClass method_clazz, String methodName) {

		Vector<IMClass> callback_classes = method_to_class.get(methodName);
		if (callback_classes==null) {
			opts.warn("method not found "+methodName+" in class " + method_clazz.getClassName() + "!");
			return;
		}

		boolean fast_prune = opts.hasOption("fast_prune");

		Enumeration<IMClass> e = callback_classes.elements();
		while (e.hasMoreElements()) {
			IMClass clazz = e.nextElement();
			if (fast_prune) {
				clazz.preParseRequiredMethodData(methodName);
			} else {
				if (!clazz.typeOf(method_clazz)) continue;
				clazz.preParseRequiredMethodData(methodName);
			}
		}
	}

	private void registerSystemServices(ClassStore repository) throws CompileException {
		List<ServiceDefinition> services
		= ServiceManager.instance.getServicesOnCurrentSystem();
		repository.systemServices = new IMPortalService[services.size()];
		for (int i = 0; i < services.size(); ++i) {
			ServiceDefinition srvDef = services.get(i);
			opts.verbose("Creating service " + srvDef);
			repository.systemServices[i] = new IMPortalService(opts, repository, srvDef);
		}
	}

	public void createPortalClasses() throws CompileException {
		opts.verbose("Creating portal classes");
		List<ImportDefinition> imps
		= ServiceManager.instance.getImportsOnCurrentSystem();

		portalProxies = new IMPortalClass[imps.size()];
		for (int i = 0; i < imps.size(); ++i) {
			ImportDefinition impDef = imps.get(i);
			opts.verbose("Creating proxy for import " + impDef);
			IMPortalClass proxy = new IMPortalClass(opts, this, impDef);
			if (classes.containsKey(proxy.getClassName()) == false) {
				addClass(proxy, "java/lang/Object", proxy.getInterfaceNames());
				portalProxies[i]=proxy;
			} else {
				portalProxies[i]=(IMPortalClass)classes.get(proxy.getClassName());
			}
		}
	}

	/**
	 * Create a new ClassStore which only contains classes needed by methods
	 * which are reachable from the start method. The start methods are the
	 * main methods of all tasks in all domains. (see requiredMethod and
	 * requiredClass for details)
	 */
	public ClassStore getPrunedClassStore() throws CompileException {
		opts.verbose("##### prune class tree");
		ClassStore new_store = new ClassStore(opts, this);
		opts.getDomains();

		// TODO: add Domain support.
		int domainID=0;

		// all class constructors are auto required!
		new_store.registerUsedMethod(0, getRootClass(), "<clinit>()V");

		/* registerSystemServices(new_store); */
		for (int i = 0; i < systemServices.length; ++i) {
			systemServices[i].require(new_store);
		}
		new_store.systemServices = systemServices;

		/* let the portal proxies survive */
		/* TODO: requireClass() in PortalService-Weavelet */
		for (int i = 0; i < portalProxies.length; ++i) {
			new_store.requireClass(domainID, portalProxies[i].getClassName());
		}

		for(FinalizeResult.ProtectedMethod pm: opts.getSysDef().getRequiredMethods()) {
			opts.verbose("Require Method: " + pm.className + "." + pm.method);

			if (pm.method==null) {
				new_store.requireClass(domainID, pm.className);
			}	else {
				new_store.requireMethod(domainID, pm.className, pm.method);
			}
		}

		new_store.registerInterfaces();
		new_store.releaseSourceRepository();

		return new_store;
	}

	private int array_start;
	private int array_end;

	public void setArrayTypeRange(int start, int end) {
		array_start = start;
		array_end = end;
	}

	/**
	 * emitClassStore is only for debug
	 */
	public void emitClassStore(StringBuffer buf) {
		int keys[] = class_type_info.sortedKeys();

		if (opts.is8BitController() || opts.hasOption("compact_header")) {
			buf.append("typedef unsigned char  class_id_t;\n");
			buf.append("typedef unsigned short array_size_t;\n\n");
			buf.append("typedef unsigned short obj_size_t;\n\n");
		} else {
			buf.append("typedef unsigned short class_id_t;\n");
			buf.append("typedef unsigned long  array_size_t;\n\n");
			buf.append("typedef unsigned int obj_size_t;\n\n");
		}

		buf.append("#define OBJECT_HEADER \\\n");
		if (opts.is8BitController() || opts.hasOption("compact_header")) {
			if (opts.getGlobalHeap().needGCInfo())
				buf.append("	unsigned char gcinfo;\\\n");
		} else {
			buf.append("	unsigned char gcinfo;\\\n");
//			buf.append("	unsigned char bitfld1;\\\n");
		}
		buf.append("	class_id_t class_id;\\\n");
		buf.append("\n");

		buf.append("typedef struct {\n");
		buf.append("OBJECT_HEADER\n");
		buf.append("} object_t;\n");
		buf.append("\n");
		buf.append("typedef struct {\n");
		buf.append("    class_id_t     type_range;\n");
		buf.append("	unsigned short size;\n");
		buf.append("	unsigned char ifaces;\n");
		buf.append("	unsigned char roff;\n");
		buf.append("} class_t;\n");

		buf.append("#define ARRAY_HEADER \\\n");
		buf.append("OBJECT_HEADER\\\n");
		buf.append("	array_size_t size;\n");

		int cls_i=0;
		for (;cls_i<keys.length;cls_i++) {
			ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(cls_i+1);
			IMClass imclass = cinfo.getIMClass();
			if (imclass.isInterface()) break;
			buf.append("#include \"");
			buf.append(imclass.getAlias());
			buf.append(".h\"\n");
		}

		buf.append("\nKESO_CONST class_t class_store[] = {\n");
		buf.append(" /* 0 RESERVED */\n");
		cls_i=0;
		for (;cls_i<keys.length;cls_i++) {
			ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(cls_i+1);
			IMClass imclass = cinfo.getIMClass();

			if (imclass.isInterface()) break;

			buf.append(" /* ");
			buf.append(cls_i+1);
			buf.append("\t");
			buf.append(imclass.getClassName());
			buf.append(" */\n");

			buf.append(" {");
			buf.append(".type_range=");
			buf.append(cinfo.getTypeRange());
			buf.append(", ");
			if (imclass!=null) {
				buf.append(".size=");
				buf.append(imclass.emitObjectSize());

				buf.append(", .ifaces=");
				buf.append((0xff & imclass.getInterfaceType()));

				buf.append(", .roff=");
				buf.append(imclass.getObjRefFieldCount());
			}
			buf.append("},\n");
		}

		buf.append("};\n");
		buf.append("#define CLASSSTORE_SIZE ");
		buf.append(cls_i);
		buf.append("\n");
		buf.append("#define CLASS(_id_) class_store[((_id_)-1)]\n");
		buf.append("#define CLS_ROFF(_cls_) CLASS(_cls_).roff\n");

		buf.append("\nchar *class_names[] = {\n");
		for (int i=0;i<keys.length;i++) {
			ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(i+1);
			IMClass imclass = cinfo.getIMClass();
			if (imclass.isInterface()) break;
			buf.append("\t\"");
			buf.append(imclass.getClassName());
			buf.append("\",\n");
		}

		buf.append("};\n");
	}

	/**
	 * translate all java classes into c
	 */
	public void translate(Coder coder, MethodTableFactory table,
	                      IntegerHashtable<ClassTypeInfo> class_type_info)
	throws CompileException {

		this.class_type_info = class_type_info;

		IMClass imclasses[] = getAllClasses();

		coder.global_header_add("/* adjust this value to your needs! */\n");
		coder.global_header_add("#define KESO_TICKS_MICRO 1995\n");

		if (opts.useGCCExtensions()) {
			coder.global_header_add("\n#define likely(x)   __builtin_expect((x),1)\n");
			coder.global_header_add("#define unlikely(x)   __builtin_expect((x),0)\n");
			coder.global_header_add("#define NORETURN      __attribute__ ((noreturn))\n");
			coder.global_header_add("#define ALIGN4        __attribute__ ((aligned (4)))\n");
			coder.global_header_add("#define ATTR_CONST    __attribute__ ((const))\n");
			coder.global_header_add("#define ATTR_PURE     __attribute__ ((pure))\n");
			coder.global_header_add("#define ATTR_MAYALIAS __attribute__ ((may_alias))\n");
			coder.global_header_add("#define KESO_INLINE   __inline__\n");
			coder.global_header_add("#define KESO_CONST    __const__\n");

		} else {
			coder.global_header_add("\n#define likely(x) x\n");
			coder.global_header_add("#define unlikely(x) x\n");
			coder.global_header_add("#define NORETURN\n");
			coder.global_header_add("#define ALIGN4\n");
			coder.global_header_add("#define ATTR_CONST\n");
			coder.global_header_add("#define ATTR_MAYALIAS\n");
			coder.global_header_add("#define ATTR_PURE\n");
			coder.global_header_add("#define KESO_INLINE\n");
			coder.global_header_add("#define KESO_CONST\n");
		}

		coder.global_define_static_assert();

		if (opts.useFixPointMath()) {
			coder.global_header_add("/* adjust this value to your needs! */\n");
			coder.global_header_add("#define KESO_FP_QF ");
			coder.global_header_add(opts.fractionBitsFloat());
			coder.global_header_add("\n");
			coder.global_header_add("#define KESO_FP_QD ");
			coder.global_header_add(opts.fractionBitsDouble());
			coder.global_header_add("\n");
			coder.global_header_add("#define KESO_FP_BLUR 0\n");
			coder.global_header_add("#define KESO_FP_FMAX 32767\n");
			coder.global_header_add("#define KESO_FP_FMIN -32767\n");
			coder.global_header_add("#define KESO_FP_DMAX 2147483647L\n");
			coder.global_header_add("#define KESO_FP_DMIN -2147483647L\n");
			coder.global_header_add("#define KESO_FP_ABS(_v_) ((_v_)<0?(-(_v_)):(_v_))\n");
			if (opts.hasOption("fp_macros")) {
				coder.global_header_add("#define KESO_FP_MACROS 1\n");
			} else {
				coder.global_header_add("/* #define KESO_FP_MACROS 1 */\n");
			}
		}

		// If you change the layout of the object header also update
		// Coder.add_initObject(). Otherwise, the init Code will break.
		int keys[] = class_type_info.sortedKeys();
		if (opts.is8BitController()) {
			coder.global_header_add("typedef char keso_machineword_t;\n");
			coder.global_header_add("typedef unsigned char keso_umachineword_t;\n");
		} else {
			coder.global_header_add("typedef int keso_machineword_t;\n");
			coder.global_header_add("typedef unsigned int keso_umachineword_t;\n");
		}

		if (opts.is8BitController()) {
			coder.global_header_add("typedef unsigned char class_id_t;\n");
			coder.global_header_add("typedef unsigned char array_size_t;\n");
			coder.global_header_add("/* max. object size is 256 Byte */\n");
			coder.global_header_add("typedef unsigned char obj_size_t;\n\n");
			coder.global_header_add("#define OBJECT_HEADER \\\n");
			if (opts.getGlobalHeap().needGCInfo())
				coder.global_header_add("	unsigned char gcinfo;\\\n");
			coder.global_header_add("	class_id_t class_id;\\\n");

			if (keys.length>254) opts.critical("too many classes for object header!");

		} else if (opts.hasOption("compact_header")) {
			coder.global_header_add("typedef unsigned char  class_id_t;\n");
			coder.global_header_add("typedef unsigned short array_size_t;\n\n");
			coder.global_header_add("/* max. object size is 64 KByte */\n");
			coder.global_header_add("typedef unsigned short obj_size_t;\n\n");
			coder.global_header_add("/* compact object header (254 classes max.) */\n");
			coder.global_header_add("#define OBJECT_HEADER \\\n");
			if (opts.getGlobalHeap().needGCInfo())
				coder.global_header_add("	unsigned char gcinfo;\\\n");
			coder.global_header_add("	class_id_t class_id;\\\n");

			if (keys.length>254) opts.critical("too many classes for compact object header!");
		} else {
			coder.global_header_add("typedef unsigned short class_id_t;\n");
			coder.global_header_add("typedef unsigned int   array_size_t;\n\n");
			coder.global_header_add("typedef unsigned int   obj_size_t;\n\n");

			/* it is important for the IRR GC that bit 0 of the first word of the
			 * object header is set to 1, in order to distinguish it from object
			 * references. The irr-alloc sets bit 0 of gcinfo to 1 on object
			 * allocation. If you change the object header, be sure not to break
			 * this. Also the endianess is of importance - gcinfo must be the least
			 * sigificant byte of the header's first word. The following header is
			 * for little endian architectures (just reverse the element order for
			 * big endian architectures).
			 */
			coder.global_header_add("#define OBJECT_HEADER \\\n");
			if (opts.getGlobalHeap().needGCInfo()) {
				coder.global_header_add("	unsigned char gcinfo;\\\n");
			}
		 //	else {
		//		coder.global_header_add("	unsigned char bitfld2;\\\n");
		//	}
		//	coder.global_header_add("	unsigned char bitfld1;\\\n");
			coder.global_header_add("	class_id_t class_id;\\\n");
		}
		coder.global_header_add("\n");

		coder.global_header_add("typedef struct {\n");
		coder.global_header_add("OBJECT_HEADER\n");
		coder.global_header_add("} object_t;\n");
		coder.global_header_add("\n");


		boolean need_roff=false;
		IntegerHashtable<IMClass> roff_index = new IntegerHashtable<IMClass>(128);
		int cls_i=0;
		for (;cls_i<keys.length;cls_i++) {
			ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(cls_i+1);
			IMClass imclass = cinfo.getIMClass();
			if (imclass.isInterface()) break;
			if (imclass.getObjRefFieldCount()!=0) {
				roff_index.put(cls_i, imclass);
			}
		}

		coder.global_header_add("#define CLS_ROFF(_cls_) ");
		if (opts.hasOption("comp_cls_store") && roff_index.size()>1) {
			coder.global_header_add("keso_roff(_cls_)\n");
			coder.global_header_add("obj_size_t keso_roff(class_id_t cls)");
			if (opts.useGCCExtensions())
				coder.global_header_add(" __attribute__ ((const)) ");
			coder.global_header_add(";\n");
			coder.global_add("obj_size_t keso_roff(class_id_t cls) {\n");
			coder.global_add("\tswitch (cls) {\n");
			int k[] = roff_index.keys();
			for (int i=0;i<k.length;i++) {
				IMClass iclass = roff_index.get(k[i]);
				coder.global_add("\t\tcase ");
				coder.global_add(iclass.getClassTypeID());
				coder.global_add(": return ");
				coder.global_add(iclass.getObjRefFieldCount());
				coder.global_add(";\n");
			}
			coder.global_add("\t}\n");
			coder.global_add("\treturn 0;\n");
			coder.global_add("}\n");
		} else {
			if (roff_index.size()>1) { need_roff=true; }
			if (need_roff) {
				coder.global_header_add("CLASS(_cls_).roff\n");
			} else {
				if (roff_index.size()==1) {
					int k[] = roff_index.keys();
					IMClass iclass = roff_index.get(k[0]);
					coder.global_header_add("((_cls_)==");
					coder.global_header_add(iclass.getClassTypeID());
					coder.global_header_add("?");
					coder.global_header_add(iclass.getObjRefFieldCount());
					coder.global_header_add(":0)\n");
				} else {
					coder.global_header_add("(0)\n");
				}
			}
		}

		coder.global_header_add("typedef struct {\n");
		//if (opts.getGlobalHeap().needObjSize() || opts.hasOption("no_inline_alloc")) {
		coder.global_header_add("	obj_size_t     size;\n");
		//}
		boolean iface_store = opts.hasOption("iface_store");
		if ((need_iface_checkcast || need_iface_instanceof) && !iface_store) {
			coder.global_header_add("/* class info extension for interfaces etc. */\n");
			coder.global_header_add("	unsigned char ifaces;\n");
		}
		if (need_roff) {
			coder.global_header_add("/* class info for bidirectional object layout */\n");
			coder.global_header_add("	unsigned char roff;\n");
		}
		coder.global_header_add("} class_t;\n\n");

		coder.global_header_add("#define ARRAY_HEADER \\\n");
		coder.global_header_add("OBJECT_HEADER\\\n");
		coder.global_header_add("	array_size_t size;\n\n");

		coder.global_header_add("typedef struct {\n");
		coder.global_header_add("ARRAY_HEADER\n");
		coder.global_header_add("} array_t;\n");
		coder.global_header_add("#define CLASS(_id_) class_store[((_id_)-1)]\n");
		coder.global_header_add("extern KESO_CONST class_t class_store[];\n");

		coder.global_header_add("/* KESO_DUP_OBJ_MAX is the maximum of recursion depth\n");
		coder.global_header_add(" * for object copy in portal calls */\n");
		coder.global_header_add("#ifndef KESO_DUP_OBJ_MAX\n");
		coder.global_header_add("#define KESO_DUP_OBJ_MAX 10\n");
		coder.global_header_add("#endif\n");

		coder.global_header_add("#define KESO_OBJ_PTR(_roff_, _mem_) (object_t*)((object_t**)(_mem_)+(_roff_))\n");
		coder.global_header_add("#define KESO_TOP_PTR(_roff_, _obj_) (jbyte*)((object_t**)(_obj_)-(_roff_))\n");
		coder.global_header_add("#define KESO_REF_PTR(_roff_, _obj_) (object_t**)((object_t**)(_obj_)-(_roff_))\n");

		if(opts.hasOption("use_coded_refs")) {
			coder.global_header_add("#define KESO_ENABLE_CODED_REFS 1\n");
		}
		if (opts.hasOption("omit_safechk")) {
			coder.global_header_add("#define KESO_OMIT_SAFECHECKS 1\n");
		}

		if (opts.hasOption("debug")) {
			coder.global_header_add("#define DEBUG 1\n");
		} else {
			if (opts.hasOption("production")) coder.global_header_add("#define KESO_PRODUCTION 1\n");
		}

		if (opts.hasOption("no_write")) {
			coder.global_header_add("#define NO_WRITE 1\n");
		}

		if (statics_to_access.size()<20) {
			coder.global_header_add("\n/* static accesses */\n");
			for (FieldAccessEntry access : statics_to_access.keySet()) {
				coder.global_header_add("/* ");
				coder.global_header_add(access.toString());
				coder.global_header_add(" */\n");
			}
		}

		if (fields_to_access.size()<20) {
			coder.global_header_add("\n/* field accesses */\n");
			for (FieldAccessEntry access : fields_to_access.keySet()) {
				coder.global_header_add("/* ");
				coder.global_header_add(access.toString());
				coder.global_header_add(" */\n");
			}
		}

		for (int i=0;i<imclasses.length;i++) { imclasses[i].global_header(coder); }

		// opaque definition of domain_t to enable use of domain_t* types
		coder.global_header_add("typedef struct DOMAINDESC domain_t;\n");

		// add keso_printf function
		coder.writeErrorHandler();

		if (sharedMemorySize>0) {
			opts.warn("Dynamic shared memory is not tested!");
			opts.warn("hint: use allocStaticMemory or allocStaticDeviceMemory!");
			coder.global_header_add("\n#define KESO_SHARED_MEM_SIZE ");
			coder.global_header_add(sharedMemorySize);
			coder.global_header_add("\n");

			if (opts.is8BitController()) {
				coder.global_header_add("\n#define KESO_MEM_NO_GC 1");
				coder.global_add_fkt("tiny_shared_memory.c");
			} else {

				/*
				 * TODO: less restrict locks
				 * These locks may protect long periods of time while
				 * searching in the free list or doing garbage collection.
				 */
				opts.warn("FIXME: Dynamic shared memory uses sub optimal locking!");
				coder.global_header_add("#define LOCK_SHARED_MEMORY() SuspendAllInterrupts()\n");
				coder.global_header_add("#define UNLOCK_SHARED_MEMORY() ResumeAllInterrupts()\n");

				/*
				 * These locks protect only short sections.
				 */
				coder.global_header_add("#define LOCK_SHARED_MEMORY_GQ() SuspendAllInterrupts()\n");
				coder.global_header_add("#define UNLOCK_SHARED_MEMORY_GQ() ResumeAllInterrupts()\n");

				coder.global_add_fkt("shared_memory.c");
			}
		}

		if (opts.hasOption("enable_pcp_emu")) {
			coder.global_header_add("extern unsigned int keso_pcp_counter;\n");
			coder.global_add("unsigned int keso_pcp_counter=0;\n");
			coder.global_header_add("#define KESO_LOCK(_obj_) do { if (!(keso_pcp_counter++)) GetResource(RES_SCHEDULER); } while (0)\n");
			coder.global_header_add("#define KESO_UNLOCK(_obj_) do { if (!(--keso_pcp_counter)) ReleaseResource(RES_SCHEDULER); } while (0)\n");
		} else {
			coder.global_header_add("#define KESO_LOCK(_obj_)\n");
			coder.global_header_add("#define KESO_UNLOCK(_obj_)\n");
		}

		IMClass obj_array = getClass("[Ljava/lang/Object;");
		int obj_array_id = 0;
		coder.global_header_add("\n#define keso_isObjectArrayClass(_class_id_) (");
		if (obj_array!=null) {
			obj_array_id = obj_array.getClassTypeInfo().getClassTypeID();
			coder.global_header_add(obj_array_id);
			coder.global_header_add("==(_class_id_))\n");
			coder.global_header_add("#define KESO_OBJECTARRAYCLASS_AVAILABLE 1\n");
		} else {
			coder.global_header_add("0)\n");
		}

		coder.global_header_add("#define keso_isArrayClass(_class_id_) (");
		if (array_start>1) {
			coder.global_header_add(array_start-1);
			coder.global_header_add("<(_class_id_)&&(_class_id_)<");
			coder.global_header_add(array_end);
			coder.global_header_add(")\n");
			coder.global_header_add("#define KESO_HAS_ARRAYS 1\n");
		} else {
			coder.global_header_add("0)\n");
		}

		IMClass mem_obj = getClass("keso/core/Memory");
		if (mem_obj!=null) {
			coder.global_header_add("\n#define KESO_MEMORYOBJECT_AVAILABLE\n");
			coder.global_header_add("#define KESO_MEMORY_ID ");
			coder.global_header_add(mem_obj.getClassTypeInfo().getClassTypeID());
			coder.global_header_add("\n#define keso_isMemoryClass(_class_id_) (");
			coder.global_header_add(mem_obj.getClassTypeInfo().getClassTypeID());
			coder.global_header_add("==(_class_id_))\n");
			coder.global_header_add("#define KESO_MEMORY_ADDR(_obj_) ");
			coder.global_header_add(coder.getField(mem_obj,"(_obj_)","addr"));
			coder.global_header_add("\n");
			coder.global_header_add("#define KESO_MEMORY_SIZE(_obj_) ");
			coder.global_header_add(coder.getField(mem_obj,"(_obj_)","size"));
			coder.global_header_add("\n");
		} else {
			coder.global_header_add("\n/* not available #define KESO_MEMORYOBJECT_AVAILABLE */\n");
			coder.global_header_add("#define keso_isMemoryClass(_class_id_) (0)\n");
			coder.global_header_add("#define KESO_MEMORY_ADDR(_obj_) (0)\n");
			coder.global_header_add("#define KESO_MEMORY_SIZE(_obj_) (0)\n");
		}

		coder.global_add("\n");

		IMClass mtclazz = getClass("keso/core/MT");
		if(mtclazz!=null) {
			coder.global_header_add_class(mtclazz);
			coder.global_header_add("#define KESO_OBJ_TO_MT(_obj_) ");
			coder.global_header_add(mtclazz.getAlias().toUpperCase());
			coder.global_header_add("_OBJ(_obj_)\n");
		}

		if (need_iface_checkcast || need_iface_instanceof) {
			coder.getIFMatrix().emitInterfaceMatrix();
		}

		if (need_class_checkcast)
			coder.global_add_fkt("exceptions/cast_chk.c");

		if (need_class_instanceof) {
			if(opts.hasOption("use_coded_refs")) {
				coder.global_header_add("#define keso_instanceof_fast(_cid_,_obj_) ((_cid_)==(NATIVE_ADDR(_obj_))->class_id)\n");
			} else {
				coder.global_header_add("#define keso_instanceof_fast(_cid_,_obj_) ((_cid_)==(_obj_)->class_id)\n");
			}
			
			coder.global_header_add("jboolean keso_instanceof(class_id_t class_id, class_id_t type_range, object_t* obj);\n");
			coder.global_add("jboolean keso_instanceof(class_id_t class_id, class_id_t type_range, object_t* obj) {\n");
			coder.global_add("	register class_id_t obj_type;\n");
			if(opts.hasOption("use_coded_refs")) {
				coder.global_add("	obj=NATIVE_ADDR(obj);\n");
			}
			coder.global_add("	if (obj==NULL) return 0;\n\n");
			coder.global_add("	obj_type=obj->class_id;\n");
			coder.global_add("	if ( obj_type<class_id || obj_type>type_range ) return 0;\n\n");
			coder.global_add("	return 1;\n}\n\n");
		}

		if(opts.hasLinkedListOfLocalReferences()) {
			coder.global_header_add("\ntypedef struct keso_stack_s {\n");
			coder.global_header_add(" domain_t *domain;\n");
			coder.global_header_add("	object_t** llrefs;\n");
			coder.global_header_add("	struct keso_stack_s KESO_CONST* next;\n");
			coder.global_header_add("} keso_stack_t;\n");

			coder.global_header_add("\n#define KESO_EOLL ((object_t *) -1) /* end of linked list */\n");
		}

		coder.global_header_add("\n/* static fields */\n");
		coder.global_header_add("#define KESO_NUM_STATIC_REFS ");
		coder.global_header_add(num_static_refs);
		coder.global_header_add("\n");
		for (IMField field : statics_to_index) {
			field.global_emit_macro(coder);
		}
		coder.global_header_add("\n");

		coder.global_add("\n");

		// translate services
		for (int srv = 0; srv < systemServices.length; ++srv) {
			systemServices[srv].translate(coder);
		}

		table.translate(coder);

		coder.global_add("\n");

		if ((need_iface_checkcast || need_iface_instanceof) && iface_store) {
			coder.global_header_add("extern unsigned char iface_store[];\n");
			coder.global_add("/* separated interface type information. */\n");
			coder.global_add("unsigned char iface_store[] = {\n");
			cls_i=0;
			int if_start = -1;
			int if_end = -1;
			int if_skip = 0;
			for (;cls_i<keys.length;cls_i++) {
				ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(cls_i+1);
				IMClass imclass = cinfo.getIMClass();

				if (imclass.isInterface()) break;

				int if_type = imclass.getInterfaceType();
				if (if_start<0) {
					if (if_type==0) continue;
					if_start = cls_i+1;
				} else {
					if (if_type==0) {
						if_skip += 1;
						continue;
					}
				}

				while (if_skip>0) {
					coder.global_add("0,\n");
					if_skip--;
				}

				coder.global_add(" /* ");
				coder.global_add(cls_i+1);
				if (imclass!=null && opts.hasDbgSymbols()) {
					coder.global_add("\t");
					coder.global_add(imclass.getClassName());
				}
				coder.global_add(" */\n");

				coder.global_add((0xff & if_type));
				coder.global_add(",\n");
				if_end = cls_i+1;
			}
			coder.global_add("};\n\n");
			if (if_start>0) {
				coder.global_header_add("#define KESO_IFACE_START ");
				coder.global_header_add(if_start);
				coder.global_header_add("\n");
				coder.global_header_add("#define KESO_IFACE_SIZE ");
				coder.global_header_add((if_end-if_start)+1);
				coder.global_header_add("\n");
			} else {
				coder.global_header_add("#define KESO_IFACE_START 0\n");
				coder.global_header_add("#define KESO_IFACE_SIZE 0\n");
			}
		}

		coder.global_add("/* use the CLASS(id) macro to access the class store! */\n");
		coder.global_add("/* use the CLS_ROFF(id) macro to access the reference offset! */\n");
		coder.global_add("KESO_CONST class_t class_store[] = {\n");

		coder.global_add(" /* 0 RESERVED */\n");
		cls_i=0;
		for (;cls_i<keys.length;cls_i++) {
			ClassTypeInfo cinfo = (ClassTypeInfo)class_type_info.get(cls_i+1);
			IMClass imclass = cinfo.getIMClass();

			if (imclass.isInterface()) break;

			coder.global_add(" /* ");
			coder.global_add(cls_i+1);
			if (imclass!=null && opts.hasDbgSymbols()) {
				coder.global_add("\t");
				coder.global_add(imclass.getClassName());
			}
			coder.global_add(" */\n");

			coder.global_add(" {");

			coder.global_header_add("#define ");
			coder.global_header_add(imclass.getAlias().toUpperCase());
			coder.global_header_add("_RANGE ");
			coder.global_header_add(cinfo.getTypeRange());
			coder.global_header_add("\n");

			if (imclass!=null) {

				//if (opts.getGlobalHeap().needObjSize() || opts.hasOption("no_inline_alloc")) {
				coder.global_add(imclass.emitObjectSize());
				//}

				if (!iface_store && (need_iface_checkcast || need_iface_instanceof)) {
					coder.global_add(", ");
					coder.global_add((0xff & imclass.getInterfaceType()));
				}

				if (need_roff) {
					coder.global_add(", ");
					coder.global_add(imclass.getObjRefFieldCount());
				}
			}
			coder.global_add("},\n");
		}

		coder.global_add("};\n");
		coder.global_header_add("#define KESO_CLASSSTORE_SIZE ");
		coder.global_header_add(cls_i-1);
		coder.global_header_add("\n");

		/* ASSERTCLASSID */
		coder.global_header_add("#define ASSERTCLASSID(_id_) KESO_ASSERT(0<(_id_) && (_id_)<");
		coder.global_header_add(cls_i);
		coder.global_header_add(")\n");

		if (opts.hasOption("debug")) {
			coder.global_header_add("#define ASSERTOBJ(_obj_) {");
			coder.global_header_add("if (_obj_!=NULL) ASSERTCLASSID((_obj_)->class_id); }\n");
		} else {
			coder.global_header_add("#define ASSERTOBJ(_obj_)\n");
		}

		for (IMClass imClass : imclasses) {
			imClass.translate(coder);
		}

		emit_Strings(coder);

		if(translation_callback.size() > 0) {
			coder.global_header_add("\n/* translation callbacks */\n");
			coder.global_add("\n/* translation callbacks */\n");
			for (TranslationCallback cb : translation_callback.values()) {
				cb.translate_global(coder);
			}
		}


		// add required c functions
		Collection<ImportDefinition> imports
		= ServiceManager.instance.getImportsOnCurrentSystem();
		for (ImportDefinition impDef : imports) {
			if (impDef.isLocal()) {
				if (!opts.isSingleDomainSystem()) {
					coder.global_add_fkt("mem_copy.c");
					coder.global_add_fkt("memobj_copy.c");
					if (need_obj_dup) coder.global_add_fkt("object_copy.c");
				}
			} else {
				if (impDef.needDataConversion()) {
					coder.global_add_fkt("mem_revcopy.c");
				} else {
					coder.global_add_fkt("mem_copy.c");

				}
			}
		}

		Collection<ServiceDefinition> services = ServiceManager.instance.getServicesOnCurrentSystem();
		for (ServiceDefinition srvDef : services) {
			if (srvDef.hasNetworkAccess()) {
				coder.global_add_fkt("mem_copy.c");
			}
		}

	}

	private void releaseSourceRepository() {
		current_src=null;
	}

	public void assert_prune() {
		if (current_src!=null) return;
		if (opts.pruneClassTree()) {
			opts.warn("require class or method before or after prune.");
			if (opts.hasOption("dbg_prune")) throw new Error();
		}
	}

	public void resetFieldAccess() {
		statics_to_access.clear();
		fields_to_access.clear();
	}

	public void destroy() {
		classes = null;
		sub_classes = null;
		impl_ifaces = null;
		rootClass = null;
		statics_to_index = null;
		statics_to_access = null;
		fields_to_access = null;
	}

	public int getNumStaticRefs() {
		return num_static_refs;
	}

	public static String getArrayElementType(String arrayType) {
		int typeIndex = arrayType.lastIndexOf('[');
		if(typeIndex == -1)
			return arrayType;
		if(arrayType.charAt(typeIndex+1) == 'L')
			return arrayType.substring(typeIndex+2, arrayType.length()-1);
		return arrayType.substring(typeIndex);
	}

	public int getSharedMemorySize() { return sharedMemorySize; }
}
