/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import keso.compiler.imcode.*;

import java.util.*;

import keso.util.*;

/**
 * This Bucket is a sorted Set. Every value appears only
 * once, and it is sorted in a FIFO.
 */
final class Bucket {

	Bitmap mark;
	int[] values;
	int size;
	int index;

	Bucket(int cap) {
		values = new int[cap];
		clear();
	}

	void add(int value) {
		if (mark.test(value)) return;
		values[size]=value;
		mark.mark(value);
		size++;
	}

	int first() {
		index = 0;
		return next();
	}

	int next() {
		if (index>=values.length) return 0;
		return values[index++];
	}

	int[] getArray() {
		if (size<values.length) {
			int[] a=new int[size];
			for (int i=0;i<size;i++) a[i]=values[i];
			values = a;
		}
		return values;
	}

	void clear() {
		mark = new Bitmap(values.length);
		size=0;
	}
}

final public class DominatorTree implements Iterable<IMBasicBlock> {

	private final class Itr implements Iterator<IMBasicBlock> {

		private int index = 1;

		@Override
		public boolean hasNext() {
			return (index < vertices.length);
		}

		@Override
		public IMBasicBlock next() {
			if (hasNext() == false)
				throw new NoSuchElementException();
			IMBasicBlock o = vertices[index++];
			while (index < vertices.length && vertices[index] == null) {
				++index;
			}
			return o;
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

	private int uniqueNumber;
	private int	bbCount;
	private IMMethod method;
	private IMBasicBlock   root;

	private IMBasicBlock[] vertices;

	private int[] parents;
	private	int[] idom; /* dominators */
	private Bucket[] df; /* dominance frontiers */
	private	int[] semi; /* semidominators */
 	private	int[] ancestor;
 	private	int[] best;

	public DominatorTree(IMMethod method, IMBasicBlock root) {
		this.method = method;
		this.root = root;
		uniqueNumber = 1;
		bbCount = 1;
	}

	/**
	 * sort all basic blocks by a depth-first search.
	 */
	private void depthFirstSearch(IMBasicBlock parent, IMBasicBlock current) throws CompileException {
		if (current.getDFNum()==0) {
			current.setDFNum(uniqueNumber);
			vertices[uniqueNumber]=current;
			if (parent!=null) parents[uniqueNumber]=parent.getDFNum();
			uniqueNumber++;
			bbCount++;

			IMBasicBlockList successor = current.getSucc();
			for (int i=0;i<successor.length();i++) {
				depthFirstSearch(current,successor.at(i));
			}
		}
	}

	/*
	 * see Lengauer and Tarjan
	 */
	public void computeDominators() throws CompileException {
		int max_number_of_bb = method.getNumberOfBasicBlocks()+1;
		if (max_number_of_bb==1) return;

		vertices = new IMBasicBlock[max_number_of_bb];
		parents  = new int[max_number_of_bb];

		depthFirstSearch(null,root);

		idom = new int[bbCount];
		semi = new int[bbCount];
		ancestor = new int[bbCount];
		best = new int[bbCount];

		Bucket[] bucket = new Bucket[bbCount];
		for (int i=0;i<bbCount;i++) bucket[i] = new Bucket(bbCount);

		int[] samedom = new int[bbCount];

		for (int i = bbCount - 1; i>1 ; i--) {
			if (vertices[i]==null) {
				System.err.println("warn: vertex["+i+"]=null number_of_bb="+bbCount);
				System.err.println(method.dumpLabelIndex());
				continue;
			}
			int current = vertices[i].getDFNum();
			int parent = parents[current];
			int semidom = parent;

			/*
			 * Calculate the semidominator of _current_, based
			 * on the semidominator theorem.
			 */

			IMBasicBlockList pred = vertices[current].getPred();
			for (int j=0;j<pred.length();j++) {
				int s;
				int v = pred.at(j).getDFNum();
				if (v <= current) {
					s = v;
				} else {
					s = semi[ancestorWithLowestSemi(v)];
				}
				if (s<semidom) {
					semidom = s;
				}
			}

			semi[current] = semidom;

			/*
			 * Calculation of _current_'s dominator is deferred
			 * until the path from _semidom_ to _current_ has been
			 * linked into the forest.
			 */

			bucket[semidom].add(current);
			link(parent, current);

			/*
			 * Now that the path form _parent_ to _v_ has been linked into
			 * the spanning forest, these lines calculate the dominator of _v_,
			 * based on the first clause the dominator theorem, or else
			 * defer the calculation untile _y_'s dominator is known.
			 */

			for (int t=bucket[parent].first();t!=0;t=bucket[parent].next()) {
				int y = ancestorWithLowestSemi(t);
				if (semi[y] == semi[t]) {
					idom[t] = parent;
				} else {
					samedom[t] = y;
				}
			}
			bucket[parent].clear();
		}
		bucket=null;

		/*
		 * Now all the deferred dominator calculation, based on
		 * the second clause of the dominator theorem, are performed.
		 */

		for (int i=1;i<bbCount;i++) {
			if (samedom[i]!=0) idom[i]=idom[samedom[i]];
		}

		parents=null;
		semi=null;
		best=null;
		ancestor=null;
	}

	/**
	 * returns the dominance frontiers of basic block (bb)
	 *
	 * The dominance frontier of a node A is the set of nodes that A does not
	 * strictly dominate, but does dominate some immediate predecessor of. From A's
	 * point of view, these are the nodes at which other control paths that don't go
	 * through A make their earliest appearance.
	 * (Wikipedia)
	 */

	public int[] getDominanceFrontierIDs(IMBasicBlock bb) {
		return getDominanceFrontierIDs(bb.getDFNum());
	}

	public int[] getDominanceFrontierIDs(int id) {
		if (id==0) return new int[0];
		if (df[id]==null) {
			System.err.println("=========================================");
			System.err.println("method = "+method.getMethodNameAndType());
			System.err.println("basic block = "+id);
			System.err.println("=========================================");
			try { computeDominanceFrontier2(); } catch (Exception ex) {}
		}
		return df[id].getArray();
	}

	/**
	 * computes the dominance frontiers using a simple algorithm
	 * published at wikipedia.
	 */
	public void computeDominanceFrontier2() throws CompileException {
		if (idom==null) computeDominators();

		df = new Bucket[bbCount];
		for (int b=1;b<bbCount;b++)
			df[b] = new Bucket(bbCount);

		// for each node b
		for (int b=1;b<vertices.length;b++) {
			if (vertices[b]==null) continue;
			IMBasicBlockList pred = vertices[b].getPred();
			// if the number of predecessors of b  2
			if (pred.length()>1) {
				//for each p in predecessors of b
				for (int i=0;i<pred.length();i++) {
					// runner := p
					int runner = pred.at(i).getDFNum();
					// while runner != idom(b)
					while (runner!=idom[b]) {
						//if (runner == b) throw new CompileException(vertex[b] + " predecessor " + i + " = " + pred.at(i));
						df[runner].add(b);
						runner = idom[runner];
						if (runner==0) break;
					}
				}
			}
		}
	}

	/**
	 * computes the dominance frontiers
	 */
	public void computeDominanceFrontier() throws CompileException {
		if (idom==null) computeDominators();
		df = new Bucket[bbCount];
		computeDF(1);
		children=null;
	}

	private void computeDF(int n) throws CompileException {
		Bucket S = new Bucket(bbCount);

		IMBasicBlockList succ = vertices[n].getSucc();
		for (int i=0;i<succ.length();i++) {
			int y = succ.at(i).getDFNum();
			if (idom[y]!=n) S.add(y);
		}

		Bucket C = childrenOf(n);
		if (C!=null) {
			for (int c=C.first();c!=0;c=C.next()) {
				computeDF(c);
				Bucket DF = df[c];
				for (int w=DF.first();w!=0;w=DF.next()) {
					if (!dom(n,w)) S.add(w);
				}
			}
		}

		df[n] = S;
	}

	private Bucket[] children = null;

	private void computeChildrenForDFS() {
		children=new Bucket[bbCount];
		for (int i=1;i<bbCount;i++) {
			int dom = idom[i];
			if (children[dom]==null) children[dom] = new Bucket(bbCount);
			children[dom].add(i);
		}
	}

	private Bucket childrenOf(int n) {
		if (children==null) computeChildrenForDFS();
		return children[n];
	}

	private void link(int parent, int current) {
		ancestor[current]=parent;
		best[current]=current;
	}

	private int ancestorWithLowestSemi(int v) {
		int a = ancestor[v];
		if (ancestor[a]!=0) {
			int b = ancestorWithLowestSemi(a);
			ancestor[v]=ancestor[a];
			if (semi[b]<semi[best[v]])
				best[v]=b;
		}
		return best[v];
	}

	/**
	 * returns true if "a" is a dominator of "b".
	 */
	private boolean dom(int a, int b) {
		if (a==1) return true;
		if (b<=1) return false;
		if (idom[b]==a || dom(a,idom[b])) return true;
		return false;
	}

	/**
	 * returns true if "a" is a dominator of "b".
	 */
	public boolean dom(IMBasicBlock a, IMBasicBlock b) {
		if (a==null) throw new Error("a is null");
		if (b==null) throw new Error("b is null");
		return dom(a.getDFNum(),b.getDFNum());
	}

	/**
	 * return the immediate dominator of basic block bb
	 */
	public IMBasicBlock getDominatorOf(IMBasicBlock bb) {
		return vertices[idom[bb.getDFNum()]];
	}

	public IMBasicBlock getIMBasicBlock(int id) {
		return vertices[id];
	}

	/**
	 * return a all basic blocks sorted by a depth-first search
	 * of the dominator tree.
	 */
	@Override
	public Iterator<IMBasicBlock> iterator() {
		return new Itr();
	}

}
