/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import keso.compiler.backend.Coder;
import keso.compiler.imcode.IMClass;
import keso.compiler.imcode.IMMethod;
import keso.compiler.imcode.MethodTable;

import java.util.*;

final public class MethodTableFactory
{
	final private static boolean debug = false;

	final private ClassStore repository;
	final private BuilderOptions opts;
	final private Map<String, MethodTable> mtables;

	public MethodTableFactory(BuilderOptions opts, ClassStore repository) {
		this.opts       = opts;
		this.repository = repository;
		this.mtables    = new HashMap<String, MethodTable>();
	}

	public void computeTables() throws CompileException {

		IMClass root = repository.getRootClass();
		opts.verbose("##### compute java.lang.Object method table");
		computeCurrentMethodTable(root);

		opts.verbose("##### compute interface tables");
		IMClass[] subs = root.getSubClasses();
		for (int i=0;subs!=null && i<subs.length;i++) {
			IMClass cclass = subs[i];
			if (!cclass.isInterface()) continue;
			computeInterfaceTable(cclass);
		}

		opts.verbose("##### compute method tables");
		for (int i=0;subs!=null && i<subs.length;i++) {
			IMClass cclass = subs[i];
			if (cclass.isInterface()) continue;
			opts.vverbose(cclass.getClassName());
			computeMethodTable(cclass);
		}
	}

	private void computeCurrentMethodTable(IMClass current) throws CompileException {
		for(IMMethod m: current.getMethods()) {
			if (m.isStatic() || m.isInit()) continue;
			if (m.getMethodTable()!=null) continue;
			String sig = m.getMethodNameAndType();
			MethodTable table = new MethodTable(current);
			mtables.put(current.getClassName()+sig, table);
			buildMethodTable(table, current, m, sig);
		}
	}

	private void computeInterfaceTable(IMClass iface) throws CompileException {
		opts.verbose("#### compute interface "+iface.getClassName());

		if (iface.getInterfaceID()==-1) {
			if ( repository.typeInUse(iface) || opts.hasOption("iface_no_omit_unused") ) {
				int ifid = repository.getNewInterfaceID();
				if (ifid==0) ifid = repository.getNewInterfaceID();
				iface.setInterfaceID(ifid);
				opts.vverbose("#? interface "+iface.getClassName()+" id="+ifid);
			} else {
				opts.vverbose("#? interface "+iface.getClassName()+" id=0 <omit>");
				iface.setInterfaceID(0);
			}
		}

		IMMethod[] methods = iface.getMethods();
		if (methods==null) return;

		IMClass[] impl_classes = repository.findImplements(iface.getClassName());

		if (impl_classes==null || impl_classes.length==0) return;

		for (int i=0;methods!=null && i<methods.length;i++) {
			IMMethod m = methods[i];
			if (m.isClassInit()) continue;
			if (m.getMethodTable()!=null) continue;
			String sig = m.getMethodNameAndType();
			opts.verbose("### compute interface method "+sig);

			MethodTable table = new MethodTable(impl_classes[0], m);

			m.setMethodTable(table);
			mtables.put(iface+sig, table);
			for (int j=0;j<impl_classes.length;j++) {
				IMClass iclass = impl_classes[j];
				IMMethod imethod = iclass.getMethod(sig);

				if (imethod==null) {
					if (iclass.isInterface()) continue;
					if (iclass.isAbstract()) continue;
					if (debug) opts.warn(iclass.getClassName()+"."+sig+" not found!");
					//addToMethodTable(table, iclass, new IMMethodNotFound(iclass, sig), sig);
					// Do not throw an exception here - this case is valid if
					// the method existed in the first place, but was removed
					// because it is unused.
					//throw new CompileException(iclass.getClassName()+"."+sig+" not found!");
				} else {
					addToMethodTable(table, iclass, imethod, sig);
				}
			}
		}

		IMClass[] subs = iface.getSubClasses();
		for (int i=0;subs!=null && i<subs.length;i++) {
			computeInterfaceTable(subs[i]);
		}
	}

	private void computeMethodTable(IMClass current) throws CompileException {
		IMMethod[] methods = current.getMethods();
		for (int i=0;methods!=null && i<methods.length;i++) {
			IMMethod m = methods[i];
			if (m.isStatic() || m.isInit()) continue;
			if (m.getMethodTable()!=null) continue;
			String sig = m.getMethodNameAndType();
			MethodTable table = new MethodTable(current);
			mtables.put(current.getClassName()+sig, table);
			buildMethodTable(table, current, m, sig);
		}
		IMClass[] subs = current.getSubClasses();
		for (int i=0;subs!=null && i<subs.length;i++) {
			computeMethodTable(subs[i]);
		}
	}

	private void addToMethodTable(MethodTable table, IMClass current, IMMethod method, String sig) throws CompileException {

		/* computation of interface method table EHT */

		MethodTable current_table = method.getMethodTable();
		if (current_table!=null) {
			if (current_table != table)
				table.joinTables(current_table);
		}

		table.addMethod(current.getClassTypeID(), method);
		method.setMethodTable(table);
		IMClass[] subs = current.getSubClasses();

		for (int i=0;subs!=null && i<subs.length;i++) {
			IMMethod next = subs[i].getMethod(sig);
			if (next==null) next = method;
			addToMethodTable(table,subs[i],next,sig);
		}
	}

	private void buildMethodTable(MethodTable table, IMClass current, IMMethod method, String sig) throws CompileException {

		/* computation of method table HT */

		table.addMethod(current.getClassTypeID(), method);
		if (method.getMethodTable()==null) method.setMethodTable(table);

		IMClass[] subs = current.getSubClasses();
		for (int i=0;subs!=null && i<subs.length;i++) {
			IMMethod next = subs[i].getMethod(sig);
			if (next==null) next = method;
			buildMethodTable(table,subs[i],next,sig);
		}
	}

	public void emitIndexTable(Coder coder, Collection<MethodTable> ctables,
			String version) throws CompileException {
		coder.global_header_add("extern ");
		coder.global_header_add(opts.declareConst());
		coder.global_header_add("code_t addr");
		coder.global_header_add(version);
		coder.global_header_add("_table[];\n");

		//coder.global_add("/* compressed dispatch_table */\n");
		coder.global_add(opts.declareConst());
		coder.global_add("code_t addr");
		coder.global_add(version);
		coder.global_add("_table[] = {\n");


		int addr_table_off = 0;
		for(MethodTable table: ctables) {
			addr_table_off = table.emit_addr_table(coder, addr_table_off);
		}
		coder.global_add("};\n");

		/* compute global index table */
		List<Integer> index_table = null;

		if (!opts.hasOption("vt_comp_sort")) {
			for(MethodTable table: ctables) {
				index_table = table.updateIndexSeq(index_table);
			}
			if (index_table!=null) {
				int n = 0;
				for(MethodTable table: ctables) {
					table.updateOffset(index_table);
					if (table.hasDispatchTable()) n++;
				}

				int bits = (index_table.size()*8) / n;

				coder.global_header_add("extern ");
				coder.global_header_add(opts.declareConst());
				if (version.equals("16")) {
					coder.global_header_add("unsigned short index");
				} else {
					coder.global_header_add("unsigned char index");
				}
				coder.global_header_add(version);
				coder.global_header_add("_table[];\n");

				coder.global_add("/* index_table size:"+index_table.size());
				coder.global_add(" ctables:"+n+" bits/table:"+bits);
				coder.global_add(" */\n");
				coder.global_add(opts.declareConst());
				if (version.equals("16")) {
					coder.global_add("unsigned short index");
				} else {
					coder.global_add("unsigned char index");
				}
				coder.global_add(version);
				coder.global_add("_table[] = {");
				for (int i=0; i < index_table.size(); i++) {
					if ((i%4)==0) coder.global_add("\n/* "+i+" */ \t");
					int ndx = index_table.get(i);
					if (ndx>=0 && ndx<10) coder.global_add(" ");
					coder.global_add(ndx);
					coder.global_add(",");

				}
				coder.global_add("\n};\n");
			}
		} else {
			LinkedList<MethodTable> ctab_sorted = new LinkedList<MethodTable>();

			for(MethodTable table: ctables) {
				int ts = table.tableSize();
				int ci=0;
				while (ci<ctab_sorted.size() &&	ctab_sorted.get(ci).tableSize()<ts) {
					ci++;
				}
				ctab_sorted.add(ci, table);
			}

			if (ctab_sorted.size()!=ctables.size()) throw new Error();

			for(MethodTable table: ctab_sorted) {
				index_table = table.updateIndexSeq(index_table);
			}
			if (index_table!=null) {
				int n = 0;

				for(MethodTable table: ctab_sorted) {
					table.updateOffset(index_table);
					if (table.hasDispatchTable()) n++;
				}

				int bits = (index_table.size()*8) / n;

				coder.global_header_add("extern ");
				coder.global_header_add(opts.declareConst());
				if (version.equals("16")) {
					coder.global_header_add("unsigned short index");
				} else {
					coder.global_header_add("unsigned char index");
				}
				coder.global_header_add(version);
				coder.global_header_add("_table[];\n");

				coder.global_add("/* index_table size:"+index_table.size());
				coder.global_add(" ctables:"+n+" bits/table:"+bits);
				coder.global_add(" */\n");
				coder.global_add(opts.declareConst());
				if (version.equals("16")) {
					coder.global_add("unsigned short index");
				} else {
					coder.global_add("unsigned char index");
				}
				coder.global_add(version);
				coder.global_add("_table[] = {");
				for (int i=0; i < index_table.size(); i++) {
					if ((i%4)==0) coder.global_add("\n/* "+i+" */ \t");
					int ndx = index_table.get(i);
					if (ndx>=0 && ndx<10) coder.global_add(" ");
					coder.global_add(ndx);
					coder.global_add(",");

				}
				coder.global_add("\n};\n");
			}
		}

	}

	public void translate(Coder coder) throws CompileException {

		if (opts.hasOption("vt_switch")) {
			for(MethodTable table: mtables.values()) {
				table.emit_switch(coder);
			}
		} else if (opts.hasOption("vt_comp")) {
			coder.global_header_add("extern ");
			coder.global_header_add(opts.declareConst());
			coder.global_header_add("code_t dispatch_table[];\n");

			coder.global_add(opts.declareConst());
			coder.global_add("code_t dispatch_table[] = {\n");
			Collection<MethodTable> ctables   = new ArrayList<MethodTable>();
			Collection<MethodTable> ctables16 = new ArrayList<MethodTable>();
			int offset = 0;

			for(MethodTable table: mtables.values()) {
				if (table.isCompressed()) {
					if (table.is16BitIndex()) {
						ctables16.add(table);
					} else {
						ctables.add(table);
					}
				} else {
					offset = table.emit_table(coder, offset);
				}
			}
			coder.global_add("};\n");

			if (ctables.size()>0) emitIndexTable(coder, ctables, "");
			if (ctables16.size()>0) emitIndexTable(coder, ctables16, "16");

		} else {
			coder.global_header_add("extern ");
			coder.global_header_add(opts.declareConst());
			coder.global_header_add("code_t dispatch_table[];\n");

			coder.global_add(opts.declareConst());
			coder.global_add("code_t dispatch_table[] = {\n");

			int offset = 0;
			for(MethodTable table: mtables.values()) {
				offset = table.emit_table(coder, offset);
			}
			coder.global_add("};\n");
		}

		coder.global_header_add("#ifndef KESO_NEED_FINALIZE\n");
		coder.global_header_add("#define KESO_INVOKE_FINALIZE(_self_)(_self_)\n");
		coder.global_header_add("#endif\n");
	}
}
