/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler;

/**
 * Analysis/transformation pass base class.
 * This class represents a pass that performs some sort of analysis or
 * transformation on the intermediate code.
 */
public abstract class Pass {

	protected PassManager pm;

	/**
	 * Returns the name of the pass.
	 * @return Pass name.
	 */
	public abstract String getName();

	public boolean isEnabled(BuilderOptions opts) {
		return true; // Enabled by default
	}

	/**
	 * Fills the passed PassInfo object with information about the pass.
	 * If this method is not overridden, the default values (the pass is not an
	 * analysis and has no dependencies) are used.
	 * @param info Pass info.
	 * @param opts Builder options.
	 */
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		// Not an analysis and and no dependencies by default
	}

	public final void setPassManager(PassManager pm) {
		this.pm = pm;
	}

	/**
	 * Initializes the pass.
	 * This method is called before the pass is executed. It can be overridden
	 * to do custom initializations. If <code>false</code> is returned, the
	 * PassManager will not run the pass.
	 * @return <code>true</code> if everything is normal, <code>false</code> if
	 *         the pass should not be executed after all.
	 */
	public boolean init() throws CompileException {
		return true;
	}

	/**
	 * Finalizes the pass.
	 * This method is called at the end of the pass execution. It can be
	 * overridden to do custom finalizations.
	 * @return <code>true</code> if the pass has made any modifications,
	 *         <code>false</code> if it has left the code untouched.
	 */
	public boolean finish() throws CompileException {
		return false;
	}

	/**
	 * Runs the pass for a specific repository.
	 * @param repository Class store.
	 * @return <code>true</code> if the pass has made any modifications,
	 *         <code>false</code> if it has left the code untouched.
	 * @throws CompileException
	 */
	public abstract boolean runOnRepository(ClassStore repository)
	throws CompileException;

}
