/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler;

import java.util.*;

/**
 * Pass information class.
 * This class holds information about a specific pass that can used by a
 * PassManager to determine the execution order of all registered passes.
 * A pass can require that other passes have to be executed before
 * ("dependency") or after ("anti-dependency") itself.
 * If the pass is marked as an analysis, it will be executed only if other
 * passes registered at the same PassManager explicitly depend on it.
 */
public final class PassInfo {

	private Pass    pass;
	private boolean isAnalysis;
	private boolean isDump;
	private Set<Class<? extends Pass>> deps
	= new HashSet<Class<? extends Pass>>();
	private Set<Class<? extends Pass>> antiDeps
	= new HashSet<Class<? extends Pass>>();
	private Set<Class<? extends Pass>> invalidated
	= new HashSet<Class<? extends Pass>>();
	private boolean doDump;
	private boolean doLoop;

	public PassInfo(Pass pass, BuilderOptions opts) {
		this.pass = pass;
		pass.getPassInfo(this, opts);
	}

	public Pass getPass() {
		return pass;
	}

	/**
	 * Sets or deletes the analysis flag.
	 * @param isAnalysis Analysis flag.
	 */
	public void setIsAnalysis(boolean isAnalysis) {
		this.isAnalysis = isAnalysis;
	}

	/**
	 * Returns the analysis flag.
	 * @return Analysis flag.
	 */
	public boolean isAnalysis() {
		return isAnalysis;
	}

	/**
	 * Sets or deletes the dump flag.
	 * @param isDump Dump flag.
	 */
	public void setIsDump(boolean isDump) {
		this.isDump = isDump;
	}

	/**
	 * Returns the dump flag.
	 * @return Dump flag.
	 */
	public boolean isDump() {
		return isDump;
	}

	/**
	 * Adds a pass to the set of passes that must be executed before the pass
	 * this PassInfo object is related to.
	 * The pass is characterized by the class that implements it.
	 * @param pass Dependency.
	 */
	public void addDependency(Class<? extends Pass> pass) {
		deps.add(pass);
	}

	/**
	 * Adds a pass to the set of passes that must be executed before the pass
	 * this PassInfo object is related to.
	 * The pass is characterized by the class that implements it.
	 * @param pass       Dependency.
	 * @param invalidate Flag indicating whether the results of the dependency
	 *                   are invalidated by the execution of this pass.
	 */
	public void addDependency(Class<? extends Pass> pass, boolean invalidate) {
		addDependency(pass);
		if (invalidate == true)
			invalidated.add(pass);
	}

	/**
	 * Returns the set of required passes.
	 * @return Dependency set.
	 */
	public Set<Class<? extends Pass>> getDependencies() {
		return deps;
	}

	/**
	 * Adds a pass to the set of passes that must be executed after the pass
	 * this PassInfo object is related to.
	 * The pass is characterized by the class that implements it.
	 * @param pass Dependency.
	 */
	public void addAntiDependency(Class<? extends Pass> pass) {
		antiDeps.add(pass);
	}

	/**
	 * Returns the set of anti-dependencies.
	 * @return Anti-dependency set.
	 */
	public Set<Class<? extends Pass>> getAntiDependencies() {
		return antiDeps;
	}

	/**
	 * Adds a pass to the set of passes whose results are invalidated by the
	 * pass this PassInfo object is related to.
	 * The pass is characterized by the class that implements it.
	 * @param pass Dependency.
	 */
	public void addInvalidated(Class<? extends Pass> pass) {
		invalidated.add(pass);
	}

	/**
	 * Returns the set of passes invalidated by this pass.
	 * @return Invalidated pass set.
	 */
	public Set<Class<? extends Pass>> getInvalidated() {
		return invalidated;
	}

	public void setDoDump(boolean doDump) {
		this.doDump = doDump;
	}

	public boolean getDoDump() {
		return doDump;
	}

	public void setDoLoop(boolean doLoop) {
		this.doLoop = doLoop;
	}

	public boolean getDoLoop() {
		return doLoop;
	}

}
