/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import java.io.*;
import java.util.*;
import java.util.Map.*;

import keso.compiler.dump.*;
import keso.compiler.imcode.*;
import keso.util.*;

/**
 * Pass manager class.
 * The pass manager is the component that manages the execution of a series of
 * passes. It utilizes the dependency information stored in the PassInfo object
 * of each pass to compute an execution plan. Cyclic dependencies will cause an
 * error.
 * Every PassManager is a Pass itself, allowing it to be registered at another
 * PassManager.
 */
public final class PassManager extends Pass {

	private final BuilderOptions opts;
	private String name = "Pass manager";
	private final Map<Class<? extends Pass>, PassInfo> passInstances
	= new HashMap<Class<? extends Pass>, PassInfo>();
	private final DirectedGraph<Class<? extends Pass>> executionGraph
	= new DirectedGraph<Class<? extends Pass>>();
	private final Set<Class<? extends Pass>> allRequired
	= new HashSet<Class<? extends Pass>>();
	private final Set<Class<? extends Pass>> available
	= new HashSet<Class<? extends Pass>>();

	public PassManager(BuilderOptions opts) {
		this.opts = opts;
	}

	public PassManager(BuilderOptions opts, String name) {
		this(opts);
		this.name = name;
	}

	public PassManager(BuilderOptions opts, String name, PassManager prevPm)
	throws CompileException {
		this(opts, name);
		for (PassInfo info : prevPm.passInstances.values()) {
			addPassInternal(info);
			Class<? extends Pass> passClass = info.getPass().getClass();
			if (prevPm.available.contains(passClass))
				available.add(passClass);
		}
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public boolean runOnRepository(ClassStore repository)
	throws CompileException {

		// Dump pass dependency graph if desired
		if (opts.hasOption("dump_pass_graph"))
			dumpDepGraph(opts.getOutputPath());

		// Perform topological sort
		List<Class<? extends Pass>> toRun;
		try {
			toRun = executionGraph.topSort(new Comparator<Class<? extends Pass>>() {
				@Override
				public int compare(Class<? extends Pass> o1,
						Class<? extends Pass> o2) {
					return o1.toString().compareTo(o2.toString());
				}
			});
		} catch (GraphException e) {
			String message = "Circular dependency in pass graph of " + name
					+ "!\n";
			Set<IMClass> circle = e.getInfo();
			message += "  Affected passes: " + circle;
			throw new CompileException(message);
		}

		// Run the passes in topological order
		boolean modified = false;
		for (Class<? extends Pass> passClass : toRun) {
			modified |= runPass(passClass, repository);
		}
		return modified;
	}

	/**
	 * Registers a pass.
	 * It is not possible to register multiple objects of the same class at the
	 * same pass manager.
	 * @param pass Pass to be registered.
	 */
	public void addPass(Pass pass) throws CompileException {

		// Add pass
		PassInfo info = new PassInfo(pass, opts);
		addPassInternal(info);

		// Update execution graph
		Class<? extends Pass> passClass = pass.getClass();
		executionGraph.addNode(passClass);
		for (Class<? extends Pass> dep : info.getDependencies()) {
			executionGraph.addEdge(dep, passClass);
		}
		for (Class<? extends Pass> antiDep : info.getAntiDependencies()) {
			executionGraph.addEdge(passClass, antiDep);
		}
	}

	/**
	 * Returns the instance of a specific Pass class. The pass must have been
	 * registered before.
	 * @param passClass Desired class.
	 * @return Pass object instantiating to the class.
	 */
	public <P extends Pass> P getPass(Class<P> passClass)
	throws CompileException {
		return passClass.cast(getInfoFor(passClass).getPass());
	}

	public void dumpDepGraph(String outPath) throws CompileException {
		String fileName = outPath + '/' + name.replace(' ', '_') + ".dot";
		try {
			DotWriter<Class<? extends Pass>> dw
			= new DotWriter<Class<? extends Pass>>(fileName);
			dw.initGraph("Passes");
			for (Entry<Class<? extends Pass>, PassInfo> entry
					: passInstances.entrySet()) {
				Class<? extends Pass> pass = entry.getKey();
				PassInfo              info = entry.getValue();
				dw.addNode(pass, pass.getSimpleName());
				for (Class<? extends Pass> dep : info.getDependencies()) {
					dw.addEdge(pass, dep);
				}
				for (Class<? extends Pass> antiDep
						: info.getAntiDependencies()) {
					dw.addEdge(antiDep, pass, "color=blue");
				}
				for (Class<? extends Pass> inv : info.getInvalidated()) {
					dw.addEdge(pass, inv, "color=red");
				}
			}
			dw.finalizeGraph();
		} catch (DotWriterException e) {
			throw new CompileException(e.getMessage());
		} catch (FileNotFoundException e) {
			throw new CompileException(e.getMessage());
		}
	}

	private void addPassInternal(PassInfo info) throws CompileException {

		// Sanity check (Yo Dawg, ...)
		Pass pass = info.getPass();
		if (pass == this)
			throw new CompileException(name + " cannot be added to its "
					+ "own list of passes");

		// Add pass to map of pass providers
		Class<? extends Pass> passClass = pass.getClass();
		passInstances.put(passClass, info);
		pass.setPassManager(this);

		// Update dependencies
		for (Class<? extends Pass> dep : info.getDependencies()) {
			allRequired.add(dep);
		}
	}

	private PassInfo getInfoFor(Class<? extends Pass> passClass)
	throws CompileException {
		PassInfo info = passInstances.get(passClass);
		if (info == null) {
			String message = "No pass registered in " + name + " for "
			+ passClass.getSimpleName()
			+ ", required by the following passes: [";
			Iterator<Class<? extends Pass>> it
			= executionGraph.getSuccessors(passClass).iterator();
			while (it.hasNext()) {
				message += it.next().getSimpleName();
				if (it.hasNext())
					message += ", ";
			}
			throw new CompileException(message + ']');
		}
		return info;
	}

	private boolean runPass(Class<? extends Pass> passClass,
			ClassStore repository) throws CompileException {

		// Skip the pass if it has been executed before and its results are
		// still available
		if (available.contains(passClass))
			return false;

		// Retrieve pass info; skip the pass if it is merely an analysis and
		// unused
		PassInfo info = getInfoFor(passClass);
		if (info.isAnalysis() && allRequired.contains(passClass) == false)
			return false;

		// Skip the pass if it is not enabled
		Pass           pass = info.getPass();
		BuilderOptions opts = repository.getOptions();
		if (pass.isEnabled(opts) == false)
			return false;

		boolean result;
		boolean modified = false;

		// If required, loop the pass until a fixed point has been reached
		do {

			// Run dependency passes
			// These passes will only be executed if necessary.
			for (Class<? extends Pass> dep : info.getDependencies()) {
				runPass(dep, repository);
			}

			// Execute the pass
			opts.phase("Executing pass: " + pass.getName());
			if (pass.init() == false)
				break;
			result    = pass.runOnRepository(repository) | pass.finish();
			modified |= result;

			// Remove the passes whose results have been invalidated from the
			// availability set; add the executed pass if it is not a dump pass
			if (result == true)
				available.removeAll(info.getInvalidated());
			if (info.isDump() == false)
				available.add(passClass);

		} while (info.getDoLoop() == true && result == true);

		// Dump the intermediate code if desired
		if (info.getDoDump() == true) {
			InterPassByteCodeDump bcd = getPass(InterPassByteCodeDump.class);
			bcd.setSuffix(passClass.getSimpleName());
			runPass(InterPassByteCodeDump.class, repository);
			bcd.setSuffix(null);
		}
		return modified;
	}

}
