/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import java.io.*;

import keso.compiler.analysis.*;
import keso.compiler.backend.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.*;
import keso.util.*;

public final class Translation extends Pass {

	@Override
	public String getName() {
		return "Translation";
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(CallGraphAnalysis.class);
		info.addDependency(ClinitAnalysis.class);
		info.addDependency(ClassTypeInfoComputation.class);
		info.addDependency(VTableComputation.class);
	}

	@Override
	public boolean runOnRepository(ClassStore repository)
	throws CompileException {

		BuilderOptions opts = repository.getOptions();
		File outPath = new File(opts.getOutputPath());
		if (outPath.exists() == false)
			outPath.mkdirs();

		// Get the coder for the selected architecture
		Coder coder;
		VTableComputation  vtc   = pm.getPass(VTableComputation.class);
		MethodTableFactory table = vtc.getTable();
		switch (opts.getTarget()) {
		case BuilderOptions.AVR_JOSEK_TARGET:
			coder = new CoderAVR_JOSEK(opts, table, repository);
			break;
		case BuilderOptions.CIAO_TARGET:
			coder = new CoderCiAO(opts, table, repository);
			break;
		case BuilderOptions.TRICORE_RTAOSEK_TARGET:
			coder = new CoderTricoreRTAOSEK(opts, table, repository);
			break;
		case BuilderOptions.X86LINUX_JOSEK_TARGET:
			coder = new CoderX86Linux_JOSEK(opts, table, repository);
			break;
		case BuilderOptions.POSIX_TRAMPOLINE_TARGET:
			coder = new CoderPOSIX_Trampoline(opts, table, repository);
			break;
		case BuilderOptions.AVR_TRAMPOLINE_TARGET:
			coder = new CoderAVR_Trampoline(opts, table, repository);
			break;
		case BuilderOptions.TRICORE_PROOSEK_TARGET:
		default:
			coder = new CoderTricore_ProOSEK(opts, table, repository);
		}

		// If enabled, pass the flow information to the backend
		FlowAnalysis fa = pm.getPass(FlowAnalysis.class);
		if (fa.isEnabled(opts) && opts.hasOption("rtcheck_opt")) {
			FlowLatticeInfoStore flowInfoStore = fa.getFlowLatticeInfoStore();
			FlowLatticeInfo combinedInfo = flowInfoStore.getForAllDomains();
			coder.setCodeInfo(new ExtendedCodeInfo(combinedInfo));
		}

		// Translate all classes and interfaces
		IntegerHashtable<ClassTypeInfo> classTypeInfo
		= pm.getPass(ClassTypeInfoComputation.class).getClassTypeInfo();
		repository.translate(coder, table, classTypeInfo);

		// Create domains
		IMDomain.translateDomains(coder, opts);

		// Write startup code
		coder.writeMainFile();

		// Clean up
		coder.close();
		opts.vverbose("Non-virtual calls: " + IMInvoke.nonVirtualCalls
				+ "\t(initially: " + IMInvoke.nonVirtualCallsInitial + ")");
		opts.vverbose("Virtual calls:     " + IMInvoke.virtualCalls
				+ "\t(initially: " + IMInvoke.virtualCallsInitial + ")");
		return false;
	}

}
