/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler;

import keso.compiler.imcode.*;

final public class VirtualOperandStack {

	private IMNode[] opr_stack;
	private int ptr;
	private IMMethod method;
	private BuilderOptions opts;

	public VirtualOperandStack(IMMethod method) {
		this(method, 10);
	}

	public VirtualOperandStack(IMMethod method, int size) {
		this.method = method;
		this.opts   = method.getOptions();
		opr_stack   = new IMNode[size+1];
		ptr=0;
	}

	/**
	 * push an operand on top of the stack.
	 */
	public void push(IMNode opr) throws CompileException {
		if (ptr>=opr_stack.length) extend(10);
		if (opr==null) throw new CompileException("store null pointer");
		opr_stack[ptr++]=opr;
	}

	/**
	 * fetch the top operand from stack.
	 */
	public IMNode pop() throws CompileException {
		if (ptr==0) throw new CompileException("operand stack underrun");
		return opr_stack[--ptr].clear_link();
	}

	/**
	 * Looks at an object on the stack without removing it.
	 * @param i Offset from the top of the stack.
	 * @return Object at the given offset.
	 * @throws CompileException
	 */
	public IMNode peek(int i) throws CompileException {
		int index = ptr - 1 - i;
		if (index < 0 || index >= ptr)
			throw new CompileException("Invalid peek index");
		return opr_stack[index];
	}

	/**
	 * clear the operand stack.
	 */
	public void clear() { ptr=0; }

	/**
	 * write the content of the current operand stack to local variables
	 * needed because of possible side effects.
	 *
	 * excluded objects are constants, read variables and read return address
	 */
	public void store(int bcPosition, IMNode extra_ops) throws CompileException {

		for (int i=0;i<ptr;i++) {
			IMNode curr = opr_stack[i];
			if (curr instanceof IMConstant) continue;
			if (curr.isVariable()) continue;
			if (curr instanceof IMPopReturnAddr) continue;
			if (curr.isMemoryType()) continue;
			if (curr.isStackVariable(i)) continue;

			IMStoreLocalVariable store_var = method.createStoreStackVariable(i,curr,bcPosition);
			opr_stack[i]=store_var.getReadOperation();

			extra_ops = extra_ops.append(store_var);
		}
	}

	/**
	 * write the content of the current operand stack to local variables
	 * needed because of possible side effects.
	 *
	 * excluded objects are constants and stack variables it self
	 */
	public void store2(int bcPosition, IMNode extra_ops, IMSlot var) throws CompileException {

		for (int i=0;i<ptr;i++) {
			IMNode curr = opr_stack[i];
			if (curr instanceof IMConstant) continue;
			// DON'T DO THIS! THIS WILL BREAK THE EXECUTION ORDER!
			// if (curr.isVariable() && curr.getIMSlot().equals(var)) continue;
			if (curr.isStackVariable(i)) continue;

			IMStoreLocalVariable store_var = method.createStoreStackVariable(i,curr,bcPosition);
			opr_stack[i]=store_var.getReadOperation();

			extra_ops = extra_ops.append(store_var);
		}
	}

	/**
	 * write the content of the current operand stack to local variables
	 * needed because of possible side effects.
	 *
	 * excluded objects are constants, read variables, read return address,
	 * getfield, getstatic
	 *
	 * The store is only executed if one of the upper stack members are
	 * a method invocation.
	 */
	public void store3(int bcPosition, IMNode extra_ops, int deep) throws CompileException {

		if (deep>0) {
			boolean flag = true;
			for (int i=0;i<deep && i<ptr;i++) {
				if (opr_stack[ptr-(i+1)] instanceof IMInvoke) flag=false;
			}
			if (flag) return;
		}

		for (int i=0;i<ptr;i++) {
			IMNode curr = opr_stack[i];
			if (curr instanceof IMConstant) continue;
			if (curr.isVariable()) continue;
			if (curr instanceof IMPopReturnAddr) continue;
			if (curr instanceof IMGetStatic) continue;
			if (curr instanceof IMGetField) continue;
			if (curr.isMemoryType()) continue;
			if (curr.isStackVariable(i)) continue;

			IMStoreLocalVariable store_var = method.createStoreStackVariable(i,curr,bcPosition);
			opr_stack[i]=store_var.getReadOperation();

			extra_ops = extra_ops.append(store_var);
		}
	}

	public void init(IMNode[] stack) throws CompileException {
		ptr=0;
		if (stack==null) return;
		for (int i=0;i<stack.length;i++) if (stack[i]!=null) push(stack[i].copy(null));
	}

	public IMNode[] leave(IMNode extra_ops) throws CompileException {
		if (ptr==0) return null;
		IMNode[] leaveStack = new IMNode[ptr];
		for (int i=0;i<ptr;i++) {
			IMNode curr = opr_stack[i];

			if (curr.isStackVariable(i)) {
				leaveStack[i]=curr;
				continue;
			}

			if (curr instanceof IMPopReturnAddr) {
				leaveStack[i] = curr;
			} else {
				IMStoreLocalVariable store_var = method.createStoreStackVariable(i,curr,-1);
				leaveStack[i] = store_var.getReadOperation();
				extra_ops = extra_ops.insert(store_var);
			}
		}
		ptr=0;
		return leaveStack;
	}

	/**
	 * Duplicates the value(s) at the top of the stack and inserts them further
	 * down.
	 * @param count    Number of stack values to be duplicated.
	 * @param offset   Number of slots between the originals and the duplicates.
	 * @param bcPos    Bytecode position.
	 * @param extraOps List of extra operations emitted due to side effects
	 *                 (output parameter).
	 * @throws CompileException
	 */
	public void dupX(int count, int offset, int bcPos, IMNode extraOps)
	throws CompileException {
		store(bcPos, extraOps);
		IMNode[] copies = new IMNode[count];
		IMNode[] values = new IMNode[count + offset];
		int i = 0;
		while (i < count) {
			copies[i] = copyDown(offset + 1, extraOps);
			values[i] = pop();
			++i;
		}
		while (i < count + offset) {
			values[i++] = pop();
		}
		for (int j = copies.length - 1; j >= 0; --j) {
			push(copies[j]);
		}
		while (--i >= 0) {
			push(values[i]);
		}
	}

	private void extend(int value) {
		opts.vwarn("virtual operand stack overrun! ("+opr_stack.length+" exceeded)");
		IMNode[] new_stack = new IMNode[opr_stack.length+value];
		System.arraycopy(opr_stack,0,new_stack,0,opr_stack.length);
		opr_stack =  new_stack;
		new_stack = null;
	}

	private IMNode copyDown(int n, IMNode extraOps) throws CompileException {
		IMNode top = opr_stack[ptr - 1].copy(null);
		if (top.isVariable() == false) // Very improbable
			return top;
		IMStoreLocalVariable store
		= method.createStoreStackVariable(ptr - n, top, top.getBCPosition());

		// Insert the store at the end (this is important!) of the list of extra
		// operations
		for (;;) {
			IMNode next = extraOps.next();
			if (next == null)
				break;
			extraOps = next;
		}
		extraOps.insert(store);
		return store.getReadOperation();
	}

	public String stackToString() {
		String ret = "";

		for (int i=0;i<ptr;i++) {
			ret += " " + opr_stack[i].toReadableString();
		}

		return ret;
	}

	@Override
	public String toString() {
		return stackToString();
	}
}
