/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import keso.compiler.*;
import keso.compiler.config.*;
import keso.compiler.config.task.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.FieldAccessReset;
import keso.util.DirectedGraph;

public final class CallGraphAnalysis extends MethodPass implements IMMethodVisitor<Boolean> {
	public static DirectedGraph<IMMethod> graph;

	/**
	 * The call graph as a directed graph, invertible, etc. Has a root node {@code null} with edges to all entry points
	 */
	private DirectedGraph<IMMethod>	callGraph;

	public CallGraphAnalysis() {
		callGraph = new DirectedGraph<IMMethod>();
		// null identifies the root node
		callGraph.addNode(null);
	}

	public DirectedGraph<IMMethod> getCallGraph() {
		return callGraph;
	}

	@Override
	public String getName() {
		return "Call-graph analysis";
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {

		info.setIsAnalysis(true);

		// ConstantFolding seems to require FieldAccessReset to be executed
		// before CallGraphAnalysis, otherwise it won't do all optimizations.
		// TODO: Verify this!
		info.addDependency(FieldAccessReset.class);
		info.addDependency(ClassTypeInfoComputation.class);
	}

	@Override
	public boolean runOnRepository(ClassStore repository) throws CompileException {
		graph = callGraph;

		IMClass[]      imClasses = repository.getAllClasses();
		BuilderOptions opts      = repository.getOptions();
		opts.verbose("####### analyse call graph ");

		opts.verbose("##### reset call graph ");
		repository.resetCallGraph();
		for (IMClass imClass : imClasses) {
			imClass.resetCallGraph();
		}

		// Call graph for class constructors
		opts.vverbose("### analyse call graph <clinit>()V");
		for (IMClass imClass : imClasses) {
			runOnClass(imClass);
		}

		// Call graph for constructor of Thread
		opts.vverbose("### analyse call graph Thread.<init>()V");
		IMClass taskClass = repository.getClass("java/lang/Thread");
		if (taskClass != null) {
			IMMethod constructor = taskClass.getMethod("<init>()V");
			if (constructor != null) {
				constructor.calledFrom(null, null);
				callGraph.addEdge(null, constructor);
			}
		}

		// Call graph for ISRs
		opts.vverbose("### analyse call graph ISR");
		for (ISRDefinition isr : opts.getJavaISRs()) {
			IMClass handlerClass = repository.getClass(isr.getHandlerClass());
			IMMethod handler = handlerClass.getMethod(isr.getHandlerMethod());
			handler.calledFrom(null, null);
			callGraph.addEdge(null, handler);
		}

		// Call graph for hooks
		opts.vverbose("### analyse call graph system hooks");
		SystemDefinition sysDef = opts.getSysDef();
		OSDefinition os = sysDef.getOSDef();
		if (os != null) {
			Object[] allHooks = os.getAllHooks();
			ComplexBoolAttribute[] hooks = (ComplexBoolAttribute[]) allHooks[0];
			for (ComplexBoolAttribute hook : hooks) {
				if (hook == null || hook.setting == false
						|| hook.getAttribute("HookMethod") == null)
					continue;
				String hc = hook.getAttribute("HookClass").valueString();
				String hm = hook.getAttribute("HookMethod").valueString();
				IMMethod method = repository.getClass(hc).getMethod(hm);
				method.calledFrom(null, null);
				callGraph.addEdge(null, method);
			}
		}

		ComplexAttribute<?> stack = sysDef.getStack();
		if (stack != null) {
			String sc = stack.getAttribute("StackChecks").valueString();
			if (sc != null && sc.equalsIgnoreCase("true")) {
				String hc = stack.getAttribute("HandlerClass").valueString();
				String hm = stack.getAttribute("HandlerMethod").valueString();
				IMMethod method = repository.getClass(hc).getMethod(hm);
				method.calledFrom(null, null);
				callGraph.addEdge(null, method);
			}
		}

		// Call graph for services
		// FIXME
		//IMPortalService services = opts.getPortalServices();
		//if (services!=null) services.calledFromSystem(repository);

		// Call graph for tasks
		for (JavaTask task : opts.getSysJavaTasks()) {
			taskClass = repository.getClass(task.getMainClassName());
			IMMethod constructor = taskClass.getMethod("<init>()V");
			constructor.calledFrom(null, null);
			callGraph.addEdge(null, constructor);
			IMMethod launch = taskClass.getMethod(task.getMainMethodName());
			if (launch == null) {
				opts.critical("main method " + task.getMainMethodName() + " not found!");
			} else {
				launch.calledFrom(null, null);
				callGraph.addEdge(null, launch);
			}
		}

		graph = null;

		opts.verbose("##### analyse call graph done");
		return false;
	}

	@Override
	public boolean runOnClass(IMClass imClass) throws CompileException {
		BuilderOptions opts = imClass.getOptions();
		opts.vverbose("#### analyse call graph " + imClass.getAlias());
		IMMethod[] methods = imClass.getMethods();
		if (methods == null)
			return false;
		for (IMMethod method : methods) {
			opts.vverbose("### " + method.getMethodNameAndType() + " " + method.getAlias());
			if (method.isClassInit()) {
				method.calledFrom(null, null);
				callGraph.addEdge(null, method);
			}
			runOnMethod(method);
		}
		return false;
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {
		return method.accept(this);
	}

	@Override
	public Boolean visit(IMMethod method) throws CompileException {

		if (method.hasCode() == false)
			return false;

		IMClass            imClass = method.getIMClass();
		BuilderOptions     opts    = method.getOptions();
		IMCallGraphVisitor visitor = new IMCallGraphVisitor(opts);
		try {
			for (IMBasicBlock bb : method) {
				bb.visitNodes(visitor);
			}
			method.setIsAnalyzed();
		} catch (CompileException e) {
			IMBasicBlock currentBb = method.getCurrentBasicBlock();
			System.err.println("======================================================");
			System.err.println("CompileException in " + imClass.getSourceFile());
			System.err.println(method.getMethodNameAndType());
			if (currentBb != null)
				System.err.println(currentBb.toLabel());
			System.err.println("======================================================");
			throw e;
		} catch (RuntimeException e) {
			IMBasicBlock currentBb = method.getCurrentBasicBlock();
			System.err.println("======================================================");
			System.err.println("RuntimeException in " + imClass.getSourceFile());
			System.err.println(method.getMethodNameAndType());
			if (currentBb != null)
				System.err.println(currentBb.toLabel());
			System.err.println("======================================================");
			throw e;
		}

		return false;
	}

	@Override
	public Boolean visit(IMLocalPortalMethod method) throws CompileException {
		ClassStore repository = method.getClassStore();
		ServiceDefinition service = method.getImportDefinition().getService();
		IMClass serviceClass = service.getServiceClass(repository);
		IMMethod callee = serviceClass.getMethod(method.getMethodNameAndType());
		callee.calledFrom(method, null);
		callGraph.addEdge(method, callee);
		if (method.getMethodFrame().hasReferenceArguments()) {
			IMClass nonCopyable = repository.getClass("keso/core/NonCopyable");
			if (nonCopyable != null)
				repository.needInstanceOf(nonCopyable);
		}
		return false;
	}

	@Override
	public Boolean visit(IMMethodNotFound method) {
		return false;
	}

	@Override
	public Boolean visit(IMPortalConstructor method) throws CompileException {

		ClassStore       repository  = method.getClassStore();
		ImportDefinition impDef      = method.getImportDefinition();
		IMClass          driverClass = repository.getClass(impDef.getDriver());
		recurse(driverClass.getMethod("<init>()V"), method);

		// Call all specified bean methods
		NetworkDefinition netDef = impDef.getNetwork();
		for (String m : netDef.getDriverInitMethods()) {
			IMMethod callee = driverClass.getMethod(m);
			recurse(callee, method);
		}

		IMClass protoClass = repository.getClass(netDef.getProtocolName());
		String ctor = "<init>(L" + netDef.getDriverInterfaceName() + ";IIIBB)V";
		recurse(protoClass.getMethod(ctor), method);

		// Allocate a buffer for sending and receiving
		recurse(protoClass.getMethod("allocReceiveBuffer(ISI)Z"), method);
		String atb = "allocTransmitBuffer(SI)Lkeso/core/Memory;";
		recurse(protoClass.getMethod(atb), method);
		return false;
	}

	@Override
	public Boolean visit(IMRemotePortalMethod method) throws CompileException {

		ImportDefinition impDef = method.getImportDefinition();
		String   protoName  = impDef.getNetwork().getProtocolName();
		IMClass  protoClass = method.getClassStore().getClass(protoName);
		recurse(protoClass.getMethod("write(ISI)Z"), method);
		recurse(protoClass.getMethod("read(I)Lkeso/core/Memory;"), method);
		return false;
	}

	private void recurse(IMMethod callee, IMMethod caller)
	throws CompileException {
		callee.calledFrom(caller, null);
		callGraph.addEdge(caller, callee);
		runOnMethod(callee);
	}

}
