/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class ClassTypeInfoComputation extends Pass {

	private IntegerHashtable<ClassTypeInfo> classTypeInfo;

	@Override
	public String getName() {
		return "Class type info computation";
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsAnalysis(true);
	}

	@Override
	public boolean runOnRepository(ClassStore repository)
			throws CompileException {

		classTypeInfo = new IntegerHashtable<ClassTypeInfo>(256);
		IMClass root = repository.getRootClass();
		if (root == null)
			throw new CompileException("class java.lang.Object not found!");

		int typeEnd = computeSubClassTypeInfo1(repository, 1, root);
		ClassTypeInfo cInfo = new ClassTypeInfo(root, 1, typeEnd - 1);
		root.setClassTypeInfo(cInfo);
		classTypeInfo.put(1, cInfo);
		return false;
	}

	public IntegerHashtable<ClassTypeInfo> getClassTypeInfo() {
		return classTypeInfo;
	}

	private int computeSubClassTypeInfo1(ClassStore repository, int start,
			IMClass current) throws CompileException {

		/*
		 * Wir sortieren zuerst die Klassen auf der 1. Ebene.
		 *
		 * 1. Werden alle Array-Klassen nebeneinander angeordnet
		 *    damit wir auf einfache Art Array-Klassen von anderen
		 *    Klassen unterscheiden koennen.
		 *
		 * 2. Werden alle Klassen mit Interfaces zusammengefasst
		 *    damit koennen wir Speicherplatz in den dispatch-Tabellen
		 *    einsparen und auch im Klassenspeicher, wenn die Typeinformation
		 *    in einem eigenen Array gespeichert wird. -X:iface_store
		 */

		// We sort the classes first
		IMClass[] subClassesL1     = current.getSubClasses();
		IMClass[] subClassesSorted = new IMClass[subClassesL1.length];
		IMClass[] subClassesRest   = new IMClass[subClassesL1.length];

		int rangeStart = start + 1;
		int arrayStart = rangeStart;
		int arrayEnd   = rangeStart;
		int pos        = 0;
		int posRest    = 0;

		// All array classes first
		for (IMClass subClass : subClassesL1) {
			if (subClass.getClassName().charAt(0)=='[') {
				subClassesSorted[pos++] = subClass;
				++arrayEnd;
			} else {
				subClassesRest[posRest++] = subClass;
			}
		}

		// All classes with interfaces next
		IMClass[] subClassesRest1 = new IMClass[posRest];
		int posRest1 = 0;

		if (repository.getOptions().hasOption("vt_fast_order") == false) {

			/*
			 * TODO:
			 * Dieser Algorithmus gruppiert die Klassen oder Klassenbaeume
			 * auf der ersten Ebene nach dem erst-besten Interface. Besser
			 * waere eine geschickte Auswahl, z. B. Interfaces nach der Anzahl
			 * der Methoden bevorzugen.
			 */
			Multimap<String, IMClass> ifClasses
			= new HashMultimap<String, IMClass>();
			for (int i = 0; i < posRest; ++i) {
				IMClass subClass = subClassesRest[i];
				if (subClass.isInterface()) {
					subClassesRest1[posRest1++] = subClass;
				} else {
					String[] ifaces = collectInterfacesInSubClasses(subClass);
					if (ifaces != null)
						ifClasses.add(ifaces[0], subClass);
					else
						subClassesRest1[posRest1++] = subClass;
				}
			}
			for (String ifKey : ifClasses.keySet()) {
				for (IMClass iface : ifClasses.get(ifKey)) {
					subClassesSorted[pos++] = iface;
				}
			}

		} else {

			/*
			 * Dieser Algorithmus gruppiert die Klassen auf der ersten
			 * Ebene nach einem ganz dummen Verfahren. Ist jedoch besser als
			 * nichts.
			 */
			for (int i = 0; i < posRest; ++i) {
				IMClass subClass = subClassesRest[i];
				if (subClass.isInterface()) {
					subClassesRest1[posRest1++] = subClass;
				} else {
					String[] ifaces = subClass.getInterfaceNames();
					if (ifaces != null)
						subClassesSorted[pos++] = subClass;
					else
						subClassesRest1[posRest1++] = subClass;
				}
			}
		}

		subClassesRest = null;

		// All classes without interfaces next
		for (int i = 0; i < posRest1; ++i) {
			subClassesSorted[pos++] = subClassesRest1[i];
		}

		// Nun kommt die eigentliche Vergabe der globalen Klassen-IDs
		for (IMClass subClass : subClassesSorted) {
			int rangeEnd = computeSubClassInfo2(rangeStart, subClass);
			ClassTypeInfo cInfo = new ClassTypeInfo(subClass, rangeStart, rangeEnd - 1);
			subClass.setClassTypeInfo(cInfo);
			classTypeInfo.put(rangeStart, cInfo);
			rangeStart = rangeEnd;
		}

		repository.setArrayTypeRange(arrayStart, arrayEnd);
		return rangeStart;
	}

	private String[] collectInterfacesInSubClasses(IMClass imClass)
	throws CompileException {

		// TODO: Find all interfaces, not only the first once
		String[] ifaces = imClass.getInterfaceNames();
		if (ifaces != null)
			return ifaces;

		IMClass[] subClasses = imClass.getSubClasses();
		if (subClasses == null)
			return null;
		for (IMClass subClass : subClasses) {
			ifaces = collectInterfacesInSubClasses(subClass);
			if (ifaces != null)
				return ifaces;
		}
		return null;
	}

	private int computeSubClassInfo2(int start, IMClass current)
	throws CompileException {

		IMClass[] subClasses = current.getSubClasses();
		int       rangeStart = start + 1;
		if (subClasses == null)
			return rangeStart;

		for (IMClass subClass : subClasses) {
			int rangeEnd = computeSubClassInfo2(rangeStart, subClass);
			ClassTypeInfo cInfo
			= new ClassTypeInfo(subClass, rangeStart, rangeEnd - 1);
			subClass.setClassTypeInfo(cInfo);
			classTypeInfo.put(rangeStart, cInfo);
			rangeStart = rangeEnd;
		}
		return rangeStart;
	}

}
