/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.*;

/**
 * Class-initialization analysis pass.
 * This pass analyzes the dependencies between class-initializer methods. If a
 * circular dependency is detected, the compilation process is aborted.
 */
public final class ClinitAnalysis extends DomainPass
implements IMMethodVisitor<Void> {

	private ReachabilityData currData;
	private Deque<IMClass> stack = new ArrayDeque<IMClass>();
	private Set<IMMethod> visitedMethods = new HashSet<IMMethod>();

	@Override
	public String getName() {
		return "Class-initialization analysis";
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsAnalysis(true);
		info.addDependency(SSADeconstruction.class);
	}

	@Override
	public boolean runOnDomain(IMDomain domain) throws CompileException {
		currData = domain.getReachabilityData();
		currData.clearClinitGraph();
		visitedMethods.clear();
		for (IMClass imClass : currData.getClasses()) {
			addClass(imClass);
		}
		return false;
	}

	@Override
	public Void visit(final IMMethod method) throws CompileException {

		// Skip the method if it has already been visited
		if (visitedMethods.contains(method))
			return null;
		visitedMethods.add(method);

		// If the method is a class initializer, push it onto the stack
		IMClass imClass  = method.getIMClass();
		boolean isClinit = method.isClassInit();
		if (isClinit)
			stack.addFirst(imClass);

		// Recursively collect the classes for field reads/writes and method
		// invocations
		for (IMBasicBlock bb : method) {

			bb.visitNodes(new IMBasicVisitor() {

				@Override
				public void visit(IMNode node) throws CompileException {
					IMField field = null;
					if (node.isFieldAccess())
						field = ((IMFieldRead) node).getField();
					else if (node instanceof IMFieldWrite)
						field = ((IMFieldWrite) node).getField();
					if (field != null) {
						addClass(field.getType());
					} else if (node instanceof IMInvoke) {
						for (IMMethod candidate
								: ((IMInvoke) node).getCandidates()) {
							addClass(candidate.getIMClass());
							if (candidate.isClassInit() == false)
								candidate.accept(ClinitAnalysis.this);
						}
					}
				}
			});
		}

		// Pop the class initializer from the stack again
		if (isClinit) {
			currData.getClinitGraph().addNode(imClass);
			stack.removeFirst();
		}
		return null;
	}

	@Override
	public Void visit(IMLocalPortalMethod method) throws CompileException {
		return null;
	}

	@Override
	public Void visit(IMMethodNotFound method) throws CompileException {
		return null;
	}

	@Override
	public Void visit(IMPortalConstructor method) throws CompileException {
		return null;
	}

	@Override
	public Void visit(IMRemotePortalMethod method) throws CompileException {
		return null;
	}

	private void addClass(IMClass imClass)
	throws CompileException {
		IMMethod clinit = imClass.getMethod("<clinit>()V");
		if (clinit == null)
			return;
		IMClass user = stack.peekFirst();
		if (user != null && user != imClass)
			currData.addClinitDependency(imClass, user);
		clinit.accept(this);
	}

}
