/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/**
 * Escape Analysis according to "Stack Allocation and Synchronization Optimizations for Java Using Escape Analysis" by
 * Choi, Gupta, Serrano, Sreedhar and Midkiff. DOI 10.1145/945885.945892
 *
 * @author Clemens Lang <sicslang@stud.cs.fau.de>
 */
package keso.compiler.analysis;

import java.io.FileNotFoundException;
import java.util.*;

import keso.compiler.BuilderOptions;
import keso.compiler.ClassStore;
import keso.compiler.CompileException;
import keso.compiler.MethodPass;
import keso.compiler.PassInfo;
import keso.compiler.analysis.connectiongraph.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.SSAConstruction;
import keso.compiler.transform.SSADeconstruction;
import keso.compiler.transform.VTableComputation;
import keso.util.DirectedGraph;
import keso.util.DotWriterException;
import keso.util.HashMultimap;
import keso.util.HashUniqueMultimap;
import keso.util.Multimap;
import keso.util.SplitKeyHashMultimap;
import keso.util.SplitKeyMultimap;

interface CallVisitor {
	public boolean process(Multimap<CGReferenceNode, CGReferenceNode> workList, Multimap<CGReferenceNode, CGReferenceNode> visited,
		SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, IMMethod caller, IMMethod callee, CGReferenceNode calleeNode,
		CGReferenceNode callerNode);
}

/**
 * Escape Analysis class for the JINO Java-to-C compiler
 */
public class ExtendedEscapeAnalysis extends MethodPass {
	private class EEAVisitor extends IMTypedVisitorSimpleAdapter {
		/**
		 * Map mapping {@code IMNode} instances to the {@code CGNode} instances of their results. Consider this: An
		 * operation represented by an {@code IMNode} pushes a reference to the "stack"; this reference is later used as
		 * an operand by a different {@code IMNode}. Using this map, the visitor visiting the second node can access the
		 * connection graph representation of the result of the first node. I.e. {@code IMNew} pushes a
		 * {@code CGObjectNode} to the stack, {@code IMAReadLocalVariable} pushes a {@code CGReferenceNode}.
		 */
		private final Map<IMNode, CGNode>		nodeToCGMap;

		/**
		 * Array of callee-side actual reference nodes for formal parameters and the return value (if any). In this
		 * array, index 0 always references the return value (and is null if this method doesn't return a reference).
		 * For the meaning of other indexes in this array, see the description of
		 * {@code ExtendedEscapeAnalysis.methodToCalleeActualsMap}.
		 */
		private final CGActualReferenceNode[]	calleeActuals;

		/**
		 * List holding references to all caller-side actual reference nodes and global reference nodes used in this
		 * method. This list is needed, because the nodes in it typically only have edges into the rest of the graph but
		 * can't be reached from the graph, making them must-visits for graph dumps or complete graph traversals.
		 */
		private final List<CGReferenceNode>		callerActualsAndGlobals;

		/**
		 * The method currently being analyzed.
		 */
		private IMMethod						currentMethod;

		/**
		 * Constructs a new visitor to be used for the method specified by {@code method}. Since this internally
		 * initializes the {@code nodeToCGMap} field used to store references to nodes used in this method the returned
		 * visitor may only be used on this method, although this won't be enforced. Be advised that you might see
		 * strange results (especially in your graph dumps), if you ignore this.
		 *
		 * @param method
		 *            The method that's going to be processed by the new visitor
		 * @throws CompileException
		 */
		public EEAVisitor(final IMMethod method) throws CompileException {
			// initialize nodeToCGMap
			Map<IMNode, CGNode> nodeToCGMap = cg.methodToCGMap.get(method);
			if (nodeToCGMap == null) {
				cg.methodToCGMap.put(method, nodeToCGMap = new HashMap<IMNode, CGNode>());
			}
			this.nodeToCGMap = nodeToCGMap;

			this.currentMethod = method;

			// initialize calleeActuals
			CGActualReferenceNode[] actuals = cg.methodToCalleeActualsMap.get(method);
			if (actuals == null) {

				// create the actual reference nodes for this method's parameters and return value
				IMSlot[] arguments = method.getMethodFrame().getOriginalArguments();
				cg.methodToCalleeActualsMap.put(method, actuals = new CGActualReferenceNode[arguments.length + 1]);

				if (method.isReference()) {
					// create actual (phantom) reference node for return value, if necessary
					actuals[0] = CGActualReferenceNode.getReturnReferenceNode(method);
				}

				for (int i = 0; i < arguments.length; ++i) {
					// we only care about references
					if (arguments[i].isReference()) {
						// create an actual (phantom) reference node ...
						actuals[i + 1] = CGActualReferenceNode.getActualReferenceNode(arguments[i]);
						// ... and a local reference node (if it is used at all) ...
						if (arguments[i].hasNoUses()) {
							actuals[i + 1].setStale(true);
						} else {
							/*
							 * this is quite the hack; CGLocalReferenceNode.getLocalReferenceNode(IMNode) expects an
							 * IMNode, but we only have an IMSlot here. Now, we could probably hack CGLocalReferenceNode
							 * to accept an IMSlot here, too, but since adding the edge is only interesting if it's
							 * actually being used and we _do_ have a list of uses available from IMSlot we can just
							 * grab the first use and use that instead, effectively creating the CGLocalReferenceNode
							 * for this local reference ahead of usage time (we would eventually be creating this node
							 * later anyway).
							 */
							IMNode usingStatement = arguments[i].getUses().iterator().next();
							CGLocalReferenceNode local = CGLocalReferenceNode.getLocalReferenceNode(usingStatement);
							// ... and add a deferred edge from the local reference node to the actual (phantom)
							// reference node
							actuals[i + 1].attach(local);
							// create phantom nodes for actual phantom nodes ahead of time
							CGObjectPhantomNode.getPhantomNodeForReference(actuals[i + 1]);
						}
					}
				}
			}
			calleeActuals = actuals;

			// initialize callerActualsAndGlobals
			List<CGReferenceNode> callerActualsAndGlobals = cg.methodToCallerActualsAndGlobalsMap.get(method);
			if (callerActualsAndGlobals == null) {
				cg.methodToCallerActualsAndGlobalsMap.put(method, callerActualsAndGlobals = new LinkedList<CGReferenceNode>());
			}
			this.callerActualsAndGlobals = callerActualsAndGlobals;
		}

		/**
		 * Handle allocation of a new object, i.e. {@code p = new T();}
		 */
		@Override
		public void visit(IMAllocation n) {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			// create an object node and attach it to this IMNode
			nodeToCGMap.put(n, CGObjectNode.getObjectNodeForAllocation(n, currentMethod));
		}

		@Override
		public void visit(IMNew n) {
			visit((IMAllocation) n);
		}

		@Override
		public void visit(IMNewArray n) {
			visit((IMAllocation) n);
		}

		@Override
		public void visit(IMNewMultiArray n) {
			visit((IMAllocation) n);
		}

		@Override
		public void visit(IMNewObjArray n) {
			visit((IMAllocation) n);
		}

		/**
		 * Handle object constants, e.g. Strings
		 */
		@Override
		public void visit(IMAConstant n) {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			nodeToCGMap.put(n, CGConstantObjectNode.getObjNodeForConstant(n));
		}

		/**
		 * Handle memory constants (used for I/O-mapped stuff)
		 */
		@Override
		public void visit(IMAMemConstant n) throws CompileException {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			nodeToCGMap.put(n, CGConstantObjectNode.getObjNodeForConstant(n));
		}

		/**
		 * Handle memory constants (used for I/O-mapped stuff)
		 */
		@Override
		public void visit(IMAMTConstant n) throws CompileException {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			nodeToCGMap.put(n, CGGlobalReferenceNode.getGlobalPhantomReferenceNode(n, callerActualsAndGlobals));
		}

		/**
		 * Handle static memory constants (used for I/O-mapped stuff)
		 */
		@Override
		public void visit(IMAMTConstantStatic n) throws CompileException {
			visit((IMAMTConstant) n);
		}

		/**
		 * Handle NULL constant (why is this not a subclass of IMAConstant again?)
		 */
		@Override
		public void visit(IMNullConstant n) {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			nodeToCGMap.put(n, CGConstantObjectNode.getObjNodeForConstant(n));
		}

		/**
		 * Handle reading from local variables, i.e. usage of a local {@code p}
		 */
		@Override
		public void visit(IMAReadLocalVariable n) {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			nodeToCGMap.put(n, CGLocalReferenceNode.getLocalReferenceNode(n));
		}

		/**
		 * Handle writing into a local variable, i.e. {@code p = q}
		 */
		@Override
		public void visit(IMAStoreLocalVariable n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}

			// creates the CGLocalReferenceNode if there's isn't any for this slot yet
			CGLocalReferenceNode localReference = CGLocalReferenceNode.getLocalReferenceNode(n);

			// contrary to what the paper says we do not need to apply byPass(p) for flow-sensitive analysis here,
			// because we're processing SSA-formed code only and assignments from multiple control flows to the same
			// variable would cause an IMPhi statement to be added. In fact we cannot apply byPass(p) here, because it
			// would cause the iterative fixpoint algorithm to never converge.

			if (nodeToCGMap.get(n.getOperand()).attach(localReference)) {
				// this node has been updated, invalidate all uses

				IMSlot slot = localReference.getNode().getIMSlot();
				if (slot != null) {
					for (IMNode use : slot.getUses()) {
						// if this use has already been processed (even if it's in the same basic block!), re-add the
						// block to the work list
						bbWorkList.add(slot.getUseBB(use));
					}
				}
			}
		}

		/**
		 * Handle reading from an array, i.e. {@code p = a[i];}
		 *
		 * NOTE: Multiple indexes are not distinguished at the moment
		 */
		@Override
		public void visit(IMAReadArray n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}

			CGLocalReferenceNode value;
			nodeToCGMap.put(n, value = CGLocalReferenceNode.getLocalReferenceNode(n));

			// IMNode indexOperand = n.getIndexOperand();

			// if this cast fails, hell breaks loose anyway ;)
			CGReferenceNode arrayBase = (CGReferenceNode) nodeToCGMap.get(n.getArrayOperand());
			Set<CGObjectNode> pointsTo = arrayBase.getPointsToRelation();
			if (pointsTo.isEmpty()) {
				// the points to relation of the array base is empty, which either means it's null or
				// it's allocated outside of this function. We assume the latter and create a phantom node.
				pointsTo.addAll(CGObjectPhantomNode.getPhantomNodeForReference(arrayBase));
			}

			// for each possible pointee of p in q = p[i]
			for (CGObjectNode objectNode : pointsTo) {
				// get it's field reference p[i] and add a deferred edge from q to p[i]
				objectNode.getFieldReferenceForField(CGObjectNode.arrayAccessIMField).attach(value);
			}
		}

		/**
		 * Handle writing into an array, i.e. {@code a[i] = q;}
		 *
		 * NOTE: Multiple indexes are not distinguished at the moment
		 */
		@Override
		public void visit(IMAStoreArray n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			// get the rvalue from the array store operation and a local reference node to it
			CGNode rvalue = nodeToCGMap.get(n.getRvalue());

			// get a local reference to the array base
			CGReferenceNode arrayBase = (CGReferenceNode) nodeToCGMap.get(n.getArrayOperand());

			Set<CGObjectNode> pointsTo = arrayBase.getPointsToRelation();
			if (pointsTo.isEmpty()) {
				// analog to operand.isFieldAccess() in the IMAStoreLocalVariable-visitor, create a
				// phantom node for the array reference if we don't know where it's pointing yet.
				pointsTo.addAll(CGObjectPhantomNode.getPhantomNodeForReference(arrayBase));
			}

			// for each entry in the pointsTo-relation of arrayReference
			for (CGObjectNode objectNode : pointsTo) {
				// add a deferred edge from the field reference to the rvalue reference
				rvalue.attach(objectNode.getFieldReferenceForField(CGObjectNode.arrayAccessIMField));
			}
		}

		/**
		 * Handle reading from object fields, i.e. {@code p = q.f}
		 */
		@Override
		public void visit(IMGetField n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			IMField field = n.getField();
			if (!field.isReference()) {
				// we only care about references
				return;
			}

			CGLocalReferenceNode localReference = CGLocalReferenceNode.getLocalReferenceNode(n);
			nodeToCGMap.put(n, localReference);

			CGReferenceNode base = (CGReferenceNode) nodeToCGMap.get(n.getBase());

			Set<CGObjectNode> pointsTo = base.getPointsToRelation();
			if (pointsTo.isEmpty()) {
				// "if U is empty, we create a phantom node and add it to U"
				pointsTo.addAll(CGObjectPhantomNode.getPhantomNodeForReference(base));
			}

			// for each possible object q referenced by base
			for (CGObjectNode objectNode : pointsTo) {
				// get its field reference q.f and add a deferred edge from p to q.f
				objectNode.getFieldReferenceForField(field).attach(localReference);
			}
		}

		/**
		 * Handle reading from static fields, i.e. {@code p = T.f}
		 */
		@Override
		public void visit(IMGetStatic n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			IMField field = n.getField();
			if (!field.isReference()) {
				// we only care about references
				return;
			}

			nodeToCGMap.put(n, CGGlobalReferenceNode.getReferenceNode(n.getClassCPEntry(), field, callerActualsAndGlobals));
		}

		/**
		 * Handle writing to object fields, i.e. {@code p.f = q}
		 */
		@Override
		public void visit(IMPutField n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			IMField field = n.getField();
			if (!field.isReference()) {
				// we only care about references
				return;
			}

			CGReferenceNode base = (CGReferenceNode) nodeToCGMap.get(n.getBase());
			CGNode pointee = nodeToCGMap.get(n.getRightNode());

			Set<CGObjectNode> pointsTo = base.getPointsToRelation();
			if (pointsTo.isEmpty()) {
				// if the points to relation of the base pointer is empty either (1) base is null, in
				// which case a NPE will be thrown or (2) this object was allocated outside of this method; we
				// conservatively assume (2) and create a phantom node for this reference
				pointsTo.addAll(CGObjectPhantomNode.getPhantomNodeForReference(base));
			}

			// for each entry in the pointsTo-relation of base
			for (CGObjectNode objectNode : pointsTo) {
				// add a deferred edge from its field reference node to the operand
				pointee.attach(objectNode.getFieldReferenceForField(field));
			}
		}

		/**
		 * Handle writing to static fields, i.e. {@code T.f = q}
		 */
		@Override
		public void visit(IMPutStatic n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			IMField field = n.getField();
			if (!field.isReference()) {
				// we only care about references
				return;
			}

			CGGlobalReferenceNode staticField = CGGlobalReferenceNode.getReferenceNode(n.getClassCPEntry(), field, callerActualsAndGlobals);
			CGNode pointee = nodeToCGMap.get(n.getRightNode());

			// add a deferred edge from T.f to q
			pointee.attach(staticField);
		}

		/**
		 * Handle returning from a method, i.e. {@code return p;}
		 */
		@Override
		public void visit(IMEpilog n) throws CompileException {
			if (debug) {
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());
			}
			if (n.usesSlot()) {
				IMSlot slot = n.getIMSlot();

				if (!slot.isReference())
					return;

				CGLocalReferenceNode.getLocalReferenceNode(n).attach(calleeActuals[0]);
			}
		}

		/**
		 * Handle {@code IMPhi} operations
		 */
		@Override
		public void visit(IMPhi n) throws CompileException {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			if (!n.isReference())
				return;

			CGLocalReferenceNode localReference = CGLocalReferenceNode.getLocalReferenceNode(n);
			nodeToCGMap.put(n, localReference);

			for (IMNode candidate : n.getSlots()) {
				nodeToCGMap.get(candidate).attach(localReference);
			}
		}

		/**
		 * Handle function calls
		 */
		@Override
		public void visit(IMInvoke n) throws CompileException {
			if (debug)
				System.out.printf("%s: %s%n", n.getClass().getSimpleName(), n.dumpOperands());

			CGActualReferenceNode[] actualArguments = CGActualReferenceNode.getArgumentActualReferenceNodes(n, nodeToCGMap,
				callerActualsAndGlobals);

			if (!n.isReference())
				return;

			nodeToCGMap.put(n, actualArguments[0]);
		}

		@Override
		public void visit(IMInvokeInterface n) throws CompileException {
			visit((IMInvoke) n);
		}

		@Override
		public void visit(IMInvokeVirtual n) throws CompileException {
			visit((IMInvoke) n);
		}

		@Override
		public void visit(IMInvokeSpecial n) throws CompileException {
			visit((IMInvoke) n);
		}

		@Override
		public void visit(IMInvokeStatic n) throws CompileException {
			visit((IMInvoke) n);
		}

		/**
		 * Handle casts, i.e. {@code p = (T) q}
		 */
		@Override
		public void visit(IMCheckCast n) throws CompileException {
			nodeToCGMap.put(n, nodeToCGMap.get(n.getRightNode()));
		}

		/**
		 * Handle exceptions, i.e. {@code throw p}
		 */
		@Override
		public void visit(IMThrow n) throws CompileException {
			nodeToCGMap.get(n.getException()).updateEscapeState(EscapeState.Global);
		}
	}

	/**
	 * Worklist for the iterative fixpoint algorithm visiting basic blocks.
	 */
	private Set<IMBasicBlock>						bbWorkList	= null;

	/**
	 * An instance of the connection graph we're creating in this class.
	 */
	private ExtendedEscapeAnalysisConnectionGraph	cg			= null;

	/**
	 * Debug flag, true if _X_dbg_ext_escape_analysis is set, see {@code isEnabled()}.
	 */
	private boolean									debug		= false;

	/**
	 * The updateNodes routine as described in the paper; it's used to update the caller's connection graph with the
	 * information from the non-local subgraph of the callee. This method uses the base case provided by
	 * {@code interproceduralAnalysis(IMMethod)} to update the graph and find new maps-to relations, which it in term
	 * uses to call itself recursively.
	 *
	 * @param mapsTo
	 *            map holding the MapsToObj(n) relation described in the paper. The map uses object nodes in the
	 *            callee's connection graph as key and object nodes in the caller's connection graph as elements in the
	 *            set of pointed-to objects.
	 * @param calleeNode
	 *            the actual reference node in the callee's connection graph currently being processed
	 * @param callerNode
	 *            the node in the caller's connection graph in the maps-to relation of the {@code calleeNode}
	 */
	@Deprecated
	private boolean updateNodes(final Map<CGObjectNode, Set<CGObjectNode>> mapsTo, final CGReferenceNode calleeNode,
		final CGReferenceNode callerNode, final Map<CGReferenceNode, Set<CGReferenceNode>> visited) {
		boolean changed = false;

		for (CGObjectNode calleeObjectNode : calleeNode.getPointsToRelation()) {
			Set<CGObjectNode> mapsToObj = mapsTo.get(calleeObjectNode);
			if (mapsToObj == null) {
				mapsTo.put(calleeObjectNode, mapsToObj = new HashSet<CGObjectNode>());
			}

			for (CGObjectNode callerObjectNode : callerNode.getPointsToRelation()) {
				boolean changedLocally = false;
				if (mapsToObj.add(callerObjectNode)) {
					changed = changedLocally = true;
				}

				// recursively find and process further mapsTo nodes
				while (true) {
					try {
						for (CGFieldReferenceNode calleeFieldReferenceNode : calleeObjectNode.getFields()) {
							CGFieldReferenceNode callerFieldReferenceNode = callerObjectNode
								.getFieldReferenceForField(calleeFieldReferenceNode.getField());

							Set<CGReferenceNode> visitedFromHere = visited.get(calleeFieldReferenceNode);
							if (visitedFromHere == null) {
								visitedFromHere = new HashSet<CGReferenceNode>();
								visited.put(calleeFieldReferenceNode, visitedFromHere);
							}
							if (!visitedFromHere.contains(callerFieldReferenceNode)) {
								visitedFromHere.add(callerFieldReferenceNode);
								System.out.println("    updateNodes(" + calleeFieldReferenceNode.hashCode() + ", "
									+ callerFieldReferenceNode.hashCode() + ")");
								changed |= updateNodes(mapsTo, calleeFieldReferenceNode, callerFieldReferenceNode, visited);
							}
						}
						break;
					} catch (ConcurrentModificationException cme) {}
				}

				if (changedLocally) {
					// update the escape state, if necessary
					if (callerObjectNode.getEscapeState() == EscapeState.Global) {
						// this code bears room for improvement: consider a method toString() that's being called with a
						// globally escaping parameter at one call site. Since we only have one connection graph per
						// calling context and this code propagates escape state into the callee this one call site
						// causes all arguments given at any call site of the same toString()-function to be marked
						// globally escaping, too.
						// Here's a possible workaround: Introducing a fourth escape state between Global and Method
						// (maybe PropagatedGlobal) whose escape state would not be propagated back up to other call
						// sites might solve the problem (although more research needs to be done on that).

						// we need to update the callee's escape state
						calleeObjectNode.updateEscapeState(EscapeState.Global);
					} else if (calleeObjectNode.getEscapeState() == EscapeState.Global) {
						// we need to update the caller's escape state
						callerObjectNode.updateEscapeState(EscapeState.Global);
					}
				}
			}
			for (CGObjectNode callerObjectNode : callerNode.getPointsToRelation()) {
				// always update edges, even if this node isn't new in the mapsToObj relation, because the mapsTo set
				// could have changed by a different call to updateNodes (but only run this once for every object node
				// in one call to updateNodes)
				for (CGFieldReferenceNode calleeFieldReferenceNode : calleeObjectNode.getFields()) {
					CGFieldReferenceNode callerFieldReferenceNode = callerObjectNode.getFieldReferenceForField(calleeFieldReferenceNode
						.getField());

					// add points-to edges where applicable
					for (CGObjectNode calleeChildObjectNode : calleeFieldReferenceNode.getPointsToRelation()) {
						if (mapsTo.get(calleeChildObjectNode) != null) {
							for (CGObjectNode callerChildObjectNode : mapsTo.get(calleeChildObjectNode)) {
								if (callerChildObjectNode.attach(callerFieldReferenceNode)) {
									System.out.println("     Added edge in caller graph: " + callerFieldReferenceNode.hashCode() + " => "
										+ callerChildObjectNode.hashCode());
									changed = true;
								}
							}
						}
					}
				}
			}
		}

		return changed;
	}

	/**
	 * Returns the name of this pass
	 *
	 * @return name
	 */
	@Override
	public String getName() {
		return "Extended escape analysis";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		debug = opts.hasOption("_X_dbg_ext_escape_analysis");
		boolean enabled = opts.hasOption("ext_escape_analysis");
		if (enabled)
			cg = ExtendedEscapeAnalysisConnectionGraph.getInstance(opts);
		return enabled;
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsAnalysis(true);
		info.addDependency(FlowAnalysis.class);
		info.addDependency(SSAConstruction.class);
		info.addDependency(VTableComputation.class);
		info.addDependency(CallGraphAnalysis.class);
		info.addAntiDependency(SSADeconstruction.class);
	}

	/**
	 * We're overriding the method traversing the repository because we need two passes over the code; intraprocedural
	 * analysis must be finished before interprocedural analysis can be started.
	 */
	@Override
	public boolean runOnRepository(ClassStore repository) throws CompileException {
		boolean result;

		result = super.runOnRepository(repository);

		// process workList created during visiting the methods
		SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo = new SplitKeyHashMultimap<IMMethod, CGObjectNode, CGObjectNode>();

		DirectedGraph<IMMethod> callGraph = pm.getPass(CallGraphAnalysis.class).getCallGraph();
		if (debug) {
			try {
				callGraph.dump(repository.getOptions().getOutputPath() + "/callGraph.dot");
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (DotWriterException e) {
				e.printStackTrace();
			}
		}
		Deque<Set<IMMethod>> sccDeque = callGraph.tarjan();
		FlowLatticeInfoStore flowInfoStore = pm.getPass(FlowAnalysis.class).getFlowLatticeInfoStore();
		interproceduralAnalysis(mapsTo, flowInfoStore, sccDeque);

		return result;
	}

	/**
	 * Run interprocedural analysis on each strongly connected component in {@code sscDeque} in a fixpoint-iterative
	 * manner.
	 *
	 * @param mapsTo
	 *            the mapsToObj relation
	 * @param flowInfoStore
	 *            flow info store holding the mapping to candidates of a given invocation
	 * @param sccDeque
	 *            queue of strongly connected components in the call graph
	 * @throws CompileException
	 */
	private void interproceduralAnalysis(SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, FlowLatticeInfoStore flowInfoStore,
		Deque<Set<IMMethod>> sccDeque) throws CompileException {
		CallVisitor updateNodesVisitor = new CallVisitor() {
			/**
			 * The updateNodes routine as described in the paper; it's used to update the caller's connection graph with
			 * the information from the non-local subgraph of the callee. This method uses the base case provided by
			 * {@code interproceduralAnalysis(IMMethod)} to update the graph and find new maps-to relations, which it in
			 * term uses to call itself recursively.
			 *
			 * @param workList
			 *            the set of pairs (calleeNode, callerNode) to be processed. This method should add pairs to
			 *            this list, if it wants them to be visited rather than recursing directly.
			 * @param visited
			 *            a set of previously visited pairs (calleeNode, callerNode). This is only provided as a
			 *            convenience and will not be enforced by the calling method. A method can use this set for
			 *            whatever purpose it sees fit.
			 * @param mapsTo
			 *            split key map holding the MapsToObj(n) relation described in the paper. The map uses object
			 *            nodes in the callee's connection graph and the caller method as key and object nodes in the
			 *            caller's connection graph as elements in the set of pointed-to objects.
			 * @param caller
			 *            the caller method, needed as key for the split key hash map {@code mapsTo}
			 * @param callee
			 *            the callee method, needed to create phantom nodes
			 * @param calleeNode
			 *            the actual reference node in the callee's connection graph currently being processed
			 * @param callerNode
			 *            the node in the caller's connection graph in the maps-to relation of the {@code calleeNode}
			 */
			@Override
			public boolean process(Multimap<CGReferenceNode, CGReferenceNode> workList, Multimap<CGReferenceNode, CGReferenceNode> visited,
				SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, IMMethod caller, IMMethod callee,
				CGReferenceNode calleeNode, CGReferenceNode callerNode) {
				boolean changed = false;

				visited.add(calleeNode, callerNode);

				if (debug)
					System.out.println("    updateNodes(" + calleeNode.hashCode() + ", " + callerNode.hashCode() + ")");

				for (CGObjectNode calleeObjectNode : calleeNode.getPointsToRelation()) {
					if (callerNode.getPointsToRelation().isEmpty()) {
						// CGObjectPhantomNode.getPhantomNode(callerNode, calleeObjectNode, callee, caller);
						calleeObjectNode.getCloneInMethod(callerNode, caller);
					}
					for (CGObjectNode callerObjectNode : callerNode.getPointsToRelation()) {
						if (mapsTo.put(caller, calleeObjectNode, callerObjectNode)) {
							changed = true;

							for (CGFieldReferenceNode calleeFieldReferenceNode : calleeObjectNode.getFields()) {
								CGFieldReferenceNode callerFieldReferenceNode = callerObjectNode
									.getFieldReferenceForField(calleeFieldReferenceNode.getField());

								if (!visited.contains(calleeFieldReferenceNode, callerFieldReferenceNode))
									workList.add(calleeFieldReferenceNode, callerFieldReferenceNode);
							}
						}
					}
				}

				return changed;
			}
		};
		CallVisitor updateEdgesVisitor = new CallVisitor() {
			/**
			 * This method is called after running {@code updateNodes()} and adds edges in the caller graph where edges
			 * are in the callee connection graph.
			 *
			 * @param workList
			 *            the set of pairs (calleeNode, callerNode) to be processed. This method should add pairs to
			 *            this list, if it wants them to be visited rather than recursing directly.
			 * @param visited
			 *            a set of previously visited pairs (calleeNode, callerNode). This is only provided as a
			 *            convenience and will not be enforced by the calling method. A method can use this set for
			 *            whatever purpose it sees fit.
			 * @param mapsTo
			 *            split key map holding the MapsToObj(n) relation described in the paper. The map uses object
			 *            nodes in the callee's connection graph and the caller method as key and object nodes in the
			 *            caller's connection graph as elements in the set of pointed-to objects.
			 * @param caller
			 *            the caller method, needed as key for the split key hash map {@code mapsTo}
			 * @param callee
			 *            the callee method
			 * @param calleeNode
			 *            the actual reference node in the callee's connection graph currently being processed
			 * @param callerNode
			 *            the node in the caller's connection graph in the maps-to relation of the {@code calleeNode}
			 */
			@Override
			public boolean process(Multimap<CGReferenceNode, CGReferenceNode> workList, Multimap<CGReferenceNode, CGReferenceNode> visited,
				SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, IMMethod caller, IMMethod callee,
				CGReferenceNode calleeNode, CGReferenceNode callerNode) {
				Multimap<CGObjectNode, CGReferenceNode> callerEdges = new HashMultimap<CGObjectNode, CGReferenceNode>();

				boolean changed = helper(workList, visited, mapsTo, caller, callee, calleeNode, callerNode, callerEdges);

				for (CGObjectNode object : callerEdges.keySet()) {
					for (CGReferenceNode reference : callerEdges.get(object)) {
						object.attach(reference);
					}
				}

				return changed;
			}

			private boolean helper(Multimap<CGReferenceNode, CGReferenceNode> workList, Multimap<CGReferenceNode, CGReferenceNode> visited,
				SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, IMMethod caller, IMMethod callee,
				CGReferenceNode calleeNode, CGReferenceNode callerNode, Multimap<CGObjectNode, CGReferenceNode> callerEdges) {
				boolean changed = false;

				visited.add(calleeNode, callerNode);

				if (false && debug)
					System.out.println("    updateEdges(" + calleeNode.hashCode() + ", " + callerNode.hashCode() + ")");

				for (CGObjectNode calleeObjectNode : calleeNode.getPointsToRelation()) {

					for (CGObjectNode callerObjectNode : callerNode.getPointsToRelation()) {
						// always update edges, even if this node isn't new in the mapsToObj relation, because
						// the mapsTo set could have changed by a different call to updateNodes
						for (CGFieldReferenceNode calleeFieldReferenceNode : calleeObjectNode.getFields()) {
							CGFieldReferenceNode callerFieldReferenceNode = callerObjectNode
								.getFieldReferenceForField(calleeFieldReferenceNode.getField());

							if (!visited.contains(calleeFieldReferenceNode, callerFieldReferenceNode))
								workList.add(calleeFieldReferenceNode, callerFieldReferenceNode);

							// add points-to edges where applicable
							for (CGObjectNode calleeChildObjectNode : calleeFieldReferenceNode.getPointsToRelation()) {
								Collection<CGObjectNode> childObjectNodes = mapsTo.get(caller, calleeChildObjectNode);
								if (childObjectNodes != null) {
									for (CGObjectNode callerChildObjectNode : childObjectNodes) {
										if (!callerFieldReferenceNode.getPointees().contains(callerChildObjectNode)) {
											changed = true;
											if (false && debug)
												System.out.println("     Added edge in caller graph: "
													+ callerFieldReferenceNode.hashCode() + " => " + callerChildObjectNode.hashCode());
											// delay actually adding the edges to prevent
											// ConcurrentModificationExceptions
											callerEdges.add(callerChildObjectNode, callerFieldReferenceNode);
										}
									}
								}
							}
						}

					}
				}

				return changed;
			}

		};
		for (Set<IMMethod> scc : sccDeque) {
			boolean changed;
			if (debug)
				System.out.println("IPA-SCC:");
			do {
				changed = false;
				for (IMMethod method : scc) {
					if (method == null)
						continue;
					changed |= interproceduralAnalysis(updateNodesVisitor, mapsTo, flowInfoStore, method);
					changed |= interproceduralAnalysis(updateEdgesVisitor, mapsTo, flowInfoStore, method);
					if (debug)
						System.out.println(" Method repeat");
				}
			} while (changed);
		}

		CallVisitor escapeStatePropagationVisitor = new CallVisitor() {
			/**
			 * Second part of the interprocedural escape analysis: A top-down pass over the call graph where all objects
			 * marked globally escaping in a caller have the objects mapping to those in the callees also marked
			 * globally escaping.
			 *
			 * @param workList
			 *            the set of pairs (calleeNode, callerNode) to be processed. This method should add pairs to
			 *            this list, if it wants them to be visited rather than recursing directly.
			 * @param visited
			 *            a set of previously visited pairs (calleeNode, callerNode). This is only provided as a
			 *            convenience and will not be enforced by the calling method. A method can use this set for
			 *            whatever purpose it sees fit.
			 * @param mapsTo
			 *            the mapsToObj relation
			 * @param flowInfo
			 *            flow info holding the mapping to candidates of a given invocation
			 * @param method
			 *            the method to process
			 * @return true, if any object was marked globally escaping
			 * @throws CompileException
			 */
			@Override
			public boolean process(Multimap<CGReferenceNode, CGReferenceNode> workList, Multimap<CGReferenceNode, CGReferenceNode> visited,
				SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo, IMMethod caller, IMMethod callee,
				CGReferenceNode calleeNode, CGReferenceNode callerNode) {
				boolean changed = false;

				visited.add(calleeNode, callerNode);

				if (debug)
					System.out.println("    updateEscapeState(" + calleeNode.hashCode() + ", " + callerNode.hashCode() + ")");

				for (CGObjectNode calleeObjectNode : calleeNode.getPointsToRelation()) {
					Collection<CGObjectNode> mapsToObj = mapsTo.get(caller, calleeObjectNode);
					if (mapsToObj != null) {
						for (CGObjectNode callerObjectNode : mapsToObj) {
							if (callerObjectNode.getEscapeState() == EscapeState.Global) {
								if (calleeObjectNode.getEscapeState() != EscapeState.Global) {
									// we need to update the callee's escape state
									changed = true;
									calleeObjectNode.updateEscapeState(EscapeState.Global);
								}
							} else {
								// if this was no globally escaping node, recurse
								// we don't have to recurse on globally escaping nodes, because {@code
								// updateEscapeState()} already does for us
								for (CGFieldReferenceNode calleeFieldReferenceNode : calleeObjectNode.getFields()) {
									CGFieldReferenceNode callerFieldReferenceNode = callerObjectNode
										.getFieldReferenceForField(calleeFieldReferenceNode.getField());

									if (!visited.contains(calleeFieldReferenceNode, callerFieldReferenceNode))
										workList.add(calleeFieldReferenceNode, callerFieldReferenceNode);
								}
							}
						}
					}
				}
				return changed;
			}
		};

		Iterator<Set<IMMethod>> revIt = sccDeque.descendingIterator();
		while (revIt.hasNext()) {
			Set<IMMethod> scc = revIt.next();
			boolean changed;
			if (debug)
				System.out.println("IPA-SSC-ESP:");
			do {
				changed = false;
				for (IMMethod method : scc) {
					if (method == null)
						continue;
					changed |= interproceduralAnalysis(escapeStatePropagationVisitor, mapsTo, flowInfoStore, method);
				}
			} while (changed);
		}
	}

	/**
	 * Run interprocedural analysis on a given method {@code method}.
	 *
	 * @param callVisitor
	 *            a closure to be called for each pair of (invocation, callee candidate)
	 * @param mapsTo
	 *            the mapsToObj relation
	 * @param flowInfoStore
	 *            flow info store holding the mapping to candidates of a given invocation
	 * @param method
	 *            the method to process
	 * @return true, if the connection graph of the given method changed
	 * @throws CompileException
	 */
	private boolean interproceduralAnalysis(CallVisitor callVisitor, SplitKeyMultimap<IMMethod, CGObjectNode, CGObjectNode> mapsTo,
		FlowLatticeInfoStore flowInfoStore, IMMethod method) throws CompileException {
		boolean changed = false;

		if (debug)
			System.out.println(" IPA: " + method.getAlias());

		// we're not using those anyway, but not clearing them might increase the maintenance overhead
		CGNode.clearEscapeLists();
		CGReferenceNode.clearReferenceNodeList();

		for (IMBasicBlock bb : method) {
			FlowLatticeInfo flowInfo = flowInfoStore.getForDomains(bb.getDomains());

			for (IMNode root : bb) {
				for (IMNode node : IMCodeFlattener.getFlattenedCode(root)) {
					if (!node.isInvoke())
						continue;

					IMInvoke invocation = (IMInvoke) node;

					if (debug)
						System.out.println(" MI: " + invocation.dumpBC());

					// if this method doesn't have a callerActualsAndGlobals list it doesn't have a connection graph
					// either and wasn't visited by {@code runOnMethod}.
					if (cg.methodToCallerActualsAndGlobalsMap.get(method) == null) {
						if (debug)
							System.out.println("  IGNORING");
						return false;
					}
					// get the caller's arguments array
					CGActualReferenceNode[] callerActualArguments = CGActualReferenceNode.getArgumentActualReferenceNodes(invocation,
						cg.methodToCGMap.get(method), cg.methodToCallerActualsAndGlobalsMap.get(method));

					for (IMMethod candidate : flowInfo.getCandidates(invocation)) {
						if (debug)
							System.out.println("  CANDIDATE: " + candidate.getAlias());

						// set up the candidates' arguments array
						// this code expects callerAlctualArguments and calleeActualArguments to have the same form
						IMSlot[] calleeSlotArguments = candidate.getMethodFrame().getOriginalArguments();
						CGActualReferenceNode[] calleeActualArguments = new CGActualReferenceNode[1 + calleeSlotArguments.length];
						if (candidate.isReference())
							calleeActualArguments[0] = CGActualReferenceNode.getReturnReferenceNode(candidate);
						for (int i = 0; i < calleeSlotArguments.length; ++i)
							if (calleeSlotArguments[i].isReference())
								calleeActualArguments[i + 1] = CGActualReferenceNode.getActualReferenceNode(calleeSlotArguments[i]);

						// both arrays should have the same size
						if (calleeActualArguments.length != callerActualArguments.length)
							throw new CompileException("caller and callee argument count doesn't match for method " + method.getAlias());

						if (debug) {
							System.out.println("  CALLEE: " + Arrays.toString(calleeActualArguments));
							System.out.println("  CALLER: " + Arrays.toString(callerActualArguments));
						}

						Multimap<CGReferenceNode, CGReferenceNode> workList = new HashUniqueMultimap<CGReferenceNode, CGReferenceNode>();
						Multimap<CGReferenceNode, CGReferenceNode> visited = new HashUniqueMultimap<CGReferenceNode, CGReferenceNode>();

						for (int i = 0; i < calleeActualArguments.length; ++i) {
							if (calleeActualArguments[i] == null && callerActualArguments[i] == null)
								continue;

							if (calleeActualArguments[i] == null || callerActualArguments[i] == null)
								throw new CompileException("Argument no. " + i + " for call to " + invocation.getCallee().getAlias()
									+ " is null on one side, but not the other");

							workList.add(calleeActualArguments[i], callerActualArguments[i]);
						}

						while (!workList.isEmpty()) {
							CGReferenceNode key = workList.keySet().iterator().next();
							Collection<CGReferenceNode> values = workList.get(key);
							CGReferenceNode value = values.iterator().next();

							workList.remove(key, value);

							changed |= callVisitor.process(workList, visited, mapsTo, method, candidate, key, value);
						}
					}
				}
			}
		}

		if (changed)
			cg.dumpGraph(method);

		return changed;
	}

	@Override
	public boolean runOnClass(IMClass imClass) throws CompileException {
		if (debug) {
			System.out.printf("%n---> Handling class %s%n", imClass.getClassName());
		}

		return super.runOnClass(imClass);
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {
		if (debug) {
			System.out.printf(" --> Handling method %s%n", method.getMethodNameAndType());
		}

		// do this _before_ creating the visitor, because the visitor's constructor already creates some nodes!
		CGNode.clearEscapeLists();
		CGReferenceNode.clearReferenceNodeList();

		final EEAVisitor visitor = new EEAVisitor(method);

		// visit the instructions

		// use an iterative fixpoint algorithm to visit basic blocks; occurrences of IMPhi referencing unvisited blocks
		// might force us to visit a block again.

		// in order to check whether adding an edge invalidates the result of a previously analyzed basic block, we have
		// to keep track which basic blocks are still in the work list (and this needs to be fast, which is the reason
		// we're using a HashSet here)
		bbWorkList = new LinkedHashSet<IMBasicBlock>();
		for (IMBasicBlock bb : method) {
			bbWorkList.add(bb);
		}

		while (!bbWorkList.isEmpty()) {
			Iterator<IMBasicBlock> bbIterator = bbWorkList.iterator();

			IMBasicBlock bb = bbIterator.next();
			bbIterator.remove();

			for (IMNode root : bb) {
				for (IMNode node : IMCodeFlattener.getFlattenedCode(root)) {
					node.accept(visitor);
					// dumpGraph();
				}
			}
		}

		// remove remaining deferred edges
		for (CGReferenceNode node : CGReferenceNode.getReferenceNodeList()) {
			node.pathCompression();
		}

		cg.dumpGraph(method);

		return false;
	}
}
