/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis;

import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import keso.compiler.BuilderOptions;
import keso.compiler.CompileException;
import keso.compiler.analysis.connectiongraph.CGActualReferenceNode;
import keso.compiler.analysis.connectiongraph.CGDotWriter;
import keso.compiler.analysis.connectiongraph.CGNode;
import keso.compiler.analysis.connectiongraph.CGReferenceNode;
import keso.compiler.imcode.IMBasicBlock;
import keso.compiler.imcode.IMCodeFlattener;
import keso.compiler.imcode.IMMethod;
import keso.compiler.imcode.IMNode;
import keso.util.DotWriterException;

/**
 * Data structures needed for {@code ExtendedEscapeAnalaysis} as described in the implemented paper. This essentially is
 * the connection graph.
 */
public final class ExtendedEscapeAnalysisConnectionGraph {
	/**
	 * Since this class is intended to be used as singleton, here's a reference to the only created object
	 */
	private static ExtendedEscapeAnalysisConnectionGraph	instance							= null;

	/**
	 * This map keeps a mapping from an {@code IMMethod} to a map mapping its {@code IMNode}s to their corresponding
	 * {@code CGNode}s in the connection graph. This keeps the maps from {@code IMNode} to {@code CGNode} small (and
	 * quick) while still preserving references to all nodes (e.g. for later reference in dumping the graph of a
	 * method).
	 */
	Map<IMMethod, Map<IMNode, CGNode>>						methodToCGMap						= new HashMap<IMMethod, Map<IMNode, CGNode>>();

	/**
	 * This map keeps a mapping from an {@code IMMethod} to an array of callee-side actual reference nodes, i.e. formal
	 * parameters and the {@code IMMethod}'s return value. Index 0 in the array is defined to hold the return reference
	 * node or null, if the method does not return a reference. For instance methods, index 1 holds the {@code this}
	 * reference and the following indexes hold the parameters in-order. For static methods, the parameters start at
	 * index 1.
	 */
	Map<IMMethod, CGActualReferenceNode[]>					methodToCalleeActualsMap			= new HashMap<IMMethod, CGActualReferenceNode[]>();

	/**
	 * This map keeps a mapping from an {@code IMMethod} to a list of caller-side actual reference nodes, i.e.
	 * parameters and return values of methods called inside a given {@code IMMethod} by means of {@code IMInvoke}, and
	 * any global reference nodes used in the method. This list is needed because these nodes only have edges into the
	 * rest of the graph, but there are no edges from the rest of the graph to these nodes. This list is the only way to
	 * get a reference to those nodes.
	 */
	Map<IMMethod, List<CGReferenceNode>>					methodToCallerActualsAndGlobalsMap	= new HashMap<IMMethod, List<CGReferenceNode>>();

	/**
	 * This map keeps track of how many dumps have been generated for a specific {@code IMMethod}'s graph
	 */
	Map<IMMethod, Integer>									methodToGraphCounterMap				= new HashMap<IMMethod, Integer>();

	/**
	 * The {@code BuilderOptions} passed in at creation of this object used to find out whether we're debugging
	 */
	private BuilderOptions									opts								= null;

	/**
	 * Helper function dumping the current graph of a method {@code method} into a graphviz dot file while appending a
	 * running number to the filename. This is especially useful when debugging.
	 *
	 * @throws CompileException
	 */
	public void dumpGraph(final IMMethod method) throws CompileException {
		if (opts.hasOption("_X_dump_ext_escape_analysis")) {
			try {
				Integer countInteger = methodToGraphCounterMap.get(method);
				int count = (countInteger != null) ? countInteger : 0;
				methodToGraphCounterMap.put(method, count + 1);

				// dump the connection graph for this method into a .dot file
				String outPath = method.getOptions().getOutputPath() + '/' + method.getAlias() + "_eea_no" + count
						+ ".dot";
				CGDotWriter dotWriter;
				dotWriter = new CGDotWriter(outPath, method.getAlias());

				// make sure the return value and the parameters are dumped, if they exist
				for (CGActualReferenceNode actual : methodToCalleeActualsMap.get(method))
					if (actual != null)
						actual.accept(dotWriter);
				// make sure caller actual reference nodes and used global reference nodes are dumped
				for (CGReferenceNode actual : methodToCallerActualsAndGlobalsMap.get(method))
					if (actual != null)
						actual.accept(dotWriter);

				for (IMBasicBlock bb : method) {
					for (IMNode root : bb) {

						List<IMNode> flattened = IMCodeFlattener.getFlattenedCode(root);
						for (IMNode node : flattened) {
							CGNode cgnode = methodToCGMap.get(method).get(node);
							if (cgnode != null)
								cgnode.accept(dotWriter);
						}
					}
				}

				dotWriter.finalize();
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (DotWriterException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Singleton accessor for the connection graph; this requires the {@code BuilderOptions} for your compiler run,
	 * because it needs to find out whether to dump graphs or not. If you're sure the graph has already been created you
	 * can also pass null here (but please don't complain if it crashes with a NPE then).
	 *
	 * @param opts
	 *            the builder options for your current compiler run
	 * @return the only instance of ExtendedEscapeAnalysisConnectionGraph
	 */
	public static ExtendedEscapeAnalysisConnectionGraph getInstance(final BuilderOptions opts) {
		if (instance == null) {
			instance = new ExtendedEscapeAnalysisConnectionGraph(opts);
		}
		return instance;
	}

	/**
	 * private constructor for this class, storing the {@code BuilderOptions} in a class member
	 *
	 * @param opts
	 *            the builder options for your current compiler run
	 */
	private ExtendedEscapeAnalysisConnectionGraph(final BuilderOptions opts) {
		this.opts = opts;
	}
}
