/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;
import java.util.Set;

import keso.classfile.*;
import keso.classfile.datatypes.*;
import keso.compiler.*;
import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.*;
import keso.util.*;

/**
 * Flow-analysis pass.
 * This pass performs an inter-procedural per-domain control and data flow
 * analysis. Its design is similar to that of the "Sparse conditional constant
 * propagation" algorithm described by Wegman and Zadeck.
 */
public final class FlowAnalysis extends ReachabilityAnalysis
implements IMMethodVisitor<Void> {

	private static final class CFGEdge {

		private final IMBasicBlock from;
		private final IMBasicBlock to;

		private CFGEdge(IMBasicBlock from, IMBasicBlock to) {
			this.from = from;
			this.to   = to;
		}

	}

	private static final class MidBBState {

		private final IMBasicBlock bb;
		private final IMNode       lastVisited;

		private MidBBState(IMBasicBlock bb, IMNode lastVisited) {
			this.bb          = bb;
			this.lastVisited = lastVisited;
		}

	}

	/** Flow-lattice information store. */
	private FlowLatticeInfoStore flowInfoStore;
	
	/** Flow-lattice information of the currently processed domain. */
	private FlowLatticeInfo currInfo;

	private final Deque<IMNode>     ssaWorkList  = new ArrayDeque<IMNode>();
	private final Deque<CFGEdge>    flowWorkList = new ArrayDeque<CFGEdge>();
	private final Deque<MidBBState> midBBStack   = new ArrayDeque<MidBBState>();

	/**
	 * Maps nodes to their respective basic block.
	 * This map only contains mappings for the root nodes of expression trees.
	 */
	private final Map<IMNode, IMBasicBlock> nodeBlocks
	= new HashMap<IMNode, IMBasicBlock>();

	private final Multimap<IMBasicBlock, IMNode> blockPhis
	= new HashUniqueMultimap<IMBasicBlock, IMNode>();

	/**
	 * Flag indicating whether the currently visited basic block contains an
	 * explicit branch.
	 */
	private boolean explicitBranch = false;

	/**
	 * Set of expression trees that had already been processed at the time the
	 * analysis of the basic block stopped at a method invocation.
	 */
	private final Set<IMNode> skippedRoots = new HashSet<IMNode>();

	/** Set of conditional branches that are already associated with a path. */
	private final Set<IMConditionalBranch> pathStarts
	= new HashSet<IMConditionalBranch>();

	@Override
	public String getName() {
		return "Flow analysis";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return (opts.hasOption("simple_reachability_analysis") == false);
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsAnalysis(false); // FIXME
		info.addDependency(ConstantFolding.class);
		info.addDependency(DomTreeComputation.class);
		info.addDependency(LivenessAnalysis.class);
		info.addDependency(VTableComputation.class);
		info.addAntiDependency(SSADeconstruction.class);
	}

	@Override
	public boolean runOnDomain(IMDomain domain) throws CompileException {

		BuilderOptions opts = domain.getOptions();
		opts.verbose("###### Processing domain " + domain);

		// Create a new flow info object for this domain
		currInfo = new FlowLatticeInfo(nodeBlocks);
		flowInfoStore = FlowLatticeInfoStore.getInstance(opts, nodeBlocks);
		flowInfoStore.putForDomain(domain, currInfo);

		// Clear the work lists just to be sure - they should be empty anyway
		flowWorkList.clear();
		ssaWorkList.clear();
		midBBStack.clear();

		// Clear the set of skipped expression trees and of path starts
		skippedRoots.clear();
		pathStarts.clear();

		// Add the non-primitive argument types of all exported service methods
		// to the set of instantiated classes
		JavaDomain domainDef = domain.getDef();
		for (ServiceDefinition serviceDef : domainDef.getServices()) {
			IMClass iface = serviceDef.getServiceInterface(classStore);
			for (IMMethod method : iface.getMethods()) {
				MethodData source = method.getSource();
				BasicTypeDescriptor[] argTypeDescs = source.getParameterTypes();
				for (BasicTypeDescriptor argTypeDesc : argTypeDescs) {
					addInstantiatedClassTree(argTypeDesc);
				}
			}
		}

		// Add the proxy classes and non-primitive return types of all imported
		// service methods to the set of instantiated classes
		for (ImportDefinition impDef : domainDef.getImports()) {
			IMClass proxyClass = classStore.getClass(impDef.getResourceClass());
			addInstantiatedClass(proxyClass);
			IMClass iface = impDef.getService().getServiceInterface(classStore);
			for (IMMethod method : iface.getMethods()) {
				addInstantiatedClassTree(method.getSource().getReturnType());
			}
		}

		// Start at the entry points of the domain
		super.runOnDomain(domain);

		// It is important to process the SSA edges first - otherwise a method
		// body may be visited before its formal arguments have been propagated.
		// Likewise, when the analysis stops in the middle of a basic block
		// because of a method invocation, the control-flow edges to the
		// callee(s) must be processed before continuing within the basic block.
		for (;;) {
			if (ssaWorkList.isEmpty() == false)
				processSSAEdge(ssaWorkList.poll());
			else if (flowWorkList.isEmpty() == false)
				processCFGEdge(flowWorkList.poll());
			else if (midBBStack.isEmpty() == false)
				processMidBBState(midBBStack.pop());
			else
				break;
		}
		return false;
	}
	
	/** Returns the flow-lattice information store. */
	public FlowLatticeInfoStore getFlowLatticeInfoStore() {
		return flowInfoStore;
	}

	@Override
	public Void visit(IMMethod method) throws CompileException {

		// If the method body is empty, do not analyze it further
		if (method.hasCode() == false)
			return null;

		// Add the first basic block to the CFG work list
		Iterator<IMBasicBlock> it = method.iterator();
		if (it.hasNext())
			flowWorkList.add(new CFGEdge(null, it.next()));
		return null;
	}

	@Override
	public Void visit(IMLocalPortalMethod method) throws CompileException {
		ClassStore repository = method.getClassStore();
		ServiceDefinition service = method.getImportDefinition().getService();
		IMClass serviceClass = service.getServiceClass(repository);
		addMethod(serviceClass.getMethod(method.getMethodNameAndType()));
		return null;
	}

	@Override
	public Void visit(IMMethodNotFound method) throws CompileException {
		return null;
	}

	@Override
	public Void visit(IMPortalConstructor method) throws CompileException {

		ClassStore       repository  = method.getClassStore();
		ImportDefinition impDef      = method.getImportDefinition();
		IMClass          driverClass = repository.getClass(impDef.getDriver());
		addMethod(driverClass.getMethod("<init>()V"));

		// Call all specified bean methods
		NetworkDefinition netDef = impDef.getNetwork();
		for (String m : netDef.getDriverInitMethods()) {
			addMethod(driverClass.getMethod(m));
		}

		IMClass protoClass = repository.getClass(netDef.getProtocolName());
		String  ifaceName  = netDef.getDriverInterfaceName();
		addMethod(protoClass.getMethod("<init>(L" + ifaceName + ";IIIBB)V"));

		// Allocate a buffer for sending and receiving
		addMethod(protoClass.getMethod("allocReceiveBuffer(ISI)Z"));
		IMMethod allocTransmitBuffer
		= protoClass.getMethod("allocTransmitBuffer(SI)Lkeso/core/Memory;");
		addMethod(allocTransmitBuffer);
		return null;
	}

	@Override
	public Void visit(IMRemotePortalMethod method) throws CompileException {
		ImportDefinition impDef    = method.getImportDefinition();
		String           protoName = impDef.getNetwork().getProtocolName();
		IMClass protoClass = method.getClassStore().getClass(protoName);
	    addMethod(protoClass.getMethod("write(ISI)Z"));
	    addMethod(protoClass.getMethod("read(I)Lkeso/core/Memory;"));
		return null;
	}

	@Override
	protected void analyzeMethod(IMMethod method) throws CompileException {
		if (method.isConstructor())
			addInstantiatedClass(method.getIMClass());
		method.accept(this);
	}

	private void processSSAEdge(IMNode root)
	throws CompileException {
		currInfo.invalidateCells(root);
		IMBasicBlock bb = nodeBlocks.get(root);
		BuilderOptions opts = root.getOptions();
		opts.vverbose("SSA edge: " + bb + " (" + root.toReadableString() + ')');
		visit(root, bb);
	}

	private void processCFGEdge(CFGEdge edge) throws CompileException {

		IMBasicBlock from = edge.from;
		IMBasicBlock bb   = edge.to;
		bb.getOptions().vverbose("CFG edge: " + from + " -> " + bb);

		// Do not visit the basic block again if the incoming CFG edge has
		// already been taken
		if (currInfo.isExecutable(from, bb))
			return;

		// Mark the current CFG edge as executable
		currInfo.markExecutable(from, bb);
		bb.addDomain(currDomain);

		// Make sure the Phi-functions are revisited
		if (from != null) {
			for (IMNode phiRoot : blockPhis.get(bb)) {
				currInfo.invalidateCells(phiRoot);
			}
		}

		processBasicBlock(bb, null);
	}

	private void processMidBBState(MidBBState state) throws CompileException {

		IMBasicBlock bb       = state.bb;
		IMNode       skipRoot = state.lastVisited;
		bb.getOptions().vverbose("Mid-BB state: " + bb);

		// Resume processing the basic block unless another Mid-BB state at the
		// same position has been restored in the meantime
		if (skippedRoots.add(skipRoot) == true)
			processBasicBlock(bb, skipRoot);
	}

	private void processBasicBlock(IMBasicBlock bb, IMNode skipRoot)
			throws CompileException {

		explicitBranch = false;

		// Visit all nodes in the basic block
		for (IMNode root : bb) {

			// If we are resuming, then skip the expression trees we have
			// already processed
			if (skipRoot != null) {
				if (root == skipRoot)
					skipRoot = null;
				continue;
			}

			int oldFlowWorkListSize = flowWorkList.size();

			// Visit the current expression tree
			visit(root, bb);

			// If a new edge has been added to the flow worklist in the
			// meantime, we have to process that edge first before resuming with
			// our current basic block. This is because we have to visit a method
			// invocation before accessing its return value.
			if (flowWorkList.size() == oldFlowWorkListSize || root.next() == null)
				continue;
			midBBStack.push(new MidBBState(bb, root));
			return;
		}

		// Add fall-through successors if the block did not contain a branch
		// expression
		if (explicitBranch == false) {
			for (IMBasicBlock succ : bb.getSucc()) {
				flowWorkList.add(new CFGEdge(bb, succ));
			}
		}
	}

	/**
	 * Visits an expression tree.
	 * @param root Root of the expression tree.
	 * @param bb   Basic block.
	 * @throws CompileException
	 */
	private void visit(final IMNode root, final IMBasicBlock bb)
	throws CompileException {

		nodeBlocks.put(root, bb);

		root.visitNodes(new IMBasicVisitor() {

			@Override
			public void visit(IMNode node) throws CompileException {

				// Field read/write: Add the field (if it is static) and the
				// class owning it to the set of reachable fields/classes
				IMField field    = null;
				boolean isStatic = false;
				if (node.isFieldAccess()) {
					field = ((IMFieldRead) node).getField();
					currInfo.addFieldUseRoot(field, root);
					if (node instanceof IMGetStatic)
						isStatic = true;
				} else if (node instanceof IMFieldWrite) {
					field = ((IMFieldWrite) node).getField();
					if (node instanceof IMPutStatic)
						isStatic = true;
				}
				if (field != null) {
					if (isStatic)
						addStaticField(field, bb.getMethod());
					addClass(field.getType());
				}

				// Method invocation: Add all possible candidates to the set of
				// reachable methods and add the corresponding edges to the work
				// lists
				if (node instanceof IMInvoke) {
					visitInvoke(node, root, bb);
					return;
				}

				// Local variable access: Add the current root node
				// to the list of trees using the variable
				if (node.usesSlot())
					currInfo.addSlotUseRoot(node.getIMSlot(), root);

				// Phi-function: Add the current root node to the list of Phi
				// roots for this basic block
				if (node instanceof IMPhi)
					blockPhis.add(bb, root);

				// Compute the new lattice cell and perform path analysis
				IMDataFlow oldCell = currInfo.getCell(node);
				IMDataFlow cell    = currInfo.computeCell(node, bb);
				if (cell == null)
					return;
				doPathAnalysis(node, bb);

				// Do nothing else if the lattice cell remains unchanged
				if (cell.equals(oldCell))
					return;

				// Store of a local variable: Add all affected node trees to the
				// SSA work list
				if (node instanceof IMStoreLocalVariable) {
					IMSlot slot = node.getIMSlot();
					/*System.err.print("UPDATING SLOT " + slot + " IN " + bb + " (");
					if (oldCell == null)
						System.err.print(oldCell);
					else
						System.err.print(oldCell.toReadableString());
					System.err.println(" -> " + cell.toReadableString() + ')');*/ // FIXME
					for (IMNode useRoot : currInfo.getSlotUseRoots(slot)) {
						ssaWorkList.add(useRoot);
					}
					return;
				}

				// Field write: Add all affected node trees to the SSA work list
				if (node instanceof IMFieldWrite) {
					//System.err.println("UPDATING FIELD " + field); // FIXME
					for (IMNode useRoot : currInfo.getFieldUseRoots(field)) {
						ssaWorkList.add(useRoot);
					}
					return;
				}

				// Branch: Add all executable successors to the flow work
				// list
				if (node.isBranch()) {

					// If the branch is conditional, evaluate the branch
					// condition and only take the ones that can actually be
					// taken
					IMBranch branch = (IMBranch) node;
					if (branch instanceof IMConditionalBranch
							&& cell.isConstant()) {
						IMBasicBlock[] targets = branch.getTargets();
						int index = cell.getValue().getIntValue();
						flowWorkList.add(new CFGEdge(bb, targets[index]));
					} else {
						for (IMBasicBlock target : branch.getUniqTargets()) {
							flowWorkList.add(new CFGEdge(bb, target));
						}
					}
					explicitBranch = true;
					return;
				}

				// "New" operation (= allocation) or constant object: Add class
				// to list of instantiated classes
				if (node.isNew() || node.hasConstant()) {
					IMClass imClass = node.determineType();
					if (imClass != null)
						addInstantiatedClass(imClass);
					return;
				}

			}
		});
	}

	private void visitInvoke(IMNode node, IMNode root, IMBasicBlock bb)
	throws CompileException {

		// If this is an interface invocation, add the interface
		// method and make sure it does not get removed
		IMInvoke invoke = (IMInvoke) node;
		if (invoke instanceof IMInvokeInterface) {
			IMMethod ifaceMethod = invoke.getCallee();
			addMethod(ifaceMethod);
			ifaceMethod.calledFrom(null, null);
		}

		// Add the root of the current expression tree to the list of call-site
		// roots for the corresponding class, so we can re-visit these sites
		// later if a new candidate class is instantiated
		currInfo.addCallSiteRoot(invoke.getCalleeClass(), root);
		for (IMMethod candidate : invoke.getCandidates()) {
			currInfo.addCallSiteRoot(candidate.getIMClass(), root);
		}

		// Iterate over all candidate methods
		for (IMMethod candidate : currInfo.computeCandidates(invoke, bb)) {

			// Add the candidate to the set of reachable methods
			addMethod(candidate);

			// If the method body is to be ignored, do not analyze
			// it further
			if (candidate.hasCode() == false)
				continue;

			// If the candidate method is non-void, connect its
			// return value to the data flow info of the invoke
			// statement
			IMSlot retVal = candidate.recomputeReturnValue();
			if (retVal != null) {
				currInfo.addSlotUseRoot(retVal, root);
				currInfo.addReturnSlot(invoke, retVal);
			}

			IMMethodFrame frame = candidate.getMethodFrame();
			IMSlot[] formalArgs = frame.getOriginalArguments();
			List<IMNode> actualArgs = getActualArgs(invoke);

			// Process the arguments
			for (int i = 0; i < formalArgs.length; ++i) {

				// Skip argument if it is unused
				IMSlot formalArg = formalArgs[i];
				if (frame.isArgUnused(formalArg))
					continue;

				// If the formal argument is undefined (i.e. this is the first
				// detected call of the candidate method), create a new special
				// Phi-function for the formal argument
				IMArgPhi argPhi;
				IMNode def = formalArg.getDefiningStatement();
				if (def == null) {
					int type = formalArg.getType();
					argPhi = new IMArgPhi(candidate, type);
					def = candidate.createStoreVariable(formalArg, argPhi, -1);
					Iterator<IMBasicBlock> it = candidate.iterator();
					IMBasicBlock bb1 = null;
					if (it.hasNext())
						bb1 = it.next();
					nodeBlocks.put(def, bb1);
					formalArg.define(bb1, def);
				} else {
					argPhi = (IMArgPhi)
							((IMStoreLocalVariable) def).getOperand();
				}

				// Connect actual argument and formal argument
				IMNode actualArg = actualArgs.get(i);
				argPhi.addInEdge(actualArg);
				ssaWorkList.add(def);

				// If the actual argument is a local variable, add
				// its use in the callee to the set of use roots for
				// that variable
				if (actualArg.isVariable())
					currInfo.addSlotUseRoot(actualArg.getIMSlot(), def);
			}
		}
	}

	private void doPathAnalysis(IMNode node, IMBasicBlock bb)
	throws CompileException {

		// Only consider conditional branches
		if (node instanceof IMConditionalBranch == false)
			return;

		IMNode     lhs        = node.getLeftNode();
		IMNode     rhs        = node.getRightNode();
		IMDataFlow lhsCell    = currInfo.getCell(lhs);
		IMDataFlow rhsCell    = currInfo.getCell(rhs);
		boolean    isLhsConst = lhsCell.isConstant();
		boolean    isRhsConst = rhsCell.isConstant();
		IMMethod   method     = node.getMethod();
		IMDataFlow trueCell   = null;
		IMDataFlow falseCell  = null;
		IMConditionalBranch condBranch = (IMConditionalBranch) node;

		if (lhs.isVariable()) {
			IMSlot slot = lhs.getIMSlot();
			if (isLhsConst == false && isRhsConst) {
				IMConstant rhsConst = rhsCell.getValue();
				int rhsValue = 0;
				if (rhs.getDatatype() != BCBasicDatatype.REFERENCE)
					rhsValue = rhsConst.getIntValue();
				if (node instanceof IMEQConditionalBranch
						|| node instanceof IMNullCheck) {           // ==
					trueCell  = createPathCell(lhs, rhs, rhsConst, bb, true);
					falseCell = createPathCell(lhs, rhs, rhsConst, bb, false);
				} else if (node instanceof IMNEConditionalBranch) { // !=
					trueCell  = createPathCell(lhs, rhs, rhsConst, bb, false);
					falseCell = createPathCell(lhs, rhs, rhsConst, bb, true);
				} else if (node instanceof IMLTConditionalBranch) { // <
					trueCell = createInterval(Integer.MIN_VALUE, rhsValue - 1,
							method);
					falseCell = createInterval(rhsValue, Integer.MAX_VALUE,
							method);
				} else if (node instanceof IMLEConditionalBranch) { // <=
					trueCell = createInterval(Integer.MIN_VALUE, rhsValue,
							method);
					falseCell = createInterval(rhsValue + 1, Integer.MAX_VALUE,
							method);
				} else if (node instanceof IMGTConditionalBranch) { // >
					trueCell = createInterval(rhsValue + 1, Integer.MAX_VALUE,
							method);
					falseCell = createInterval(Integer.MIN_VALUE, rhsValue,
							method);
				} else if (node instanceof IMGEConditionalBranch) { // >=
					trueCell = createInterval(rhsValue, Integer.MAX_VALUE,
							method);
					falseCell = createInterval(Integer.MIN_VALUE, rhsValue - 1,
							method);
				}
			}
			setPathCells(condBranch, bb, slot, trueCell, falseCell);
		}

		if (rhs.isVariable()) {
			IMSlot slot = rhs.getIMSlot();
			if (isLhsConst && isRhsConst == false) {
				IMConstant lhsConst = lhsCell.getValue();
				int lhsValue = 0;
				if (lhs.getDatatype() != BCBasicDatatype.REFERENCE)
					lhsValue = lhsConst.getIntValue();
				if (node instanceof IMEQConditionalBranch
						|| node instanceof IMNullCheck) {           // ==
					trueCell  = createPathCell(rhs, lhs, lhsConst, bb, true);
					falseCell = createPathCell(rhs, lhs, lhsConst, bb, false);
				} else if (node instanceof IMNEConditionalBranch) { // !=
					trueCell  = createPathCell(rhs, lhs, lhsConst, bb, false);
					falseCell = createPathCell(rhs, lhs, lhsConst, bb, true);
				} else if (node instanceof IMLTConditionalBranch) { // <
					trueCell = createInterval(lhsValue + 1, Integer.MAX_VALUE,
							method);
					falseCell = createInterval(Integer.MIN_VALUE, lhsValue,
							method);
				} else if (node instanceof IMLEConditionalBranch) { // <=
					trueCell = createInterval(lhsValue, Integer.MAX_VALUE,
							method);
					falseCell = createInterval(Integer.MIN_VALUE, lhsValue + 1,
							method);
				} else if (node instanceof IMGTConditionalBranch) { // >
					trueCell = createInterval(Integer.MIN_VALUE, lhsValue - 1,
							method);
					falseCell = createInterval(lhsValue, Integer.MAX_VALUE,
							method);
				} else if (node instanceof IMGEConditionalBranch) { // >=
					trueCell = createInterval(Integer.MIN_VALUE, lhsValue,
							method);
					falseCell = createInterval(lhsValue - 1, Integer.MAX_VALUE,
							method);
				}
			}
			setPathCells(condBranch, bb, slot, trueCell, falseCell);
		}
	}

	private void setPathCells(IMConditionalBranch condBranch, IMBasicBlock bb,
			IMSlot slot, IMDataFlow trueCell, IMDataFlow falseCell)
					throws CompileException {
		setPathCells(condBranch, bb, 1, slot, trueCell);
		setPathCells(condBranch, bb, 0, slot, falseCell);
		pathStarts.add(condBranch);
	}

	private void setPathCells(IMConditionalBranch condBranch, IMBasicBlock bb,
			int targetIndex, IMSlot slot, IMDataFlow cell)
	throws CompileException {

		// If the conditional branch does not have a "true-target", there is
		// nothing we can (or need to) do
		IMBasicBlock[] targets = condBranch.getTargets();
		if (targetIndex >= targets.length)
			return;

		// If any of the target's predecessor blocks is dominated by another
		// successor of the basic block that contains the conditional branch,
		// there is another path to the target for which the cell is not valid -
		// so we can't do anything here.
		IMMethod      method   = condBranch.getMethod();
		DominatorTree domTree  = method.getDomTree();
		IMBasicBlock  targetBB = targets[targetIndex];
		for (IMBasicBlock targetPred : targetBB.getPred()) {
			for (IMBasicBlock branchSucc : bb.getSucc()) {
				if (branchSucc == targetPred
						|| domTree.dom(branchSucc, targetPred))
					return;
			}
		}

		// If the target is not strictly dominated by the branch block, empty
		// the path cell - there is a live path around the branch block that
		// leads to the target.
		if (targetBB == bb || domTree.dom(bb, targetBB) == false)
			cell = null;

		// Setting an empty path cell only makes sense if we explicitly want to
		// remove it - which is the case if we are setting an empty cell for a
		// path whose start (conditional branch) we have already visited
		if (cell == null && pathStarts.contains(condBranch) == false)
			return;

		// Set the path cell for the target block and all other blocks dominated
		// by it
		Set<IMBasicBlock> changedBBs = new HashSet<IMBasicBlock>();
		for (IMBasicBlock domBB : method) {
			if ((domBB.getLiveInSlots().contains(slot) == false
					&& domBB.getPhiInSlots().contains(slot) == false)
					|| (domBB != targetBB
							&& domTree.dom(targetBB, domBB) == false))
				continue;
			if (currInfo.setPathCell(slot, domBB, targetBB, cell) == true)
				changedBBs.add(domBB);
			//System.err.println("SETTING " + slot + " TO " + ((cell != null) ? cell.toReadableString() : cell) + " IN " + domBB + " (comparison in " + bb + ")"); // FIXME
		}

		// Add all trees to the SSA work list that use the given slot and reside
		// in basic blocks whose path cell has changed
		for (IMNode useRoot : currInfo.getSlotUseRoots(slot)) {
			if (changedBBs.contains(nodeBlocks.get(useRoot)))
				ssaWorkList.add(useRoot);
		}
	}

	private void addInstantiatedClassTree(BasicTypeDescriptor typeDesc) {

		// Find the class corresponding to the descriptor (if any)
		if (typeDesc.isPrimitive())
			return;
		String className;
		if (typeDesc.isArray())
			className = typeDesc.getFullArrayTypeDesc();
		else
			className = typeDesc.getClassName();
		IMClass type = classStore.getClass(className);
		if (type == null)
			return;

		// Add the class and all of its sub-classes
		for (IMClass imClass : classStore.getAllClasses()) {
			if (imClass.typeOf(type) == true)
				addInstantiatedClass(imClass);
		}
	}

	private void addInstantiatedClass(IMClass imClass) {

		// Recursively add all super-classes.
		// If the classes have not been instantiated before, trigger the
		// re-evaluation of all virtual call sites of their methods.
		for (IMClass superClass = imClass.getSuperClass(); superClass != null;
				superClass = superClass.getSuperClass()) {
			if (currInfo.addInheritedClass(superClass) == true)
				triggerReevaluation(superClass);
		}

		// An abstract class cannot be instantiated
		if (imClass.isAbstract())
			return;

		// Add the class itself.
		// If the class has not been instantiated before, trigger the
		// re-evaluation of all virtual call sites of its methods.
		if (currInfo.addInstantiatedClass(imClass) == true)
			triggerReevaluation(imClass);
	}

	private void triggerReevaluation(IMClass imClass) {
		for (IMNode callSiteRoot : currInfo.getCallSiteRoots(imClass)) {
			ssaWorkList.add(callSiteRoot);
		}
	}

	private IMDataFlow createPathCell(IMNode varNode, IMNode constNode,
			IMConstant constVal, IMBasicBlock bb, boolean equals)
	throws CompileException {

		// Primitive data type
		if (constVal.getDatatype() != BCBasicDatatype.REFERENCE) {
			if (equals == true) // == constVal
				return new IMDataFlowConstValue(constVal);
			else                // != constVal
				return null;
		}

		// Object
		IMClass type = currInfo.getCell(varNode).getType();
		if (constVal.isNullValue()) {
			if (equals == true) // == null
				return IMDataFlow.createDataFlowObj(constNode, bb, type);
			else                // != null
				return IMDataFlow.createValidDataFlowObj(varNode, bb, type);
		} else {
			if (equals == true) // != null
				return IMDataFlow.createValidDataFlowObj(varNode, bb, type);
			else                // == null
				return IMDataFlow.createDataFlowObj(constNode, bb, type);
		}
	}

	private static List<IMNode> getActualArgs(IMInvoke invoke) {
		List<IMNode> actualArgs = new ArrayList<IMNode>();
		IMNode object = invoke.getObject();
		if (object != null)
			actualArgs.add(object);
		for (IMNode arg : invoke.getArguments()) {
			if (arg != null)
				actualArgs.add(arg);
		}
		return actualArgs;
	}

	private static IMDataFlow createInterval(int lowerBound, int upperBound,
			IMMethod method) {
		if (lowerBound > upperBound) // Over-/underflow
			return null;
		if (lowerBound == upperBound) {
			IMConstant constant = method.createIMIConstant(lowerBound, -1);
			return new IMDataFlowConstValue(constant);
		}
		BuilderOptions opts = method.getOptions();
		return new IMDataFlowIntRange(lowerBound, upperBound, opts);
	}

}
