/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;
import java.util.Map.*;

import keso.classfile.constantpool.*;
import keso.classfile.datatypes.*;
import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.kni.*;
import keso.util.*;

/** Flow-lattice information class. */
public final class FlowLatticeInfo {

	private final Map<IMNode, IMBasicBlock> nodeBlocks;

	private final Map<IMNode, IMDataFlow> cells = new HashMap<IMNode, IMDataFlow>();
	private final Set<IMNode> invalidated = new HashSet<IMNode>();

	private final Map<IMSlot, IMDataFlow> slots = new HashMap<IMSlot, IMDataFlow>();

	private final Multimap<IMInvoke, IMSlot> retVals
	= new HashUniqueMultimap<IMInvoke, IMSlot>();
	private Map<IMField, IMArgPhi> fields = new HashMap<IMField, IMArgPhi>();

	private final SplitKeySet<IMBasicBlock, IMBasicBlock> executableEdges
	= new SplitKeyHashSet<IMBasicBlock, IMBasicBlock>();

	private final Multimap<IMSlot, IMNode> slotUseRoots
	= new HashUniqueMultimap<IMSlot, IMNode>();
	private final Multimap<IMField, IMNode> fieldUseRoots
	= new HashUniqueMultimap<IMField, IMNode>();
	private final Multimap<IMClass, IMNode> callSiteRoots
	= new HashUniqueMultimap<IMClass, IMNode>();

	private final SlotInfo pathInfo = new SlotInfo();

	// Instantiated and inherited classes.
	private final Collection<IMClass> instClasses      = new HashSet<IMClass>();
	private final Collection<IMClass> inheritedClasses = new HashSet<IMClass>();
	
	// Mapping (call site -> candidate methods).
	private final Multimap<IMInvoke, IMMethod> invokeCandidates
	= new HashMultimap<IMInvoke, IMMethod>();

	/**
	 * Constructor.
	 * @param nodeBlocks Mapping from nodes to basic blocks containing them.
	 */
	FlowLatticeInfo(Map<IMNode, IMBasicBlock> nodeBlocks) {
		this.nodeBlocks = nodeBlocks;
	}

	/**
	 * Returns the lattice cell associated with a given node.
	 * @param node The node.
	 * @return The lattice cell if it has already been computed,
	 *         <code>null</code> otherwise.
	 */
	public IMDataFlow getCell(IMNode node) {
		return cells.get(node);
	}

	/**
	 * Determines if a given class is instantiated in this domain.
	 * @param imClass The class.
	 * @return <tt>true</tt> if the class is instantiated, <tt>false</tt>
	 *         otherwise.
	 */
	public boolean isClassInstantiated(IMClass imClass) {
		return instClasses.contains(imClass);
	}
	
	/**
	 * Returns the list of candidates for a method invocation.
	 * @param invoke Invocation instruction.
	 * @return List of candidate methods.
	 */
	public Collection<IMMethod> getCandidates(IMInvoke invoke)
			throws CompileException {
		Collection<IMMethod> candidates = invokeCandidates.get(invoke);
		if (candidates.size() == 0) { // Sanity check
			if (invoke instanceof IMInvokeNonUnique) {
				throw new CompileException("No candidates for "
						+ invoke.toReadableString() + " in "
						+ invoke.getMethod());
			}
			candidates.add(invoke.getCandidates().iterator().next());
		}
		return candidates;
	}
	
	/**
	 * Replaces an instruction node with another node.
	 * @param oldNode Old instruction.
	 * @param newNode New instruction.
	 */
	void replaceNode(IMNode oldNode, IMNode newNode) {
		
		// Update the flow-lattice cell map
		IMDataFlow cell = cells.remove(oldNode);
		if (cell != null)
			cells.put(newNode, cell);
		
		// If the node is an invocation, also update the callee-candidate map
		if (oldNode instanceof IMInvoke) {
			Collection<IMMethod> candidates
			= invokeCandidates.removeAll((IMInvoke) oldNode);
			if (candidates != null)
				invokeCandidates.addAll((IMInvoke) newNode, candidates);
		}
	}

	/**
	 * Invalidates all nodes of an expression tree so their cells will be
	 * re-computed when accessed for the next time.
	 * @param root Root node of the tree.
	 * @throws CompileException
	 */
	void invalidateCells(IMNode root) throws CompileException {
		root.visitNodes(new IMBasicVisitor() {
			@Override
			public void visit(IMNode node) throws CompileException {
				invalidated.add(node);
			}
		});
	}

	/**
	 * Computes the lattice cell for a given node.
	 * If the cell has already been computed, a cached result is returned.
	 * @param node The node.
	 * @param bb   Basic block containing the node.
	 * @return The computed lattice cell.
	 * @throws CompileException
	 */
	IMDataFlow computeCell(IMNode node, IMBasicBlock bb)
	throws CompileException {

		// If the cell has already been computed and has not been invalidated
		// since then, just return its saved value. Otherwise, recompute it.
		IMDataFlow cell = null;
		if (invalidated.contains(node) == false) {
			cell = getCell(node);
			if (cell != null)
				return cell;
		} else {
			invalidated.remove(node);
		}

		IMMethod method = node.getMethod();
		int bcPos = node.getBCPosition();

		if (node.usesSlot()) {

			cell = getSlotCell(node.getIMSlot(), bb);

		} else if (node instanceof IMInvoke) {

			cell = visitInvoke((IMInvoke) node, bb);

		} else if (node instanceof IMStoreLocalVariable) {

			IMStoreLocalVariable store = (IMStoreLocalVariable) node;
			IMNode operand = store.getOperand();
			if (operand instanceof IMPhi)
				cell = visitPhi(store, bb);
			else if (operand instanceof IMArgPhi)
				cell = visitArgPhi(store, bb);
			else
				cell = computeCell(operand, bb);
			setSlotCell(node.getIMSlot(), cell);

		} else if (node.isFieldAccess()) {

			cell = visitFieldRead((IMFieldRead) node, bb);

		} else if (node instanceof IMFieldWrite) {

			IMField field = ((IMFieldWrite) node).getField();
			if (field.isFinal()) {
				IMArgPhi fieldPhi = fields.get(field);
				if (fieldPhi == null) {
					fieldPhi = new IMArgPhi(method, node.getDatatype());
					fields.put(field, fieldPhi);
				}
				IMNode rValue = node.getRightNode();
				fieldPhi.addInEdge(rValue);
				cell = computeCell(rValue, bb);
			}

		} else if (node instanceof IMNewArray) {

			IMNode size = ((IMNewArray) node).getSize();
			cell = IMDataFlow.createArrayDataFlowObj(node, size, bb);

		} else if (node instanceof IMNewObjArray) {

			IMNode size = ((IMNewObjArray) node).getSize();
			cell = IMDataFlow.createArrayDataFlowObj(node, size, bb);

		} else if (node instanceof IMArrayLength) {

			IMNode     array     = ((IMArrayLength) node).getArray();
			IMDataFlow arrayCell = computeCell(array, bb);
			if (arrayCell.isArray())
				cell = ((IMDataFlowArray) arrayCell).getLength();

		} else if (node instanceof IMCheckCast) {

			cell = computeCell(node.getRightNode(), bb);
			IMClass castType = node.determineType();
			IMClass opType   = cell.getType();
			if (castType.typeOf(opType) == true) // Downcast
				cell = null;

		} else if (node instanceof IMAMemConstant) {

			IMAMemConstant memConst = (IMAMemConstant) node;
			IMNode addr = method.createIMIConstant(memConst.getAddr(), bcPos);
			IMNode size = method.createIMIConstant(memConst.getSize(), bcPos);
			cell = IMDataFlow.createAMemDataFlowObj(node, addr, size, bb);

		} else if (node instanceof IMUnaryNode) {

			cell = evaluate(node, bb, computeCell(node.getRightNode(), bb));

		} else if (node instanceof IMBinaryNode
				|| node instanceof IMConditionalBranch) {

			cell = visitBinaryNode(node, bb);

		} else if (node instanceof IMInstanceOf) {

			IMClass opType = computeCell(node.getRightNode(), bb).getType();
			if (opType != null) {
				IMClass checkedType = node.determineType();
				int value = -1;
				if (opType.typeOf(checkedType) == true)
					value = 1;
				else if (checkedType.typeOf(opType) == false)
					value = 0;
				if (value != -1) {
					IMConstant result = method.createIMIConstant(value, bcPos);
					cell = IMDataFlow.createDataFlowObj(result, bb);
				}
			}
		}

		return putCell(cell, node, bb);
	}

	/**
	 * Marks a CFG as executable.
	 * @param from Source block of the edge.
	 * @param to   Destination block of the edge.
	 */
	void markExecutable(IMBasicBlock from, IMBasicBlock to) {
		executableEdges.add(from, to);
	}

	/**
	 * Determines if a CFG edge is marked as executable.
	 * @param from Source block of the edge.
	 * @param to   Destination block of the edge.
	 * @return <code>true</code> if the edge has been marked as executable,
	 *         <code>false</code> otherwise.
	 */
	boolean isExecutable(IMBasicBlock from, IMBasicBlock to) {
		return executableEdges.contains(from, to);
	}

	/**
	 * For a given stack slot, adds a syntax tree to the list of trees in which
	 * that slot is used.
	 * @param slot The slot.
	 * @param root Root node of the tree.
	 */
	void addSlotUseRoot(IMSlot slot, IMNode root) {
		slotUseRoots.add(slot, root);
	}

	/**
	 * Returns the root nodes of all syntax trees in which a given stack slot is
	 * used.
	 * @param slot The slot.
	 * @return List of root nodes.
	 */
	Collection<IMNode> getSlotUseRoots(IMSlot slot) {
		return slotUseRoots.get(slot);
	}

	/**
	 * For a given field, adds a syntax tree to the list of trees in which that
	 * field is used.
	 * @param field The field.
	 * @param root Root node of the tree.
	 */
	void addFieldUseRoot(IMField field, IMNode root) {
		fieldUseRoots.add(field, root);
	}

	/**
	 * Returns the root nodes of all syntax trees in which a given field is
	 * used.
	 * @param field The field.
	 * @return List of root nodes.
	 */
	Collection<IMNode> getFieldUseRoots(IMField field) {
		return fieldUseRoots.get(field);
	}

	void addCallSiteRoot(IMClass imClass, IMNode root) {
		callSiteRoots.add(imClass, root);
	}

	Collection<IMNode> getCallSiteRoots(IMClass imClass) {
		return callSiteRoots.get(imClass);
	}

	/**
	 * Adds a return slot for a given method invocation.
	 * The slot is the local variable within the method body that receives the
	 * return value.
	 * @param invoke The invocation instruction.
	 * @param slot   The slot.
	 */
	void addReturnSlot(IMInvoke invoke, IMSlot slot) {
		retVals.add(invoke, slot);
	}

	/**
	 * Computes the meet between a set of lattice cells.
	 * @param cells The lattice cells to be met.
	 * @return Result of the meet operator.
	 * @throws CompileException
	 */
	IMDataFlow meet(IMDataFlow... cells) throws CompileException {

		if (cells.length == 0)
			return null;

		// Compute the meet between the lattice cells
		IMDataFlow cell = cells[0];
		for (int i = 1; i < cells.length; ++i) {
			if (cell == null)
				cell = cells[i];
			else if (cells[i] != null)
				cell = cell.fold(cells[i]);
		}
		return cell;
	}

	boolean setPathCell(IMSlot slot, IMBasicBlock bb, IMBasicBlock target,
			IMDataFlow pathCell) throws CompileException {
		return pathInfo.setPathCell(slot, bb, target, pathCell);
	}

	boolean addInstantiatedClass(IMClass imClass) {
		return (instClasses.add(imClass) == true
				&& inheritedClasses.contains(imClass) == false);
	}

	boolean addInheritedClass(IMClass imClass) {
		return (inheritedClasses.add(imClass) == true
				&& instClasses.contains(imClass) == false);
	}
	
	/**
	 * Computes the list of candidates for a method invocation.
	 * @param invoke Invocation instruction.
	 * @param bb     Basic block in which the invocation takes place.
	 * @return List of candidate methods.
	 */
	Collection<IMMethod> computeCandidates(IMInvoke invoke, IMBasicBlock bb)
			throws CompileException {
		
		// If the invocation is non-virtual, add the one candidate
		Iterable<IMMethod> allCandidates = invoke.getCandidates();
		if (invoke instanceof IMInvokeUnique) {
			for (IMMethod candidate : allCandidates) { // Should only be 1 cand.
				invokeCandidates.add(invoke, candidate);
			}
			return invokeCandidates.get(invoke);
		}

		IMNode  obj         = invoke.getObject();
		IMClass objType     = computeCell(obj, bb).getType();
		String  nameAndType = invoke.getCallee().getMethodNameAndType();

		// Look at all classes
		for (IMClass imClass : bb.getMethod().getClassStore().getAllClasses()) {

			// If the class matches, is instantiated and offers a candidate, add
			// the candidate
			boolean isInstantiated = isClassInstantiated(imClass);
			boolean isOfType       = imClass.typeOf(objType);
			if (isOfType) {
				IMMethod method = imClass.getMethod(nameAndType);
				JoinPointChecker joinPoints = imClass.getJoinPoints();
				if (method != null && (isInstantiated
						|| joinPoints.ignoreMethodBody(imClass, method)
						   == true)) {
					if (invokeCandidates.get(invoke).contains(method) == false)
						invokeCandidates.add(invoke, method);
					continue;
				}
			}

			// Otherwise, if the class is instantiated, matches and one of its
			// super-classes offers a candidate, add that candidate
			if (isInstantiated == false || isOfType == false)
				continue;
			for (imClass = imClass.getSuperClass(); imClass != null;
					imClass = imClass.getSuperClass()) {
				IMMethod method = imClass.getMethod(nameAndType);
				if (method == null)
					continue;
				if (invokeCandidates.get(invoke).contains(method) == false)
					invokeCandidates.add(invoke, method);
			}
		}

		return invokeCandidates.get(invoke);
	}
	
	/**
	 * Merges this flow-lattice information with another set of information.
	 * @param other The other flow-lattice information.
	 */
	void merge(FlowLatticeInfo other) throws CompileException {
		for (Entry<IMNode, IMDataFlow> entry : other.cells.entrySet()) {
			IMNode     node  = entry.getKey();
			IMDataFlow oCell = entry.getValue();
			IMDataFlow cell  = getCell(node);
			cells.put(node, meet(cell, oCell));
		}
		for (Entry<IMSlot, IMDataFlow> entry : other.slots.entrySet()) {
			IMSlot slot = entry.getKey();
			setSlotCell(slot, meet(slots.get(slot), entry.getValue()));
		}
		if (other.instClasses != null)
			instClasses.addAll(other.instClasses);
		for (Entry<IMInvoke, Collection<IMMethod>> entry :
			other.invokeCandidates.entrySet()) {
			IMInvoke invoke = entry.getKey();
			for (IMMethod candidate : entry.getValue()) {
				if (invokeCandidates.get(invoke).contains(candidate) == false)
					invokeCandidates.add(invoke, candidate);
			}
		}

		// TODO: fields?
	}

	/**
	 * Sets the lattice cell of a given stack slot.
	 * @param slot The slot.
	 * @param cell The lattice cell.
	 */
	private void setSlotCell(IMSlot slot, IMDataFlow cell) {
		slots.put(slot, cell);
	}

	private IMDataFlow getSlotCell(IMSlot slot, IMBasicBlock bb)
	throws CompileException {

		// Get the slot cell
		IMDataFlow slotCell = slots.get(slot);
		if (slotCell == null)
			return null;

		// If a path cell exists, intersect the slot cell with it
		IMDataFlow bbCell = pathInfo.getBBCell(slot, bb);
		if (bbCell == null)
			return slotCell;
		slotCell = slotCell.intersect(bbCell);

		// If the slot is the first argument of an instance method, it is
		// definitely valid and an instance of the method class
		IMMethod method = bb.getMethod();
		if (method.isStatic() || slot.getVarNumber() != 0
				|| slot.getType() != BCBasicDatatype.REFERENCE)
			return slotCell;
		IMClass dynamicType = slotCell.getType();
		IMClass staticType  = method.getIMClass();
		if (dynamicType.typeOf(staticType) == false)
			dynamicType = staticType;
		return new IMDataFlowValidObj(bb.getOptions(), dynamicType);
	}

	private IMDataFlow visitPhi(IMStoreLocalVariable store, IMBasicBlock bb)
	throws CompileException {
		IMPhi      phi  = (IMPhi) store.getOperand();
		IMDataFlow cell = null;
		IMBasicBlockList preds = bb.getPred();
		for (int i = 0; i < preds.length(); ++i) {
			/*System.err.println("PHI: " + preds.at(i) + " (" + preds.at(i).getDFNum() + ") -> " + bb + " (" + bb.getDFNum() + ")"); // FIXME
			if ((preds.at(i) == bb || bb.getMethod().getDomTree().dom(bb, preds.at(i)) == true))
				System.err.println("  PHI: backward edge detected!");
			if (isExecutable(preds.at(i), bb))
				System.err.println("  PHI: edge " + i + " is executable");*/
			if (isExecutable(preds.at(i), bb))
				cell = meet(cell, computeCell(phi.getSource(i), bb));
		}
		return putCell(cell, phi, bb);
	}

	private IMDataFlow visitArgPhi(IMStoreLocalVariable store, IMBasicBlock bb)
	throws CompileException {

		// Compute the meet of the lattice cells of all actual arguments.
		// Since the ArgPhi may contain references to nodes that are only
		// reachable in other domains, skip those which have not been visited
		// (yet) in the current domain.
		IMArgPhi   argPhi = (IMArgPhi) store.getOperand();
		IMDataFlow cell   = null;
		for (IMNode node : argPhi.getInEdges()) {
			if (getCell(node) != null)
				cell = meet(cell, computeCell(node, bb));
		}

		// If the cell belongs to the first formal argument of a non-static
		// method, it can only be valid
		if (cell != null && cell.isValid() == false) {
			IMMethod method = store.getMethod();
			if (method.isStatic() == false) {
				IMSlot[] args = method.getMethodFrame().getOriginalArguments();
				if (store.getIMSlot() == args[0]) {
					IMClass type = cell.getType();
					cell = new IMDataFlowValidObj(type.getOptions(), type);
				}
			}
		}
		return putCell(cell, argPhi, bb);
	}

	private IMDataFlow visitInvoke(IMInvoke invoke, IMBasicBlock bb)
	throws CompileException {

		// Compute the meet of the return values of all candidates that have
		// already been visited in this domain
		IMDataFlow cell = null;
		for (IMSlot slot : retVals.get(invoke)) {
			IMDataFlow slotCell = getSlotCell(slot, bb);
			if (slotCell != null)
				cell = meet(cell, slotCell);
		}
		if (cell == null)
			return null;

		// If the callee returns an object, the return value is guaranteed to be
		// at least of the formal return type
		IMMethod callee    = invoke.getCallee();
		int      basicType = callee.getBasicReturnType();
		if (basicType == BCBasicDatatype.REFERENCE) {
			IMClass formalType = invoke.getReturnType();
			if (formalType != null && formalType.typeOf(cell.getType()))
				cell = cell.copy(formalType);
		}

		// Special hard-coded cases: If the invocation is a shared-memory
		// lookup, remember the name of the SHM object. Likewise, if it is a
		// message allocation or reception, remember the name of the message
		// object.
		Collection<IMMethod> candidates = computeCandidates(invoke, bb);
		if (candidates.size() != 1)
			return cell;
		IMMethod candidate  = candidates.iterator().next();
		String   className  = candidate.getIMClass().getClassName();
		String   methodName = candidate.getMethodNameAndType();
		if (className.equals("ciao/sharedmemory/SharedMemoryService")
				&& methodName.equals("getSharedMemoryByName(Ljava/lang/String;)"
						+ "Lkeso/core/Memory;")) {
			IMDataFlow nameCell = computeCell(invoke.getArguments()[0], bb);
			return new IMDataFlowSHM(cell, nameCell);
		} else if (className.startsWith("ciao/messageports/MemorySenderPort_")
				&& methodName.equals("allocateMessage()Lkeso/core/Memory;")) {
			return createMessageCell(invoke, bb, cell, className,
					"ciao/messageports/MemorySenderPort_");
		} else if (className.startsWith("ciao/messageports/MemoryReceiverPort_")
				&& methodName.equals("receiveMessage()Lkeso/core/Memory;")) {
			return createMessageCell(invoke, bb, cell, className,
					"ciao/messageports/MemoryReceiverPort_");
		}
		return cell;
	}

	private IMDataFlow visitFieldRead(IMFieldRead read, IMBasicBlock bb)
	throws CompileException {
		IMField field = read.getField();
		if (field.isFinal() == false)
			return null;
		IMArgPhi fieldPhi = fields.get(field);
		if (fieldPhi == null)
			return null;
		IMDataFlow cell = null;
		for (IMNode node : fieldPhi.getInEdges()) {
			cell = meet(cell, computeCell(node, nodeBlocks.get(node)));
		}
		return cell;
	}

	private IMDataFlow visitBinaryNode(IMNode node, IMBasicBlock bb)
	throws CompileException {

		IMNode lhsNode = node.getLeftNode();
		IMNode rhsNode = node.getRightNode();
		IMDataFlow lhsCell = computeCell(lhsNode, bb);
		IMDataFlow rhsCell = computeCell(rhsNode, bb);
		BuilderOptions opts = node.getOptions();

		// If at least one of the two operand cells is a set of constant values,
		// the result is a set as well
		if (lhsCell.isSet()) {
			IMDataFlow set = new IMDataFlowSet(opts, node.getDatatype());
			for (IMDataFlow lhsElement : (IMDataFlowSet) lhsCell) {
				if (rhsCell.isSet()) {
					for (IMDataFlow rhsElement : (IMDataFlowSet) rhsCell) {
						IMDataFlow resultCell
						= evaluate(node, bb, lhsElement, rhsElement);
						if (resultCell.isConstant() == false)
							return evaluate(node, bb, lhsCell, rhsCell);
						set = set.addConstValue(resultCell);
					}
				} else {
					IMDataFlow resultCell
					= evaluate(node, bb, lhsElement, rhsCell);
					if (resultCell.isConstant() == false)
						return evaluate(node, bb, lhsCell, rhsCell);
					set = set.addConstValue(resultCell);
				}
			}
			return set;
		} else if (rhsCell.isSet()) {
			IMDataFlow set = new IMDataFlowSet(opts, node.getDatatype());
			for (IMDataFlow rhsElement : (IMDataFlowSet) rhsCell) {
				IMDataFlow resultCell = evaluate(node, bb, lhsCell, rhsElement);
				if (resultCell.isConstant() == false)
					return evaluate(node, bb, lhsCell, rhsCell);
				set = set.addConstValue(resultCell);
			}
			return set;
		}

		// If at least one of the two operand cells is an integer range, the
		// result is a range as well
		if (isSlotWithPathCell(lhsNode, bb) && isSlotWithPathCell(rhsNode, bb)
				&& (lhsCell.hasFixedRange() || rhsCell.hasFixedRange())) {
			Integer lhsMin = getMin(lhsCell);
			Integer lhsMax = getMax(lhsCell);
			Integer rhsMin = getMin(rhsCell);
			Integer rhsMax = getMax(rhsCell);
			if (lhsMin != null && lhsMax != null
					&& rhsMin != null && rhsMax != null) {
				Integer min
				= node.evaluateMin(lhsMin, lhsMax, rhsMin, rhsMax);
				Integer max
				= node.evaluateMax(lhsMin, lhsMax, rhsMin, rhsMax);
				if (min != null && max != null) {
					if (min != max)
						return new IMDataFlowIntRange(min, max, opts);
					IMMethod method = node.getMethod();
					int bcPos = node.getBCPosition();
					IMConstant value
					= method.createIMIConstant(min, bcPos);
					return IMDataFlow.createDataFlowObj(value, bb);
				}
			}
		}

		if (node instanceof IMConditionalBranch) {
			IMDataFlow resultCell = null;
			if (isNullReference(lhsCell))
				resultCell = evaluateNullCheck(node, rhsCell, bb);
			else if (isNullReference(rhsCell))
				resultCell = evaluateNullCheck(node, lhsCell, bb);
			if (resultCell != null)
				return resultCell;
		}

		return evaluate(node, bb, lhsCell, rhsCell);
	}

	private IMDataFlow evaluate(IMNode node, IMBasicBlock bb,
			IMDataFlow... argCells) throws CompileException {

		// If any of the arguments is not a constant, return
		IMConstant[] args = new IMConstant[argCells.length];
		for (int i = 0; i < argCells.length; ++i) {
			if (argCells[i].isConstant() == false)
				return IMDataFlow.createDataFlowObj(node, bb);
			args[i] = argCells[i].getValue();
		}

		// Evaluate the expression
		IMConstant result = node.evaluate(args);
		if (result == null)
			return IMDataFlow.createDataFlowObj(node, bb);
		return IMDataFlow.createDataFlowObj(result, bb);
	}

	private IMDataFlow putCell(IMDataFlow cell, IMNode node, IMBasicBlock bb)
	throws CompileException {

		// If the cell is uninitialized, initialize it
		if (cell == null)
			cell = IMDataFlow.createDataFlowObj(node, bb);

		// If the body of this method is ignored, nothing is for sure
		JoinPointChecker joinPoints = node.getOptions().getJoinPoints();
		IMMethod method = node.getMethod();
		if (joinPoints.ignoreMethodBody(method.getIMClass(), method) == true)
			cell = IMDataFlow.createInvalidDataFlowObj(node, bb);

		// Put the cell
		cells.put(node, cell);
		return cell;
	}

	private boolean isSlotWithPathCell(IMNode node, IMBasicBlock bb) {
		return (node.isVariable()
				&& pathInfo.getBBCell(node.getIMSlot(), bb) != null);
	}

	private static Integer getMin(IMDataFlow cell) throws CompileException {
		if (cell.hasFixedRange())
			return cell.minValue();
		else if (cell.isConstant())
			return cell.getValue().getIntValue();
		else
			return null;
	}

	private static Integer getMax(IMDataFlow cell) throws CompileException {
		if (cell.hasFixedRange())
			return cell.maxValue();
		else if (cell.isConstant())
			return cell.getValue().getIntValue();
		else
			return null;
	}

	private static boolean isNullReference(IMDataFlow cell)
	throws CompileException {
		return (cell instanceof IMDataFlowConstObj
				&& cell.getValue().isNullValue());
	}

	private static IMDataFlow evaluateNullCheck(IMNode node, IMDataFlow cell,
			IMBasicBlock bb) throws CompileException {
		if (cell.isValid() == false)
			return null;
		int result;
		if (node instanceof IMEQConditionalBranch
				|| node instanceof IMNullCheck)
			result = 0;
		else if (node instanceof IMNEConditionalBranch)
			result = 1;
		else
			return null;
		int bcPos = node.getBCPosition();
		IMConstant constRes = node.getMethod().createIMIConstant(result, bcPos);
		return IMDataFlow.createDataFlowObj(constRes, bb);
	}

	private static IMDataFlow createMessageCell(IMInvoke invoke,
			IMBasicBlock bb, IMDataFlow cell, String className,
			String classPrefix) throws CompileException {
		String name = className.replaceFirst(classPrefix, "");
		FakeStringCPEntry cpEntry = new FakeStringCPEntry(name);
		IMMethod   method    = invoke.getMethod();
		int        bcPos     = invoke.getBCPosition();
		IMConstant nameConst = method.createIMAConstant(cpEntry, bcPos);
		IMDataFlow nameCell  = IMDataFlow.createValidDataFlowObj(nameConst, bb);
		return new IMDataFlowMessage(cell, nameCell);
	}

}
