/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;

import keso.compiler.*;
import keso.compiler.imcode.*;

/**
 * Flow-lattice information store.
 * This is the central place where the computed flow-lattice information for all
 * domains is stored. It allows querying the info for a single domain, the
 * combined info for a set of domains and the combined info for the entire
 * program. Results are cached in order to keep the overhead for querying low.
 */
public final class FlowLatticeInfoStore {
	
	private static final int MAX_DOMAIN_COUNT = 8;
	private static final int MAX_COMBO_COUNT  = (1 << MAX_DOMAIN_COUNT) - 1;
	
	private static FlowLatticeInfoStore instance = null;
	
	private final Map<IMNode, IMBasicBlock> nodeBlocks;
	
	private final Map<IMDomain, FlowLatticeInfo> flowInfos
	= new HashMap<IMDomain, FlowLatticeInfo>();
	
	private FlowLatticeInfo combinedInfo = null;
	
	private final FlowLatticeInfo[] comboMap;
	
	/**
	 * Constructor.
	 * @param opts       Builder options.
	 * @param nodeBlocks Map that holds for each node the basic block in which
	 *                   it is located.
	 */
	private FlowLatticeInfoStore(BuilderOptions opts,
			Map<IMNode, IMBasicBlock> nodeBlocks) throws CompileException {
		this.nodeBlocks = nodeBlocks;
		int maxDomainID = 0;
		for (IMDomain domain : opts.getDomains()) {
			int id = domain.getDomainID();
			if (id > maxDomainID)
				maxDomainID = id;
		}
		if (maxDomainID < MAX_COMBO_COUNT)
			comboMap = new FlowLatticeInfo[MAX_COMBO_COUNT];
		else
			comboMap = null;
	}
	
	/**
	 * Returns the computed flow-lattice info for a single domain.
	 * @param domain The domain.
	 */
	public FlowLatticeInfo getForDomain(IMDomain domain) {
		return flowInfos.get(domain);
	}
	
	/**
	 * Returns the combined flow-lattice info for a set of domains.
	 * @param domains Set of domains.
	 */
	public FlowLatticeInfo getForDomains(Collection<IMDomain> domains)
			throws CompileException {
		
		if (comboMap == null || domains.size() == 0)
			return getForAllDomains();
		
		int index = 0;
		for (IMDomain domain : domains) {
			index |= 1 << domain.getDomainID();
		}
		FlowLatticeInfo info = comboMap[index];
		
		if (info == null) {
			info = new FlowLatticeInfo(nodeBlocks);
			for (IMDomain domain : domains) {
				info.merge(getForDomain(domain));
			}
			comboMap[index] = info;
		}
		return info;
	}
	
	/** Returns the combined flow-lattice info for all domains. */
	public FlowLatticeInfo getForAllDomains() throws CompileException {
		if (combinedInfo == null) {
			combinedInfo = new FlowLatticeInfo(nodeBlocks);
			for (FlowLatticeInfo flowInfo : flowInfos.values()) {
				combinedInfo.merge(flowInfo);
			}
		}
		return combinedInfo;
	}
	
	/**
	 * Replaces an instruction node with another node.
	 * @param oldNode Old instruction.
	 * @param newNode New instruction.
	 */
	public IMNode replaceNode(IMNode oldNode, IMNode newNode) {
		for (FlowLatticeInfo info : flowInfos.values()) {
			info.replaceNode(oldNode, newNode);
		}
		if (comboMap != null) {
			for (FlowLatticeInfo info : comboMap) {
				if (info != null)
					info.replaceNode(oldNode, newNode);
			}
		}
		if (combinedInfo != null)
			combinedInfo.replaceNode(oldNode, newNode);
		return newNode;
	}
	
	void putForDomain(IMDomain domain, FlowLatticeInfo info) {
		
		// Reset all combined flow-lattice info
		for (int i = 0; i < comboMap.length; ++i) {
			comboMap[i] = null;
		}
		combinedInfo = null;
		
		flowInfos.put(domain, info);
	}
	
	static FlowLatticeInfoStore getInstance(BuilderOptions opts,
			Map<IMNode, IMBasicBlock> nodeBlocks) throws CompileException {
		if (instance == null)
			instance = new FlowLatticeInfoStore(opts, nodeBlocks);
		return instance;
	}

}
