/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.util.*;

/**
 * Liveness-analysis pass.
 * This pass analyzes all basic blocks and creates liveness information for all
 * local variables. Each basic block is annotated with a LiveIn and a LiveOut
 * set. This information is used by several transformation and dump passes.
 * The liveness analysis is agnostic about the intermediate code representation
 * (SSA/non-SSA); it can handle both forms.
 */
public final class LivenessAnalysis extends MethodPass {

	@Override
	public String getName() {
		return "Liveness analysis";
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsAnalysis(true);
		info.addDependency(CallGraphAnalysis.class);
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		Multimap<IMBasicBlock, IMSlot> gen
		= new HashUniqueMultimap<IMBasicBlock, IMSlot>();
		Multimap<IMBasicBlock, IMSlot> kill
		= new HashUniqueMultimap<IMBasicBlock, IMSlot>();
		Multimap<IMBasicBlock, IMPhi> phiMap
		= new HashUniqueMultimap<IMBasicBlock, IMPhi>();

		// Get Gen and Kill sets for all basic blocks
		for (IMBasicBlock bb : method) {

			bb.clearLivenessInfo();

			for (IMNode root : bb) {

				for (IMNode node : IMCodeFlattener.getFlattenedCode(root)) {

					if (node instanceof IMPhi) {                       // Phi

						// Add the node to the set of Phi-functions of this
						// basic block and remove all its operands from the set
						// of non-Phi slots
						IMPhi phi = (IMPhi) node;
						phiMap.add(bb, phi);
						for (int i = 0; i < bb.getPred().length(); ++i) {
							IMSlot slot = phi.getSourceVar(i);
							if (slot == null)
								continue;
							bb.addPhiIn(slot);
							gen.remove(bb, slot);
						}

					} else if (node.usesSlot()) {                      // Use

						// Update Gen set
						IMSlot slot = node.getIMSlot();
						if (slot != null) {
							if (kill.get(bb).contains(slot) == false)
								gen.add(bb, slot);
						}

					} else if (node instanceof IMStoreLocalVariable) { // Def.

						// Update Kill set
						kill.add(bb, node.getIMSlot());
					}
				}
			}
		}

		boolean changed;

		// Fixed-point algorithm
		do {

			changed = false;

			// Analyze all basic blocks
			for (IMBasicBlock bb : method) {

				// Update LiveOut
				for (IMBasicBlock succ : bb.getSucc()) {
					for (IMSlot slot : succ.getLiveInSlots()) {
						bb.addLiveOut(slot);
					}
					IMBasicBlockList succPreds = succ.getPred();
					for (IMPhi phi : phiMap.get(succ)) {
						for (int i = 0; i < succPreds.length(); ++i) {
							if (succPreds.at(i) != bb)
								continue;
							IMSlot slot = phi.getSourceVar(i);
							if (slot != null)
								bb.addLiveOut(slot);
						}
					}
				}

				// Update LiveIn
				int prevLiveInCount = bb.getLiveInSlots().size();
				bb.addLiveIn(gen.get(bb));
				Collection<IMSlot> liveOut = bb.getLiveOutSlots();
				Collection<IMSlot> addIn   = new HashSet<IMSlot>(liveOut);
				addIn.removeAll(kill.get(bb));
				bb.addLiveIn(addIn);

				// See if anything changed
				if (bb.getLiveInSlots().size() != prevLiveInCount)
					changed = true;
			}

		} while (changed == true);

		return false;
	}

}
