/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import keso.compiler.*;
import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.imcode.*;

/** Reachability-analysis base class. */
public abstract class ReachabilityAnalysis extends DomainPass {

	protected IMDomain currDomain;

	@Override
	public final boolean init() {

		// Reset the reachability info
		IMDomain.getOverallReachabilityData().clear();
		ReachabilityData rData = IMDomain.getOverallReachabilityData();
		for (IMMethod method : rData.getMethods()) {
			for (IMBasicBlock bb : method) {
				bb.clearDomains();
			}
		}
		return true;
	}

	/**
	 * Runs the pass for a specific domain.
	 * @param domain The domain.
	 * @return <code>true</code> on success, <code>false</code> on error.
	 */
	@Override
	public boolean runOnDomain(IMDomain domain) throws CompileException {

		domain.getReachabilityData().clear();
		currDomain = domain;

		// Tasks: default constructor + entry point
		BuilderOptions opts = domain.getOptions();
		for (IMTask task : domain.getTasks()) {
			IMClass imClass = classStore.getClass(task.getMainClass());
			opts.vverbose("Task constructor:");
			addMethod(imClass.getMethod("<init>()V"));
			opts.vverbose("Task entry point:");
			addMethod(imClass.getMethod(task.getMainMethod()));
		}

		// ISRs: handler method
		JavaDomain domainDef = domain.getDef();
		for (ISRDefinition isr : domainDef.getISRs()) {
			IMClass imClass = classStore.getClass(isr.getHandlerClass());
			opts.vverbose("ISR:");
			addMethod(imClass.getMethod(isr.getHandlerMethod()));
		}

		// Exported services: all methods
		for (ServiceDefinition service : domainDef.getServices()) {
			IMClass imClass = service.getServiceClass(classStore);
			for (IMMethod method : imClass.getMethods()) {
				opts.vverbose("Service:");
				addMethod(method);
			}
		}

		// Operating system hooks: all hook methods
		SystemDefinition sysDef = opts.getSysDef();
		OSDefinition     os     = sysDef.getOSDef();
		if (os != null) {
			Object[] allHooks = os.getAllHooks();
			ComplexBoolAttribute[] hooks = (ComplexBoolAttribute[]) allHooks[0];
			for (ComplexBoolAttribute hook : hooks) {
				if (hook == null || hook.setting == false
						|| hook.getAttribute("HookMethod") == null)
					continue;
				String hc = hook.getAttribute("HookClass").valueString();
				String hm = hook.getAttribute("HookMethod").valueString();
				addMethod(classStore.getClass(hc).getMethod(hm));
			}
		}

		// Stack checks: all handler methods
		ComplexAttribute<?> stack = sysDef.getStack();
		if (stack != null) {
			String sc = stack.getAttribute("StackChecks").valueString();
			if (sc != null && sc.equalsIgnoreCase("true")) {
				String hc = stack.getAttribute("HandlerClass").valueString();
				String hm = stack.getAttribute("HandlerMethod").valueString();
				addMethod(classStore.getClass(hc).getMethod(hm));
			}
		}

		// Weavelets: all explicitly requested methods
		for (IMMethod method : opts.getJoinPoints().keepMethods()) {
			addMethod(method);
		}
		return false;
	}

	/**
	 * Starts the analysis at a given entry point (i. e., a method).
	 * @param method Entry point.
	 * @throws CompileException
	 */
	protected final void addMethod(IMMethod method) throws CompileException {

		// Skip the method if it has already been marked as reachable
		ReachabilityData currData = currDomain.getReachabilityData();
		if (currData.getMethods().contains(method))
			return;

		method.getOptions().vverbose("Method " + method);

		// Skip the method if it is a portal to another domain
		IMClass imClass   = method.getIMClass();
		IMClass serviceIf = method.getIMClass("keso/core/Service");
		if (serviceIf != null && imClass.implementsInterface(serviceIf)
				&& hasService(currDomain, imClass) == false)
			return;

		// Add the method
		currData.addMethod(method);
		IMDomain.getOverallReachabilityData().addMethod(method);
		method.addDomain(currDomain);

		// Add the class owning the method
		addClass(imClass);

		// Analyze the method
		analyzeMethod(method);
	}

	protected final void addClass(IMClass imClass) throws CompileException {

		// Skip the class if it has already been marked as reachable
		ReachabilityData currData = currDomain.getReachabilityData();
		if (currData.getClasses().contains(imClass))
			return;

		// Add the class
		currData.addClass(imClass);
		IMDomain.getOverallReachabilityData().addClass(imClass);

		// Add the class's static initializer (if any)
		IMMethod clinit = imClass.getMethod("<clinit>()V");
		if (clinit != null)
			addMethod(clinit);
	}

	protected final void addStaticField(IMField field, IMMethod user) {
		BuilderOptions opts = user.getOptions();
		opts.vverbose("Static field " + field + " used by " + user);
		currDomain.getReachabilityData().addStaticField(field, user);
		IMDomain.getOverallReachabilityData().addStaticField(field, user);
	}

	private static boolean hasService(IMDomain domain, IMClass imClass) {
		for (ServiceDefinition service : domain.getDef().getServices()) {
			if (imClass.getClassName().equals(service.getServiceClassName()))
				return true;
		}
		return false;
	}

	protected abstract void analyzeMethod(IMMethod method)
			throws CompileException;

}
