/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import java.util.*;

import keso.compiler.imcode.*;
import keso.util.*;

/** Reachability-analysis data class. */
public final class ReachabilityData {

	private final Set<IMClass>  classes = new HashSet<IMClass>();
	private final Set<IMMethod> methods = new HashSet<IMMethod>();
	private final Multimap<IMField, IMMethod> staticFieldUsers
	= new HashUniqueMultimap<IMField, IMMethod>();
	private final DirectedGraph<IMClass> clinitGraph
	= new DirectedGraph<IMClass>();

	/**
	 * Returns the set of reachable classes.
	 * @return Class set.
	 */
	public Collection<IMClass> getClasses() {
		return Collections.unmodifiableCollection(classes);
	}

	/**
	 * Returns the set of reachable methods.
	 * @return Method set.
	 */
	public Collection<IMMethod> getMethods() {
		return Collections.unmodifiableCollection(methods);
	}

	/**
	 * Returns the set of reachable static fields.
	 * @return Static field set.
	 */
	public Collection<IMField> getStaticFields() {
		return Collections.unmodifiableCollection(staticFieldUsers.keySet());
	}

	/**
	 * Returns the set of methods using a certain static field.
	 * @param field Static field.
	 * @return User set.
	 */
	public Collection<IMMethod> getStaticFieldUsers(IMField field) {
		return staticFieldUsers.get(field);
	}

	/**
	 * Returns the class-initialization dependency graph.
	 * @return The graph.
	 */
	public DirectedGraph<IMClass> getClinitGraph() {
		return clinitGraph;
	}

	/**
	 * Adds a class to the set of reachable classes.
	 * @param imClass Class to be added.
	 */
	void addClass(IMClass imClass) {
		classes.add(imClass);
	}

	/**
	 * Adds a method to the set of reachable methods.
	 * @param method Method to be added.
	 */
	void addMethod(IMMethod method) {
		methods.add(method);
	}

	/**
	 * Adds a static field to the set of reachable static fields.
	 * @param field Field to be added.
	 * @param user  Method using the static field.
	 */
	void addStaticField(IMField field, IMMethod user) {
		staticFieldUsers.add(field, user);
	}

	/**
	 * Adds a class-initialization dependency.
	 * @param before Class that has to be initialized before the second one.
	 * @param after  Class that has to be initialized after the first one.
	 */
	void addClinitDependency(IMClass before, IMClass after) {
		if (clinitGraph.isEdge(before, after) == false)
			clinitGraph.addEdge(before, after);
	}

	/** Clears the reachability data. */
	void clear() {
		classes.clear();
		methods.clear();
		staticFieldUsers.clear();
		clearClinitGraph();
	}

	/**
	 * Clears the class-initialization dependency graph.
	 */
	void clearClinitGraph() {
		clinitGraph.clear();
	}

}
