/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.analysis;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.transform.*;

/** Simple (control-flow-agnostic) reachability analysis pass. */
public final class SimpleReachabilityAnalysis extends ReachabilityAnalysis {

	@Override
	public String getName() {
		return "Simple reachability analysis";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.hasOption("simple_reachability_analysis");
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {

		info.setIsAnalysis(true);
		info.addDependency(CallGraphAnalysis.class);

		// This is not really a dependency, but as constant folding invalidates
		// the reachability results, it only makes sense to execute this pass
		// after constant folding.
		info.addDependency(ConstantFolding.class);
	}

	@Override
	protected void analyzeMethod(final IMMethod method)
	throws CompileException {

		for (IMBasicBlock bb : method) {

			bb.addDomain(currDomain);

			// Analyze field accesses
			bb.visitNodes(new IMBasicVisitor() {

				@Override
				public void visit(IMNode node) throws CompileException {

					// Find all field accesses
					IMField field;
					boolean isStatic = false;
					if (node instanceof IMGetField) {
						field = ((IMGetField) node).getField();
					} else if (node instanceof IMGetStatic) {
						field = ((IMGetStatic) node).getField();
						isStatic = true;
					} else if (node instanceof IMPutField) {
						field = ((IMPutField) node).getField();
					} else if (node instanceof IMPutStatic) {
						field = ((IMPutStatic) node).getField();
						isStatic = true;
					} else {
						return;
					}

					// Add the field (if it is static) and the class it belongs to
					if (isStatic)
						addStaticField(field, method);
					addClass(field.getType());
				}
			});
		}

		// Add all called methods
		for (IMMethod callee : method.getCalledMethods()) {
			method.getOptions().vverbose("Called from " + method + ':');
			addMethod(callee);
		}
	}

}
