/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMInvoke;
import keso.compiler.imcode.IMMethod;
import keso.compiler.imcode.IMNode;
import keso.compiler.imcode.IMSlot;
import keso.util.DotWriterException;

/**
 * A actual reference in the connection graph, i.e. arguments and return values, which are object references
 */
public class CGActualReferenceNode extends CGReferenceNode {
	/**
	 * Map mapping slots to their corresponding actual reference nodes
	 */
	private static final Map<IMSlot, CGActualReferenceNode>		slotToActualReferenceMap		= new HashMap<IMSlot, CGActualReferenceNode>();

	/**
	 * Map mapping methods to their corresponding return value actual reference nodes
	 */
	private static final Map<IMMethod, CGActualReferenceNode>	methodToReturnMap				= new HashMap<IMMethod, CGActualReferenceNode>();

	/**
	 * Map mapping method invocation sites to their corresponding caller side return value actual reference nodes
	 */
	private static Map<IMInvoke, CGActualReferenceNode>			invocationToReturnMap			= new HashMap<IMInvoke, CGActualReferenceNode>();

	/**
	 * Map mapping method invocation sites to their corresponding caller side actual reference nodes pointing to
	 * parameters
	 */
	private static Map<IMInvoke, CGActualReferenceNode[]>		invocationToActualReferenceMap	= new HashMap<IMInvoke, CGActualReferenceNode[]>();

	/**
	 * The slot that contains this parameter
	 */
	private final IMSlot										slot;

	/**
	 * The invocation site this node was created for
	 */
	private final IMInvoke										invocation;

	/**
	 * A number {@code >= 0} numbering the parameter or -1 for the return node, if {@code invocation != null}. If
	 * {@code invocation} is {@code null}, this field is defined to be -1 for the node representing the return value and
	 * {@code != -1} otherwise.
	 */
	private final int											paramNumber;

	/**
	 * Creates a {@code CGActualReferenceNode} from an {@code IMSlot}. This is done for all formal parameters for a
	 * method in the analysis.
	 *
	 * @param slot
	 *            the slot that contains the parameter inside the method
	 * @param paramNumber
	 *            the running number of the parameter that causes the creation of this node
	 */
	private CGActualReferenceNode(final IMSlot slot) {
		this.slot = slot;
		this.invocation = null;
		this.paramNumber = Integer.MIN_VALUE;
		updateEscapeState(EscapeState.Method);
	}

	/**
	 * Creates a {@code CGActualReferenceNode} from an {@code IMMethod}. This is done for return values of a method,
	 * where the method does not have an {@code IMSlot} assigned for its return value, as is the case with abstract
	 * methods and interface methods.
	 *
	 * @param method
	 *            the method that contains the return value, for which to create an actual reference node
	 */
	private CGActualReferenceNode(final IMMethod method) {
		this.slot = null;
		this.invocation = null;
		this.paramNumber = -1;
		updateEscapeState(EscapeState.Method);
	}

	/**
	 * Creates a caller-side {@code CGActualReferenceNode} from an {@code IMInvoke}. This is done at each call site of a
	 * method to create an anchor point for interprocedural analysis.
	 *
	 * @param invocation
	 *            the invocation of a method for which to create an actual reference node
	 * @param paramNumber
	 *            the number of the parameter being processed. Pass -1 for the return value, 0 for the this reference
	 *            and numbers starting at 1 for parameters.
	 */
	private CGActualReferenceNode(final IMInvoke invocation, final int paramNumber) {
		this.slot = null;
		this.invocation = invocation;
		this.paramNumber = paramNumber;
		// the default escape state for these methods is local!
	}

	/**
	 * Singleton factory returning the {@code CGActualReferenceNode} for a formal parameter {@code IMSlot}.
	 *
	 * @param argument
	 *            the slot that contains the parameter inside the method
	 * @param paramNumber
	 *            the running number of the parameter to create this node for
	 * @return a actual (phantom) reference node corresponding to the formal parameter of a method
	 */
	public static CGActualReferenceNode getActualReferenceNode(final IMSlot slot) {
		CGActualReferenceNode actualReference = slotToActualReferenceMap.get(slot);

		if (actualReference == null) {
			slotToActualReferenceMap.put(slot, actualReference = new CGActualReferenceNode(slot));
			referenceNodeList.add(actualReference);
		}

		return actualReference;
	}

	/**
	 * Singleton factory returning the {@code CGActualReferenceNode} for a return value of a method {@code method} on
	 * the callee side.
	 *
	 * @param method
	 *            the method of this return value
	 * @return the actual (phantom) reference node corresponding to the return value of the given method
	 */
	public static CGActualReferenceNode getReturnReferenceNode(final IMMethod method) {
		// do not try to use the slot of the return value here, as it might be the same of a parameter, causing cycles
		// in the connection graph
		CGActualReferenceNode actualReference = methodToReturnMap.get(method);

		if (actualReference == null) {
			methodToReturnMap.put(method, actualReference = new CGActualReferenceNode(method));
			referenceNodeList.add(actualReference);
		}

		return actualReference;
	}

	/**
	 * Helper method creating and returning actual reference nodes pointing to the real parameters on the caller side
	 *
	 * @param invocation
	 *            the invocation site to create nodes for
	 * @param nodeToCGNodeMap
	 *            map mapping {@code IMNode}s to their corresponding {@code CGNode}s in the connection graph
	 * @param callerActualsAndGlobals
	 *            the list of caller-side actual reference nodes and global reference nodes used in the currently
	 *            analyzed method; this list is used to register and newly created reference nodes.
	 * @return array of actual reference nodes representing the parameters and return value of the given invocation on
	 *         the caller-side
	 * @throws CompileException
	 */
	public static CGActualReferenceNode[] getArgumentActualReferenceNodes(final IMInvoke invocation,
			final Map<IMNode, CGNode> nodeToCGNodeMap, final List<CGReferenceNode> callerActualsAndGlobals)
			throws CompileException {
		// treat the 'this' parameter no different from other arguments
		CGActualReferenceNode[] actualArguments = invocationToActualReferenceMap.get(invocation);
		IMNode[] arguments = invocation.getArguments();

		if (actualArguments == null) {
			int arrayLength = arguments.length + 1;

			if (invocation.getObject() != null)
				arrayLength++;

			invocationToActualReferenceMap.put(invocation, actualArguments = new CGActualReferenceNode[arrayLength]);

			// IMInvoke nodes do not have slots assigned anyway, so there's no point in trying to use them as keys in a
			// hash map

			actualArguments[0] = invocationToReturnMap.get(invocation);
			if (actualArguments[0] == null && invocation.getCallee().isReference()) {
				invocationToReturnMap.put(invocation, actualArguments[0] = new CGActualReferenceNode(invocation, -1));
				callerActualsAndGlobals.add(actualArguments[0]);
			}

			int idx = 1;

			if (invocation.getObject() != null) {
				if (actualArguments[idx] == null) {
					actualArguments[idx] = new CGActualReferenceNode(invocation, idx - 1);
					callerActualsAndGlobals.add(actualArguments[idx]);
				}
				nodeToCGNodeMap.get(invocation.getObject()).attach(actualArguments[idx]);

				idx++;
			}

			for (int i = 0; i < arguments.length; ++i) {
				if (arguments[i].isReference()) {
					if (actualArguments[idx] == null) {
						actualArguments[idx] = new CGActualReferenceNode(invocation, idx - 1);
						callerActualsAndGlobals.add(actualArguments[idx]);
					}
					nodeToCGNodeMap.get(arguments[i]).attach(actualArguments[idx]);
				}

				idx++;
			}
		}

		return actualArguments;
	}

	/**
	 * The slot that contains this argument at run-time
	 *
	 * @return the {@code IMSlot} that was used to create this actual reference nodeF
	 */
	public IMSlot getSlot() {
		return slot;
	}

	/**
	 * The invocation site used to create this node or {@code null} if this is not a caller-side node.
	 *
	 * @return the {@code IMInvoke} used to create this node or {@code null} if this is not a caller-side node.
	 */
	public IMInvoke getInvocation() {
		return invocation;
	}

	/**
	 * Getter for the running parameter number
	 *
	 * @return the parameter number, see the field for possible values
	 */
	public int getParamNumber() {
		return paramNumber;
	}

	@Override
	public void pathCompression() {
		// handle pathCompression() on actual reference nodes different, because we never want to mark these nodes as
		// stale which would prevent them from showing up in the graph dump and
		super.pathCompression();
		setStale(false);
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}

	@Override
	public String toString() {
		String label;

		if (invocation != null) {
			// yes, it's non-trivial to get just the name of the called method from an invocation...
			try {
				label = invocation.getCandidates().iterator().next().getMethodName();
			} catch (CompileException e) {
				label = "exception";
			}

			if (paramNumber == -1) {
				label += "(ret)";
			} else {
				label += "(param" + paramNumber + ")";
			}
		} else {
			if (paramNumber == -1) {
				label = "ret";
			} else {
				label = slot.toString();
			}
		}

		return label;
	}
}
