/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.HashMap;
import java.util.Map;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMConstant;
import keso.compiler.imcode.IMMethod;
import keso.util.DotWriterException;

/**
 * constant object node in the connection graph
 */
public class CGConstantObjectNode extends CGObjectNode {
	/**
	 * Map mapping constants to their corresponding object nodes in the connection graph
	 */
	private static Map<IMConstant, CGConstantObjectNode>	constantToObjectMap	= new HashMap<IMConstant, CGConstantObjectNode>();

	/**
	 * Holds a reference to the constant this object describes
	 */
	private final IMConstant								constant;

	/**
	 * Return the object node associated with a given constant. Creates a new {@code CGConstantObjectNode} if none
	 * exists yet.
	 *
	 * @param n
	 *            the constant to return the object node for
	 * @return the object node in the connection graph corresponding to the constant
	 */
	public static CGConstantObjectNode getObjNodeForConstant(final IMConstant constant) {
		CGConstantObjectNode objectNode = constantToObjectMap.get(constant);

		// if there's no object node created for this constant yet, create one
		if (objectNode == null) {
			constantToObjectMap.put(constant, objectNode = new CGConstantObjectNode(constant));
		}

		return objectNode;
	}

	/**
	 * Construct a new constant object node in the connection graph; in order to be able to distinguish between
	 * constants and null, store the constant that was used to create this constant object node
	 *
	 * @param constant
	 *            {@code IMConstant} object, which describes the intermediate code instruction creating this constant
	 */
	protected CGConstantObjectNode(final IMConstant constant) {
		super(null);
		this.constant = constant;
	}

	/**
	 * Returns the constant intermediate code node used to create this constant object node.
	 *
	 * @return constant
	 */
	public IMConstant getConstant() {
		return constant;
	}

	@Override
	public boolean attach(final CGReferenceNode reference) {
		if (constant.isNullValue()) {
			return false;
		}

		// we're not pointing to null values, since that's of no interest to us
		return super.attach(reference);
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}

	@Override
	public CGConstantObjectNode getCloneInMethod(CGReferenceNode reference, IMMethod method) {
		// NOTE: This will return the same object node in all connection graphs!
		CGConstantObjectNode object = getObjNodeForConstant(constant);
		object.attach(reference);
		return object;
	}
}
