/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.io.FileNotFoundException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMAllocation;
import keso.compiler.imcode.IMConstant;
import keso.compiler.imcode.IMMethod;
import keso.util.DotWriter;
import keso.util.DotWriterException;

/**
 * Dump class writing a connection graph to a graphviz dot file. This internally uses the DotWriter class from
 * keso.util.
 */
public class CGDotWriter {
	private final DotWriter<CGNode>	writer;
	private final Set<CGNode>		visited;

	public CGDotWriter(final String fileName, final String name) throws FileNotFoundException, DotWriterException {
		this.writer = new DotWriter<CGNode>(fileName);
		this.visited = new HashSet<CGNode>();

		this.writer.initGraph(name);
	}

	private void fieldReference(CGObjectNode from, CGFieldReferenceNode frn) throws DotWriterException,
			CompileException {
		frn.accept(this);
		String edgeLabel = "F";
		if (frn.getField() == null) {
			edgeLabel += "[]";
		}
		writer.addEdge(from, frn, "label=\"" + edgeLabel + "\"");
	}

	private void reference(CGReferenceNode n) throws DotWriterException, CompileException {
		for (CGObjectNode on : n.getPointees()) {
			on.accept(this);
			writer.addEdge(n, on, "label=P");
		}

		for (CGReferenceNode def : n.getDeferredEdges()) {
			def.accept(this);
			writer.addEdge(n, def, "label=D, style=dashed");
		}

		for (CGReferenceNode rdef : n.reverseDeferredEdges) {
			rdef.accept(this);
		}
	}

	private String escapeState(final CGNode node) {
		return escapeState(node, "");
	}

	private String escapeState(final CGNode node, final String style) {
		switch (node.getEscapeState()) {
			case Local:
				return "fillcolor=\"/rdgy3/2\", style=\"filled, " + style + "\"";
			case Method:
				return "fillcolor=\"/rdgy3/3\", style=\"filled, " + style + "\"";
			default:
				return "fillcolor=\"/rdgy3/1\", style=\"filled, " + style + "\"";
		}
	}

	public void visit(final CGObjectNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			IMAllocation a = n.getAllocation();
			String label;

			if (a != null) {
				String name = a.dumpOperands();
				int idx = name.lastIndexOf('/');
				if (idx != -1)
					label = name.substring(idx + 1);
				else
					label = name;
			} else {
				label = "<noalloc>";
			}

			writer.addNode(n, label + "\\n" + n.hashCode(), "shape=rectangle, color=\"/ylgn5/5\", " + escapeState(n));

			for (CGFieldReferenceNode frn : n.getFields()) {
				fieldReference(n, frn);
			}
		}
	}

	public void visit(final CGConstantObjectNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);
			IMConstant c = n.getConstant();

			// do not print null values
			if (c.isNullValue())
				return;

			String label = c.toReadableString().replaceAll("[^\\w ]", "");
			writer.addNode(n, label + "\\n" + n.hashCode(), "shape=rectangle, color=\"/ylgn5/5\", " + escapeState(n));

			for (CGFieldReferenceNode frn : n.getFields()) {
				fieldReference(n, frn);
			}
		}
	}

	public void visit(final CGObjectPhantomNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			writer.addNode(n, "<phantom>" + "\\n" + n.hashCode(), "shape=rectangle, color=\"/ylgn5/5\", " + escapeState(n, "dashed"));

			for (CGFieldReferenceNode frn : n.getFields()) {
				fieldReference(n, frn);
			}
		}
	}

	public void visit(final CGLocalReferenceNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			String label;

			if (n.getNode().getIMSlot() != null) {
				label = n.getNode().getIMSlot().toString();
			} else {
				label = "tmp_" + n.getNode().getClass().getSimpleName();
			}

			writer.addNode(n, label + "\\n" + n.hashCode(), "color=\"/spectral4/4\", " + escapeState(n));

			reference(n);
		}
	}

	public void visit(final CGFieldReferenceNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			String label;

			if (n.getField() != null) {
				label = n.getField().getFieldName();
			} else {
				label = "[?]";
			}

			writer.addNode(n, label + "\\n" + n.hashCode(), "color=\"/spectral4/1\", " + escapeState(n));

			reference(n);
		}
	}

	public void visit(final CGGlobalReferenceNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			String label = n.getName();

			writer.addNode(n, label + "\\n" + n.hashCode(), "color=\"/spectral4/1\", " + escapeState(n, "dashed"));

			reference(n);
		}
	}

	public void visit(final CGActualReferenceNode n) throws DotWriterException, CompileException {
		if (!n.isStale() && !visited.contains(n)) {
			visited.add(n);

			String label;
			String style;

			if (n.getInvocation() != null) {
				style = "dotted";

				// yes, it's non-trivial to get just the name of the called method from an invocation...
				label = "unknown";
				Iterator<IMMethod> candidateIterator = n.getInvocation().getCandidates().iterator();
				if (candidateIterator.hasNext()) {
					label = candidateIterator.next().getMethodName();
				}

				if (n.getParamNumber() == -1) {
					label += "(ret)";
				} else {
					label += "(param" + n.getParamNumber() + ")";
				}
			} else {
				style = "dashed";

				if (n.getParamNumber() == -1) {
					label = "ret";
				} else {
					label = n.getSlot().toString();
				}
			}

			writer.addNode(n, label + "\\n" + n.hashCode(), "color=\"/spectral4/4\", " + escapeState(n, style));

			reference(n);
		}
	}

	@Override
	public void finalize() throws DotWriterException {
		writer.finalizeGraph();
	}
}
