/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import keso.classfile.constantpool.ClassCPEntry;
import keso.compiler.CompileException;
import keso.compiler.imcode.IMConstant;
import keso.compiler.imcode.IMField;
import keso.util.DotWriterException;

/**
 * A global reference in the connection graph, i.e. static fields that are object references
 */
public class CGGlobalReferenceNode extends CGReferenceNode {
	/**
	 * Multimap mapping class table entry and field description to their corresponding global reference nodes
	 */
	private static Map<ClassCPEntry, Map<IMField, CGGlobalReferenceNode>>	classAndFieldToGlobalReferenceMap	= new HashMap<ClassCPEntry, Map<IMField, CGGlobalReferenceNode>>();

	/**
	 * The class containing the static field this global reference node object represents
	 */
	private final ClassCPEntry												classEntry;

	/**
	 * the field in the class this global reference node object represents
	 */
	private final IMField													field;

	/**
	 * The constant reference we created this global reference node for, or null if none.
	 */
	private final IMConstant												constant;

	/**
	 * Singleton factory returning global reference nodes for static fields.
	 *
	 * @param classCPEntry
	 *            the class containing the field
	 * @param field
	 *            the field
	 * @param callerActualsAndGlobals
	 *            the list of caller-side actual reference nodes and global reference nodes used in the currently
	 *            analyzed method; this list is used to register and newly created reference nodes.
	 * @return a {@code CGGlobalReferenceNode} representing the static field in the connection graph
	 */
	public static CGGlobalReferenceNode getReferenceNode(final ClassCPEntry classCPEntry, final IMField field,
			final List<CGReferenceNode> callerActualsAndGlobals) {
		CGGlobalReferenceNode globalReference;
		Map<IMField, CGGlobalReferenceNode> fieldToGlobalReferenceMap = classAndFieldToGlobalReferenceMap
				.get(classCPEntry);

		if (fieldToGlobalReferenceMap == null) {
			classAndFieldToGlobalReferenceMap.put(classCPEntry,
					fieldToGlobalReferenceMap = new HashMap<IMField, CGGlobalReferenceNode>());
			fieldToGlobalReferenceMap.put(field, globalReference = new CGGlobalReferenceNode(classCPEntry, field));
		} else {
			globalReference = fieldToGlobalReferenceMap.get(field);

			if (globalReference == null) {
				fieldToGlobalReferenceMap.put(field, globalReference = new CGGlobalReferenceNode(classCPEntry, field));
			}
		}

		callerActualsAndGlobals.add(globalReference);

		// do this every time to make sure it's being added to the current method's global escape list, ensuring it's
		// being processed in the reachability analysis and dumped in the DotWriter.
		globalReference.updateEscapeState(EscapeState.Global);

		return globalReference;
	}

	/**
	 * Singleton factory returning global reference nodes for constant references
	 *
	 * @param constant
	 *            the constant reference to create a connection graph representation for
	 * @param callerActualsAndGlobals
	 *            the list of caller-side actual reference nodes and global reference nodes used in the currently
	 *            analyzed method; this list is used to register and newly created reference nodes.
	 * @return a {@code CGGlobalReferenceNode} representing the constant reference in the connection graph
	 */
	public static CGGlobalReferenceNode getGlobalPhantomReferenceNode(final IMConstant constant,
			final List<CGReferenceNode> callerActualsAndGlobals) {
		CGGlobalReferenceNode globalReference = new CGGlobalReferenceNode(constant);
		callerActualsAndGlobals.add(globalReference);
		return globalReference;
	}

	/**
	 * Creates a new global reference node in the connection graph. Global reference nodes describe static class
	 * members, which are the only global variables in Java.
	 *
	 * @param classEntry
	 *            The class table entry describing the class
	 * @param field
	 *            the field reference object describing the field to access
	 */
	private CGGlobalReferenceNode(final ClassCPEntry classEntry, final IMField field) {
		this.classEntry = classEntry;
		this.field = field;
		this.constant = null;
	}

	/**
	 * Creates a new global reference node in the connection graph. Global reference nodes describe static references;
	 * in this case this is used for the {@code IMAMTConstant} classes in KESO.
	 *
	 * @param constant
	 *            the {@code IMConstant} class describing the global reference we're creating this global reference node
	 *            for
	 */
	private CGGlobalReferenceNode(final IMConstant constant) {
		this.classEntry = null;
		this.field = null;
		this.constant = constant;
	}

	/**
	 * Returns this field's name
	 *
	 * @return the class and field name of this global reference
	 */
	public String getName() {
		if (constant != null)
			return constant.toReadableString();
		return classEntry.getClassName() + "." + field.getFieldName();
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}
}
