/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.HashMap;
import java.util.Map;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMNode;
import keso.compiler.imcode.IMSlot;
import keso.util.DotWriterException;

/**
 * A local reference in the connection graph, i.e. those local variables, that are object references
 */
public class CGLocalReferenceNode extends CGReferenceNode {
	/**
	 * Map mapping reference variable slots to their corresponding local reference nodes in the connection graph
	 */
	private static Map<IMSlot, CGLocalReferenceNode>	slotToLocalReferenceMap	= new HashMap<IMSlot, CGLocalReferenceNode>();

	/**
	 * Map mapping nodes using reference variables (usually in the way of returning them) to the corresponding local
	 * reference nodes in the connection graph
	 */
	private static Map<IMNode, CGLocalReferenceNode>	nodeToLocalReferenceMap	= new HashMap<IMNode, CGLocalReferenceNode>();

	/**
	 * Holds a reference to the node that first used this local reference
	 */
	private IMNode										node;

	/**
	 * Returns a local reference node in the connection graph for a given node, and creates it if none exists for this
	 * node yet. This is required for a number of operations returning a pointer, but not writing it to an
	 * {@code IMSlot} directly. If the given node has an {@code IMSlot} assigned, the slot will be used to 1-limit the
	 * creation of reference nodes and the the node returned will correspond to this {@code IMSlot}.
	 */
	public static CGLocalReferenceNode getLocalReferenceNode(final IMNode node) {
		CGLocalReferenceNode localReferenceNode;

		if (node.getIMSlot() == null) {
			localReferenceNode = nodeToLocalReferenceMap.get(node);

			// if there's no local reference node in the connection graph for this node yet, create one
			if (localReferenceNode == null) {
				nodeToLocalReferenceMap.put(node, localReferenceNode = new CGLocalReferenceNode(node));
				referenceNodeList.add(localReferenceNode);
			}
		} else {
			localReferenceNode = slotToLocalReferenceMap.get(node.getIMSlot());

			// if there's no local reference node in the connection graph for this slot yet, create one
			if (localReferenceNode == null) {
				slotToLocalReferenceMap.put(node.getIMSlot(), localReferenceNode = new CGLocalReferenceNode(node));
				referenceNodeList.add(localReferenceNode);
			}
		}

		return localReferenceNode;
	}

	/**
	 * Return the {@code CGLocalReferenceNode} corresponding to a given {@code IMSlot}. This does _not_ create the node,
	 * if none exists. Do not use this method in {@code ExtendedEscapeAnalysis}, it is intended for the stack allocation
	 * transformation only to determine which allocations are still alive.
	 *
	 * @param slot
	 *            the slot to get the local reference node for
	 * @return the local reference node in the connection graph corresponding to the given slot
	 */
	public static CGLocalReferenceNode getLocalReferenceNodeForSlot(final IMSlot slot) {
		return slotToLocalReferenceMap.get(slot);
	}

	/**
	 * Creates a new local reference node in the connection graph
	 *
	 * @param node
	 *            the node that first uses the reference
	 */
	private CGLocalReferenceNode(final IMNode node) {
		this.node = node;
	}

	/**
	 * Returns the slot that contains this local reference
	 *
	 * @return slot containing the reference
	 */
	public IMNode getNode() {
		return node;
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}
}
