/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import keso.compiler.CompileException;
import keso.util.DotWriterException;

/**
 * Abstract superclass for all nodes in the connection graph
 */
abstract public class CGNode {
	/**
	 * Set serving as starting point for reachability analysis on method escaping level after intraprocedural analysis.
	 * This is updated by {@code CGNode.setEscapeState()}.
	 */
	private static Set<CGNode>	methodEscapeList	= new HashSet<CGNode>();

	/**
	 * Accessor method for the list of method escaping nodes
	 *
	 * @return an unmodifiable set of method escaping nodes
	 */
	public static Set<CGNode> getMethodEscapeList() {
		return Collections.unmodifiableSet(methodEscapeList);
	}

	/**
	 * Set serving as starting point for reachability analysis on globally escaping level after intraprocedural
	 * analysis. This is updated by {@code CGNode.setEscapeState()}.
	 */
	private static Set<CGNode>	globalEscapeList	= new HashSet<CGNode>();

	/**
	 * Accessor method for list list of globally escaping objects
	 *
	 * @return an unmodifiable set of globally escaping objects
	 */
	public static Set<CGNode> getGlobalEscapeList() {
		return Collections.unmodifiableSet(globalEscapeList);
	}

	/**
	 * Clears the list of escaping nodes; call this for each new method
	 */
	public static void clearEscapeLists() {
		methodEscapeList.clear();
		globalEscapeList.clear();
	}

	/**
	 * Property signaling whether this node is no longer being referenced in the graph and can be omitted when dumping
	 * the graph in order to simplify
	 */
	private boolean	stale	= false;

	/**
	 * Returns true, if this node can be omitted from dumping, because it is no longer referenced by any other CG node
	 *
	 * @return boolean indicating whether this node is superfluous
	 */
	public boolean isStale() {
		return stale;
	}

	/**
	 * Mark this node as stale or not stale causing it to be ignored when dumping the graph.
	 *
	 * @param stale
	 *            the new stale state, {@code true} to hide the node, {@code false} to show it
	 */
	public void setStale(boolean stale) {
		this.stale = stale;
	}

	/**
	 * Escape state of the object; This value contains the information we try to calculate. A value of
	 * {@code EscapeState.Local} denotes a non-escaping (i.e. method local) object, a value of
	 * {@code EscapeState.Method} signifies the object escapes a method, but does not globally escape (i.e. into a
	 * different thread) and a value of {@code EscapeState.Global} means this object is accessible from a global context
	 * and should not be optimized in any way.
	 */
	private EscapeState	escapeState	= EscapeState.Local;

	/**
	 * @return the escape state associated with this object
	 */
	public EscapeState getEscapeState() {
		return escapeState;
	}

	/**
	 * Update the escape state of this object with a new value. This is being done with respect to the old escape state,
	 * i.e. you cannot lower the escape state using this function.
	 *
	 * @param escapeState
	 *            the new value for the escape state you are trying to set
	 */
	public void updateEscapeState(final EscapeState escapeState) {
		boolean change = false;
		switch (escapeState) {
			case Global:
				if (this.escapeState == EscapeState.Method)
					methodEscapeList.remove(this);

				this.escapeState = escapeState;
				change = globalEscapeList.add(this);

				break;
			case Method:
				// only increase escape state, never decrease
				if (this.escapeState == EscapeState.Global)
					break;

				this.escapeState = escapeState;
				change = methodEscapeList.add(this);

				break;
			case Local:
				// do nothing
		}

		if (change) {
			// and recursively update dependent nodes with the new escape state
			for (CGNode childNode : getDescendants()) {
				childNode.updateEscapeState(escapeState);
			}
		}
	}

	/**
	 * Sets the escape state to the specified value, without respect to the old escape state. This method can be used to
	 * lower the escape state of a node. In most cases, you should not use this method, but {@code updateEscapeState()}.
	 * <strong>Only use this method if you know what you are doing!</strong>
	 *
	 * @param escapeState
	 *            the new escape state
	 */
	public void setEscapeState(final EscapeState escapeState) {
		this.escapeState = escapeState;
	}

	/**
	 * Attach this node to the given {@code CGReferenceNode}, adding a points-to edge if {@code this} is a
	 * {@code CGObjectNode} or a deferred edge if {@code this} is a {@code CGReferenceNode}.
	 *
	 * @param reference
	 *            the reference node to attach to {@code this}
	 * @return true, if the operation added a new edge
	 * @throws CompileException
	 */
	public abstract boolean attach(final CGReferenceNode reference) throws CompileException;

	/**
	 * Attach this node to the given {@code CGObjectNode}, adding a points-to edge if {@code this} is a
	 * {@code CGReferenceNode} or throwing an exception if {@code this} is a {@code CGObjectNode} (because you can't
	 * attach object nodes to other object nodes directly)
	 *
	 * @param objectNode
	 *            the object node to attach to {@code this}
	 * @return true, if the operation added a new edge
	 * @throws CompileException
	 */
	public abstract boolean attach(final CGObjectNode objectNode) throws CompileException;

	/**
	 * Implementation of the visitor pattern for {@code CGNode}s, allowing the dot writer to walk the tree of
	 * {@code CGNode}s.
	 *
	 * @param visitor
	 *            visitor visiting the nodes
	 * @throws CompileException
	 */
	public abstract void accept(final CGDotWriter visitor) throws DotWriterException, CompileException;

	/**
	 * Method called after processing a method, applying {@code byPass()} to remove the remaining deferred edges, unless
	 * a deferred edge cannot be resolved (which happens when the node pointed to by the deferred edge doesn't have any
	 * outgoing edges). This method does nothing for object nodes, but is still defined on object nodes to save
	 * {@code instanceof CGReferenceNode} and a cast to {@code CGReferenceNode}.
	 */
	public void pathCompression() {}

	/**
	 * Method returning all nodes reachable from this node
	 *
	 * @return a collection of nodes that is in any way reachable from this node
	 */
	public abstract Collection<? extends CGNode> getDescendants();
}
