/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMAllocation;
import keso.compiler.imcode.IMField;
import keso.compiler.imcode.IMMethod;
import keso.util.DotWriterException;
import keso.util.SplitKeyHashMap;
import keso.util.SplitKeyMap;

/**
 * Object node in the connection graph
 */
public class CGObjectNode extends CGNode {
	public static final IMField											arrayAccessIMField			= null;

	/**
	 * Map mapping allocation sites to their corresponding object nodes in the connection graph
	 */
	private static SplitKeyMap<IMAllocation, IMMethod, CGObjectNode>	allocationToObjectMap		= new SplitKeyHashMap<IMAllocation, IMMethod, CGObjectNode>();

	/**
	 * Map mapping fields to the corresponding field reference nodes in the connection graph
	 */
	private final Map<IMField, CGFieldReferenceNode>					fieldToFieldReferenceMap	= new HashMap<IMField, CGFieldReferenceNode>();

	/**
	 * Holds a reference to the allocation site of this object
	 */
	private final IMAllocation											allocation;

	/**
	 * List of field references in this object; those are the field edges in the collection graph as specified by the
	 * implemented paper. For each field in this object there should be an entry in this list to a
	 * {@code CGFieldReferenceNode} describing referencing the field in question.
	 */
	private List<CGFieldReferenceNode>									fields						= new LinkedList<CGFieldReferenceNode>();

	/**
	 * Return the object node associated with a given allocation site. Creates a new {@code CGObjectNode} if none exists
	 * yet.
	 *
	 * @param allocation
	 *            the allocation site to return the object node for
	 * @return the object node in the connection graph corresponding to the allocation site
	 */
	public static CGObjectNode getObjectNodeForAllocation(IMAllocation allocation, IMMethod method) {
		CGObjectNode objectNode = allocationToObjectMap.get(allocation, method);

		// if there's no object node created for this allocation site yet, create one
		if (objectNode == null) {
			allocationToObjectMap.put(allocation, method, objectNode = new CGObjectNode(allocation));
		}

		return objectNode;
	}

	/**
	 * Return a field reference node to the specified field {@code field} inside this object. This method will construct
	 * a CGFieldReferenceNode if none exists for this combination of object and field yet.
	 *
	 * @param field
	 *            the field referenced
	 * @return a {@code CGFieldReferenceNode} representing an edge from this object to the field
	 */
	public CGFieldReferenceNode getFieldReferenceForField(final IMField field) {
		CGFieldReferenceNode fieldReferenceNode = fieldToFieldReferenceMap.get(field);

		// if there's no field reference node in the connection graph for this combination of object and field yet,
		// create one
		if (fieldReferenceNode == null) {
			fieldToFieldReferenceMap.put(field, fieldReferenceNode = new CGFieldReferenceNode(this, field));
			addField(fieldReferenceNode);
		}

		return fieldReferenceNode;
	}

	/**
	 * Construct a new object node in the connection graph; since object nodes are created by an allocation operation
	 * and we want to transform non-escaping objects' allocations into stack allocations we need a reference to the
	 * allocation site, which is given as a parameter to the constructor
	 *
	 * @param allocation
	 *            {@code IMAllocation} object, which describes the intermediate code instruction allocation this object
	 */
	protected CGObjectNode(final IMAllocation allocation) {
		this.allocation = allocation;
	}

	/**
	 * Add a field reference from this object to a field denoted by {@code field}.
	 *
	 * @param field
	 *            The field reference that describes the field
	 */
	public void addField(final CGFieldReferenceNode field) {
		fields.add(field);
	}

	/**
	 * Returns an unmodifiable list of field references in this object
	 *
	 * @return list of field references
	 */
	public List<CGFieldReferenceNode> getFields() {
		return Collections.unmodifiableList(fields);
	}

	/**
	 * Returns the allocating intermediate code instruction for this object. Note that this method can return
	 * {@code null} and will do so for objects representing storage associated with a class rather than an instance of a
	 * class.
	 *
	 * @return allocation instruction
	 */
	public IMAllocation getAllocation() {
		return allocation;
	}

	/**
	 * Returns a representation of a given object node in the connection graph of a given method {@code method}. Creates
	 * at most one object node per allocation site in a method's connection graph.
	 *
	 * @param method
	 *            the method, who's connection graph shall hold the new object node
	 * @return the new object node
	 */
	public CGObjectNode getCloneInMethod(CGReferenceNode reference, IMMethod method) {
		CGObjectNode object = getObjectNodeForAllocation(allocation, method);
		object.attach(reference);
		return object;
	}

	@Override
	public boolean attach(final CGReferenceNode reference) {
		boolean wasAdded = reference.pointsTo(this);

		if (wasAdded)
			updateEscapeState(reference.getEscapeState());

		return wasAdded;
	}

	@Override
	public boolean attach(final CGObjectNode objectNode) throws CompileException {
		throw new CompileException("CGObjectNodes cannot point to CGObjectNodes directly!");
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}

	@Override
	public Collection<? extends CGNode> getDescendants() {
		return getFields();
	}

	/**
	 * @return true, if this object is a phantom node and can safely be cast to {@code CGObjectPhantomNode}
	 */
	public boolean isPhantom() {
		return false;
	}
}
