/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import keso.compiler.CompileException;
import keso.compiler.imcode.IMMethod;
import keso.util.DotWriterException;

/**
 * Phantom object node in the connection graph. Phantom nodes are used when an object node should be referenced by a
 * reference node, but we don't know where it actually comes from (this can happen, if a reference always is null, in
 * which case we conservatively assume we just don't know about the object in question). Phantom nodes are also used for
 * return values and parameters.
 */
public class CGObjectPhantomNode extends CGObjectNode {
	/**
	 * Map mapping reference nodes to existing phantom object nodes for those references
	 */
	private static Map<CGReferenceNode, CGObjectPhantomNode>				referenceToPNMap		= new HashMap<CGReferenceNode, CGObjectPhantomNode>();

	/**
	 * A map mapping a tuple of (object node, method) to a phantom node that may already have been created for this
	 * tuple.
	 */
	private static Map<CGObjectNode, Map<IMMethod, CGObjectPhantomNode>>	objectToMethodToPNMap	= new HashMap<CGObjectNode, Map<IMMethod, CGObjectPhantomNode>>();

	/**
	 * The node in the connection graph of a called method that caused the creation of this phantom node, or
	 * {@code null}. Note that this might be another [@code CGObjectPhantomNode}!
	 */
	private CGObjectNode													source;

	/**
	 * The method that contains {@code source}, or null
	 */
	private IMMethod														sourceMethod;

	/**
	 * Constructs a new phantom object node; since there is no allocation site for a phantom node, we call
	 * {@code super(null);}, i.e. set the allocation site to null
	 */
	private CGObjectPhantomNode(final CGReferenceNode reference) {
		// there is no allocation site for phantom nodes
		super(null);
		// add a points-to edge from the reference to this phantom node
		attach(reference);
	}

	/**
	 * Constructs a new phantom object node from a given source object in a given source method
	 *
	 * @param source
	 *            the object that caused the creation of this phantom node
	 * @param sourceMethod
	 *            the method that contains the source
	 */
	private CGObjectPhantomNode(final CGObjectNode source, final IMMethod sourceMethod) {
		super(null);
		this.source = source;
		this.sourceMethod = sourceMethod;
	}

	/**
	 * Returns a phantom node for the given reference {@code reference}, creating one if none exists yet.
	 *
	 * @param reference
	 *            the reference to obtain the phantom node for
	 * @return a phantom node for the given reference.
	 */
	public static Collection<CGObjectPhantomNode> getPhantomNodeForReference(final CGReferenceNode reference) {
		Collection<CGObjectPhantomNode> phantomNodes = new LinkedList<CGObjectPhantomNode>();

		if (reference.getDeferredEdges().isEmpty()) {
			CGObjectPhantomNode phantomNode = referenceToPNMap.get(reference);

			if (phantomNode == null) {
				referenceToPNMap.put(reference, phantomNode = new CGObjectPhantomNode(reference));
			}

			phantomNodes.add(phantomNode);

			return phantomNodes;
		} else {
			// add the phantom nodes to the ends of the deferred edges instead
			for (CGReferenceNode def : reference.getDeferredEdges()) {
				phantomNodes.addAll(CGObjectPhantomNode.getPhantomNodeForReference(def));
			}

			return phantomNodes;
		}
	}

	public static CGObjectNode getPhantomNode(final CGReferenceNode reference, final CGObjectNode calleeObjectNode, final IMMethod callee,
		final IMMethod caller) {
		if (callee == caller) {
			calleeObjectNode.attach(reference);
			return calleeObjectNode;
		}

		Map<IMMethod, CGObjectPhantomNode> methodToPNMap = objectToMethodToPNMap.get(calleeObjectNode);
		if (methodToPNMap == null) {
			objectToMethodToPNMap.put(calleeObjectNode, methodToPNMap = new HashMap<IMMethod, CGObjectPhantomNode>());
		}

		CGObjectPhantomNode phantomNode = methodToPNMap.get(caller);
		if (phantomNode == null) {
			// let's see if this phantom node was caused by itself somewhere down the call-graph (and if it was, don't
			// create a new node)
			CGObjectNode n = calleeObjectNode;
			while (true) {
				if (n != null && n.isPhantom()) {
					CGObjectPhantomNode pn = ((CGObjectPhantomNode) n);
					if (pn.source != null) {
						if (pn.sourceMethod == callee && pn.source == calleeObjectNode) {
							calleeObjectNode.attach(reference);
							return calleeObjectNode;
						}
						n = pn.source;
						continue;
					}
				}
				break;
			}

			phantomNode = new CGObjectPhantomNode(calleeObjectNode, callee);
			methodToPNMap.put(caller, phantomNode);
		}

		phantomNode.attach(reference);

		return phantomNode;
	}

	/**
	 * Moves a phantom node, if one exists for a node {@code node} to a descendant reference node {@code value}. This
	 * method is called when a new deferred edge is added to a node, i.e. when we encounter an assignment {@code p = q;}
	 * and there already is a phantom node for {@code p}.
	 *
	 * @param node
	 *            the node the deferred edge is being added to and that potentially has the phantom object node
	 * @param value
	 *            the pointee of the new deferred edge
	 */
	public static void movePhantomNode(final CGReferenceNode node, final CGReferenceNode value) {
		CGObjectPhantomNode phantom = referenceToPNMap.get(node);
		if (phantom == null)
			return;

		referenceToPNMap.remove(node);
		node.pointsTo.remove(phantom);

		Set<CGObjectNode> pointsTo = value.getPointsToRelation();

		if (pointsTo.isEmpty()) {
			// we need a phantom node there
			pointsTo.addAll(getPhantomNodeForReference(value));
		}

		for (CGObjectNode objectNode : pointsTo) {
			// add all outgoing edges of the old phantom node to whatever is in the points-to-relation of the new
			// pointee of the deferred edge
			for (CGFieldReferenceNode fieldReference : phantom.getFields()) {
				CGFieldReferenceNode newFieldReference = objectNode.getFieldReferenceForField(fieldReference.getField());

				// move incoming deferred edges
				for (CGReferenceNode revDeferredEdge : fieldReference.reverseDeferredEdges) {
					revDeferredEdge.deferredEdges.remove(fieldReference);
					newFieldReference.attach(revDeferredEdge);
				}
				fieldReference.reverseDeferredEdges.clear();

				// move deferred edges
				for (CGReferenceNode deferredEdge : fieldReference.getDeferredEdges()) {
					deferredEdge.attach(newFieldReference);
				}

				// move points-to edges
				for (CGObjectNode pointsToEdge : fieldReference.getPointees()) {
					pointsToEdge.attach(newFieldReference);
				}
			}
		}
	}

	@Override
	public CGObjectNode getCloneInMethod(CGReferenceNode reference, IMMethod method) {
		return getPhantomNode(reference, source, sourceMethod, method);
	}

	@Override
	public void accept(final CGDotWriter visitor) throws DotWriterException, CompileException {
		visitor.visit(this);
	}

	@Override
	public boolean isPhantom() {
		return true;
	}
}
