/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.analysis.connectiongraph;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Abstract superclass for all reference nodes in the connection graph
 */
abstract public class CGReferenceNode extends CGNode {
	/**
	 * List holding references to all reference nodes created in the currently processed method; clear this list after
	 * each method by calling {@code clearReferenceNodeList()}.
	 */
	protected static List<CGReferenceNode>	referenceNodeList	= new LinkedList<CGReferenceNode>();

	/**
	 * Return a list of reference nodes created in the currently processed method
	 *
	 * @return unmodifiable list of reference nodes
	 */
	public static List<CGReferenceNode> getReferenceNodeList() {
		return Collections.unmodifiableList(referenceNodeList);
	}

	/**
	 * Clears the list of reference nodes since the last clear; call this once for each method you're processing.
	 */
	public static void clearReferenceNodeList() {
		referenceNodeList.clear();
	}

	/**
	 * List of pointees pointed to by points-to edges in the collection graph. Points-to edges are represented by this.
	 * All entries in this list are pointees and form a points-to edge from {@code this} to the pointee.
	 */
	protected Set<CGObjectNode>		pointsTo				= new HashSet<CGObjectNode>();

	/**
	 * List of reference nodes that this reference node has deferred edges to. A deferred edge in the connection graph
	 * is generated when a reference variable is assigned to another reference variable, i.e. {@code lvalue = rvalue},
	 * where {@code lvalue} and {@code rvalue} are references. For such an assignment, a deferred edge from the
	 * {@code CGReferenceNode} representing {@code lvalue} to the {@code CGReferenceNode} representing {@code rvalue}
	 * would be added.
	 *
	 * These edges are used to minimize the effort and computing power used for rebuilding the graph, because rebuilding
	 * parts of the graph as would usually be required for an assignment is deferred (hence the name) until the
	 * {@code byPass} method is called. This method copies the entries in {@code deferredEdges[i].pointsTo} into
	 * {@code this.pointsTo} and removes the deferred edge.
	 */
	protected Set<CGReferenceNode>	deferredEdges			= new HashSet<CGReferenceNode>();

	/**
	 * List of reference nodes that have deferred edges pointing to this node. This list is needed to be able to apply
	 * the {@code byPass()} function.
	 */
	protected Set<CGReferenceNode>	reverseDeferredEdges	= new HashSet<CGReferenceNode>();

	/**
	 * Add a new points-to edge from this object to {@code pointee}.
	 *
	 * @param pointee
	 *            {@code CGObjectNode} to point to
	 * @return true, if the pointee was added, false if it already existed
	 */
	public boolean pointsTo(final CGObjectNode pointee) {
		return pointsTo.add(pointee);
	}

	/**
	 * Add a new deferred edge from this object to {@code value}, which usually is the right-hand side of an assignment,
	 * i.e. an rvalue.
	 *
	 * @param value
	 *            {@code CGReferenceNode} that has been assigned to this reference
	 * @return true, if the call modified the graph, false, if the deferred edge was already present
	 */
	public boolean addDeferredEdge(final CGReferenceNode value) {
		boolean added;

		// don't add self-referential edges, they're useless
		if (this == value)
			return false;

		added = deferredEdges.add(value);
		added |= value.reverseDeferredEdges.add(this);

		if (added && !pointsTo.isEmpty()) {
			CGObjectPhantomNode.movePhantomNode(this, value);
		}

		return added;
	}

	/**
	 * Returns an unmodifiable set of object nodes for which a points-to edge from this object to the specified pointee
	 * exists.
	 *
	 * @return set of pointee objects
	 */
	public Set<CGObjectNode> getPointees() {
		return Collections.unmodifiableSet(pointsTo);
	}

	/**
	 * Returns a set of object nodes that this reference points to directly and/or indirectly (via deferred edges). This
	 * is the PointsTo(this) relation as described in the paper.
	 *
	 * @return set of pointees
	 */
	public Set<CGObjectNode> getPointsToRelation() {
		return getPointsToRelation(new HashSet<CGObjectNode>(), new HashSet<CGReferenceNode>());
	}

	private Set<CGObjectNode> getPointsToRelation(Set<CGObjectNode> pointees, Set<CGReferenceNode> visited) {
		pointees.addAll(pointsTo);
		visited.add(this);

		for (CGReferenceNode deferredEdge : deferredEdges) {
			if (visited.contains(deferredEdge))
				continue;
			visited.add(deferredEdge);

			deferredEdge.getPointsToRelation(pointees, visited);
		}

		return pointees;
	}

	/**
	 * Returns an umodifiable list of reference nodes for which a deferred edge from this object to the returned pointee
	 * exists.
	 *
	 * @return list of pointed-to reference nodes
	 */
	public Set<CGReferenceNode> getDeferredEdges() {
		return Collections.unmodifiableSet(deferredEdges);
	}

	/**
	 * Resolves the deferred edges and adds points-to edges instead
	 */
	@Deprecated
	public void byPass() {
		Iterator<CGReferenceNode> it = reverseDeferredEdges.iterator();

		// for all incoming deferred edges...
		while (it.hasNext()) {
			CGReferenceNode deferredEdgeSource = it.next();

			// ... remove the incoming edge ...
			deferredEdgeSource.deferredEdges.remove(this);
			it.remove();

			// ... and add edges to this object's pointees instead
			for (CGReferenceNode deferredEdge : deferredEdges) {
				deferredEdge.attach(deferredEdgeSource);
			}
			for (CGObjectNode objectNode : pointsTo) {
				objectNode.attach(deferredEdgeSource);
			}
		}

		deferredEdges.clear();
		pointsTo.clear();
	}

	@Override
	public void pathCompression() {
		if (!reverseDeferredEdges.isEmpty()) {
			// this node has incoming deferred edges and thus needs to be processed
			if (deferredEdges.isEmpty() && pointsTo.isEmpty()) {
				// handle terminal nodes by creating phantom nodes
				// we know this can only contain one item
				CGObjectPhantomNode.getPhantomNodeForReference(this).iterator().next();

				pathCompressionHelper();
				// We don't need to ensure
				// "However, we keep the points-to edges from the terminal nodes to the phantom nodes."
				// manually, because {@code pathCompressionHelper()} doesn't clear the pointsTo list.
			} else {
				// this node is not a terminal node
				pathCompressionHelper();
				setStale(true);
			}
		}
	}

	/**
	 * Helper function removing all incoming deferred edges pointing to {@code this} node by rewriting all incoming
	 * edges to point to this node's descendants. Only call this method after having ensured there is at least one
	 * outgoing edge from this node.
	 */
	private void pathCompressionHelper() {
		Iterator<CGReferenceNode> it = reverseDeferredEdges.iterator();

		// for all incoming deferred edges...
		while (it.hasNext()) {
			CGReferenceNode source = it.next();
			// remove the incoming edge, leave the one outgoing to this node, because we need it in the stack allocation
			// transformation to determine whether a given allocation is reachable from the set of live variables
			source.deferredEdges.remove(this);
			it.remove();

			// and add edges to this object's pointees instead
			for (CGReferenceNode deferredEdge : deferredEdges) {
				deferredEdge.attach(source);
			}
			for (CGObjectNode objectNode : pointsTo) {
				objectNode.attach(source);
			}
		}
	}

	@Override
	public boolean attach(final CGReferenceNode reference) {
		boolean wasAdded = reference.addDeferredEdge(this);

		if (wasAdded)
			updateEscapeState(reference.getEscapeState());

		return wasAdded;
	}

	@Override
	public boolean attach(final CGObjectNode objectNode) {
		boolean wasAdded = pointsTo.add(objectNode);

		if (wasAdded)
			objectNode.updateEscapeState(getEscapeState());

		return wasAdded;
	}

	@Override
	public Collection<? extends CGNode> getDescendants() {
		LinkedList<CGNode> list = new LinkedList<CGNode>(pointsTo);
		list.addAll(deferredEdges);
		return list;
	}
}
