/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.backend;

import keso.classfile.datatypes.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.messageport.*;
import keso.compiler.config.task.*;
import keso.compiler.imcode.*;
import keso.util.*;

import keso.compiler.backend.gradmp.GradMPChecks;

import java.io.*;

import java.util.*;
import java.util.Set;

public abstract class Coder {

	private GradMPChecks	gradmpc;

	protected IMModule module;
	protected IMMethod currentMethod;
	protected BuilderOptions opts;

	protected CodeInfo codeInfo = new CodeInfo();

	protected String code_alias;
	protected StringBuffer mbody;
	protected StringBuffer mbodyOuter;
	protected StringBuffer mbodyFrame;
	protected StringBuffer local;
	protected StringBuffer line;

	protected StringBuffer cheader;
	protected StringBuffer chline;

	// add some code to the beginning of the startup hook
	protected StringBuffer startupHook = new StringBuffer();

	/**
	 * Includes for the currently processed function module.
	 */
	protected SortedSet<String> includes;

	protected String indent;

	protected List<MakeRule> rules = new ArrayList<MakeRule>();
	protected StringBuffer global;
	protected StringBuffer global_header;

	/**
	 * Includes for global.c
	 */
	protected SortedSet<String> global_includes;

	protected Set<IMMethod> classConstructors = new HashSet<IMMethod>();

	protected static class StaticAllocObjectConstructor {
		final private String alias;
		final private IMClass type;
		final private Set<IMDomain> domains;

		StaticAllocObjectConstructor(String alias, IMClass type, Set<IMDomain> domains) {
			this.alias = alias;
			this.type = type;
			this.domains = domains;
		}

		public boolean equals(Object o) {
			if(o instanceof StaticAllocObjectConstructor) {
				return alias.equals( ((StaticAllocObjectConstructor)o).alias );
			}
			return false;
		}

		public int hashCode() {
			return alias.hashCode();
		}

		public void write(PrintStream out, BuilderOptions opts, Coder coder) throws CompileException {
			IMMethod init_method = type.getMethod("<init>()V");
			if (init_method == null)
				throw new CompileException("the class " + type + " does not provide a default constructor <init>()V, but the execution of such a constructor was requested for the statically allocated object " + alias);

			// do not call empty constructors
			if(init_method.getCosts() == 0)
				return;

			for(IMDomain dom: domains) {
				if (!opts.isSingleDomainSystem()) {
					out.println("\tKESO_CURRENT_DOMAIN=" + coder.getDDescPtr(dom.getDef()) + ";");
				}

				coder.writeCallToInit(out, init_method, "\t", alias);
			}
		}
	}

	private Set<StaticAllocObjectConstructor> static_alloc_object_constructors
	= new HashSet<StaticAllocObjectConstructor>();

	/**
	 * Register a constructor call for a statically allocated object that needs to be called on
	 * startup.
	 *
	 * Note that the constructor will be called in each of the given domains on the same object,
	 * which is only safe in limited cases. For the general case, you should create a separate
	 * object for each of the domains and register a constructor call for each of these objects
	 * in its (single) domain.
	 *
	 * The constructor invoked will be the default constructor (<init>()V), so the given class
	 * must provide such a constructor.
	 *
	 * @param alias  name (in C code) of the statically allocated object
	 * @param type   class of the statically allocated object
	 * @param doms   set of domains in that the constructor is to be invoked
	 */
	public void register_static_object_constructor(String alias, IMClass type, Set<IMDomain> doms) {
		static_alloc_object_constructors.add(new StaticAllocObjectConstructor(alias,type,doms));
	}

	/**
	 * Includes for global.h
	 */
	protected SortedSet<String> global_header_includes;

	protected MethodTableFactory mdispatch;
	protected ClassStore repository;

	protected IMBasicBlock currentBB;

	protected InterfaceTypeMatrix ifmatrix;

	private SortedSet<String> already_added_funcs = new TreeSet<String>();

	protected Coder(BuilderOptions opts, MethodTableFactory table,
			ClassStore store) throws CompileException {
		global = new StringBuffer();
		global_header = new StringBuffer();
		global_includes = new TreeSet<String>();
		global_header_includes = new TreeSet<String>();
		mdispatch = table;
		repository = store;
		this.opts=opts;
		ifmatrix = new InterfaceTypeMatrix(opts, this, store);
		gradmpc = new GradMPChecks( store, opts );
	}

	public InterfaceTypeMatrix getIFMatrix() {
		return ifmatrix;
	}

	public void setCurrentBasicBlock(IMBasicBlock bb) {
		currentBB = bb;
	}

	public IMBasicBlock getCurrentBasicBlock() {
		return currentBB;
	}

	public ClassStore getClassStore() {
		return repository;
	}

	/**
	 * Returns a set of all domains that the currently processed basic block is reachable from.
	 */
	public Set<IMDomain> getCurrentDomains() {
		Set<IMDomain> doms = getCurrentBasicBlock().getDomains();
		return doms;
	}

	/**
	 * Returns the domain that the currently processed basic block is reachable from.
	 *
	 * Throws an exception if the basic block is reachable from more than one domain.
	 * Do not call this function unless you expect that the basic block is reachable
	 * from only one domain. Otherwise, use {@link #getCurrentDomains()}.
	 */
	public IMDomain getCurrentDomain() throws CompileException {
		Set<IMDomain> doms = getCurrentBasicBlock().getDomains();
		if(doms.size() == 0)
			return null;
		if(doms.size() > 1)
			throw new CompileException("getCurrentDomain() called from context that is not unique to a domain");
		return doms.iterator().next();
	}

	public void setCodeInfo(CodeInfo codeInfo) {
		this.codeInfo = codeInfo;
	}

	public CodeInfo getCodeInfo() {
		return codeInfo;
	}

	/**
	 * Add an int to the globals of the C-Target
	 */
	public void global_add(int str) throws CompileException {
		global.append(str);
	}

	/**
	 * Add a line to the globals of the C-Target
	 */
	public void global_add(String str) throws CompileException {
		global.append(str);
	}

	public void global_add(Object str) throws CompileException {
		global.append(str);
	}

	public void global_addln(String str) throws CompileException {
		global.append(str);
		global.append("\n");
	}

	public void global_addln(int str) throws CompileException {
		global.append(str);
		global.append("\n");
	}

	public void global_addln(Object str) throws CompileException {
		global.append(str);
		global.append("\n");
	}

	public void global_addln() throws CompileException {
		global.append("\n");
	}

	public void global_header_add_asm(String asm) throws CompileException {
		global_header.append("__asm__ __volatile__ (");
		global_header.append(asm);
		global_header.append(");");
	}

	/**
	 * Add an int to the globals of the C-Target
	 */
	public void global_header_add(int str) throws CompileException {
		global_header.append(str);
	}

	/**
	 * Add a line to the globals of the C-Target
	 */
	public void global_header_add(String str) throws CompileException {
		global_header.append(str);
	}

	public void beginClass(IMModule module) throws CompileException {
		this.module  = module;
		this.opts   = module.getOptions();

		this.local   = new StringBuffer();
		this.chline  = new StringBuffer();

		beginHeader(module);
	}

	public void beginHeader(IMModule module) throws CompileException {
		this.module  = module;

		this.cheader = new StringBuffer();

		header_add("/* Class: ");
		header_add(module.getClassName());
		header_addln(" */\n");
		header_add("#ifndef __");
		header_add(module.getAlias());
		header_addln("_H__");
		header_add("#define __");
		header_add(module.getAlias());
		header_addln("_H__ 1");
		header_addln("#include \"keso_types.h\"");
		header_addln("#include \"global.h\"");
	}

	public void addInvokeMakro(IMMethod method) throws CompileException {
		MethodTable table = method.getMethodTable();
		if (table==null) return;
		table.emit_invoke_macro(method, this);
	}

	public void beginClassMethod(IMModule module) throws CompileException {
		mbody = new StringBuffer();
		line  = new StringBuffer();
		includes = new TreeSet<String>();
		code_alias = module.getAlias();

		indent = "";
	}

	public String extraProxyHelperArgString() throws CompileException {
		return ", domain_t *CALLER";
	}

	public String extraProxyHelperArgs(ServiceDefinition srv) throws CompileException {
		return ", KESO_CURRENT_DOMAIN";
	}
	// Domain state recovery in case of replica fault
	public void add_StateRecoveryMethod(IMMethod proxy, ServiceDefinition srv) throws CompileException {

	}
	// Generates the voting method (for module replication). Necessary for complex data types
	public void add_PortalVoterMethod(IMMethod proxy, ServiceDefinition srv) throws CompileException {
		boolean has_return_value = (proxy.getBasicReturnType()!=BCBasicDatatype.VOID);
		if(has_return_value==false)
			throw new CompileException("Configured replicated service does not have a return value");
		StringBuffer helper = new StringBuffer();
		StringBuffer arg = null;
		// get the class type of the method's return value
		String classname = proxy.getMethodData().getReturnType().getClassName();
		IMClass returnClass = repository.getClass(classname);
		// get all primitive class fields
		Vector<IMField> primFields = returnClass.getPrimitiveAttributes();
		String returnType=null;
		// Get all domains, a service is defined in
		List<JavaDomain> services_domains = srv.getDomains();
		int replica_cnt = services_domains.size();
		List<StringBuffer> arguments = new ArrayList<StringBuffer>();

		returnType=proxy.getReturnType();
		helper.append(returnType);
		helper.append(" "+proxy.toString()+"_vote(");
	  // set all arguments
		for(int i=0; i < services_domains.size(); i++) {
				helper.append(returnType);
				helper.append(" ");
				arg=new StringBuffer("ret_value_");
				arg.append(getDomainSymbolPrefix(services_domains.get(i)));
				arguments.add(arg);
				helper.append(arg);
				arg=null;
				// last argument
				if(i==(services_domains.size()-1)){
					break;
				}
				helper.append(", ");
		}
		helper.append(") {\n");
		helper.append("\t" +returnType + " ret;\n");
		for(int i=0; i<primFields.size(); i++) {
			helper.append("\t" + primFields.get(i).getTypeString() + " r_" +i + ";\n");
		}
		// fastVote over object fields
		for(int i=0; i<primFields.size(); i++) {
			String fieldname = primFields.get(i).getFieldName();
			//helper.append("\t" + getField(returnClass, "ret", fieldname) + " = fastVote_");
			helper.append("\tr_" + i + " = "+proxy.toString()+"_fastVote_");
			helper.append(primFields.get(i).getTypeString() + "(");
			// get all arguments
			for(int j=0; j<services_domains.size(); j++) {
				helper.append(getField(returnClass, arguments.get(j).toString(), fieldname));
				// last argument
      	if(j==(services_domains.size()-1)){
					break;
				}
				helper.append(", ");
			}
			helper.append(");\n");
		}
		// voting ok
		helper.append("\tif(");
		for(int i=0; i<primFields.size(); i++) {
			helper.append("r_" + i + " !=-1");
			if(i == primFields.size()-1) {
				break;
			}
			helper.append(" && ");
		}
		helper.append(") {\n\t\treturn " + arguments.get(0).toString() + ";\n\t}\n");
		// signal fatal error
		helper.append("\treturn NULL;\n}\n\n");
		mbodyOuter.append(helper);
	}


	// Generates the fast voting method (for module replication)
	public void add_PortalFastVoterMethod(IMMethod proxy, ServiceDefinition srv) throws CompileException {
		boolean has_return_value = (proxy.getBasicReturnType()!=BCBasicDatatype.VOID);
		String arrayType=null;
		if(has_return_value==false)
			throw new CompileException("Configured replicated service does not have a return value");
		IMMethod service = srv.getServiceClass(repository).getMethod(proxy.getMethodNameAndType());
		StringBuffer helper = new StringBuffer();
		StringBuffer arg = null;
		IMClass returnClass = null;
		BasicTypeDescriptor typeDesc = proxy.getMethodData().getReturnType();
		// get the class type of the method's return value
		String classname = typeDesc.getClassName();

		// get correct identifier for array types
		if(typeDesc.isArray() == true) {
			returnClass =	proxy.getIMClass(typeDesc.getFullArrayTypeDesc());
			add_class(returnClass.getAlias());
		}	else { // for any other types
			returnClass = repository.getClass(classname);
			add_class(returnClass);
		}

		// Get all domains, a service is defined in
		List<JavaDomain> services_domains = srv.getDomains();
		int replica_cnt = services_domains.size();
		List<StringBuffer> arguments = new ArrayList<StringBuffer>();
		List<StringBuffer> domid = new ArrayList<StringBuffer>();
		//String rettype=null;
		// if the return value is an object, get all different object field types
		Set<String> fieldtypes = new HashSet<String>();
		// fastVote must be called on primitive data, so get all primitive object field types
		if(proxy.getBasicReturnType()==BCBasicDatatype.REFERENCE && typeDesc.isArray() == false) {
			Vector<IMField> primFields = returnClass.getPrimitiveAttributes();
			for(IMField field : primFields) {
				fieldtypes.add(field.getTypeString());
			}
		}
		// handle arrays
		if(typeDesc.isArray() == true){
			arrayType = ((IMArrayClass)returnClass).getElementType();
			fieldtypes.add(arrayType);
			//throw new CompileException("Replica support for arrays is not implemented yet " + ((IMArrayClass)returnClass).getElementType());
		}
		// return type is primitive
		if (typeDesc.isArray() == false){
			if(proxy.getReturnType().compareTo("object_t*") != 0)
				fieldtypes.add(proxy.getReturnType());
		}
		//rettype=proxy.getReturnType();
		// create a method for each datatype
		// if the return type is an array type, fieldtypes contains the type of the array
		int was_in=0;
		for(String returnType : fieldtypes) {
			if(typeDesc.isArray() == true) {
				helper.append("object_t* " +proxy.toString()+"_arrayVote(");
			}
			if(typeDesc.isArray() == false) {
				helper.append(returnType);
				helper.append(" "+proxy.toString()+"_fastVote_" + returnType +"(");
			}
			for(int i=0; i < services_domains.size(); i++) {
				if(typeDesc.isArray() == true) {
					helper.append("object_t* ");
				}
				if(typeDesc.isArray() == false) {
					helper.append(returnType);
					helper.append(" ");
				}
				arg=new StringBuffer("ret_value_");
				arg.append(getDomainSymbolPrefix(services_domains.get(i)));
				domid.add(new StringBuffer(getDomainSymbolPrefix(services_domains.get(i))));
				if(was_in == 0)
					arguments.add(arg);
				helper.append(arg);
				arg=null;
				// last argument
				if(i==(services_domains.size()-1)){
					break;
				}
				helper.append(", ");
			}
			helper.append(") {\n");

			List<StringBuffer> list_acc;
			List<StringBuffer> list_acc_2;
			List<StringBuffer> vars=null;
			if(typeDesc.isArray() == true) {
				list_acc = new ArrayList<StringBuffer>();
				list_acc_2 = new ArrayList<StringBuffer>();
				vars = new ArrayList<StringBuffer>();
			}
			else {
				list_acc = new ArrayList<StringBuffer>(arguments);
				list_acc_2 = arguments;
			}

			/*List<StringBuffer> list_acc = new ArrayList<StringBuffer>();
				List<StringBuffer> list_acc_2;
				if(typeDesc.isArray() == true)
				list_acc_2 = new ArrayList<StringBuffer>();
				else
				list_acc_2 = arguments;*/
			//if(typeDesc.isArray() == true) {
			if(typeDesc.isArray() == true) {
				String array_ptr = new String(returnClass.getAlias()+"_t*");
				helper.append("\tint size = (("+array_ptr+")"+arguments.get(0)+")->size;\n");
				// omit array header
				// cast to correct array types
				for(int i=0; i < services_domains.size(); i++) {
					helper.append("\t" +array_ptr + " a_" + i + " = ("+array_ptr+")"+arguments.get(i)+";\n");
					// if arrays have to be compared
					StringBuffer temp=new StringBuffer("a_"+i);
					vars.add(0, temp);
					//temp.append("->data[i]");
					list_acc.add(0, temp);
					list_acc_2.add(0, temp);
				}
			}
			// support NMR
			List<StringBuffer> worked_list = new ArrayList<StringBuffer>();
			List<StringBuffer> failed_replicas = new ArrayList<StringBuffer>();
			String arr_el=null;
			if(typeDesc.isArray() == true) {
				arr_el=new String("->data[i]");
			} else {
				arr_el=new String("");
			}
			//int r_cnt = list.size();
			int nFaults = (replica_cnt-1)/2;
			int success_r = replica_cnt-nFaults;
			// this is nFaults out of replica_cnt, that is
			//int comp_levels = fac(replica_cnt)/fac(nFaults)/fac(replica_cnt-nFaults);
			// start comparison
			if(typeDesc.isArray() == true) {
				helper.append("\tobject_t* result;\n");
				helper.append("\tfor(int i=0; i<size; i++) {\n");
			} else {
				helper.append("\t"+returnType+" result;\n");
			}
			for(int l=0; l<replica_cnt-1; l++) {
				// get the first parameter to start the comparison
				StringBuffer current = list_acc.remove(l);
				worked_list.add(0,current);// first level
				for(int z=0; z<list_acc.size(); z++) {
					if(worked_list.contains(list_acc.get(z)))
						continue;
					if(z==0) {
						helper.append("\tif("+current+arr_el+"=="+list_acc.get(z)+arr_el+") {\n");
					}
					else {
						helper.append("\telse if("+current+arr_el+"=="+list_acc.get(z)+arr_el+") {\n");
					}
					// begin array loop, if data to be compared is an array
					worked_list.add(0, list_acc.remove(z));
					// the number of sublevel comparisons are dependend on the possible success rate for a 2n+1 replica system
					/*for(int k=0; k<nFaults-1; k++) {
						helper.append("\t\tif("+current+"=="+list.get(k)+") {\n");
						current=list.remove(k);
						worked_list.add(current);
						}*/
					helper.append("\t\t\tresult="+current+";\n");
					for(int m=0; m<nFaults; m++) {
						helper.append("\t\t\tif("+worked_list.get(0)+arr_el+"!="+list_acc.get(m)+arr_el+") {\n");
						failed_replicas.add(list_acc.get(m));
						// recover domain state (restricted domain scope GC)
						helper.append("\t\t\t\t/* State recovery: Copy heap (restricted domain scope), static class fields and other domain information */ \n");
						// get current domain identifier
						int idx=list_acc_2.indexOf(worked_list.get(0));
						StringBuffer src = domid.get(idx);
						idx=list_acc_2.indexOf(list_acc.get(m));
						StringBuffer dest=domid.get(idx);
						helper.append("\t\t\t\tkeso_memcpy((jbyte*)(&"+dest+"_DDesc), (jbyte*)");
						helper.append("(&"+src+"_DDesc), sizeof(struct domain_rds_t));\n");
						// deep copy references
						// GROESSE? FIXME IMDomain.heap_size
						helper.append("\t\t\t\tkeso_memcpy((jbyte*)(&"+dest+"_heap), (jbyte*)(&"+src+"_heap), "+src+"_DDesc.hd.heap_end-"+src+"_heap);\n");
						helper.append("\t\t\t\tkeso_memcpy((jbyte*)(&"+dest+"_DDesc.static_ref), (jbyte*)(&"+src+"_DDesc.static_ref), sizeof(object_t*) * KESO_NUM_STATIC_REFS);\n");
						helper.append("\t\t\t\tfor(int k=0; k<KESO_NUM_STATIC_REFS;k++) {\n");
						helper.append("\t\t\t\t\tkeso_deep_copy(&"+dest+"_DDesc.static_ref[k], &"+src+"_DDesc.static_ref[k]);\n");
						if(typeDesc.isArray() == true)
							helper.append("\t\t\t\t}\n\t\t\t\tbreak;\n\t\t\t}\n\t\t}\n");
						else
							helper.append("\t\t\t\t}\n\t\t\t}\n\t\t}\n");
					}
					list_acc.add(z,worked_list.remove(0));
					/*if(typeDesc.isArray() == true) {
						// finish array loop
						helper.append("\t}\n");
						// finish else if block
						helper.append("\t}\n");
					}*/
				}
				list_acc.add(l,worked_list.get(0));
			}
			if(typeDesc.isArray() == true) {
				// finish array loop
				helper.append("\t}\n");
				helper.append("\treturn result;\n");
			}
			// finish function
			helper.append("}\n\n");
			was_in=1;
			}
		mbodyOuter.append(helper);
		}


		/**
	 * Generates a helper method that is called from the proxy method of
	 * local portal calls and basically just invokes the service method.
	 *
	 * The purpose of this function is to provide an explicit join point.
	 * The helper method is passed the Caller and Callee domain in addition
	 * to the actual parameters of the service method.
	 */
	public void add_PortalHelperMethod(IMMethod proxy, ServiceDefinition srv)
		throws CompileException {

		IMMethod service = srv.getServiceClass(repository).getMethod(proxy.getMethodNameAndType());
		boolean has_llref = (service.canBlock() && opts.hasLinkedListOfLocalReferences());
		boolean has_return_value = (proxy.getBasicReturnType()!=BCBasicDatatype.VOID);
		IMClass task_clazz = proxy.getIMClass("java/lang/Thread");

		StringBuffer helper = new StringBuffer();
		helper.append("static KESO_INLINE ");
		helper.append(proxy.getReturnType());
		helper.append(" ");
		helper.append(proxy.getIdentifier());
		helper.append("_helper(domain_t *CALLEE");

		helper.append(extraProxyHelperArgString());
		helper.append(", ");
		helper.append(proxy.getArgString());
		helper.append(") {\n");

		final String[] args = proxy.getMethodFrame().getArgNames();
		final int[] types = proxy.getMethodFrame().getArgTypes();

		if(has_llref) {
			int refs = 0;

			for (int i=0;i<types.length;i++) {
				if (types[i]==BCBasicDatatype.REFERENCE)
					refs++;
			}

			if(refs > 1) { // has references other than "this"
				helper.append("\tobject_t *obj[");
				helper.append(refs);
				helper.append("];\n");
				helper.append("\t*pref = (object_t*) ((int)obj | 1);\n"); // FIXME other marker variants
				helper.append("\tpref = &obj[");
				helper.append(refs-1);
				helper.append("];\n");
			}
		}

		add_portal_enter(helper,task_clazz, srv.getDomain());

		// copy reference parameters to heap of service domain
		for (int i=0,j=0;i<args.length;i++) {
			if (args[i].equals("obj0")) continue;
			if (types[i]==BCBasicDatatype.REFERENCE) {
				if (has_llref) {
					helper.append("\tobj[");
					helper.append(j++);
					helper.append("]");
				} else {
					helper.append("\t");
					helper.append(args[i]);
				}
				helper.append(" = ");

				if (proxy.getMethodFrame().getArgumentClass(has_llref?(i-1):i).equals("keso/core/Memory")) {
					helper.append("keso_dup_mem(");
					helper.append(args[i]);
					helper.append(")");
				} else {
					helper.append("keso_dup_object(");
					helper.append(args[i]);
					helper.append(")");
					repository.needObjDUP();
				}
				helper.append(";\n");
			}
		}

		helper.append("\t");
		if (has_return_value) {
			helper.append(proxy.getReturnType());
			helper.append(" ret_value = ");
		}
		helper.append(service.getIdentifier());
		helper.append('(');

		for (int i=0,j=0;i<args.length;i++) {
			if (i>0) helper.append(',');
			if (types[i]==BCBasicDatatype.REFERENCE && !args[i].equals("obj0") && has_llref) {
					helper.append("(object_t*)");
					helper.append("obj[");
					helper.append(j++);
					helper.append("]");
			} else {
				helper.append(args[i]);
			}
		}
		helper.append(");\n\n");

		add_portal_leave(helper,task_clazz);

		if(has_return_value) {
			if (proxy.getBasicReturnType()==BCBasicDatatype.REFERENCE) {
				helper.append("\treturn keso_dup_object(ret_value);\n");
				repository.needObjDUP();
			} else {
				helper.append("\treturn ret_value;\n");
			}
		}

		helper.append("}\n\n");
		mbodyOuter.append(helper);
	}

	public void beginMethod(IMMethod method) throws CompileException {
		currentMethod = method;
		code_alias = method.getAlias();
		mbody = new StringBuffer();
		mbodyOuter = new StringBuffer();
		line  = new StringBuffer();
		includes = new TreeSet<String>();

		indent = "";

		StringBuffer prototype = new StringBuffer();

		prototype.append(method.getReturnType());
		prototype.append(" ");
		prototype.append(method.getIdentifier());
		prototype.append("(");
		prototype.append(method.getArgString());
		prototype.append(")");

		if (opts.hasDbgSymbols()) {
			local_add("/* ");
			local_add(method.getClassName());
			local_add(".");
			local_add(method.getMethodNameAndType());
			if (method.isPure())
				local_add(" PURE");
			if (method.isConstant())
				local_add(" CONST");
			local_add(" */\n");
		}

		addln(prototype);
		addln('{');

		indent = "\t";

		if (opts.hasDbgSymbols())
		{
			header_add("/* ");
			header_add(method.getMethodNameAndType());
			header_addln(" */");
		}
		header_add(prototype);
		if (opts.useGCCExtensions()) {
			if (method.isConstant()) {
				header_add(" ATTR_CONST ");
			} else if (method.isPure()) {
				header_add(" ATTR_PURE ");
			}
		}
		header_addln(";");

	}

	public void add_rule(MakeRule rule) throws CompileException {
		rules.add(rule);
	}

	public void add_class(IMModule module) throws CompileException {
		add_class(module.getAlias());
	}
	public void add_class(String classname) throws CompileException {
		includes.add(classname);
	}

	public void global_add_class(IMModule module) throws CompileException {
		global_add_class(module.getAlias());
	}
	public void global_add_class(String classname) throws CompileException {
		global_includes.add(classname);
	}

	public void global_header_add_class(IMModule module) throws CompileException {
		global_header_add_class(module.getAlias());
	}
	public void global_header_add_class(String classname) throws CompileException {
		if (global_header_includes.add(classname)) {
			global_header_add("#include \"");
			global_header_add(classname);
			global_header_add(".h\"\n");
		}
	}

	/**
	 * Create C-code for a static memory allocation.
	 *
	 * coder.add("object_t* memobj = ");
	 * coder.add_allocStaticMem(1024);
	 * coder.addln(";");
	 */
	private int identifierCounter=0;
	public void add_allocStaticMem(int size) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		add_class(memory_class);
		local_add("static jbyte staticMemArea");
		local_add(identifierCounter);
		local_add("[");
		local_add(size);
		local_add("];\n");

		add_allocConstObject(memory_class, "staticMem"+identifierCounter,null);
		add_init_field(memory_class, "addr", "("+opts.targetAddrType()+")staticMemArea"+identifierCounter);
		add_init_field(memory_class, "size", java.lang.Integer.toString(size));
		add_init_end();

		identifierCounter++;
	}

	public String add_initArray(IMClass clazz, int arrsize, boolean staticInitialization) throws CompileException {
		StringBuffer initStr = new StringBuffer(add_initObject(clazz,staticInitialization));

		initStr.append(arrsize);
		initStr.append(" /* size */, ");
		initStr.append(" /* data */ ");
		return initStr.toString();
	}

	public String add_initObject(IMClass clazz, boolean staticInitialization) throws CompileException {
		StringBuffer initStr = new StringBuffer(30);

		// null all references within the object
		int numrefs = clazz.getObjRefFieldCount();
		while(numrefs --> 0)
			initStr.append("NULL, ");

		initStr.append(add_initObjectHeader(clazz, staticInitialization));

		return initStr.toString();
	}

	public String add_initObjectHeader(IMClass clazz, boolean staticInitialization) throws CompileException {
		StringBuffer initStr = new StringBuffer(30);

		if(opts.getGlobalHeap().needGCInfo()) {
			initStr.append(staticInitialization ? "1" : "KESO_CURRENT_DOMAIN->colorbit");
			initStr.append(" /* gcinfo */,\n");
		}

		initStr.append(clazz.getClassID());
		initStr.append(" /* class_id */,\n");
		return initStr.toString();
	}

	/**
	 * Create C-code for constant object. The Object must be real constant
	 * not only the object reference. The memory is allocated statically at
	 * compile time.
	 *
	 * After calling add_allocConstObject the init process of the object must
	 * be finished. See example:
	 *
	 * IMClass clazz = method.getIMClass("java/lang/String");
	 * coder.add_class(clazz.getAlias());
	 *
	 * coder.add("new_string =");
	 *
	 * coder.add_allocConstObject(clazz, "my_empty_string");
	 * coder.add_init_field("value","(object_t*)0");
	 * coder.add_init_end();
	 *
	 * coder.addln(";");
	 */
	public void add_allocConstObject(IMClass clazz, String static_id, String refs) throws CompileException {
		add_allocConstObject(true, clazz, static_id, refs);
	}

	public void add_allocConstObject(boolean global, IMClass clazz, String static_id, String refs) throws CompileException {
		IMMethod constructor = clazz.getMethod("<init>()V");
		if(constructor!=null && constructor.getCosts()>0)
			opts.warn("constructor for const object " + static_id + " of class " + clazz + " will *NOT* be invoked!");

		if (global) {
			header_add("#include \"");
			header_add(clazz.getAlias());
			header_add(".h\"\n");
			header_add("extern ");
			header_add(clazz.getClassTypeString());
			header_add(" ");
			header_add(static_id);
			header_add(";\n");
		} else {
			local_add("static ");
		}
		local_add(clazz.getClassTypeString());
		local_add(" ALIGN4 ");
		local_add(static_id);
		local_add("={\n");

		// reference values
		if(clazz.getObjRefFieldCount()>0 && refs==null) {
			local_add(add_initObject(clazz,true));
		} else {
			local_add(refs);
			local_add(add_initObjectHeader(clazz,true));
		}

		add(add_accessStaticObject(clazz, static_id));

		// non reference fields need to be finished by the caller
	}

	public void add_init_field(IMClass clazz, String fieldName, String value) throws CompileException {
		local_add(value);
		local_add("/* ");
		local_add(fieldName);
		local_add(" */\n");
	}

	public void add_init_field(IMClass clazz, String fieldName, int value) throws CompileException {
		local_add(", ");
		local_add(value);
		local_add("\n");
	}

	public void add_init_field_hex(IMClass clazz, String fieldName, String typecast, int value) throws CompileException {
		local_add(", ");
		local_add(typecast);
		local_add("0x");
		local_add(Integer.toHexString(value));
		local_add("\n");
	}

	public void add_init_end() throws CompileException {
		local_add("};\n");
	}

	public void add_allocGlobalConstObject(IMClass clazz, String alias, IMNode callback) throws CompileException {
		add_class(clazz);
		repository.registerGlobalTranslationCallback(alias, callback);
		add(add_accessStaticObject(clazz, alias));
	}

	/**
	 * Create C-code for constant object. The Object must be real constant
	 * not only the object reference. The memory is allocated statically at
	 * compile time.
	 */
	public void global_add_allocConstObject(IMClass clazz, String static_id) throws CompileException {
		global_add_allocConstObject(clazz,static_id, null);
	}

	public void global_add_allocConstObject(IMClass clazz, String static_id, String refInit) throws CompileException {
		global_header_add_class(clazz);
		global_header_add("extern ");
		global_header_add(clazz.getClassTypeString());
		global_header_add(" ALIGN4 ");
		global_header_add(static_id);
		global_header_add(";\n");

		global_add(clazz.getClassTypeString());
		global_add(" ALIGN4 ");
		global_add(static_id);
		global_add("={\n");
		if(refInit != null) {
			global_add(refInit);
			global_add(add_initObjectHeader(clazz,true));
		} else {
			global_add(add_initObject(clazz,true));
		}
	}

	public void global_add_init_field_hex(String prefix, String field, String typecast, int value) throws CompileException {
		global_add(typecast);
		global_add("0x");
		global_add(Integer.toHexString(value));
		global_add(" /* ");
		global_add(prefix);
		global_add(field);
		global_add(" */,\n");
	}

	public void global_add_init_field(String prefix, String field, String value) throws CompileException {
		global_add(value);
		global_add(" /* ");
		global_add(prefix);
		global_add(field);
		global_add(" */,\n");
	}

	public void global_add_init_end() throws CompileException {
		global_add("};\n");
	}

	public void endOfLocals() throws CompileException {
		if (mbodyFrame==null) {
			mbodyFrame = mbody;
			mbody = new StringBuffer();
		}
	}

	private int stack_obj_counter = 0;

	public void add_allocStackObjectInlined(IMClass clazz) throws CompileException {
		mbodyFrame.append(indent);
		mbodyFrame.append(clazz.getClassTypeString());
		mbodyFrame.append(" stack_obj");
		mbodyFrame.append(stack_obj_counter);
		mbodyFrame.append(" = { ");
		mbodyFrame.append(add_initObject(clazz,false));
		for(int i=0; i<clazz.getPrimitiveFieldCount(); i++)
			mbodyFrame.append("0,");
		mbodyFrame.append(" };\n");

		add("((object_t*)&stack_obj");
		add(stack_obj_counter++);
		add(")");
	}

	public void add_allocStackObject(IMClass clazz) throws CompileException {
		mbodyFrame.append(indent);
		mbodyFrame.append(clazz.getClassTypeString());
		mbodyFrame.append(" stack_obj");
		mbodyFrame.append(stack_obj_counter);
		mbodyFrame.append(";\n");
		add("keso_alloc_stack((object_t*)&stack_obj");
		add(stack_obj_counter++);
		add(",");
		add(clazz.getClassID());
		if (!opts.hasOption("no_inline_alloc")) {
			add(',');
			add(clazz.emitObjectSize());
			add(',');
			add(clazz.getObjRefFieldCount());
		}
		add(")");
	}

	/**
	 * Create C-code to alloc a static object. The object must be initialized
	 * by calling the constructor!
	 */
	public void add_allocStaticObject(IMClass clazz, String static_id) throws CompileException {
		local_add("static ");
		local_add(clazz.getClassTypeString());
		local_add(" ALIGN4 ");
		local_add(static_id);
		local_add("={\n");
		local_add(add_initObject(clazz,true));
		local_add("};\n");
		add(add_accessStaticObject(clazz, static_id));
	}

	/**
	 * Generates a codestring that resembles a reference to a statically allocated object.
	 * This method takes care that the offset to the OBJECT_HEADER is properly
	 * added to the address if applicable.
	 * This method should always be used to access statically allocated objects,
	 * including system resources like Tasks, Resources and Alarms.
	 *
	 * @param  clazz   The IMClass the accessed object belongs to. This has alway
	 *				 to be the most special class of the object, ie may not be any
	 *				 of its superclasses, so that the offset may be calculated
	 *				 correctly. For Task objects, e.g., it has always to be the
	 *				 Task's mainClass and not the generic Task class.
	 * @param  objName The name of the accessed object in the C-Code.
	 * @return String that can be used to access the object.
	 */
	public String add_accessStaticObject(IMClass clazz, String objName) throws CompileException {
		StringBuffer result = new StringBuffer();
		result.append("((object_t*) ");
		if(clazz.getObjRefFieldCount()>0)
			result.append("((object_t**) ");
		result.append("&");
		result.append(objName);
		if(clazz.getObjRefFieldCount()>0) {
			result.append("+");
			result.append(clazz.getObjRefFieldCount());
			result.append(")");
		}
		result.append(")");
		return result.toString();
	}

	public void add_allocObject(IMClass clazz, StringBuffer buf) throws CompileException {
		if (!opts.hasOption("no_inline_alloc")) {
			buf.append("KESO_ALLOC(");
			buf.append(clazz.getClassID());
			buf.append(',');
			buf.append(clazz.emitObjectSize());
			buf.append(',');
			buf.append(clazz.getObjRefFieldCount());
			buf.append(')');
		} else {
			buf.append("keso_allocObject(");
			buf.append(clazz.getClassID());
			buf.append(')');
		}
	}
	/**
	 * Create C-code to alloc a object. It must be initialized
	 * by calling the constructor!
	 */
	public void add_allocObject(IMClass clazz) throws CompileException {
		if (!opts.hasOption("no_inline_alloc")) {
			add("KESO_ALLOC(");
			add(clazz.getClassID());
			add(',');
			add(clazz.emitObjectSize());
			add(',');
			add(clazz.getObjRefFieldCount());
			add(')');
		} else {
			add("keso_allocObject(");
			add(clazz.getClassID());
			add(')');
		}
	}

	/**
	 * Create C Code to access an object field.
	 *
	 * This is unsafe! The object reference is not checked. Use chk_ref()
	 * to do so.
	 */
	public String getField(IMClass clazz, String obj, String fieldName) throws CompileException {

		StringBuffer field = new StringBuffer("(");
		field.append(clazz.getAlias().toUpperCase());
		// parity check ext
		if (opts.hasOption("use_coded_refs")) { 
			field.append("_OBJ(NATIVE_ADDR(");
		}
		else { 
			field.append("_OBJ(");
		}
		field.append(obj);
		// parity check extension
		if (opts.hasOption("use_coded_refs")) {
			field.append("))->");
		}
		else {
			field.append(")->");
		}
		field.append(clazz.emitField(fieldName));
		field.append(')');

		return field.toString();
	}

	public void add_getField(IMClass imClass, IMNode obj, String fieldName)
			throws CompileException {
		add_getField(imClass, obj, fieldName, true);
	}

	/**
	 * Add C Code to access an object field.
	 *
	 * This is unsafe! The object reference is not checked. Use chk_ref()
	 * to do so.
	 */
	public void add_getField(IMClass clazz, IMNode obj, String fieldName,
			boolean needsCheck) throws CompileException {
		if (needsCheck && checked_obj != obj) {
			add_before("\t/* Warn: The next reference access is potentially UNSAFE! */\n");
			opts.vwarn("Potentially unsafe reference access!\n");
		}
		add("(");
		add(clazz.getAlias().toUpperCase());
		// parity check extension
		if(opts.hasOption("use_coded_refs")) {
			add("_OBJ(NATIVE_ADDR(");
		}
		else { 
			add("_OBJ(");
		}
		obj.translate(this);
		// parity check extension
		if(opts.hasOption("use_coded_refs")) {
			add("))->");
		}
		else {
			add(")->");
		}
		add(clazz.emitField(fieldName));
		add(')');
	}

	final public void add_putField(IMClass clazz, IMNode obj, String fieldName, int datatype, IMNode value) throws CompileException {

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add("KESO_WR_GUARD(");

		if (opts.getGlobalHeap().needWriteBarriers() && !omitWriteBarrier(datatype)) {
			add("KESO_WRBR(");
			add_getField(clazz, obj, fieldName);
			add(",(");
			add(BuilderOptions.typeToString(datatype));
			add(")");
			value.translate(this);
			add(")");
		} else {
			add_getField(clazz, obj, fieldName);
			add(" = ");
			//if (datatype==BCBasicDatatype.REFERENCE) add("(object_t*)");
			add("(");
			add(BuilderOptions.typeToString(datatype));
			add(")");
			value.translate(this);
		}

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add(")");
	}

	final public void add_putField_NoWrBr(IMClass clazz, IMNode obj, String fieldName, int datatype, IMNode value) throws CompileException {

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add("KESO_WR_GUARD(");

		add_getField(clazz, obj, fieldName);
		add(" = ");
		//if (datatype==BCBasicDatatype.REFERENCE) add("(object_t*)");
		add("(");
		add(BuilderOptions.typeToString(datatype));
		add(")");
		value.translate(this);

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add(")");
	}

	public void add_portal_enter(StringBuffer tgt, IMClass task_clazz, DomainDefinition<?> target_domain) throws CompileException {
		if (opts.isSingleDomainSystem()) return;

		/* switch domain */
		tgt.append("\n\tKESO_CURRENT_DOMAIN = KESO_CURRENT_TASK->");
		tgt.append(task_clazz.emitField("e_domain"));
		tgt.append(" = CALLEE;\n\n");
	}

	public void add_portal_leave(StringBuffer tgt, IMClass task_clazz) throws CompileException {
		if (opts.isSingleDomainSystem()) return;

		/* switch domain */
		tgt.append("\tKESO_CURRENT_DOMAIN = KESO_CURRENT_TASK->");
		tgt.append(task_clazz.emitField("e_domain"));
		tgt.append(" = CALLER;\n\n");
	}

	protected String getNodeDDescPtr(IMNode node) throws CompileException {
		if (currentBB != null) {
			Set<IMDomain> domains = currentBB.getDomains();
			if (domains.size() == 1)
				return getDDescPtr(domains.iterator().next().getDef());
		}
		return "KESO_CURRENT_DOMAIN";
	}

	public void add_getStatic(IMGetStatic node, IMField field, int datatype) throws CompileException {
		field.emit_macro(this, getNodeDDescPtr(node));
	}

	public void add_putStatic(IMPutStatic node, IMField field, int datatype, IMNode value) throws CompileException {
		String ddptr = getNodeDDescPtr(node);

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add("KESO_WR_GUARD(");

		if (opts.getGlobalHeap().needWriteBarriers() && !omitWriteBarrier(datatype)) {
			add("KESO_WRBR(");
			field.emit_macro(this,ddptr);
			add(",(");
			add(BuilderOptions.typeToString(datatype));
			add(")");
			value.translate(this);
			add(")");
		} else {
		/*	if(opts.hasOption("use_coded_refs")) {
				field.emit_macro(this,ddptr);
				add(" = ");
				if(node.field_emitted() == 0) {
					add("(");
					add(BuilderOptions.typeToString(datatype));
					add(")");
					value.translate(this);
				}
				// term has already been evaluated by IMPutStatic 
				else {
					add("obj_f;\n");
				}
*/
			// normal refs
	//		} else {
				field.emit_macro(this,ddptr);
				add(" = ");
				add("(");
				add(BuilderOptions.typeToString(datatype));
				add(")");
				value.translate(this);
		//	}
		}

		if (opts.needGuardedMemoryAccess() && datatype==BCBasicDatatype.REFERENCE)
			add(")");
	}

	/**
	 * Add C Code to access an object field.
	 */
	public void add_getField(IMClass clazz, IMMethod method, String obj, String fieldName,String reason) throws CompileException {
		method.nullCheckEmitted++;
		//TODO falsch mitgezaehlt, sobald gradmp funktioniert (chk_ref omitted evtl)
		chk_ref(obj, reason, clazz.getField( fieldName ));
		add(getField(clazz, obj, fieldName));
	}

	/**
	 * Add C Code to access an object field.
	 *
	 * This is unsafe! The object reference _must_ always be valid.
	 */
	public void add_getField_fast(IMClass clazz, String obj, String fieldName,String reason) throws CompileException {
		if (opts.hasDebugCore()) chk_ref(obj, reason);
		add(getField(clazz, obj, fieldName));
	}

	public void add(char str) throws CompileException {
		line.append(str);
	}

	/**
	 * Add a whole function to the Coder.
	 * The contents of the specified file are appended as-is
	 * to output.
	 *
	 * @param filename File to read the code from.
	 */
	protected void add_fkt(String filename, StringBuffer hdr, StringBuffer src) throws CompileException {
		StringBuffer target = null;
		String srcline = "";
		boolean withInCopyText = false;

		try {
			java.io.BufferedReader fktfile = new java.io.BufferedReader(new java.io.FileReader(opts.getCorePath()+"/c-templates/"+filename));

			while( (srcline = fktfile.readLine()) != null) {
				if (srcline.equals("/**(c)")) {
					withInCopyText = true;
				} else if (srcline.equals("  (c)**/")) {
					withInCopyText = false;
				} else if(withInCopyText) {
					continue;
				} else if (srcline.equals("/*KESO--CFILE--KESO*/")) {
					target=src;
				} else if (srcline.equals("/*KESO--HEADER--KESO*/")) {
					target=hdr;
				} else {
					target.append(srcline);
					target.append("\n");
				}
			}
			fktfile.close();
		} catch (NullPointerException e) {
			throw new CompileException("File " + filename + " contains line within unknown section: " + srcline);
		} catch (Exception e) {
			throw new CompileException(e.getMessage());
		}
	}

	/**
	 * Add a whole function to the Coder.
	 * The contents of the specified file are appended as-is
	 * to output.
	 *
	 * @param filename File to read the code from.
	 */
	public void add_fkt(String filename) throws CompileException {
		add_fkt(filename, chline, local);
	}

	/**
	 * Add a whole function to the Coder.
	 * The contents of the specified file are appended as-is
	 * to output.
	 *
	 * @param filename File to read the code from.
	 */
	public void global_add_fkt(String filename) throws CompileException {
		if (already_added_funcs.contains(filename)) return;
		already_added_funcs.add(filename);

		if (opts.hasDbgSymbols()) {
			global_header_add("\n/* BEGIN ");
			global_header_add(filename);
			global_header_add(" */\n");
		}

		add_fkt(filename, global_header, global);

		if (opts.hasDbgSymbols()) {
			global_header_add("\n/* END ");
			global_header_add(filename);
			global_header_add(" */\n");
		}
	}

	public void add(int str) throws CompileException {
		line.append(str);
	}

	public void add(Object str) throws CompileException {
		line.append(str.toString());
	}

	public void add(StringBuffer str) throws CompileException {
		line.append(str);
	}

	public void add_hex(int value) throws CompileException {
		line.append("0x");
		line.append(Integer.toHexString(value));
	}

	public void add(String str) throws CompileException {
		line.append(str);
	}

	public void add_comment(String msg) throws CompileException {
		if (!opts.hasDbgSymbols()) return;
		line.append("/* ");
		line.append(msg);
		line.append(" */");
		addln();
	}

	public void add_asm(String asm) throws CompileException {
		line.append("__asm__ __volatile__ (\"");
		line.append(asm);
		line.append("\")");
	}

	public void addln(char str) throws CompileException {
		line.append(str); addln();
	}

	public void addln(StringBuffer str) throws CompileException {
		line.append(str); addln();
	}

	public void addln(String str) throws CompileException {
		line.append(str); addln();
	}

	public void chk_ref(String obj, String reason) throws CompileException {
		chk_ref( obj, reason, null );
	}

	public void chk_ref(String obj, String reason, IMField field) throws CompileException {
		// check parity of reference
		if(opts.hasOption("use_coded_refs")) {
			mbody.append("\tKESO_CHECK_"+reason+"_PARITY(");
			mbody.append(obj);
			if (opts.useExceptionStrings()) {
				mbody.append(",\"");
				mbody.append(code_alias);
				mbody.append("\",");
			} else {
				mbody.append(",(char*) 0,");
			}
			mbody.append(-1);
			mbody.append(", ");
			mbody.append(getCheckID());
			mbody.append(");\n");
			// parity check end
		}
		if(gradmpc.omitNullcheck(this, reason, field)) {
			//method.nullCheckEmitted--; //TODO count?
			return;
		}

		mbody.append(indent);
		mbody.append("\tKESO_CHECK_"+reason+"_NULLPOINTER(");
		mbody.append(obj);

		if (opts.useExceptionStrings()) {
			mbody.append(",\"");
			mbody.append(code_alias);
			mbody.append("\",");
		} else {
			mbody.append(",(char*) 0,");
		}

		mbody.append(-1);
		mbody.append(", ");
		mbody.append(getCheckID());
		mbody.append(");\n");
	}

	private IMNode checked_obj = null;
	public void chk_ref(IMNode obj, IMMethod method, int bcPosition, String reason) throws CompileException {
		chk_ref(obj, method, bcPosition, reason, null);
	}

	public void chk_ref(IMNode obj, IMMethod method, int bcPosition,
			String reason, IMField field) throws CompileException {

		if (obj == null)
			throw new CompileException("obj == null " + method.getMethodName()
					+ " -> " + method.getAlias());

		checked_obj = obj;

		if (obj instanceof IMNullConstant) {
			opts.warn("static null pointer in "+method.javaLineInfo(bcPosition));
			method.nullCheckAOTFail++;
			mbody.append(indent);
			mbody.append("\t\t\tKESO_CHECK_"+reason+"_NULLPOINTER((object_t*)0, (char*)0, ");
			mbody.append(bcPosition);
			mbody.append(", ");
			mbody.append(getCheckID());
			mbody.append(");\n");
			return;
		}

		if (obj instanceof IMAConstant) {
			opts.vwarn("check reference: const object used omit check");
			method.nullCheckAOTSuccess++;
			return;
		}

		boolean isChecked = codeInfo.isChecked(obj, this);
		if (isChecked) {
			method.nullCheckAOTSuccess++;
		}

		if (opts.hasDebugCore() || !isChecked) {
			if (!(obj instanceof IMAReadLocalVariable || obj.isFinalStatic())) {
				/*
				 * processStack calls the store method before every method
				 * invocation. It is not possible to get something other than
				 * a local variable!
				 */
				throw new CompileException("obj is not a variable! "+method.getMethodName()+" "+obj.toReadableString());
			}
			mbody.append(indent);

			boolean skip_check_params = false;
			if (isChecked) {
				mbody.append("KESO_ASSERT_NULLPOINTER(");
			} else {
				String rtoc = gradmpc.reasonToOmitNullcheck(this,reason,field);
				if( rtoc != null ) { //omit nullcheck
					//TODO count omitted checks?
					skip_check_params = true;
					if( gradmpc.preservesErrorDetection( rtoc ) ) {
						obj.setChecked(this);
					}
				} else {
					method.nullCheckEmitted++;
					
					// parity check
					if(opts.hasOption("use_coded_refs")) {
						mbody.append("\tKESO_CHECK_"+reason+"_PARITY(");
						if( !skip_check_params ) {
							mbody.append(obj.emit());
							mbody.append(",\"");
							mbody.append(method.toString());
							mbody.append("\",");
							mbody.append(bcPosition);
							mbody.append(", ");
							mbody.append(getCheckID());
							mbody.append(");\n");
						}
						// parity check end
					}
					mbody.append("\t\t\tKESO_CHECK_"+reason+"_NULLPOINTER(");
					obj.setChecked(this);
				}
			}

			if( !skip_check_params ) {
				mbody.append(obj.emit());
				mbody.append(",\"");
				mbody.append(method.toString());
				mbody.append("\",");
				mbody.append(bcPosition);
				mbody.append(", ");
				mbody.append(getCheckID());
				mbody.append(");\n");
			}

		} else {
			if (!(obj instanceof IMAReadLocalVariable || obj.isFinalStatic())) {
				opts.warn("obj is not a variable! "+method.javaLineInfo(bcPosition)+" "+obj.toReadableString());
			}
		}
	}

	public void chk_array(IMNode self, IMNode obj, IMNode index,
			IMMethod imethod, int bcPosition, String reason)
	throws CompileException {
		try {
			mbody.append(indent);

			if (obj instanceof IMNullConstant) {
				opts.warn("array reference is null! "+imethod.getMethodName());

				imethod.nullCheckAOTFail++;

				mbody.append("/* Sorry, but this array reference is always null! */");
				mbody.append(indent);
				mbody.append("\t\t\tKESO_CHECK_"+reason+"_NULLPOINTER((object_t*)0,\"");
				mbody.append(imethod.toString());
				mbody.append("\",");
				mbody.append(bcPosition);
				mbody.append(", ");
				mbody.append(getCheckID());
				mbody.append(");\n");
				return;
			}

			if (!(obj instanceof IMAReadLocalVariable))
				throw new CompileException("obj is not a variable! "
						+imethod.getMethodName()+" "+obj.toReadableString()+" "+obj);

			// If the bounds check can be omitted altogether, do it!
			boolean isChecked = codeInfo.isChecked(obj, this);
			if (isChecked && codeInfo.canOmitArrayBoundsCheck(obj, index)) {
				nullCheckSuccess(self, imethod);
				imethod.boundsCheckOmitted++;
				return;
			}


			boolean skip_check_params = false;
			String
				rtoc_null	= gradmpc.reasonToOmitNullcheck(this,reason, null),
				rtoc_bound	= gradmpc.reasonToOmitBoundcheck(this, reason);


			int length = codeInfo.getArrayLength(obj);
			if (length != -1) {
				nullCheckSuccess(self, imethod);
				// array != null <-- kein setChecked && length bekannt
				if( rtoc_bound != null ) { //omit boundcheck (and nullcheck)
					imethod.boundsCheckOmitted++; //count?
					skip_check_params = true;
				} else {
					imethod.boundsCheckKnownIndex++;
					mbody.append("\t\t\tKESO_CHECK_"+reason+"_ARRAY_KNOWNBOUNDS(");
					mbody.append(length);
				}
			} else {
				if (isChecked || rtoc_null != null) { //omit nullcheck
					nullCheckSuccess(self, imethod);
					if( rtoc_bound != null ) { //omit boundcheck
						imethod.boundsCheckOmitted++; //count?
						skip_check_params = true;
					} else {
						imethod.boundsCheckFull++;
						mbody.append("\t\t\tKESO_CHECK_"+reason+"_ARRAY_NOTNULL(");
					}
					if( !isChecked && gradmpc.preservesErrorDetection(rtoc_null) ) {
						obj.setChecked(this);
					}
				} else { //nullcheck required
					if( rtoc_bound != null ) { //omit boundcheck
						//TODO count omitted checks? (nullcheck without boundcheck)
						skip_check_params = true;
						if (self.hasToBeChecked())
							chk_ref(obj, imethod, bcPosition, "ARRAYLENGTH"); //TODO nullcheck?
					} else {
						imethod.nullCheckEmitted++;
						imethod.boundsCheckFull++;
						mbody.append("\t\t\tKESO_CHECK_"+reason+"_ARRAY(");
					}
					obj.setChecked(this);
				}
				if(!skip_check_params) {
					mbody.append(((IMAReadLocalVariable)obj).getIMSlot().toString());
				}
			}

			if( !skip_check_params ) {
				mbody.append(",");
				mbody.append(index.toReadableString());

				if (opts.useExceptionStrings()) {
					mbody.append(",\"");
					mbody.append(imethod.toString());
					mbody.append("\",");
				} else {
					mbody.append(", (char*) 0,");
				}

				mbody.append(bcPosition);
				mbody.append(", ");
				mbody.append(getCheckID());
				mbody.append(");\n");
			}

		} catch (ClassCastException ex) {
			System.err.println("obj = "+obj.toReadableString());
			System.err.println("index = "+index.toReadableString());
			throw ex;
		}
	}

	/**
	 * Checks the offset to a (direct) Memory access via the keso/core/Memory
	 * object.
	 *
	 * @param obj The Memory object to which the access is made
	 * @param index The offset of the access
	 * @param accesssize Size in bytes of the access
	 * @param imethod calling method
	 * @param bcPosition bytecode position
	 */
	public void chk_range(IMNode obj, IMNode index, int accesssize,
			IMMethod imethod, int bcPosition, String type)
	throws CompileException {

		IMClass mem_class = repository.getClass("keso/core/Memory");
		add_class(mem_class);

		if (!(obj instanceof IMAReadLocalVariable) && !obj.isConstant())
			throw new CompileException("obj is not a variable! "
					+imethod.getMethodName()+" "+obj.toReadableString()+" "+obj);

		// null reference check
		if (!obj.isConstant()) {
			chk_ref(obj, imethod, bcPosition, type);
		} else if (obj instanceof IMNullConstant) {
			opts.warn("static null pointer in "+imethod.javaLineInfo(bcPosition));
			imethod.nullCheckAOTFail++;
			mbody.append(indent);
			mbody.append("\t\t\tKESO_CHECK_"+type+"_NULLPOINTER((object_t*)0, (char*)0, ");
			mbody.append(bcPosition);
			mbody.append(", ");
			mbody.append(getCheckID());
			mbody.append(");\n");
			return;
		}

		// If the range check can be omitted altogether, do it!
		if (codeInfo.isChecked(obj, this)
				&& codeInfo.canOmitMemRangeCheck(obj, index, accesssize)) {
			imethod.nullCheckAOTSuccess++;
			imethod.boundsCheckOmitted++;
			return;
		}

		String shmName  = codeInfo.getSHMName(obj);
		String portName = codeInfo.getPortName(obj);

		if (index instanceof IMIConstant) {
			int off = ((IMIConstant)index).getValue();
			if(off<0) throw new CompileException("negative memory offset "
					+imethod.getMethodName()+" "+index.toReadableString()+" "+index);

			if (obj.isConstant()) {
				IMConstant const_node = obj.nodeToConstant();
				opts.vverbose("perform range check "+const_node.toReadableString()+" ["+off+"]");
				if ((accesssize-1+off)>((IMAMemConstant)const_node).getSize())
					throw new CompileException("access out of range");
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckOmitted++;
				return;
			} else if (shmName != null) {
				mbody.append(indent + openSHMCheck(shmName));
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckConstant++;
			} else if (portName != null) {
				mbody.append(indent + openMessageCheck(portName));
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckConstant++;
			} else {
				mbody.append(indent);
				mbody.append("\t\t\tKESO_CHECK_BOUNDS(");
				mbody.append(mem_class.getClassTypeString());
				mbody.append(',');
				mbody.append(((IMAReadLocalVariable)obj).getIMSlot().toString());
				imethod.nullCheckEmitted++;
				imethod.boundsCheckKnownIndex++;
			}
			mbody.append(',');
			mbody.append(accesssize-1+off);
			mbody.append(',');
			if (opts.useExceptionStrings()) {
				mbody.append('"');
				mbody.append(imethod.toString());
				mbody.append("\",");
			} else {
				mbody.append("(char *) 0,");
			}
			mbody.append(bcPosition);
			mbody.append(", ");
			mbody.append(getCheckID());
			mbody.append(");\n");

		} else {

			mbody.append(indent);
			if (obj instanceof IMAMemConstant) {
				mbody.append("\t\t\tKESO_CHECK_ARRAY_KNOWNBOUNDS(");
				mbody.append(((IMAMemConstant)obj).getSize());
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckKnownIndex++;
			} else if (shmName != null) {
				mbody.append(openSHMCheck(shmName));
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckKnownIndex++;
			} else if (portName != null) {
				mbody.append(openMessageCheck(portName));
				imethod.nullCheckAOTSuccess++;
				imethod.boundsCheckKnownIndex++;
			} else {
				mbody.append("\t\t\tKESO_CHECK_BOUNDS(");
				mbody.append(mem_class.getClassTypeString());
				mbody.append(',');
				mbody.append(((IMAReadLocalVariable)obj).getIMSlot().toString());
				imethod.nullCheckEmitted++;
				imethod.boundsCheckFull++;
			}
			mbody.append(',');
			mbody.append(accesssize-1);
			mbody.append("+");
			mbody.append(((IMIReadLocalVariable)index).getIMSlot().toString());
			mbody.append(',');
			if (opts.useExceptionStrings()) {
				mbody.append('"');
				mbody.append(imethod.toString());
				mbody.append("\",");
			} else {
				mbody.append("(char *) 0,");
			}
			mbody.append(bcPosition);
			mbody.append(", ");
			mbody.append(getCheckID());
			mbody.append(");\n");
		}
	}

	public void scall(IMMethod method, String vname,  IMNode[] args) throws CompileException {
		add(method.getAlias());
		add('(');
		add(vname);
		if (args.length>0) {
			add(',');
			add_args(args);
		}
		add(')');
		return;
	}

	public void add_args(IMNode[] args) throws CompileException {
		for (int i=0;i<args.length;i++)  {
			if (i>0) add(',');
			args[i].translate(this);
		}
	}

	public void vcall(IMMethodFrame frame, IMMethod callee, IMNode self, IMNode[] args) throws CompileException {

		if (self.hasSideEffect()) throw new CompileException("this-pointer is not a pure node! "
				+code_alias+" "+self.toReadableString()+" "+self.dumpBC());

		MethodTable table = callee.getMethodTable();
		if (table==null) {
			opts.vwarn("interface method "
					+callee.getClassName()+"."+callee.getMethodName()+" not impl.!");
			add("keso_throw_method_not_implemented(\"");
			add(code_alias);
			add("\",");
			add(self.getBCPosition());
			add(")");
			return;
		}

		if (table.emit_call(this,self)) {
			callee.emitArguments(this, frame, self, args);
		}
	}

	public void invoke(IMMethod method, Object parameters)
		throws CompileException {

		boolean llrefs = opts.hasLinkedListOfLocalReferences() && method.canBlock();

		add("\t");
		add(method.getIdentifier());
		add("(");

		if (llrefs) {
			add("pref");
			if (parameters!=null) add(",");
		}

		if (parameters!=null) add(parameters.toString());
		add(");\n");

		if (llrefs) add("\t*pref=KESO_EOLL;\n");
	}

	/**
	 * Generate code to execute a root controlflow.
	 * A root control flow means that the executed function is the first using a
	 * fresh stack. If the linked list of local references mode is enabled,
	 * exec_task will pass the initial pref to the function.
	 *
	 * @param buffer      StringBuffer that the code will be written to.
	 * @param method      The method that shall be executed.
	 * @param pref        The initial pref Attribute, or null if not applicable.
	 * @param parameters  Any parameters that method might expect, or null if
	 *                    none. parameters.toString() must return a valid list of
	 *                    parameters in C-Syntax.
	 * @return buffer with the code appended.
	 */
	public StringBuffer exec_task(StringBuffer buffer, IMMethod method, String pref, Object parameters)
		throws CompileException {
		buffer.append("\t");
		buffer.append(method.getIdentifier());
		buffer.append("(");

		if (opts.hasLinkedListOfLocalReferences() && method.canBlock()) {
			if(pref==null) {
				throw new CompileException(method.getAlias() + " is not allowed to block (pref set to null)");
			} else {
				buffer.append("(object_t**)&(");
				buffer.append(pref);
				if (parameters!=null) buffer.append("),"); else buffer.append(")");
			}
		}

		if (parameters!=null) buffer.append(parameters.toString());
		buffer.append(");\n");

		if (opts.hasLinkedListOfLocalReferences() && pref!=null) {
			buffer.append("\t");
			buffer.append(pref);
			buffer.append("=NULL;\n");
		}

		return buffer;
	}

	final public void add_before(int value) throws CompileException {
		mbody.append(value);
	}

	final public void add_before(StringBuffer str) throws CompileException {
		mbody.append(str);
	}

	final public void add_before(String str) throws CompileException {
		mbody.append(str);
	}

	final public void addln_before(StringBuffer str) throws CompileException {
		if (str.length()>0 && str.charAt(0)!='#') mbody.append(indent);
		mbody.append(str);
		mbody.append("\n");
	}

	final public void addln_before(String str) throws CompileException {
		if (str.length()>0 && str.charAt(0)!='#') mbody.append(indent);
		mbody.append(str);
		mbody.append("\n");
	}

	final public void addln() throws CompileException {
		if (line.length()>0 && line.charAt(0)!='#') mbody.append(indent);
		mbody.append(line);
		mbody.append("\n");
		line.setLength(0);
	}

	final public void header_add(int str) throws CompileException {
		chline.append(str);
	}

	final public void header_add(StringBuffer str) throws CompileException {
		chline.append(str);
	}

	final public void header_add(String str) throws CompileException {
		chline.append(str);
	}

	final public void header_addln(StringBuffer str) throws CompileException {
		cheader.append(chline);
		cheader.append(str);
		cheader.append("\n");
		chline.setLength(0);
	}

	final public void header_addln(String str) throws CompileException {
		cheader.append(chline);
		cheader.append(str);
		cheader.append("\n");
		chline.setLength(0);
	}

	final public void local_add(int str) throws CompileException {
		local.append(str);
	}

	final public void local_add(String str) throws CompileException {
		local.append(str);
	}

	final public void local_add(StringBuffer str) throws CompileException {
			local.append(str);
	}

	public void endMethod() throws CompileException {
		indent = "";
		addln("}");
		if (mbodyFrame!=null) {
			mbodyFrame.append(mbody);
		} else {
			mbodyFrame = mbody;
		}
		mbodyOuter.append(mbodyFrame);
		writeCodeFile(mbodyOuter);
		currentMethod.storeMethodBody(mbodyFrame, includes);
		mbody = null;
		mbodyFrame = null;
		mbodyOuter = null;
		currentMethod = null;
	}

	public void endClassMethod() throws CompileException {
		indent = "";
		writeCodeFile(mbody);
	}

	public void endHeader() throws CompileException {
		flushCodeFile();
		header_addln("\n#endif");
		writeHeaderFile(cheader);
		cheader=null;
	}
	public void endClass() throws CompileException {
		endHeader();
	}

	public void close() throws CompileException {
		writeGlobalFiles();
	}

	public void writeErrorHandler() throws CompileException {
		if(opts.hasOption("no_write")) {
			global_header_add("#define KESO_WRITE(msg,len) \n");
			global_header_add("#define KESO_PRINTF(...) \n");
		} else {
			emitPrintf();
		}

		// count the runtime invocations of every check
		if(opts.hasOption("count_checks")) {
			global_add("unsigned int keso_checkcounts[KESO_NUM_CHECKS];\n");

			global_header_add("extern unsigned int keso_checkcounts[];\n");
			global_header_add("#define KESO_COUNT_CHECK(CHK_ID) keso_checkcounts[CHK_ID]++\n");
		} else {
			global_header_add("#define KESO_COUNT_CHECK(CHK_ID) \n");
		}

		global_add_fkt("exceptions/null_chk.c");
		global_add_fkt("exceptions/array_chk.c");
		if (opts.hasOption("use_coded_refs")) {
			global_add_fkt("exceptions/parity_chk.c");
		}
		global_add_fkt("exceptions/error_chk.c");
	}

	public void writeInfoFile() throws CompileException {
		try {
			String mainfile = opts.getOutputPath()+"/kesoinfo.c";
			PrintStream out = new PrintStream(new FileOutputStream(mainfile));

			out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");

			out.println("#include <stdio.h>");
			out.println("#include <stdlib.h>");
			out.println("#include <string.h>");
			out.println("#include \"keso_types.h\"\n");

			out.println("#define _GLOBAL_H_ 1");
			out.println("#define __domains_H__ 1\n");
			out.println("#define KESO_CONST\n");
			out.println("#define TaskType int\n");

			StringBuffer buf=new StringBuffer();
			repository.emitClassStore(buf);
			out.print(buf);

			out.print("\nvoid dump_class_store() {\n");
			out.print("\tint i;\n");
			out.print("\tprintf(\"cls:type    :size:ifs:roff:classname\\n\");\n");
			out.print("\tfor (i=0;i<CLASSSTORE_SIZE;i++) {\n");
			out.print("\t\tint cid = i+1;\n");
			out.print("\t\tclass_t cls = CLASS(cid);\n");
			out.print("\t\tprintf(\"%3d:%3d -%3d:%4d:%3d:%4d:\",cid,cid,cls.type_range,cls.size,cls.ifaces,CLS_ROFF(cid));\n");
			out.print("\t\tprintf(\"%s\\n\",class_names[i]);\n");
			out.print("\t}\n");
			out.print("}\n\n");

			out.print("\nvoid dump_gc_stat() {\n");
			if (opts.hasOption("gc_log_i4trace")) {
				/* work todo */
			} else {
				out.print("\tFILE *log;\n");
				out.print("\tif (NULL==(log=fopen(\"gc.log\",\"r\"))) { perror(\"gc.log\"); exit(1); }\n");
				out.print("\tprintf(\"E: time :total:free:slots:\\n\");\n");
				out.print("\twhile (!feof(log)) {\n");
				out.print("\t\tchar event;\n");
				out.print("\t\tunsigned int time, total, free, size, did, tid, pos, cid;\n");
				out.print("\t\tfscanf(log,\"%c:%u:%u:%u:%u:%u:%u:%u:%u\\n\",&event,&time,&total,&free,&size,&did,&tid,&pos,&cid);\n");
				out.print("\t\tprintf(\"%c:%6u:%5u:%4u:%5u:\",event,time,total,free,size);\n");
				out.print("\t\tswitch (event) {\n");
				out.print("\t\tcase 'a':\n");
				out.print("\t\t\tprintf(\"task %u alloc(%s)\\n\",tid,class_names[cid-1]);\n");
				out.print("\t\t\tbreak;\n");
				out.print("\t\tdefault:\n");
				out.print("\t\t\tprintf(\"domain %u\\n\",did);\n");
				out.print("\t\t}\n");
				out.print("\t}\n");
				out.print("\tfclose(log);\n");
			}
			out.print("}\n\n");

			out.print("\nint main(int argc, char* argv[]) {\n");
			out.print("\tif (argc<2) { printf(\"usage: %s [cls|gc]\\n\",argv[0]); exit(1); }\n");
			out.print("\tif (strncmp(argv[1], \"cls\", 3)==0) {\n");
			out.print("\t\tdump_class_store(); exit(0);\n");
			out.print("\t} else if (strncmp(argv[1], \"gc\", 2)==0) {\n");
			out.print("\t\tdump_gc_stat();exit(0);\n");
			out.print("\t}\n");
			out.print("\treturn 1;\n");
			out.print("}\n");

			out.close();

		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	/**
	 * Write the file with the main function. This method depends on the architecture of the target system.
	 */
	public abstract void writeMainFile() throws CompileException;

	/**
	 * Write the global files needed to build the system. This method depends on the architecture of the target system.
	 */
	protected abstract void writeGlobalFiles() throws CompileException;

	protected void writeTypeHeader() throws CompileException {
		try {
			String filename = opts.getOutputPath()+"/keso_types.h";
			PrintStream out = new PrintStream(new FileOutputStream(filename));
			out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
			out.println("#ifndef _KESO_TYPES_H_");
			out.println("#define _KESO_TYPES_H_ 1");
			out.println("#include <stdint.h>");

			// FIXME: Find a better place for this
			out.println("#ifdef __APPLE__");
			out.println("\t#include <mach/mach_time.h>");
			out.println("#endif");

			out.println("typedef char* code_t;");

			if(opts.hasOption("wide_char")) {
				out.println("typedef uint16_t             jchar;");
			} else {
				out.println("typedef uint8_t              jchar;");
			}
			//BuilderOptions uses jbyte for boolean --> jboolean not mapped to stdint.
			out.println("typedef int                    jboolean;");
			out.println("typedef int8_t                 jbyte;");
			out.println("typedef uint8_t                jubyte;");
			out.println("typedef int16_t                jshort;");
			out.println("typedef uint16_t               jushort;");
			out.println("typedef int32_t                jint;");
			out.println("typedef uint32_t               juint;");
			out.println("typedef int64_t                jlong;");
			out.println("typedef uint64_t               julong;");
			out.println("typedef float                  jfloat;");
			out.println("typedef double                 jdouble;");

			out.println("typedef union { unsigned char __c[4]; float __d; } __huge_val_t;");
			out.println("extern __huge_val_t __huge_val;");
			out.println("extern __huge_val_t __nan_val;");

			out.println("#define INFINITY (__huge_val.__d)");
			out.println("#define NAN (__nan_val.__d)");

			out.println("#endif");
			out.close();
		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	protected void writeRuntimeData(PrintStream out) throws CompileException {
		IMClass task_class = repository.getClass("java/lang/Thread");
		int task_refs = task_class.getObjRefFieldCount();

		// add an alias to the task class type name for use by c-template
		global_header_add_class(task_class);

		global_header_add("#define KESO_TASKCLASSTYPE ");
		global_header_add(task_class.getClassTypeString());
		global_header_add("\n");
		global_header_add("#define KESO_TASKCLASSID ");
		global_header_add(task_class.getClassID());
		global_header_add("\n");

		allocUserTaskObjects();

		/* global task pointer, updated by PreTaskHook */

		// get keso_curr_task as a TASKCLASSTYPE pointer pointing to the base address of the object;
		// don't use this as a reference value for Java code
		global_header_add("#define KESO_CURRENT_TASK keso_curr_task\n");

		// the INVALID_TASK value as a TASKCLASSTYPE pointer, i.e. the null value minus the reference offset
		// in Java code, INVALID TASK is represented by the null reference
		global_header_add("#define KESO_INVALID_TASK ((KESO_TASKCLASSTYPE*) ((object_t**)NULL - " +task_refs+ "))\n");

		// sets keso_curr_task from a Thread Java object reference by subtracting the reference offset
		global_header_add("#define KESO_SET_CURRENT_TASK(_x_) keso_curr_task= (KESO_TASKCLASSTYPE*) ((object_t**)(_x_) - " +task_refs+ ")\n");

		// gets keso_curr_task as a Thread Java object reference by adding the reference offset
		global_header_add("#define KESO_GET_CURRENT_TASK() (object_t*) ((object_t**) keso_curr_task + " +task_refs+ ")\n");

		writeCurrTaskDef();

		/* keso_curr_task_fkt() will return a reference to the running
		 * task with the help of OSEK GetTaskID(). It will be used in
		 * the PreTaskHook to update the keso_curr_task variable.
		 */
		writeCurrTaskFkt();
	}

	protected void writeCurrTaskFkt() throws CompileException {
		// should be overridden
		opts.warn("keso.compiler.backend.Coder.writeCurrTaskFkt() not implemented for current target.");
	}

	protected void writeCurrTaskDef() throws CompileException {
		global_header_add("extern KESO_TASKCLASSTYPE *keso_curr_task;\n");
		global_add("KESO_TASKCLASSTYPE *keso_curr_task;\n");
	}

	protected void writeHeaderFile(StringBuffer str) throws CompileException {
		try {
			String filename = opts.getOutputPath()+"/"+module.getAlias()+".h";
			PrintStream out = new PrintStream(new FileOutputStream(filename));
			out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
			out.println(str);
			out.close();
		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	protected void writeIncludes(PrintStream out, SortedSet<String> includes) {
		for (String s : includes) {
			out.print("#include \"");
			out.print(s);
			out.println(".h\"");
		}
	}

	protected String codeFileSuffix() { return ".c"; }

	protected void writeCodeFile(StringBuffer str) throws CompileException {
		try {
			if (!opts.hasOption("single_file")) {
				add_rule(new MakeRule(opts, code_alias, module.getAlias()+".h", module instanceof IMClass));
				String filename = opts.getOutputPath()+"/"+code_alias+codeFileSuffix();
				PrintStream out = new PrintStream(new FileOutputStream(filename));

				printDebugLine(out, 0);

				out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
				out.println("#include \"keso_types.h\"");
				out.println("#include \"global.h\"");
				out.println("#include \"domains.h\"");

				writeIncludes(out, includes);

				out.println("\n");
				out.println(local.toString());
				local.setLength(0);
				out.println(str);
				out.close();
			} else {
				writeSingleCodeFile(str);
			}
		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	final protected void printDebugLine(PrintStream out, int line) throws CompileException, IOException {
		if (opts.hasOption("dbg_line") && currentMethod!=null) {
			int dbg_line = currentMethod.getLineNumber(line);
			if (dbg_line>0) {
				out.print("#line ");
				out.print(dbg_line-1);
				out.print(" \"src/");
				out.print(currentMethod.getIMClass().getSourceFile());
				out.println("\"");
			}
		}
	}


	private PrintStream cout = null;
	final protected void writeSingleCodeFile(StringBuffer str) throws CompileException {
		try {
			if (cout==null) {
				String filename = opts.getOutputPath()+"/";
				if (currentMethod!=null) {
					add_rule(new MakeRule(opts, currentMethod.getIMClass().getAlias(), module.getAlias()+".h", module instanceof IMClass));
					filename += currentMethod.getIMClass().getAlias()+codeFileSuffix();
				} else {
					add_rule(new MakeRule(opts,module.getAlias()+".o",
								module.getAlias()+codeFileSuffix(), module.getAlias()+".h", module instanceof IMClass));
					filename += code_alias+codeFileSuffix();
				}
				cout = new PrintStream(new FileOutputStream(filename));

				printDebugLine(cout, 0);

				cout.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
				cout.println("#include \"keso_types.h\"");
				cout.println("#include \"global.h\"");
				cout.println("#include \"domains.h\"");
			} else {
				printDebugLine(cout, 0);
			}


			writeIncludes(cout, includes);

			cout.println("\n");
			cout.println(local.toString());
			local.setLength(0);
			cout.println(str);

		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	final protected void flushCodeFile() throws CompileException {
		if (cout!=null) {
			cout.close();
			cout=null;
		}
	}

	/**
	 * Return true if no write barriers are needed for the method
	 * (generally code) that is currently translated.
	 */
	public boolean omitWriteBarrier(int datatype) {

		if (datatype!=BCBasicDatatype.REFERENCE) return true;

		if(currentMethod==null) return false;

		// All class initializers are called in the startup hook, GC is never running
		if(currentMethod.isClassInit())
			return true;

		// All Task/Resource/Alarm Constructors are also called in StartupHook
		if(currentMethod.isInit()) {
			if(currentMethod.getIMClass().getClassName().equals("keso/core/Alarm"))
				return true;
			if(currentMethod.getIMClass().getClassName().equals("keso/core/Resource"))
				return true;
			if(currentMethod.getIMClass().isSubClassOf("java/lang/Thread"))
				return true;
		}

		return false;
	}

	/**
	 * Initializes the task index keso_task_index.
	 *
	 * Call this from the StartupHook before running any user code.
	 *
	 * The task index is a constant time way of reverse resolving OSEK task ids
	 * to KESO system objects. As an alternative, the keso_task_reverse_lookup
	 * function may be used for saving RAM or for OSEK systems that do not
	 * sequentially enumerate task identifiers.
	 */
	protected void createTaskIndex(PrintStream out) throws CompileException {
		List<? extends TaskDefinition> tasks = opts.getSysAllTasks();

		IMClass resource_class
		= repository.getClass(tasks.get(0).getResourceClass());

		// create task_index
		global_header_add("extern ");
		global_header_add(resource_class.getClassTypeString());
		global_header_add("* keso_task_index[KESO_MAX_TASK];\n");

		global_add(resource_class.getClassTypeString());
		global_add("* keso_task_index[KESO_MAX_TASK];\n");

		// init task_index
		out.println("\t/* Initialize keso_task_index */");
		for(TaskDefinition td: tasks) {
			out.print("\tkeso_task_index[");
			out.print(td.getIdentifier());
			out.print("] = " );
			if(td instanceof JavaTask) {
				out.print("&");
				out.print(getResObjSymbol(td));
			} else {
				out.print("KESO_INVALID_TASK");
			}
			out.println(";");
		}
	}

	public String getResObjSymbol(SystemObject<?> res) throws CompileException {
		StringBuffer sym = new StringBuffer(16);

		sym.append(getDomainSymbolPrefix( res.getDomain() ));
		sym.append("_");
		sym.append(res.getIdentifier());
		sym.append("_obj");
		return sym.toString();
	}

	public String getUserResObjSymbol(UserObject res) throws CompileException {
		StringBuffer sym = new StringBuffer(16);
		sym.append(getDomainSymbolPrefix( res.getDomain() ));
		sym.append("_");
		sym.append(res.getIdentifier());
		sym.append("_user_obj");
		return sym.toString();
	}

	public void registerClassInit(IMMethod clinit) {
		if (classConstructors.add(clinit) == true)
			opts.verbose("register class init for " + clinit.getClassName());
	}

	public void writeHeaderClassConstructors(PrintStream out)
	throws CompileException {
		if (classConstructors.isEmpty())
			return;
		for (IMMethod m : classConstructors) {
			out.print("#include \"" + m.getIMClass().getAlias() + ".h\"\n");
		}
	}

	private void writeCallToInit(PrintStream out, IMMethod m, String indent, String alias) throws CompileException {
		if (m.canBlock()) {
			opts.warn("Method " + m + " might block during initialization!");
		}

		if (m.getCosts()==0)
			return;

		out.print(indent);
		out.print(m.getAlias());
		out.print("(");
		out.print( add_accessStaticObject(m.getIMClass(), alias));
		out.println(");");
	}

	private void writeCallToClInit(PrintStream out, IMMethod m, String indent) throws CompileException {
		if (m.canBlock() && !opts.hasOption("_X_all_CanBlock")) {
			opts.warn("Method " + m + " might block during class initialization!");
		}

		if (m.getCosts()==0)
			return;

		if (opts.hasOption("inline_clinit")) {
			StringBuffer body = m.getMethodBodyPretty();
			out.println(body);
		} else {
			if (m.canBlock() && opts.hasLinkedListOfLocalReferences()) {
				out.print("{object_t* unsafe_llref;");
				out.print(indent);
				out.print(m.getAlias());
				out.print("(&unsafe_llref);}\n");
			} else {
				out.print(indent);
				out.print(m.getAlias());
				out.println("();");
			}
		}
	}

	/**
	 * Writes calls to all registered static object constructors to the given PrintStream.
	 *
	 * @param out   destination for the generated code
	 */
	public void writeCallStaticObjectConstructors(PrintStream out) throws CompileException {
		if(static_alloc_object_constructors.size() <= 0)
			return;

		// TODO group by domain so that we do not reset KESO_CURRENT_DOMAIN more often than needed
		out.println("\t/* Call constructors of statically allocated objects */");

		for(StaticAllocObjectConstructor sc: static_alloc_object_constructors) {
			sc.write(out,opts,this);
		}
	}

	public void writeCallClassConstructors(PrintStream out) throws CompileException {

		if (classConstructors.isEmpty())
			return;

		if (opts.hasOption("_X_all_CanBlock") == false) {
			for (IMMethod m : classConstructors) {
				if (m.canBlock()) {
					opts.warn("Method " + m + " might block during class initialization!");
				}
			}
		}

		out.println("\t/* Call class constructors */");
		out.print("\t{");

		// Create mapping (class -> clinit)
		Map<IMClass, IMMethod> clinitMap = new HashMap<IMClass, IMMethod>();
		for (IMMethod clinit : classConstructors) {
			clinitMap.put(clinit.getIMClass(), clinit);
		}

		for (JavaDomain dd : opts.getSysDef().getJavaDomains()) {

			// Get initialization order by topologically sorting the dependency
			// graph
			IMDomain domain = IMDomain.getIMDomain(dd);
			ReachabilityData rData = domain.getReachabilityData();
			DirectedGraph<IMClass> clinitGraph = rData.getClinitGraph();
			List<IMClass> sequence;
			/*try {
				clinitGraph.dump(opts.getOutputPath() + '/' + domain + ".dot");
			} catch (FileNotFoundException e) {
				throw new CompileException(e.getMessage());
			} catch (DotWriterException e) {
				throw new CompileException(e.getMessage());
			}*/
			try {
				sequence = clinitGraph.topSort(new Comparator<IMClass>() {
					@Override
					public int compare(IMClass o1, IMClass o2) {
						return o1.getClassTypeID() - o2.getClassTypeID();
					}
				});
			} catch (GraphException e) {
				String message = "Circular dependency between class "
					+ "initializers!\n";
				Set<IMClass> circle = e.getInfo();
				message += "  Affected classes: " + circle;
				throw new CompileException(message);
			}

			boolean first = true;

			// Emit the calls to the class initializers in the computed order
			for (IMClass imClass : sequence) {

				// Continue if the class has no or no reachable static
				// initializer for this domain
				IMMethod clinit = clinitMap.get(imClass);
				if (clinit == null
						|| clinit.getDomains().contains(domain) == false)
					continue;

				if (first) {
					first = false;
					out.println("\n\t\t/* Domain: " + dd.toString() + " */");
					if (opts.isSingleDomainSystem() == false)
						out.println("\t\tKESO_CURRENT_DOMAIN="
								+ getDDescPtr(dd) + ';');
				}

				writeCallToClInit(out, clinit, "\t\t");
			}
		}

		out.println("\t}");
	}

	/**
	 *	Creates runtime data structures for a class of system objects.
	 *
	 *  In particular, the following tasks will be performed:
	 *    - static allocation of the system objects
	 *		- emission of a vector containing references to all system objects (keso_<resClass>_index)
	 *		- definition of the KESO_MAX_<resClass> macro containing the number of system objects of this class
	 */
	protected void addManagedResource(List<? extends SystemObject<?>> resources, String resClass) throws CompileException {

		// only used if there are user defined resources
		if(resources.size()<=0)
			return;

		IMClass resource_class
		= repository.getClass(resources.get(0).getResourceClass());
		if(resource_class == null) {
			throw new CompileException("Resource class "
					+ resources.get(0).getResourceClass().toString()
					+ " not found");
		}

		global_header_add("\n\n/* "+resClass+" management */\n\n");
		global_add("\n/* "+resClass+" management */\n");

		global_header_add_class(resource_class);

		// create global resource object
		if(opts.hasOption("init_systemobjects_on_startup")) {
			startupHook.append("\t/* Init " + resClass + " system objects */\n");
		}
		for(SystemObject<?> res: resources) {
			// global objects vs. domain local objects
			String res_obj = getResObjSymbol(res);
			global_add_allocConstObject(resource_class, res_obj);

			if (res instanceof TaskDefinition) {
				TaskDefinition td = (TaskDefinition) res;
				global_add_init_field("", resource_class.emitField("prio"), Integer.toString(td.getPriority()));

				// for tasks, init domain and effective domain fields if applicable
				if(!opts.isSingleDomainSystem()) {
					DomainDefinition<?> dd = res.getDomain();
					String ddesc = (res instanceof JavaTask) ? getDDescPtr(dd) : "NULL";

					global_add_init_field("", resource_class.emitField("domain"),ddesc);
					if(opts.hasLocalPortalUse())
						global_add_init_field("", resource_class.emitField("e_domain"), ddesc);
				}

				if(opts.hasLinkedListOfLocalReferences()) {
					global_add_init_field("", resource_class.emitField("stack"), "NULL");
				}

				// add reference to user object
				String user_obj = "NULL";
				if(res instanceof JavaTask) {
					JavaTask jtd = (JavaTask) res;
					IMClass mainClass = repository.getClass(jtd.getUserResourceClass());
					user_obj = add_accessStaticObject(mainClass, getUserResObjSymbol(jtd));
				}
				global_add_init_field("", resource_class.emitField("user_object"), user_obj);
			}

			// init ID field
			if(opts.hasOption("init_systemobjects_on_startup")) {
				startupHook.append("\t");
				startupHook.append(res_obj);
				startupHook.append(".");
				startupHook.append(resource_class.emitField(resClass.toLowerCase()+"_id"));
				startupHook.append(" = ");
				startupHook.append(res.getIdentifier());
				startupHook.append(";\n");
			} else {
				global_add_init_field("", resource_class.emitField(resClass.toLowerCase()+"_id"), res.getIdentifier());
			}

			global_add_init_end();
		}
	}

	protected void allocUserTaskObjects() throws CompileException {
		global_header_add("\n\n/* usercode task objects */\n\n");
		global_add("\n/* usercode task objects */\n");

		Collection<JavaTask> jts = opts.getSysJavaTasks();

		for(JavaTask jt: jts) {
			IMClass tclass = repository.getClass(jt.getMainClassName());

			// global objects vs. domain local objects
			String tobj = getUserResObjSymbol(jt);
			global_add_allocConstObject(tclass, tobj);
			global_add_init_end();
		}
	}

	/**
	 * Call from the StartUpHook creating code.
	 * Adds code to call the constructor of managed Resources.
	 */
	protected void callManagedResourceConstructors(Collection<? extends UserObject> resources, String resClass, PrintStream outstream)
		throws CompileException {
		if (resources.size()==0) return;

		StringBuffer out = new StringBuffer(500);
		boolean ccalled=false;

		DomainDefinition<?> lastdomain=null;

		out.append("\n\t/* Call each "+resClass+"'s constructor */\n");

		for(UserObject res: resources) {
			IMClass resource_class = repository.getClass(res.getUserResourceClass());
			IMMethod constructor = resource_class.getMethod("<init>()V");

			if(constructor==null)
				continue;

			/* Set the current domain id correctly. We do not need to set current
			 * Task because no system services that need the current task must be
			 * invoked from the StartupHook.
			 */
			if (!opts.isSingleDomainSystem() && res.getDomain() != lastdomain ) {
				lastdomain = res.getDomain();
				out.append("\tKESO_CURRENT_DOMAIN = ");
				out.append(getDDescPtr(lastdomain));
				out.append(";\n");
			}

			final String thisRef = add_accessStaticObject(resource_class, getUserResObjSymbol(res));

			if (constructor.canBlock() && opts.hasLinkedListOfLocalReferences()) {
				if(!opts.hasOption("_X_all_CanBlock")) {
					opts.warn("Method " + constructor + " might block during class initialization!");
				}
				out.append("{object_t* unsafe_llref;");
				out.append(indent);
				out.append(constructor.getAlias());
				out.append("(&unsafe_llref, ");
				out.append(thisRef);
				out.append(");}\n");
			} else {
				out.append(indent);
				out.append(constructor.getAlias());
				out.append("(");
				out.append(thisRef);
				out.append(");\n");
			}

			ccalled=true;
		}

		/* No need to reset the current domain id, the PreTaskHook will set
		 * it correctly before running any Task */
		if(ccalled) outstream.println(out.toString());
	}

	public String getDomainSymbolPrefix(DomainDefinition<?> dom) throws CompileException {
		return dom.getIdentifier();
	}

	public String getDDescPtr(DomainDefinition<?> dom) throws CompileException {
		StringBuffer s = new StringBuffer(16);
		s.append("((domain_t *) &");
		s.append(getDDesc(dom));
		s.append(")");
		return s.toString();
	}

	public String getDDesc(DomainDefinition<?> dom) throws CompileException {
		StringBuffer s = new StringBuffer(16);
		s.append(getDomainSymbolPrefix(dom));
		s.append("_DDesc");
		return s.toString();
	}

	protected void writePrintCharFunction(PrintStream out) throws IOException {
		// this default printc function just immediately aborts the output
		out.println("int __keso_printc( jchar c ) {");
		out.println("\treturn 0;");
		out.println("}");
	}

	public void emitPrintf() throws CompileException {
		global_add_fkt("keso_printf.c");
		global_header_add("#define KESO_WRITE(msg,len) keso_write(msg,len)\n");
		global_header_add("#define KESO_PRINTF(...) keso_printf(__VA_ARGS__)\n");
	}

	public void emitSystemTimeNanos() throws CompileException {
		throw new CompileException("system time retrieval is not supported on this target");
	}

	// tricore specific
	public String lock_sysreg() throws CompileException {
		throw new CompileException("target does not support system register locking");
	}
	public String unlock_sysreg() throws CompileException {
		throw new CompileException("target does not support system register locking");
	}

	public String openSHMCheck(String shmName)
	throws CompileException {
		SharedMemoryDefinition shm = opts.getSysDef().getSharedMemory(shmName);
		String shmType = shm.getCType();
		String header = shm.getAttribute("cheaderfile").valueString();
		add_class(header.substring(0, header.length() - 2)); // Add header
		return "\t\tKESO_CHECK_ARRAY_KNOWNBOUNDS(sizeof(" + shmType + ')';
	}

	public String openMessageCheck(String portName)
	throws CompileException {
		ReceiverPortDefinition rpd = opts.getSysDef().getReceiverPort(portName);
		SenderPortDefinition   spd = rpd.getSenderPort();
		String msgType = spd.getMessageType();
		String header = spd.getAttribute("msgcheaderfile").valueString();
		add_class(header.substring(0, header.length() - 2)); // Add header
		return "\t\tKESO_CHECK_ARRAY_KNOWNBOUNDS(sizeof(" + msgType + ')';
	}


	public void addOffsetofAsserts( final IMClass clazz ) throws CompileException {
		//TODO schlecht, Coder hat Wissen ueber Klasse...
		gradmpc.addOffsetofAsserts( clazz, this );
	}
	public void global_define_static_assert() throws CompileException {
		//there is no static assert in c99. using static assert found at:
		//"http://stackoverflow.com/questions/3385515/static-assert-in-c"
		global_header_add("#define KESO_STATIC_ASSERT_BASE0(COND,MSG) typedef char keso_static_assertion_##MSG[(!!(COND))*2-1]\n");
		global_header_add("#define KESO_STATIC_ASSERT_BASE1(X,L,MSG) KESO_STATIC_ASSERT_BASE0(X,at_line_##L##_##MSG)\n");
		global_header_add("#define KESO_STATIC_ASSERT_BASE2(X,L,MSG) KESO_STATIC_ASSERT_BASE1(X,L,MSG)\n");
		global_header_add("#define KESO_STATIC_ASSERT(COND,MSG) KESO_STATIC_ASSERT_BASE2(COND,__LINE__,MSG)\n");
		global_header_add("#include <stddef.h>\n");
		global_header_add("#define KESO_OFFSETOF_ASSERT( S, M, O ) KESO_STATIC_ASSERT( offsetof( S, M ) == (O), wrong_offset_for_##S##_##M )\n");
	}
	private static boolean is_alpha( char c ) {
		return ('A' <= c && c <= 'Z' ) || ('a' <= c && c <= 'z' );
	}
	private static String force_c_identifier( String msg ) {
		if( msg.length() == 0 ) return msg;
		final char[] buf = new char[ msg.length() ];
		char first = msg.charAt(0);
		buf[0] = is_alpha(first) ? first : '_';
		for( int i = 1; i < msg.length(); ++i ) {
			char c = msg.charAt(i);
			buf[i] = (is_alpha(c) || ('0' <= c && c <= '9') ) ? c : '_';
		}
		return new String(buf);
	}
	public void header_add_static_assert(String check, String msg) throws CompileException {
		header_add( "KESO_STATIC_ASSERT(" + check + ", " + force_c_identifier(msg) + ");\n" );
	}
	public void header_add_offsetof_assert( String structn, String member, int offset ) throws CompileException {
		header_add( "KESO_OFFSETOF_ASSERT(" + structn + ", " + member + ", " + offset + ");\n" );
	}


	protected static int check_id = 1;
	public static int getCheckID() {
		return check_id++;
	}

	private static void nullCheckSuccess(IMNode self, IMMethod method) {
		if (self.hasToBeChecked())
			method.nullCheckAOTSuccess++;
	}

	// add external heap definitions to file
	public void add_domainHeapDefinition(List<IMDomain> list) throws CompileException {
		StringBuffer helper = new StringBuffer();
		helper.append("/* define external heap symbols to gain access to domain heaps for state recovery */\n");
		for(IMDomain d : list) {
			helper.append("extern char " + getDomainSymbolPrefix(d.getDef())+ "_heap[" + d.getHeap().getSize() +"];\n");
		}
		helper.append("\n");
		local_add(helper);
	}
}
