/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.backend;

import keso.classfile.datatypes.*;

import keso.compiler.*;
import keso.compiler.config.*;
import keso.compiler.config.task.*;
import keso.compiler.imcode.*;

import java.io.*;

import java.util.*;

public abstract class CoderOSEK extends Coder {
	protected HashMap<String,HashSet<String>> supportedAttributes = new HashMap<String,HashSet<String>>();

	public boolean supportsOILAttribute(String category, String name, StringBuffer oil) {
		HashSet<String> cat = supportedAttributes.get(category);
		if ( cat == null)
			return false;
		return cat.contains(name);
	}

	protected CoderOSEK(BuilderOptions opts, MethodTableFactory table, ClassStore store)
		throws CompileException {
		super(opts,table,store);

		HashSet<String> supAttrs = new HashSet<String>();
		supportedAttributes.put("OS", supAttrs);
		supAttrs.add("ERRORHOOK");
		supAttrs.add("POSTTASKHOOK");
//		supAttrs.add("PRETASKHOOK");  // always generated
		supAttrs.add("SHUTDOWNHOOK");
//		supAttrs.add("STARTUPHOOK");  // always generated
		supAttrs.add("STATUS");
		supAttrs.add("USEGETSERVICEID");
		supAttrs.add("USEPARAMETERACCESS");
		supAttrs.add("USERESSCHEDULER");

		supAttrs = new HashSet<String>();
		supportedAttributes.put("TASK", supAttrs);
		supAttrs.add("PRIORITY");
		supAttrs.add("SCHEDULE");
		supAttrs.add("ACTIVATION");
//		supAttrs.add("AUTOSTART");
//		supAttrs.add("RESOURCE");
//		supAttrs.add("EVENT");
		supAttrs.add("MESSAGE");

		supAttrs = new HashSet<String>();
		supportedAttributes.put("ALARM_AUTOSTART", supAttrs);
		supAttrs.add("ALARMTIME");
		supAttrs.add("CYCLETIME");
		supAttrs.add("APPMODE");

		supAttrs = new HashSet<String>();
		supportedAttributes.put("COUNTER", supAttrs);
		supAttrs.add("MINCYCLE");
		supAttrs.add("TICKSPERBASE");
		supAttrs.add("MAXALLOWEDVALUE");

		supAttrs = new HashSet<String>();
		supportedAttributes.put("ISR", supAttrs);
		supAttrs.add("CATEGORY");
	//	supAttrs.add("RESOURCE");
	//	supAttrs.add("MESSAGE");
	}

	protected void writeOILFile(String oilHeader) throws CompileException, IOException {
		String oilfile_name = opts.getOutputPath()+"/keso_main.oil";

		StringBuffer buf_oil = new StringBuffer(1000);

		if (oilHeader!=null) buf_oil.append(oilHeader);
		opts.getSysDef().toOIL(buf_oil,"",this);
		String oil = buf_oil.toString();

		File oilfile = new File(oilfile_name);

		if (BuilderOptions.smartFileUpdate(oilfile, oil)) {
			System.err.println("writing OIL file '" + oilfile_name + "'");
			PrintStream out = new PrintStream(new FileOutputStream(oilfile));
			out.print(oil);
			out.close();
		}
	}

	/**
	 * The keso_task_reverse_lookup function maps OSEK task identifiers to the
	 * corresponding KESO system objects.
	 *
	 * This mechanism is an alternative to the keso_task_index direct lookup
	 * that trades execution time for RAM. This function must also be used on
	 * OSEK systems that do not sequentially enumerate their task identifiers.
	 *
	 * It is sufficient to reverse resolve JavaTasks - for any other
	 * task, the current task reference is not accessed and can be
	 * set to INVALID_TASK.
	 */
	protected void writeTaskReverseLookupFkt() throws CompileException {
		global_header_add("KESO_TASKCLASSTYPE* keso_task_reverse_lookup(TaskType taskID);\n");

		Collection<? extends TaskDefinition> tasks = opts.getSysJavaTasks();

		global_add("\n");
		global_add("KESO_TASKCLASSTYPE* keso_task_reverse_lookup(TaskType taskID) {\n");

		for(TaskDefinition td: tasks) {
			global_add("\tif(taskID == ");
			global_add(td.getIdentifier());
			global_add(") {\n");
			global_add("\t\treturn &");
			global_add(getResObjSymbol(td));
			global_add(";\n\t}\n");
		}
		global_add("\treturn KESO_INVALID_TASK;\n"); // invalid or native task
		global_add("}\n\n");
	}

	protected void writeCurrTaskFkt() throws CompileException {
		global_header_add("KESO_TASKCLASSTYPE *keso_curr_task_fkt(void);\n");

		global_add("\n\n");
		global_add("/* Returns a pointer to the currently running Task\n");
		global_add(" * keso_task_index is initialised with NULL, which\n");
		global_add(" * represents the INVALID_TASK Java task object.\n");
		global_add(" */\n");
		global_add("KESO_TASKCLASSTYPE *keso_curr_task_fkt(void) {\n");
		global_add("\tTaskType taskID;\n");
		global_add("\t");
		global_add("GetTaskID(&taskID);\n");
		if(opts.hasOption("task_rlookup_fkt")) {
			global_add("\treturn keso_task_reverse_lookup(taskID);");
		} else {
			global_add("\treturn keso_task_index[taskID];");
		}
		global_add("\n}\n\n");
	}

	protected void writeRuntimeData(PrintStream out) throws CompileException {
		// TASK MANAGEMENT
		// we create system objects for all tasks in the system to enable Java
		// applications to use system services on imported native tasks
		addManagedResource(opts.getSysJavaReachableTasks(), "Task");

		// RESOURCE MANAGEMENT
		addManagedResource(opts.getSysJavaReachableResources(), "Resource");

		// ALARM MANAGEMENT
		addManagedResource(opts.getSysJavaReachableAlarms(), "Alarm");

		super.writeRuntimeData(out);
	}

	protected String extraObjectFiles() throws CompileException { return ""; }
	protected String extraSourceFiles() throws CompileException { return ""; }
	protected String extraMakeMacros() throws CompileException { return ""; }
	protected String extraGeneratedFiles() throws CompileException { return ""; }
	protected String extraAllTargets() throws CompileException { return ""; }

	protected void writeMakefile(PrintStream out) throws CompileException {
		out.println("# THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT\n\n");
		out.println(".PHONY: clean all recreate\n");
		out.print("CC="); out.println(opts.getCC());
		out.print("STRIP="); out.println(opts.getStrip());
		out.print("SIZE="); out.println(opts.getObjSize());
		out.print("OBJCPY="); out.println(opts.getObjCopy());
		out.print("OBJDUMP="); out.println(opts.getObjDump());

		if (opts.getLinkerOptions()!=null) {
			out.print("LDSCR="); out.println(opts.getLinkerOptions());
		}

		out.print(opts.getCFlagsVarName() + "= ");
		out.println(opts.getCFlags());
		out.print("LDFLAGS=");
		out.println(opts.getLDFlags());
		out.println(extraMakeMacros());

		out.print("KESOSRC=global" +  codeFileSuffix() );
		out.print(" ");
		String src_files[] = new String[rules.size()];
		for (int i=0;i<rules.size();i++) {
			src_files[i] = rules.get(i).getSource();
			out.print(src_files[i]);
			out.print(" ");
		}
		out.print(extraSourceFiles());

		out.print("\nall: keso_main.elf "+extraAllTargets()+"\n");

		out.print("\nkeso_main.elf keso_main_g.elf: $(KESOSRC) ");
		out.print(extraSourceFiles());
		out.print("\n\t@echo CC+LD  keso_main.elf");
		out.print("\n\t@$(CC) $(" + opts.getCFlagsVarName() + ") -o keso_main.elf $(KESOSRC) $(LDFLAGS) ");
		if (opts.getCoreLibName()!=null) {
			out.print(" -L");
			out.print(opts.getCorePath());
			out.print(" -l");
			out.print(opts.getCoreLibName());
		}
		out.print("\n\t@cp keso_main.elf keso_main_g.elf");
		out.print("\n\t@$(STRIP) keso_main.elf");
		out.print("\n\t@$(OBJCPY) -R .comment keso_main.elf");
		out.print("\n\t@echo $(KESORC) : $(JINOFLAGS)");
		out.println("\n\t@$(SIZE) keso_main.elf");

		out.println("clean: ");
		out.println("\t@rm -rf keso_main.elf keso_main_g.elf *.o *.dot "+extraGeneratedFiles()+" \n");

		out.println("recreate: ");
		out.println("\tcd ../.. ; make\n");

		out.println("\nkesoinfo: kesoinfo.c");
		out.println("\tgcc -O2 -ansi -Wall -o kesoinfo kesoinfo.c");

		out.println("tags:");
		out.println("\t$(CTAGBIN) *.h *" + codeFileSuffix() );

		out.println("\nkeso_main.hex: keso_main.elf");
		out.println("\t$(OBJCPY) -O ihex -R .eeprom keso_main.elf keso_main.hex");
	}

	protected void writeStartupHook(PrintStream out,
			ComplexBoolAttribute hookDef, String[][] hookParams,
			ComplexBoolAttribute eHookDef, String[][] eHookParams
		) throws CompileException {
		out.println(startupHook.toString());

		// Initialize heaps
		IMDomain[] doms = opts.getDomains();
		out.println("\t/* Initialize heaps */");
		for(IMDomain dom: doms) {
			IMHeap dheap = dom.getHeap();
			dheap.initHeap(this, out);
		}
		out.println();

		// We should do all the initialization work before calling
		// user code (if possible). Constructors are user code..
		global_header_add("\n#define KESO_MAX_TASK ");
		global_header_add(opts.getSysAllTasks().size()+1); // +1 for INVALID_XXX NULL reference
		global_header_add("\n");

		if(opts.hasOption("task_rlookup_fkt")) {
			writeTaskReverseLookupFkt();
		} else {
			createTaskIndex(out);
		}

		// Early Startup Hook - native code invoked before running user Java code
		if (eHookDef!=null && eHookDef.setting) {
			if(eHookDef.getAttribute("HookClass") == null) {
				out.print("\n\t/* user-defined early startup hook (non-Java method) */\n");
				addHookCode(out,eHookDef, eHookParams);
			} else {
				throw new CompileException("only a native hook method may be used for the early startup hook");
			}
		}

		// call class constructors
		writeCallClassConstructors(out);

		// call constructors of statically allocated objects
		writeCallStaticObjectConstructors(out);

		// call service object constructors
		Collection<ServiceDefinition> services = opts.getSysDef().getServices();
		callManagedResourceConstructors(services, "Service", out);

		// Call constructors of the resource objects
		// We call the tasks' constructors last since we don't know what they'll be
		// doing We do not call the constructors of Alarms and Resources anymore
		// because they are empty
		callManagedResourceConstructors(opts.getSysJavaTasks(), "Task", out);

		// call user-defined hook here
		if (hookDef!=null && hookDef.setting) {
			out.print("\n\t/* user defined startup hook */\n");
			addHookCode(out,hookDef, hookParams);
		}
	}

	protected void writeHookRoutines(PrintStream out) throws CompileException {
		OSDefinition osdef = opts.getSysDef().getOSDef();

		// allHooks will contain {hooks, hookdescs, hookparams}
		Object[] allHooks = osdef.getAllHooks();
		ComplexBoolAttribute[] hooks = (ComplexBoolAttribute[]) allHooks[0];
		String[] hookDescs = (String[]) allHooks[1];
		String[][][] hookParams = (String[][][]) allHooks[2];

		out.println("#include \"domains.h\"");

		if (opts.hasOption("inline_clinit")) {
			for (IMMethod m : classConstructors) {
				writeIncludes(out, m.getMethodIncludes());
			}
		}

		// Startup Hook specially handled
		out.println("void StartupHook() {");
		writeStartupHook(out, hooks[0], hookParams[0], hooks[1], hookParams[1]);
		out.println("}\n");

		// PreTaskHook will be used to init keso_curr_task and keso_curr_domain
		out.println("void PreTaskHook() {");
		out.println("\t/* Init keso_curr_task and keso_curr_domain */");
		out.println("\tKESO_TASKCLASSTYPE *ct = keso_curr_task_fkt();");
		if (!opts.isSingleDomainSystem()) {
			if(opts.hasLocalPortalUse()) {
				out.println("\tKESO_CURRENT_DOMAIN = (ct!=KESO_INVALID_TASK) ? ct->_e_domain : NULL;");
			} else {
				out.println("\tKESO_CURRENT_DOMAIN = (ct!=KESO_INVALID_TASK) ? ct->_domain : NULL;");
			}
		}
		out.println("\tkeso_curr_task = ct;");

		// call user defined hook
		for(int i=2; i<hooks.length; i++) {
			if(hookDescs[i].compareTo("PreTaskHook")!=0) continue;
			if(hooks[i]==null || !hooks[i].setting) break;
			out.print("\n\t/* user defined pretask hook */\n");
			addHookCode(out,hooks[i], hookParams[i]);
		}
		out.println("}\n");

		/* PostTaskHook:
		 *  for some reason I don't know ProOSEK wants to have a PostTaskHook
		 *  when a PreTaskHook is used. So for god's sake, in case the user doe
		 *  not use one we will add an empty one here. We will, however, not enable
		 *  the OIL attribute
		 * ErrorHook:
		 *  It may happen that currently _no_ task is running when the ErrorHook is
		 *  called. We must ensure that in this case the errorhook runs in domain
		 *  zero and keso_curr_task is set to KESO_INVALID_TASK
		 */
		for(int i=2; i<hooks.length; i++) {
			if(hookDescs[i].compareTo("PostTaskHook")==0) {
				if(hooks[i]==null || !hooks[i].setting)
					out.println("void PostTaskHook() { /* empty PostTaskHook to satisfy ProOSEK */ }\n");
			} else if(hookDescs[i].compareTo("ErrorHook")==0) {
				if(hooks[i]!=null && hooks[i].setting) {
					out.println("void ErrorHook(");
					out.print(hookParams[i][0][0]);
					out.print(" ");
					out.print(hookParams[i][0][1]);
					out.print(") {");
					out.println("\tif(keso_curr_task_fkt() == KESO_INVALID_TASK) {");
					out.println("\t\tKESO_CURRENT_DOMAIN = DOMAIN_ZERO;");
					out.println("\t\tkeso_curr_task = KESO_INVALID_TASK;");
					out.println("\t}");
					out.println("\t/* user defined ErrorHook */");
					addHookCode(out, hooks[i], hookParams[i]);
					out.println("\t/* No need to restore task/domain environment, since it was only modified");
					out.println("\t * in case there was no task running. In this case the PreTaskHook will restore");
					out.println("\t * the environment when the next task is launched.");
					out.println("\t */");
					out.println("}\n");
				}
			}
		}

		// the other hooks are purely user defined (for now ;-) )

		// [0-1] is the StartupHook, skip it
		for (int i=2; i<hooks.length; i++) {
			if(hooks[i]==null || !hooks[i].setting) continue;

			// skip specially handled hooks here
			// be sure to enable the always enabled Hook
			// in the OIL file by modifying OSDefinition.toOIL()
			if(hookDescs[i].compareTo("PreTaskHook")==0) continue;
			if(hookDescs[i].compareTo("ErrorHook")==0) continue;

			// handle non modified hook
			out.print("void " );
			out.print( hookDescs[i] );
			out.print( "(");
			if (hookParams[i]!=null) {
				for (int j=0; j<hookParams[i].length; j++) {
					if(j>0) out.print(", ");
					out.print(hookParams[i][j][0]); // param type
					out.print(" ");
					out.print(hookParams[i][j][1]); // param's name
				}
			}
			out.println(") {");

			/* just a slight mod for the ShutdownHook ;-)
			 * make sure it runs in DomainZero
			 * ShutdownHook must not call getTaskID, keso_curr_task
			 * does therefore not have to be set
			 */
			if(hookDescs[i].compareTo("ShutdownHook")==0)
				out.println("\tKESO_CURRENT_DOMAIN = DOMAIN_ZERO;");

			addHookCode(out, hooks[i], hookParams[i]);
			out.println("}\n");
		}
	}

	protected String ISRMacro(int ISRCategory) {
		return "ISR";
	}

	protected void addISRHeader(PrintStream out, ISRDefinition isr) throws CompileException {
		out.print(ISRMacro(isr.getCategory()));
		out.print("(");
		out.print(isr.ident);
		out.println(") {");
	}

	protected void writeISRs(PrintStream out) throws CompileException {

		boolean inline_isr = opts.hasOption("inline_isr");

		for (ISRDefinition isr : opts.getJavaISRs()) {

			try {
				int cat = isr.getCategory();
				if(cat!=1 && cat!=2) throw new CompileException(isr + " does not have a valid Category attribute defined!");
				IMClass handlerClass = repository.getClass(isr.getAttribute("HandlerClass").valueString());
				IMMethod handlerMethod = handlerClass.getMethod(isr.getAttribute("HandlerMethod").valueString());
				if (inline_isr) writeIncludes(out, handlerMethod.getMethodIncludes());

				out.print("#include \"");
				out.print(handlerMethod.getIMClass().getAlias());
				out.print(".h\"\n");


				addISRHeader(out,isr);

				/* Domain environment
				 * we do not rely on keso_curr_task being set because it might be NULL
				 * instead we backup the domainid to restore it afterwards
				 *
				 * Task environment
				 * if the currently running task is in a domain different to the ISR'
				 * domain, we must ensure keso_curr_task is set to KESO_INVALID_TASK.
				 * If we did not do this, a task reference could move to a different
				 * domain enabling unauthorized use of services like SetEvent/ChainTask
				 */

				if (!opts.isSingleDomainSystem()) {
					String isrDomain = getDDescPtr(isr.parent);

					out.println("\tdomain_t *domain_backup = KESO_CURRENT_DOMAIN;");
					out.println("\tKESO_TASKCLASSTYPE *task_backup = KESO_CURRENT_TASK;");

					out.print("\tKESO_CURRENT_DOMAIN = ");
					out.print(isrDomain);
					out.print(";\n");
					out.println("\tKESO_CURRENT_TASK = NULL;");
				}

				if (inline_isr) {
					out.println(handlerMethod.getMethodBodyPretty());
				} else {
					out.println(exec_task(new StringBuffer(), handlerMethod, null, null));
				}

				if (!opts.isSingleDomainSystem()) {
					out.println("\tKESO_CURRENT_DOMAIN = domain_backup;");
					out.println("\tKESO_CURRENT_TASK = task_backup;");
				}
				out.println("}\n");
			} catch (NullPointerException e) {
				e.printStackTrace();
				throw new CompileException(isr + " does not have all required attributes defined!");
			}
		}
	}

	/**
	 * Return true if no write barriers are needed for the method
	 * (generally code) that is currently translated.
	 */
	public boolean omitWriteBarrier(int datatype) {
		if (datatype!=BCBasicDatatype.REFERENCE) return true;

		if(currentMethod==null) return false;

		// All class initializers are called in the startup hook, GC is never running
		if(currentMethod.isClassInit())
			return true;

		// All Task/Resource/Alarm Constructors are also called in StartupHook
		if(currentMethod.isInit()) {
			if(currentMethod.getIMClass().getClassName().equals("keso/core/Alarm"))
				return true;
			if(currentMethod.getIMClass().getClassName().equals("keso/core/Resource"))
				return true;
			if(currentMethod.getIMClass().isSubClassOf("keso/core/Task"))
				return true;
		}

		return false;
	}

	public void registerClassInit(IMMethod clinit) {
		if (classConstructors.add(clinit) == true)
			opts.verbose("register class init for " + clinit.getClassName());
	}

	public void writeHeaderClassConstructors(PrintStream out)
	throws CompileException {
		for (IMMethod m : classConstructors) {
			out.print("#include \"");
			out.print(m.getIMClass().getAlias());
			out.print(".h\"\n");
		}
	}

	/**
	 * adds a function call to the user defined hook method.
	 */
	protected void addHookCode(PrintStream out, ComplexBoolAttribute hook, String[][] params) throws CompileException {
		Attribute<?> hookClassName = hook.getAttribute("HookClass");
		Attribute<?> hookMethodName = hook.getAttribute("HookMethod");
		Attribute<?> nativeMethodName = hook.getAttribute("NativeHookMethod");
		IMMethod hookMethod = null;

		if ( (hookMethodName==null || hookClassName==null) && nativeMethodName==null )
			throw new CompileException("HookClass/HookMethod not specified for " + hook.ident);

		if(hookClassName != null) {
			IMClass hookClass = repository.getClass(hookClassName.valueString());
			if (hookClass==null)
				throw new CompileException("HookClass "+hookClassName.valueString()+" specified for " + hook.ident + " does not exist");
			hookMethod = hookClass.getMethod(hookMethodName.valueString());
			if (hookMethod==null)
				throw new CompileException("HookMethod "+hookMethodName.valueString()+" specified for " + hook.ident + " does not exist");
			if (!hookMethod.isStatic()) {
				throw new CompileException("HookMethod specified for " + hook.ident + " is not static");
			}
		}

		// add call
		if(hookMethod != null) {
			StringBuffer para = null;
			if(params != null) {
				for (int i=0; i<params.length; i++) {
				/* specifying a hook method with the correct signature
				 * is the responsibility of the application developer.
				 * If he specified a method with the wrong signature,
				 * the generated KESO system will not compile, just a
				 * an OSEK system would not compile if hooks with
				 * signatures that don't fit were specified
				 */
					if (para==null) para = new StringBuffer();
					else para.append(',');
					para.append(params[i][1]); // param name
				}
			}
			out.print(exec_task(new StringBuffer(), hookMethod, null, para));
		}
		if(nativeMethodName != null) {
			out.println("\t"+nativeMethodName.valueString() + "();");
		}
	}

	/**
	 * Add code for all Alarm callback functions.
	 *
	 * We ensure that Alarm callbacks run in the domain they were configured in.
	 * Alarm callbacks may not call getTaskID and we trust the application
	 * developer that he does not do that because it would possibly migrate a Task
	 * reference to a domain that should not be able to access it.
	 */
	protected void writeAlarmCallbacks(PrintStream out,
			Collection<AlarmDefinition> alarmDefs) throws CompileException {
		for (AlarmDefinition adef : alarmDefs) {
			ComplexAttribute<?> action = adef.getAction();

			if(action.value.compareToIgnoreCase("ALARMCALLBACK")!=0) continue;

			out.print("ALARMCALLBACK(");
			out.print(adef.getIdentifier());
			out.println("_callback) {");

			Attribute<?> cbClass = action.getAttribute("AlarmCallbackClass");
			Attribute<?> cbMethod = action.getAttribute("AlarmCallbackMethod");

			if (cbClass==null || cbMethod==null)
				throw new CompileException("Alarm " + adef.getIdentifier() + " needs AlarmCallbackClass and AlarmCallbackMethod defined");

			IMClass cbIMClass = repository.getClass(cbClass.valueString());
			if (cbIMClass == null)
				throw new CompileException("Alarm " + adef.getIdentifier() + "'s callback class is not known");

			IMMethod cbIMMethod = cbIMClass.getMethod(cbMethod.valueString());
			if(cbIMMethod == null || !cbIMMethod.isStatic())
				throw new CompileException("Alarm " + adef.getIdentifier() + "'s callback method is not known or not static");

			/* Alarm callbacks must not use getTaskID, thus we do not need to adjust the current task here
			 * We don't need to change the domain in a single domain system
			 */
			if (!opts.isSingleDomainSystem()) {
				String cbDomain = getDDescPtr( adef.parent );
				out.println("\tdomain_t *domain_backup = KESO_CURRENT_DOMAIN;");
				out.print("\tKESO_CURRENT_DOMAIN = ");
				out.print(cbDomain);
				out.println(";");
			}
			out.print(exec_task(new StringBuffer(), cbIMMethod, null, null));

			// We can probably save this too, PreTaskHook should run because
			// Alarm Callbacks seem to run on ISR Level, and Rescheduling is
			// performed when returning from Interrupt to Task Level, but
			// not 100% sure about this so I choose the save bet for now
			if (!opts.isSingleDomainSystem()) {
				out.println("\tKESO_CURRENT_DOMAIN = domain_backup;");
			}
			out.println("}");
		}
	}

	protected void addTaskHeader(PrintStream out, TaskDefinition task) throws CompileException {
		out.print("TASK (");
		out.print(task.getIdentifier());
		out.print(") {\n");
	}

	protected void addTaskCode(PrintStream out, IMMethod mainMethod, JavaTask task, IMClass mainClass) throws CompileException {
		if(mainMethod.isStatic()) {
			throw new CompileException("Task entry functions must not be static! (Task " + task + ")");
		}

		if(mainMethod.hasReturnValue()) {
			throw new CompileException("Task entry functions should not return a value! (Task " + task + ")");
		}
		if(!mainMethod.getMethodNameAndType().endsWith("()V")) {
			throw new CompileException("Task entry functions must not accept parameters! (Task " + task + ")");
		}

		boolean inline_task = (opts.hasOption("inline_task") && !mainMethod.canBlock());
		if (inline_task) writeIncludes(out, mainMethod.getMethodIncludes());

		addTaskHeader(out,task);

		if(opts.hasLinkedListOfLocalReferences()) {
			out.print("\tkeso_stack_t stack;\n");
			out.print("\tstack.domain = ");
			out.print(getDDescPtr(task.parent));
			out.print(";\n");
			out.print("\tstack.llrefs = (object_t**) KESO_EOLL;\n");
			out.print("\tstack.next   = NULL;\n");

			out.print("\tKESO_CURRENT_TASK->_stack = &stack;\n\n");
		}

		if (opts.getGlobalHeap().deferTasksOnDemand()) {
			out.print("\tKESO_GC_DEFER_TASK();\n");
		}

		if (inline_task) {
			StringBuffer body = mainMethod.getMethodBodyPretty();
			int pos=-1;
			pos = body.indexOf("obj0");
			if (pos!=-1) {
				out.print("\tobject_t* obj0 = ");
				out.print(add_accessStaticObject(mainClass, task.getIdentifier()+"_obj"));
				out.println(";");
			}
			out.println(body);
		} else {
			out.print(exec_task(new StringBuffer(), mainMethod, "stack.llrefs",
						add_accessStaticObject(mainClass, getUserResObjSymbol(task))));
		}

		if(opts.hasLinkedListOfLocalReferences()) {
			out.print("\tKESO_CURRENT_TASK->_stack = NULL;\n");
		}

		// for backup, if application does not do it though it should!
		out.print("\tTerminateTask();\n}\n\n");
	}

	protected void writeOSEKIncludes(PrintStream out) {
		out.print("#include \"os.h\"\n");

		out.print("#ifndef NULL\n");
		out.print("#ifndef __cplusplus\n");
		out.print("#define NULL ((void*)0)\n");
		out.print("#else\n");
		out.print("#define NULL (0)\n");
		out.print("#endif\n");
		out.print("#endif\n");
	}

	protected void writeMainfileIncludes(PrintStream out) {
	}

	public void writeMainFile() throws CompileException {
		try {
			String mainfile = opts.getOutputPath()+"/keso_main" + codeFileSuffix();
			PrintStream out = new PrintStream(new FileOutputStream(mainfile));
			Collection<JavaTask> tasks = opts.getSysJavaTasks();

			out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");

			/* include the headers for each task's class in global.h */
			out.print("#include \"global.h\"\n");
			writeMainfileIncludes(out);

			/*
			for(ServiceDefinition serv: opts.getSysDef().getServices()) {
				IMClass mainClass = repository.getClass(serv.getUserResourceClass());
				String include = "#include \"" + mainClass.getAlias() + ".h\"\n";
				global_header_add(include);
			}
			*/

			writeHeaderClassConstructors(out);

			// includes for hooks
			OSDefinition os = opts.getSysDef().getOSDef();
			if (os!=null) {
				Object[] allHooks = os.getAllHooks();
				ComplexBoolAttribute[] hooks = (ComplexBoolAttribute[]) allHooks[0];
				java.util.Set<String> hookClasses = new HashSet<String>();

				for(int i=0;i<hooks.length;i++) {
					if (hooks[i]==null || !hooks[i].setting) continue;

					if(hooks[i].getAttribute("HookClass") != null) {
						String hc = hooks[i].getAttribute("HookClass").valueString();
						if(hookClasses.add(hc)) {
							// this header must be included
							out.print("#include \"");
							out.print(repository.getClass(hc).getAlias());
							out.print(".h\"\n");
						}
					}
					if(hooks[i].getAttribute("NativeHookMethod")!=null) {
						out.print("void ");
						out.print(hooks[i].getAttribute("NativeHookMethod").valueString());
						out.println("();");
					}
				}
			} else {
				opts.warn("No OSEK Section found!");
			}

			maincHook(out);

			out.print("\n\n");

			// Runtime information such as Task objects, Resource objects, etc.
			writeRuntimeData(out);

			// Hook routines
			writeHookRoutines(out);

			// ISRs
			writeISRs(out);

			// alarm callback functions
			writeAlarmCallbacks(out, opts.getSysJavaReachableAlarms());

			// emit code for all tasks
			final IMClass runnable_if = repository.getClass("java/lang/Runnable");
			for(JavaTask task: tasks) {
				IMClass mainClass = repository.getClass(task.getMainClassName());
				if(!mainClass.implementsInterface(runnable_if)) {
					throw new CompileException("Task main class " + task.getMainClassName() + " does not implement the Runnable interface!");
				}
				IMMethod mainMethod = mainClass.getMethod(task.getMainMethodName());
				addTaskCode(out, mainMethod, task, mainClass);
			}

			add_rule(new MakeRule(opts, "keso_main.o",
						"keso_main"+codeFileSuffix(), module.getAlias()+".h", false));

			out.close();

		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}

		writeInfoFile();
	}

	protected void maincHook(PrintStream out) {
		//override to add some arch specific code in keso_main.c
	}

	protected void writeGlobalFiles() throws CompileException {
		writeGlobalFiles("");
	}

	protected void writeGlobalFiles(String OILprefix) throws CompileException {
		try {
			writeTypeHeader();
			writeOILFile(OILprefix);

			// Makefile
			String makefile = opts.getOutputPath()+"/Makefile";
			PrintStream out = new PrintStream(new FileOutputStream(makefile));
			writeMakefile(out);
			out.close();

			// Write global.h
			String filename = opts.getOutputPath()+"/global.h";
			PrintStream glb_out = new PrintStream(new FileOutputStream(filename));
			glb_out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
			glb_out.println("#ifndef _GLOBAL_H_");
			glb_out.println("#define _GLOBAL_H_ 1\n");

			glb_out.println("#include \"keso_types.h\"");
			glb_out.println("#define keso_throw_method_not_implemented(method, bcpos) (keso_throw_error(\"method \" method \" not implemented\"), NULL)");
			writeOSEKIncludes(glb_out);
			glb_out.println(global_header);
			glb_out.println("#define KESO_NUM_CHECKS " + Coder.check_id);
			glb_out.println("\n#endif");
			glb_out.close();

			// Write global.c
			filename = opts.getOutputPath()+"/global"+codeFileSuffix();
			glb_out = new PrintStream(new FileOutputStream(filename));
			glb_out.println("/* THIS FILE IS AUTO-GENERATED BY KESO! DON'T EDIT */\n");
			glb_out.println("#include \"global.h\"");
			glb_out.println("#include \"domains.h\"");

			writeIncludes(glb_out, global_includes);

			if(!opts.hasOption("no_write")) {
				writePrintCharFunction(glb_out);
			}
			glb_out.println(global);

			glb_out.close();

		} catch (IOException ex) {
			throw new CompileException(ex.toString());
		}
	}

	public boolean emptyObjectsNeedBraces() {
		return false;
	}
}
