/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.backend;

import keso.compiler.*;
import keso.compiler.config.AlarmDefinition;
import keso.compiler.config.ResourceDefinition;
import keso.compiler.config.task.TaskDefinition;
import java.io.*;
import java.util.HashSet;

public abstract class CoderTrampoline extends CoderOSEK {
	private boolean doneOSMod = false;

	public boolean supportsOILAttribute(String category, String name, StringBuffer oil) {
		if(!doneOSMod && category.compareToIgnoreCase("OS")==0) {
			doneOSMod=true;

			oil.append("\t\tAPP_NAME = \"");
			oil.append(opts.getSystemName());
			oil.append(".elf\";\n");
			String trampolinePath = System.getenv("TRAMPOLINE_BASE_PATH");
			if (trampolinePath ==null) {
				opts.warn("TRAMPOLINE_BASE_PATH unknown! Please edit enviroment variable or edit OIL file");
			}
			oil.append("\t\tTRAMPOLINE_BASE_PATH = \"");
			oil.append(trampolinePath);
			oil.append("\";\n");

			oil.append("\t\tCOMPILER = \"");
			oil.append(opts.getCC());
			oil.append("\";\n");

			oil.append("\t\tLINKER = \"");
			oil.append(opts.getCC());
			oil.append("\";\n");
			oil.append("\t\tASSEMBLER = \"");
			oil.append(opts.getCC());
			oil.append("\";\n");

			oil.append("\t\tCFLAGS = \"");
			oil.append(opts.getCFlags());
			oil.append(" -U_FORTIFY_SOURCE"); // causes problems with longjmp on ubuntu systems
			oil.append("\";\n");
			oil.append("\t\tASFLAGS = \"");
			oil.append(opts.getASFlags());
			oil.append(" -U_FORTIFY_SOURCE"); // causes problems with longjmp on ubuntu systems
			oil.append(" ");
			oil.append(opts.getCFlags());
			oil.append("\";\n");
			oil.append("\t\tLDFLAGS = \"");
			oil.append(opts.getLDFlags());
			oil.append("\";\n");

			oil.append("\t\tAPP_SRC = \"global.c\";\n");

			for (int i=0;i<rules.size();i++) {
				oil.append("\t\tAPP_SRC = \"");
				oil.append(rules.get(i).getSource());
				oil.append("\";\n");
			}
		}
		return super.supportsOILAttribute(category,name,oil);
	}

	public CoderTrampoline(BuilderOptions opts, MethodTableFactory table, ClassStore store)
		throws CompileException {
		super(opts,table,store);

		// Trampoline does not sequentially enumerate task identifiers, we therefore
		// need the switch-function variant of the reverse lookup
		opts.setExtraOption("task_rlookup_fkt");

		// Trampoline does not use constant value system objects, therefore we cannot
		// statically initialize our system objects but need to postpone it to the
		// startup phase
		opts.setExtraOption("init_systemobjects_on_startup");

		// trampoline build system compiles unit-wise
		opts.setExtraOption("gcc-no-whole-program");

		HashSet<String> supAttrs = supportedAttributes.get("TASK");
		supAttrs.add("STACKSIZE");

		supAttrs = supportedAttributes.get("ISR");
		supAttrs.add("PRIORITY");
		supAttrs.add("STACKSIZE");
		supAttrs.add("SOURCE");

		supAttrs = supportedAttributes.get("COUNTER");
		supAttrs.add("SOURCE");
	}

	protected void writeOSEKIncludes(PrintStream out) {
		out.println("#include \"tpl_os.h\"");
		out.println("#include \"tpl_app_config.h\"");

		// declare tasks
		for(TaskDefinition td: opts.getSysAllTasks()) {
			out.print("DeclareTask(");
			out.print(td.getIdentifier());
			out.println(");");
		}

		// declare alarms
		for(AlarmDefinition ad: opts.getSysAllAlarms()) {
			out.print("DeclareAlarm(");
			out.print(ad.getIdentifier());
			out.println(");");
		}

		// declare resources
		for(ResourceDefinition rd: opts.getSysAllResources()) {
			out.print("DeclareResource(");
			out.print(rd.getIdentifier());
			out.println(");");
		}
	}

	public void writeInfoFile() throws CompileException { }

	protected void writeGlobalFiles() throws CompileException {
		writeGlobalFiles("OIL_VERSION = \"2.5\" : \"test\" ;");
	}

	protected void maincHook(PrintStream out) {
		out.println("int main(void) {");
		out.println("\tStartOS(OSDEFAULTAPPMODE);");
		out.println("\treturn 0;");
		out.println("}");
	}

	protected void writeMakefile(PrintStream out) throws CompileException {
		out.println("ifeq ($(TPL_GOIL_TEMPLATEDIR),)");
		out.println("\tTPL_GOIL_TEMPLATEDIR := $(error Environment variable TPL_GOIL_TEMPLATEDIR is not set)");
		out.println("endif");

		out.println(".PHONY: all");
		out.println("all: keso_main.oil");
		out.print("\tgoil --target=");
		out.print(getGOILArchitecture());
		out.println("	-v --templates=\"$(TRAMPOLINE_BASE_PATH)/goil/templates\" -g keso_main.oil");
		out.println("\t$(MAKE)");
	}

	protected abstract String getGOILArchitecture();

	public boolean emptyObjectsNeedBraces() {
		return true;
	}
}

