/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.backend;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;

public final class ExtendedCodeInfo extends CodeInfo {

	private final FlowLatticeInfo flowInfo;

	public ExtendedCodeInfo(FlowLatticeInfo flowInfo) {
		this.flowInfo = flowInfo;
	}

	@Override
	public int getArrayLength(IMNode array) throws CompileException {
		IMDataFlow cell = getCell(array);
		if (cell.isArray() == false)
			return super.getArrayLength(array);
		IMDataFlow sizeCell = ((IMDataFlowArray) cell).getLength();
		if (sizeCell.isConstant() == false)
			return -1;
		return sizeCell.getValue().getIntValue();
	}

	@Override
	public boolean isChecked(IMNode object, Coder coder)
	throws CompileException {
		if (getCell(object).isValid())
			return true;
		return super.isChecked(object, coder);
	}

	@Override
	public boolean isConstant(IMNode object)
	throws CompileException {
		return getCell(object).isConstant();
	}

	@Override
	public boolean canOmitArrayBoundsCheck(IMNode array, IMNode index)
	throws CompileException {
		IMDataFlow arrayCell = getCell(array);
		if (arrayCell.isArray() == false)
			return false;
		IMDataFlow lengthCell = ((IMDataFlowArray) arrayCell).getLength();
		if (isConstOrInterval(lengthCell) == false)
			return false;
		IMDataFlow indexCell = getCell(index);
		if (isConstOrInterval(indexCell) == false)
			return false;
		return (indexCell.minValue() >= 0
				&& indexCell.maxValue() < lengthCell.minValue());

	}

	@Override
	public boolean canOmitMemRangeCheck(IMNode memory, IMNode offset,
			int chunkSize) throws CompileException {
		IMDataFlow memoryCell = getCell(memory);
		if (memoryCell instanceof IMDataFlowAMem == false)
			return false;
		IMDataFlow sizeCell = ((IMDataFlowAMem) memoryCell).getSize();
		if (isConstOrInterval(sizeCell) == false)
			return false;
		IMDataFlow offsetCell = getCell(offset);
		if (isConstOrInterval(offsetCell) == false)
			return false;
		return (offsetCell.minValue() >= 0
				&& offsetCell.maxValue() + chunkSize <= sizeCell.minValue());
	}

	@Override
	public String getSHMName(IMNode memory) throws CompileException {
		IMDataFlow memoryCell = getCell(memory);
		if (memoryCell instanceof IMDataFlowSHM == false)
			return null;
		IMDataFlow nameCell = ((IMDataFlowSHM) memoryCell).getName();
		if (nameCell.isConstant() == false)
			return null;
		return nameCell.getValue().getCPEntry().getDescription();
	}

	@Override
	public String getPortName(IMNode memory) throws CompileException {
		IMDataFlow memoryCell = getCell(memory);
		if (memoryCell instanceof IMDataFlowMessage == false)
			return null;
		IMDataFlow nameCell = ((IMDataFlowMessage) memoryCell).getName();
		if (nameCell.isConstant() == false)
			return null;
		return nameCell.getValue().getCPEntry().getDescription();
	}

	private IMDataFlow getCell(IMNode node) {
		IMDataFlow cell = flowInfo.getCell(node);
		if (cell == null) // Defensive programming, just to make sure
			return IMDataFlow.ANY;
		return cell;
	}

	private static boolean isConstOrInterval(IMDataFlow cell) {
		return (cell.isConstant() || cell.hasFixedRange());
	}

}
