/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.backend;

import keso.compiler.*;
import keso.compiler.imcode.*;

import keso.util.Bitmap;

import java.util.*;

/**
 *
 * extra compiler options:
 * -X:iface_force_matrix
 */
public class InterfaceTypeMatrix {

	private Collection<Bitmap> iface_types = new ArrayList<Bitmap>();
	private Map<Bitmap, Integer> iface_index = new HashMap<Bitmap, Integer>();

	private Coder coder;
	private BuilderOptions opts;
	private ClassStore repository;

	public InterfaceTypeMatrix(BuilderOptions opts, Coder coder, ClassStore store) {
		this.coder=coder;
		this.opts=opts;
		this.repository=store;
	}

	public int registerInterfaceTypeInfo(Bitmap bitmap) throws CompileException {

		int max_ifaces = repository.getMaxInterfaceID();
		if (inlineIFMatrix(opts,max_ifaces)) return bitmap.value(0);

		Integer index = iface_index.get(bitmap);
		if (index==null) {
			iface_types.add(bitmap);
			index = new Integer(iface_types.size());
			iface_index.put(bitmap,index);
		}

		return index.intValue();
	}

	public void emitCheckInterfaceMakro(IMClass iface, ClassTypeInfo type_info) throws CompileException {
		int max_ifaces = repository.getMaxInterfaceID();
		int id   = iface.getInterfaceID();

		coder.header_add("#define CHK_IFACE_ID");
		coder.header_add(type_info.getClassTypeID());
		if(opts.hasOption("use_coded_refs")) {
			coder.header_add("(_obj_) keso_checkiface((object_t*)(NATIVE_ADDR(_obj_)),");
		} else {
			coder.header_add("(_obj_) keso_checkiface(_obj_,");
		}
		if (max_ifaces>31) {
			coder.header_add(id);
		} else {
			coder.header_add(1<<id);
		}
		coder.header_add(")\n");

		coder.header_add("#define INSTANCEOF_IFACE_ID");
		coder.header_add(type_info.getClassTypeID());
		if(opts.hasOption("use_coded_refs")) {
			coder.header_add("(_obj_) keso_instanceof_iface((object_t*)(NATIVE_ADDR(_obj_)),");
		} else{
			coder.header_add("(_obj_) keso_instanceof_iface(_obj_,");
		}
		if (max_ifaces>31) {
			coder.header_add(id);
		} else {
			coder.header_add(1<<id);
		}
		coder.header_add(")\n");
	}

	public void emitInterfaceMatrix() throws CompileException {
		int max_ifaces = repository.getMaxInterfaceID();

		coder.global_header_add("#define keso_isNonCopyable(_class_id_) ");

		IMClass noncopyable = repository.getClass("keso/core/NonCopyable");
		if (noncopyable!=null) {
			int id   = noncopyable.getInterfaceID();
			coder.global_header_add("keso_implements_iface(_class_id_,");
			if (max_ifaces>31) {
				coder.global_header_add(id);
			} else {
				coder.global_header_add(1<<id);
			}
			coder.global_header_add(")\n");
		} else {
			coder.global_header_add("(0)\n");
		}

		if (!inlineIFMatrix(opts,max_ifaces)) {
			coder.global_add("\n/* the interface type matrix */\n");
			coder.global_add("/* length: ");
			coder.global_add(iface_types.size());
			coder.global_add(" */\n");
			coder.global_add(opts.declareConst());
			coder.global_add("unsigned long iface_type_matrix[] = {\n");
			for (Bitmap iftype : iface_types) {
				coder.global_add("\t");
				coder.global_add(iftype.emit_flat());
				coder.global_add("\n");
			}
			coder.global_add("};\n");
		}

		/* Option iface_store
		 *
		 * Die Bitvektoren, welche Information ueber die von einer Klasse implementierten
		 * Schnittstellen enthalten, werden nicht im class_store, sondern in einem eigenen
		 * Feld iface_store gespeichert. Dafuer werden alle Klassen so sortiert, dass Klassen
		 * die Interfaces implementieren, niedrigere IDs als die Klassen, die keine
		 * Schnittstellen implementieren erhalten. Der iface_store muss dann nur fuer diese
		 * Klassen einen Eintrag mit dem Bitvektor enthalten. Ohne die Option wird der
		 * Bitvektor direkt mit jeder Klasse im class_store gespeichert.
		 *
		 * Vorteil: Weniger RAM Verbrauch
		 * Nachteil: Es wird ein zusaetzlicher Ueberpruefung benoetigt, ob die Klasse einen Eintrag
		 *   im iface_store hat, oder nicht. Diese Ueberpruefung und ein zusaetzliche Subtraktion
                 *   kostet Laufzeit und ein paar Byte im Programmspeicher (ROM/Flash).
		 *
		 * Die erste ID einer Klasse, die Schnittstellen implementiert, ist KESO_IFACE_START,
		 * die Anzahl der Klassen, die Schnittstellen implementieren, ist KESO_IFACE_SIZE.
		 */
		boolean iface_store = opts.hasOption("iface_store");

		coder.global_header_add("object_t *keso_checkiface(object_t* obj, unsigned int id);\n");
		coder.global_add("\nobject_t *keso_checkiface(object_t* obj, unsigned int id) {\n");
		coder.global_add("	if (obj==NULL) return obj;\n");
		coder.global_add("	if (keso_implements_iface(obj->class_id, id)==0)\n");
		coder.global_add("		KESO_THROW_ERROR(\"checkcast\");\n");
		coder.global_add("	return obj;\n}\n");

		coder.global_header_add("jboolean keso_instanceof_iface(object_t* obj, unsigned int id);\n");
		coder.global_add("\njboolean keso_instanceof_iface(object_t* obj, unsigned int id) {\n");
		coder.global_add("	if (obj==NULL) return 1;\n");
		coder.global_add("	return keso_implements_iface(obj->class_id, id);\n}\n");

		coder.global_header_add("jboolean keso_implements_iface(class_id_t cls, unsigned int id);\n");
		coder.global_add("\njboolean keso_implements_iface(class_id_t cls, unsigned int id) {\n");
		coder.global_add("	unsigned int info;\n\n");
		if (iface_store) {
			coder.global_add("	if ((unsigned int)(cls-KESO_IFACE_START)>=KESO_IFACE_SIZE)\n");
			coder.global_add("		return 0;\n\n");
			coder.global_add("	info = iface_store[cls-KESO_IFACE_START];\n");
		} else {
			coder.global_add("	info = CLASS(cls).ifaces;\n");
		}
		if (inlineIFMatrix(opts,max_ifaces)) {
			coder.global_add("	if (!(info & id))\n");
		} else {
			if (max_ifaces<32) {
				coder.global_add("\tif (info==0 || !(iface_type_matrix[info-1] & id))\n");
			} else {
				opts.warn("Have "+max_ifaces+" interfaces! (>32)");
				coder.global_add("\tif (info==0 || !(iface_type_matrix[");
				coder.global_add("((info-1)*");
				coder.global_add((max_ifaces+31)/32+1);
				coder.global_add(")+(id/32)] & (1<<(id%32))))\n");
			}
		}
		coder.global_add("		return 0;\n\n");
		coder.global_add("	return 1;\n}\n");
	}

	private static boolean inlineIFMatrix(BuilderOptions sopts, int max_ifaces) {
		if (sopts.hasOption("iface_force_matrix")) return false;
		return (max_ifaces<8);
	}
}
