/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/*
	M.Strotz 2011
*/

package keso.compiler.backend.gradmp;

import java.util.*;

import keso.compiler.backend.gradmp.util.*;

import keso.compiler.*;
import keso.compiler.imcode.*;

import keso.compiler.backend.*;



final class ClassLayout {

	public final static class ClassInfo {
		ClassInfo superclass;
		long	last_ref,	//offset der letzten Referenz
				next_prim;	//offset des naechsten Felds (primitiv)
		ClassInfo( ClassInfo supi , long ref, long prim ) {
			superclass = supi;
			last_ref = ref;
			next_prim = prim;
		}
		ClassInfo( ClassInfo supi ) {
			this( supi, supi.last_ref, supi.next_prim );
		}
	};

	private final HashMap<IMField,FieldInfo>	fieldmap;
	private final HashMap<IMClass,ClassInfo>	classmap;

	private FieldInfo m_objecthead, m_classdesc, m_arraylength;

	public ClassLayout() {
		fieldmap = new HashMap<IMField,FieldInfo>();
		classmap = new HashMap<IMClass,ClassInfo>();
	}


	public Collection<FieldInfo> getFieldInfos() { return fieldmap.values(); }
	public Map<IMField,FieldInfo> getFieldMap() { return fieldmap; }


	public FieldInfo mapField( IMField field ) {
		return fieldmap.get( field );
	}
	private ClassInfo mapClass( IMClass clazz ) {
		return classmap.get( clazz );
	}

	public void setupFrom( ClassStore cs, BuilderOptions opt ) {
		setupMagic( cs, opt );
		scan(cs);
	}


	//TODO entsprechende Methoden in ClassStore einfuegen!
	//TODO Kommentar zum Classlayout einfuegen (wie Coder.add_initObject)
	public void setupMagic( ClassStore st, BuilderOptions opt ) {
		boolean need_gc_info = opt.getGlobalHeap().needGCInfo();
		long sizeof_gc_info, sizeof_class_id, sizeof_arraylength;
		sizeof_gc_info = 1;
		if( opt.is8BitController() ) {
			sizeof_class_id = 1;
			sizeof_arraylength = 1;
		} else if( opt.hasOption("compact_header") ) {
			sizeof_class_id = 1;
			sizeof_arraylength = 2;
		} else {
			sizeof_class_id = 2;
			sizeof_arraylength = 4;
		}
		{
			long offs = 0;
			long size;
			if( need_gc_info ) {
				size = sizeof_gc_info;
				offs = add_aligned(offs,size);
				log("gc info: " + (offs-size) + ", " + size);
			}
			size = sizeof_class_id;
			offs = add_aligned(offs,size);
			log("class id: " + (offs-size) + ", " + size);
			m_classdesc = new FieldInfo( "__class_id", new Interval(offs - size, size) ); //class_id
			m_objecthead = new FieldInfo( "__object_head", new Interval(0, offs) );
		}
		{
			long offs = objecthead().sizeof();
			long size = sizeof_arraylength;
			offs = add_aligned(offs,size);
			log("array length: " + (offs-size) + ", " + size);
			m_arraylength = new FieldInfo( "__arraylength", new Interval(offs - size, size) ); //size
		}
	}

	public FieldInfo objecthead() {
		return m_objecthead;
	}
	public FieldInfo classdesc() {
		return m_classdesc;
	}
	public FieldInfo arraylength() {
		return m_arraylength;
	}

	public void scanRootClass( IMClass object ) { scan_down_rec( null, object ); }
	public void scan( ClassStore store ) { scanRootClass( store.getRootClass() ); }

	private void scan_down_rec( ClassInfo supi, IMClass clazz ) {
		ClassInfo info;
		if( supi == null || null == clazz.getSuperClass()) { //java.lang.Object
			log("scanning java.lang.Object: " + clazz );
			info = new ClassInfo( null, 0, objecthead().sizeof() );
		} else {
			info = new ClassInfo( supi );
			scan( info, clazz );
		}
		classmap.put( clazz, info );
		if( clazz.hasSubClasses() && !clazz.isInterface() ) {
			IMClass[] subs = clazz.getSubClasses();
			for( int i = 0; i < subs.length; ++i ) {
				scan_down_rec( info, subs[i] );
			}
		}
	}


	/*
		//ANNAHME:
			C-Compiler sortiert Felder nicht um.
			Auszug aus dem C99 Standard:
			Within a structure object, the non-bit-field members
			and the units in which bit-fields reside have addresses
			that increase in the order in which they are declared.
			A pointer to a structure object, suitably converted,
			points to its initial member (or if that member is a bit-field,
			then to the unit in which it resides), and vice versa.
			There may be unnamed padding within a structure object,
			but not at its beginning.

		//ANNAHME - FALSCH:
			Alle Typen besitzen "natuerliches" Alignment.
			z.B.: int64 ist auch auf 16-Bit-Systemen auf 8 Byte ausgerichtet.
			//TODO ueberpruefen...
			DONE: ueberprueft.
				ANNAHME: Maximales Alignment ist Pointer-Breite ...?
	*/

	private void scan( ClassInfo info, IMClass clazz ) {
		log("scanning layout: " + clazz );
		//IMClass sclass = getSuperClass();
		//if (fieldsMT.size()>0 && !sclass.hasBaseField() && !opts.hasOption("mt_no_save_memobj")) {
		//	buffer.append("\tobject_t* memobj;\n");
		//}
		scanRefs( info, clazz );
		//if (fieldsMT.size()>0 && !sclass.hasBaseField()) {
		//	buffer.append("\tvoid* base;\n");
		//}
		scanPrims( info, clazz );
		//TODO RawFields
	}

	private void dump_fieldmap_entries() {
		System.err.println("fieldmap-entries: ");
		for( Map.Entry<IMField, FieldInfo> e : fieldmap.entrySet() ) {
			System.err.println( "> " + e.getKey().hashCode() + "> " + e.getKey() + " --> " + e.getValue() );
		}
	}

	private void scanRefs( ClassInfo info, IMClass cl ) {
		//for( IMField field : cl.getFieldsRef() ) { // die referenzen sind genau rueckwaerts angeordnet!!
		Vector<IMField> vec = cl.getAddedRefFields();
		if( vec == null ) return;
		for( int i = vec.size() - 1; i >= 0; --i ) {
			IMField field = vec.get( i );
			info.last_ref -= size_of_ref();
			FieldInfo fi = new FieldInfo( field, new Interval(info.last_ref, size_of_ref()) );
			fieldmap.put( field, fi );
		}
	}

	private void scanPrims( ClassInfo info, IMClass cl ) {
		long offs = info.next_prim;
		Vector<IMField> vec = cl.getAddedPrimitiveFields();
		if( vec == null ) return;
		for( IMField field : vec ) {
			int size = size_of( field );
			//if( size > size_max ) size_max = size;
			offs = add_aligned(offs,size); //TODO hat int64 auf 16Bit-Sys nat Align?
			FieldInfo fi = new FieldInfo( field, new Interval(offs - size, size) );
			fieldmap.put( field, fi );
		}
		info.next_prim = offs;
	}

	private boolean has_no_java_fields( final IMClass clazz ) {
		for( IMClass curc = clazz; curc != null; curc = curc.getSuperClass() ) {
			Vector<IMField> vec = curc.getAddedRefFields();
			if( vec != null && !vec.isEmpty() ) {
				return false;
			}
			vec = curc.getAddedPrimitiveFields();
			if( vec != null && !vec.isEmpty() ) {
				return false;
			}
		}
		return true;
	}

	public void addOffsetofAsserts( final IMClass clazz, final Coder coder ) throws CompileException {
		if( has_no_java_fields( clazz ) ) return;

		coder.header_add("\n/*if C-compiler failes here rerun jino with \"-gradmp inactive\"*/\n");
		final ClassInfo ci = mapClass( clazz );
		final int translation = (int)ci.last_ref;
		//final int translation = clazz.getObjRefFieldCount() * size_of_ref();

		for( IMClass curc = clazz; curc != null; curc = curc.getSuperClass() ) {
			Vector<IMField> vec = curc.getAddedRefFields();
			if( vec != null ) {
				for( IMField f : vec ) {
					add_offsetof_assert( clazz, f, translation, coder );
				}
			}
			vec = curc.getAddedPrimitiveFields();
			if( vec != null ) {
				for( IMField f : vec ) {
					add_offsetof_assert( clazz, f, translation, coder );
				}
			}
		}

		int cid_offs = (int)(m_classdesc.offset()) - translation;
		log("class id assert: " + cid_offs + ", " + clazz.getAlias() );
		coder.header_add_offsetof_assert( clazz.getAlias() + "_t", "class_id", cid_offs );
	}

	private void add_offsetof_assert( IMClass clazz, IMField field, int translation, Coder coder ) throws CompileException {
		FieldInfo fi = mapField( field );
		if( fi == null ) return;
		coder.header_add_offsetof_assert( clazz.getAlias() + "_t", field.getAlias(), (int)(fi.offset()) - translation );
	}

	private void log( String s ) {
		BuilderOptions.getOpts().vverbose
		("gradmp: " + s );
	}


	private static long align( long offs, long size_ ) {
		final long asize = alignof(size_);
		return ((offs + asize - 1)/asize)*asize;
	}
	private static long add_aligned( long offs, long size ) {
		return align(offs,size) + size;
	}
	private static long alignof( long size ) {
		//ANNAHME: pointers have largest alignment.
		final long refsize = size_of_ref();
		return (size < refsize) ? size : refsize;
	}


	private static String typestring( IMField field ) {
		return BuilderOptions.getOpts().typeToString( field.getDatatype() );
	}
	private static int size_of( IMField field ) {
		String ts = typestring( field );
		return BuilderOptions.getOpts().getTypeSize( ts );
	}
	private static int size_of_ref() {
		return BuilderOptions.getOpts().targetAddrSize();
	}

};
