/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/*
	M.Strotz 2012
*/

package keso.compiler.backend.gradmp;

import keso.compiler.imcode.*;
import keso.compiler.*;
import keso.compiler.backend.*;

public class Defs {

	/**
		Optimization-Levels
	*/
	public final static int
		NOPT = 0,
		SDET = 1,
		WDET = 2,
		JUPR = 3,
		KILLALL = 4;

	/**
		Optimization-Levels
	*/
	public final static int
		ALLCHECKS = NOPT,
		INACTIVE = NOPT,
		WEAK_DETECTION = WDET,
		HWOFFLOADING = SDET,
		CONTAINMENT = JUPR;

	private final static String
		STR_ALLCHECKS = "allchecks",
		STR_INACTIVE = "inactive",
		STR_HWOFFLOADING = "hwoffloading",
		STR_CONTAINMENT = "containment";


	
	public static String levelToString( int level ) {
		switch( level ) {
			case INACTIVE: return STR_INACTIVE;
			case HWOFFLOADING: return STR_HWOFFLOADING;
			//case WDET: return STR_WEAK_DETECTION;
			case CONTAINMENT: return STR_CONTAINMENT;
			case KILLALL: return "KILLALL";
			default: return "error__unknown_level";
		}
	}

	public static int stringToLevel(String str) {
		if(null == str) return INACTIVE;
		str = str.toLowerCase();
		if( str.equals(STR_ALLCHECKS) ) return INACTIVE;
		if( str.equals(STR_HWOFFLOADING) ) return HWOFFLOADING;
		if( str.equals(STR_CONTAINMENT) ) return CONTAINMENT;
		//if( str.equals("KILLALL") ) return CONTAINMENT; //KILLALL;
		warn("Unknown GradMP-Level: \"" + str + "\".\n");
		return INACTIVE;
	}

	
	// ---- ----

	private static final String
		OMIT_BEG = ""; //"nullcheck omitted: ";


	public static final String

		READ_TRAP = OMIT_BEG + "reading this field from null will trap.",
		READ_NULL_NULL = OMIT_BEG + "reading this reference from null will read null.",
		READ_LOCAL = OMIT_BEG + "reading invalid primitive data has just local impact.",

		READ_EMPTY_ARRAY = OMIT_BEG + "array at null will appear empty.",

		READ_NOT_ARRAYLENGTH = OMIT_BEG + "arraylength won't be read (local impact).",

		WRITE_TRAP = OMIT_BEG + "null-write will trap.",
		WRITE_IGNORE = OMIT_BEG + "null-write will be ignored.",

		READ_ERROR_CLASSDESC = OMIT_BEG + "null-read will read errorhandling-classdescriptor.",

		NO_VALID_THIS = OMIT_BEG + "no valid this-ref.";


	public static final String
		BOUND_READ_LOCAL = "(bounds) " + READ_LOCAL; //"bounds-check omitted: reading invalid primitive data has just local impact.";


	public static boolean preservesErrorDetection(int level) {
		return level <= HWOFFLOADING;
	}
	public boolean preservesErrorDetection(final String rtoc) {
		return	rtoc == READ_TRAP || rtoc == WRITE_TRAP || rtoc == READ_ERROR_CLASSDESC;
	}


	
	
	
	protected static void log( String s ) {
		BuilderOptions.getOpts().verbose
		( "gradmp: " + s );
	}
	protected static void vlog( String s ) {
		BuilderOptions.getOpts().vverbose
		( "gradmp: " + s );
	}
	protected static void warn( String s ) {
		BuilderOptions.getOpts().warn
		( "gradmp: " + s );
	}
	protected static void vwarn( String s ) {
		BuilderOptions.getOpts().vwarn
		( "gradmp: " + s );
	}
	
	
	
	public static int getCurrentSafetyLevel(final Coder coder) {
		int minlevel = CONTAINMENT; //returns "containment" if Block is not reachable...
		for(IMDomain dom : coder.getCurrentDomains()) {
			keso.compiler.config.Attribute<?> attr = dom.getDef().getAttribute("Safety");
			int curlevel = ALLCHECKS;
			if(null != attr) {
				curlevel = stringToLevel(attr.valueString());
			}
			if( curlevel < minlevel ) minlevel = curlevel;
		}
		return minlevel;
	}


	
};
