/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/*
	M.Strotz 2011
*/

package keso.compiler.backend.gradmp;

import keso.compiler.imcode.*;
import keso.compiler.*;
import keso.compiler.backend.*;

public final class GradMPChecks extends Defs {

	// ---- ----
	private CheckChecker	chk;
	private ClassLayout		cll;

	private boolean no_valid_this_ptr = false;
	private boolean no_omit_nullchecks = false;
	private boolean no_omit_boundchecks = false;
	private boolean read_has_global_impact = false;
	private boolean unexpected_nullrefs = false;
	private boolean m_active = false;

	// ---- ----

	public GradMPChecks() { }

	public GradMPChecks(ClassStore cs, BuilderOptions opt) {
		this();
		setupFrom( cs, opt );
	}

	// ---- ----

	public boolean active() { return m_active; }

	// ---- ----
	// Setup
	// ---- ----
	public void setupFrom(ClassStore cs, BuilderOptions opt) {

		no_valid_this_ptr = opt.hasOption("no_valid_this_ptr");
		no_omit_nullchecks = opt.hasOption("gradmp_no_omit_nullchecks");
		no_omit_boundchecks = opt.hasOption("gradmp_no_omit_boundchecks");		
		unexpected_nullrefs = opt.hasOption("gradmp_allow_unexpected_nullrefs");

		read_has_global_impact =
			opt.hasOption("gradmp_read_has_global_impact")
			|| !opt.hasOption("gradmp_read_has_local_impact"); //dominant: true;

		m_active = opt.hasOption("gradmp") && !opt.hasOption("gradmp_inactive"); //dominant: false;

		createChecker( opt );
		if(!active()) return;
		cll = new ClassLayout();
		cll.setupFrom(cs, opt);
		scanFields();
	}
	
	private MemoryDescription createMemoryDescription(final BuilderOptions opt) {
		MemoryDescription md = null;
		try {
			md = opt.getSysDef().getMemoryDescription();
		} catch( NullPointerException e ) {}
		if(md == null) {
			if( active() ) warn("no memory-description.");
			//m_active = false;
			md = new MemoryDescription(); //dummy
			//return;
		}
		md.setup();
		if( m_active ) {
			vlog("got memory-description:\n" + md.getDescriptionString() );
		}
		return md;
	}

	private void createChecker(final BuilderOptions opt) {
		MemoryDescription md = createMemoryDescription(opt);
		chk = new CheckChecker(md,unexpected_nullrefs,read_has_global_impact);
	}

	private void scanFields() { //CHG
		chk.checkClassDesc(cll.classdesc());
		vlog("checked " + cll.classdesc().toStringDesc());
		chk.checkArrayLength(cll.arraylength());
		vlog("checked " + cll.arraylength().toStringDesc());
		
		for(FieldInfo fi : cll.getFieldInfos()) {
			chk.check(fi);
			vlog("checked " + fi.toStringDesc());
		}
	}


	// ---- ----
	// Coder
	// ---- ----

	public void addOffsetofAsserts(final IMClass clazz, final Coder coder) throws CompileException {
		if( !active() ) return;
		log("add offsetof-asserts: " + clazz );
		cll.addOffsetofAsserts( clazz, coder );
	}


	// ---- ----
	// Coder
	// ---- ----

	/**
		@param mnem		name of the bytecode operation that causes the check ("reason")
		@param field	field referenced by GETFIELD / PUTFIELD
	*/
	public boolean omitNullcheck(final Coder coder, final String mnem, final IMField field) {
		return null != reasonToOmitNullcheck(coder, mnem, field );
	}
	/**
		@param mnem		name of the bytecode operation that causes the check ("reason")
	*/
	public boolean omitBoundcheck(final Coder coder, final String mnem) {
		return null != reasonToOmitBoundcheck(coder, mnem);
	}

	// ---- ----

	private boolean is_reference_operation(final String mnem) {
		char c = mnem.charAt(0);
		return c == 'L'; //nicht A!
	}

	/*private boolean is_primitive_operation( final String mnem ) {
		char c = mnem.charAt(0);
		switch(c) {
			case 'B': //byte
			case 'C': //char
			case 'S': //short
			case 'I': //int
			case 'J': //long
			case 'F': //float
			case 'D': //double
				return true;
			default:
				return false;
		}
	}*/

	// ---- ----

	/**
		@param mnem		name of the bytecode operation that causes the check ("reason")
		@param field	field referenced by GETFIELD / PUTFIELD or null if no field is accessed.
		@return	returns the reason to omit the nullcheck for this operation as a string,
				or null if the check may not be omitted.
	*/
	public String reasonToOmitNullcheck(final Coder coder, final String mnem, final IMField field) {
		vlog( "nullcheck: " + mnem + " " + ((field != null)? field.getAlias() : "null") );
		String res = reasonToOmitNullcheck_(coder, mnem, field);
		vlog( res );
		return res;
	}
	private String reasonToOmitNullcheck_(final Coder coder, final String mnem, final IMField field) {
		if( !active() ) return null;
		if( no_omit_nullchecks ) return null;
		
		int safety_level = getCurrentSafetyLevel(coder);
		//if( safety_level == KILLALL ) return "KILLALL";

		if( mnem.length() < 2 ) {
			vwarn("unknown operation: " + mnem );
			return null; //unbekannter befehl
		}

		if( mnem.equals("ARRAYLENGTH") ) return chk.reasonToOmitCheckOnArrayLength(safety_level);
		if( mnem.equals("INVOKEVIRTUAL") ) return chk.reasonToOmitCheckOnInvokeVirtual(safety_level);
		if( mnem.equals("INVOKEINTERFACE") ) return chk.reasonToOmitCheckOnInvokeInterface(safety_level);
		if( mnem.equals("INVOKESPECIAL") ) return chk.reasonToOmitCheckOnInvokeSpecial(safety_level, no_valid_this_ptr);

		final String family = mnem.substring(1);
		boolean is_ref = is_reference_operation(mnem);

		if( family.equals("ALOAD") )	return chk.reasonToOmitNullcheckOnArrayLoad(safety_level, is_ref);
		if( family.equals("ASTORE") )	return chk.reasonToOmitNullcheckOnArrayStore(safety_level, is_ref);

		if( family.length() < 4 ) {
			vwarn("unknown operation: " + mnem );
			return null;
		}

		if( family.substring(3).equals("FIELD") ) {
			// PUTFIELD / GETFIELD
			if( field == null ) {
				vwarn("no field specified: " + mnem );
				return null;
			}
			FieldInfo fi = cll.mapField( field );
			if( fi == null ) {
				vwarn("unknown field: " + field );
				return null;
			}
			final String dir = family.substring(0,3);
			if( dir.equals("GET") )	return fi.reasonToOmitCheckOnGet(safety_level);
			if( dir.equals("PUT") ) return fi.reasonToOmitCheckOnPut(safety_level);
		}

		{
			vwarn("unknown operation: " + mnem );
			return null;
		}
	}


	/**
		@param mnem		name of the bytecode operation that causes the check ("reason")
		@return	returns the reason to omit the boundcheck for this operation as a string,
				or null if the check may not be omitted.
	*/
	public String reasonToOmitBoundcheck(final Coder coder, final String mnem) {
		vlog( "boundcheck: " + mnem );
		String res = reasonToOmitBoundcheck_(coder,mnem);
		vlog( res );
		return res;
	}
	private String reasonToOmitBoundcheck_(final Coder coder, final String mnem) {
		if( !active() ) return null;
		if( no_omit_boundchecks ) return null;
		
		int safety_level = getCurrentSafetyLevel(coder);
		//if( safety_level == KILLALL ) return "KILLALL";

		if( mnem.length() < 4 ) {
			vwarn("unknown operation: " + mnem );
			return null;
		}

		if( mnem.equals("ARRAYLENGTH") ) {
			//TODO was bedeutet ARRAYLENGTH in verbindung mit boundcheck?
			vwarn( mnem + " on boundcheck.");
			return null;
		}

		final String family = mnem.substring(1);
		boolean is_ref = is_reference_operation(mnem);

		if( family.equals("ALOAD") )	return chk.reasonToOmitBoundcheckOnArrayLoad(safety_level, is_ref);
		if( family.equals("ASTORE") )	return chk.reasonToOmitBoundcheckOnArrayStore(safety_level, is_ref);

		{
			vwarn("unknown operation: " + mnem );
			return null;
		}
	}

};
