/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/*
	M.Strotz 2011
*/

package keso.compiler.backend.gradmp;

import keso.compiler.backend.gradmp.util.*;

import java.util.*;

public final class MemoryBlock implements Comparable< MemoryBlock >
{

	/**
		Attribute Flags (LIRWX)
	*/
	public final static int

		//! executable
		X = 0x01,
		//! read/write
		W = 0x02,
		//! readonly
		R = 0x04,
		//! allocatable
		A = 0x08,
		//! initialized
		I = 0x10,
		L = I,
		//TODO P?
		P = 0x20,

		//! inverted
		IX = X << 8,
		IW = W << 8,
		IR = R << 8,
		IA = A << 8,
		II = I << 8,
		IL = L << 8,
		IP = P << 8,

		DEFAULT_ATTRIB = 0;

	public final static int

		RW_UNSPEC = 0,
		RW_CONST = 1,
		RW_TRAP = 2,
		RW_NORMAL = 3,
		RW_ERROR = 4,

		READ_NORMAL = RW_NORMAL,
		WRITE_NORMAL = RW_NORMAL,
		READ_CONST = RW_CONST,
		WRITE_IGNORE = RW_CONST,
		READ_TRAP = RW_TRAP,
		WRITE_TRAP = RW_TRAP,
		READ_UNSPEC = RW_UNSPEC,
		WRITE_UNSPEC = RW_UNSPEC,
		READ_ERROR = RW_ERROR,
		WRITE_ERROR = RW_ERROR;

	private final static String
		STR_NORMAL = "normal",
		STR_READ_NORMAL = STR_NORMAL,
		STR_WRITE_NORMAL = STR_NORMAL,
		STR_READ_CONST = "const",
		STR_WRITE_IGNORE = "ignore",
		STR_TRAP = "trap",
		STR_READ_TRAP = STR_TRAP,
		STR_WRITE_TRAP = STR_TRAP,
		STR_UNSPEC = "unspec",
		STR_READ_UNSPEC = STR_UNSPEC,
		STR_WRITE_UNSPEC = STR_UNSPEC,
		STR_ERROR = "desc_error",
		STR_READ_ERROR = STR_ERROR,
		STR_WRITE_ERROR = STR_ERROR;



	public MemoryBlock( int on_read, int on_write, String name, int attr, Interval i, byte[] const_mem ) {
		m_onr = on_read;
		m_onw = on_write;
		m_name = name;
		m_attr = attr;
		m_interval = i;
		m_cbs = const_mem;
		m_sub = new ArrayList<MemoryBlock>();
	}

	public MemoryBlock( int on_read, int on_write, String name, int attr, long org, long len, byte const_byte ) {
		this( on_read, on_write, name, attr, new Interval( org, len ), new byte[]{ const_byte } );
	}

	public MemoryBlock( int on_read, int on_write, String name, int attr, long org, long len ) {
		this( on_read, on_write, name, attr, org, len, (byte)0 );
	}

	public MemoryBlock( int on_read, int on_write, String name, long org, long len ) {
		this(
			on_read, on_write,
			name,
			0, //get_attr_from( on_read, on_write ),
			org, len
		);
	}

	public MemoryBlock( String name, int attr, long org, long len ) {
		this(
			get_on_read_from( attr ),
			get_on_write_from( attr ),
			name, attr, org, len
		);
	}
	public MemoryBlock( String name, String attr, long org, long len ) {
		this( name, parseAttributes(attr), org, len );
	}
	public MemoryBlock( String name, long org, long len ) {
		this( name, DEFAULT_ATTRIB, org, len );
	}
	public MemoryBlock( String name ) {
		this( RW_UNSPEC, RW_UNSPEC, name, 0, null, null );
	}

	public String name() { return m_name; }
	public int attributes() { return m_attr; }

	public Interval interval() { return m_interval; }
	public long origin() { return (null != m_interval)?m_interval.origin():0l; }
	public long length() { return (null != m_interval)?m_interval.length():0l; }
	/** first address beyond this memory-block */
	public long beyond() { return (null != m_interval)?m_interval.beyond():0l; }

	public boolean getL() { return (m_attr & L) != 0; }
	public boolean getI() { return (m_attr & I) != 0; }
	public boolean getA() { return (m_attr & A) != 0; }
	public boolean getR() { return (m_attr & R) != 0; }
	public boolean getW() { return (m_attr & W) != 0; }
	public boolean getX() { return (m_attr & X) != 0; }
	public boolean getP() { return (m_attr & P) != 0; }


	/** does reading this memory-block cause a trap?*/
	public boolean trapsOnRead() { return onRead() == READ_TRAP; }
	/** does writing this memory-block cause a trap?*/
	public boolean trapsOnWrite() { return onWrite() == WRITE_TRAP; }

	/** liefern Lesezugriffe immer einen konstanten Wert? */
	public boolean constOnRead() { return onRead() == READ_CONST; }
	/** does this memory-block ignore write-access? */
	public boolean ignoresWrite() { return onWrite() == WRITE_IGNORE; }
	
	/** has reading/writing unspecified behaviour?*/
	public boolean unspecOnRead() { return onRead() == READ_UNSPEC; }
	public boolean unspecOnWrite() { return onWrite() == WRITE_UNSPEC; }
	
	/** has reading/writing "normal" behaviour?*/
	public boolean normalOnRead() { return onRead() == READ_NORMAL; }
	public boolean normalOnWrite() { return onWrite() == WRITE_NORMAL; }

	/** behavior on read-access */
	public int onRead() { return (int)m_onr; }
	/** behavior on write-access */
	public int onWrite() { return (int)m_onw; }

	/** set behavior on read-access */
	public void setOnRead( int behv ) { m_onr = behv; }
	/** set behavior on write-access */
	public void setOnWrite( int behv ) { m_onw = behv; }

	/** set interval */
	public void setInterval( Interval i ) {
		m_interval = i;
	}
	public void setInterval( long org, long len ) {
		m_interval = new Interval( org, len );
	}
	public void setOrigin( long org ) { setInterval( org, length() ); }
	public void setLength( long len ) { setInterval( origin(), len ); }
	public boolean hasInterval() { return null != m_interval; }

	/** set attributes */
	public void setAttributes( int attr ) { m_attr = attr; }


	public String onReadToString() {
		switch( onRead() ) {
			case READ_TRAP: return STR_READ_TRAP;
			case READ_CONST: return STR_READ_CONST;
			case READ_NORMAL: return STR_READ_NORMAL;
			case READ_UNSPEC: return STR_READ_UNSPEC;
			default: return STR_READ_ERROR;
		}
	}
	public String onWriteToString() {
		switch( onWrite() ) {
			case WRITE_TRAP: return STR_WRITE_TRAP;
			case WRITE_IGNORE: return STR_WRITE_IGNORE;
			case WRITE_NORMAL: return STR_WRITE_NORMAL;
			case WRITE_UNSPEC: return STR_WRITE_UNSPEC;
			default: return STR_WRITE_ERROR;
		}
	}

	/*
		R: speicher ist lesbar. (READ_NORMAL)
		W: speicher ist beschreibbar und lesbar (WRITE_NORMAL/READ_NORMAL)
		!R: speicher ist nicht lesbar
		!W: speicher ist nicht schreibbar (?)
		TODO
		Annahme(!): Ungueltige Zugriffe verursachen Trap.
		Alternativ:
			unguelt. read: default
			unguelt. write: (ignore) / default
	*/
	private static int get_on_read_from( int attr ) {
		return ( (attr & (R|W)) == 0 )? READ_UNSPEC : READ_NORMAL;
	}
	private static int get_on_write_from( int attr ) {
		return ( (attr & W) == 0 )? WRITE_UNSPEC : WRITE_NORMAL;
	}
	private static int get_attr_from( int onr, int onw ) {
		int erg = 0;
		if( onr == READ_NORMAL ) erg |= R;
		if( onw == WRITE_NORMAL ) erg |= W;
		return erg;
	}

	/** */
	public void correctRWBehavior_rec() throws RuntimeException {
		correctRWBehavior();
		if( onRead() == READ_ERROR ) {
			throw new RuntimeException("memoryblock \"" + name() + "\" has contradicting read-behavior.");
		}
		if( onWrite() == WRITE_ERROR ) {
			throw new RuntimeException("memoryblock \"" + name() + "\" has contradicting write-behavior.");
		}
		for( MemoryBlock sub : subblocks() ) {
			sub.correctRWBehavior_rec();
		}
	}

	/** */
	public void correctRWBehavior() {
		setOnRead( correctRWBehavior( onRead(), get_on_read_from( attributes() ) ) );
		setOnWrite( correctRWBehavior( onWrite(), get_on_write_from( attributes() ) ) );
	}

	private static int correctRWBehavior( int cur, int fa ) {
		switch( cur ) {
		case RW_UNSPEC:
			return fa;

		case RW_TRAP:
			if( fa == RW_UNSPEC ) {
				return cur;
			} //fa == RW_NORMAL
			return RW_ERROR;

		case RW_CONST: //IGNORE
			if( fa == RW_UNSPEC ) {
				return cur;
			} //fa == RW_NORMAL
			return RW_ERROR;

		case RW_NORMAL:
			return cur;

		default:
			return RW_ERROR;
		}
	}


	/** does this memory-block include address x?*/
	public boolean includes( long address ) {
		return interval().includes( address );
	}
	public boolean includes( long address, int size ) {
		return interval().includes_org_len( address, size );
	}
	public boolean includes( MemoryBlock r ) {
		return interval().includes( r.interval() );
	}

	public int compareTo( MemoryBlock mb ) {
		return (int)( origin() - mb.origin() );
	}

	public String attributesToString() {
		final char[] c = new char[] {'-','-','-','-','-','-'};
		if( getR() ) c[0] = 'r';
		if( getW() ) c[1] = 'w';
		if( getX() ) c[2] = 'x';
		if( getA() ) c[3] = 'a';
		if( getI() ) c[4] = 'i';
		if( getP() ) c[5] = 'p';
		return new String( c );
	}

	public String toString() {
		return "MemBlock( "
			+ name()
			+ ", " + attributesToString()
			+ ", 0x" + Long.toHexString(origin()) + ", " + length()
			+ ", " + onReadToString() + ", " + onWriteToString()
		+ ")";
	}
	public String toStringDesc( String preln ) {
		String res = preln + this;
		if( constOnRead() && m_cbs != null ) {
			res += "{ ";
			for( int i = 0; i < m_cbs.length; ++i ) {
				res += Integer.toHexString(m_cbs[i] & 0xff) + " ";
			}
			res += "}";
		}
		preln += "  ";
		for( MemoryBlock mb : subblocks() ) {
			res += "\n" + mb.toStringDesc(preln);
		}
		return res;
	}
	public String toStringConfig() {
		return name() + " = {\n"
			+ "\torigin = " + origin() + ";\n"
			+ "\tlength = " + length() + ";\n"
			+ "\tread = \"" + onReadToString() + "\";\n"
			+ "\twrite = \"" + onWriteToString() + "\";\n"
			+ ((!isNested())?"":( "\tsuper = " + superblock().name() + ";\n" ))
			+ "}\n";
	}
	public String toStringConfigAll() {
		String erg = toStringConfig();
		for( MemoryBlock mb : subblocks() ) {
			erg += mb.toStringConfigAll();
		}
		return erg;
	}

	/** liest man in diesem block
		- unabhaenging von offset und endianess -
		immer einen konstanten wert?
	*/
	public boolean hasConstInt() {
		return constOnRead() && m_cbs.length == 1;
	}

	/** siehe hasConstInt */
	public long readConstInt( int size ) {
		long val = m_cbs[0] & 0xffl;
		if( size > 1 ) val |= val << 8;
		if( size > 3 ) val |= val << 16;
		if( size > 7 ) val |= val << 32;
		return val;
	}

	/** liest aus einem 'constOnRead'-Block */
	public boolean readConstString( long offs, final byte[] buf ) {
		int offset = (int)(offs - origin());
		if( null == m_cbs || offset + buf.length > m_cbs.length ) {
			return false;
		}
		if( hasConstInt() ) {
			for( int i = 0; i < buf.length; ++i ) buf[i] = m_cbs[0];
		} else {
			for( int i = 0; i < buf.length; ++i ) {
				buf[i] = m_cbs[offset + i];
			}
		}
		return true;
	}

	/** */
	public boolean setConsts( final byte[] mem ) { //TODO DEAD
		if( !constOnRead() ) return false;
		if( !hasInterval() ) return false;
		if( mem.length > 1 && mem.length != length() ) return false;
		m_cbs = mem;
		return true;
	}
	public void justSetConsts( final byte[] mem ) {
		m_cbs = mem;
	}
	public boolean checkConsts() {
		//System.err.println("checkConsts"); //TODO DEBUG
		if( m_cbs == null || !constOnRead() ) return true;
		//System.err.println("has const bytes: " + m_cbs.length ); //TODO DEBUG
		//System.err.println("is const on read"); //TODO DEBUG
		if( !hasInterval() ) return false;
		//System.err.println("has interval"); //TODO DEBUG
		if( m_cbs.length > 1 && m_cbs.length != length() ) return false;
		//System.err.println("checkConsts success"); //TODO DEBUG
		return true;
	}

	/**
		sucht nach einem Sub-Block, der die angegebene Adresse beinhaltet
		und gibt diesen zurueck (oder sich selbst).
	*/
	public MemoryBlock getBlock( long address, int size ) {
		//TODO O(n) ... //TODO size...
		if( !includes( address, size ) ) return null;
		for( MemoryBlock mb : m_sub ) {
			MemoryBlock sub = mb.getBlock(address,size);
			if( sub != null ) return sub;
		}
		return this;
	}
	public MemoryBlock getBlock( final String name ) {
		//TODO O(n) ... //TODO size...
		if( name().equals( name ) ) return this;
		for( MemoryBlock mb : m_sub ) {
			MemoryBlock sub = mb.getBlock(name);
			if( sub != null ) return sub;
		}
		return null;
	}

	public void add( MemoryBlock mb ) { m_sub.add( mb ); }

	public List<MemoryBlock> subblocks() {
		return m_sub;
	}

	public MemoryBlock superblock() { return m_superblock; }
	public boolean isNested() { return null != superblock(); }
	public void setSuperblock( MemoryBlock superblock ) { m_superblock = superblock; }

	public void checkIntervals() throws RuntimeException {
		final int min_length = 4;
		if( !hasInterval() ) throw new RuntimeException("memoryblock \"" + name() + "\" has no interval." );
		for( MemoryBlock sub : subblocks() ) {
			sub.checkIntervals();
		}
		for( int i = 0; i < subblocks().size(); ++i ) {
			final MemoryBlock sub = subblocks().get(i);
			if( !includes( sub ) ) {
				throw new RuntimeException("memoryblock \"" + name() + "\" doesn't include its subblock \"" + sub.name() + "\"." );
			}
			//if( sub.length() < min_length ) { //besser in includes/getBlock testen?
			//	throw new RuntimeException("memoryblock \"" + name() + "\" is too small, must be at least " + min_length + " bytes long." );
			//}
			//TODO O(n**2)
			for( int j = i+1; j < subblocks().size(); ++j ) {
				final MemoryBlock sub2 = subblocks().get(j);
				if( sub.interval().intersects( sub2.interval() ) ) {
					throw new RuntimeException("memoryblock \"" + sub.name() + "\" intersects with block \"" + sub2.name() + "\"." );
				}
			}
		}
	}


	private final String	m_name;
	private int				m_attr;
	private	Interval		m_interval;
	private int				m_onr, m_onw;
	private byte[]			m_cbs;
	private final ArrayList<MemoryBlock>	m_sub;
	private MemoryBlock		m_superblock;

	public static int parseOnReadDesc( String s ) {
		s = s.toLowerCase();
		if( s.equals( STR_READ_NORMAL ) ) return READ_NORMAL;
		if( s.equals( STR_READ_CONST ) ) return READ_CONST;
		if( s.equals( STR_READ_TRAP ) ) return READ_TRAP;
		if( s.equals( STR_READ_UNSPEC ) ) return READ_UNSPEC;
		throw new IllegalArgumentException("unknown read behavior description: \"" + s + "\"." );
	}

	public static int parseOnWriteDesc( String s ) {
		s = s.toLowerCase();
		if( s.equals( STR_WRITE_NORMAL ) ) return WRITE_NORMAL;
		if( s.equals( STR_WRITE_IGNORE ) ) return WRITE_IGNORE;
		if( s.equals( STR_WRITE_TRAP ) ) return WRITE_TRAP;
		if( s.equals( STR_WRITE_UNSPEC ) ) return WRITE_UNSPEC;
		throw new IllegalArgumentException("unknown write behavior description: \"" + s + "\"." );
	}

	public static int parseAttributes( String s ) {
		if( s.equals("") ) return DEFAULT_ATTRIB;
		if( s.charAt(0) == '(' ) {
			s = s.substring( 1, s.length() - 1 );
		}
		if( s.equals("") ) return DEFAULT_ATTRIB;
		s = s.toLowerCase();
		boolean inv = false;
		int flags = 0;
		for( int i = 0; i < s.length(); ++i ) {
			char c = s.charAt(i);
			if( !inv ) {
				switch( c ) {
				case 'x': flags |= X; break;
				case 'w': flags |= W; break;
				case 'r': flags |= R; break;
				case 'a': flags |= A; break;
				case 'i':
				case 'l': flags |= I; break;
				case 'p': flags |= P; break;
				case '!': inv = true; break;
				}
			} else {
				switch( c ) {
				case 'x': flags &= ~X; flags |= IX; break;
				case 'w': flags &= ~W; flags |= IW; break;
				case 'r': flags &= ~R; flags |= IR; break;
				case 'a': flags &= ~A; flags |= IA; break;
				case 'i':
				case 'l': flags &= ~I; flags |= II; break;
				case 'p': flags &= ~P; flags |= IP; break;
				}
			}
		}
		return flags;
	}


};
