/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/*
	M.Strotz 2011
*/

package keso.compiler.backend.gradmp;

import java.util.*;
import java.io.*;

import keso.compiler.*;


public final class MemoryDescription
{

	public static final String VOID_BLOCK_NAME = "__void_block";


	// Default-Block
	private final MemoryBlock m_void_block;
	//keso.compiler.BuilderOptions m_bopt;
	// Blockliste
	private final MemoryBlock m_bnull;

	private String m_linkerscript_path;

	private boolean m_has_unspec_read_block;

	// ----

	public MemoryDescription( MemoryBlock void_block ) {
		m_void_block = void_block;
		m_bnull = new MemoryBlock( MemoryBlock.READ_UNSPEC, MemoryBlock.WRITE_UNSPEC, "", 0, Long.MAX_VALUE );
	}

	public MemoryDescription() {
		this( new MemoryBlock( MemoryBlock.READ_UNSPEC, MemoryBlock.WRITE_UNSPEC, VOID_BLOCK_NAME, 0, Long.MAX_VALUE ) );
	}

	public String getDescriptionString() {
		String erg =
			"address-width: " + getAddressWidth()
			+ "\nendian: " + (isLittleEndian()?"little":"big")
			+ "\nBlocks:\n" + m_void_block + "\n";
		for( MemoryBlock mb : blocks() ) {
			erg += mb.toStringDesc("") + "\n";
		}
		return erg;
	}
	public String getConfigString() {
		String erg =
			"MemoryDescription "
			+ ((hasLinkerscriptPath())?( "(\"" + getLinkerscriptPath() + "\")" ):"")
			+ "{\n";
		for( MemoryBlock b : blocks() ) {
			erg += "\n" + b.toStringConfigAll();
		}
		return erg + "}\n";
	}

	// ---- Blocks

	public Collection<MemoryBlock> blocks() {
		return m_bnull.subblocks();
	}
	public MemoryBlock add( MemoryBlock b ) {
		blocks().add(b);
		return b;
	}
	public MemoryBlock getBlock( long address_param, int size ) {
		long address = toAddress( address_param );
		MemoryBlock erg = m_bnull.getBlock(address,size);
		if( erg == m_bnull ) return m_void_block;
		else return erg;
	}
	public MemoryBlock getBlock( final String name ) {
		//TODO O(n) ...
		if( m_void_block.name().equals( name ) ) return m_void_block;
		return m_bnull.getBlock(name);
	}
	public long readConstant( MemoryBlock mb, long offset, int size ) {
		if( mb.hasConstInt() ) {
			return mb.readConstInt( size );
		} else {
			final byte[] buf = new byte[size];
			mb.readConstString( offset, buf );
			return makeLong(buf);
		}
	}


	// ---- Merge

	public HashMap<MemoryBlock,MemoryBlock> startMerge( MemoryDescription md ) {
		HashMap<MemoryBlock,MemoryBlock> map = new HashMap<MemoryBlock,MemoryBlock>();
		LinkedList<MemoryBlock> to_add = new LinkedList<MemoryBlock>();
		for( MemoryBlock other : blocks() ) {
			MemoryBlock mine = getBlock( other.name() );
			if( mine == null ) {
				to_add.add( other );
			} else {
				map.put( mine, other );
			}
		}
		blocks().addAll( to_add );
		return map;
	}

	public static void getSubBlocks( Map<MemoryBlock,MemoryBlock> map ) {
		for( Map.Entry<MemoryBlock, MemoryBlock> e : map.entrySet() ) {
			MemoryBlock
				mine = e.getKey(),
				other = e.getValue();
			mine.subblocks().addAll( other.subblocks() );
		}
	}

	public static void getIntervals( Map<MemoryBlock,MemoryBlock> map ) {
		for( Map.Entry<MemoryBlock, MemoryBlock> e : map.entrySet() ) {
			MemoryBlock
				mine = e.getKey(),
				other = e.getValue();
			if( mine.hasInterval() ) {
				if( other.hasInterval() ) {
					if( !mine.interval().equals( other.interval() ) ) {
						throw new RuntimeException("contradicting intervals for memoryblock \"" + mine.name() + "\"");
					}
				}
			} else {
				if( other.hasInterval() ) {
					mine.setInterval( other.interval() );
				}
			}
		}
	}

	public static void getAttributesFromLinkerscript( Map<MemoryBlock,MemoryBlock> map ) {
		for( Map.Entry<MemoryBlock, MemoryBlock> e : map.entrySet() ) {
			MemoryBlock
				mine = e.getKey(),
				other = e.getValue();
			mine.setAttributes( other.attributes() );
		}
	}

	public static void getRWBehaviorFromConfig( Map<MemoryBlock,MemoryBlock> map ) {
		for( Map.Entry<MemoryBlock, MemoryBlock> e : map.entrySet() ) {
			MemoryBlock
				mine = e.getKey(),
				other = e.getValue();
			mine.setOnRead( other.onRead() );
			mine.setOnWrite( other.onWrite() );
		}
	}

	public void correctRWBehavior() {
		for( MemoryBlock mb : blocks() ) {
			mb.correctRWBehavior_rec();
		}
	}

	public void checkIntervals() {
		m_bnull.checkIntervals();
	}

	public void mergeWithLinkerscript( MemoryDescription md ) {
		final Map<MemoryBlock,MemoryBlock>  map = startMerge( md );
		getIntervals( map );
		checkIntervals();
		getAttributesFromLinkerscript( map );
		correctRWBehavior();
	}
	public void mergeWithConfig( MemoryDescription md ) {
		final Map<MemoryBlock,MemoryBlock> map = startMerge( md );
		getSubBlocks( map );
		getIntervals( map );
		checkIntervals();
		getRWBehaviorFromConfig( map );
		correctRWBehavior();
	}

	public void correctBlocks() {
		checkIntervals();
		correctRWBehavior();
	}

	public void setup() {
		correctBlocks(); //TODO abgespeckte variante...
		setup_unspec_read();
	}

	// ---- Parse Config

	public void setProperty( MemoryBlock b, String id, String rhs, List<Integer> iarray ) { //TODO DEAD
		id = id.toLowerCase();
		boolean illegal_array = false;

		if( id.equals("origin") || id.equals("org") ) {
			b.setOrigin( Long.parseLong(rhs) );
			illegal_array = true;

		} else if( id.equals("length") || id.equals("len") ) {
			b.setLength( Long.parseLong(rhs) );
			illegal_array = true;

		} else if( id.equals("read") ) {
			b.setOnRead( MemoryBlock.parseOnReadDesc( strip_quotes(rhs) ) );
			illegal_array = false;
			if( iarray != null ) {
				final byte[] mem = new byte[ iarray.size() ];
				for( int i = 0; i < mem.length; ++i ) {
					int var = iarray.get(i);
					//if( var < -128 || 255 < var ) //schade...
					mem[i] = (byte)var;
				}
				if( !b.setConsts( mem ) ) {
					throw new IllegalArgumentException("unable to set constant memory for block \"" + b.name() + "\".");
				}
			}

		} else if( id.equals("super") ) {
			MemoryBlock super_block = getBlock(rhs);
			if( null == super_block ) {
				throw new IllegalArgumentException("unknown block: \"" + rhs + "\".");
			}
			super_block.add(b);
			b.setSuperblock( super_block );
			//check super_block.includes(b) <-- checkIntervals...

		} else if( id.equals("write") ) {
			b.setOnWrite( MemoryBlock.parseOnWriteDesc( strip_quotes(rhs) ) );
			illegal_array = true;

		} else {
			throw new IllegalArgumentException("memory-blocks have no property called \"" + id + "\".");
		}

		if( illegal_array && null != iarray ) {
			throw new IllegalArgumentException("cannot assign arrays to \"" + id + "\".");
		}
	}

	public void setProperties( MemoryBlock memblock, keso.compiler.config.Set<?> set) {
		Collection<keso.compiler.config.Attribute<?>> attribs = set.getAllAttributes();
		for( keso.compiler.config.Attribute<?> attr : attribs ) {
			setProperty(memblock,attr);
		}
		if( !memblock.checkConsts() ) {
			throw new IllegalArgumentException("unable to set constant memory for block \"" + memblock.name() + "\".");
		}
	}

	public void setProperty( MemoryBlock b, keso.compiler.config.Attribute<?> attr) {
		String id = attr.ident.toLowerCase();
		if( id.equals("origin") || id.equals("org") ) {
			b.setOrigin( attr.valueInt() & 0xffffffffL );
		} else if( id.equals("length") || id.equals("len") ) {
			b.setLength( attr.valueInt() & 0xffffffffL );
		} else if( id.equals("read") ) {
			if( attr instanceof keso.compiler.config.ArrayAttr<?> ) {
				b.setOnRead( MemoryBlock.READ_CONST );
				keso.compiler.config.ArrayAttr<?> arr = (keso.compiler.config.ArrayAttr<?>) attr;
				List<String> sarray = arr.valueArray();
				final byte[] mem = new byte[ sarray.size() ];
				for( int i = 0; i < mem.length; ++i ) {
					int var = Integer.parseInt( sarray.get(i) );
					//if( var < -128 || 255 < var ) //schade...
					mem[i] = (byte)var;
				}
				b.justSetConsts( mem );
			} else {
				b.setOnRead( MemoryBlock.parseOnReadDesc( attr.valueString() ) );
			}
		} else if( id.equals("super") ) {
			MemoryBlock super_block = getBlock( attr.valueString() );
			if( null == super_block ) {
				throw new IllegalArgumentException("unknown block: \"" + attr.valueString() + "\".");
			}
			super_block.add(b);
			b.setSuperblock( super_block );
			//check super_block.includes(b) <-- checkIntervals...
		} else if( id.equals("write") ) {
			b.setOnWrite( MemoryBlock.parseOnWriteDesc( attr.valueString() ) );
		} else {
			throw new IllegalArgumentException("memory-blocks have no property called \"" + id + "\".");
		}
	}


	// ---- LinkerScriptPath

	public boolean hasLinkerscriptPath() {
		return null != m_linkerscript_path;
	}
	public String getLinkerscriptPath() {
		return m_linkerscript_path;
	}
	public void setLinkerscriptPath( String s ) {
		m_linkerscript_path = strip_quotes(s);
	}

	// ---- Parse

	public void parseFile( String path ) throws FileNotFoundException {
		blocks().clear();
		continueParsingFile(path);
	}
	public void continueParsingFile( String path ) throws FileNotFoundException {
		//Dateiendungen unterscheiden...
		continueParsingLinkerscriptFile(path);
	}

	public void parseLinkerscript( Reader reader, String name ) {
		blocks().clear();
		continueParsingLinkerscript( reader, name );
	}
	public void parseLinkerscriptFile( String path ) throws FileNotFoundException {
		BufferedReader in = new BufferedReader(new FileReader(path));
		parseLinkerscript( in, path );
	}
	public String parseLinkerscriptFile( List<String> paths, String name ) throws FileNotFoundException {
		for( String path : paths ) {
			try {
				String script = path + "/" + name;
				parseLinkerscriptFile( script );
				return script;
			} catch( FileNotFoundException fnfe ) {
			}
		}
		throw new FileNotFoundException("could not open \"" + name + "\" anywhere.");
	}
	public String parseLinkerscriptFile( List<String> paths ) throws FileNotFoundException {
		if( !hasLinkerscriptPath() ) return "";
		return parseLinkerscriptFile( paths, getLinkerscriptPath() );
	}
	/*public void parseLinkerscriptFile() throws FileNotFoundException {
		parseLinkerscriptFile( bopts().getLinkerscriptPaths() );
	}*/

	public void continueParsingLinkerscript( Reader reader, String name ) {
		keso.compiler.backend.gradmp.ldparser.Parser parser = new keso.compiler.backend.gradmp.ldparser.Parser( reader );
		parser.start(name); //catch...
		keso.compiler.backend.gradmp.ldparser.ast.LinkerScript ls = (keso.compiler.backend.gradmp.ldparser.ast.LinkerScript) parser.getRoot();
		ls.addMemoryBlocksTo( blocks() );
	}
	public void continueParsingLinkerscriptFile( String path ) throws FileNotFoundException {
		BufferedReader in = new BufferedReader(new FileReader(path));
		continueParsingLinkerscript( in, path );
	}

	public String parseAndMergeLinkerscript( List<String> paths ) throws FileNotFoundException {
		final MemoryDescription left = new MemoryDescription( m_void_block );
		left.setLinkerscriptPath( getLinkerscriptPath() );
		final String script = left.parseLinkerscriptFile( paths );
		mergeWithLinkerscript( left );
		return script;
	}

	// ---- Adressbreite

	public int getAddressSize() {
		return bopts().targetAddrSize();
	}
	public int getAddressWidth() {
		return getAddressSize() * 8;
	}
	public long getAddressMask() { return (1l << getAddressWidth()) - 1l; }
	public long toAddress( long x ) { return x & getAddressMask(); }

	// ---- Endianness

	public boolean isLittleEndian() {
		return bopts().hasLittleEndianMemory();
	}
	public boolean isMSBF() {
		return !isLittleEndian();
	}
	public long makeLong( final byte[] buf ) {
		long value = 0;
		if( isMSBF() ) {
			for( int i = 0; i < buf.length; ++i ) {
				value |= (buf[i] & 0xff) << ((buf.length-1-i)*8);
			}
		} else {
			for( int i = 0; i < buf.length; ++i ) {
				value |= (buf[i] & 0xff) << (i*8);
			}
		}
		return value;
	}

	// ---- unspec-read --> read has global impact
	private void setup_unspec_read() {
		if(m_void_block.unspecOnRead()) {
			m_has_unspec_read_block = true;
			return;
		}
		for(MemoryBlock block : blocks()) {
			if(block.unspecOnRead()) {
				m_has_unspec_read_block = true;
				break;
			}
		}
	}
	public boolean hasUnspecReadBlock() {
		return m_has_unspec_read_block;
	}

	// ----

	private BuilderOptions bopts() {
		return BuilderOptions.getOpts();
	}

	// ----

	private static String strip_quotes( final String s ) {
		if( s !=null && s.length() > 1 && s.charAt(0) == '\"' ) {
			return s.substring(1, s.length() - 1);
		} else {
			return s;
		}
	}

};
