/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.task.*;
import keso.compiler.config.parser.ParseException;
import keso.compiler.backend.CoderOSEK;

/**
 * Differs from ComplexAttribute in that it handles some attributes
 * of the complex Alarm-Action attribute specially when outputting OIL.
 */
public class AlarmActionAttribute extends ComplexAttribute<AlarmDefinition> {
	public AlarmActionAttribute(AlarmDefinition parent, String name, String setting, int line) throws ParseException {
		super(parent, name, setting, line);
	}

	private void checkConfig() throws CompileException {
		if (value.compareToIgnoreCase("ALARMCALLBACK")==0) {
			// TODO specify function in attribute
		}

		if (value.compareToIgnoreCase("ACTIVATETASK")==0 ||
				value.compareToIgnoreCase("SETEVENT")==0) {
			// need TASK attribute
			Attribute<?> task = getAttribute("UseTask");
			if(task == null) throw new CompileException("Alarm "+parent.getIdentifier()+"'s action needs the UseTask attribute specified.");
			TaskDefinition tdef = parent.parent.getTask(task.valueString());
			if(tdef == null)
				throw new CompileException("Alarm "+getIdentifier()+"'s UseTask specifies a Task not known in domain "+
						parent.parent.getIdentifier());
		}

		if (value.compareToIgnoreCase("SETEVENT")==0) {
			// need EVENT attribute
			Attribute<?> ev = getAttribute("UseEvent");
			if (ev == null) throw new CompileException("Alarm "+parent.getIdentifier()+"'s action needs the UseEvent attribute specified.");

			EventDefinition edef = parent.parent.getSystem().getEvent(ev.valueString());
			if(edef==null)
				throw new CompileException("Alarm "+getIdentifier()+"'s UseEvent specifies an unknown Event " + ev.valueString());
		}
	}

	/**
	 * Return an OIL representation of this AlarmAction attribute.
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {
		checkConfig();

		oil.append(prefix);
		oil.append(ident.toUpperCase());
		oil.append(" = ");
		oil.append(value.toUpperCase());

		// Dump the subattributes
		oil.append(" {\n");

		if (value.compareToIgnoreCase("ALARMCALLBACK")==0) {
			// need Alarmcallbackname
			oil.append(prefix+"\tALARMCALLBACKNAME = ");
			oil.append(parent.getIdentifier());
			oil.append("_callback;\n");
		}

		if (value.compareToIgnoreCase("ACTIVATETASK")==0 ||
				value.compareToIgnoreCase("SETEVENT")==0) {
			// need TASK attribute
			Attribute<?> task = getAttribute("UseTask");
			TaskDefinition tdef = parent.parent.getTask(task.valueString());
			oil.append(prefix);
			oil.append("\tTASK = ");
			oil.append(tdef.getIdentifier());
			oil.append(";\n");
		}

		if (value.compareToIgnoreCase("SETEVENT")==0) {
			// need EVENT attribute
			Attribute<?> ev = getAttribute("UseEvent");
			EventDefinition edef = parent.parent.getSystem().getEvent(ev.valueString());
			oil.append(prefix);
			oil.append("\tEVENT = ");
			oil.append(edef.getIdentifier());
			oil.append(";\n");
		}

		oil.append(prefix);
		oil.append("};\n");
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		checkConfig();

		String action=null, task=null, event=null;

		if (value.compareToIgnoreCase("ACTIVATETASK")==0 || value.compareToIgnoreCase("SETEVENT")==0) {
			action = "activatetask=\"true\"";

			Attribute<?> taskn = getAttribute("UseTask");
			TaskDefinition tdef = parent.parent.getTask(taskn.valueString());
			task = tdef.getIdentifier();
		}

		if (value.compareToIgnoreCase("SETEVENT")==0) {
			action = "setevent=\"true\"";

			Attribute<?> ev = getAttribute("UseEvent");
			EventDefinition edef = parent.parent.getSystem().getEvent(ev.valueString());
			event = edef.getIdentifier();
		}

		buf.append(prefix + action);
		if(task!=null) buf.append(prefix + "task=\"" + task +"\"");
		if(event!=null) buf.append(prefix + "event=\"" + event +"\"");
	}
}

