/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.parser.ParseException;
import keso.compiler.backend.CoderOSEK;

import java.util.*;

public class AlarmDefinition extends SystemObject<DomainDefinition<?>> {
	/* used as index into keso_alarm_index to
	 * reference the entry associated to an invalid
	 * domain lookup (null)
	 *
	 * Stable after config file has been parsed completely.
	 */
	public static byte INVALID_ALARM=0;

	private AlarmActionAttribute action;
	private ComplexBoolAttribute autoStart;

	public AlarmDefinition(DomainDefinition<?> dom, String name, int line) {
		super(dom, name, line);
		INVALID_ALARM++;
	}

	public ComplexBoolAttribute getAutoStart() {
		return this.autoStart;
	}

	public void setComplex(ComplexAttribute cattr) throws ParseException {
		if((cattr instanceof ComplexBoolAttribute) && (cattr.ident.compareToIgnoreCase("Autostart")==0)) {
			if (autoStart!=null) throw new ParseException("Multiple definition of Autostart in Alarm " + getIdentifier());
			autoStart = (ComplexBoolAttribute) cattr;
		} else if (cattr instanceof AlarmActionAttribute && cattr.ident.compareToIgnoreCase("Action")==0) {
			if (action!=null) throw new ParseException("Multiple definition of Action in Alarm " + getIdentifier());
			action = (AlarmActionAttribute) cattr;
		} else throw new ParseException("Unknown ComplexAttribute " + cattr.ident + " in Alarm " + getIdentifier());
	}

	/**
	 * Generate a ALARM block in OIL syntax out of this ALARM Definition.
	 *
	 * Known and supported attributes
	 * OIL standard:
	 *   COUNTER    (Reference to a defined counter)
	 *   AUTOSTART  (ComplexBoolAttribute)
	 *     - ALARMTIME (time when alarm shall expire first)
	 *     - CYCLETIME (cycletime of a cyclic alarm)
	 *     - APPMODE   (Multiref to Appmodes where autostart applies)
	 *   ACTION     (ComplexAttribute{ALARMCALLBACK, SETEVENT, ACTIVATETASK})
	 *     - TASK   (Reference to a task)
	 *     - EVENT  (Reference to an event)
	 *     - ALARMCALLBACKCLASS (class with callback method)
	 *     - ALARMCALLBACKMETHOD (name+signature of callback method)
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {

		Attribute<?> counter = getAttribute("UseCounter");
		if (counter == null)
			throw new CompileException("Alarm "+getIdentifier()+" does not have an associated counter.");

		SystemDefinition  sdef = parent.getSystem();
		CounterDefinition cdef = sdef.getCounter(counter.valueString());

		if (cdef == null)
			throw new CompileException("Alarm "+getIdentifier()+"'s associated counter "+counter.valueString()+" is now known to the system.");

		oil.append(prefix+"ALARM " + getIdentifier() + " {\n");
		oil.append(prefix+"\t");
		oil.append("COUNTER = ");
		oil.append(cdef.getIdentifier());
		oil.append(";\n");

		// Autostart
		autoStart.toOIL(oil, true, prefix+"\t", "ALARM_AUTOSTART", coder);

		// Action
		action.toOIL(oil, prefix+"\t", coder);

		oil.append(prefix+"};\n\n");
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		Map<String, String> osekIncompatAttributes = new HashMap<String, String>();
		osekIncompatAttributes.put("autostart","armed");
		osekIncompatAttributes.put("alarmtime","absolutetime");
		osekIncompatAttributes.put("usetask","task");
		osekIncompatAttributes.put("useevent","event");

		Attribute<?> counter = getAttribute("UseCounter");
		if (counter == null)
			throw new CompileException("Alarm "+getIdentifier()+" does not have an associated counter.");

		SystemDefinition  sdef = parent.getSystem();
		CounterDefinition cdef = sdef.getCounter(counter.valueString());
		if (cdef == null)
			throw new CompileException("Alarm "+getIdentifier()+"'s associated counter "+counter.valueString()+" is not known to the system.");

		buf.append(prefix+"<alarm name=\"" + getIdentifier() + "\" counterref=\"");
		buf.append(cdef.getIdentifier());
		buf.append("\" ");

		// Autostart
		buf.append("armed=\"");
		buf.append(autoStart.setting?"true":"false");
		buf.append("\" ");
		autoStart.toCiAOConfig(buf," ",osekIncompatAttributes);

		// Action
		action.toCiAOConfig(buf, " ");

		buf.append(" />\n\n");
	}

	public ComplexAttribute<?> getAction() { return action; }

	protected String typeID() { return "a_"; }

	public String getResourceClass() { return "keso/core/Alarm"; }

	protected void finalizeCfg(FinalizeResult res) throws ParseException {
		res.requireMethod("keso/core/Alarm", null);

		// if callback action require callback method, too
		if(action.value.compareToIgnoreCase("ALARMCALLBACK")==0) {
			Attribute<?> classAttr = action.getAttribute("AlarmCallbackClass");
			Attribute<?> methodAttr = action.getAttribute("AlarmCallbackMethod");

			if(methodAttr==null || classAttr==null)
				throw new ParseException("Alarm " + getIdentifier() + " needs AlarmCallbackClass and AlarmCallbackMethod specified");

			res.requireMethod(classAttr.valueString(), methodAttr.valueString());
		}
	}

	public DomainDefinition<?> getDomain() {
		return parent;
	}
}
