/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.parser.*;
import keso.compiler.config.task.*;
import keso.compiler.backend.CoderOSEK;

import java.lang.Integer;
import java.util.*;

public class EventDefinition extends Set<SystemDefinition> {
	private static int numEvents;

	/** Mask of this event. A power of 2 */
	private int mask;

	private boolean unused = false;

	/** Contains TaskDefinitions of all owner Tasks.
	 * Used to find a mask
	 */
	private Collection<TaskDefinition> ownerTasks;

	public static int numEvents() { return numEvents; }

	public EventDefinition(SystemDefinition parent, String name, int line) {
		super(parent, name, line);
		ownerTasks = new ArrayList<TaskDefinition>();
		numEvents++;
	}

	public EventDefinition(SystemDefinition parent, String name) {
		super(parent, name, -1);
		ownerTasks = new ArrayList<TaskDefinition>();
		numEvents++;
	}

	public void addAttribute(String ident, Attribute value) throws ParseException {
		super.addAttribute(ident, value);

		if (ident.compareToIgnoreCase("IntendedUnused")==0) {
			BuilderOptions.getOpts().vwarn("Event \""+ getIdentifier() + "\" is marked as unused! Won't create OIL entry.");
			unused = true;
			return;
		}

		// AUTO masks need to be calculated in the finalizer when
		// all owning tasks are known
		if(ident.compareToIgnoreCase("Mask")!=0) return;
		String maskstr = value.valueString();
		if(maskstr.compareToIgnoreCase("AUTO")==0) return;

		try {
			if (maskstr.startsWith("0x")) {
				mask = Integer.parseInt(maskstr.substring(2), 16);
			} else {
				mask = Integer.parseInt(maskstr);
			}
		} catch (java.lang.NumberFormatException e) {
			mask = 0;
			BuilderOptions.getOpts().verbose("Event " + getIdentifier() + "'s mask cannot be parsed. AUTO assigned.");
			return;
		}

		if (bitCount(mask)!=1) {
			BuilderOptions.getOpts().verbose("Event " + getIdentifier() + "'s mask is not a power of 2. AUTO assigned.");
			mask = 0;
		}
	}

	/**
	 * Determine the number of bits set in an int.
	 *
	 * @param i is the Integer whose bits will be checked
	 * @return the number of bits set in i
	 */
	private static int bitCount(int i) {
		int cnt=0;

		do {
			cnt += i&1;
			i >>>= 1;
		} while (i!=0);

		return cnt;
	}

	public String toString() { return "Event:"+ ident; }

	/**
	 * Generate a EVENT block in OIL syntax out of this EVENT Definition.
	 *
	 * May be used only if already finalized
	 *
	 * Known and supported attributes
	 * OIL standard:
	 *   MASK		(UINT64)
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {

		if (unused) return;

		if (mask==0)
			throw new CompileException("Mask not set for Event "+getIdentifier()+". Finalize first!");

		oil.append(prefix+"EVENT " + getIdentifier() + " {\n");
		oil.append(prefix+"\tMASK = 0x");
		oil.append(Integer.toHexString(mask));
		oil.append(";\n");
		oil.append(prefix+"};\n\n");
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {

		if (unused) return;

		if (mask==0) throw new CompileException("Mask not set for Event "+getIdentifier()+". Finalize first!");

		buf.append(prefix+"<event name=\"");
		buf.append(getIdentifier());
		buf.append("\" mask=\"0x");
		buf.append(Integer.toHexString(mask));
		buf.append("\" />\n");
	}


	public String getIdentifier() {
		return ident;
	}

	protected void finalizeCfg(FinalizeResult res) throws keso.compiler.config.parser.ParseException {
		// find out mask, all owning tasks must have registered themselves
		if (mask == 0) {
			int combinedMask=0;
			for(TaskDefinition task: ownerTasks)
				combinedMask |= task.getEvBits();

			// now find the least significant cleared bit
			for(int i=0;i<32;i++) {
				mask = 1<<i;
				if (0== (combinedMask & mask)) break;
			}
		}

		// mark mask's bit used at all owner tasks
		for(TaskDefinition task: ownerTasks)
			task.registerEvMask(mask);
	}

	public int getMask() throws CompileException {
		if(mask==0) throw new CompileException("getMask(): finalize first");
		return mask;
	}

	protected String typeID() { return "e_"; }

	public void registerTask(TaskDefinition task) {
		ownerTasks.add(task);
	}
}
