/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.parser.*;
import keso.compiler.config.messageport.*;
import keso.compiler.config.domain.*;

import java.io.*;

import java.util.*;

/**
 * Executable Class that will generate parts of the Keso java
 * libs with information from kesorc.
 */
public class GenerateKesoClasses {

	private static SystemDefinition sdef;

	static abstract class Template {
		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			generateSource(basedir, templatedir, srcdir, this, pkgdir + "keso/core");
		}

		abstract void genCode(Writer srcfile) throws Exception;
		abstract String getGeneratedClassName();
		abstract String getTemplateName();
		abstract boolean isIdentifierLine(String line);

		/**
		 * Default: Check if generated file is newer than KESORC.
		 */
		boolean upToDate(File gen_file) {
			if (mtime > 0 && gen_file.exists() && gen_file.lastModified() > mtime) {
				return true;
			}
			return false;
		}
	}

	static class EventsTemplate extends Template {
		String getGeneratedClassName() { return "Events"; }
		String getTemplateName() { return "Events"; }

		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_EVENT_DEFINITIONS*/")==0);
		}

		void genCode(Writer srcfile) throws Exception {
			for (EventDefinition ev : GenerateKesoClasses.sdef.getEvents()) {
				srcfile.write("\tpublic static final int ");
				srcfile.write(ev.ident);
				srcfile.write(" = 0x");
				srcfile.write(java.lang.Integer.toHexString(ev.getMask()));
				srcfile.write(";\n");
			}
		}
	}

	static class AppmodesTemplate extends Template {
		String getGeneratedClassName() { return "Appmodes"; }
		String getTemplateName() { return "Appmodes"; }
		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_APPMODE_DEFINITIONS*/")==0);
		}

		void genCode(Writer srcfile) throws Exception {
			List<AppmodeDefinition> appmodes = GenerateKesoClasses.sdef.getAppmodes();

			for(int i=0; i<appmodes.size(); i++) {
				AppmodeDefinition am = appmodes.get(i);
				if(am.ident.compareToIgnoreCase("OSDEFAULTAPPMODE")!=0) {
					writeIntConstant(srcfile, am.ident, i+1);
				}
			}
		}
	}

	static class CountersTemplate extends Template {
		String getGeneratedClassName() { return "Counters"; }
		String getTemplateName() { return "Counters"; }
		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_COUNTER_DEFINITIONS*/")==0);
		}

		void genCode(Writer srcfile) throws Exception {
			List<CounterDefinition> counters
			= GenerateKesoClasses.sdef.getCounters();

			/* FIXME
			 * I'm not sure how the ProOSEK configurator calculates the values of the
			 * System timer. For now I guess that it's the properties of the first
			 * defined counter or all zero if there is no counter defined
			 */
			int osmaxallowedvalue=0, osticksperbase=0, osmincycle=0, ostickduration=0;

			for(int i=0; i<counters.size(); i++) {
				CounterDefinition cnt = counters.get(i);
				int maxallowedvalue, ticksperbase, mincycle;
				//int tickduration;

				try {
					maxallowedvalue = Integer.parseInt(cnt.getAttribute("MAXALLOWEDVALUE").valueString());
					ticksperbase = Integer.parseInt(cnt.getAttribute("TICKSPERBASE").valueString());
					mincycle = Integer.parseInt(cnt.getAttribute("MINCYCLE").valueString());
					//tickduration = Integer.parseInt(cnt.getAttribute("TIME_IN_NS").valueString());
				} catch (Exception e) {
					throw new CompileException("Counter " + cnt.getIdentifier() + " has a required property not defined");
				}

				if(i==0) {
					osmaxallowedvalue=-maxallowedvalue;
					osticksperbase=ticksperbase;
					osmincycle=mincycle;
					//ostickduration=tickduration;
				}

				writeIntConstant(srcfile, "OSTICKSPERBASE_"+cnt.ident, ticksperbase);
				writeIntConstant(srcfile, "OSMAXALLOWEDVALUE_"+cnt.ident, maxallowedvalue);
				writeIntConstant(srcfile, "OSMINCYCLE_"+cnt.ident, mincycle);
				srcfile.write("\n");
			}
			writeIntConstant(srcfile, "OSTICKSPERBASE", osticksperbase);
			writeIntConstant(srcfile, "OSMAXALLOWEDVALUE", osmaxallowedvalue);
			writeIntConstant(srcfile, "OSMINCYCLE", osmincycle);
			writeIntConstant(srcfile, "OSTICKDURATION", ostickduration);
		}
	}

	static class ReceiverPortTemplate extends Template {
		protected String portName;

		String getGeneratedClassName() { return getFilenamePrefix()+portName; }
		String getTemplateName() { return "MemoryReceiverPort"; }
		static String getFilenamePrefix() { return "MemoryReceiverPort_"; }

		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_CLASSNAME*/")==0);
		}
		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			HashSet<String> keep_list = new HashSet<String>();
			for(DomainDefinition<keso.compiler.config.task.TaskDefinition> jd: sdef.getAllSysDomains()) {
				for(ReceiverPortDefinition rpd: jd.getReceiverPorts()) {
					portName = rpd.ident;
					keep_list.add(getGeneratedClassName() + ".java");
					generateSource(basedir,templatedir,srcdir, this, pkgdir + "ciao/messageports");
				}
			}
			// fake message ports
			for(String fakeMessagePort: sdef.getFakeMessagePorts()) {
				portName = fakeMessagePort;
				keep_list.add(getGeneratedClassName() + ".java");
				generateSource(basedir,templatedir,srcdir, this, pkgdir + "ciao/messageports");
			}
			deleteCoreClasses(basedir+"/"+srcdir+"/"+pkgdir+"/ciao/messageports", ReceiverPortTemplate.getFilenamePrefix(), keep_list);
		}
		void genCode(Writer srcfile) throws Exception {
			srcfile.write(getGeneratedClassName());
		}
	}

	static class SenderPortTemplate extends ReceiverPortTemplate {
		String getGeneratedClassName() { return getFilenamePrefix()+portName; }
		String getTemplateName() { return "MemorySenderPort"; }
		static String getFilenamePrefix() { return "MemorySenderPort_"; }
		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			HashSet<String> keep_list = new HashSet<String>();
			for(DomainDefinition<keso.compiler.config.task.TaskDefinition> jd: sdef.getAllSysDomains()) {
				for(SenderPortDefinition spd: jd.getSenderPorts()) {
					portName = spd.ident;
					keep_list.add(getGeneratedClassName() + ".java");
					generateSource(basedir,templatedir,srcdir, this, pkgdir + "ciao/messageports");
				}
			}
			// fake message ports
			for(String fakeMessagePort: sdef.getFakeMessagePorts()) {
				portName = fakeMessagePort;
				keep_list.add(getGeneratedClassName() + ".java");
				generateSource(basedir,templatedir,srcdir, this, pkgdir + "ciao/messageports");
			}
			deleteCoreClasses(basedir+"/"+srcdir+"/"+pkgdir+"ciao/messageports", SenderPortTemplate.getFilenamePrefix(), keep_list);
		}
	}

	static class CiAOConfigTemplate extends Template {
		String getGeneratedClassName() { return "CiAOConfig"; }
		String getTemplateName() { return "CiAOConfig"; }
		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_CIAOCONFIG_DEFINITIONS*/")==0);
		}

		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			generateSource(basedir, templatedir, srcdir, this, pkgdir + "ciao");
		}

		boolean upToDate(File gen_file) {
			File ciao_config = ciaoConfig();

			if (ciao_config!=null && ciao_config.exists() && gen_file.exists()
					&& gen_file.lastModified() > ciao_config.lastModified()) {
				return true;
			}
			return false;
		}

		private File ciaoConfig() {
			BuilderOptions opts = BuilderOptions.getOpts();

			String ciao_dir = System.getenv("KESOOUTPATH");
			if(ciao_dir == null) {
				opts.warn("$KESOOUTPATH not set, ciao.CiAOConfig will not be generated!");
				return null;
			}

			String app_name = ciao_dir;
			if(	ciao_dir.lastIndexOf('/') >= 0 ) {
				app_name = ciao_dir.substring( ciao_dir.lastIndexOf('/') + 1 );
			}

			File ciao_config = new File(ciao_dir, app_name + ".config");
			return ciao_config;
		}

		void genCode(Writer srcfile) throws Exception {
			BuilderOptions opts = BuilderOptions.getOpts();

			File ciao_config = ciaoConfig();
			if(ciao_config==null || !ciao_config.exists()) {
				opts.warn(ciao_config + " does not exist. ciao.CiAOConfig will not be generated!");
				return;
			}

			BufferedReader cc = new BufferedReader(new FileReader(ciao_config));
			String line;

			while( (line = cc.readLine()) != null) {
				if(!line.startsWith("CONFIG_"))
					continue;

				int index = line.indexOf('=');
				if(index < 0) continue;
				String key,value;
				key = line.substring("CONFIG_".length(), index);
				value = line.substring(index+1);

				if(value.startsWith("0x")) {
					long val = Long.parseLong( value.substring(2), 16 );
					if (val > 0xffffffffL) { // assume that sign does not matter...
						srcfile.write("\tpublic static final long " + key + " = " + value + "L;\n");
					} else {
						srcfile.write("\tpublic static final int " + key + " = " + value + ";\n");
					}

				} else if (value.matches("^\\d+$")) {
					long val = Long.parseLong( value, 10 );
					if (val > 0xffffffffL) { // assume that sign does not matter...
						srcfile.write("\tpublic static final long " + key + " = " + value + "L;\n");
					} else {
						srcfile.write("\tpublic static final int " + key + " = " + value + ";\n");
					}

				} else if (value.matches("^[yn]$")) {
					srcfile.write("\tpublic static final boolean " + key + " = ");
					if(value.equals("y"))
						srcfile.write("true;\n");
					else
						srcfile.write("false;\n");

				} else if (value.matches("^\"\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\"$") ) { // IP-Address, convert to 32-bit integer
					String ipstr = value.substring(1, value.length() -1);
					boolean valid = true;
					int addr = 0;

					for(int i=0; i<32; i+=8) {
						int dindex = ipstr.indexOf('.');
						String oct = ipstr;
						if(dindex > 0) {
							oct = ipstr.substring(0,dindex);
							ipstr = ipstr.substring(dindex+1);
						}
						int oct_i = Integer.parseInt(oct);
						if(oct_i > 255)	{
							opts.warn("CiAOConfig: invalid IP address: " + key + " = " + value);
							valid = false;
							break;
						}
						addr |= oct_i << i;
					}

					if(valid) {
						srcfile.write("\tpublic static final int " + key + " = 0x" + Integer.toHexString(addr) + ";\n");
					}

				} else if (value.matches("^\".*\"$") ) { // String, ignore for now

				} else {
					opts.warn("CiAOConfig: ignoring config option of unknown format: " + line);
				}
			}
		}
	}

	static class SPIComProcessTemplate extends Template {
		String getGeneratedClassName() { return "SPIComProcess"; }
		String getTemplateName() { return "SPIComProcess"; }
		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_MESSAGEPORTS*/")==0);
		}
		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			generateSource(basedir, templatedir, srcdir, this, pkgdir + "spicom");
		}

		void genCode(Writer srcfile) throws Exception {
			for(DomainDefinition<keso.compiler.config.task.TaskDefinition> jd: sdef.getAllSysDomains()) {
				for(ReceiverPortDefinition rpd: jd.getReceiverPorts()) {
					if(rpd.getIdentifier().endsWith("SPICom") == true) {
						String name = rpd.getIdentifier().substring(0, (rpd.getIdentifier().length()-6));
						srcfile.write(
							"			{//" + name + "\n" +
							"			MemoryReceiverPort_" + name + "SPICom rec =  MemoryReceiverPort_" + name + "SPICom.lookup();\n" +
							"			msg = rec.receiveMessage();\n" +
							"			if ( msg != null ) {\n" +
							"				msgReceived = true;\n" +
							"				result.copyFromSHM(msg);\n" +
							"				rec.releaseMessage();\n\n" +
							"				MemorySenderPort_SPICom" + name + " send = MemorySenderPort_SPICom" + name + ".lookup();\n" +
							"				Memory answer_mem = send.allocateMessage();\n" +
							"				if ( answer_mem != null ) {\n" +
							"					answer.copyFrom(result);\n" +
							"					spicom.sendReceive(answer);\n" +
							"					answer.copyToSHM(answer_mem);\n" +
							"					send.sendMessage();\n" +
							"				}\n" +
							"			}}\n");
						srcfile.write("\n");
					}
				}
			}
		}
	}

	static class EthernetTelemetry_collectorTemplate extends Template {
		String getGeneratedClassName() { return "EthernetTelemetry_collector"; }
		String getTemplateName() { return "EthernetTelemetry_collector"; }
		boolean isIdentifierLine(String line) {
			return (line.compareTo("/*KESO_INSERT_MESSAGEPORTS*/")==0);
		}
		void generate(SystemDefinition sdef, String basedir, String templatedir, String srcdir, String pkgdir) throws Exception {
			generateSource(basedir, templatedir, srcdir, this, pkgdir + "ethernet");
		}

		void genCode(Writer srcfile) throws Exception {
			for(DomainDefinition<keso.compiler.config.task.TaskDefinition> jd: sdef.getAllSysDomains()) {
				for(ReceiverPortDefinition rpd: jd.getReceiverPorts()) {
					if(rpd.getIdentifier().endsWith("Ethernet") == true) {
						String name = rpd.getIdentifier().substring(0, (rpd.getIdentifier().length()-"Ethernet".length()));
						srcfile.write(
							"			{//" + name + "\n" +
							"			MemoryReceiverPort_" + name + "Ethernet rec =  MemoryReceiverPort_" + name + "Ethernet.lookup();\n" +
							"			msg = rec.receiveMessage();\n" +
							"			if ( msg != null ) {\n" +
							"				msgReceived = true;\n" +
							"				eth.appendPacket(msg);\n" +
							"				rec.releaseMessage();\n" +
							"			}}\n");
						srcfile.write("\n");
					}
				}
			}
		}
	}

	private static void writeIntConstant(Writer srcfile, String name, int value) throws Exception {
		srcfile.write("\tpublic static final int ");
		srcfile.write(name);
		srcfile.write(" = 0x");
		srcfile.write(Integer.toHexString(value));
		srcfile.write(";\n");
	}

	private static void deleteCoreClasses(final String basedir, final String name, Collection<String> keep_list) throws Exception {
		File dir = new File(basedir);
		File[] files = dir.listFiles(new FilenameFilter() {
			public boolean accept(File dir, String fname) {
				return fname.startsWith(name);
			}
		});

		for(File f: files) {
			if(!keep_list.contains(f.getName())) {
				f.delete();
			}
		}
	}

	// KESORC modification time
	private static long mtime = 0;

	public static void generateModuleClass(WorldDefinition wdef,
			String basedir, String classname, String tmplatedir, String srcdir, String pkgdir,
			boolean samebuild) throws Exception {
		if(samebuild) {
			mtime = wdef.lastModified();
		}
		wdef.rewindSystemDef();
		while ((sdef = wdef.nextSystemDef())!=null) {
			try {
				Template tmplate = (Template) Class.forName("keso.compiler.config.GenerateKesoClasses$" + classname).newInstance();
				tmplate.generate(sdef, basedir, tmplatedir, srcdir, pkgdir);
			} catch(ClassNotFoundException e) {
				System.out.println("No template for class "+classname);
			}
		}
	}

	private static void generateSource(String basedir, String tmplate, String srcdir, Template generator, String pkgdir) throws Exception {
		File gen_class_file = new File(basedir+"/"+srcdir+"/"+pkgdir+"/"+generator.getGeneratedClassName()+".java");

		BuilderOptions opts = BuilderOptions.getOpts();
		if(generator.upToDate(gen_class_file)) {
			opts.verbose("Skip generating "+pkgdir+"/"+generator.getGeneratedClassName() + " (up to date)");
			return;
		}

		BufferedReader rtemplate =
			new BufferedReader(new java.io.FileReader(basedir+"/"+tmplate+"/"+generator.getTemplateName()+".java.template"));

		//BufferedWriter wsrcfile = new BufferedWriter(new java.io.FileWriter(gen_class_file));
		StringWriter tmpClass = new StringWriter();

		while (true) {
			String line = rtemplate.readLine();
			if (line == null) break;

			if(generator.isIdentifierLine(line)) {
				generator.genCode(tmpClass);
			} else {
				tmpClass.write(line);
			}
			//tmpClass.newLine();
			tmpClass.write("\n");
		}
		rtemplate.close();

		String tmpOut = tmpClass.toString();
		if (BuilderOptions.smartFileUpdate(gen_class_file, tmpOut)) {
			System.out.println("Generating "+pkgdir+"/"+generator.getGeneratedClassName());
			PrintStream out = new PrintStream(new FileOutputStream(gen_class_file));
			out.print(tmpOut);
			out.close();
		} else {
			//System.err.println("file '" + gen_class_file + "' is up-to-date.");
		}
	}
}
