/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.parser.ParseException;
import keso.compiler.backend.CoderOSEK;

import java.util.*;

public class ISRDefinition extends Set<DomainDefinition<?>> {
	private static int numISRs;

	private List<Attribute<?>> usedMessages;
	private List<Attribute<?>> usedResources;

	public ISRDefinition(DomainDefinition<?> parent, String name, int line) {
		super(parent, name, line);
		usedMessages = new ArrayList<Attribute<?>>();
		usedResources = new ArrayList<Attribute<?>>();
		numISRs++;
	}

	public static int getNumISRs() { return numISRs; }

	public String toString() { return "ISR:"+ ident; }

	public void addAttribute(String ident, Attribute value) throws ParseException {
	  if(ident.compareToIgnoreCase("UseResource")==0) {
	    usedResources.add(value);
	  } else if(ident.compareToIgnoreCase("UseMessage")==0) {
	    usedMessages.add(value);
	  } else super.addAttribute(ident,value);
	}

	public int getCategory() {
		Attribute<?> cat = getAttribute("category");
		if (cat == null) return 2;
		return Integer.parseInt(cat.valueString());
	}

    public String getHandlerClass() {
        Attribute<?> attr = getAttribute("HandlerClass");
		if (attr==null) return null;
		return attr.valueString();
	}

    public String getHandlerMethod() {
        Attribute<?> attr = getAttribute("HandlerMethod");
		if (attr==null) return null;
		return attr.valueString();
	}

    public String getName() {
		return ident;
	}

	public List<Attribute<?>> getUsedMessages() {
		return usedMessages;
	}

	public List<Attribute<?>> getUsedResources() {
		return usedResources;
	}


	/**
	 * Generate a ISR block in OIL syntax out of this ISR Definition.
	 *
	 * Known and supported attributes
	 * OIL standard:
	 *   CATEGORY         (UINT32)
	 *   RESOURCE         (Reference to an accessed resource)
	 *   MESSAGE          (Reference to a message accessed by this ISR)
	 *
	 * ProOSEK extensions:
	 *   TRICORE_VECTOR   (specifies the SRR the ISR will be attached to)
	 *   TRICORE_IRQLEVEL (UINT32 greater 0)
	 *   TRICORE_NUM_CSA  (UINT32)
	 *   STACKSIZE        (UINT32)
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {
		oil.append(prefix+"ISR " + ident + " {\n");
		super.toOIL(oil, prefix+"\t", "ISR", coder);

		// Resources
		DomainDefinition<?> dom = parent;
		SystemDefinition sysdef = (SystemDefinition) dom.parent;

		for(Attribute<?> a: usedResources) {
			ResourceDefinition res = sysdef.getSysResource(a.valueString());
			if (res == null)
				throw new CompileException("ISR " + getIdentifier() + " references unknown Resource " + a.valueString());

			multiRefToOIL(oil, prefix+"\t", "RESOURCE", res.getIdentifier());
		}

		oil.append(prefix+"};\n\n");
	}

	protected void finalizeCfg(FinalizeResult res) throws ParseException {
		if(parent instanceof JavaDomain) {
			Attribute<?> classAttr = getAttribute("HandlerClass");
			Attribute<?> methodAttr = getAttribute("HandlerMethod");

				if(methodAttr==null || classAttr==null)
					throw new ParseException("ISR " + getIdentifier() + " needs HandlerClass and HandlerMethod specified");

				res.requireMethod(classAttr.valueString(), methodAttr.valueString());
		}
	}

	protected String typeID() { return "i_"; }

	public String getComponent() throws CompileException {
		return ((DomainDefinition<?>) parent).getComponent();
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		buf.append(prefix+"<isr name=\"");
		buf.append(getIdentifier());
		buf.append("\" category=\"");
		buf.append(getCategory());
		buf.append("\" device=\"");
		buf.append(getAttribute("device").valueString());
		buf.append("\" function=\"functionISR");
		buf.append(getIdentifier());
		buf.append("\" />\n");
	}

	public String getIdentifier() {
		return ident;
	}
}
