/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.config.domain.*;
import keso.compiler.imcode.*;

import keso.compiler.*;

public class ImportDefinition extends SystemObject<DomainDefinition<?>> {

	public int identifier;

	public ImportDefinition(JavaDomain dom, String name, int line) {
		super(dom, name, line);
		parent.registerFinalize(this);
	}

	public String getServiceName() {
		return ident;
	}

	public boolean hasNetworkAccess() {
		return getNetwork() != null;
	}

	public boolean isLocal() {
		return getService().getSystem() == parent.parent;
	}

	public boolean needDataConversion() {
		return parent.parent.hasLittleEndianMemory() != getService().getSystem().hasLittleEndianMemory();
	}

	public String getDriver() {
		Attribute<?> attr = getAttribute("Driver");
		if (attr == null) {
			return null;
		}
		return attr.valueString();
	}

	public void checkClasses(ClassStore repository) throws CompileException {
		if (!isLocal()) {
			IMClass driverIface = repository.getClass(getNetwork().getDriverInterfaceName());
			IMClass driverClass = repository.getClass(getDriver());

			if (driverClass == null) {
				throw new CompileException("The Driver " + getDriver() + " specified for import " + this + " was not found.");
			}

			if (!driverClass.implementsInterface(driverIface)) {
				throw new CompileException("The Driver " + getDriver() + " specified for import " + this + " doesn't implement the specified DriverInterface.");
			}
		}
	}


	public int getCallIdentifier(String methodNameAndType) {
		return ServiceManager.instance.getCallIdentifier(this, methodNameAndType);
	}

	public int getReturnIdentifier(String methodNameAndType) {
		return ServiceManager.instance.getReturnIdentifier(this, methodNameAndType);
	}

	public NetworkDefinition getNetwork() {
		return ServiceManager.instance.lookupNetwork(getAccess());
	}

	public int getReadTimeout() {
		Attribute<?> attr = getAttribute("ReadTimeout");
		if (attr == null) {
			return 0;   // NO_TIMEOUT
		}
		return attr.valueInt();
	}

	public int getWriteTimeout() {
		Attribute<?> attr = getAttribute("WriteTimeout");
		if (attr == null) {
			return 0;   // NO_TIMEOUT
		}
		return attr.valueInt();
	}

	public int getAllocTimeout() {
		Attribute<?> attr = getAttribute("AllocTimeout");
		/*
		 * Man kann im Builder nicht auf Klassen zugreifen, die
		 * aus der Bibliothek stammen. Das soll auch _nicht_
		 * funktionieren!
		 *
		 * Eine Moeglichkeit waere im ConstantenPool nachzuschauen.
		 *
		 * IMClass memService = repository.getClass("keso/core/MemoryService");
		 *
		 * Aber das ist auch nicht einfach.
		 */
		//return keso.core.MemoryService.BLOCK;
		if (attr == null) {
			return -1;      // BLOCK
		}
		return attr.valueInt();
	}

	public String getAccess() {
		Attribute<?> attr = getAttribute("Access");
		if (attr == null) {
			return null;
		}
		return attr.valueString();
	}

	public ServiceDefinition getService() {
		return ServiceManager.instance.lookupService(getServiceName());
	}

	protected void finalizeCfg(FinalizeResult res) throws keso.compiler.config.parser.ParseException {
		System.err.println("finalizeCfg ImportDef. " + ident);
	}

	public String toString() { return ident + " (domain: " + parent + ")"; }

	public DomainDefinition<?> getDomain() {
		return parent;
	}

	public String getResourceClass() {
		return keso.compiler.imcode.IMPortalClass.createClassName(this);
	}
}
