/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.parser.ParseException;
import keso.compiler.imcode.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

final public class NetworkDefinition extends Set {

    public int identifier;
    private List<ServiceDefinition> services
    = new ArrayList<ServiceDefinition>();
    private final static String[] knownAttributes = new String[] {
        "Protocol",
        "DriverInterface"
    };

    private List<String> initMethods = new ArrayList<String>();
    private List<Integer> initParams = new ArrayList<Integer>();

    public NetworkDefinition(Set<?> parent, String name, int line) throws ParseException {
        super(parent, name, line);
    }

    public String getNetworkName() {
        return ident;
    }

    public void addService(ServiceDefinition def) {
        services.add(def);
    }

    public List<ServiceDefinition> getServices() {
        return services;
    }

    public void checkClasses(ClassStore repository) throws CompileException {

        IMClass driverIface = repository.getClass(getDriverInterfaceName());
        IMClass protocol = repository.getClass(getProtocolName());
        IMClass packetStream = repository.getClass("keso/core/io/PacketStream");

        if (driverIface == null) {
            throw new CompileException("The DriverInterface " + getDriverInterfaceName() + " specified for network " + this + " was not found.");
        }

        // check required bean methods for this interface
        Collection<String> attrs = getDriverAttributes();

        for (String id : attrs) {
            String methodName = "set" + id + "(I)V";

            //TODO: Check other int types
	    if (driverIface.getMethod(methodName) == null) {
		    System.err.println("The method " + methodName + " was not found in interface " + getDriverInterfaceName());
	    } else {
		    initMethods.add(methodName);
		    initParams.add(new Integer(getAttribute(id).valueInt()));
	    }

        }

        if (packetStream == null) {
            throw new CompileException("The interface PacketStream was not found.");
        }

        if (protocol == null) {
            throw new CompileException("The Protocol " + getProtocolName() + " specified for network " + this + " was not found.");
        }

        if (!protocol.implementsInterface(packetStream)) {
            throw new CompileException("The Protocol " + getProtocolName() + " specified for network " + this + " doesn't implement PacketStream.");
        }

    }

    public String getProtocolName() {
        Attribute<?> attr = getAttribute("Protocol");
        if (attr == null) {
            return null;
        }
        return attr.valueString();
    }

    //TODO access to driver attributes
    public String getDriverInterfaceName() {
        Attribute<?> attr = getAttribute("DriverInterface");
        if (attr == null) {
            return null;
        }
        return attr.valueString();
    }

    public List<String> getDriverAttributes() {
        List<String> result = new ArrayList<String>();
				Collection<String> ids = getAllIdentifiers();
				for(String id: ids) {
            boolean found = false;

						for(String kid: knownAttributes) {
                if (kid.equals(id)) {
                    found = true;
                    break;
                }
            }

            if (!found) {
               result.add(id);
            }
        }
        return result;
    }

    public List<String> getDriverInitMethods() {
        return initMethods;
    }

    public List<Integer> getDriverInitParameters() {
        return initParams;
    }


}
