/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.TaskDefinition;
import keso.compiler.backend.CoderOSEK;

public class ResourceDefinition extends SystemObject<DomainDefinition<?>> {
	/* used as index into keso_resource_index to
	 * reference the entry associated to an invalid
	 * domain lookup (null)
	 *
	 * Stable after config file has been parsed completly.
	 */
	public static byte INVALID_RESOURCE=0;

	private int ceilingPriority = 0;

	public ResourceDefinition(DomainDefinition<?> dom, String name, int line) {
		super(dom, name, line);
		INVALID_RESOURCE++;
	}

	/**
	 * Generate a RESOURCE block in OIL syntax out of this RESOURCE Definition.
	 *
	 * Known and supported attributes
	 * OIL standard:
	 *   RESOURCEPROPERTY		(STANDARD, LINKED, INTERNAL)
	 *
	 *   If ResourceProperty==LINKED, it has a subattribute
	 *   LINKED_RESOURCE		(RESOURCETYPE)
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {
		Attribute<?> property = getAttribute("ResourceProperty");

		if (property == null)
			throw new CompileException("ResourceProperty of Resource "+
					getIdentifier()+" not defined.");

		oil.append(prefix+"RESOURCE " + getIdentifier() + " {\n");
		oil.append(prefix+"\t"+property.ident.toUpperCase());
		oil.append(" = ");
		oil.append(property.valueString().toUpperCase());

		if (property.valueString().compareToIgnoreCase("Linked")==0) {
			Attribute<?> linkedRes = getAttribute("LinkedResource");
			if(linkedRes==null)
				throw new CompileException("LinkedResource attribute of LINKED Resource "+
						getIdentifier()+" not defined.");

			oil.append(" {\n" + prefix + "\t\t");
			oil.append(linkedRes.ident.toUpperCase());
			oil.append(" = ");

			ResourceDefinition lres = parent.getResource(linkedRes.valueString());

			if (lres==null)
				throw new CompileException("Unknown resource " + linkedRes.valueString()+" linked to "+getIdentifier());

			oil.append(lres.getIdentifier());
			oil.append(";\n"+prefix+"\t};\n");
		} else oil.append(";\n");

		oil.append(prefix+"};\n\n");
	}

	protected String typeID() { return "r_"; }

	public void recalcCeilingPriority() {
		// determine ceiling priority
		for(DomainDefinition<? extends TaskDefinition> dd: parent.getSystem().getAllSysDomains()) {
			for(TaskDefinition td: dd.getTasks()) {
				for(String ur: td.getResources()) {
					if(ur.compareToIgnoreCase(ident) == 0 && ceilingPriority < td.getPriority()) {
						ceilingPriority = td.getPriority();
					}
				}
			}
		}
	}

	protected void finalizeCfg(FinalizeResult res) {
		// require Resource class
		res.requireMethod("keso/core/Resource", null);

		recalcCeilingPriority();
	}

	public String getResourceClass() { return "keso/core/Resource"; }

	public DomainDefinition<?> getDomain() {
		return parent;
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {

		buf.append(prefix+"<resource name=\"");
		buf.append(getIdentifier());
		buf.append("\" ceiling=\"");
		buf.append(ceilingPriority);
		buf.append("\" />\n");
	}


}
