/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.config.parser.ParseException;
import keso.compiler.CompileException;
import keso.compiler.backend.CoderOSEK;

import java.util.*;

public abstract class Set<ParentType extends Set> extends Attribute<ParentType> {
	private Map<String, Attribute<?>> entries;

	public Set(String ident, int line) { this(null, ident, line); }

	public Set(ParentType parent, String ident, int line) {
		super(parent,ident,line);
		entries = new HashMap<String, Attribute<?>>();
	}

	public int valueInt() { throw new Error(ident); }
	public double valueDouble() { throw new Error(ident); }
	public String valueString() { throw new Error(ident); }
	public boolean valueBoolean() { throw new Error(ident); }

	public void addAttribute(String ident, Attribute<?> value) throws ParseException {
		entries.put(ident.toUpperCase(), value);
	}

	public Attribute<?> getAttribute(String ident) {
		return entries.get(ident.toUpperCase());
	}

	public Collection<Attribute<?>> getAllAttributes() {
		return entries.values();
	}

	public Collection<String> getAllIdentifiers() {
		return entries.keySet();
	}

	public void setComplex(ComplexAttribute<?> cattr) throws ParseException {
		addAttribute(cattr.toString(), cattr);
	}

	/**
	 * General OIL conversion method that will return all subattributes
	 * of this Set in OIL format.
	 * Specializations may overload this method as appropriate but still
	 * use it to dump their subattributes.
	 * The parameter supportedAttributes contains the name of all known
	 * Attributes for this Set. Only those attributes will be dumped.
	 * If supportedAttributes is the null reference, all attributes will
	 * be dumped.
	 * The parameter prefix will be prepended to every generated line
	 * of output, thus allowing to give hints for correct identation.
	 */
	public void toOIL(StringBuffer oil, String prefix, String oilCategory, CoderOSEK coder)
		throws CompileException {

		for(Attribute<?> a: entries.values()) {
			String attrName = a.ident.toUpperCase();
			if(coder.supportsOILAttribute(oilCategory, attrName, oil))
					a.toOIL(oil, prefix,coder);
		}
	}

	public void toCiAOConfig(StringBuffer buf, String prefix, Map<String, String> idmap)
		throws CompileException {
		for(Attribute<?> a: entries.values()) {
			a.toCiAOConfig(buf, prefix,idmap);
		}
	}

	/**
	 * Finds an Attribute by its identifier in a Collection of Attributes.
	 * @param attributes A Collection with Attribute-only elements that will be
	 *                   searched.
	 * @param ident      The identifier of the Attribute that shall be looked
	 *                   up.
	 * @return null if the Attribute was not found, the Attribute if it was
	 *         found.
	 */
	public <T extends Attribute<?>> T
	findAttributeInCollection(Collection<T> attributes, String ident) {
		for (T a : attributes) {
			if(a.ident.compareTo(ident) == 0) return a;
		}
		return null;
	}

	/**
	 * finalizeCfg will be called after the config file has entirely been read.
	 * Any subclass may overload it to do work that can only be done with
	 * knowledge of the complete config.
	 *
	 * This is the standard finalizeCfg function that just does not do anything
	 */
	protected void finalizeCfg(FinalizeResult res) throws keso.compiler.config.parser.ParseException {}

	final protected void finalizeCollection(Collection<? extends Set> sets,
			FinalizeResult res)
	throws keso.compiler.config.parser.ParseException {
		for(Set<?> s: sets)
			s.finalizeCfg(res);
	}

	protected void multiRefToOIL(StringBuffer oil, String prefix, String attrName, String value) {
		oil.append(prefix);
		oil.append(attrName.toUpperCase());
		oil.append(" = ");
		oil.append(value);
		oil.append(";\n");
	}

	public String toString() { return ident; }
}

