/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config;

import keso.compiler.*;
import keso.compiler.backend.CoderOSEK;
import keso.compiler.config.parser.ParseException;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.*;
import keso.compiler.config.messageport.*;

import java.util.*;

import keso.compiler.backend.gradmp.MemoryDescription;

public class SystemDefinition extends Set<WorldDefinition> {
	private static int SYSTEM_ID = 0;
	private final int systemID;

	private final DomainZero domainzero;
	private final List<DomainDefinition> all_domains
	= new ArrayList<DomainDefinition>();
	private final List<DomainDefinition<?>> app_domains
	= new ArrayList<DomainDefinition<?>>();
	private List<JavaDomain> java_domains = new ArrayList<JavaDomain>();

	private final List<AppmodeDefinition> appmodes
	= new ArrayList<AppmodeDefinition>();
	private final List<CounterDefinition> counters
	= new ArrayList<CounterDefinition>();
	private final List<EventDefinition> events
	= new ArrayList<EventDefinition>();
	private final List<ServiceDefinition> services
	= new ArrayList<ServiceDefinition>();
	private final List<Set<?>> finalizeCallback = new ArrayList<Set<?>>();
	private final List<SharedMemoryDefinition> sharedmemory
	= new ArrayList<SharedMemoryDefinition>();

	private Map<String, Map<String, String>> usedIdentifiers
	= new HashMap<String, Map<String, String>>();

	private MemoryDescription m_memdesc;

	public void registerIdentifier(String type, String ident) throws ParseException {
		type = type.toUpperCase();

		Map<String, String> typeht = usedIdentifiers.get(type);
		if(typeht == null) {
			typeht = new HashMap<String, String>();
			usedIdentifiers.put(type, typeht);
		}
		if(typeht.put(ident, ident) != null) {
			throw new ParseException("Duplicate " + type + " identifier: " + ident);
		}
	}

	public String[] getUsedIdentifiers(String type) {
		type = type.toUpperCase();
		Map<String, String> typeht = usedIdentifiers.get(type);
		if(typeht == null) return null;

		java.util.Set<String> values = typeht.keySet();
		String[] ids = new String[values.size()];
		ids = values.toArray(ids);
		return ids;
	}

	private OSDefinition os;
	private ComplexAttribute<?> stack;

	FinalizeResult finalizeResult;

	public SystemDefinition(WorldDefinition parent, String name, int line) throws ParseException {
		super(parent, name, line);

		DomainDefinition.resetID();
		AlarmDefinition.INVALID_ALARM=0;
		ResourceDefinition.INVALID_RESOURCE=0;

		systemID = SYSTEM_ID++;
		domainzero = new DomainZero(this);
		all_domains.add(domainzero);
		appmodes.add(new AppmodeDefinition(this, "OSDEFAULTAPPMODE", 0));
	}

	public int getSystemID() {
		return systemID;
	}

	public boolean hasLittleEndianMemory() {
		return BuilderOptions.hasLittleEndianMemory(getTarget());
	}

	public int getTarget() {
		Attribute<?> attr = getAttribute("Target");
		if (attr == null) {
			return BuilderOptions.DEFAULT_TARGET;
		}
		return BuilderOptions.getTargetByName(attr.valueString());
	}

	public String getProcessorType() {
		Attribute<?> main = getAttribute("ProcessorType");
		if (main==null) return null;
		return main.valueString();
	}

	public void addJavaDomain(JavaDomain dom) {
		BuilderOptions.getOpts().verbose("java domain: "+dom);
		all_domains.add(dom);
		app_domains.add(dom);
		java_domains.add(dom);
	}

	public void addNativeDomain(NativeDomain dom) {
		BuilderOptions.getOpts().verbose("native domain: "+dom);
		all_domains.add(dom);
		app_domains.add(dom);
	}

	public List<DomainDefinition<?>> getAppDomains() {
		return app_domains;
	}

	public List<JavaDomain> getJavaDomains() {
		return java_domains;
	}

	public List<DomainDefinition> getAllSysDomains() {
		return all_domains;
	}

	public DomainDefinition<?> getAppDomain(String dName) {
		return findAttributeInCollection(app_domains, dName);
	}

	public JavaDomain getJavaDomain(String dName) {
		return findAttributeInCollection(java_domains, dName);
	}

	public void addAppmode(AppmodeDefinition amode) {
		BuilderOptions.getOpts().verbose("Appmode: "+amode);
		appmodes.add(amode);
	}

	public ResourceDefinition getSysResource(String ident) {
		for(DomainDefinition<?> dom: all_domains) {
			ResourceDefinition res = dom.getResource(ident);
			if(res != null)
				return res;
		}
		return null;
	}

	public AlarmDefinition getSysAlarm(String ident) {
		for(DomainDefinition<?> dom: app_domains) {
			AlarmDefinition ad = dom.getAlarm(ident);
			if(ad != null)
				return ad;
		}
		return null;
	}

	public List<AppmodeDefinition> getAppmodes() { return appmodes; }

	public AppmodeDefinition getAppmode(String amName) {
		return findAttributeInCollection(appmodes, amName);
	}

	public void addCounter(CounterDefinition cdef) {
		BuilderOptions.getOpts().verbose("counter: "+cdef);
		counters.add(cdef);
	}

	public List<CounterDefinition> getCounters() {	return counters; }

	public CounterDefinition getCounter(String cName) {
		return findAttributeInCollection(counters, cName);
	}

	public void addEvent(EventDefinition ev) {
		BuilderOptions.getOpts().verbose("Event: "+ev);
		events.add(ev);
	}

	public void addSharedMemory(SharedMemoryDefinition shm) {
		BuilderOptions.getOpts().verbose(shm.toString());
		sharedmemory.add(shm);
	}

	public List<EventDefinition> getEvents() { return events; }

	public EventDefinition getEvent(String evName) {
		return findAttributeInCollection(events, evName);
	}

	public void setOS(OSDefinition os) throws ParseException {
		if(this.os != null)
			throw new ParseException("Duplicate definition of OsekOS encountered (kesorc:)"+os.getLine()+")");
		this.os = os;
	}

	public void registerFinalize(Set set) {
		finalizeCallback.add(set);
	}

	private EventDefinition _gcrunevent = null;
	public EventDefinition createGCRunEvent() {
		if(_gcrunevent == null) {
			_gcrunevent = new EventDefinition(this, "KESO_GCRun");
			events.add(_gcrunevent);
		}

		return _gcrunevent;
	}

	public void finalizeCfg() throws keso.compiler.config.parser.ParseException {
		finalizeResult = new FinalizeResult();

		if (os!=null) {
			os.finalizeCfg(finalizeResult);
		}
		finalizeCollection(appmodes, finalizeResult);

		for(DomainDefinition<?> dom: all_domains) {
			dom.finalizeCfg(finalizeResult);
		}

		finalizeCollection(counters, finalizeResult);
		finalizeCollection(finalizeCallback, finalizeResult);

		// all tasks need to have finalized before finalizing Events
		finalizeCollection(events, finalizeResult);

		// create system services vector
		for(JavaDomain dom: getJavaDomains()) {
			services.addAll(dom.getServices());
		}
	}

	public java.util.Set<FinalizeResult.ProtectedMethod> getRequiredMethods() {
		return finalizeResult.requiredMethods;
	}

	public OSDefinition getOSDef() { return os; }

	/**
	 * Generates an OIL file out of the SystemDefinition.
	 * The information of the subobjects will be included,
	 * therefore the returned String contains a whole OIL file.
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {
		oil.append(prefix+"\nCPU " + ident + " {\n");

		// OS block
		if (os == null)
			throw new CompileException("OsekOS Block not defined in kesorc");
		os.toOIL(oil, prefix+"\t", coder);

		// Domains
		if (app_domains.size()==0)
			throw new CompileException("At least one domain must be defined in kesorc");

		// ISRs belong to domains, so they are also handled here
		for (DomainDefinition<?> dom: all_domains) {
			dom.toOIL(oil, prefix+"\t",coder);
		}

		// appmodes
		for (AppmodeDefinition ad : appmodes) {
			ad.toOIL(oil, prefix+"\t",coder);
		}

		// events
		for (EventDefinition ed : events) {
			ed.toOIL(oil, prefix+"\t",coder);
		}

		// counters
		for (CounterDefinition cd : counters) {
			cd.toOIL(oil, prefix+"\t",coder);
		}

		// System Tasks
		//for(int i=0; i<sysTasks.size(); i++)
		//	 sysTasks.elementAt(i).toOIL(oil, prefix+"\t",coder);

		// System Resources
		//for(int i=0; i<sysResources.size(); i++)
		//	sysResources.elementAt(i).toOIL(oil, prefix+"\t",coder);

		oil.append(prefix+"};\n");
	}

	private final java.util.Set<String> fakeMessagePorts = new HashSet<String>();
	public void addAttribute(String ident, Attribute<?> value) throws ParseException {
		if(ident.compareToIgnoreCase("FakeMessagePort")==0) {
			fakeMessagePorts.add(value.valueString());
		}
		super.addAttribute(ident,value);
	}

	public java.util.Set<String> getFakeMessagePorts() { return fakeMessagePorts; }

	public void setComplex(ComplexAttribute cattr) throws ParseException {
		if (cattr.ident.compareToIgnoreCase("Stack")==0) {
			stack = cattr;
			return;
		}
		addAttribute(cattr.ident, cattr);
	}

	public ComplexAttribute<?> getStack() { return stack; }

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		buf.append(prefix+"\n<ciaoApp  name=\"" + ident + "\"\n");
		buf.append("\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n");
		buf.append("\txsi:noNamespaceSchemaLocation=\"../../../../../../../../tools/xml/ciaoApp.xsd\">\n\n");

		// OS block
		if (os == null) throw new CompileException("OsekOS Block not defined in kesorc");
		os.toCiAOConfig(buf, prefix+"\t");

		// Domains
		if (app_domains.size()==0)
			throw new CompileException("At least one domain must be defined in kesorc");

		// ISRs belong to domains, so they are also handled here
		for(DomainDefinition<?> dd: all_domains)
			dd.toCiAOConfig(buf, prefix+"\t");

		// appmodes
		for(AppmodeDefinition amd: appmodes)
			amd.toCiAOConfig(buf, prefix+"\t");

		// shared memory
		for(SharedMemoryDefinition shm: sharedmemory)
			shm.toCiAOConfig(buf, prefix+"\t");

		// events
		for(EventDefinition edef: events)
			edef.toCiAOConfig(buf, prefix+"\t");

		// counters
		for (CounterDefinition cdef: counters)
			cdef.toCiAOConfig(buf,prefix+"\t");

		// Application Headers
		buf.append(prefix+"\t<applicationheaders>\n");
		Attribute<?> appheader = getAttribute("AppHeader");
		if (appheader!=null) {
			buf.append(prefix+"\t\t<applicationheader name=\"");
			buf.append(appheader.valueString());
			buf.append("\"/>\n");
		}

		buf.append(prefix+"\t\t<applicationheader name=\"ciao_appcode.h\"/>\n");
//		buf.append(prefix+"\t\t<applicationpointcuts name=\"keso_pointcuts.ah\"/>\n");
		buf.append(prefix+"\t</applicationheaders>\n");

		buf.append("\n</ciaoApp>\n");
	}

	public DomainZero getDomainZero() { return domainzero; }

	public List<ServiceDefinition> getServices() {
		return services;
	}

	public ReceiverPortDefinition getReceiverPort(String name) {
		for(DomainDefinition<?> dd: app_domains) {
			Collection<ReceiverPortDefinition> rpds = dd.getReceiverPorts();
			for(ReceiverPortDefinition rp: rpds) {
				if(rp.ident.compareTo(name) == 0)
					return rp;
			}
		}
		return null;
	}

	public SharedMemoryDefinition getSharedMemory(String shmName) {
		return findAttributeInCollection(sharedmemory, shmName);
	}

	public List<SharedMemoryDefinition> getSharedMemories() {
		return sharedmemory;
	}

	public TaskDefinition getSysTask(String taskName) {
		for(DomainDefinition<?> dd: app_domains) {
			TaskDefinition td = dd.getTask(taskName);
			if(td != null)
				return td;
		}
		return null;
	}


	public MemoryDescription getMemoryDescription() {
		return m_memdesc;
	}
	public void setMemoryDescription( MemoryDescription md ) {
		m_memdesc = md;
	}
}

