/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config.domain;

import keso.compiler.*;
import keso.compiler.backend.CoderOSEK;
import keso.compiler.config.*;
import keso.compiler.config.task.*;
import keso.compiler.config.messageport.*;
import keso.compiler.config.parser.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;

public abstract class DomainDefinition<TaskDefType extends TaskDefinition> extends Set<SystemDefinition> {
	private static byte INVALID_DOMAIN=0;

	protected final List<AlarmDefinition> alarms
	= new ArrayList<AlarmDefinition>();
	protected final Collection<ComplexAttribute<DomainDefinition<?>>> usedSharedMemory
	= new ArrayList<ComplexAttribute<DomainDefinition<?>>>();
	protected final List<ISRDefinition> isrs = new ArrayList<ISRDefinition>();
	protected final List<ReceiverPortDefinition> receiverPorts
	= new ArrayList<ReceiverPortDefinition>();
	protected final List<ResourceDefinition> resources
	= new ArrayList<ResourceDefinition>();
	protected final List<SenderPortDefinition> senderPorts
	= new ArrayList<SenderPortDefinition>();
	protected final List<TaskDefType> tasks = new ArrayList<TaskDefType>();

	public final byte domainid;

	public DomainDefinition(SystemDefinition system, String name, int line) {
		super(system, name, line);
		domainid = INVALID_DOMAIN++;
	}

	public void registerFinalize(Set set) { parent.registerFinalize(this); }

	public void addTask(TaskDefType task) throws ParseException {
		BuilderOptions.getOpts().verbose("task: "+task);
		parent.registerIdentifier("Task", task.ident);
		tasks.add(task);
	}

	public TaskDefType getTask(String taskName) {
		for(TaskDefType task: tasks) {
			if(task.ident.compareTo(taskName) == 0)
				return task;
		}
		return null;
	}

	public void addISR(ISRDefinition isr) {
		BuilderOptions.getOpts().verbose("ISR: "+isr);
		isrs.add(isr);
	}

	public List<ISRDefinition> getISRs() { return isrs; }

	public ISRDefinition getISR(String isrName) {
		for(ISRDefinition isr: isrs) {
			if(isr.ident.compareTo(isrName) == 0) return isr;
		}
		return null;
	}

	public void addResource(ResourceDefinition res) throws ParseException {
		BuilderOptions.getOpts().verbose("Resource: "+res);
		parent.registerIdentifier("Resource", res.ident);
		resources.add(res);
	}

	public void addSenderPort(SenderPortDefinition spd) {
		BuilderOptions.getOpts().verbose("SenderPort: "+spd);
		senderPorts.add(spd);
	}

	public void addReceiverPort(ReceiverPortDefinition rpd) {
		BuilderOptions.getOpts().verbose("ReceiverPort: "+rpd);
		receiverPorts.add(rpd);
	}

	public List<TaskDefType> getTasks() { return tasks; }
	public List<ResourceDefinition> getResources() { return resources; }

	public ResourceDefinition getResource(String resName) {
		return findAttributeInCollection(resources, resName);
	}

	public void addAlarm(AlarmDefinition adef) throws ParseException {
		BuilderOptions.getOpts().verbose("global alarm: "+adef);
		parent.registerIdentifier("Alarm", adef.ident);
		alarms.add(adef);
	}

	public List<AlarmDefinition> getAlarms() {	return alarms; }
	public List<ReceiverPortDefinition> getReceiverPorts() { return receiverPorts; }
	public List<SenderPortDefinition> getSenderPorts() { return senderPorts; }

	public AlarmDefinition getAlarm(String aName) {
		return findAttributeInCollection(alarms, aName);
	}

	public String toString() { return ident; }

	/**
	 * Return the information from this domain and its subobjects in a
	 * format suitable for inclusion within an OIL file.
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder)
		throws CompileException {
		for(TaskDefType task: tasks) {
			task.toOIL(oil, prefix, coder);
		}

		for(ResourceDefinition res : resources) {
			res.toOIL(oil, prefix, coder);
		}

		for(AlarmDefinition ad : alarms) {
			ad.toOIL(oil, prefix, coder);
		}

		for(ISRDefinition id : isrs) {
			id.toOIL(oil, prefix, coder);
		}
	}

	public void finalizeCfg(FinalizeResult res) throws ParseException {
		finalizeCollection(tasks, res);
		finalizeCollection(resources, res);
		finalizeCollection(alarms, res);
		finalizeCollection(isrs, res);
		finalizeCollection(senderPorts, res);
		finalizeCollection(receiverPorts, res);

		for(ComplexAttribute<DomainDefinition<?>> useshm: usedSharedMemory) {
			SharedMemoryDefinition shm = parent.getSharedMemory(useshm.value);
			if(shm == null)
				throw new ParseException("shared memory " + useshm.value + " used but never defined");
			String amode  = useshm.getAttribute("mode").valueString();
			boolean read  = amode.indexOf('r') >= 0;
			boolean write = amode.indexOf('w') >= 0;
			shm.registerUser(this, read, write);
		}
	}

	protected String typeID() { return "d_"; }

	public void setComplex(ComplexAttribute cattr) throws ParseException {
		if(cattr.ident.compareToIgnoreCase("UseSharedMemory") == 0) {
			usedSharedMemory.add(cattr);
		} else {
			throw new ParseException("Domain supports no complex attribute " + cattr.ident);
		}
	}

	public static void resetID() {
		INVALID_DOMAIN=0;
	}

	public String getComponent() throws CompileException {
		try {
			return getAttribute("CiAOComponent").valueString();
		} catch(NullPointerException e) {
			throw new CompileException("Domain " + getIdentifier() + " does not have CiAOComponent specified\n");
		}
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		buf.append(prefix+ "<osapplication name=\"");
		buf.append(getIdentifier());
		buf.append("\" trusted=\"");
		buf.append(getAttribute("CiAOTrusted").valueString());
		buf.append("\">\n");

		buf.append(prefix+"\t<componentref name=\"" + getComponent() + "\"/>\n");
		buf.append(prefix + "</osapplication>\n\n");

		buf.append(prefix + "<component name=\"");
		buf.append(getComponent());
		buf.append("\">\n");

		for(TaskDefType td: tasks) {
			buf.append(prefix+"\t<taskref name=\"");
			buf.append(td.getIdentifier());
			buf.append("\"/>\n");
		}

		for(ReceiverPortDefinition rpd: receiverPorts) {
			buf.append(prefix+"\t<receiverport name=\"");
			buf.append(rpd.getIdentifier());
			buf.append("\" ");
			rpd.toCiAOConfig(buf, " ", null);
			buf.append(" />\n");
		}

		for(SenderPortDefinition spd: senderPorts) {
			buf.append(prefix+"\t<senderport name=\"");
			buf.append(spd.getIdentifier());
			buf.append("\" ");
			spd.toCiAOConfig(buf, " ", null);
			buf.append(" />\n");
		}

		for(ISRDefinition id: isrs) {
			buf.append(prefix+"\t<isrref name=\"");
			buf.append(id.getIdentifier());
			buf.append("\"/>\n");
		}

		for(ComplexAttribute<DomainDefinition<?>> useshm: usedSharedMemory) {
			buf.append(prefix+"\t<shmref name=\"");
			buf.append(useshm.value);
			buf.append("\" mode=\"");
			buf.append(useshm.getAttribute("mode").valueString());
			buf.append("\" />\n");
		}
		buf.append(prefix+"</component>\n\n");

		for(TaskDefType td: tasks)
			td.toCiAOConfig(buf, prefix);

		for(ResourceDefinition rd: resources)
			rd.toCiAOConfig(buf, prefix);

		for(ISRDefinition id: isrs)
			id.toCiAOConfig(buf, prefix);

		for(SenderPortDefinition spd: senderPorts)
			spd.toCiAOConfig(buf,prefix);

		buf.append("\n");

		//for(int i=0;i<resources.size();i++) {
		//	ResourceDefinition res = (ResourceDefinition)resources.elementAt(i);
		//	res.toOIL(oil, prefix);
		//}

		for(AlarmDefinition ad : alarms) {
			ad.toCiAOConfig(buf, prefix);
		}

		//for(int i=0;i<isrs.size();i++)
		//	((ISRDefinition) isrs.elementAt(i)).toOIL(oil, prefix);
	}

	// HACK transitive parent accesses (e.g. access parent.parent inside a
	// Taskdefinition) are of type Set rather than SystemDefinition, since
	// DomainDefinition<> is not fully specialized within TaskDefinition
	public final SystemDefinition getSystem() {
		return parent;
	}

	private HashSet<String> importedTasks     = new HashSet<String>();
	private HashSet<String> importedResources = new HashSet<String>();
	private HashSet<String> importedAlarms    = new HashSet<String>();

	public java.util.Set<String> getTaskImports() {
		return importedTasks;
	}

	public java.util.Set<String> getResourceImports() {
		return importedResources;
	}

	public java.util.Set<String> getAlarmImports() {
		return importedAlarms;
	}

	public void addAttribute(String ident, Attribute<?> value) throws ParseException {
		if(ident.compareToIgnoreCase("ImportTask") == 0) {
			importedTasks.add(value.valueString());
		} else if(ident.compareToIgnoreCase("ImportResource") == 0) {
			importedResources.add(value.valueString());
		} else if(ident.compareToIgnoreCase("ImportAlarm") == 0) {
			importedAlarms.add(value.valueString());
		} else {
			super.addAttribute(ident,value);
		}
	}

	public List<SharedMemoryDefinition> getSharedMemories() {
		List<SharedMemoryDefinition> shms
		= new ArrayList<SharedMemoryDefinition>();

		for(ComplexAttribute<DomainDefinition<?>> useshm: usedSharedMemory) {
			SharedMemoryDefinition shm = parent.getSharedMemory(useshm.value);
			shms.add(shm);
		}

		return shms;
	}

	public String getIdentifier() { return ident; }
}
