/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.config.task;

import keso.compiler.*;
import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.parser.ParseException;
import keso.compiler.backend.CoderOSEK;

import java.util.*;

public abstract class TaskDefinition extends SystemObject<DomainDefinition<?>> {
	protected ComplexBoolAttribute autoStart;

	protected int priority=0;
	protected boolean preemptable=false;
	protected boolean callsched=false;
	protected int activation=1;
	protected int stacksize=128;

	protected java.util.Set<String> usedResources;
	protected java.util.Set<String> usedEvents;
	protected java.util.Set<Attribute<?>> usedMessages;

	private int evBitsUsed;
	public static byte INVALID_TASK=0;

	public TaskDefinition(DomainDefinition<?> dom, String name, int line) {
		super(dom, name, line);
		usedResources = new HashSet<String>();
		usedEvents = new HashSet<String>();
		usedMessages = new HashSet<Attribute<?>>();
		evBitsUsed=0;
		INVALID_TASK++;
	}

	public boolean getAutoStart() {
		return autoStart==null?false:autoStart.setting;
	}

	public String getAutoStartAppmode() {
		if(autoStart.setting) {
			try {
				return autoStart.getAttribute("APPMODE").valueString();
			} catch (NullPointerException e) { }
		}
		return "OSDEFAULTAPPMODE";
	}

	public void setComplex(ComplexAttribute<?> cattr) throws ParseException {
		if ( (!(cattr instanceof ComplexBoolAttribute)) ||
				(cattr.ident.compareToIgnoreCase("autostart")!=0))
			throw new ParseException("TaskDefinition does not support the ComplexAttribute + " + cattr.ident);
		autoStart = (ComplexBoolAttribute) cattr;
	}

	public void addAttribute(String ident, Attribute<?> value) throws ParseException {
		if(ident.compareToIgnoreCase("UseResource")==0) {
			usedResources.add(value.valueString());
		} else if(ident.compareToIgnoreCase("UseMessage")==0) {
			usedMessages.add(value);
		} else if(ident.compareToIgnoreCase("UseEvent")==0) {
			usedEvents.add(value.valueString());

		} else if(ident.compareToIgnoreCase("Priority")==0) {
			setPriority(value.valueInt());
		} else if(ident.compareToIgnoreCase("Stacksize")==0) {
			setStackSize(value.valueInt());
		} else if(ident.compareToIgnoreCase("CallScheduler")==0) {
			callsched = value.valueBoolean();
		} else if(ident.compareToIgnoreCase("Activation")==0) {
			activation = value.valueInt();
		} else if(ident.compareToIgnoreCase("Schedule")==0) {
			if(value.valueString().compareToIgnoreCase("FULL")==0)
				preemptable = true;
			else if(value.valueString().compareToIgnoreCase("NON")==0)
				preemptable = false;
			else throw new ParseException("Schedule needs to be FULL or NON");
		}
		super.addAttribute(ident,value);
	}

	public java.util.Set<String> getResources()	{return usedResources;}
	public java.util.Set<String> getEvents()	{return usedEvents; }
	public java.util.Set<Attribute<?>> getMessages()	{return usedMessages; }

	public void addResource(ResourceDefinition res) throws ParseException {
		addAttribute("UseResource", new StringAttr(this, "UseResource", res.ident,0));
	}

	public String toString() { return ident + " (domain: " + parent + ")"; }

	public int getEvBits() { return evBitsUsed; }

	// called by an Event to register its mask as used so other Events can chose
	// another one
	public void registerEvMask(int evMask) { evBitsUsed |= evMask; }

	/**
	 * Generate a TASK block in OIL syntax out of this TaskDefinition.
	 *
	 * Known and supported attribute
	 * OIL standard:
	 *   PRIORITY			(UINT32)
	 *   SCHEDULE			(NON, FULL)
	 *   ACTIVATION			(UINT32)
	 *   AUTOSTART			(boolean, if true has APPMODE subattrs)
	 *   RESOURCE			(multiple references to accessed Resources)
	 *   EVENT			(multiple references to used events)
	 *   MESSAGE			(multiple references to used messages)
	 * ProOSEK extensions:
	 *   TYPE			(BASIC, EXTENDED)
	 *   STACKSIZE			(UINT32)
	 *   CALLSCHEDULER		(boolean)
	 */
	public void toOIL(StringBuffer oil, String prefix, CoderOSEK coder) throws CompileException {
		oil.append(prefix+"TASK " + getIdentifier() + " {\n");

		// autostart attribute handled specially
		oil.append("\t\tAUTOSTART = ");
		if(getAutoStart()) {
			oil.append("TRUE {\n\t\t\tAPPMODE = ");
			oil.append(getAutoStartAppmode());
			oil.append(";\n\t\t};\n");
		} else {
			oil.append("FALSE;\n");
		}

		// Resources, Events and Messages are multiple reference
		DomainDefinition<?> dom = parent;
		SystemDefinition sysdef = (SystemDefinition) dom.parent;

		for(String r: usedResources) {
			ResourceDefinition res = sysdef.getSysResource(r);
			if (res == null)
				throw new CompileException("Task " + getIdentifier() + " references unknown Resource " + r);

			multiRefToOIL(oil, prefix+"\t", "RESOURCE", res.getIdentifier());
		}

		for(String e: usedEvents) {
			EventDefinition ev = sysdef.getEvent(e);

			if(ev == null)
				throw new CompileException("Task " + getIdentifier() + " references unknown Event " + e);

			multiRefToOIL(oil, prefix+"\t", "EVENT", ev.getIdentifier());
		}

		super.toOIL(oil, prefix + "\t", "TASK", coder);
		oil.append(prefix+"};\n\n");
	}

	public void finalizeCfg(FinalizeResult res) throws keso.compiler.config.parser.ParseException {
		SystemDefinition sysdef = (SystemDefinition) parent.parent;

		res.requireMethod("java/lang/Thread", null);

		// register at all used events
		for(String e: usedEvents) {
			EventDefinition ev = sysdef.getEvent(e);
			if (ev==null)
				throw new keso.compiler.config.parser.ParseException("Unknown Event " + e	+ " referenced from Task " + getIdentifier());
			ev.registerTask(this);
		}
	}

	protected String typeID() { return "t_"; }

	public String getComponent() throws CompileException {
		return ((DomainDefinition<?>) parent).getComponent();
	}

	public void setAutoStart(boolean aS) throws ParseException {
		autoStart = new ComplexBoolAttribute(this, "Autostart", aS, -1);
	}

	public int getPriority() { return priority; }
	public void setPriority(int prio) { priority = prio; }
	public void setStackSize(int size) { stacksize = size; }
	public void setSchedule(boolean preempt) { preemptable = preempt; }
	public void addEvent(String ev) {
		usedEvents.add(ev);
	}

	public void toCiAOConfig(StringBuffer buf, String prefix) throws CompileException {
		buf.append(prefix+"<task name=\"");
		buf.append(getIdentifier());
		buf.append("\" ");

		buf.append("priority=\""   + priority + "\" " );
		buf.append("schedule=\""   + (preemptable ?"full":"non") + "\" " );
		buf.append("activation=\"" + activation + "\" " );
		buf.append("stacksize=\""  + stacksize + "\" " );
		buf.append(">\n");
		for(String eve: usedEvents)
			buf.append(prefix + "\t<eventref name=\"" + eve + "\"/>\n");
		buf.append(prefix + "</task>\n");
	}

	public String getResourceClass() { return "java/lang/Thread"; }

	public DomainDefinition<?> getDomain() { return parent; }
}
