/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.dump;

import java.io.*;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class ClassTreeWriter extends Pass {

	@Override
	public String getName() {
		return "Class tree writer";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return (opts.getOutputClassTree() != null);
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsDump(true);
	}

	@Override
	public boolean runOnRepository(ClassStore repository)
	throws CompileException {

		BuilderOptions opts     = repository.getOptions();
		String         filename = opts.getOutputClassTree();
		if (filename == null)
			return false;

		try {

			String outPath = "";
			if ((new File(filename)).isAbsolute() == false)
				outPath = opts.getOutputPath() + '/';
			DotWriter<IMClass> dw = new DotWriter<IMClass>(outPath + filename);
			dw.initGraph("ClassTree");
			dw.writeLine("ranksep=\"2.0 equally\";");
			dw.setNodeStyle("shape=box height=.5 width=.5 fontsize=10");

			// Write edges
			IMClass root = repository.getClass("java/lang/Object");
			if (root == null)
				writeClassAndSubClasses(repository, dw, "java/lang/Object");
			else
				writeClassAndSubClasses(repository, dw, root);
			IMClass imClasses[] = repository.getAllClasses();

			// Write nodes for array types
			dw.initSubGraph("cluster_arrays");
			dw.writeLine("style=filled;");
			dw.writeLine("color=white;");
			for (IMClass imClass : imClasses) {
				if (imClass.getClassName().startsWith("["))
					dw.addNode(imClass,
							imClass.getClassTypeInfo().toDigString());
			}
			dw.finalizeSubGraph();

			// Write nodes for interfaces
			dw.initSubGraph("cluster_ifaces");
			dw.writeLine("style=filled;");
			dw.writeLine("color=white;");
			for (IMClass imClass : imClasses) {
				if (imClass.isInterface()
						&& imClass.getClassName().startsWith("[") == false)
					dw.addNode(imClass, imClass.getAlias(), "shape=hexagon");
			}
			dw.finalizeSubGraph();

			// Write nodes for classes
			dw.initSubGraph("cluster_classes");
			dw.writeLine("style=filled;");
			dw.writeLine("color=white;");
			for (IMClass imClass : imClasses) {
				if (imClass.getClassName().startsWith("[") == false
						&& imClass.isInterface() == false)
					dw.addNode(imClass, imClass.getClassTypeInfo().toDigString()
							+ "\\nmethods: " + imClass.getMethods().length);
			}
			dw.finalizeSubGraph();
			dw.finalizeGraph();

		} catch (Exception e) {
			Debug.out.println(e);
			e.printStackTrace();
		}

		return false;
	}

	private static void writeClassAndSubClasses(ClassStore repository,
			DotWriter<IMClass> dw, String rootName)
	throws DotWriterException {
		IMClass subs[] = repository.getSubClasses(rootName);
		if (subs == null)
			return;
		dw.addNode(null, rootName, "rank=source");
		for (IMClass sub : subs) {
			dw.addEdge(null, sub);
			writeClassAndSubClasses(repository, dw, sub);
		}
	}

	private static void writeClassAndSubClasses(ClassStore repository,
			DotWriter<IMClass> dw, IMClass imClass)
	throws DotWriterException {
		String className = imClass.getClassName();
		IMClass subs[] = repository.getSubClasses(className);
		if (subs != null) {
			for (IMClass sub : subs) {
				dw.addEdge(sub, imClass, "arrowhead=empty");
				writeClassAndSubClasses(repository, dw, sub);
			}
		}
		IMClass types[] = repository.findImplements(className);
		if (types != null) {
			for (IMClass type : types) {
				dw.addEdge(type, imClass, "arrowhead=empty style=dotted");
				//writeClassAndSubClasses(repository, dw, type);
			}
		}
	}

}
