/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.dump;

import java.io.*;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class DomTreeWriter extends MethodPass {

	@Override
	public String getName() {
		return "Dominator tree writer";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.createDomTree();
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsDump(true);
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method instanceof IMSystemMethod || method.hasCode() == false)
			return false;

		String outPath  = method.getClassStore().getOptions().getOutputPath();
		String fileName = outPath + "/dom_" + method.getAlias() + ".dot";
		try {
			DotWriter<IMBasicBlock> dw = new DotWriter<IMBasicBlock>(fileName);
			dw.initGraph("DomTree");
			dw.writeLine("size=\"7,10\";");
			dw.writeLine("page=\"8.25,11.75\";");
			dw.setNodeStyle("shape=box height=.5 width=.5 fontsize=10");
			DominatorTree domTree = method.getDomTree();
			for (IMBasicBlock bb : method) {
				dw.addNode(bb, getDomBoxLabel(bb, method), "fontsize=8");
				IMBasicBlock dom = domTree.getDominatorOf(bb);
				if (dom != null)
					dw.addEdge(dom, bb);
			}
			dw.finalizeGraph();
		} catch (DotWriterException e) {
			throw new CompileException(e.getMessage());
		} catch (FileNotFoundException e) {
			throw new CompileException(e.getMessage());
		}
		return false;
	}

	@Override
	protected boolean skipClass(IMClass imClass) {
		return (imClass instanceof IMArrayClass);
	}

	private static String getDomBoxLabel(IMBasicBlock bb, IMMethod cMethod)
	throws CompileException {
		IMMethod method = bb.getMethod();
		if (method != cMethod)
			bb.getOptions().critical("Basic block mix up: "
					+ bb.toReadableString() + '\n' + method.getAlias() + "!="
					+ cMethod.getAlias());
		String label = bb.toLabel() + "\\l " + bb.getDFNum();
		int[]  df    = method.getDomTree().getDominanceFrontierIDs(bb);
		if (df != null && df.length > 0) {
			label += " DF:";
			for (int id : df) {
				label += " " + id;
			}
		}
		return label;
	}

}
