/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.dump;

import java.util.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.config.ISRDefinition;
import keso.compiler.config.ServiceDefinition;
import keso.compiler.config.domain.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class ReachabilityGraphWriter extends DomainPass {

	@Override
	public String getName() {
		return "Reachability graph writer";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.createReachabilityGraph();
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsDump(true);
		info.addDependency(FlowAnalysis.class);
		info.addDependency(SimpleReachabilityAnalysis.class);
	}

	@Override
	public boolean runOnDomain(IMDomain domain) throws CompileException {

		Set<IMMethod> tasks = new HashSet<IMMethod>();
		for (IMTask task : domain.getTasks()) {
			IMClass imClass = classStore.getClass(task.getMainClass());
			tasks.add(imClass.getMethod(task.getMainMethod()));
		}
		JavaDomain    domainDef = domain.getDef();
		Set<IMMethod> isrs      = new HashSet<IMMethod>();
		for (ISRDefinition isr : domainDef.getISRs()) {
			IMClass imClass = classStore.getClass(isr.getHandlerClass());
			isrs.add(imClass.getMethod(isr.getHandlerMethod()));
		}
		Set<IMMethod> serviceMethods = new HashSet<IMMethod>();
		for (ServiceDefinition service : domainDef.getServices()) {
			IMClass imClass = service.getServiceClass(classStore);
			for (IMMethod method : imClass.getMethods()) {
				serviceMethods.add(method);
			}
		}

		ReachabilityData     data          = domain.getReachabilityData();
		Collection<IMMethod> rMethods      = data.getMethods();
		Collection<IMField>  rStaticFields = data.getStaticFields();
		Multimap<IMClass, IMField> fields
		= new HashMultimap<IMClass, IMField>();
		for (IMField field : rStaticFields) {
			fields.add(field.getType(), field);
		}

		String outPath  = classStore.getOptions().getOutputPath();
		String domainID = domain.getIdentifier();
		String fileName = outPath + "/reachability_" + domainID + ".dot";

		try {

			DotWriter<Object> dw = new DotWriter<Object>(fileName);
			dw.initGraph("ReachabilityGraph");
			dw.writeLine("ranksep=2;");
			dw.setNodeStyle("shape=box");

			// Iterate over all classes
			for (IMClass imClass : data.getClasses()) {

				// Cluster all methods and fields of the current class
				dw.initSubGraph("cluster_" + imClass.getClassID());
				dw.writeLine("label=\"" + imClass.getClassName() + "\";");
				IMMethod[] methods = imClass.getMethods();
				if (methods != null) {
					for (IMMethod method : methods) {
						if (rMethods.contains(method) == false)
							continue;
						String style = "";
						if (tasks.contains(method))
							style = "color=darkgreen, fontcolor=darkgreen";
						else if (isrs.contains(method))
							style = "color=red, fontcolor=red";
						else if (serviceMethods.contains(method))
							style = "color=blue, fontcolor=blue";
						dw.addNode(method, method.getMethodNameAndType(), style);
					}
				}
				for (IMField field : fields.get(imClass)) {
					dw.addNode(field, field.getFieldName(), "shape=ellipse");
				}
				dw.finalizeSubGraph();

				// Write edges to all callees for each method of the current class
				for (IMMethod method : methods) {
					if (rMethods.contains(method) == false)
						continue;
					for (IMMethod callee : method.getCalledMethods()) {
						if (rMethods.contains(callee))
							dw.addEdge(method, callee);
					}
				}
			}

			for (IMField field : rStaticFields) {
				for (IMMethod user : data.getStaticFieldUsers(field)) {
					dw.addEdge(user, field, "style=dotted");
				}
			}
			dw.finalizeGraph();

		} catch (Exception e) {
			e.printStackTrace();
			throw new CompileException(getName() + " failed");
		}

		return false;
	}

}
