/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.dump;

import java.io.*;
import java.util.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class StatisticsDump extends MethodPass {

	private String outPath = ".";
	private IMDomain[] domains;

	private final Multimap<IMDomain, IMClass> classMap
	= new HashUniqueMultimap<IMDomain, IMClass>();
	private final Map<IMDomain, Integer> uniqueDomClasses
	= new HashMap<IMDomain, Integer>();
	private int totalClasses = 0;
	private final List<Set<IMClass>> domCountClasses
	= new ArrayList<Set<IMClass>>();
	
	private final Multimap<IMDomain, IMMethod> methodMap
	= new HashUniqueMultimap<IMDomain, IMMethod>();
	private final Map<IMDomain, Integer> uniqueDomMethods
	= new HashMap<IMDomain, Integer>();
	private int totalMethods = 0;
	private final List<Set<IMMethod>> domCountMethods
	= new ArrayList<Set<IMMethod>>();

	private final Multimap<IMDomain, IMBasicBlock> bbMap
	= new HashUniqueMultimap<IMDomain, IMBasicBlock>();
	private final Map<IMDomain, Integer> uniqueDomBBs
	= new HashMap<IMDomain, Integer>();
	private int totalBBs = 0;
	private final List<Set<IMBasicBlock>> domCountBBs
	= new ArrayList<Set<IMBasicBlock>>();

	@Override
	public String getName() {
		return "Statistics dump";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.hasOption("dump_stats");
	}

	@Override
	public boolean finish() throws CompileException {

		PrintStream out;
		try {
			out = new PrintStream(outPath + "/domain-code-stats.txt");
		} catch (FileNotFoundException e) {
			throw new CompileException(e.getMessage());
		}
		out.println("------------------------------------------");
		out.println("Total number of classes: " + totalClasses);
		out.println();
		out.println("Number of classes per domain:");
		for (IMDomain domain : domains) {
			out.println("  " + domain + ": " + classMap.get(domain).size()
					+ " (unique: " + uniqueDomClasses.get(domain) + ")");
		}
		out.println();
		out.println("Number of classes live in n domains:");
		for (int i = 1; i < domCountClasses.size(); ++i) {
			out.println("  " + i + ": " + domCountClasses.get(i).size());
		}
		out.println();

		out.println("------------------------------------------");
		out.println("Total number of methods " + totalMethods);
		out.println();
		out.println("Number of methods per domain:");
		for (IMDomain domain : domains) {
			out.println("  " + domain + ": " + methodMap.get(domain).size()
					+ " (unique: " + uniqueDomMethods.get(domain) + ")");
		}
		out.println();
		out.println("Number of methods live in n domains:");
		for (int i = 1; i < domCountMethods.size(); ++i) {
			out.println("  " + i + ": " + domCountMethods.get(i).size());
		}
		out.println();

		out.println("------------------------------------------");
		out.println("Total number of basic blocks: " + totalBBs);
		out.println();
		out.println("Number of basic blocks per domain:");
		for (IMDomain domain : domains) {
			out.println("  " + domain + ": " + bbMap.get(domain).size()
					+ " (unique: " + uniqueDomBBs.get(domain) + ")");
		}
		out.println();
		out.println("Number of basic blocks live in n domains:");
		for (int i = 1; i < domCountBBs.size(); ++i) {
			out.println("  " + i + ": " + domCountBBs.get(i).size());
		}
		out.close();

		try {
			out = new PrintStream(outPath + "/domain-code-stats.tex");
		} catch (FileNotFoundException e) {
			throw new CompileException(e.getMessage());
		}
		out.println("\\begin{tabular}{r|rr|rr|rr}");
		out.println("  \\textbf{Domain} & "
				+ "\\multicolumn{2}{c|}{\\textbf{Classes (Unique)}} & "
				+ "\\multicolumn{2}{c|}{\\textbf{Methods (Unique)}} & "
				+ "\\multicolumn{2}{c}{\\textbf{BBs (Unique)}} \\\\");
		out.println("  \\hline");
		for (IMDomain domain : domains) {
			int classes       = classMap.get(domain).size();
			int uniqueClasses = uniqueDomClasses.get(domain);
			int methods       = methodMap.get(domain).size();
			int uniqueMethods = uniqueDomMethods.get(domain);
			int bbs           = bbMap.get(domain).size();
			int uniqueBBs     = uniqueDomBBs.get(domain);
			out.println("  " + domain.toString().replace("_", "\\_") + " & "
					+ classes + " & (" + uniqueClasses + ", "
					+ getPercent(uniqueClasses, classes) + "~\\%) & " + methods
					+ " & (" + uniqueMethods + ", " + getPercent(uniqueMethods, methods)
					+ "~\\%) & " + bbs
					+ " & (" + uniqueBBs + ", " + getPercent(uniqueBBs, bbs)
					+ "~\\%) \\\\");
		}
		out.println("\\end{tabular}");
		out.close();

		try {
			out = new PrintStream(outPath + "/code-dist-stats.tex");
		} catch (FileNotFoundException e) {
			throw new CompileException(e.getMessage());
		}
		out.println("\\begin{tabular}{c|rl|rl|rl}");
		out.println("  \\textbf{\\# of domains} & "
				+ "\\multicolumn{2}{c|}{\\textbf{Classes}} & "
				+ "\\multicolumn{2}{c|}{\\textbf{Methods}} & "
				+ "\\multicolumn{2}{c}{\\textbf{BBs}} \\\\");
		out.println("  \\hline");
		for (int i = 1; i < domCountClasses.size(); ++i) {
			int classes  = domCountClasses.get(i).size();
			int methods  = domCountMethods.get(i).size();
			int bbs      = domCountBBs.get(i).size();
			out.println("  " + i + " & " 
					+ classes + " & (" + getPercent(classes, totalClasses)+ "~\\%) & " 
					+ methods + " & (" + getPercent(methods,totalMethods) + "~\\%) & " 
					+ bbs     + " & (" + getPercent(bbs, totalBBs) + "~\\%) \\\\");
		}
		out.println("  \\hline");
		out.println("  \\textbf{Total} & " + totalClasses + " & & " + totalMethods + " & " + totalBBs
				+ " & \\\\");
		out.println("\\end{tabular}");
		out.close();

		return false;
	}

	@Override
	public boolean runOnRepository(ClassStore repository)
			throws CompileException {
		BuilderOptions opts = repository.getOptions();
		outPath = opts.getOutputPath();
		domains = opts.getDomains();
		int domainCount = domains.length;
		domCountClasses.add(null);
		domCountMethods.add(null);
		domCountBBs.add(null);
		for (int i = 0; i < domainCount; ++i) {
			uniqueDomClasses.put(domains[i], 0);
			domCountClasses.add(new HashSet<IMClass>());
			uniqueDomMethods.put(domains[i], 0);
			domCountMethods.add(new HashSet<IMMethod>());
			uniqueDomBBs.put(domains[i], 0);
			domCountBBs.add(new HashSet<IMBasicBlock>());
		}
		return super.runOnRepository(repository);
	}

	@Override
	public boolean runOnClass(IMClass imClass) throws CompileException {
		int      classDomCount = 0;
		IMDomain classDom      = null;
		for (IMDomain domain : domains) {
			ReachabilityData rData = domain.getReachabilityData();
			if (rData.getClasses().contains(imClass) == false)
				continue;
			classMap.add(domain, imClass);
			++classDomCount;
			classDom = domain;
		}
		if (classDomCount > 0) {
			++totalClasses;
			if (classDomCount == 1)
				uniqueDomClasses.put(classDom, uniqueDomClasses.get(classDom) + 1);
			domCountClasses.get(classDomCount).add(imClass);
		}
		return super.runOnClass(imClass);
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {
		Set<IMDomain> methodDomains = new HashSet<IMDomain>();

		for (IMBasicBlock bb : method) {
			Set<IMDomain> bbDomains = bb.getDomains();

			methodDomains.addAll(bbDomains);

			for (IMDomain domain : bbDomains) {
				bbMap.add(domain, bb);
			}
			int bbDomCount = bbDomains.size();
			if (bbDomCount <= 0)
				continue;
			++totalBBs;
			if (bbDomCount == 1) {
				IMDomain domain = bbDomains.iterator().next();
				uniqueDomBBs.put(domain, uniqueDomBBs.get(domain) + 1);
			}
			domCountBBs.get(bbDomCount).add(bb);
		}

		if(methodDomains.size() > 0) {
			++totalMethods;
			for (IMDomain domain : methodDomains) {
				methodMap.add(domain, method);
			}

			if(methodDomains.size() == 1) {
				IMDomain domain = methodDomains.iterator().next();
				uniqueDomMethods.put(domain, uniqueDomMethods.get(domain) + 1);
			}
			domCountMethods.get(methodDomains.size()).add(method);
		}

		return false;
	}

	private static int getPercent(int numerator, int denominator) {
		return Math.round(100f * numerator / denominator);
	}

}
