/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.dump;

import java.io.*;

import keso.classfile.*;
import keso.classfile.datatypes.*;
import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class UMLDiagramWriter extends MethodPass {

	private PrintStream out;

	@Override
	public String getName() {
		return "UML diagram writer";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.createUMLDiagram();
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.setIsDump(true);
	}

	@Override
	public boolean runOnClass(IMClass imClass) throws CompileException {

		if (imClass instanceof IMArrayClass)
			return false;

		BuilderOptions opts = imClass.getOptions();
		String         name = imClass.getClassName();
		opts.verbose("###### Writing Classes as TeX " + name);

		String outPath      = opts.getOutputPath();
		String escClassName = DecoratedNames.toEscString(name);
		try {
			out = new PrintStream(outPath + "/UML_" + escClassName + ".tex");
		} catch (IOException e) {
			throw new CompileException(e.getMessage());
		}

		out.println("\\documentclass[a4paper,14pt]{book}");

		out.println("\\usepackage{geometry}");
		out.println("\\usepackage{tabularx}");
		out.println("\\begin{document}");

		out.print("%% UML Diagram ");
		out.println(name);
		out.println("\\begin{tabular}{|l|}");
		out.println("\\hline");
		String cName = DecoratedNames.baseName(name);
		out.print(DecoratedNames.toEscTexString(cName));
		out.println(" \\\\");
		out.println("\\hline");

		FieldData[] fields = imClass.getFieldData();
		if (fields == null || fields.length == 0) {
			out.println(" \\\\");
		} else {
			for (FieldData field : fields) {
				String type = field.getType().substring(0, 1);
				out.print(DecoratedNames.toEscTexString(type));
				out.print(" ");
				out.print(DecoratedNames.toEscTexString(field.getName()));
				out.println(" \\\\");
			}
		}
		out.println("\\hline");
		for (IMMethod method : imClass.getMethods()) {
			runOnMethod(method);
		}
		out.println("\\hline");
		out.println("\\end{tabular}");
		out.println("\\end{document}");

		out.close();
		return false;
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method.isClassInit())
			return false;

		MethodData source = method.getSource();
		if (method.isConstructor()) {
			String cName = DecoratedNames.baseName(method.getClassName());
			out.print(DecoratedNames.toEscTexString(cName));
		} else {
			String type = source.getReturnType().getJavaLanguageType();
			out.print(DecoratedNames.toEscTexString(type));
			out.print(' ');
			out.print(DecoratedNames.toEscTexString(method.getMethodName()));
		}
		out.print('(');
		BasicTypeDescriptor[] params = source.getParameterTypes();
		for (int i = 0; i < params.length; ++i) {
			if (i > 0)
				out.print(',');
			String type = params[i].getJavaLanguageType();
			out.print(DecoratedNames.toEscTexString(type));
		}
		out.print(')');
		out.println(" \\\\");
		return false;
	}

}
