/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.datatypes.*;
import keso.compiler.*;
import keso.compiler.backend.*;

import java.io.*;
import java.util.*;

public class IMBasicBlock extends IMNode implements Iterable<IMNode> {

	private final class Itr implements Iterator<IMNode> {

		private IMNode curr;

		private Itr() {
			if (bcList != null)
				curr = bcList.next();
		}

		@Override
		public boolean hasNext() {
			return (curr != null);
		}

		@Override
		public IMNode next() {
			if (hasNext() == false)
				throw new NoSuchElementException();
			IMNode prev = curr;
			curr = curr.next();
			return prev;
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

	private final static boolean verbose_label = true;

	boolean done = false;
	private boolean warn_once = false;

	private String copylabel = null;

	private int dfnum = 0;

	private IMBasicBlockList pred = new IMBasicBlockList();
	private IMBasicBlockList succ = new IMBasicBlockList();

	private IMNode[] enterStack;
	private IMNode[] leaveStack;
	private IMNode   bcList;

	private IMExceptionHandler exphandler;
	private boolean  subroutine;

	private SortedSet<IMSlot> liveIn  = new TreeSet<IMSlot>();
	private SortedSet<IMSlot> phiIn   = new TreeSet<IMSlot>();
	private SortedSet<IMSlot> liveOut = new TreeSet<IMSlot>();

	/** Set of domains where this basic block is live. */
	private SortedSet<IMDomain> domains = new TreeSet<IMDomain>();

	protected IMBasicBlock(IMMethod new_method, IMBasicBlock orig) {
		super(orig.src_method,0,-1,orig.src_bcPosition);
		method = new_method;
	}

	public IMBasicBlock(IMMethod method, int bcpos) {
		super(method,0,-1,bcpos);
		subroutine = false;
	}

	public void setDFNum(int dfnum) {
		this.dfnum=dfnum;
	}

	public int getDFNum() {
		return dfnum;
	}

	/**
	 * Insert a list of IMNode objects. The list must start with
	 * an IMListHead object which is skipped.
	 */
	public void insertBCBefore(IMNode preList) throws CompileException {
		if (preList == null || preList.next() == null)
			return;
		if (bcList == null || bcList.next() == null) {
			bcList = preList;
			return;
		}
		IMNode node = preList;
		while (node.next() != null) {
			node = node.next();
		}
		node.append(bcList.next());
		bcList = preList;
	}

	/**
	 * Append an IMNode object at the end of the basic block.
	 */
	public void appendBC(IMNode node) throws CompileException {
		IMNode last = getLastInstruction();
		if (last == null) {
			last   = new IMListHead();
			bcList = last;
		}
		last.append(node);
	}

	public void appendBCList(IMNode postList) {
		IMNode last = getLastInstruction();
		if (last == null) {
			last   = new IMListHead();
			bcList = last;
		}
		last.append_list(postList);
	}

	/**
	 * Append an IMNode object at the end of the basic block.
	 */
	public void appendBeforeBranch(IMNode node) throws CompileException {
		IMNode prev = bcList;
		while (prev!=null && prev.next()!=null && !prev.next().isBranch()) prev = prev.next();
		if (prev!=null) node.append(prev.next());
		prev = prev.append(node);
	}

	/**
	 * Append an IMNode at the top of a basic block.
	 */
	public void placePhi(IMNode node) throws CompileException {
		if (bcList == null)
			bcList = new IMListHead();
		else
			node.append(bcList.next());
		bcList.append(node);
	}

	public void addBC(IMNode node) throws CompileException {
		if (bcList == null)
			bcList = new IMListHead();
		bcList.append(node);
	}

	public IMBasicBlockList processBasicBlock(VirtualOperandStack stack) throws CompileException {
		done = true;

		stack.init(enterStack);

		if (exphandler!=null) stack.push(new IMCaughtException(method, exphandler, bcPosition));

		IMNode new_list = new IMListHead();
		IMNode new_list_prev = null;
		IMNode new_list_end = new_list;

		IMNode prev = bcList;
		IMListHead extra_ops = new IMListHead();
		while (prev!=null && prev.next() != null) {
			IMNode curr = prev.next();

			if (opts.vvverbose()) opts.vvverbose("# VS "+method.getMethodName()+":"+curr.getBCPosition()+" "+curr.dumpBC());
			IMNode n = curr.processStack(stack, extra_ops);

			new_list_end = new_list_end.append_list(extra_ops);
			extra_ops.clear_link();

			if (n!=null) {
				new_list_prev = new_list_end;
				curr = new_list_end = new_list_end.append_list(n);
			}

			if (curr instanceof IMReturnSubroutine) {
				subroutine = true;
				succ.add(((IMBranch)curr).getUniqTargets());
				curr.clear_link();
				break;
			}

			if (curr.isBranch()) {
				succ.add(((IMBranch)curr).getUniqTargets());
				curr.clear_link();
				opts.vvverbose("# branch (goto, if, switch)");
				break;
			}

			if (curr instanceof IMThrow) {
				curr.clear_link();
				opts.vvverbose("# end of basic block (throw)");
				new_list_prev = new_list_end;
				break;
			}

			if (curr.isEndOfBasicBlock()) {
				succ.add((IMBasicBlock)next());
				//curr.clear_link();
				opts.vvverbose("# end of basic block ("+curr.dumpBC()+")");
				new_list_end.clear_link();
				new_list_prev = new_list_end;
				break;
			}

			prev = curr;
		}

		bcList = new_list;

		if (new_list_prev==null) new_list_prev = new_list_end;
		leaveStack = stack.leave(new_list_prev);
		for (int i=0;i<succ.length();i++) {
			IMBasicBlock s = succ.at(i);
			s.addPred(this);
			if (s.done) continue;
			s.setInitStack(leaveStack);
		}

		if (opts.vvverbose()) opts.vvverbose("# END succ "+succ+" leave "+dumpArray(leaveStack));
		return succ;
	}

	@Override
	public boolean isBasicBlock() { return true; }

	final public IMBasicBlockList getSucc() { return succ; }

	final public IMBasicBlockList getPred() { return pred; }

	final public void addPred(IMBasicBlock pred) { this.pred.add(pred); }

	final public IMBasicBlock removeDeadBasicBlocks(IMBasicBlock pre) throws CompileException {

		if (pre==null) return this;
		if (pred.length()==0) {
			if (succ.length()!=0)
			{
				opts.warn("call graph mix up! "+method.lineInfo(bcPosition));
			}
			pre.next_node = next_node;

			IMDeadCodeVisitor dv = new IMDeadCodeVisitor(method);
			IMNode prev = bcList;
			while (prev!=null && prev.next() != null) {
				prev = prev.next();
				prev.visitNodes(dv);
			}

			bcList = null;
			return pre;
		}

		return this;
	}

	final public void unlinkPred(IMBasicBlock bb) {
		unlinkPred(bb, true);
	}

	final public void unlinkPred(IMBasicBlock bb, boolean affectSuccs) {
		pred.remove(bb);
		// TODO: update Phi-functions, remove vars
		if (affectSuccs == true && pred.length() == 0) {
			unlinkAllSucc();
		}
	}

	final public void unlinkAllSucc() {
		for (int i=0; i<succ.length(); i++) {
			IMBasicBlock nb = succ.at(i);
			nb.unlinkPred(this);
		}
		succ.clear();
	}

	final public void unlinkSucc(int nr) {

		IMBasicBlock nb = succ.at(nr);

		// The sequence of the next two statements matters!
		// If inverting it, we would run into a StackOverflowException if two
		// blocks form a small loop.
		succ.remove(nb);
		nb.unlinkPred(this);
	}

	final public void setInitStack(IMNode[] stack) {
		enterStack = stack;
	}

	final public IMNode[] getLeaveStack() {
		return leaveStack;
	}

	public void setExceptionHandler(IMExceptionHandler exphandler) {
		this.exphandler = exphandler;
		if (exphandler.getClassName()!=null) {
			method.requireClass(0, exphandler.getClassName());
		}
	}

	public boolean isExceptionHandler() {
		return exphandler!=null;
	}

	public boolean isSubroutine() {
		return subroutine;
	}

	/**
	* The return value is true if the basic block bb is the direct and only
	* predecessor of the basic block and if it is possible to dismiss the label
	* of the basic block.
	*/
	public boolean isFallingThrough(IMBasicBlock bb) {

		if (next()!=bb) return false;

		if (succ.at(0)!=bb) return false;

		IMNode prev = bcList;
		while (prev!=null && prev.next() != null) {
			IMNode curr = prev.next();
			// The switches won't allow a fall through.
			if (curr instanceof IMLookupSwitch || curr instanceof IMTableSwitch) return false;
			prev = curr;
		}

		return true;
	}

	public boolean empty() {
		return (bcList==null || bcList.next()==null);
	}

	public String toNodeName() {
		return toLabel();
	}

	private String label=null;

	public String toLabel() {
		if (label==null) {
			String ret = "";
			if (copylabel != null)
				ret += copylabel;
			if (src_method != null)
				ret += src_method.getAlias();
			else
				ret += "unknown";
			ret += "_B" + getNonNegative(src_bcPosition);
			label = ret;
		}
		return label;
	}

	final public String getShortLabel() {
		String result = "_";
		if (copylabel != null)
			result += copylabel;
		return result + "B" + getNonNegative(src_bcPosition);
	}

	@Override
	final public String toReadableString() {

		StringBuffer s = new StringBuffer();

		s.append(toLabel());
		s.append(":\n");

		IMNode prev = bcList;
		while (prev!=null && prev.next() != null) {
			IMNode curr = prev.next();
			try {
				s.append("\t");
				s.append(curr.toReadableString());
				s.append(";\n");
			} catch (NullPointerException ex) {
				s.append("\t<NullPointerException> "+curr.getBCPosition()+": "+curr.dumpBC());
			}
			prev = curr;
		}

		return s.toString();
	}

	@Override
	public void visitNodes(IMVisitor visitor) throws CompileException {
		if (!done) {
			if (!warn_once) opts.warn(method.lineInfo(getBCPosition())+": basic block not processed yet! - ignore");
			warn_once = true;
			return;
		}
		IMNode prev = bcList;

		while (prev!=null && prev.next() != null) {
			try {
				prev = prev.next();
				prev.visitNodes(visitor);
			} catch (RuntimeException ex) {
				printError(ex, prev);
				throw ex;
			} catch (CompileException ex) {
				printError(ex, prev);
				throw ex;
			}
		}
	}

	public final void removeNode(IMNode remove) throws CompileException {
		IMNode prev = bcList;
		while (prev!=null && prev.next() != null) {
			if (prev.next() == remove) {
				prev.append(remove.next());
				return;
			}
			prev = prev.next();
		}
	}

	public final void replaceNode(IMNode oldNode, IMNode newNode)
	throws CompileException {
		if (oldNode == newNode)
			return;
		if (newNode != null)
			oldNode.insert(newNode);
		removeNode(oldNode);
	}

	public void markAsCopied() {
		copylabel = method.getNewCopyLabel();
	}

	@Override
	public IMNode copy(IMCopyVisitor visitor) throws CompileException {
		return visitor.updateBlock(this);
	}

	final public IMBasicBlock createStub(IMMethod host_method, IMCopyVisitor visitor) throws CompileException {
		IMBasicBlock nbb = new IMBasicBlock(host_method, this);
		nbb.subroutine = subroutine;
		return nbb;
	}

	final public IMNode updateStub(IMBasicBlock nbb, IMCopyVisitor visitor) throws CompileException {
		nbb.copyNodeValues(visitor, this);
		nbb.done = done;
		// FIXME: exception handler not supported yet.
		if (exphandler!=null) throw new Error();
		nbb.exphandler = null;
		nbb.subroutine = subroutine;
		nbb.pred = pred.copy(visitor);
		nbb.succ = succ.copy(visitor);
		nbb.bcList = copyBCList(visitor);
		if (next_node!=null) {
			nbb.next_node = ((IMBasicBlock)next_node).copy(visitor);
		}
		return nbb;
	}

	final public IMNode copyBCList(IMCopyVisitor visitor) throws CompileException {
		IMNode prev = bcList;

		if (bcList==null) return null;

		if (!(prev instanceof IMListHead)) throw new CompileException("not a list "+bcList);

		IMNode new_list = new IMListHead();
		IMNode tail=new_list;
		while (prev!=null && prev.next() != null) {
			IMNode curr = prev.next();
			tail = tail.append(curr.copy(visitor));
			prev = curr;
		}

		return new_list;
	}

	/**
	 * Insert the code from the callee into this basic block by splitting
	 * the basic block at prev_pos.
	 *
	 * prev_pos: ...
	 * current : INVOKE callee <=> body callee
	 * tail	   : ...
	 *
	 */
	final public IMBasicBlock insertCode(IMNode prev_pos, IMNode call, IMMethod callee, IMNode self, IMNode args[]) throws CompileException {
		IMNode current = prev_pos.next();

		// split basic block at the prev_pos statement
		IMBasicBlock tail = split(prev_pos);
		callee.insertCode(method, current, call, self, args, this, tail);

		return tail;
	}

	final public void link_nofollow(IMBasicBlock next_bb) throws CompileException {
		if (succ.length()>1) throw new CompileException();
		succ = new IMBasicBlockList(next_bb);
		next_bb.pred = new IMBasicBlockList(this);
	}

	/**
	 * link the given basic block (next_bb) at (this) basic block.
	 */
	final public void link(IMBasicBlock next_bb) throws CompileException {
		if (succ.length()>1) throw new CompileException();
		succ = new IMBasicBlockList(next_bb);
		if (next_bb.pred.length()>0) {
			next_bb.addPred(this);
		} else {
			next_bb.pred = new IMBasicBlockList(this);
		}
		next_node = next_bb;
	}

	final public void updatePred(IMBasicBlock old_bb, IMBasicBlock new_bb) {
		pred.update(old_bb, new_bb);
	}

	final public void updateSucc(IMBasicBlock old_bb, IMBasicBlock new_bb) {
		succ.update(old_bb, new_bb);
	}

	/**
	 * Splits the basic block into two pieces. The byte code is split up behind
	 * the position given by prevPos. The prevPos node itself will neither
	 * belong to the first nor to the second basic block.
	 * @param prevPos Node after which the block is to be split.
	 * @return The second block, starting after prevPos.
	 */
	final public IMBasicBlock split(IMNode prevPos) throws CompileException {

		// TODO: exception handler not supported yet.
		if (exphandler!=null) throw new Error();

		IMBasicBlock tail = new IMBasicBlock(method,this);

		tail.done = done;
		tail.subroutine = subroutine;

		// predecessor
		tail.pred = new IMBasicBlockList(this);

		// successor
		for (int i=0;i<succ.length();i++) { succ.at(i).updatePred(this,tail); }
		tail.succ = succ;
		succ = new IMBasicBlockList(tail);
		// update next_node
		tail.next_node = next_node;
		next_node = tail;

		// split byte code list at current position
		if (prevPos == null)
			prevPos = bcList;
		IMNode curr = prevPos.next();
		prevPos.clear_link();
		opts.vverbose("split at "+curr.lineInfo()+" "+curr.toReadableString());

		tail.src_bcPosition = curr.src_bcPosition;

		if (curr.next()==null) {
			tail.bcList = null;
		} else {
			tail.bcList = new IMListHead();
			tail.bcList.append(curr.next());
			curr.clear_link();
		}

		return tail;
	}

	/**
	 * Merges this basic block with the next one if this block has exactly one
	 * successor and at the same time is the only predecessor of it.
	 */
	final public IMBasicBlock melt() throws CompileException {

		if (succ.length() == 1 && succ.at(0)==next_node) {
			
			// If the basic block contains a conditional branch whose targets
			// are all identical, replace it with a Goto to mark the place where
			// the code of the successor block is to be appended. If this
			// conversion were not performed, we would end up with a branch into
			// Nirvana.
			for (IMNode root : this) {
				if (root.isBranch() == false)
					continue;
				IMBasicBlock[] targets      = ((IMBranch) root).getTargets();
				IMBasicBlock   firstTarget  = targets[0];
				boolean        equalTargets = true;
				for (int i = 1; i < targets.length; ++i) {
					if (targets[i] != firstTarget) {
						equalTargets = false;
						break;
					}
				}
				if (equalTargets == false)
					continue;
				replaceNode(root,
						method.createGoto(firstTarget, root.getBCPosition()));
			}
			
			IMBasicBlock nb = succ.at(0);
			if (nb==method.getEndBlock()) return nb;
			if (nb.pred.length()==1 && nb.pred.at(0)==this) {
				succ = nb.succ;
				for (int i=0;i<succ.length();i++) {
					succ.at(i).updatePred(nb,this);
				}
				next_node = nb.next_node;

				IMNode node = bcList;
				while (node !=null && node.next()!=null) {
					if (node.next() instanceof IMGoto) break;
					node=node.next();
				}
				node.append_list(nb.bcList);

				return this;
			}
		}

		return (IMBasicBlock)next_node;
	}

	/**
	 * Perform constant folding and other optimizations.
	 */
	@Override
	final public IMNode ssa_optimize() throws CompileException {

		if (!done) {
			if (!warn_once) opts.warn(method.lineInfo(getBCPosition())+": basic block not processed yet! - ignore");
			warn_once = true;
			return next();
		}

		if (bcList==null || bcList.next()==null) return (IMBasicBlock)next();

		IMNode prev = bcList;
		for (IMNode curr=prev.next();curr!=null;curr=curr.next()) {
			try {
				IMNode updatenode = curr.ssa_optimize();
				if (updatenode==null) {
					// remove current node
					prev.append(curr.next());
					curr=prev;
				} else if (updatenode!=curr) {
					// replace current node with a new node or a list
					// of new nodes.
					prev=prev.append_list(updatenode);
					prev.append(curr.next());
					curr=prev;
				}
				prev = curr;
			} catch (RuntimeException ex) {
				printError(ex, curr);
				throw ex;
			} catch (CompileException ex) {
				printError(ex, curr);
				throw ex;
			}
		}

		if (opts.hasOption("ssa_melt")) return melt();
		return next();
	}

	final public IMBasicBlock inlineMethodCalls() throws CompileException {

		IMNode curr=null;
		try {
			if (bcList==null || bcList.next()==null) return (IMBasicBlock)next();

			IMNode prev = bcList;
			for (curr=prev.next();curr!=null;curr=curr.next()) {
				IMBasicBlock tail=curr.inlineMethodCalls(this, prev);
				if (tail!=null) {
					return tail;
				}
				prev = curr;
			}

		} catch (RuntimeException ex) {
			printError(ex, curr);
			throw ex;
		} catch (CompileException ex) {
			printError(ex, curr);
			throw ex;
		}

		return (IMBasicBlock)next();
	}

	/**
	 * Compute a heuristic value for the complexity of the basic block.
	 */
	@Override
	final public int costs() throws CompileException {
		int costs = 0;
		IMNode prev = bcList;

		while (prev!=null && prev.next() != null) {
			prev = prev.next();
			costs += prev.costs();
		}

		return costs;
	}

	final private void emit_source_line(String pre, StringBuffer buffer, int line) throws CompileException {
		if (line<0) return;
		try {
			IMClass clazz = method.getIMClass();
			StringBuffer src = new StringBuffer();
			src.append(clazz.readSourceLine(line));
			int pos=0;
			while ((pos=src.indexOf("*/"))!=-1)
				src.setCharAt(pos,'o');
			while ((pos=src.indexOf("/*"))!=-1)
				src.setCharAt(pos+1,'o');
			while ((pos=src.indexOf("\t"))!=-1)
				src.setCharAt(pos,' ');
			buffer.append(pre);
			buffer.append(" ");
			buffer.append(line);
			buffer.append(": ");
			buffer.append(src);
			buffer.append("\n");
		} catch (Exception ex) {
			opts.vwarn(ex+" "+method.getMethodNameAndType()+" line: "+line);
		}
	}

	final public int emit_source(Coder coder, int pos) throws CompileException {
		IMNode curr=null;
		int curr_line = pos;
		try {
			boolean has_info=false;
			StringBuffer dbg=new StringBuffer();
			dbg.append("\n/*==============================================================/\n");
			IMNode prev = bcList;
			while (prev!=null && prev.next() != null) {
				curr = prev.next();

				int line = method.getLineNumber(curr.getBCPosition());
				if (line>curr_line) {
					if (curr_line!=-1) {
						for (int i=4;i>0;i--)
							if (line>(curr_line+i)) emit_source_line(":",dbg, line-i);
					}
					emit_source_line("|",dbg, line);
					curr_line=line;
					has_info=true;
				}
				prev = curr;
			}
			dbg.append("/==============================================================*/");
			if (has_info) coder.addln(dbg);
		} catch (RuntimeException ex) {
			printError(ex, curr);
			throw ex;
		} catch (CompileException ex) {
			printError(ex, curr);
			throw ex;
		}

		return curr_line;
	}

	final public int translate(Coder coder, int curr_line) throws CompileException {

		if (!done) {
			if (!warn_once) opts.warn(method.lineInfo(getBCPosition())+": basic block not processed yet! - ignore");
			warn_once = true;
			return curr_line;
		}

		if ( pred.length()==0 ||
				(pred.length()==1 && pred.at(0).isFallingThrough(this))) {
			if (opts.hasDbgSymbols()) {
				coder.add("\n/* ");
				coder.add(toLabel());
				coder.add(": ");
				if (verbose_label) {
					coder.add(" Pred: ");
					coder.add(pred);
					coder.add(" No: ");
					coder.add(dfnum);
					coder.add((done?" done":" NOT done"));
				}
				coder.addln(" */");
			}
		} else {
			coder.add("\n");
			coder.add(toLabel());

			if (verbose_label) {
				coder.add(": /* Pred: ");
				coder.add(pred);
				coder.add(" No: ");
				coder.add(dfnum);
				coder.add((done?" done":" NOT done"));
				coder.addln(" */");
			} else {
				coder.addln(":");
			}
		}

		if (opts.hasOption("dbg_src")) curr_line=emit_source(coder,curr_line);

		if (bcList==null || bcList.next()==null) return curr_line;

		coder.setCurrentBasicBlock(this);

		opts.vvverbose("# BB "+toReadableString());

		IMNode curr=null;
		try {
			IMNode prev = bcList;
			while (prev!=null && prev.next() != null) {
				curr = prev.next();

				if (opts.hasOption("dbg_line")) {
					int line = method.getLineNumber(curr.getBCPosition());
					if (line>=0 && line!=curr_line) {
						IMClass clazz = method.getIMClass();
						coder.add("#line ");
						coder.add(line);
						coder.add(" \"src/");
						coder.add(clazz.getSourceFile());
						coder.addln("\"");
						curr_line=line;
					}
				}
				if (false && opts.hasVerbose()) {
					if (!(curr instanceof IMGoto))
						coder.add_before("/* "+curr.toReadableString()+" */");
				}
			
				curr.translate(coder);
				//System.err.println("IMINSTANCE IS " + curr.toString());
				if (curr.needSemicolon()) coder.addln(";");
			
				if(opts.hasOption("use_coded_refs")) {
					/* Get the object identifier (for parity calculation) */
					//if(curr instanceof IMStoreLocalVariable || curr instanceof IMStoreArray && curr.datatype == BCBasicDatatype.REFERENCE) {
					if(curr instanceof IMStoreLocalVariable && curr.datatype == BCBasicDatatype.REFERENCE) {
						IMStoreLocalVariable obj_n = (IMStoreLocalVariable) curr;
						if(obj_n.operand instanceof IMNew || obj_n.operand instanceof IMNewArray || obj_n.operand instanceof IMNewObjArray || obj_n.operand instanceof IMNewMultiArray) {
							if(opts.hasOption("debug_FT")) {
								System.err.println("BASIC " + curr.toReadableString() + " " + obj_n.var + " " + obj_n.operand);
								coder.add("KESO_PRINTF(\"1 obj1_0 adress is %x\\n\","+ obj_n.var+");\n");
							}
							// compute parity of reference and set parity bit
							coder.add("\tSET_PARITY(" + obj_n.var + ");\n");
							if(opts.hasOption("debug_FT")) {
								coder.add("KESO_PRINTF(\"2 obj1_0 adress is %x\\n\","+ obj_n.var+");\n");
								coder.add("KESO_PRINTF(\"native adress is %x\\n\", NATIVE_ADDR("+ obj_n.var+"));\n");
							}
						}
					}

					//if(curr instanceof IMStoreArray) {
					//	IMStoreArray obj_n = (IMStoreArray) curr;
					//	if(obj_n.aOpr instanceof IMNew || obj_n.aOpr	instanceof IMNewArray || obj_n.aOpr instanceof IMNewObjArray || obj_n.aOpr instanceof IMNewMultiArray) {
					//		System.err.println("FOO " + curr.toReadableString() + " " + obj_n.rvalue + " " + obj_n.rvalue);
					// 		//compute parity of reference and set parity bit
					//		coder.add("KESO_PRINTF(\"1 obj1_0 adress is %x\\n\","+ obj_n.var+");\n");
					//		coder.add("\tFOO(" + obj_n.rvalue + ");\n");
					//		coder.add("KESO_PRINTF(\"2 obj1_0 adress is %x\\n\","+ obj_n.var+");\n");
					//		coder.add("KESO_PRINTF(\"native adress is %x\\n\", NATIVE_ADDR("+ obj_n.var+"));\n");
					//}
					//	}
					if(opts.hasOption("debug_FT")) {
						if(curr instanceof IMInvokeVirtual || curr instanceof IMInvokeInterface) {
							int mt_off = method.getMethodTable().get_mt_offset();
							IMInvoke invoke_node = (IMInvoke) curr;
							coder.add("\nKESO_PRINTF(\"Invoke method on obj 0x%x with class_id %d with number %d\\n\", (NATIVE_ADDR(");
							coder.add(invoke_node.obj.toReadableString());
							coder.add(")), (NATIVE_ADDR(");
							coder.add(invoke_node.obj.toReadableString());
							coder.add("))->class_id, (NATIVE_ADDR(");
							coder.add(invoke_node.obj.toReadableString());
							coder.add("))->class_id+"+mt_off+");\n");
						}
					}	
					}
				// parity end
				prev = curr;
			}
		} catch (RuntimeException ex) {
			printError(ex, curr);
			throw ex;
		} catch (CompileException ex) {
			printError(ex, curr);
			throw ex;
		}

		return curr_line;
	}

	final private void printError(Throwable ex, IMNode curr) {
		System.err.println("========================================================");
		System.err.println(ex+" in: ");
		System.err.print(method.getClassName());
		System.err.print(".java  ");
		System.err.print(method.getMethodNameAndType());
		if (curr!=null) {
			int line = method.getLineNumber(curr.getBCPosition());
			if (line>=0) System.err.println(" (line: "+line+")");
			else System.err.println(" (bc: "+curr.getBCPosition()+")");
		}
		System.err.print(method.getAlias());
		System.err.println(".c");
		System.err.println("========================================================");
	}

	@Override
	final public String dumpBC() {
		String result = getShortLabel() + ": (";
		if (done == false)
			result += "not ";
		return result + "done; LiveIn: " + liveIn + "; PhiIn: " + phiIn
		+ "; LiveOut: " + liveOut + ')';
	}

	final public void writeCFGEdges(PrintStream out) {
		for (int i=0; i<succ.length();i++) {
			IMBasicBlock nb = succ.at(i);
			out.print("\t");
			out.print(toNodeName());
			out.print("->");
			out.print(nb.toNodeName());
			out.print(" [");
			if (succ.length()>1) {
				out.print("label=\" ");
				out.print(i);
				out.print("\"");
			}
			out.print(" style=bold]");

			out.println(";");
		}
		if (opts.hasOption("cfg_pred")) {
			for (int i=0; i<pred.length();i++) {
				IMBasicBlock nb = pred.at(i);
				out.print("\t");
				out.print(toNodeName());
				out.print("->");
				out.print(nb.toNodeName());
				out.print(" [label=\" ");
				out.print(i);
				out.print("\" style=dotted]");
				out.println(";");
			}
		}

		if (opts.hasOption("cfg_show_order")) {
			IMBasicBlock nb = (IMBasicBlock)next();
			if (nb!=null) {
				out.print("\t");
				out.print(toNodeName());
				out.print("->");
				out.print(nb.toNodeName());
				out.print(" [style=dashed]");
				out.println(";");
			}
		} else if (opts.hasOption("cfg_in_order")) {
			IMBasicBlock nb = (IMBasicBlock)next();
			if (nb!=null) {
				if (!nb.equals(succ.at(0))) {
					out.print("\t");
					out.print(toNodeName());
					out.print("->");
					out.print(nb.toNodeName());
					out.print(" [style=invis]");
					out.println(";");
				}
			}
		}
		out.println();
	}

	final public void writeCFGBoxes(PrintStream out) throws CompileException {
		out.print("\t");
		out.print(toNodeName());

		if (opts.hasOption("cfg_small")) {
			out.print(" [label=\"");
			out.print(toLabel());
			out.print("\" fontsize=8 ");
			if (copylabel==null) out.print(" style=rounded ");
			if (pred.length()==0 || succ.length()==0) {
				out.print("shape=ellipse");
			} else {
				if (succ.length()==2) out.print("shape=invhouse");
				if (succ.length()>2) out.print("shape=hexagon");
			}
			out.println("]");
		} else {
			writePseudoCode(out);
		}
	}

	@Override
	final public String toString() {
		return toLabel();
	}

	final private void writePseudoCode(PrintStream out) {
		boolean str=false;

		out.print(" [label=\"");
		out.print("\\N:\\l ");

		if (!opts.hasOption("cfg_no_code")) {
			IMNode prev = bcList;
			while (prev!=null && prev.next() != null) {
				IMNode curr = prev.next();
				StringBuffer s = new StringBuffer(curr.toReadableString());
				for (int i=0;i<s.length();i++) {
					switch (s.charAt(i)) {
						case '"':
							s.setCharAt(i,'\'');
							str=(str?false:true);
							break;
						case '\t':
							s.setCharAt(i,' ');
							s.insert(i,"    ");
							break;
						case '\n':
							s.setCharAt(i,'.');
							break;
					}
				}
				out.print(s.toString());
				out.print("\\l ");
				prev = curr;
			}
		}

		out.print("\" fontsize=9 ");

		out.print("fillcolor=grey98, style=\"filled\", ");
		if (pred.length()==0 || succ.length()==0) {
			out.print("shape=ellipse");
		} else {
			if (succ.length()==2) out.print("shape=invhouse");
			if (succ.length()>2) out.print("shape=hexagon");
		}
		out.println("]");
	}

	private String dumpArray(IMNode[] arr) {
		if (arr==null) return "null";

		if (arr.length<1) return " - ";

		StringBuffer ret = new StringBuffer((arr[0]==null?" --- ":arr[0].toReadableString()));
		for (int i=1;i<arr.length;i++) { ret.append(','); ret.append((arr[i]==null?" --- ":arr[i].toReadableString())); }

		return ret.toString();
	}

	@Override
	public Iterator<IMNode> iterator() {
		return new Itr();
	}

	public void clearLivenessInfo() {
		liveIn.clear();
		phiIn.clear();
		liveOut.clear();
	}

	public void addLiveIn(IMSlot slot) {
		liveIn.add(slot);
	}

	public void addLiveIn(Collection<IMSlot> slots) {
		liveIn.addAll(slots);
	}

	public void removeLiveIn(IMSlot slot) {
		liveIn.remove(slot);
	}

	public void addPhiIn(IMSlot slot) {
		phiIn.add(slot);
	}

	public void removePhiIn(IMSlot slot) {
		phiIn.remove(slot);
	}

	public void addLiveOut(IMSlot slot) {
		liveOut.add(slot);
	}

	public void removeLiveOut(IMSlot slot) {
		liveOut.remove(slot);
	}

	public Set<IMSlot> getLiveInSlots() {
		return Collections.unmodifiableSet(liveIn);
	}

	public Set<IMSlot> getPhiInSlots() {
		return Collections.unmodifiableSet(phiIn);
	}

	public Set<IMSlot> getLiveOutSlots() {
		return Collections.unmodifiableSet(liveOut);
	}

	/** Clears the set of domains where this basic block is live. */
	public void clearDomains() {
		domains.clear();
	}

	/**
	 * Adds a domain to the set of domains where this basic block is live.
	 * @param domain Domain to be added.
	 */
	public void addDomain(IMDomain domain) {
		domains.add(domain);
		method.addDomain(domain);
	}

	/**
	 * Returns the set of domains where this basic block is live.
	 * @return Domain set.
	 */
	public Set<IMDomain> getDomains() {
		return Collections.unmodifiableSet(domains);
	}

	private IMNode getLastInstruction() {
		IMNode curr = bcList;
		while (curr != null && curr.next() != null) {
			curr = curr.next();
		}
		return curr;
	}

	private static int getNonNegative(int x) {
		return (x >= 0) ? x : 0;
	}

}
