/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import java.io.*;
import java.util.*;

import keso.compiler.*;

/** Intermediate code writer class. */
public class IMCodeWriter extends IMBasicVisitor {

	private Deque<IMNode> stack = new ArrayDeque<IMNode>();

	@Override
	public void visit(IMNode node) throws CompileException {

		// All IMNodes are visited before their operands, so we push the visited
		// nodes onto a stack and print them in reverse order.
		// The only exception are the subclasses of IMStoreLocalVariable, which
		// do it the other way round (what the heck, why?) - so we have to add
		// them to the bottom of the stack.
		if (node instanceof IMStoreLocalVariable)
			stack.addLast(node);
		else
			stack.push(node);
	}

	public void writeMethod(IMMethod method, PrintStream out)
	throws CompileException {

		out.println(method.getMethodNameAndType() + " {");
		int count = 0;

		for (IMBasicBlock bb : method) {

			// Print basic block label
			out.println(bb.dumpBC());

			for (IMNode root : bb) {

				// Collect all nodes in the current basic block
				List<IMNode> flattened
				= IMCodeFlattener.getFlattenedCode(root);

				// Assign virtual register IDs
				// Not all instructions actually produce a result on the stack -
				// however, for the sake of simplicity and readability, a
				// virtual register is printed out for them as well.
				for (IMNode f : flattened) {
					f.setRegId(count++);
				}

				// Print the nodes in their correct order
				for (IMNode node : flattened) {
					out.printf("%6d: %5s = %s", node.getBCPosition(),
							node.getRegId(), node.dumpBC());
					String extra = getExtraString(node);
					if (extra != null)
						out.print(extra);
					out.println();
				}
			}
		}

		out.println('}');
	}

	protected String getExtraString(IMNode node) {
		return null;
	}

}
