/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.datatypes.*;
import keso.compiler.*;

public class IMDataFlow {

	public final static IMDataFlowAny ANY = new IMDataFlowAny();

	protected final BuilderOptions opts;

	public IMDataFlow(BuilderOptions opts) {
		this.opts = opts;
	}

	/**
	 * Determines whether the cell is valid - i.e., it is guaranteed
	 * to be non-<code>null</code>.
	 * @return <code>true</code> if the cell is valid, <code>false</code>
	 *         otherwise.
	 */
	public boolean isValid() {
		return false;
	}

	public boolean isAny() {
		return false;
	}

	public boolean isSet() {
		return false;
	}

	public boolean isConstant() {
		return false;
	}

	public boolean hasFixedRange() {
		return false;
	}

	public boolean isArray() {
		return false;
	}

	public int maxValue() throws CompileException {
		throw new CompileException("no fixed range");
	}

	public int minValue() throws CompileException {
		throw new CompileException("no fixed range");
	}

	public IMDataFlow addConstValue(IMDataFlow df) throws CompileException {
		return null;
	}

	public String toReadableString() {
		return "???";
	}

	public IMDataFlow copy() throws CompileException {
		// TODO: copy?
		return this;
	}

	public IMDataFlow copy(IMClass type) throws CompileException {
		return copy();
	}

	public IMDataFlow fold(IMDataFlow flow) throws CompileException {
		throw new CompileException("not implemented!");
	}

	public IMDataFlow intersect(IMDataFlow flow) throws CompileException {
		throw new CompileException("not implemented!");
	}

	public IMConstant getValue() throws CompileException {
		throw new CompileException("no constant value "+this);
	}

	public IMConstant getConstValue(IMMethod method, int bcpos) throws CompileException {
		throw new CompileException("no constant value "+this);
	}

	@Override
	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		return toReadableString().equals(((IMDataFlow) obj).toReadableString());
	}

	public IMClass getType() {
		return null;
	}

	public static IMDataFlow createValidDataFlowObj(IMNode node,
			IMBasicBlock currentBB) throws CompileException {
		return createValidDataFlowObj(node, currentBB, determineType(node));
	}

	public static IMDataFlow createValidDataFlowObj(IMNode node,
			IMBasicBlock currentBB, IMClass type) throws CompileException {
		if (node.getDatatype() != BCBasicDatatype.REFERENCE)
			throw new CompileException("Must be a reference");
		if (node.isConstant() && !(node instanceof IMAMemConstant
				|| node instanceof IMAMTConstant))
			return new IMDataFlowConstObj(node.nodeToConstant(), type);
		if (node.isChecked(currentBB))
			return new IMDataFlowValidObj(node.getOptions(), type);
		if (!node.getOptions().hasOption("no_valid_this_ptr"))
			return new IMDataFlowValidObj(node.getOptions(), type);
		return new IMDataFlowInvalidObj(node.getOptions(), type);
	}

	public static IMDataFlow createInvalidDataFlowObj(IMNode node,
			IMBasicBlock currentBB) throws CompileException {
		if (node.getDatatype() == BCBasicDatatype.REFERENCE)
			return createInvalidDataFlowObj(node, currentBB, determineType(node));
		return new IMDataFlowAny(node, currentBB);
	}

	public static IMDataFlow createInvalidDataFlowObj(IMNode node,
			IMBasicBlock currentBB, IMClass type) throws CompileException {
		if (node.getDatatype() == BCBasicDatatype.REFERENCE)
			return new IMDataFlowInvalidObj(node.getOptions(), type);
		return new IMDataFlowAny(node, currentBB);
	}

	public static IMDataFlow createDataFlowObj(IMNode node,
			IMBasicBlock currentBB) throws CompileException {

		switch (node.getDatatype()) {
		case BCBasicDatatype.REFERENCE:
			return createDataFlowObj(node, currentBB, determineType(node));
		case BCBasicDatatype.BOOLEAN:
		case BCBasicDatatype.INT:
		case BCBasicDatatype.SHORT:
		case BCBasicDatatype.BYTE:
		case BCBasicDatatype.CHAR:
			if (node.isConstant())
				return new IMDataFlowConstValue(node.nodeToConstant());
		}

		// return ANY;
		return new IMDataFlowAny(node, currentBB);
	}

	public static IMDataFlow createDataFlowObj(IMNode node,
			IMBasicBlock currentBB, IMClass type) throws CompileException {
		if (node.getDatatype() != BCBasicDatatype.REFERENCE)
			throw new CompileException("Must be a reference");
		if (node.isConstant() && !(node instanceof IMAMemConstant
				|| node instanceof IMAMTConstant))
			return new IMDataFlowConstObj(node.nodeToConstant(), type);
		if (node.isChecked(currentBB))
			return new IMDataFlowValidObj(node.getOptions(), type);
		return new IMDataFlowInvalidObj(node.getOptions(), type);
	}

	public static IMDataFlow createArrayDataFlowObj(IMNode array, IMNode length,
			IMBasicBlock currentBB) throws CompileException {
		IMDataFlow arrayFlow  = createDataFlowObj(array,  currentBB);
		IMDataFlow lengthFlow = createDataFlowObj(length, currentBB);
		return new IMDataFlowArray(arrayFlow, lengthFlow);
	}

	public static IMDataFlow createAMemDataFlowObj(IMNode object,
			IMNode address, IMNode size, IMBasicBlock currentBB)
	throws CompileException {
		IMDataFlow objectFlow  = createDataFlowObj(object, currentBB);
		IMDataFlow addressFlow = createDataFlowObj(address, currentBB);
		IMDataFlow sizeFlow    = createDataFlowObj(size,  currentBB);
		return new IMDataFlowAMem(objectFlow, addressFlow, sizeFlow);
	}

	private static IMClass determineType(IMNode node) {
		IMClass type = node.determineType();
		if (type == null)
			return node.getMethod().getIMClass("java/lang/Object");
		return type;
	}

}
