/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.classfile.datatypes.*;

final public class IMDataFlowConstValue extends IMDataFlow {

	private IMConstant node;
	private String value_str;

	public IMDataFlowConstValue(IMConstant node) {
		super(node.getOptions());
		this.node = node;
	}

	@Override
	public IMDataFlow fold(IMDataFlow flow) throws CompileException {
		if (flow instanceof IMDataFlowConstValue
				&& node.compareTo(flow.getValue()) == 0)
			return this;
		IMDataFlow dfSet = flow.addConstValue(this);
		if (dfSet != null)
			return dfSet;
		return IMDataFlow.ANY;
	}

	@Override
	public IMDataFlow intersect(IMDataFlow flow) throws CompileException {
		if (flow instanceof IMDataFlowConstValue) {
			if (node.compareTo(flow.getValue()) == 0)
				return this;
		} else if (flow instanceof IMDataFlowSet) {
			for (IMDataFlow subFlow : (IMDataFlowSet) flow) {
				if (node.equalValue(subFlow.getValue()) == true)
					return this;
			}
		} else if (flow instanceof IMDataFlowIntRange) {
			int value = node.getIntValue();
			if (value >= flow.minValue() && value <= flow.maxValue())
				return this;
		} else if (flow instanceof IMDataFlowAny) {
			return this;
		}
		return IMDataFlow.ANY;
	}

	/**
	 * create a new set of constant values.
	 * @throws CompileException
	 */
	@Override
	public IMDataFlow addConstValue(IMDataFlow df) throws CompileException {
		return new IMDataFlowSet(this, df);
	}

	@Override
	public boolean isConstant() {
		return true;
	}

	@Override
	public int maxValue() throws CompileException {
		switch (node.getDatatype()) {
		case BCBasicDatatype.BOOLEAN:
		case BCBasicDatatype.BYTE:
		case BCBasicDatatype.CHAR:
		case BCBasicDatatype.SHORT:
		case BCBasicDatatype.INT:
			return node.getIntValue();
		default:
			return super.maxValue();
		}
	}

	@Override
	public int minValue() throws CompileException {
		return maxValue();
	}

	@Override
	public IMConstant getValue() throws CompileException {
		return node;
	}

	@Override
	public IMConstant getConstValue(IMMethod method, int bcPos)
	throws CompileException {
		int dataType = node.getDatatype();
		switch (dataType) {
		case BCBasicDatatype.BOOLEAN:
		case BCBasicDatatype.BYTE:
		case BCBasicDatatype.CHAR:
		case BCBasicDatatype.SHORT:
		case BCBasicDatatype.INT:
			return method.createIMIConstant(node.getIntValue(), bcPos);
		case BCBasicDatatype.LONG:
			return method.createIMLConstant(node.getLongValue(), bcPos);
		case BCBasicDatatype.FLOAT:
			return method.createIMFConstant(node.getFloatValue(), bcPos);
		case BCBasicDatatype.DOUBLE:
			return method.createIMDConstant(node.getDoubleValue(), bcPos);
		default:
			throw new CompileException("unkown datatype constant: "
					+ BCBasicDatatype.toString(dataType));
		}
	}

	@Override
	public boolean isValid() {
		return true;
	}

	@Override
	public int hashCode() {
		return toReadableString().hashCode();
	}

	@Override
	public String toReadableString() {
		if (value_str == null)
			value_str = node.toReadableString();
		return node.toReadableString();
	}

}
