/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import java.util.*;

import keso.compiler.*;
import keso.util.*;
import keso.classfile.datatypes.*;

public class IMDataFlowSet extends IMDataFlow implements Iterable<IMDataFlow> {

	int datatype = -1;
	Set<IMDataFlow> dataflows = new HashSet<IMDataFlow>();
	IntegerHashtable<IMDataFlow> ht = new IntegerHashtable<IMDataFlow>();

	/**
	 * Maximum number of elements allowed before using an
	 * {@link IMDataFlowIntRange} instead of a value set. */
	private static final int MAX_ELEMENT_COUNT = 16;

	public IMDataFlowSet(BuilderOptions opts, int datatype) {
		super(opts);
		this.datatype = datatype;
	}

	public IMDataFlowSet(IMDataFlow df, IMDataFlow df2)
	throws CompileException {
		super(df.opts);
		if (df instanceof IMDataFlowConstValue) {
			try {
				datatype = ((IMDataFlowConstValue) df).getValue().getDatatype();
			} catch (CompileException ex) {
			}
		}
		addConstValue(df);
		addConstValue(df2);
	}

	@Override
	public IMDataFlow fold(IMDataFlow flow) throws CompileException {
		if (flow.isAny())
			return flow;
		if (flow instanceof IMDataFlowIntRange)
			return flow.fold(this);
		return addConstValue(flow);
	}

	@Override
	public IMDataFlow intersect(IMDataFlow flow) throws CompileException {

		List<IMDataFlow> intersection;

		if (flow instanceof IMDataFlowConstValue) {
			return flow.intersect(this);
		} else if (flow instanceof IMDataFlowSet) {
			intersection = new ArrayList<IMDataFlow>();
			for (IMDataFlow subFlow : this) {
				for (IMDataFlow oSubFlow : (IMDataFlowSet) flow) {
					if (subFlow.equals(oSubFlow))
						intersection.add(subFlow);
				}
			}
		} else if (flow instanceof IMDataFlowIntRange) {
			intersection = new ArrayList<IMDataFlow>();
			for (IMDataFlow subFlow : this) {
				int value = subFlow.getValue().getIntValue();
				if (value >= flow.minValue() && value <= flow.maxValue())
					intersection.add(subFlow);
			}
		} else if (flow instanceof IMDataFlowAny) {
			return this;
		} else {
			return IMDataFlow.ANY;
		}

		if (intersection.isEmpty())
			return IMDataFlow.ANY;
		if (intersection.size() == 1)
			return intersection.get(0);
		IMDataFlowSet result = new IMDataFlowSet(opts, datatype);
		for (IMDataFlow subFlow : intersection) {
			result.addConstValue(subFlow);
		}
		return result;
	}

	@Override
	public boolean isSet() {
		return true;
	}

	@Override
	public boolean hasFixedRange() {
		if (isNumber()) return true;
		return false;
	}

	@Override
	public int maxValue() throws CompileException {
		int ks[] = ht.sortedKeys();
		return ks[ks.length-1];
	}

	@Override
	public int minValue() throws CompileException {
		int ks[] = ht.sortedKeys();
		return ks[0];
	}

	@Override
	public IMDataFlow addConstValue(IMDataFlow df) throws CompileException {
		if (df.isSet()) {
			for (IMDataFlow dfi : (IMDataFlowSet) df) {
				dataflows.add(dfi);
				if (isNumber()) {
					try {
						int value = df.getValue().getIntValue();
						ht.put(value,df);
					} catch (CompileException ex) {
					}
				}
			}
			if (ht.size() > MAX_ELEMENT_COUNT)
				return new IMDataFlowIntRange(minValue(), maxValue(), opts);
			return this;
		}
		dataflows.add(df);
		if (isNumber()) {
			try {
				int value = df.getValue().getIntValue();
				ht.put(value,df);
			} catch (CompileException ex) {
			}
		}
		if (ht.size() > MAX_ELEMENT_COUNT)
			return new IMDataFlowIntRange(minValue(), maxValue(), opts);
		return this;
	}

	private boolean isNumber() {
		if (datatype==BCBasicDatatype.INT) return true;
		if (datatype==BCBasicDatatype.SHORT) return true;
		if (datatype==BCBasicDatatype.BYTE) return true;
		return false;
	}

	@Override
	public String toReadableString() {
		if (isNumber()) {
			StringBuffer str = new StringBuffer("[");
			int ks[] = ht.sortedKeys();
			if (ks.length<4) {
				for (int i=0;i<ht.size();i++) {
					IMDataFlow df = ht.get(ks[i]);
					if (i>0) str.append(",");
					str.append(df.toReadableString());
				}
			} else {
				str.append(ks[0]);
				str.append("-");
				str.append(ks[ks.length-1]);
			}
			return str.append("]").toString();
		} else {
			Iterator<IMDataFlow> it = iterator();
			StringBuffer str = new StringBuffer("[");
			while (it.hasNext()) {
				IMDataFlow df = it.next();
				str.append(df.toReadableString());
				if (it.hasNext()) str.append(",");
			}
			return str.append("]").toString();
		}
	}

	@Override
	public Iterator<IMDataFlow> iterator() {
		return dataflows.iterator();
	}

}
