/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.Coder;
import java.util.*;

/**
 * The elements of this class represent descriptor fields of the domain_desc
 * table.
 */
final class DescriptorField {
	String name;
	String type;
	String value;

	/**
	 * Indicates if this DescriptorField's type is a function pointer.
	 * In this case, the name of the field is included in the type
	 * and does not need to be printed separately.
	 */
	boolean fnPointer;

	/**
	 * Indicates if this DescriptorField's value shall be prefixed with
	 * the domain symbol prefix.
	 * Useful, if the field is initialized with the value of a domain
	 * local symbol.
	 */
	boolean domPrefix;

	/**
	 * Indicates if this DescriptorField's value needs to be cast when
	 * initializing this field.
	 */
	boolean initWithCast;

	DescriptorField(String type, String name, String value, boolean fnPointer, boolean domPrefix, boolean iwc) {
		this.type = type;
		this.name = name;
		this.value = value;
		this.fnPointer = fnPointer;
		this.domPrefix = domPrefix;
		this.initWithCast = iwc;
	}

	void emit_declare(Coder coder) throws CompileException {
		coder.header_add(type);
		if(!fnPointer) {
			coder.header_add(" ");
			coder.header_add(name);
		}
		coder.header_add(";");
	}

	String emit_define(String dsp) {
		String v=value;
		if (v==null) v="0";
		if(domPrefix) v = dsp + v;
		if(initWithCast) v = "("+ type  +")"+v;
		return v+" /* "+ name +" */,\n";
	}

	final public int hashCode() {
		return name.hashCode();
	}

	final public boolean equals(Object obj) {
		if (obj instanceof DescriptorField) return name.equals(((DescriptorField)obj).name);
		return false;
	}
}

/**
 * Abstract superclass for all IMModul children that contain fields of the
 * domain descriptor (or maybe other descriptors in the future).
 */
public abstract class IMDescriptorContainer extends IMModule {
	/**
	 * Contains the descriptor fields.
	 */
	private Map<String, DescriptorField> descFields
	        = new HashMap<String, DescriptorField>();

	protected IMDescriptorContainer(String name, String alias, BuilderOptions opts) {
		super(name,alias,opts);
	}

	/**
	 * Add a new descriptor field.
	 *
	 * @param type C-Type of the field.
	 * @param name Name of the field.
	 */
	public void addDescriptorField(String type, String name, boolean fnPointer)  {
		addDescriptorField(type,name,null, fnPointer);
	}

	/**
	 * Add a new descriptor field and assign an initial value.
	 *
	 * @param type  C-Type of the field.
	 * @param name  Name of the field.
	 * @param value Initial value of the field.
	 */
	public void addDescriptorField(String type, String name, String value, boolean fnPointer)  {
		addDescriptorField(type,name,value,fnPointer,false,false);
	}

	/**
	 * Add a new descriptor field and assign an initial value.
	 *
	 * @param type  C-Type of the field.
	 * @param name  Name of the field.
	 * @param value Initial value of the field.
	 * @param fnPointer Specify if the type of this field is a function pointer.
	 * @param domPrefix Specify if the value should be prefixed with the domain symbol prefix.
	 */
	public void addDescriptorField(String type, String name, String value, boolean fnPointer, boolean domPrefix, boolean initWithCast)  {
		descFields.put(name,new DescriptorField(type, name, value, fnPointer,domPrefix,initWithCast));
	}

	/**
	 * Change the value of a descriptor field.
	 *
	 * @param name  Name of the field.
	 * @param value New value of the field.
	 *
	 * @throws CompileException A field with the specified name does not exist.
	 */
	public void setFieldValue(String name, String value)  throws CompileException {
		DescriptorField f=null;
		if ((f=descFields.get(name))==null) throw new CompileException("unkown descriptor field");
		f.value = value;
	}

	public void emit_declareFields(Coder coder, String prefix, String suffix) throws CompileException {
		for(DescriptorField f : descFields.values()) {
			coder.header_add(prefix);
			f.emit_declare(coder);
			coder.header_add(suffix);
		}
	}

	/**
	 * Generate a String with the specified prefix (indentation) that
	 * contains the definition of all fields in this container.
	 *
	 * @param  prefix String that will be prefixed to each line, primarily for indentation.
	 * @param  domSymPrefix Symbol prefix of the Domain that the field value may refer to.
	 * @return String containing the definitions, one per line, prefixed with the specified prefix.
	 */
	public String emit_defineFields(String prefix, String domSymPrefix) {
		StringBuffer definition = new StringBuffer();

		for(DescriptorField f : descFields.values()) {
			definition.append(prefix);
			definition.append(f.emit_define(domSymPrefix));
		}
		return definition.toString();
	}
}
