/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.*;
import keso.compiler.analysis.*;
import keso.compiler.backend.*;
import java.util.*;

/**
 * Internal representation of a domain.
 */
final public class IMDomain extends IMDescriptorContainer
implements Comparable<IMDomain> {

	private ReachabilityData rData = new ReachabilityData();

	private static Map<String, IMDomain[]> all_doms
	= new HashMap<String, IMDomain[]>();
	private	static Map<String, IMClass> usedModules
	= new HashMap<String, IMClass>();
	private static ReachabilityData overallRData = new ReachabilityData();
	private static Map<DomainDefinition, IMDomain> def2im
	= new HashMap<DomainDefinition, IMDomain>();

	/**
	 * System-wide unique identifier for this domain.
	 */
	private int domain_id;

	/**
	 * Internal representations of all tasks belonging to this domain.
	 */
	private IMTask[] task_list;

	/**
	 * The Heap of this domain.
	 */
	private IMHeap heap;

	/**
	 * The representation of this domain's definition in
	 * the KESO configuration file.
	 */
	private JavaDomain myDef;

	/**
	 * Initializes a newly created IMDomain instance.
	 *
	 * @param domainDef Representation of the domain's definition from the configuration file.
	 */
	private IMDomain(BuilderOptions opts, JavaDomain domainDef) throws CompileException {
		super("domains","domains",opts);

		myDef = domainDef;
		domain_id = domainDef.domainid;
		def2im.put(domainDef,this);

		heap = opts.getGlobalHeap().createHeap(this, domainDef.getHeap());

		List<JavaTask> tasks = domainDef.getTasks();
		task_list = new IMTask[tasks.size()];
		for (int i=0; i<tasks.size(); i++) {
			JavaTask task_def = tasks.get(i);
			task_list[i] = new IMTask(task_def);
		}
	}

	public void addUsedModule(IMClass mod) {
		usedModules.put(mod.getClassTypeString(), mod);
	}

	private static Map<String, IMClass> classesWithStatics
	= new HashMap<String, IMClass>();

	public static void addClassWithStatics(IMClass clazz) {
		classesWithStatics.put(clazz.getAlias(), clazz);
	}

	/**
	 * Get the system-wide unique identifier of this domain.
	 *
	 * @return A system-wide unique identifier of this domain.
	 */
	public int getDomainID() { return domain_id; }

	/**
	 * Get the definition of this domain in the configuration file.
	 *
	 * @return Definition of this domain in the configuration file.
	 */
	public JavaDomain getDef() { return myDef; }

	/**
	 * Get unique String identifier for this domain.
	 *
	 * @return System-wide unique String identifying this domain.
	 */
	public String getIdentifier() { return myDef.getIdentifier(); }

	/**
	 * Get the internal representations of all Tasks belonging to this domain.
	 *
	 * @return Array with the internal representations of all Tasks belonging to this domain.
	 */
	public IMTask[] getTasks() { return task_list; }

	/**
	 * Get the heap of this domain.
	 *
	 * @return Heap of this domain.
	 */
	public IMHeap getHeap() { return heap; }

	/**
	 * Get the internal representation of a Task belonging to this domain
	 * specified by its name.
	 *
	 * @param  taskName Name of the wanted task as specified in the keso configuration.
	 * @return The internal representation of the requested Task, or null if no tasks with the specified name exists in this domain.
	 */
	public IMTask getTask(String taskName) {
		for (int j=0;j<task_list.length;j++) {
			if (task_list[j].getDef().ident.equals(taskName)) return task_list[j];
		}
		return null;
	}

	public ReachabilityData getReachabilityData() {
		return rData;
	}

	/**
	 * Generate code that initializes the domain_desc structure of this domain.
	 *
	 * @param coder The Coder that the code will be emitted to.
	 */
	private void translate_desc(Coder coder) throws CompileException {
		int nsr = coder.getClassStore().getNumStaticRefs();

		coder.local_add("\t/* init base fields */\n");
		coder.local_add("\t{ /* static refs */ ");
		while(nsr-->0) coder.local_add("NULL, ");
		coder.local_add("},\n");
		coder.local_add(emit_defineFields("\t", coder.getDomainSymbolPrefix(myDef)));
		coder.local_add("\t{ /* init heap (type " + heap.name + ") */\n");
		coder.local_add(heap.emit_defineFields("\t\t", coder.getDomainSymbolPrefix(myDef)));
		coder.local_add("\t},\n");
		coder.add_class(heap);
	}

	/**
	 * Generate supplemental code for this domain, including the heap of the
	 * domain.
	 *
	 * @param coder The Coder that the code will be emitted to.
	 */
	public void translate(Coder coder) throws CompileException {
		// add shared domain fields
		if (opts.getGlobalHeap().multiImpl()) {
			/* all domain_descs contain a field with a pointer to
			 * their heap's allocator function
			 */
			addDescriptorField("object_t* (*allocator) (class_id_t, obj_size_t, obj_size_t)", "allocator", heap.getAllocatorName(), true);
		}
		addDescriptorField("const unsigned int", "heap_size", Integer.toString(heap.getSize()),false);
		addDescriptorField("unsigned int", "heap_free", Integer.toString(heap.getSize()),false);

		if (opts.getGlobalHeap().needGCInfo()) {
			addDescriptorField("unsigned char", "colorbit", Integer.toString(heap.getInitialColor()),false);
		}

		heap.translate(coder);
	}

	/**
	 * Generate code for system-wide domain management.
	 * The domain_desc type domain_t will also be generated by this method.
	 *
	 * @param coder The Coder that the code will be emitted to.
	 * @param opts  BuilderOptions instance.
	 */
	public static void translateDomains(Coder coder, BuilderOptions opts) throws CompileException {
		IMDomain[] domains = opts.getDomains();

		for (int i=0;i<domains.length;i++) {
			domains[i].translate(coder);
		}

		/* The domain_t will contain common fields that are independent of the
		 * underlying heap implementation. For each heap type, a heap specific
		 * domain descriptor type will be created with the fields of that heap
		 * type. The specific types will be used to instantiate the actual
		 * domain descriptors.
		 */
		coder.beginHeader(domains[0]);
		coder.beginClassMethod(domains[0]);

		coder.header_add("#include \"global.h\"\n");

		for (IMModule mod: usedModules.values()) {
			coder.header_add("#include \"");
			coder.header_add(mod.getAlias());
			coder.header_add(".h\"\n");
		}

		for (String className: classesWithStatics.keySet()) {
			coder.header_add("#include \"" + className + ".h\"\n");
		}

		opts.getGlobalHeap().emit_header(coder);

		coder.header_add("\n#define DOMAIN_DESC_BASEFIELDS \\\n");
		// static reference fields
		coder.header_add("\tobject_t * volatile static_ref[KESO_NUM_STATIC_REFS];\\\n");
		// domain descriptor fields required by other system parts
		domains[0].emit_declareFields(coder,"\t","\\\n");

		/*
		 * At -O2 and higher, recent GCC versions do strict type aliasing,
		 * i.e. objects may be pointed to only by one type of pointer (in
		 * addition to char* and void*). domain_t* are used to point to any
		 * specific domain descriptor type and thus break this rule, which
		 * results in compile warnings when using a recent GCC version. GCC
		 * comes with a type attribute may_alias that allows pointers of
		 * this type to point to any object. An alternate and possibly more
		 * portable solution would be combining all domain descriptor types
		 * in a union domain_t.
		 */
		coder.header_add("\nstruct DOMAINDESC {\n");
		coder.header_add("\tDOMAIN_DESC_BASEFIELDS\n");
		coder.header_add("} ATTR_MAYALIAS;\n");

		// create heap specific domain descriptor types
		opts.getGlobalHeap().emit_heapdesc_types(coder);

		coder.header_add("\n\n#define KESO_NUM_JAVADOMAINS ");
		coder.header_add(domains.length);
		coder.header_add("\n");

		// FIXME point this to an extra domain descriptor if needed
		coder.header_add("#define DOMAIN_ZERO NULL\n");

		coder.header_add("#define STATICREF(DD) DD->static_ref\n");

		if (opts.isSingleDomainSystem()) {
			coder.header_add("#define KESO_CURRENT_DOMAIN (");
			coder.header_add(coder.getDDescPtr(domains[0].getDef()));
			coder.header_add(")\n");
		} else {
			coder.header_add("#define KESO_CURRENT_DOMAIN keso_curr_domain\n");
			coder.header_add("extern domain_t *keso_curr_domain;\n");
			coder.local_add("domain_t *keso_curr_domain = DOMAIN_ZERO;\n");
		}

		opts.getGlobalHeap().translate(coder,domains);

		// Define and init domain descriptors
		for (int i=0;i<domains.length;i++) {
			String dd_type = "struct domain_"+domains[i].heap.name+"_t ";

			// declare
			coder.header_add("extern ");
			coder.header_add(dd_type);
			coder.header_add(coder.getDDesc(domains[i].getDef()));
			coder.header_add(";\n");

			// define
			coder.local_add(dd_type);
			coder.local_add(coder.getDDesc(domains[i].getDef()));
			coder.local_add(" = {\n");
			domains[i].translate_desc(coder);
			coder.local_add("};\n");
		}

		coder.endClassMethod();
		coder.endHeader();
	}

	@Override
	public String toString() {
		return getIdentifier();
	}

	@Override
	public int compareTo(IMDomain o) {
		return domain_id - o.domain_id;
	}

	/**
	 * Constructs the internal representations of all domains
	 * using the DomainDefinitions of the system configuration.
	 *
	 * @param  opts BuilderOptions instance.
	 * @return Array with the internal domain representations.
	 */
	public static IMDomain[] createDomains(String sys_name, BuilderOptions opts) throws CompileException {
		IMDomain[] doms= all_doms.get(sys_name);

		if (doms==null) {
			List<JavaDomain> domains = opts.getSysDef().getJavaDomains();
			doms = new IMDomain[domains.size()];

			opts.setupGlobalHeap();

			for(int i=0; i<domains.size(); i++) {
				doms[i] = new IMDomain(opts, domains.get(i));
			}

			all_doms.put(sys_name,doms);
		}

		return doms;
	}

	public static ReachabilityData getOverallReachabilityData() {
		return overallRData;
	}

	public static IMDomain getIMDomain(DomainDefinition ddef) {
		return def2im.get(ddef);
	}

	// get IMDomain representation of a JavaDomain list
  public static List<IMDomain> get_IMDomainList(List<JavaDomain> list, BuilderOptions opts) throws CompileException {
    IMDomain[] doms = opts.getDomains();
    List<IMDomain> imdomains=new ArrayList<IMDomain>();
    for(int k=0; k<list.size();k++) {
      for(IMDomain d : doms) {
        if(d.getDef()==list.get(k)) {
          imdomains.add(d);
        }
      }
    }
    return imdomains;
  }


}
