/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.*;
import keso.classfile.datatypes.*;

import keso.compiler.*;
import keso.compiler.backend.*;
import keso.util.*;

/**
 * Internal representation of an object field.
 */
public final class IMField implements Comparable<IMField> {

	private BuilderOptions  opts;

	private IMClass	  	fieldClass;
	private int             datatype;
	private int		static_index = -1;
	private FieldData       fieldDesc;

	private String	  fieldName;
	private String	  aliasName;
	private String	  macroName;
	private String	  typeString=null;

	public IMField(BuilderOptions opts, IMClass field_class, FieldData fieldDesc) {
		this.opts      = opts;
		this.fieldDesc = fieldDesc;
		this.fieldClass= field_class;

		String typeDesc = fieldDesc.getType();

		this.datatype  = BasicTypeDescriptor.toBasicDatatype(typeDesc.charAt(0));
		this.fieldName = fieldDesc.getName();
		this.aliasName = DecoratedNames.createFieldAlias(opts, field_class, fieldDesc.getName());

		StringBuffer key = new StringBuffer("S");
		key.append(field_class.getAlias());
		key.append('_');
		key.append(aliasName);
		macroName = key.toString().toUpperCase();

		if (datatype==BCBasicDatatype.REFERENCE) {
			int len = typeDesc.length();
			typeString = typeDesc.substring(1,len-1);
		}
	}

	public boolean hasStaticIndex() {
		return static_index!=-1;
	}

	public void setStaticIndex(int value) {
		static_index=value;
	}

	public int getDatatype() {
		return datatype;
	}

	public String getTypeString() {
		if (isVolatile()) {
			return "volatile "+BuilderOptions.typeToString(datatype);
		} else {
			return BuilderOptions.typeToString(datatype);
		}
	}

	public boolean isReference() { return datatype==BCBasicDatatype.REFERENCE; }
	public boolean isArray() { return fieldDesc.getType().charAt(0) == '['; }
	public boolean isStatic() { return fieldDesc.isStatic(); }
	public boolean isFinal() { return fieldDesc.isFinal(); }
	public boolean isPrivate() { return fieldDesc.isPrivate(); }
	public boolean isVolatile() { return fieldDesc.isVolatile(); }
	public boolean isTransient() { return fieldDesc.isTransient(); }

	public String getPrivateFieldName() {
		if (isPrivate()) {
			return aliasName;
		} else {
			return fieldName;
		}
	}

	public String getFieldName() {
		return fieldName;
	}

	public String getAlias() {
		return aliasName;
	}

	public String getMacroName() {
		return macroName;
	}

	public IMClass getType() {
		return fieldClass;
	}

	public String getRefClassName() {
		return typeString;
	}

	public void header_emit_mt(Coder coder) throws CompileException {
		if (typeString.equals("keso/core/MT")) return;

		coder.header_add("\t");
		if (typeString.equals("keso/core/MT_U32")) {
			coder.header_add("volatile unsigned long");
		} else if (typeString.equals("keso/core/MT_SPACE32")) {
			coder.header_add("const unsigned long");
		} else if (typeString.equals("keso/core/MT_U32RO")) {
			coder.header_add("volatile unsigned long");
		} else if (typeString.equals("keso/core/MT_U16")) {
			coder.header_add("volatile unsigned short");
		} else if (typeString.equals("keso/core/MT_U8")) {
			coder.header_add("volatile unsigned char");
		} else if (typeString.equals("keso/core/MT_BOOLEAN")) {
			coder.header_add("volatile bool");
		} else if (typeString.equals("keso/core/MT_FLOAT")) {
			coder.header_add("volatile float");
		} else {
			throw new CompileException("unkown memory type! "+typeString);
		}
		coder.header_add(" ");
		coder.header_add(aliasName);
		coder.header_add(";\n");
	}

	public void emit_macro(Coder coder, String DDPTR) throws CompileException {
		coder.add(macroName);
		coder.add("(");
		coder.add(DDPTR);
		coder.add(")");
	}

	public void global_emit_macro(Coder coder) throws CompileException {
		if (!isStatic()) throw new CompileException("is not a static field");
		coder.global_header_add("#define ");
		coder.global_header_add(macroName);
		coder.global_header_add("(DDPTR) ");

		if (datatype==BCBasicDatatype.REFERENCE) {
			coder.global_header_add(" STATICREF(DDPTR)");
			coder.global_header_add("[");
			coder.global_header_add(static_index);
			coder.global_header_add("]");

		} else {
			coder.global_header_add(" (DDPTR->");
			coder.global_header_add(fieldClass.getClassStaticsString());
			coder.global_header_add(".");
			coder.global_header_add(aliasName);
			coder.global_header_add(")");
		}
		coder.global_header_add("\n");
	}

	public void header_emit(Coder coder) throws CompileException {
		coder.header_add("\t");
		if (isVolatile())
			coder.header_add("volatile ");
		coder.header_add(BuilderOptions.typeToString(datatype));
		coder.header_add(" ");
		coder.header_add(aliasName);
		coder.header_add(";");
	}

	public void emit(StringBuffer buffer) {
		buffer.append("\t");
		if (isVolatile())
			buffer.append("volatile ");
		buffer.append(BuilderOptions.typeToString(datatype));
		buffer.append(" ");
		buffer.append(aliasName);
		buffer.append(";\n");
	}

	public void emit_field(StringBuffer buffer) throws CompileException {
		buffer.append(aliasName);
	}

	public String toString() {
		if (typeString!=null)
			return typeString+" "+fieldName;
		return "??? "+fieldName;
	}

	@Override
	public int compareTo(IMField o) {
		if (fieldClass != o.fieldClass)
			return fieldClass.getAlias().compareTo(o.fieldClass.getAlias());
		return macroName.compareTo(o.macroName);
	}

}
