/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.*;
import keso.compiler.config.*;

import java.util.*;

/**
 * This class manages the system-wide configured heaps.
 */
public final class IMGlobalHeap extends IMModule {

	/**
	 * Maps heaptypes to one created instance of that heaptype.
	 */
	private Map<String, IMHeap> impls = new HashMap<String, IMHeap>();

	public IMGlobalHeap(BuilderOptions opts) {
		super("global_heap","global_heap",opts);
	}

	/**
	 * Indicates if the used heap types require the use of writebarriers.
	 */
	private boolean needWrBr=false;

	/**
	 * Indicates if the used heap types require stack scan.
	 */
	private boolean needStackScan=false;

	/**
	 * Indicates if the used heap types require the use of allocation colors.
	 */
	private boolean needObjColor=false;

	/**
	 * Indicates if the allocation of an object can block.
	 */
	private boolean allocBlocks = false;

	/**
	 * Indicates if tasks could be defered by the GC.
	 */
	private boolean deferTasks = false;

	/**
	 * Indicates if we need object size.
	 */
	private boolean needObjSize = false;

	/**
	 * Creates a new heap of the given heap definition.
	 *
	 * @param dom     The domain the newly created Heap will be assigned to
	 * @param heapdef The heap's definition from the configuration file.
	 * @return The newly created Heap
	 */
	public IMHeap createHeap(IMDomain dom, ComplexAttribute<?> heapdef)
			throws CompileException {
		int size=0;
		IMHeap heap=null;

		if (heapdef!=null) {
			/* more different types */
			size = heap_size(heapdef);
			opts.vverbose("#### "+heapdef.value+" (size: "+size+")");

			if (heapdef.value.equals("IdleRoundRobin")) {
				heap = new IMHeapIdleRoundRobin(dom, opts, size, heapdef);
			} else if (heapdef.value.equals("CoffeeBreak")) {
				heap = new IMHeapCoffeeBreak(dom, opts, size, heapdef);
			} else if (heapdef.value.equals("RestrictedDomainScope")) {
				heap = new IMHeapRestrictedDomainScope(dom, opts, size, heapdef);
			}
		}

		if (heap==null) {
			/* default heap implementation very restrict but also small */
			/* dynamic objects are only seen by the creating task (restricted task scope)*/
			/* the objects are collected on task termination */
			opts.warn("## no heap definition found! (use default rds)");
			size = 64*1024;
			heap = new IMHeapRestrictedDomainScope(dom, opts, size, heapdef);
		} else {
			impls.put(heap.name,heap);
		}


		return heap;
	}

	void emit_heapdesc_types(Coder coder) throws CompileException {

		for (IMHeap thisHeap : impls.values()) {

			// heapdesc type
			coder.header_add("struct domain_");
			coder.header_add(thisHeap.name);
			coder.header_add("_t {\n");
			coder.header_add("\tDOMAIN_DESC_BASEFIELDS\n");
			coder.header_add("\tstruct {\n");
			thisHeap.emit_declareFields(coder,"\t\t","\n");
			coder.header_add("\t} hd;\n");
			coder.header_add("};\n");
		}
	}

	public void emit_header(Coder coder) throws CompileException {
		for (IMHeap thisHeap : impls.values()) {
			coder.header_add("#include \"");
			coder.header_add(thisHeap.getAlias());
			coder.header_add(".h\"\n");
		}
	}

	/**
	 * Translate function for the case in which only one heap implementation is
	 * being used.
	 * KESO_ALLOC will use always the same allocator function and no indirection
	 * via the domain_desc structure will be necessary.
	 *
	 * @param coder The Coder to which the code will be generated.
	 */
	private void translate_singular(Coder coder, IMDomain[] doms) throws CompileException {
		IMHeap heap_impl = doms[0].getHeap();
		String single_heap = heap_impl.getAlias();

		coder.header_add("#include \"");
		coder.header_add(single_heap);
		coder.header_add(".h\"\n");

		coder.header_add("\n#define KESO_ALLOC(_cls_,_size_,_roff_) ");
		coder.header_add("keso_");
		coder.header_add(single_heap);
		coder.header_add("_alloc(_cls_,_size_,_roff_)\n");
	}

	/**
	 * Translate function for the case in which more than one heap implementation is
	 * being used.
	 * KESO_ALLOC will use lookup the allocator via the domain descof the current domain.
	 *
	 * @param coder The Coder to which the code will be generated.
	 */
	private void translate_non_singular(Coder coder) throws CompileException {
		coder.header_add("\n#define KESO_ALLOC(_cls_,_size_,_roff_) ");
		coder.header_add("KESO_CURRENT_DOMAIN->allocator(_cls_,_size_,_roff_)\n");
	}

	/**
	 * Generate code for all created heaps.
	 * The generated code will depend on whether there is more than one Heap
	 * used or not.
	 * <p>
	 * Furthermore, all used heap implementation's translate function will
	 * be called.
	 * @param coder Code will be generated to this Coder.
	 * @param doms  Array with interal representations of all domains
	 * @see #translate_non_singular(Coder coder)
	 * @see #translate_singular(Coder coder, IMDomain[] doms)
	 * @see IMHeap#translate(Coder coder)
	 */
	public void translate(Coder coder, IMDomain[] doms) throws CompileException {
		opts.verbose("###### create Heaps ");

		// heap memory
		for(int i=0; i<doms.length; i++) {
			coder.local_add("char ");
			coder.local_add(coder.getDomainSymbolPrefix(doms[i].getDef()));
			coder.local_add("_heap[");
			coder.local_add(doms[i].getHeap().heap_size);
			coder.local_add("] ALIGN4;\n");
		}

		// global heap support functions
		if (impls.size()==1) {
			translate_singular(coder, doms);
		} else {
			translate_non_singular(coder);
		}

		if (needObjSize) {
			coder.header_add("obj_size_t keso_objSize(obj_size_t slot_size, object_t *obj);\n");
			coder.local_add("\n/* Determine the size in slots of the object\n");
			coder.local_add(" * pointed to by obj\n");
			coder.local_add(" */\n");
			coder.local_add("obj_size_t keso_objSize(obj_size_t slot_size, object_t *obj) {\n");
			coder.local_add("	obj_size_t size = CLASS(obj->class_id).size;\n\n");
			coder.local_add("	if(keso_isArrayClass(obj->class_id)) {\n");
			coder.local_add("		/* element size * element count */\n");
			if(opts.hasOption("use_coded_refs")){
				coder.local_add("		size *= ((array_t*) (NATIVE_ADDR(obj)))->size;\n");
			}
			else{	
				coder.local_add("		size *= ((array_t*) obj)->size;\n");
			}
			coder.local_add("		/* plus size of array header */\n");
			if(coder.getClassStore().getClass("[Ljava/lang/Object;") != null) {
				coder.add_class("object_array");
				coder.local_add("\t\tif(keso_isObjectArrayClass(obj->class_id))\n");
				coder.local_add("\t\t\tsize += sizeof(object_array_t);\n");
				coder.local_add("\t\telse\n\t");
			}
			coder.local_add("		size += sizeof(array_t);\n");
			coder.local_add("	}\n\n");
			coder.local_add("	size += slot_size-1;\n");
			coder.local_add("	size /= slot_size;\n");
			coder.local_add("	return size;\n");
			coder.local_add("}\n");
		}

		if (opts.needGuardedMemoryAccess()) {
			coder.header_add("#ifndef KESO_OMIT_MEMORY_GUARD\n");
			coder.header_add("#define KESO_WR_GUARD(_store_) { ");
			coder.header_add("SuspendAllInterrupts(); _store_ ; ResumeAllInterrupts(); ");
			coder.header_add("}\n");
			coder.header_add("#else\n");
			coder.header_add("#define KESO_WR_GUARD(_store_) _store_\n");
			coder.header_add("#endif\n");

			coder.header_add("#ifndef KESO_OMIT_MEMORY_GUARD\n");
			coder.header_add("#define KESO_RD_GUARD(_read_) { ");
			coder.header_add("SuspendAllInterrupts(); _read_ ; ResumeAllInterrupts(); ");
			coder.header_add("}\n");
			coder.header_add("#else\n");
			coder.header_add("#define KESO_RD_GUARD(_read_) _read_\n");
			coder.header_add("#endif\n");
		}

		if (needWrBr) {
			coder.header_add("#ifndef KESO_OMIT_WRITEBARRIERS\n");
			coder.header_add("#if KESO_WRBR_INLINE_LEVEL==1\n");
			coder.header_add("#define KESO_WRBR(_field_,_value_) { object_t** ref=(object_t**) &(_field_); ");
			coder.header_add("if(KESO_CURRENT_DOMAIN==(domain_t*)keso_irr_gc_active_domain) keso_irr_pushObject(*ref); ");
			coder.header_add("*ref =  _value_; }\n");
			coder.header_add("#elif KESO_WRBR_INLINE_LEVEL==0\n");
			coder.header_add("#define KESO_WRBR(_field_,_value_) keso_irr_writebarrier(&(_field_), _value_)\n");
			coder.header_add("#endif\n");
			coder.header_add("#else\n");
			coder.header_add("#define KESO_WRBR(_field_,_value_) (_field_ =  _value_)\n");
			coder.header_add("#endif\n");
		}

		// generic keso_allocObject function
		if (opts.hasOption("no_inline_alloc")) {
			coder.header_add("\nobject_t *keso_allocObject(class_id_t class_id);\n");
			coder.local_add("\nobject_t *keso_allocObject(class_id_t class_id) {\n");
			coder.local_add("\treturn KESO_ALLOC(class_id,CLASS(class_id).size,CLS_ROFF(class_id));\n");
			coder.local_add("}\n");
		} else { // workaround because keso_allocObject is used in some c-templates
			coder.header_add("\n#define keso_allocObject(ID) KESO_ALLOC(ID, CLASS(ID).size, CLS_ROFF(ID))\n");
		}
	}

	/**
	 * Query if multiple different heap types are used in the system.
	 *
	 * @return true if multiple different heap types are configured for the domains
	 */
	public boolean multiImpl() {
		return impls.size()>1;
	}

	/**
	 * Get an Enumeration of all used heap implementations
	 */
	public Collection<IMHeap> usedImpls() {
		return impls.values();
	}

	public void enableStackScan() {
		needStackScan=true;
		needObjSize=true;
	}

	public boolean needStackScan() {
		return needStackScan;
	}

	public boolean needObjSize() {
		return needObjSize;
	}

	public void enableNeedObjSize() {
		needObjSize = true;
	}

	public void enableWriteBarriers() {
		needWrBr=true;
	}

	public void enableObjColor() {
		needObjColor=true;
	}

	public boolean needGCInfo() {
		return (needObjColor || needStackScan);
	}

	public boolean needWriteBarriers() {
		return needWrBr;
	}

	public void enableBlockingAllocation() {
		allocBlocks = true;
	}

	public boolean allocationCanBlock() {
		return allocBlocks;
	}

	public void enableDeferTasks() {
		deferTasks = true;
	}

	public boolean deferTasksOnDemand() {
		return deferTasks;
	}

	private static int heap_size(ComplexAttribute<?> heapdef) {
		Attribute<?> size_attr = heapdef.getAttribute("HeapSize");
		if (size_attr!=null) return size_attr.valueInt();

		size_attr = heapdef.getAttribute("HeapSizeKB");
		if (size_attr!=null) return (size_attr.valueInt() * 1024);

		size_attr = heapdef.getAttribute("HeapSizeMB");
		if (size_attr!=null) return (size_attr.valueInt() * 1024 * 1024);

		return 1024;
	}
}
