/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.*;
import keso.compiler.config.*;
import java.util.*;
import java.io.PrintStream;

/**
 * The coffee break GC.
 *
 * Heap is divided into slots of a configurable size with a simple garbage
 * collector shared among all instances of this Heap implementation that
 * cleans memory of participating domains in a round robin manner when the
 * system is idle.
 * <p>
 * The Garbage collector uses a bitmap that contains one bit for each slot.
 * The bitmap will be shared for all domains and must, therefore the domain with
 * the most slots determines its size.
 * <p>
 * The Garbage collector runs with the lowest priority in the system and
 * therefore only when all the other tasks are in suspended or waiting state.
 * The GC therefore only needs to scan static fields and immortal objects. If
 * there are waiting tasks, their stack also needs to be scanned which can be
 * achieved with use of the linked object list.
 * <p>
 * Free Memory is managed with a linked list, contained directly in the free
 * memory slots (slotsize must at least be 4 bytes, which is the current size
 * of the object header). The linked list is regenerated from the bitmap
 * during each GC cycle in order to merge free slots.
 * The listdata will be saved in a list_el structure at the beginning of each
 * free block, containing the size of the block in slots and the offset in
 * slots to the next free block.
 */
public class IMHeapCoffeeBreak extends IMHeapSlottedGC {
	/**
	 * Contains the domain_ids of all domains that
	 * make use of this Heaptype.
	 */
	private static List<IMDomain> domains  = new ArrayList<IMDomain>();

	/**
	 * Flag to indicate if one-time translation parts like the implementations
	 * functions have already been emitted or not.
	 */
	private static SystemDefinition once = null;

	public IMHeapCoffeeBreak(IMDomain dom, BuilderOptions opts, int size, ComplexAttribute<?> heapdef)
			throws CompileException {
		super("coffee", "coffee", size,opts,dom,heapdef);

		// pointer to the beginning of the heap
		addDescriptorField("coffee_listel_t*", "heap_top", "_heap", false, true, true);
		// pointer to the list of free blocks
		addDescriptorField("coffee_listel_t*", "freemem", "_heap", false, true, true);
		domains.add(dom);
	}

        /**
         * Check if a slot size is valid.
         *
         * slot size must be a power of 2.
         * minimal slot size is 4.
         * maximal slot size is 128.
         */
	@Override
        protected int alignSlotSize(int ss) {
                for (int mins=4;mins<=128;mins=mins<<1) {
                        if (ss<=mins) return mins;
                }
                return 128;
        }


	public void translate(Coder coder) throws CompileException {
		// Write functions for this heap (allocator, etc.)
		// only done once per implementation
		SystemDefinition sysDef = opts.getSysDef();
		if (once != sysDef) {
			once = sysDef;

			coder.beginHeader(this);
			coder.beginClassMethod(this);

			super.translate(coder);

			if (opts.hasOption("gc_log_unix")) {
				coder.add_fkt("heap/log-unix.c");
			} else if (opts.hasOption("gc_log_i4trace")) {
				coder.add_fkt("heap/log-i4trace.c");
			} else {
				coder.add_fkt("heap/log-none.c");
			}

			coder.add_fkt("heap/coffee-break-alloc.c");

			coder.add_fkt("heap/coffee-break-gc.c");
			if (opts.hasOption("gc_dbg_bitmap"))
				coder.add_fkt("heap/bitmap_dbg.c");

			coder.endClassMethod();
			coder.endHeader();
		}
	}

	public void initHeap(Coder coder, PrintStream out) throws CompileException {
		// init the head element of the freemem list
		String ddesc = "\t" + coder.getDDesc(domain.getDef()) + ".hd.freemem->";

		out.println(ddesc+ "size = " + slotCnt + ";");
		out.println(ddesc+ "next = (coffee_listel_t*) NULL;");
	}

	protected List<IMDomain> domains() {
		return domains;
	}
}

