/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.*;
import keso.compiler.config.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.parser.*;
import keso.compiler.config.task.*;
import java.util.*;
import java.io.PrintStream;

/**
 * Heap divided into slots of a configurable size with a simple garbage
 * collector shared among all instances of this Heap implementation that
 * cleans memory of participating domains in a round robin manner when the
 * system is idle.
 * <p>
 * The Garbage collector uses a bitmap that contains one bit for each slot.
 * The bitmap will be shared for all domains and must, therefore the domain with
 * the most slots determines its size.
 * <p>
 * The Garbage collector runs with the lowest priority in the system and
 * therefore only when all the other tasks are in suspended or waiting state.
 * The GC therefore only needs to scan static fields and immortal objects. If
 * there are waiting tasks, their stack also needs to be scanned which can be
 * achieved with use of the linked object list.
 * <p>
 * Free Memory is managed with a linked list, contained directly in the free
 * memory slots (slotsize must at least be 4 bytes, which is the current size
 * of the object header). The linked list is regenerated from the bitmap
 * during each GC cycle in order to merge free slots.
 * The listdata will be saved in a list_el structure at the beginning of each
 * free block, containing the size of the block in slots and the offset in
 * slots to the next free block.
 */
public class IMHeapIdleRoundRobin extends IMHeapSlottedGC {

	/* Resources used to synchronize stack scanning */
	private final static Map<String, NativeResource> gcresources
	        = new HashMap<String, NativeResource>();

	/**
	 * Contains the domain_ids of all domains that
	 * make use of this Heaptype.
	 */
	private final static List<IMDomain> domains = new ArrayList<IMDomain>();
	private final static List<String> domainresources = new ArrayList<String>();

	/**
	 * Flag to indicate if one-time translation parts like the implementations
	 * functions have already been emitted or not.
	 */
	private static SystemDefinition once = null;

	private NativeResource gcresource;

	public IMHeapIdleRoundRobin(IMDomain dom, BuilderOptions opts, int size, ComplexAttribute<?> heapdef)
			throws CompileException {
		super("irr", "irr", size,opts,dom,heapdef);

		IMGlobalHeap gheap = opts.getGlobalHeap();
		gheap.enableWriteBarriers();

		// pointer to the beginning of the heap
		addDescriptorField("irr_listel_t*", "heap_top", "_heap", false, true, true);
		// pointer to the list of free blocks
		addDescriptorField("volatile irr_listel_t*", "freemem", "_heap", false, true, true);

		// add Resource to tasks that use waitevent
		for(TaskDefinition task: dom.getDef().getTasks()) {
			if(task.getEvents().size()>0) {
				try {
					task.addResource(gcresource);
				} catch (Exception e) {
					throw new CompileException(e.getMessage());
				}
			}
		}

		if(gcresource!=null) {
			domainresources.add( gcresource.getIdentifier() );
			gcresource.recalcCeilingPriority(); // fix ceiling
		}
		domains.add(dom);
	}

	public void translate(Coder coder) throws CompileException {
		// Write functions for this heap (allocator, etc.) only done once per implementation
		SystemDefinition sysDef = opts.getSysDef();
		if (once != sysDef) {
			once = sysDef;

			coder.beginHeader(this);
			coder.beginClassMethod(this);

			super.translate(coder);

			// generate module with this heap's functions
			coder.add_fkt("heap/irr-alloc.c");
			coder.add_fkt("heap/irr-gc.c");
			coder.endClassMethod();
			coder.endHeader();
		}
	}

	protected void addDomainArray(Coder coder) throws CompileException {
		boolean multGCRes = domainresources.size() > 1;

		coder.local_add("#define MANAGEDDOMAINS ");
		coder.local_add(domains.size());
		coder.local_add("\n");

		// provides the currently processed domain's descriptor
		if(domains.size() == 1) {
			coder.local_add("#define DOMAINID() (&");
			coder.local_add(coder.getDDesc(domains.get(0).getDef()));
			coder.local_add(")\n");
		} else {
			coder.local_add("#define DOMAINID() (domains[keso_curdom]");
			if(multGCRes) coder.local_add("[0]");
			coder.local_add(")\n");

			coder.local_add("static domain_irr_t * KESO_CONST domains[MANAGEDDOMAINS]");
			if(multGCRes) coder.local_add("[2]");
			coder.local_add(" = {\n");

			for(int i=0; i<domains.size(); i++) {
				coder.local_add("\t");
				if(multGCRes) coder.local_add("{ ");

				coder.local_add("&");
				coder.local_add( coder.getDDesc(domains.get(i).getDef()) );

				if(multGCRes) {
					coder.local_add(", ");
					coder.local_add(domainresources.get(i));
					coder.local_add("}");
				}
				coder.local_add(",\n");
			}
			coder.local_add("};\n");
		}

		// provides the currently processed domain's sync resource
		if(domainresources.size() > 1) {
				coder.local_add("#define GCRESOURCE() (domains[keso_curdom][1])\n");
		} else if(domainresources.size() == 1) {
				coder.local_add("#define GCRESOURCE() ");
				coder.local_add(domainresources.get(0));
				coder.local_add("\n");
		}
	}

	protected void enhanceSysDef() throws CompileException {
		super.enhanceSysDef();

		String groupname;
		boolean createResource=false;
		DomainZero domz = opts.getSysDef().getDomainZero();

		// Check if this domain needs a gc resource
		// the domain needs a resource if it contains blocking tasks
		for(TaskDefinition task: domain.getDef().getTasks()) {
			if(task.getEvents().size()>0) { createResource=true; break; }
		}
		if(createResource==false) return;

		// create the resource for the sync group if not done previously
		try {
			groupname = heapdef.getAttribute("Group").valueString();
		} catch (NullPointerException e) {
			groupname="Default";
		}
		gcresource = gcresources.get(groupname);
		if(gcresource != null) return;

		// create GC Resource
		gcresource = new NativeResource(domz, "IRRHEAP_"+groupname);
		gcresources.put(groupname, gcresource);

		try {
			domz.addResource(gcresource);
		} catch (ParseException e) {
			throw new CompileException("application use of reserved identifier:\n" + e.getMessage());
		}

		try {
			gctask.addResource(gcresource);
		} catch (Exception e) {
			throw new CompileException(e.getMessage());
		}
	}

	public void initHeap(Coder coder, PrintStream out) throws CompileException {
		// init the head element of the freemem list
		String ddesc = "\t" + coder.getDDesc(domain.getDef()) + ".hd.freemem->";

		out.println(ddesc+ "size = " + slotCnt + ";");
		out.println(ddesc+ "next = (irr_listel_t*) -1;");
		out.println(ddesc+ "flags = 0;");
	}

	protected List<IMDomain> domains() { return domains; }
}
