/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.*;
import keso.compiler.config.*;
import keso.compiler.config.parser.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.*;

import java.util.*;

/**
 * Code common to IRR and Coffeebreak GC.
 */
public abstract class IMHeapSlottedGC extends IMHeap {
	/* Definition of the GC Task */
	protected static NativeTask gctask;

	/**
	 * Size of a slot on the heap.
	 */
	public final int slotSize;

	/**
	 * Number of slots used by the heap.
	 */
	public final int slotCnt;

	/**
	 * max #slots of any instance created.
	 */
	protected static int maxslots=0;

	/**
	 * number of reference slots on the working stack used for scanning.
	 */
	protected static int workingStackSize;

	/**
	* min slot size of any instance created.
	*/
	protected static int minSlotSize = 128;


	/**
	 * Remember the gctmode of previous domains
	 */
	protected final static int ANY                = -1;
	protected final static int LAZY               = 0;
	protected final static int WORKAHOLIC         = 1;
	protected final static int ENFORCEONDEMAND    = 2;
	protected final static int DEFERTASKSONDEMAND = 3;

	protected static final String[] modes = {
		"LAZY",
		"WORKAHOLIC",
		"ENFORCEONDEMAND",
		"DEFERTASKSONDEMAND",
	};

	protected static int gctmode = ANY;

	protected IMHeapSlottedGC(String name, String alias, int size, BuilderOptions opts,
			IMDomain dom, ComplexAttribute<?> heapdef) throws CompileException {

		super(name, alias, size,opts,dom,heapdef);

		if (heapdef.getAttribute("slotSize")==null) {
			opts.warn("no slot size specified; using default slot size of 8 bytes");
			this.slotSize = 8;
		} else {
			int ss = heapdef.getAttribute("slotSize").valueInt();
			this.slotSize = alignSlotSize(ss);
			if (this.slotSize!=ss) {
				opts.warn("Given slot size is invalid! using: "+this.slotSize);
			}
		}

		if (this.slotSize<minSlotSize) {
			minSlotSize = this.slotSize;
		}

		slotCnt = size/slotSize;
		if (slotCnt>maxslots)
			maxslots=slotCnt;

		if(slotCnt * slotSize != heap_size) {
			opts.warn("heap size " + heap_size + " is not a multiple of the slot size " +this.slotSize);
			heap_size = slotCnt * slotSize;
		}

		if (heapdef.getAttribute("WorkingStackSize")==null) {
			workingStackSize = slotCnt;
		} else {
			workingStackSize = heapdef.getAttribute("WorkingStackSize").valueInt();
		}

		if (heapdef.getAttribute("GCTMode") != null ) {
			String gctmodestring = heapdef.getAttribute("GCTMode").valueString().toUpperCase();

			int gmode=0;
			while (gmode<modes.length) {
				if (modes[gmode].equals(gctmodestring))
					break;
				gmode++;
			}

			if (gmode>=modes.length) {
				opts.warn("Unknown GC mode: "+gctmodestring);
			} else if (gctmode == ANY) {
				gctmode = gmode;
			} else if (gctmode != gmode) {
				throw new CompileException("All Domains need the same GC mode!");
			}
		}

		enhanceSysDef();

		IMGlobalHeap gheap = opts.getGlobalHeap();
		if(needStackScan()) {
			gheap.enableStackScan();
		} else {
			gheap.enableNeedObjSize();
		}

		gheap.enableObjColor();

		if (gctmode==ENFORCEONDEMAND)
			gheap.enableBlockingAllocation();

		if (gctmode==DEFERTASKSONDEMAND)
			gheap.enableDeferTasks();

		// free slots on the heap
		//addDescriptorField("keso_gc_slotcnt_t", "freeslots", Integer.toString(slotCnt), false);
		// slots allocated since last scan
		addDescriptorField("keso_gc_slotcnt_t", "sasls", "0", false);
		// heapsize in bytes
		//addDescriptorField("unsigned int", "heap_size", String.valueOf(heap_size), false);
		// size of one slot in bytes
		addDescriptorField("unsigned char", "slotSize", new Integer(slotSize).toString(), false);
	}

	protected boolean needStackScan() {
		for(TaskDefinition tdef: domain.getDef().getTasks()) {
			if(tdef.getEvents().size() > 0) {
				return true;
			}
		}
		return false;
	}

	private void needStackScan(Coder coder) throws CompileException {
		/*
		 * Innerhalb einer Methode werden Objektreferenzen in einem Objekt-Array
		 * registriert. So ist es fuer den GC moeglich die lebenden Objekte beim
		 * Stack-Scan sicher zu finden. Dieser Ansatz ist architektur-unabhaengig
		 * und *sollte* mit jedem C-Compiler funktionieren.
		 *
		 * Die Objekt-Arrays der einzelnen Methoden sind wiederrum zu einer
		 * Liste verlinkt. Ein Feld im Array kann folgende Inhalte besitzen.
		 *
		 * 1. NULL (leer)
		 * 2. Zeiger auf ein Objekt (Objekt-Zeiger)
		 * 3. Zeiger auf das naechste Objekt-Array (Stack-Zeiger)
		 * 4. KESO_EOLL (end of linked list)
		 *
		 * Der GC muss beim Stack-Scan diese Inhalte unterscheiden. Die Inhalte
		 * (1) und (4) sind dabei trivial. Fuer (2) und (3) ist das Makro
		 * keso_isStackRef() zustaendig. Hier fuer gibt es unterschiedliche
		 * Implementierungen.
		 *
		 * 1. Im Stack-Zeiger wird das niederwertigste Bit als Marker gesetzt.
		 *    (Overhead zur Laufzeit und geht nicht bei 8-Bit-Architekturen.)
		 *    Anmerkung: Durch die Slot-Einteilung ist ein Objekt immer ein vielfaches
		 *    von 2-Byte gross, daher funktioniert dieser Mechanismus (ein 16-bit
		 *    Alignment des Gesamtheaps vorausgesetzt) auch auf 8-Bit MCUs.
		 * 2. Dem Stack-Zeiger folgt ein Markersymbol.
		 *    (leicht erhöhter Stack-Bedarf.)
		 */
		if (opts.useLLRefMarker2()) { // Variant 2
			opts.warn("keso_isStackRef is using untested llref marker version 2!");
			coder.global_header_add("#define KESO_FRAME_END_MARK ((object_t*) -2)\n");
			coder.global_header_add("#define keso_isStackRef(_ref_) ( ((object_t**)&(_ref_))[1]==KESO_FRAME_END_MARK)\n");
			coder.global_header_add("#define keso_unpackStackRef(_ref_) (_ref_)\n");

		} else { // Variant 1
			coder.global_header_add("#define keso_isStackRef(_ref_) ");
			coder.global_header_add("((unsigned int)(_ref_)&0x1)\n");
			coder.global_header_add("#define keso_unpackStackRef(_ref_) ");
			coder.global_header_add("(object_t**)((unsigned int)(_ref_)-1)\n");
		}
	}

	public void translate(Coder coder) throws CompileException {
		super.translate(coder);

		int rootset_size = addRootDescriptorField(domains(),coder);

		if (gctmode==ANY) gctmode = LAZY;
		coder.global_header_add("#define KESO_GCMODE_");
		coder.global_header_add(modes[gctmode]);
		coder.global_header_add(" 1\n");

		if (minSlotSize<8) {
			coder.global_header_add("#define KESO_HEAP_HAS_TINY_SLOTS 1\n");
		}

		//////////////////////////////////////////////////////////////////////////////////////
		/////////////////////////   Revisit this   ///////////////////////////////////////////
		//////////////////////////////////////////////////////////////////////////////////////
		coder.local_add("#define KESO_SEND_GCRUN_EVENTS() ");
		if (gctmode==ENFORCEONDEMAND || gctmode==DEFERTASKSONDEMAND) {
			coder.local_add("{ int _t; for (_t=0;_t<KESO_MAX_TASK;_t++) ");
			coder.local_add("SetEvent(_t, KESO_GCRun); }");
		}
		coder.local_add("\n");

		coder.local_add("#define KESO_GC_DEFER_TASK() ");
		if (gctmode==DEFERTASKSONDEMAND) {
			coder.local_add("{ if (KESO_CURRENT_DOMAIN->heap_size <");
			coder.local_add("(KESO_CURRENT_DOMAIN->heap_size>>1)");
			coder.local_add(") keso_defer_task(); }");
			coder.local_add("\nvoid keso_defer_task();");
			coder.local_add("void keso_defer_task() {\n");
			coder.local_add("\tWaitEvent(KESO_GCRun);ClearEvent(KESO_GCRun);\n");
			coder.local_add("}\n");
		}
		coder.local_add("\n");
		//////////////////////////////////////////////////////////////////////////////////////

		coder.header_addln("#include <stdint.h>");
		if(maxslots < (256)) {
			coder.header_addln("typedef uint8_t keso_gc_slotcnt_t;");
		} else if (maxslots < (65535)) {
			coder.header_addln("typedef uint16_t keso_gc_slotcnt_t;");
		} else {
			coder.header_addln("typedef uint32_t keso_gc_slotcnt_t;");
		}

		coder.local_add("#include \"");
		coder.local_add(alias);
		coder.local_add(".h\"\n\n");

		// define size of the root set
		coder.local_add("#define KESO_GC_ROOTSETSIZE " + rootset_size + "\n");

		// define the maximum number of slots
		coder.local_add("#define KESO_GC_MAXSLOTS ");
		coder.local_add(maxslots);
		coder.local_add("\n");

		coder.local_add("#define KESO_GC_WORKINGSTACK_SIZE ");
		coder.local_add(workingStackSize);
		coder.local_add("\n");

		coder.local_add("static unsigned int keso_gc_bitmap[(KESO_GC_MAXSLOTS+sizeof(unsigned int)*8-1)/(sizeof(unsigned int)*8)];\n");

		addDomainArray(coder);

		// create list of all blocking tasks whose stack may need scanning
		if(needStackScan()) {
		 	needStackScan(coder);

			int blocktasks = 0;
			coder.local_add("static KESO_TASKCLASSTYPE * KESO_CONST blocking_tasks[] = {\n");
			for(JavaTask jd: opts.getSysJavaTasks()) {
				if(jd.getEvents().size() > 0) {
					coder.local_add("\t&");
					coder.local_add(coder.getResObjSymbol(jd));
					coder.local_add(",\n");
					blocktasks++;
				}
			}
			coder.local_add("};\n");
			coder.local_add("#define NUM_BLOCKING_TASKS " + blocktasks + "\n");
		}
		coder.add_fkt("heap/gc-common-llrefs.c");
	}

	protected void addDomainArray(Coder coder) throws CompileException {
		coder.local_add("#define MANAGEDDOMAINS ");
		coder.local_add(domains().size());
		coder.local_add("\n");

		if(domains().size() == 1) {
			coder.local_add("#define DOMAINID() (&");
			coder.local_add(coder.getDDesc(domains().get(0).getDef()));
			coder.local_add(")\n");
		} else {
			coder.local_add("#define DOMAINID() (domains[keso_curdom])\n");

			coder.local_add("static struct domain_"+name+"_t * KESO_CONST domains[MANAGEDDOMAINS] = {");
			for(int i=0; i<domains().size(); i++) {
				coder.local_add("&");
				coder.local_add( coder.getDDesc(domains().get(i).getDef()) );
				coder.local_add(", ");
			}
			coder.local_add(" };\n");
		}
	}

	protected void enhanceSysDef() throws CompileException {
		// create GC Task
		if(gctask!=null)
			return;

		DomainZero domz  = opts.getSysDef().getDomainZero();
		gctask = new NativeTask(domz, "keso_"+name+"_gc",-1);
		gctask.setSchedule(true);
		gctask.setStackSize(256);
		try {
			gctask.setAutoStart(true);
			domz.addTask(gctask);
		} catch(ParseException e) {
			throw new CompileException("application use of reserved identifier:\n" + e.getMessage());
		}
	}

	/**
	 * Check if a slot size is valid.
	 *
	 * slot size must be a power of 2.
	 * minimal slot size is 8.
	 * maximal slot size is 128.
	 */
	protected int alignSlotSize(int ss) {
		for (int mins=8;mins<=128;mins=mins<<1) {
			if (ss<=mins) return mins;
		}
		return 128;
	}

	protected int addRootDescriptorField(List<IMDomain> managedDomains,
			Coder coder) throws CompileException {

		Collection<Collection<UserObject>> uobjs
			= new ArrayList<Collection<UserObject>>();

		List<Collection<UserObject>> refuobjs
			= new ArrayList<Collection<UserObject>>();

		int rootset_size = 0;
		ClassStore repository = coder.getClassStore();

		// determine user objects for each domain
		for(IMDomain imd: managedDomains) {
			JavaDomain jd = imd.getDef();
			Collection<UserObject> domobjs = new ArrayList<UserObject>();

			for(JavaTask jt: jd.getTasks())
				domobjs.add(jt);

			for(ServiceDefinition sd: jd.getServices())
				domobjs.add(sd);
			uobjs.add(domobjs);
		}

		// determine user objects with references for each domain and size of the
		// root set
		for(Collection<UserObject> uov: uobjs) {
			Collection<UserObject> refdomuobjs = new ArrayList<UserObject>();
			for(UserObject uo: uov) {
				IMClass mainClass = repository.getClass(uo.getUserResourceClass());
				if(mainClass.getObjRefFieldCount() != 0)
					refdomuobjs.add(uo);
			}
			refuobjs.add(refdomuobjs);
			if(refdomuobjs.size() > rootset_size)
				rootset_size = refdomuobjs.size();
		}

		if(rootset_size == 0)
			return 0;

		for(int i=0; i<refuobjs.size(); i++) {
			Collection<UserObject> uov = refuobjs.get(i);
			StringBuffer initVal = new StringBuffer("{ ");
			for(UserObject uo: uov) {
				IMClass mainClass = repository.getClass(uo.getUserResourceClass());
				initVal.append(coder.add_accessStaticObject(mainClass, coder.getUserResObjSymbol(uo)) + ",");
			}
			for(int j=uov.size(); j<rootset_size; j++)
				initVal.append("NULL,");

			initVal.setCharAt(initVal.length()-1, '}');

			managedDomains.get(i).getHeap().addDescriptorField("object_t * KESO_CONST",
					"rootset["+rootset_size+"]",
					initVal.toString(), false);
		}

		return rootset_size;
	}

	protected abstract List<IMDomain> domains();
}

