/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.*;
import keso.classfile.datatypes.*;
import java.util.*;
import keso.compiler.*;
import keso.compiler.config.domain.*;
import keso.compiler.backend.Coder;
import keso.compiler.config.*;

/**
 * This class implements the portal call to a service.
 */
public class IMLocalPortalMethod extends IMSystemMethod {

	private ImportDefinition importDefinition;
	private ClassStore repository;

	/**
	  IMPortalMethod repraesentiert eine Portal-Methode und all ihre Daten in der
	  Zwischendarstellung des Compilers. Die Klasse der Methode und die Daten aus der
	  Klassendatei werden als Parameter bei der Objekt-Erzeugung uebergeben. Der
	  Bytecode wird nicht gelesen. Die Methode dient als Proxy fuer die eigentliche
	  Implementierung.
	  */
	public IMLocalPortalMethod(IMPortalClass clazz, MethodData source, ClassStore repository, ImportDefinition impDef) throws CompileException {
		super(clazz);
		this.source = source;
		hasNoCode = false;
		this.frame  = new IMMethodFrame(this, source.getInitialVariableTypeDescs(), 0);
		this.repository = repository;
		importDefinition = impDef;
	}

	@Override
	public boolean canBlock() throws CompileException {
		ServiceDefinition srv = importDefinition.getService();
		IMClass srv_class = srv.getServiceClass(repository);

		return srv_class.getMethod(getMethodNameAndType()).canBlock();
	}

	@Override
	public void translate(Coder coder) throws CompileException {
		ServiceDefinition srv = importDefinition.getService();

		// Get all domains, a service is defined in
		List<JavaDomain> services_domains = srv.getDomains();
		List<IMDomain> imdomains = IMDomain.get_IMDomainList(services_domains, opts);

		IMClass srv_class = srv.getServiceClass(repository);
		BasicTypeDescriptor typeDesc=null;
		IMClass task_clazz = getIMClass("java/lang/Thread");
		IMMethod callee = srv_class.getMethod(getMethodNameAndType());
		boolean has_return_value = (getBasicReturnType()!=BCBasicDatatype.VOID);
		boolean has_llref = (callee.canBlock() && opts.hasLinkedListOfLocalReferences());
		int domain_count=services_domains.size();
		String rettype=null;
		try {
			// get external heap definitions of replicated service domains
			if(domain_count > 1) {
				coder.add_domainHeapDefinition(imdomains);
			}
			coder.addInvokeMakro(this);

			coder.beginMethod(this);
			coder.add_PortalHelperMethod(this, srv);

			// if a service is replicated
			if(domain_count > 1) {
				// This method handles primitive and array types
				coder.add_PortalFastVoterMethod(this, srv);
				typeDesc = this.getMethodData().getReturnType();
				if(getBasicReturnType()==BCBasicDatatype.REFERENCE && typeDesc.isArray()==false) {
					// this method handles other class types
					coder.add_PortalVoterMethod(this, srv);
				}
			}

			coder.add_class(clazz.getAlias());
			coder.add_class(srv_class.getAlias());
			coder.add_class(task_clazz.getAlias());

			frame.translate(coder);
			// if it is a replicated service
			for(JavaDomain jd : services_domains) {
				if (has_return_value) {
					rettype=getReturnType();
					coder.add(rettype);
					coder.add(" ret_value_");
					coder.add(coder.getDomainSymbolPrefix(jd));
					coder.addln(";");
				}
			}
			for(JavaDomain jd : services_domains) {

				coder.addln("/* call service object */");
				coder.add("obj0 = ");
				coder.add(coder.getDomainSymbolPrefix(jd).toUpperCase() + "_");
				coder.add(IMPortalService.getGetMacro(srv.getServiceName()));
				coder.addln(";");

				String[] args = frame.getArgNames();

			if (has_llref) {
				coder.addln("KESO_CONST keso_stack_t stack = {");
				coder.add("\t/* domain_desc */ ");
				coder.add(coder.getDDescPtr(srv.getDomain()));
				coder.addln(",");
				coder.addln("\t/* llrefs */ (object_t **) KESO_EOLL,");
				coder.addln("\t/* next */   KESO_CURRENT_TASK->_stack");
				coder.addln("};");

				// terminate llref of previous partition
				coder.addln("*pref=KESO_EOLL;");
				// tail of the new llref
				coder.addln("pref = (object_t**) &stack.llrefs;");

				// push new stack partition
				coder.addln("KESO_CURRENT_TASK->_stack = &stack;\n");
						}

				if (has_return_value) {
					coder.add("ret_value_");
					coder.add(coder.getDomainSymbolPrefix(jd));
					coder.add(" = ");
				}
				coder.add(getIdentifier());
				coder.add("_helper(");
				coder.add(coder.getDDescPtr(jd));
				coder.add(coder.extraProxyHelperArgs(srv));
				for (int i=0;i<args.length;i++) {
					coder.add(", ");
					coder.add(args[i]);
				}
				coder.addln(");\n");
			}

			// pop stack partition
			if(has_llref)
				coder.addln("KESO_CURRENT_TASK->_stack = stack.next;");
			if (has_return_value) {
				if(domain_count==1) {
					coder.add("return ret_value_");
					coder.add(coder.getDomainSymbolPrefix(services_domains.get(0)));
					coder.addln(";");
				}
				// vote over results in replicated mode (multiple domains)
				else {
					// if the return type is an array, we can fastVote in a loop over the array elements
					if(typeDesc.isArray() == true) {
						coder.add("return "+this.toString()+"_arrayVote(");
					}
					// if we have a primitive data type, we can fastVote now
					if(getBasicReturnType()!=BCBasicDatatype.REFERENCE) {
						coder.add("return "+this.toString()+"_fastVote(");
					}
				  // we have to split the vote
					if(getBasicReturnType()==BCBasicDatatype.REFERENCE && typeDesc.isArray() == false) {
						coder.add("return "+this.toString()+"_vote(");
					}
					for(int i=0; i < services_domains.size(); i++) {
						coder.add("ret_value_");
						coder.add(coder.getDomainSymbolPrefix(services_domains.get(i)));
						// last argument
						if(i==(services_domains.size()-1)){
							break;
						}
						coder.add(", ");
					}
					coder.addln(");");
				}
			}
			coder.endMethod();
			// create vote method
			if(domain_count > 1) {
				coder.add(rettype);
				//coder.add(" vote(
			}
		} catch (CompileException ex) {
			System.err.println("CompileException in "+getClassName()+"."+getMethodNameAndType());
			throw ex;
		}
	}

	public ImportDefinition getImportDefinition() {
		return importDefinition;
	}

	@Override
	public <R> R accept(IMMethodVisitor<R> visitor) throws CompileException {
		return visitor.visit(this);
	}

}
