/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import java.util.*;

import keso.classfile.*;
import keso.classfile.constantpool.*;
import keso.classfile.datatypes.*;
import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.backend.*;
import keso.compiler.kni.*;
import keso.util.*;

final class JumpStack {

	private IMBasicBlock[] stack;
	private int pos;

	public JumpStack() {
		stack = new IMBasicBlock[60];
		pos = 0;
	}

	public void push(IMBasicBlockList bl) {
		for (int i=0;i<bl.length();i++) push(bl.at(i));
	}

	public void push(IMBasicBlock node) {
		stack[pos++] = node;
		if (pos==stack.length) {
			IMBasicBlock[] new_stack = new IMBasicBlock[stack.length+10];
			for (int i=0;i<stack.length;i++) new_stack[i] = stack[i];
			stack = new_stack;
			new_stack=null;
		}
	}

	public IMBasicBlock pop() {
		if (pos<=0) return null;
		return stack[--pos];
	}
}

final class MethodEntry {

	String className;
	String nameAndType;
	private int hash_value;

	MethodEntry(String c, String n) {
		className=c;
		nameAndType=n;
		hash_value = className.hashCode() + nameAndType.hashCode();
	}

	public int hashCode() {
		return hash_value;
	}

	public boolean equals(java.lang.Object obj) {
		if (!(obj instanceof MethodEntry)) return false;
		MethodEntry e = (MethodEntry) obj;
		if (!e.className.equals(className)) return false;
		if (e.nameAndType.equals(nameAndType)) return true;
		return false;
	}
}

final class AliasEntry {

	IMBasicBlock bb;
	IMSlot var;
	IMNode value;

	public AliasEntry(IMBasicBlock bb, IMSlot var, IMNode value) {
		this.bb=bb;
		this.var=var;
		this.value=value;
	}
}

public class IMMethod implements Iterable<IMBasicBlock>, Comparable<IMMethod> {

	private final class Itr implements Iterator<IMBasicBlock> {

		private IMBasicBlock curr = labelList;

		@Override
		public boolean hasNext() {
			currentBB = null;
			return (curr != null);
		}

		@Override
		public IMBasicBlock next() {
			if (curr == null)
				throw new NoSuchElementException();
			IMBasicBlock prev = curr;
			curr      = (IMBasicBlock) curr.next();
			currentBB = prev;
			return prev;
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

	}

	final static private boolean dbg_process_vstack = false;

	// BEGIN RT check statistics
	public int nullCheckEmitted;
	public int nullCheckAOTSuccess;
	public int nullCheckAOTFail;
	public int boundsCheckFull;
	public int boundsCheckKnownIndex;
	public int boundsCheckConstant;
	public int boundsCheckOmitted;
	public int castCheckEmitted;
	public int castCheckFast;
	public int castCheckIFace;
	// END RT check statistics

	private static final int[] aType = new int[] {
		-1,-1,-1,-1,
		BCBasicDatatype.BOOLEAN, BCBasicDatatype.CHAR,
		BCBasicDatatype.FLOAT, BCBasicDatatype.DOUBLE,
		BCBasicDatatype.BYTE, BCBasicDatatype.SHORT,
		BCBasicDatatype.INT, BCBasicDatatype.LONG };

	private boolean stack_processed = false;
	protected String 	alias;
	protected IMClass	clazz;
	protected MethodData   source;
	private   StringBuffer methodBody;
	private   SortedSet<String> includes;
	protected IMMethodFrame   frame;
	protected BuilderOptions  opts;
	protected ConstantPool    cPool;
	protected MethodTable	mtable;
	protected Set<MethodEntry> calledMethods;
	protected Collection<IMCallerEntry> calledFrom;

	private IMBasicBlock	 currentBB;
	private IMBasicBlock     labelList;
	private IntegerHashtable<IMBasicBlock> label_index;
	private IMExceptionHandler[]     catch_list;
	private IntegerHashtable<String> local_exception_calls;
	private IMBasicBlock	 endBlock;
	private IMSlot		 returnSlot;
	private int 		 addr_slot;

	private DominatorTree domTree;

	protected boolean isInit;
	protected boolean analyzed=false;
	protected boolean	isPure=true;
	protected boolean	isConst=true;
	protected boolean hasNoCode;
	private static boolean analyseCalleesDone = false;
	private boolean already_inlined = false;

	private IntegerHashtable<LocalVariableData> localVariableByIndex;
	private HashSet<String> usedLocalVariableNames;

	protected  IMMethod(IMClass clazz) throws CompileException {
		this.clazz  = clazz;
		this.opts   = clazz.getOptions();
		this.cPool  = clazz.getConstantPool();
	}

	/**
	  IMMethod repraesentiert eine Methode und all ihre Daten in der Zwischendarstellung. Die
	  Klasse der Methode und die Daten aus der Klassendatei werden als Parameter bei der
	  Objekt-Erzeugung uebergeben. Enthaelt die Methode ausfuehrbaren Bytecode, so wird dieser
	  gelesen und in eine Zwischendarstellung umgewandelt.
	 */
	public IMMethod(IMClass clazz, MethodData source) throws CompileException {
		this(clazz);

		this.source = source;

		isInit = (source.getMethodName().equals("<init>")
				|| source.getMethodName().equals("<clinit>"));

		if (source.getCode()!=null) {
			hasNoCode=false;
			this.frame  = new IMMethodFrame(this,
					source.getInitialVariableTypeDescs(),
					source.getNumLocalVariables());
			parseBC(source);
		} else {
			hasNoCode=true;
			if (source.isAbstract() || source.isNative()) {
				this.frame  = new IMMethodFrame(this,
						source.getInitialVariableTypeDescs(),
						0);
			}

			if (source.isNative()) {
				isPure = false;
				isConst = false;
			}
		}
	}

	public boolean isLocalVariableNameUsed(String name) {
		try {
			if(usedLocalVariableNames==null) {
				usedLocalVariableNames = new HashSet<String>();
				LocalVariableData[] lvdata = source.getLocalVariableTable();

				for(int i=0; i<lvdata.length; i++) {
					if((lvdata[i].name.equals("this"))) {
						usedLocalVariableNames.add("thisp");
					} else {
						usedLocalVariableNames.add(lvdata[i].name);
					}
				}
			}
		} catch(NullPointerException e) { }

		return usedLocalVariableNames.contains(name);
	}

	public String getLocalVariableName(int lv_idx) {
		String name = null;

		try {
			if(localVariableByIndex==null) {
				localVariableByIndex = new IntegerHashtable<LocalVariableData>();
				LocalVariableData[] lvdata = source.getLocalVariableTable();

				for(int i=0; i<lvdata.length; i++) {
					localVariableByIndex.put(lvdata[i].index,lvdata[i]);
				}
			}

			LocalVariableData lv = localVariableByIndex.get(lv_idx);

			// this is reserved in C++
			name = lv.name.equals("this") ? "thisp" : lv.name;
		} catch(NullPointerException e) { }

		return name;
	}

	/**
	 * Translate the byte code position into a line number.
	 */
	public int getLineNumber(int bcpos) {
		int line = -1;

		LineAttributeData linedata[] = source.getLineNumberTable();
		if (linedata==null) return -1;

		for (int i=0;i<linedata.length;i++) {
			if (bcpos<linedata[i].startBytecodepos) return line;
			line = linedata[i].lineNumber;
		}

		return line;
	}

	/**
	* Create a new uniq label.
	*/
	public String getNewCopyLabel() {
		return clazz.getNewCopyLabel();
	}

	/**
	 * Register a class as used.
	 */
	public void requireClass(int domainID,String className) {
		clazz.requireClass(domainID,className);
	}

	public void call(String className, String methodName) {
		if (calledMethods==null) calledMethods=new HashSet<MethodEntry>();
		MethodEntry e = new MethodEntry(className, methodName);
		if (calledMethods.contains(e)==false)
			calledMethods.add(e);
	}

	/**
	 * Register a method as used.
	 */
	public void requireMethod(int domainID, String className, String methodName) {
		if (opts.pruneClassTree()) clazz.requireMethod(domainID, className, methodName);
		call(className, methodName);
	}

	/**
	 * Returns an unmodifiable view of the methods called by this IMMethod.
	 * @return Collection of this method's callees.
	 */
	public IMMethod[] getCalledMethods() throws CompileException {
		if (calledMethods == null)
			return new IMMethod[0];
		IMMethod[] ret = new IMMethod[calledMethods.size()];
		int        i   = 0;
		for (MethodEntry e : calledMethods) {
			ret[i++] = findMethod(e.className, e.nameAndType);
		}
		return ret;
	}

	public void removeCalledMethod(IMMethod method) throws CompileException {
		if (calledMethods == null)
			return;
		Iterator<MethodEntry> it = calledMethods.iterator();
		while (it.hasNext()) {
			MethodEntry entry = it.next();
			if (findMethod(entry.className, entry.nameAndType) == method) {
				it.remove();
				break;
			}
		}
	}

	/**
	 * Register a method as used.
	 */
	final public void requireMethod(int domainID, MethodRefCPEntry cpEntry) {
		requireMethod(domainID, cpEntry.getClassName(), cpEntry.getMemberName()+cpEntry.getMemberTypeDesc());
	}

	/**
	 * Register a method as used.
	 */
	final public void requireMethod(int domainID, InterfaceMethodRefCPEntry cpEntry) {
		requireMethod(domainID, cpEntry.getClassName(), cpEntry.getMemberName()+cpEntry.getMemberTypeDesc());
	}

	public DominatorTree getDomTree() throws CompileException {
		if (domTree==null) computeDomTree();
		return domTree;
	}

	public void setMethodTable(MethodTable table) {
		if (table != null)
			table.tableIsUsedBy(this);
		mtable = table;
	}

	public MethodTable getMethodTable() {
		return mtable;
	}

	public void setIsAnalyzed() {
		analyzed = true;
	}

	public boolean isAnalyzed() {
		return analyzed;
	}

	public boolean isInit() {
		return isInit;
	}

	final public boolean isClassInit() {
		return getMethodNameAndType().equals("<clinit>()V");
	}

	/**
	 * GCC does better work when it knows more about the function,
	 * therefore we will analyse if a method is _pure_ and use the
	 * GCC function attributes for annotation.
	 *
	 * = gcc info pages ===============================================
	 * Many functions have no effects except the return value and their
	 * return value depends only on the parameters and/or global
	 * variables. Such a function can be subject to common subexpression
	 * elimination and loop optimization just as an arithmetic operator
	 * would be. These functions should be declared with the attribute
	 * `pure'. For example,
	 *
	 * int square (int) __attribute__ ((pure));
	 *
	 * says that the hypothetical function `square' is safe to call fewer
	 * times than the program says.
	 *
	 * Some of common examples of pure functions are `strlen' or `memcmp'.
	 * Interesting non-pure functions are functions with infinite loops
	 * or those depending on volatile memory or other system resource,
	 * that may change between two consecutive calls (such as `feof' in a
	 * multithreading environment).
	 * ================================================================
	 */
	public boolean isPure() throws CompileException {
		if (opts.hasOption("_X_gcc_all_pure")) return true;
		if (opts.hasOption("gcc_no_pure")) return false;
		if (source.isNative()) return false;
		if (Weavelet.FALSE==getJoinPoints().checkAttribute(this, Weavelet.ATTR_PURE)) return false;
		if (isPure && !analyseCalleesDone)
			determineBlockingMethods();
		return isPure;
	}

	/**
	 * GCC does better work when it knows more about the function,
	 * therefore we will analyse if a method is const and use the
	 * GCC function attributes for annotation.
	 *
	 * = gcc info pages ===============================================
	 * Many functions do not examine any values except their arguments,
	 * and have no effects except the return value. Basically this is
	 * just slightly more strict class than the `pure' attribute above,
	 * since function is not allowed to read global memory.
	 *
	 * Note that a function that has pointer arguments and examines the
	 * data pointed to must _not_ be declared `const'. Likewise, a
	 * function that calls a non-`const' function usually must not be
	 * `const'. It does not make sense for a `const' function to return
	 * `void'.
	 * ================================================================
	 */
	public boolean isConstant() throws CompileException {
		if (opts.hasOption("_X_gcc_all_const")) return true;
		if (opts.hasOption("gcc_no_const")) return false;
		if (!isPure()) return false;
		if (Weavelet.FALSE==getJoinPoints().checkAttribute(this, Weavelet.ATTR_CONST)) return false;
		if (isConst && !analyseCalleesDone)
			determineBlockingMethods();
		return isConst;
	}

	public boolean unused() throws CompileException {
		if (calledFrom==null) return true;
		return false;
	}

	public boolean clearUnused() throws CompileException {

		String className = getClassName();

		// FIXME: String methods are invoked in weavelets
		if (className.equals("java/lang/String"))
			return false;
		// FIXME: bug within IMThrow?!?
		if (className.equals("java/lang/Error"))
			return false;

		ReachabilityData rData = IMDomain.getOverallReachabilityData();
		if (false && rData.getMethods().contains(this) == false)
			return true;
		if (opts.hasOption("clear_unused_methods") == false)
			return false;
		return unused();
	}

	public boolean hasCode() {
		return !hasNoCode;
	}

	/**
	 * Analyse if a task can block by calling the method represented the
	 * IMMethod object.
	 */
	private String canBlockIn = blocking_undecided;
	private final static String blocking_undecided = "#undecided";
	private final static String none_blocking = "#none_blocking";
	private final static String blocks_in_new = "#new_operation";
	private final static String blocks_in_self = "#self";

	public boolean canBlock() throws CompileException {
		boolean ret = true;

		if(!analyseCalleesDone)
			determineBlockingMethods();

		if(canBlockIn == blocking_undecided)
			throw new CompileException("Blocking Method analysis has not yet been done");

		if (opts.hasOption("_X_all_CanBlock"))
			ret = true;

		else if (canBlockIn==none_blocking)
			ret = false;

		return ret;
	}

	/**
	 * mark Method as blocking in object allocation
	 */
	public void blockInNew() { canBlockIn = blocks_in_new; }

	/**
	 * BuilderOptions
	 */
	public BuilderOptions getOptions() {
		return opts;
	}

	public IMSlot getMethodArgument(int index) throws CompileException {
		IMSlot[] arg_slots = frame.getMethodArguments();
		if (canBlock() && opts.hasLinkedListOfLocalReferences()) index++;
		return arg_slots[index];
	}

	public IMMethodFrame getMethodFrame() {
		return frame;
	}

	public int getSizeOfReturnValue() throws CompileException {
		int rsize = BuilderOptions.getTypeSize(BuilderOptions.typeToString(getBasicReturnType()));
		return rsize;
	}

	public int getNumberOfArguments() throws CompileException {
		return frame.getMethodArguments().length;
	}

	public int getSizeOfArguments() throws CompileException {
		IMSlot[] argSlots = frame.getMethodArguments();
		int total = 0;

		for (int i = 1; i < argSlots.length; ++i) {
			total += BuilderOptions.getTypeSize(argSlots[i].cType());
		}

		return total;
	}

	/**
	 * getArg(type,nr) liefert ein IMNode-Object fuer den
	 * Zugriff auf die Parameter der Methode.
	 *
	 * @param type
	 * 	(BCBasicDatatype.REFERENCE, ...)
	 * @param nr
	 * 	Bei virtuellen Methoden ist 0 der this-Pointer, alle
	 * 	weiteren Parameter folgen.
	 */
	public IMNode getArg(int type, int nr) throws CompileException {
		return frame.readArgument(type, nr);
	}

	/**
	 * getIMClass() liefert die definierende Klasse der Methode.
	 */
	public IMClass getIMClass() {
		return clazz;
	}

	public void storeMethodBody(StringBuffer body, SortedSet<String> includes) {
		int pos=-1;
		pos=body.indexOf("{");
		if (pos!=-1) {
			methodBody = body.replace(0,pos,"");
		}
		this.includes = includes;
	}

	public StringBuffer getMethodBody() {
		return methodBody;
	}

	public StringBuffer getMethodBodyPretty() {
		int pos=-1;
		methodBody.insert(0,"\t/* "+getAlias()+" */");
		while ((pos=methodBody.indexOf("return;")) != -1) {
			//methodBody.delete(pos,pos+7);
			//methodBody=methodBody.replace(pos,pos+7,"{int bug=1;}/* make f*** gcc happy */");
			methodBody=methodBody.replace(pos,pos+7,"while(0); /* make gcc happy */");
		}
		pos=-1;
		while ((pos=methodBody.indexOf("\n",pos+1)) != -1) { methodBody=methodBody.replace(pos,pos+1,"\n\t"); }
		return methodBody;
	}

	// HACK: inlining for dummies
	public SortedSet<String> getMethodIncludes() {
		return includes;
	}

	/**
	 * Get the global join points.
	 * This is just delegated to the class.
	 * @see IMClass#getJoinPoints()
	 */
	public JoinPointChecker getJoinPoints() {
		return clazz.getJoinPoints();
	}

	/**
	 * Get the global ClassStore.
	 * This is just delegated to the class.
	 * @see IMClass#getClassStore()
	 */
	public ClassStore getClassStore() {
		return clazz.getClassStore();
	}

	/**
	 * Get the IMClass object of a given class.
	 * This is just delegated to the current class store.
	 * @see ClassStore#getClass(String)
	 */
	public IMClass getIMClass(String classname) {
		return clazz.findClass(classname);
	}

	/**
	 * Get the class name of the defining class.
	 */
	public String getClassName() {
		return clazz.getClassName();
	}

	/**
	 * Lookup the IMMethod object by a constant pool entry.
	 */
	final public IMMethod findMethod(MethodRefCPEntry cpEntry) throws CompileException {
		return findMethod(cpEntry.getClassName(), cpEntry.getMemberName()+cpEntry.getMemberTypeDesc());
	}

	/**
	 * Lookup the IMMethod object by a constant pool entry.
	 */
	final public IMMethod findMethod(InterfaceMethodRefCPEntry cpEntry) throws CompileException {
		return findMethod(cpEntry.getClassName(), cpEntry.getMemberName()+cpEntry.getMemberTypeDesc());
	}

	/**
	 * Lookup the IMMethod object by the symbolic name and type.
	 */
	final public IMMethod findMethod(String classname, String nameAndType) throws CompileException {

		IMClass cls = clazz.findClass(classname);
		if (cls==null) throw new CompileException("Class not found "+classname);

		if (nameAndType==null) throw new CompileException("Undefined method name.");

		for (IMClass scls=cls;scls!=null;scls=scls.getSuperClass()) {
			IMMethod m = scls.getMethod(nameAndType);
			if (m!=null) return m;

			IMClass[] _ifaces = cls.getInterfaces();
			for (int i=0;_ifaces!=null && i<_ifaces.length;i++) {
				m = _ifaces[i].getMethod(nameAndType);
				if (m!=null) return m;
			}
		}

		throw error(0,"Method "+classname+"." +nameAndType+" not found!");
	}

	/**
	 * Is true if the method is not void.
	 */
	public boolean hasReturnValue() {
		return (getBasicReturnType()!=BCBasicDatatype.VOID);
	}

	/**
	 * Get the BCBasicDatatype of the return value.
	 */
	public int getBasicReturnType() {
	    return source.getBasicReturnType();
	}

	/**
	 * Returns true, if this method returns a reference type
	 */
	public boolean isReference() {
		return getBasicReturnType() == BCBasicDatatype.REFERENCE;
	}

	/**
	 * Get the C type of the return value.
	 */
	public String getReturnType() {
		return BuilderOptions.typeToString(source.getBasicReturnType());
	}

	/**
	 * Get the whole C like argument list.
	 */
	public String getArgString() throws CompileException {
		return frame.getArgString();
	}

	public String getArgReCallString() throws CompileException {
		return frame.getArgReCallString();
	}

	/**
	 * Get the alias name of the method, which is used in the C source.
	 */
	public String getAlias() {
		if (alias==null) alias=DecoratedNames.createMethodAlias(opts,this);
		return alias;
	}

	public MethodData getSource() {
		return source;
	}

	/**
	 * return the C code identifier.
	 */
	public String getIdentifier() {
		return getAlias();
	}

	/**
	 * return the C code identifier for a virtual method invocation.
	 */
	public String getInvokeAlias() {
		return "INVOKE_"+getIdentifier().toUpperCase();
	}

	public void setAddrSlot(int value) {
		addr_slot = value;
	}

	public int getAddrSlot() {
		return addr_slot;
	}

	public String getMethodName() {
		return source.getMethodName();
	}

	public String getMethodNameAndType() {
		return source.getMethodNameAndType();
	}

	final public boolean termed(String symbol) {
		String term = getMethodNameAndType();
		if (term.equals(symbol)) return true;
		term = getMethodName();
		if (term.equals(symbol)) return true;
		return false;
	}

	public boolean isConstructor() {
		// TODO: remove duplicated method (see isInit)
		return getMethodName().equals("<init>");
	}

	public boolean isStatic() { return source.isStatic(); }

	public boolean isVirtual() {
		return (!source.isStatic() && !isFinal() && !source.isPrivate());
	}

	public boolean isPrivate() { return source.isPrivate(); }

	public boolean isFinal() {
		if (source.isFinal() || clazz.isFinal()) return true;
		if (clazz.hasSubClasses()) {
			IMClass[] subs = clazz.getSubClasses();
			for (int i=0;i<subs.length;i++) {
				if (subs[i].getMethod(getMethodNameAndType())!=null) {
					if (isPrivate()) {
						opts.warn("private method "+getMethodNameAndType()+
								" is overridden in "+subs[i].getClassName());
						return true;
					}
					return false;
				}
			}
			return true;
		}
		return false;
	}

	public int getNumberOfBasicBlocks() {
		int number=0;
		for (IMBasicBlock bb=labelList;bb!=null;bb=(IMBasicBlock)bb.next()) {
			number++;
		}
		return number;
	}

	public String dumpLabelIndex() {
		return label_index.toString();
	}

	public IMBasicBlock createBBlock(int bcpos) {
		return new IMBasicBlock(this,bcpos);
	}

	public IMAMTConstant createMTHeapConstant(IMNode self, IMClass type, int addr) {
		return new IMAMTConstant(this, -1, self.getBCPosition(), type, addr);
	}

	public IMAMTConstantStatic createMTStaticConstant(IMNode self, IMClass type, int addr) {
		return new IMAMTConstantStatic(this, -1, self.getBCPosition(), type, addr);
	}

	public IMAMemConstant createMemConstant(IMNode self, int addr, int size) {
		return new IMAMemConstant(this, -1, self.getBCPosition(), addr, size);
	}

	public IMAObjConstant createObjConstant(IMNode self, String alias, IMClass type) {
		return new IMAObjConstant(this, -1, self.getBCPosition(), alias, type);
	}

	public IMStoreLocalVariable createStoreStackVariable(int stackpos, IMNode expr, int bcpos)
		throws CompileException {

		int type = expr.getDatatype();
		IMSlot vslot = frame.getStackVariable(type, stackpos);
		vslot.setAlias(expr);
		IMStoreLocalVariable nvar = storeLocal(bcpos, vslot);
		nvar.setOperand(expr);
		nvar.clear_link();

		return nvar;
	}

	public IMStoreLocalVariable createStoreVariable(IMSlot vslot, IMNode expr, int bcpos) throws CompileException {
		IMStoreLocalVariable nvar = storeLocal(bcpos, vslot);
		nvar.setOperand(expr);
		nvar.clear_link();
		return nvar;
	}

	public IMReadLocalVariable readLocal(int bcpos, IMSlot vslot) throws CompileException {
		if (vslot==null) throw new CompileException("unknown slot");
		int type = vslot.getType();
		IMReadLocalVariable nvar;
		switch (type) {
			case BCBasicDatatype.BOOLEAN:
			case BCBasicDatatype.BYTE:
			case BCBasicDatatype.CHAR:
			case BCBasicDatatype.SHORT:
			case BCBasicDatatype.INT:
				nvar=new IMIReadLocalVariable(this,21,bcpos,vslot);
				break;
			case BCBasicDatatype.LONG:
				nvar=new IMLReadLocalVariable(this,22,bcpos,vslot);
				break;
			case BCBasicDatatype.FLOAT:
				nvar=new IMFReadLocalVariable(this,23,bcpos,vslot);
				break;
			case BCBasicDatatype.DOUBLE:
				nvar=new IMDReadLocalVariable(this,24,bcpos,vslot);
				break;
			case BCBasicDatatype.REFERENCE:
				nvar=new IMAReadLocalVariable(this,25,bcpos,vslot);
				break;
			default:
				throw error(bcpos,"unknown type: "+type);
		}
		return nvar;
	}

	public IMPopReturnAddr createPopReturnAddr(int bcpos) {
		return new IMPopReturnAddr(this,-1,-1,bcpos);
	}

	public IMConstant createIMIConstant(int value, int bcpos) {
		return new IMIConstant(this,19,bcpos,value);
	}

	public IMConstant createIMLConstant(long value, int bcpos) {
		return new IMLConstant(this,20,bcpos,value);
	}

	public IMConstant createIMFConstant(float value, int bcpos) {
		return new IMFConstant(this,19,bcpos,value);
	}

	public IMConstant createIMDConstant(double value, int bcpos) {
		return new IMDConstant(this,20,bcpos,value);
	}

	public IMConstant createIMAConstant(ConstantPoolEntry cpEntry, int bcpos) {
		return new IMAConstant(this,19,bcpos,cpEntry);
	}

	public IMConstant createIMNullConstant(int bcpos) {
		return new IMNullConstant(this,19,bcpos);
	}

	public IMNode createGoto(IMBasicBlock label, int bcpos) {
		return new IMGoto(this, 167, bcpos, label);
	}

	final public IMSlot getReturnValue() {
		return returnSlot;
	}

	/**
	 * Re-computes the return slot.
	 * This is necessary after any transformation that might do variable
	 * renaming, for example SSA construction or copy propagation.
	 * @return The new return slot.
	 * @throws CompileException
	 */
	final public IMSlot recomputeReturnValue() throws CompileException {

		for (IMBasicBlock bb : this) {

			if (bb.next() != null)
				continue;

			bb.visitNodes(new IMBasicVisitor() {

				@Override
				public boolean skipNode(IMNode node) {
					return !(node instanceof IMEpilog
							|| node instanceof IMReturnSubroutine);
				}

				@Override
				public void visit(IMNode node) throws CompileException {
					if (node instanceof IMEpilog
							|| node instanceof IMReturnSubroutine)
						returnSlot = node.getIMSlot();
				}
			});
		}

		return returnSlot;
	}

	final public IMBasicBlock getEndBlock() {
		return endBlock;
	}

	final public void dumpRawBC() {
		dumpRawBC(source);
	}

	final public void dumpRawBC(MethodData src) {
		BytecodeInputStream bcStream = new BytecodeInputStream(src.getBytecode());
		int i=0;
		while (bcStream.hasMore()) {
			if (i==0) System.err.println();
			int bc = bcStream.readUnsignedByte();
			System.err.print("0x");
			if (bc<16) System.err.print("0");
			System.err.print(Integer.toHexString(bc));
			System.err.print(" ");
			if (bc<10) System.err.print(" ");
			if (bc<100) System.err.print(" ");
			System.err.print(bc);
			System.err.print(":");
			i++;
			if (i>=16) i=0;
		}
		System.err.println();
	}

	final public IMCallerEntry getFoldedCallerInfo() throws CompileException {
		return IMCallerEntry.getFoldedCallerEntry(calledFrom);
	}

	final public IntegerHashtable<IMBasicBlock> getReturnTargets() {
		return ret_index;
	}

	private Map<String, AliasEntry> var_alias_unsafe;
	private Map<String, AliasEntry> var_alias_safe;

	public void registerAlias(IMBasicBlock bb, IMSlot var, IMNode value) throws CompileException {

		// volatile fields are unsafe for alias forwarding
		if (value.isVolatile()) return;

		if (var_alias_unsafe==null)
			var_alias_unsafe = new HashMap<String, AliasEntry>();
		if (var_alias_safe==null)
			var_alias_safe = new HashMap<String, AliasEntry>();

		Map<String, AliasEntry> var_alias;
		if (value.isFinalStatic())
			var_alias = var_alias_safe;
		else
			var_alias = var_alias_unsafe;

		AliasEntry aliasEnt=null;
		if ((aliasEnt=var_alias.get(value.toMemAlias()))==null) {
			aliasEnt = new AliasEntry(bb,var,value);
			var_alias.put(value.toMemAlias(), aliasEnt);
			return;
		}

		if (aliasEnt.bb!=bb) {
			aliasEnt = new AliasEntry(bb,var,value);
			var_alias.put(value.toMemAlias(), aliasEnt);
		}
	}

	/**
	* Clear all alias entries for none static final fields.
	*/
	public void clearUnsafeAliases() {
		var_alias_unsafe=null;
	}

	public void removeAlias(IMBasicBlock bb, IMNode value) throws CompileException {
		Map<String, AliasEntry> var_alias;
		if (value.isFinalStatic())
			var_alias = var_alias_safe;
		else
			var_alias = var_alias_unsafe;
		if (var_alias!=null) var_alias.remove(value.toMemAlias());
	}

	public IMSlot lookupAlias(IMBasicBlock bb, IMNode value) throws CompileException {
		Map<String, AliasEntry> var_alias = var_alias_unsafe;
		if (value.isFinalStatic())
			var_alias = var_alias_safe;
		else
			var_alias = var_alias_unsafe;

		if (var_alias==null) return null;

		AliasEntry aliasEnt = var_alias.get(value.toMemAlias());

		if (aliasEnt==null) return null;

		if (aliasEnt.value==value) return null;

		if (aliasEnt.bb!=bb) {
			/* The alias was defined in a different basic block. */
			if (opts.hasOption("ssa_alias_less")) {
				/* The defining basic block is the only and direct predecessor */
				IMBasicBlockList pb = bb.getPred();
				if (pb.length()==1 && pb.at(0)==aliasEnt.bb) {
					aliasEnt.bb = bb;
					return aliasEnt.var;
				}
			} else {
				/* The defining basic block is a dominator */
				DominatorTree dt = getDomTree();
				if (dt.dom(aliasEnt.bb,bb)) return aliasEnt.var;
			}
			return null;
		}

		return aliasEnt.var;
	}


	/**
	 * Scan all basic blocks and perform constant folding if possible.
	 */

	private boolean another_ssa_run;
	private boolean[] arg_escape;

	public void doAnotherSSARun() {
		another_ssa_run = true;
	}

	private int arg_cycle_l1 = -1;
	private int arg_cycle_l2 = -1;
	public boolean doArgumentEscape(int arg) throws CompileException {

		IMSlot[] args = frame.getOriginalArguments();
		if (args[arg].getType()!=BCBasicDatatype.REFERENCE)
			opts.warn(getAlias()+" args["+arg+"]=>"+args[arg]+" is not a reference!");

		if (opts.hasOption("_X_ssa_less_astack")) return true;

		/* cycles are evil! */
		if (arg_cycle_l1!=-1) {
			/* if it is the same argument we won't escape for now! */
			if (arg_cycle_l1==arg) return false;

			if (arg_cycle_l2!=-1) {
				if (arg_cycle_l2==arg) return false;
			} else {
				arg_cycle_l2=arg;
				boolean arg_tmp = args[arg].doesEscape(true);
				arg_cycle_l2=-1;
				return arg_tmp;
			}

			opts.warn("Can`t resolve cycle in escape analysis!");
			return true;
		}

		if (arg_escape==null) {
			/* no code no escape ;-) */
			if (hasNoCode) return false;

			/* may be we have some hints */
			if (Weavelet.TRUE==opts.getJoinPoints().checkAttribute(this,Weavelet.ATTR_NOESCAPE)) return false;

			/* we are pessimistic if we cannot analyse the method body */
			if (opts.getJoinPoints().ignoreMethodBody(clazz,this)) return true;

			/* do the ssa optimization and escape analysis for the method.
			 * the result is cached in arg_escape[...] */
			arg_cycle_l1 = arg;
			doEscapeAnalyses();
			arg_cycle_l1 = -1;

			if (arg_escape==null) {
				opts.warn("Can't do escape analysis on "+getClassName()+"."+getMethodName());
				return true;
			}
		}

		return arg_escape[arg];
	}

	public boolean doConstantFolding() throws CompileException {

		another_ssa_run  = false;
		var_alias_unsafe = null;
		var_alias_safe   = null;

		// Optimize
		for (IMBasicBlock bb = labelList; bb != null;) {
			currentBB = bb;
			bb = (IMBasicBlock) bb.ssa_optimize();
		}

		// The uses are buggy
		IMUpdateUsesVisitor updateVisitor = new IMUpdateUsesVisitor(this);
		updateVisitor.begin();
		for (IMBasicBlock bb : this) {
			bb.visitNodes(updateVisitor);
		}
		updateVisitor.end();

		// Collect dead code
		IMBasicBlock pb = null;
		for (IMBasicBlock bb = labelList; bb != null; bb = (IMBasicBlock) pb.next()) {
			currentBB = bb;
			pb = (IMBasicBlock) bb.removeDeadBasicBlocks(pb);
		}
		currentBB = null;
		return another_ssa_run;
	}

	public void initEscapeInfo() throws CompileException {
		arg_escape = new boolean[frame.getOriginalArguments().length];
		for (int i = 0; i < arg_escape.length; ++i) {
			arg_escape[i] = true;
		}
	}

	public void setEscapeInfo(int argIndex, boolean doesEscape) {
		arg_escape[argIndex] = doesEscape;
	}

	/**
	 * Scan all basic blocks and inline method calls if it makes
	 * sense.
	 */
	final public void inlineMethodCalls() throws CompileException {

		opts.vverbose("&& inlineMethodCalls in "+getAlias());

		if (already_inlined) return;
		already_inlined = true;

		if (hasNoCode) return;

		if (opts.getJoinPoints().ignoreMethodBody(clazz,this)) return;

		for (IMBasicBlock bb=labelList;bb!=null;) {
			currentBB=bb;
			bb=bb.inlineMethodCalls();
		}

		if (!opts.hasOption("dont_merge_basic_blocks")) {
			IMBasicBlock bb=labelList;
			while (bb!=null) {
				currentBB=bb;
				bb = bb.melt();
			}
		}

		currentBB=null;
	}

	final public int getCosts() throws CompileException {
		int costs = 0;
		for (IMBasicBlock bb=labelList;bb!=null;bb=(IMBasicBlock)bb.next()) {
			currentBB = bb;
			costs += bb.costs();
		}
		currentBB = null;
		return costs;
	}

	private int inline_loop_level = 0;
	public boolean isInlineCandidate() throws CompileException {

		opts.vverbose("## inline ? "+getClassName()+"."+getMethodNameAndType());

		if (source.isNative() || labelList==null) {
			return false;
		}

		// ignore methods with exception handlers
		ExceptionHandlerData[] handler = source.getExceptionHandlers();
		if (handler!=null && handler.length>0) {
			opts.vverbose("# has exception handler");
			return false;
		}

		// preserve join points
		if (opts.getJoinPoints().ignoreMethodBody(clazz,this)) {
			opts.vverbose("# is joint point ");
			return false;
		}

		// Bug: long and double arguments are not managed in the right way!
		if (hasLongOrDoubleArgument()) {
			opts.vwarn("TODO: Don't ignore "+getAlias()+" for in-line expansion.");
			return false;
		}

		if (opts.forceInline(getClassName()+"."+getMethodNameAndType())) {
			opts.warn(getAlias()+" inlining forced");
			return true;
		}

		// first try to inline all methods
		if (inline_loop_level>0) {
			opts.vverbose("# inline loop "+getAlias());
			return false;
		}
		inline_loop_level++;
		if (!already_inlined) inlineMethodCalls();
		inline_loop_level--;

		// calculate inline costs
		int max_costs = opts.maxInlineCosts();
		if (calledFrom!=null && calledFrom.size()==1) {
			// single caller sides are prefered
			max_costs = opts.maxInlineCostsSingle();
		}
		int costs = getCosts();
		if (costs>max_costs) {
			opts.vverbose("# "+getAlias()+" inline costs too high "+costs+">"+max_costs);
			return false;
		}

		opts.vverbose("# "+getAlias()+" is candidate");
		return true;
	}

	/**
	 * Make a copy of the imcode and insert it into the host method.
	 *
	 * host_method  the new method to place the code
	 * current 	current line to place the code
	 * call		the invokation object it self
	 * self		this-pointer of the method call
	 * args		arguments of the method call
	 *
	 * prev_bb	the prev. basic block of the host method
	 * next_bb 	the next basic block of the host method
	 *
	 * callee.insertCode(method, current, call, self, args, this, tail);
	 */
	public boolean insertCode(IMMethod host_method,
			IMNode current, IMNode call, IMNode self, IMNode[] args,
			IMBasicBlock prev_bb, IMBasicBlock next_bb) throws CompileException {

		opts.vverbose("### "+host_method.getAlias());
		opts.vverbose("## "+call.lineInfo()+" inline "+call.toReadableString());

		int bcpos = call.getBCPosition();

		next_bb.getPred().remove(prev_bb);

		//  create a copy of all basic blocks
		IMInlineVisitor visitor;
		if (current != call)
			visitor = new IMInlineVisitor(host_method, self, args, bcpos, next_bb, current, call);
		else
			visitor = new IMInlineVisitor(host_method, self, args, bcpos, next_bb, null, null);

		IMMethodFrame host_frame = host_method.getMethodFrame();

		host_frame.beginInlineForeignFrame();

		IMBasicBlock code = (IMBasicBlock)labelList.copy(visitor);

		/* seek end of code */
		IMBasicBlock cpos = code;
		while (cpos!=null && cpos.next()!=null) {
			if (code.empty()) {
				cpos.getPred().remove(code);
				code=cpos; /* skip emtpy basic blocks */
			}
			cpos=(IMBasicBlock)cpos.next();
		}

		if (code.empty()) {
			opts.vverbose("= empty ==");
		} else {
			cpos.link(next_bb);
			opts.vvverbose("= insert ==");
			opts.vvverbose(code.toReadableString());
			opts.vvverbose("--------");
			IMBasicBlockList cpred = code.getPred();
			opts.vvverbose("-- insert code "+getAlias()+" with pred "+cpred);
			prev_bb.link(code);

			int off=0;

			// store the arguments in new local variables
			// this pointer
			if (self!=null) {
				IMSlot obj0 = frame.getLocalVariable(self.getDatatype(),0);
				if (host_frame.hasUpdate(obj0)) {
					IMStoreLocalVariable store_var = host_method.createStoreVariable(host_frame.adjustSlot(obj0), self, bcpos);
					prev_bb.appendBC(store_var);
				}
				off++;
			}

			// args
			for (int i=0;i<args.length;i++) {
				IMSlot arg = frame.getLocalVariable(args[i].getDatatype(),i+off);
				if (host_frame.hasUpdate(arg)) {
					IMStoreLocalVariable store_var_arg = host_method.createStoreVariable(host_frame.adjustSlot(arg), args[i], bcpos);
					prev_bb.appendBC(store_var_arg);
				}
			}
		}

		host_frame.endInlineForeignFrame();

		opts.vverbose("= copy ==");
		for (IMBasicBlock pos = prev_bb;pos!=null && pos!=next_bb.next();pos=(IMBasicBlock)pos.next()) {
			pos.method=host_method;
			opts.vverbose(pos.getMethod());
			opts.vverbose(pos.toReadableString());
		}
		opts.vverbose("--------");

		return true;
	}

	public IMBasicBlock getCurrentBasicBlock() {
		return currentBB;
	}

	public void emitArguments(Coder coder, IMMethodFrame caller_frame,
			IMNode self, IMNode[] args) throws CompileException {
		frame.emitArguments(coder, caller_frame, self, args);
	}

	public IMExceptionHandler[] getExceptionHandler() {
		return catch_list;
	}

	public String lookupExceptionHandler(IMClass type, int bcpos) throws CompileException {
		String ret=null;
		int range = 0;

		if (catch_list==null) return null;

		for (int i=0;i<catch_list.length;i++) {
			IMExceptionHandler exp_handler = catch_list[i];
			if (exp_handler.inRegion(bcpos)) {
				ret = getAlias()+"_eh_"+exp_handler.getStartPos()+"_"+exp_handler.getEndPos();
				range = exp_handler.getRangeID();
				break;
			}
			// TODO: local short cuts
		}

		if (ret==null) return null;

		if (local_exception_calls==null) {
			local_exception_calls = new IntegerHashtable<String>();
		}

		if (local_exception_calls.get(range)==null) {
			local_exception_calls.put(range, ret);
		}

		return ret;
	}

	public String getLocalExceptionLabel(int range_id) {
		return local_exception_calls.get(range_id);
	}

	public int[] getLocalExceptionCalls() throws CompileException {
		if (local_exception_calls==null) return null;
		return local_exception_calls.sortedKeys();
	}

	/**
	 * This method translates the imcode to the backend language.
	 *
	 * The method is overridden in IMPortalMethod.
	 */
	public void translate(Coder coder) throws CompileException {

		try {

			if (source!=null) coder.addInvokeMakro(this);

			if (hasNoCode) {
				coder.header_add("/* abstract method ");
				coder.header_add(getAlias());
				coder.header_addln(" */");
				coder.header_add("#define ");
				coder.header_add(getAlias());
				coder.header_add(" ((");
				coder.header_add(getReturnType());
				coder.header_add(" (*) (");
				coder.header_add(getArgString());
				coder.header_addln("))keso_throw_method_not_implemented(\"\",0))");
				return;
			}

			if (!stack_processed) {
				throw new CompileException("not ready!");
			}

			if (unused()) {
				coder.local_add("/*\n * call analysis 0 ");
				coder.local_add(getAlias());
				coder.local_add("\n */\n");
			} else {
				IMCallerEntry.emitCallerInfo(coder, this, calledFrom);
			}

			if (clearUnused()) {
				opts.verbose("# omit "+getAlias());
				coder.local_add("/* omit unused method "+getAlias()+" */\n");
			} else {
				coder.beginMethod(this);

				coder.add_class(clazz.getAlias());

				if (ret_index!=null) {
					coder.addln("jshort ret_addr;");
				}

				int curr_line = -1;
				if (!opts.getJoinPoints().affectMethod(clazz, this, coder)) {
					frame.translate(coder);
					for (IMBasicBlock bb=labelList;bb!=null;bb=(IMBasicBlock)bb.next()) {
						currentBB = bb;
						curr_line=bb.translate(coder, curr_line);
					}
					currentBB = null;
				}

				coder.endMethod();
			}

			if (!hasReturnValue()) {
				if (isPure() || isConstant()) {
					opts.vwarn("constant or pure method '"+getAlias()+"' is void!");
				}
			}
		} catch (CompileException ex) {
			System.err.println("======================================================");
			System.err.println("CompileException in "+clazz.getSourceFile());
			System.err.println(getMethodNameAndType());
			if (currentBB!=null)  System.err.println(currentBB.toLabel());
			System.err.println("======================================================");
			throw ex;
		} catch (RuntimeException ex) {
			System.err.println("======================================================");
			System.err.println("RuntimeException in "+clazz.getSourceFile());
			System.err.println(getMethodNameAndType());
			if (currentBB!=null)  System.err.println(currentBB.toLabel());
			System.err.println("======================================================");
			throw ex;
		}
	}

	final public void calledFrom(IMMethod caller, IMDataFlow[] argInfo) {
		if (calledFrom == null)
			calledFrom = new ArrayList<IMCallerEntry>();
		calledFrom.add(new IMCallerEntry(caller, argInfo));
	}

	public Collection<IMMethod> getCallers() {
		if (calledFrom == null)
			return Collections.emptyList();
		Collection<IMMethod> callers = new ArrayList<IMMethod>();
		for (IMCallerEntry entry : calledFrom) {
			if (entry.from != null)
				callers.add(entry.from);
		}
		return callers;
	}

	final public void resetCallerAnalysis() {
		calledFrom    = null;
		calledMethods = null;
	}

	final public void analyseFieldAccess(IMVisitor visitor)
	throws CompileException {
		if (hasNoCode)
			return;
		for (IMBasicBlock bb : this) {
			bb.visitNodes(visitor);
		}
	}

	/**
	 * compute the dominator tree and the dominance frontier of
	 * the method.
	 */
	final public void computeDomTree() throws CompileException {
		domTree = new DominatorTree(this, labelList);

		/* compute the dominator tree */
		domTree.computeDominators();

		/* now compute dominance frontiers */
		if (opts.hasOption("dom_use_alternative_algo")) {
			domTree.computeDominanceFrontier2();
		} else {
			domTree.computeDominanceFrontier();
		}
	}

	final public String javaLineInfo(int bcpos) {
		int line = getLineNumber(bcpos);
		StringBuffer infomsg = new StringBuffer();
		infomsg.append(getClassName());
		infomsg.append(".class in ");
		infomsg.append(getMethodName());
		if (line<0) {
			infomsg.append(" (bc ");
			infomsg.append(bcpos);
		} else {
			infomsg.append(" (line ");
			infomsg.append(line);
		}
		infomsg.append(")");
		return infomsg.toString();
	}

	public String toString() {
		return getAlias();
	}

	public int hashCode() {
		if (alias==null) alias=DecoratedNames.createMethodAlias(opts,this);
		return alias.hashCode();
	}

	public boolean equals(java.lang.Object o) {
		if (o==this) return true;
		if (o instanceof IMMethod) {
			if (alias==null) alias=DecoratedNames.createMethodAlias(opts,this);
			return alias.equals(((IMMethod)o).getAlias());
		}
		return false;
	}

	@Override
	public Iterator<IMBasicBlock> iterator() {
		return new Itr();
	}

	public <R> R accept(IMMethodVisitor<R> visitor) throws CompileException {
		return visitor.visit(this);
	}

	public void removeBasicBlock(IMBasicBlock bb) {
		IMNode prev = null;
		IMNode curr = labelList;
		while (curr != null) {
			if (curr.equals(bb)) {
				if (prev == null)
					labelList = (IMBasicBlock) curr.next();
				else
					prev.append(curr.next());
				break;
			}
			prev = curr;
			curr = curr.next();
		}
	}

	public void insertBCBefore(IMNode preList) throws CompileException {
		labelList.insertBCBefore(preList);
	}

	public boolean isEligibleForOptimization() throws CompileException {
		if (hasNoCode)
			return false;
		return !opts.getJoinPoints().ignoreMethodBody(clazz, this);
	}

	/**
	 * Adds a domain to the set of domains where this method is live.
	 * @param domain Domain to be added.
	 */
	public void addDomain(IMDomain domain) {
		domains.add(domain);
	}

	/**
	 * Returns the set of domains where this method is live.
	 * @return Domain set.
	 */
	public Set<IMDomain> getDomains() {
		return Collections.unmodifiableSet(domains);
	}

	public MethodData getMethodData() {
		return source;
	}

	/**
	 * Declares the method to be abstract by setting the hasNoCode flag to
	 * <tt>true</tt>. This is the preferred way to get rid of an unused method
	 * which cannot be dropped altogether because it is still referenced in a
	 * vtable.
	 */
	public void makeAbstract() {
		hasNoCode = true;
		labelList = null;
	}

	protected IMBasicBlock getBasicBlock(int ip, int offset) throws CompileException {
		int bcpos = ip + offset;

		IMBasicBlock label = label_index.get(bcpos);
		if (label!=null) return label;

		label = createBBlock(bcpos);
		label_index.put(bcpos,label);
		addBasicBlock(label);

		return label;
	}

	protected void parseBC(MethodData source) throws CompileException {

		label_index = new IntegerHashtable<IMBasicBlock>();
		labelList  = null;

		BytecodeInputStream bcStream = new BytecodeInputStream(source.getBytecode());

		// ==========================================
		// 1. PASS
		// ==========================================
		// in the first pass we will translate the
		// bytecode into imcode objects, and create
		// the basic blocks
		// ==========================================

		getBasicBlock(bcStream.getCurrentPosition(),0);
		endBlock = getBasicBlock(bcStream.getCurrentPosition(), bcStream.getNumBytes()+1);
		if (getBasicReturnType()==BCBasicDatatype.VOID) {
			endBlock.addBC(new IMEpilog(this, bcStream.getNumBytes()+1));
		} else {
			returnSlot = frame.createNewVariable(getBasicReturnType());
			endBlock.addBC(new IMEpilog(this, bcStream.getNumBytes()+1));
		}
		endBlock.done=true;

		ExceptionHandlerData[] handler = source.getExceptionHandlers();
		catch_list = null;
		if (handler!=null && handler.length>0) {
			catch_list = new IMExceptionHandler[handler.length];
			for (int i=0;i<handler.length;i++) {
				getBasicBlock(handler[i].getStartBCIndex(),0);
				getBasicBlock(handler[i].getEndBCIndex(),0);
				catch_list[i] = new IMExceptionHandler(this, handler[i]);
				catch_list[i].setHandler(getBasicBlock(handler[i].getHandlerBCIndex(),0));
			}
		}

		IMNode bc_list = new IMListHead();
		processBCStream(bcStream, bc_list);


		// ===============================================
		// 2. Pass
		// ===============================================
		// merge basic block list and byte code list
		// ===============================================

		IMBasicBlock bb = labelList;
		int bbpos = bb.getBCPosition();
		while (bc_list!=null && bc_list.next()!=null) {
			IMNode curr = bc_list.next();

			if (curr.getBCPosition()>=bbpos) {
				bb.addBC(curr);

				bc_list.setEndOfBasicBlock();
				bb=(IMBasicBlock)bb.next();
				bbpos = bb.getBCPosition();
				// seek end of basic block
				while (curr.next()!=null && curr.next().getBCPosition()<bbpos) { curr = curr.next(); };
			}

			bc_list = curr;
		}
		if (bc_list!=null) bc_list.setEndOfBasicBlock();
		bc_list = null;

		// ===============================================
		// 3. Pass
		// ===============================================
		// now we will process all basic blocks in a depth
		// first order of execution
		// ===============================================

		VirtualOperandStack oprStack = new VirtualOperandStack(this,source.getNumStackSlots());
		JumpStack stack = new JumpStack();

		if (dbg_process_vstack)
			opts.setCurrentClass(getClassName(),0,"vstack");

		stack.push(labelList);
		while ((bb=stack.pop())!=null) {
			if (bb.done) continue;
			currentBB = bb;
			stack.push(bb.processBasicBlock(oprStack));
		}

		if (dbg_process_vstack)
			opts.setCurrentClass("xxx",0,null);

		currentBB = null;
		stack_processed = true;

		/* Remove dead code which is not processed yet because it was not
		 * reached. This could happen if we have disabled the exception
		 * handler or if the java compiler creates brain dead code. */
		if (!opts.hasOption("no_bb_cleanup")) {
			IMBasicBlock pb=(IMBasicBlock)labelList;
			while (pb!=null&&pb.next()!=null) {
				IMBasicBlock cb = (IMBasicBlock)pb.next();
				if (!cb.done) {
					IMNode nb=cb;
					while (nb.next()!=null && !(nb.next() instanceof IMBasicBlock)) nb = nb.next();
					pb.append(nb.next());
				} else {
					pb=cb;
				}
			}
		}
	}

	final protected CompileException error() {
		return error(-1,"");
	}

	final protected CompileException error(int bcpos, String text) {
		return new CompileException(lineInfo(bcpos)+": "+text);
	}

	final protected String lineInfo(int bcpos) {
		int line = getLineNumber(bcpos);
		StringBuffer infomsg = new StringBuffer();
		if (false) {
			infomsg.append(getClassName());
			infomsg.append(".");
			infomsg.append(getMethodName());
		} else {
			infomsg.append(getAlias());
		}
		if (line<0) {
			infomsg.append(" (bc ");
			infomsg.append(bcpos);
		} else {
			infomsg.append(" (line ");
			infomsg.append(line);
		}
		infomsg.append(")");
		return infomsg.toString();
	}

	private void addBasicBlock(IMBasicBlock bb) {
		int bcpos = bb.getBCPosition();

		if (labelList==null) {
			labelList=bb;
			return;
		}

		if (labelList.getBCPosition() > bcpos) {
			bb.next_node = labelList;
			labelList = bb;
			return;
		}

		IMBasicBlock c;
		for (c=labelList;c.next_node!=null&&c.next_node.getBCPosition()<bcpos;c=(IMBasicBlock)c.next_node);

		bb.next_node = c.next_node;
		c.next_node = bb;

		return;
	}

	private IMStoreLocalVariable storeLocal(int bcpos, IMSlot vslot) throws CompileException {
		int type = vslot.getType();
		IMStoreLocalVariable nvar;
		switch (type) {
			case BCBasicDatatype.BOOLEAN:
			case BCBasicDatatype.BYTE:
			case BCBasicDatatype.CHAR:
			case BCBasicDatatype.SHORT:
				nvar=new IMIStoreLocalVariable(this,54,type,bcpos,vslot);
				break;
			case BCBasicDatatype.INT:
				nvar=new IMIStoreLocalVariable(this,54,bcpos,vslot);
				break;
			case BCBasicDatatype.LONG:
				nvar=new IMLStoreLocalVariable(this,55,bcpos,vslot);
				break;
			case BCBasicDatatype.FLOAT:
				nvar=new IMFStoreLocalVariable(this,56,bcpos,vslot);
				break;
			case BCBasicDatatype.DOUBLE:
				nvar=new IMDStoreLocalVariable(this,57,bcpos,vslot);
				break;
			case BCBasicDatatype.REFERENCE:
				nvar=new IMAStoreLocalVariable(this,58,bcpos,vslot);
				break;
			default:
				throw error(bcpos,"unknown type: "+type);
		}
		return nvar;
	}

	private IMConstant readConstantPool(int bc,int bcpos,int cindex) {
		ConstantPoolEntry cpEntry = cPool.constantEntryAt(cindex);
		if (cpEntry instanceof NumericCPEntry) {
			try {
				BCBasicDatatype cpValue  = ((NumericCPEntry)cpEntry).value();
				switch (cpValue.type()) {
					case BCBasicDatatype.INT:
						return createIMIConstant(((BCInteger)cpValue).value(),bcpos);
					case BCBasicDatatype.LONG:
						return createIMLConstant(((BCLong)cpValue).value(),bcpos);
					case BCBasicDatatype.FLOAT:
						return createIMFConstant(((BCFloat)cpValue).value(),bcpos);
					case BCBasicDatatype.DOUBLE:
						return createIMDConstant(((BCDouble)cpValue).value(),bcpos);
					default:
						throw new CompileException("unkown datatype constant");
				}
			} catch (Exception ex) {
				System.err.println("!! "+ex);
				System.exit(-1);
			}
		}
		int domainID=0;
		if (cpEntry instanceof StringCPEntry) {
			requireClass(domainID, "java/lang/String");
			requireClass(domainID, "[C");
		}
		return createIMAConstant(cpEntry,bcpos);
	}

	/**
	 * read bc stream and create list of im-objects
	 */
	private void processBCStream(BytecodeInputStream bcStream, IMNode bc_list) throws CompileException {
		IMBasicBlock label;
		boolean wide_prefix = false;
		while (bcStream.hasMore()) {
			int ip,bc,vi,offset;

			ip = bcStream.getCurrentPosition();
			bc = bcStream.readUnsignedByte();

			switch (bc) {
				case 1: // (0x01) aconst_null
					bc_list=bc_list.append(new IMNullConstant(this,bc,ip));
					break;
				case 2: // (0x02) iconst_m1
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,-1));
					break;
				case 3: // (0x03) iconst_0
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,0));
					break;
				case 4: // (0x04) iconst_1
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,1));
					break;
				case 5: // (0x05) iconst_2
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,2));
					break;
				case 6: // (0x06) iconst_3
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,3));
					break;
				case 7: // (0x07) iconst_4
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,4));
					break;
				case 8: // (0x08) iconst_5
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,5));
					break;
				case 9: // (0x09) lconst_0
					bc_list=bc_list.append(new IMLConstant(this,bc,ip,0));
					break;
				case 10: // (0x0a) lconst_1
					bc_list=bc_list.append(new IMLConstant(this,bc,ip,1));
					break;
				case 11: // (0x0b) fconst_0
					bc_list=bc_list.append(new IMFConstant(this,bc,ip,0));
					break;
				case 12: // (0x0c) fconst_1
					bc_list=bc_list.append(new IMFConstant(this,bc,ip,1));
					break;
				case 13: // (0x0d) fconst_2
					bc_list=bc_list.append(new IMFConstant(this,bc,ip,2));
					break;
				case 14: // (0x0e) dconst_0
					bc_list=bc_list.append(new IMDConstant(this,bc,ip,0));
					break;
				case 15: // (0x0f) dconst_1
					bc_list=bc_list.append(new IMDConstant(this,bc,ip,1));
					break;
				case 16: // (0x10) bipush
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,bcStream.readByte()));
					break;
				case 17: // (0x11) sipush
					bc_list=bc_list.append(new IMIConstant(this,bc,ip,bcStream.readShort()));
					break;
				case 18: // (0x12) ldc
					bc_list=bc_list.append(readConstantPool(bc,ip,bcStream.readUnsignedByte()));
					break;
				case 19: // (0x13) ldc_w  // push 32 bit value
				case 20: // (0x14) ldc2_w // push 64 bit value long or double
					bc_list=bc_list.append(readConstantPool(bc,ip,bcStream.readUnsignedShort()));
					break;
				case 21: // (0x15) iload
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMIReadLocalVariable(this,bc,ip,vi));
					break;
				case 22: // (0x16) lload
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMLReadLocalVariable(this,bc,ip,vi));
					break;
				case 23: // (0x17) fload
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMFReadLocalVariable(this,bc,ip,vi));
					break;
				case 24: // (0x18) dload
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMDReadLocalVariable(this,bc,ip,vi));
					break;
				case 25: // (0x19) aload
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMAReadLocalVariable(this,bc,ip,vi));
					break;
				case 26: // (0x1a) iload_0
					//testLocalVar(0);
					bc_list=bc_list.append(new IMIReadLocalVariable(this,bc,ip,0));
					break;
				case 27: // (0x1b) iload_1
					bc_list=bc_list.append(new IMIReadLocalVariable(this,bc,ip,1));
					break;
				case 28: // (0x1c) iload_2
					bc_list=bc_list.append(new IMIReadLocalVariable(this,bc,ip,2));
					break;
				case 29: // (0x1d) iload_3
					bc_list=bc_list.append(new IMIReadLocalVariable(this,bc,ip,3));
					break;
				case 30: // (0x1e) lload_0
					bc_list=bc_list.append(new IMLReadLocalVariable(this,bc,ip,0));
					break;
				case 31: // (0x1f) lload_1
					bc_list=bc_list.append(new IMLReadLocalVariable(this,bc,ip,1));
					break;
				case 32: // (0x20) lload_2
					bc_list=bc_list.append(new IMLReadLocalVariable(this,bc,ip,2));
					break;
				case 33: // (0x21) lload_3
					bc_list=bc_list.append(new IMLReadLocalVariable(this,bc,ip,3));
					break;
				case 34: // (0x22) fload_0
					bc_list=bc_list.append(new IMFReadLocalVariable(this,bc,ip,0));
					break;
				case 35: // (0x23) fload_1
					bc_list=bc_list.append(new IMFReadLocalVariable(this,bc,ip,1));
					break;
				case 36: // (0x24) fload_2
					bc_list=bc_list.append(new IMFReadLocalVariable(this,bc,ip,2));
					break;
				case 37: // (0x25) fload_3
					bc_list=bc_list.append(new IMFReadLocalVariable(this,bc,ip,3));
					break;
				case 38: // (0x26) dload_0
					bc_list=bc_list.append(new IMDReadLocalVariable(this,bc,ip,0));
					break;
				case 39: // (0x27) dload_1
					bc_list=bc_list.append(new IMDReadLocalVariable(this,bc,ip,1));
					break;
				case 40: // (0x28) dload_2
					bc_list=bc_list.append(new IMDReadLocalVariable(this,bc,ip,2));
					break;
				case 41: // (0x29) dload_3
					bc_list=bc_list.append(new IMDReadLocalVariable(this,bc,ip,3));
					break;
				case 42: // (0x2a) aload_0
					bc_list=bc_list.append(new IMAReadLocalVariable(this,bc,ip,0));
					break;
				case 43: // (0x2b) aload_1
					bc_list=bc_list.append(new IMAReadLocalVariable(this,bc,ip,1));
					break;
				case 44: // (0x2c) aload_2
					bc_list=bc_list.append(new IMAReadLocalVariable(this,bc,ip,2));
					break;
				case 45: // (0x2d) aload_3
					bc_list=bc_list.append(new IMAReadLocalVariable(this,bc,ip,3));
					break;
				case 46: // (0x2e) iaload
					bc_list=bc_list.append(new IMIReadArray(this,bc,ip));
					break;
				case 47: // (0x2f) laload
					bc_list=bc_list.append(new IMLReadArray(this,bc,ip));
					break;
				case 48: // (0x30) faload
					bc_list=bc_list.append(new IMFReadArray(this,bc,ip));
					break;
				case 49: // (0x31) daload
					bc_list=bc_list.append(new IMDReadArray(this,bc,ip));
					break;
				case 50: // (0x32) aaload
					bc_list=bc_list.append(new IMAReadArray(this,bc,ip));
					break;
				case 51: // (0x33) baload
					bc_list=bc_list.append(new IMBReadArray(this,bc,ip));
					break;
				case 52: // (0x34) caload
					bc_list=bc_list.append(new IMCReadArray(this,bc,ip));
					break;
				case 53: // (0x35) saload
					bc_list=bc_list.append(new IMSReadArray(this,bc,ip));
					break;
				case 54: // (0x36) istore
					if (wide_prefix)  vi = bcStream.readUnsignedShort();
					else vi = bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,bc,ip,vi));
					break;
				case 55: // (0x37) lstore
					if (wide_prefix)  vi = bcStream.readUnsignedShort();
					else vi = bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMLStoreLocalVariable(this,bc,ip,vi));
					break;
				case 56: // (0x38) fstore
					if (wide_prefix)  vi = bcStream.readUnsignedShort();
					else vi = bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMFStoreLocalVariable(this,bc,ip,vi));
					break;
				case 57: // (0x39) dstore
					if (wide_prefix)  vi = bcStream.readUnsignedShort();
					else vi = bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMDStoreLocalVariable(this,bc,ip,vi));
					break;
				case 58: // (0x3a) astore
					if (wide_prefix)  vi = bcStream.readUnsignedShort();
					else vi = bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMAStoreLocalVariable(this,bc,ip,vi));
					break;
				case 59: // (0x3b) istore_0
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,bc,ip,0));
					break;
				case 60: // (0x3c) istore_1
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,bc,ip,1));
					break;
				case 61: // (0x3d) istore_2
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,bc,ip,2));
					break;
				case 62: // (0x3e) istore_3
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,bc,ip,3));
					break;
				case 63: // (0x3f) lstore_0
					bc_list=bc_list.append(new IMLStoreLocalVariable(this,bc,ip,0));
					break;
				case 64: // (0x40) lstore_1
					bc_list=bc_list.append(new IMLStoreLocalVariable(this,bc,ip,1));
					break;
				case 65: // (0x41) lstore_2
					bc_list=bc_list.append(new IMLStoreLocalVariable(this,bc,ip,2));
					break;
				case 66: // (0x42) lstore_3
					bc_list=bc_list.append(new IMLStoreLocalVariable(this,bc,ip,3));
					break;
				case 67: // (0x43) fstore_0
					bc_list=bc_list.append(new IMFStoreLocalVariable(this,bc,ip,0));
					break;
				case 68: // (0x44) fstore_1
					bc_list=bc_list.append(new IMFStoreLocalVariable(this,bc,ip,1));
					break;
				case 69: // (0x45) fstore_2
					bc_list=bc_list.append(new IMFStoreLocalVariable(this,bc,ip,2));
					break;
				case 70: // (0x46) fstore_3
					bc_list=bc_list.append(new IMFStoreLocalVariable(this,bc,ip,3));
					break;
				case 71: // (0x47) dstore_0
					bc_list=bc_list.append(new IMDStoreLocalVariable(this,bc,ip,0));
					break;
				case 72: // (0x48) dstore_1
					bc_list=bc_list.append(new IMDStoreLocalVariable(this,bc,ip,1));
					break;
				case 73: // (0x49) dstore_2
					bc_list=bc_list.append(new IMDStoreLocalVariable(this,bc,ip,2));
					break;
				case 74: // (0x4a) dstore_3
					bc_list=bc_list.append(new IMDStoreLocalVariable(this,bc,ip,3));
					break;
				case 75: // (0x4b) astore_0
					bc_list=bc_list.append(new IMAStoreLocalVariable(this,bc,ip,0));
					break;
				case 76: // (0x4c) astore_1
					bc_list=bc_list.append(new IMAStoreLocalVariable(this,bc,ip,1));
					break;
				case 77: // (0x4d) astore_2
					bc_list=bc_list.append(new IMAStoreLocalVariable(this,bc,ip,2));
					break;
				case 78: // (0x4e) astore_3
					bc_list=bc_list.append(new IMAStoreLocalVariable(this,bc,ip,3));
					break;
				case 79: // (0x4f) iastore
					bc_list=bc_list.append(new IMIStoreArray(this,bc,ip));
					break;
				case 80: // (0x50) lastore
					bc_list=bc_list.append(new IMLStoreArray(this,bc,ip));
					break;
				case 81: // (0x51) fastore
					bc_list=bc_list.append(new IMFStoreArray(this,bc,ip));
					break;
				case 82: // (0x52) dastore
					bc_list=bc_list.append(new IMDStoreArray(this,bc,ip));
					break;
				case 83: // (0x53) aastore
					bc_list=bc_list.append(new IMAStoreArray(this,bc,ip));
					break;
				case 84: // (0x54) bastore
					bc_list=bc_list.append(new IMBStoreArray(this,bc,ip));
					break;
				case 85: // (0x55) castore
					bc_list=bc_list.append(new IMCStoreArray(this,bc,ip));
					break;
				case 86: // (0x56) sastore
					bc_list=bc_list.append(new IMSStoreArray(this,bc,ip));
					break;
				case 87: // (0x57) pop
					bc_list=bc_list.append(new IMStackOperationPOP(this,bc,ip));
					break;
				case 88: // (0x58) pop2
					bc_list=bc_list.append(new IMStackOperationPOP2(this,bc,ip));
					break;
				case 89: // (0x59) dup
					bc_list=bc_list.append(new IMStackOperationDUP(this,bc,ip));
					break;
				case 90: // (0x5a) dup_x1
					bc_list=bc_list.append(new IMStackOperationDUP_X1(this,bc,ip));
					break;
				case 91: // (0x5b) dup_x2
					bc_list=bc_list.append(new IMStackOperationDUP_X2(this,bc,ip));
					break;
				case 92: // (0x5c) dup2
					bc_list=bc_list.append(new IMStackOperationDUP2(this,bc,ip));
					break;
				case 93: // (0x5d) dup2_x1
					bc_list=bc_list.append(new IMStackOperationDUP2_X1(this,bc,ip));
					break;
				case 94: // (0x5e) dup2_x2
					bc_list=bc_list.append(new IMStackOperationDUP2_X2(this,bc,ip));
					break;
				case 95: // (0x5f) swap
					bc_list=bc_list.append(new IMStackOperationSWAP(this,bc,ip));
					break;
				case 96: // (0x60) iadd
					bc_list=bc_list.append(new IMIAdd(this,bc,ip));
					break;
				case 97: // (0x61) ladd
					bc_list=bc_list.append(new IMLAdd(this,bc,ip));
					break;
				case 98: // (0x62) fadd
					bc_list=bc_list.append(new IMFAdd(this,bc,ip));
					break;
				case 99: // (0x63) dadd
					bc_list=bc_list.append(new IMDAdd(this,bc,ip));
					break;
				case 100: // (0x64) isub
					bc_list=bc_list.append(new IMISub(this,bc,ip));
					break;
				case 101: // (0x65) lsub
					bc_list=bc_list.append(new IMLSub(this,bc,ip));
					break;
				case 102: // (0x66) fsub
					bc_list=bc_list.append(new IMFSub(this,bc,ip));
					break;
				case 103: // (0x67) dsub
					bc_list=bc_list.append(new IMDSub(this,bc,ip));
					break;
				case 104: // (0x68) imul
					bc_list=bc_list.append(new IMIMul(this,bc,ip));
					break;
				case 105: // (0x69) lmul
					bc_list=bc_list.append(new IMLMul(this,bc,ip));
					break;
				case 106: // (0x6a) fmul
					bc_list=bc_list.append(new IMFMul(this,bc,ip));
					break;
				case 107: // (0x6b) dmul
					bc_list=bc_list.append(new IMDMul(this,bc,ip));
					break;
				case 108: // (0x6c) idiv
					bc_list=bc_list.append(new IMIDiv(this,bc,ip));
					break;
				case 109: // (0x6d) ldiv
					bc_list=bc_list.append(new IMLDiv(this,bc,ip));
					break;
				case 110: // (0x6e) fdiv
					bc_list=bc_list.append(new IMFDiv(this,bc,ip));
					break;
				case 111: // (0x6f) ddiv
					bc_list=bc_list.append(new IMDDiv(this,bc,ip));
					break;
				case 112: // (0x70) irem
					bc_list=bc_list.append(new IMIRem(this,bc,ip));
					break;
				case 113: // (0x71) lrem
					bc_list=bc_list.append(new IMLRem(this,bc,ip));
					break;
				case 114: // (0x72) frem
					bc_list=bc_list.append(new IMFRem(this,bc,ip));
					break;
				case 115: // (0x73) drem
					bc_list=bc_list.append(new IMDRem(this,bc,ip));
					break;
				case 116: // (0x74) ineg
					bc_list=bc_list.append(new IMINeg(this,bc,ip));
					break;
				case 117: // (0x75) lneg
					bc_list=bc_list.append(new IMLNeg(this,bc,ip));
					break;
				case 118: // (0x76) fneg
					bc_list=bc_list.append(new IMFNeg(this,bc,ip));
					break;
				case 119: // (0x77) dneg
					bc_list=bc_list.append(new IMDNeg(this,bc,ip));
					break;
				case 120: // (0x78) ishl
					bc_list=bc_list.append(new IMIShiftLeft(this,bc,ip));
					break;
				case 121: // (0x79) lshl
					bc_list=bc_list.append(new IMLShiftLeft(this,bc,ip));
					break;
				case 122: // (0x7a) ishr
					bc_list=bc_list.append(new IMIShiftRight(this,bc,ip));
					break;
				case 123: // (0x7b) lshr
					bc_list=bc_list.append(new IMLShiftRight(this,bc,ip));
					break;
				case 124: // (0x7c) iushr
					bc_list=bc_list.append(new IMIShiftRight2(this,bc,ip));
					break;
				case 125: // (0x7d) lushr
					bc_list=bc_list.append(new IMLShiftRight2(this,bc,ip));
					break;
				case 126: // (0x7e) iand
					bc_list=bc_list.append(new IMIBitAnd(this,bc,ip));
					break;
				case 127: // (0x7f) land
					bc_list=bc_list.append(new IMLBitAnd(this,bc,ip));
					break;
				case 128: // (0x80) ior
					bc_list=bc_list.append(new IMIBitOr(this,bc,ip));
					break;
				case 129: // (0x81) lor
					bc_list=bc_list.append(new IMLBitOr(this,bc,ip));
					break;
				case 130: // (0x82) ixor
					bc_list=bc_list.append(new IMIBitXor(this,bc,ip));
					break;
				case 131: // (0x83) lxor
					bc_list=bc_list.append(new IMLBitXor(this,bc,ip));
					break;
				case 132: // (0x84) iinc
					int value;
					if (wide_prefix) {
						vi = bcStream.readUnsignedShort();
						value = bcStream.readShort();
					} else {
						vi = bcStream.readUnsignedByte();
						value = bcStream.readByte();
					}
					bc_list=bc_list.append(new IMIReadLocalVariable(this,21,ip,vi));
					bc_list=bc_list.append(createIMIConstant(value,ip));
					bc_list=bc_list.append(new IMIAdd(this,bc,ip));
					bc_list=bc_list.append(new IMIStoreLocalVariable(this,54,ip,vi));
					break;
				case 133: // (0x85) i2l
					bc_list=bc_list.append(new IMCastI2L(this,bc,ip));
					break;
				case 134: // (0x86) i2f
					bc_list=bc_list.append(new IMCastI2F(this,bc,ip));
					break;
				case 135: // (0x87) i2d
					bc_list=bc_list.append(new IMCastI2D(this,bc,ip));
					break;
				case 136: // (0x88) l2i
					bc_list=bc_list.append(new IMCastL2I(this,bc,ip));
					break;
				case 137: // (0x89) l2f
					bc_list=bc_list.append(new IMCastL2F(this,bc,ip));
					break;
				case 138: // (0x8a) l2d
					bc_list=bc_list.append(new IMCastL2D(this,bc,ip));
					break;
				case 139: // (0x8b) f2i
					bc_list=bc_list.append(new IMCastF2I(this,bc,ip));
					break;
				case 140: // (0x8c) f2l
					bc_list=bc_list.append(new IMCastF2L(this,bc,ip));
					break;
				case 141: // (0x8d) f2d
					bc_list=bc_list.append(new IMCastF2D(this,bc,ip));
					break;
				case 142: // (0x8e) d2i
					bc_list=bc_list.append(new IMCastD2I(this,bc,ip));
					break;
				case 143: // (0x8f) d2l
					bc_list=bc_list.append(new IMCastD2L(this,bc,ip));
					break;
				case 144: // (0x90) d2f
					bc_list=bc_list.append(new IMCastD2F(this,bc,ip));
					break;
				case 145: // (0x91) i2b
					bc_list=bc_list.append(new IMCastI2B(this,bc,ip));
					break;
				case 146: // (0x92) i2c
					bc_list=bc_list.append(new IMCastI2C(this,bc,ip));
					break;
				case 147: // (0x93) i2s
					bc_list=bc_list.append(new IMCastI2S(this,bc,ip));
					break;
				case 148: // (0x94) lcmp
					bc_list=bc_list.append(new IMLCompare(this,bc,ip));
					break;
					/*
					   Notes The fcmpg and fcmpl instructions differ only in their treatment of a comparison
					   involving NaN. NaN is unordered, so any float comparison fails if either or both
					   of its operands are NaN. With both fcmpg and fcmpl available, any float comparison
					   may be compiled to push the same result onto the operand stack whether the comparison
					   fails on non-NaN values or fails because it encountered a NaN. For more information,
					   see Section 7.5, "More Control Examples."
					 */
				case 149: // (0x95) fcmpl
					bc_list=bc_list.append(new IMFLCompare(this,bc,ip));
					break;
				case 150: // (0x96) fcmpg
					bc_list=bc_list.append(new IMFGCompare(this,bc,ip));
					break;
				case 151: // (0x97) dcmpl
					bc_list=bc_list.append(new IMDLCompare(this,bc,ip));
					break;
				case 152: // (0x98) dcmpg
					bc_list=bc_list.append(new IMDGCompare(this,bc,ip));
					break;
				case 153: // (0x99) ifeq
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMEQConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 154: // (0x9a) ifne
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMNEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 155: // (0x9b) iflt
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMLTConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 156: // (0x9c) ifge
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMGEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 157: // (0x9d) ifgt
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMGTConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 158: // (0x9e) ifle
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMLEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 159: // (0x9f) if_icmpeq
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMEQConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 160: // (0xa0) if_icmpne
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMNEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 161: // (0xa1) if_icmplt
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMLTConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 162: // (0xa2) if_icmpge
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMGEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 163: // (0xa3) if_icmpgt
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMGTConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 164: // (0xa4) if_icmple
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMLEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 165: // (0xa5) if_acmpeq
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMEQConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 166: // (0xa6) if_acmpne
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMNEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 167: // (0xa7) goto
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMGoto(this,bc,ip,label));
					break;
				case 168: // (0xa8) jsr
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMCall(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 169: // (0xa9) ret
					if (wide_prefix) vi=bcStream.readUnsignedShort();
					else vi=bcStream.readUnsignedByte();
					bc_list=bc_list.append(new IMReturnSubroutine(this,bc,ip,vi));
					break;
				case 170: // (0xaa) tableswitch
					// skip padding zeroes
					while (bcStream.getCurrentPosition() % 4 != 0) bcStream.readByte();
					offset = bcStream.readInt();
					int low    = bcStream.readInt();
					int high   = bcStream.readInt();
					int tsize = high-low;
					IMBasicBlock jmptable[] = new IMBasicBlock[tsize+2];
					jmptable[0] = getBasicBlock(ip,offset);
					for (int i=1;i<tsize+2;i++) jmptable[i]=getBasicBlock(ip,bcStream.readInt());
					bc_list=bc_list.append(new IMTableSwitch(this,bc,ip,low,high,jmptable));
					break;
				case 171: // (0xab) lookupswitch
					// skip padding zeroes
					while (bcStream.getCurrentPosition() % 4 != 0) bcStream.readByte();
					offset = bcStream.readInt();
					int len    = bcStream.readInt();
					int values[] = new int[len];
					IMBasicBlock addrs[] = new IMBasicBlock[len+1];
					addrs[0] = getBasicBlock(ip,offset);
					for (int i=0;i<len;i++) {
						values[i] = bcStream.readInt();
						addrs[i+1] = getBasicBlock(ip,bcStream.readInt());
					}
					bc_list=bc_list.append(new IMLookupSwitch(this,bc,ip,values,addrs));
					break;
				case 172: // (0xac) ireturn
					bc_list=bc_list.append(storeLocal(ip, getReturnValue()));
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 173: // (0xad) lreturn
					bc_list=bc_list.append(storeLocal(ip,getReturnValue()));
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 174: // (0xae) freturn
					bc_list=bc_list.append(storeLocal(ip,getReturnValue()));
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 175: // (0xaf) dreturn
					bc_list=bc_list.append(storeLocal(ip,getReturnValue()));
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 176: // (0xb0) areturn
					bc_list=bc_list.append(storeLocal(ip,getReturnValue()));
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 177: // (0xb1) return
					bc_list=bc_list.append(new IMGoto(this,167,ip,getEndBlock()));
					break;
				case 178: // (0xb2) getstatic
					{
						FieldRefCPEntry cpEntry=cPool.fieldRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMGetStatic(this,bc,ip,cpEntry));
						break;
					}
				case 179: // (0xb3) putstatic
					{
						FieldRefCPEntry cpEntry=cPool.fieldRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMPutStatic(this,bc,ip,cpEntry));
						break;
					}
				case 180: // (0xb4) getfield
					{
						FieldRefCPEntry cpEntry=cPool.fieldRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMGetField(this,bc,ip,cpEntry));
						break;
					}
				case 181: // (0xb5) putfield
					{
						FieldRefCPEntry cpEntry=cPool.fieldRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMPutField(this,bc,ip,cpEntry));
						break;
					}
				case 182: // (0xb6) invokevirtual
					{
						MethodRefCPEntry cpEntry=cPool.methodRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMInvokeVirtual(this,bc,ip,cpEntry));
						++IMInvoke.virtualCallsInitial;
						break;
					}
				case 183: // (0xb7) invokespecial
					{
						MethodRefCPEntry cpEntry=cPool.methodRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMInvokeSpecial(this,bc,ip,cpEntry));
						++IMInvoke.nonVirtualCallsInitial;
						break;
					}
				case 184: // (0xb8) invokestatic
					{
						MethodRefCPEntry cpEntry
							= cPool.methodRefEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMInvokeStatic(this,bc,ip,cpEntry));
						++IMInvoke.nonVirtualCallsInitial;
						break;
					}
				case 185: // (0xb9) invokeinterface
					{
						InterfaceMethodRefCPEntry cpEntry
							= cPool.InterfaceMethodRefEntryAt(bcStream.readUnsignedShort());
						bcStream.readByte(); // read useless arg length
						bcStream.readByte(); // read useless padding zero
						bc_list=bc_list.append(new IMInvokeInterface(this,bc,ip,cpEntry));
						++IMInvoke.virtualCallsInitial;
						break;
					}
				case 186: // (0xba) xxxunusedxxx
					break;
				case 187: // (0xbb) new
					{
						ClassCPEntry cpEntry
							= cPool.classEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMNew(this,bc,ip,cpEntry));
						break;
					}
				case 188: // (0xbc) newarray
					{
						int type = bcStream.readByte();
						bc_list=bc_list.append(new IMNewArray(this,bc,ip,aType[type]));
						break;
					}
				case 189: // (0xbd) anewarray
					{
						ClassCPEntry cpEntry
							= cPool.classEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMNewObjArray(this,bc,ip,cpEntry));
						break;
					}
				case 190: // (0xbe) arraylength
					bc_list=bc_list.append(new IMArrayLength(this,bc,ip));
					break;
				case 191: // (0xbf) athrow
					bc_list=bc_list.append(new IMThrow(this,bc,ip));
					break;
				case 192: // (0xc0) checkcast
					{
						ClassCPEntry cpEntry
							= cPool.classEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMCheckCast(this,bc,ip,cpEntry));
						break;
					}
				case 193: // (0xc1) instanceof
					{
						ClassCPEntry cpEntry
							= cPool.classEntryAt(bcStream.readUnsignedShort());
						bc_list=bc_list.append(new IMInstanceOf(this,bc,ip,cpEntry));
						break;
					}
				case 194: // (0xc2) monitorenter
					bc_list=bc_list.append(new IMMonitor(this,bc,ip));
					break;
				case 195: // (0xc3) monitorexit
					bc_list=bc_list.append(new IMMonitor(this,bc,ip));
					break;
				case 196: // (0xc4) wide
					wide_prefix = true;
					continue;
				case 197: // (0xc5) multianewarray
					{
						ClassCPEntry cpEntry
							= cPool.classEntryAt(bcStream.readUnsignedShort());
						int dim = bcStream.readUnsignedByte();
						bc_list=bc_list.append(new IMNewMultiArray(this,bc,ip,cpEntry,dim));
					}
					break;
				case 198: // (0xc6) ifnull
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMEQConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 199: // (0xc7) ifnonnull
					label=getBasicBlock(ip,bcStream.readShort());
					bc_list=bc_list.append(new IMNEConditionalBranch(this,bc,ip,label,getBasicBlock(ip,3)));
					break;
				case 200: // (0xc8) goto_w
					label=getBasicBlock(ip,bcStream.readInt());
					bc_list=bc_list.append(new IMGoto(this,bc,ip,label));
					break;
				case 201: // (0xc9) jsr_w
					label=getBasicBlock(ip,bcStream.readInt());
					bc_list=bc_list.append(new IMCall(this,bc,ip,label,getBasicBlock(ip,5)));
					break;
				case 202: // (0xca) breakpoint
					/* _quick opcodes: */
				case 203: // (0xcb) ldc_quick
				case 204: // (0xcc) ldc_w_quick
				case 205: // (0xcd) ldc2_w_quick
				case 206: // (0xce) getfield_quick
				case 207: // (0xcf) putfield_quick
				case 208: // (0xd0) getfield2_quick
				case 209: // (0xd1) putfield2_quick
				case 210: // (0xd2) getstatic_quick
				case 211: // (0xd3) putstatic_quick
				case 212: // (0xd4) getstatic2_quick
				case 213: // (0xd5) putstatic2_quick
				case 214: // (0xd6) invokevirtual_quick
				case 215: // (0xd7) invokenonvirtual_quick
				case 216: // (0xd8) invokesuper_quick
				case 217: // (0xd9) invokestatic_quick
				case 218: // (0xda) invokeinterface_quick
				case 219: // (0xdb) invokevirtualobject_quick
				case 221: // (0xdd) new_quick
				case 222: // (0xde) anewarray_quick
				case 223: // (0xdf) multianewarray_quick
				case 224: // (0xe0) checkcast_quick
				case 225: // (0xe1) instanceof_quick
				case 226: // (0xe2) invokevirtual_quick_w
				case 227: // (0xe3) getfield_quick_w
				case 228: // (0xe4) putfield_quick_w
					/* Reserved opcodes */
				case 254: // (0xfe) impdep1
				case 255: // (0xff) impdep2
					throw new CompileException ("not implemeted byte code! ("+bc+")");
			}

			wide_prefix = false;
		}
	}

	private IntegerHashtable<IMBasicBlock> ret_index = null;
	final public int registerReturnAddr(IMBasicBlock retbb) {
		if (ret_index==null) ret_index = new IntegerHashtable<IMBasicBlock>();
		int index = retbb.getBCPosition();
		ret_index.put(index, retbb);
		return index;
	}

	private void doEscapeAnalyses() throws CompileException {
		/* now we do the escape analyses */
		IMVisitor v1 = new IMEscapeVisitor(opts,IMEscapeVisitor.PREPARE);
		for (IMBasicBlock bb=labelList;bb!=null;bb=(IMBasicBlock)bb.next()) {
			currentBB = bb;
			bb.visitNodes(v1);
		}
		IMVisitor v2 = new IMEscapeVisitor(opts,IMEscapeVisitor.ANALYSE);
		for (IMBasicBlock bb=labelList;bb!=null;bb=(IMBasicBlock)bb.next()) {
			currentBB = bb;
			bb.visitNodes(v2);
		}
		currentBB = null;

		/*
		 * Analyse if method arguments escape and save the
		 * result in arg_escape[...]
		 */
		IMSlot args[] = frame.getOriginalArguments();
		initEscapeInfo();
		if (!opts.hasOption("_X_ssa_less_astack")) {
			for (int i=0;i<args.length;i++) {
				opts.vverbose("## test escape "+getAlias()+" "+args[i]);
				if (args[i].getType()==BCBasicDatatype.REFERENCE) {
					arg_escape[i] = args[i].doesEscape(true);
				}
			}
		}
	}

	/**
	 * calculate the method costs and decide if the method is a
	 * candidate for inlining.
	 */
	private boolean hasLongOrDoubleArgument() throws CompileException {
		IMSlot[] argSlots = frame.getMethodArguments();
		for (int i=0;i<argSlots.length; ++i) {
			int t = argSlots[i].getType();
			if (t==BCBasicDatatype.LONG) return true;
			if (t==BCBasicDatatype.DOUBLE) return true;
		}
		return false;
	}

	final private void analyseCalledMethod(IMMethod m) throws CompileException {
		if (!m.isPure()) {
			isPure=false;
			isConst=false;
		} else if (!m.isConstant()) {
			isConst=false;
		}
	}

	/**
	 * Determines the "Blocking" state of all methods in the class store.
	 *
	 * A method is a blocking method if it calls a blocking system service
	 * (WaitEvent) or another blocking method.
	 *
	 * To cope with cycles the algorithm passes all methods until no more
	 * changes happen in a pass. All methods are initially marked as being
	 * undecided. During a pass, a method is marked as
	 *   blocking if
	 *     - there is a KNI aspect that marks the method as blocking (syscalls)
	 *     - one of the method's callees has already been marked as blocking
	 *   not blocking if
	 *     - the method does not have any callees and no matching KNI aspect
	 *     - the method does not have any undecided callees and non of the
	 *       the callees is a blocking method
	 */
	private void determineBlockingMethods() throws CompileException {
		// collect all methods in a Set methods
		IMClass[] classes = getClassStore().getAllClasses();
		HashSet<IMMethod> methods = new HashSet<IMMethod>();
		for(IMClass c: classes) {
			if(c.getMethods() == null) continue;
			for(IMMethod m: c.getMethods()) {
				methods.add(m);
			}
		}

		// loop until no more changes occur
		for(boolean haveChanges=true; haveChanges; ) {
			haveChanges = false;
			for(IMMethod m: methods) {
				boolean hasUndecided = false;

			 	// skip already decided
				if (m.canBlockIn != blocking_undecided)
					continue;

				MethodTable mTable = m.getMethodTable();

				// mark directly blocking methods (decided by KNI) as blocking
				if (Weavelet.TRUE == m.getJoinPoints().checkAttribute(m, Weavelet.ATTR_BLOCK)) {
					m.decideBlocking( blocks_in_self );
					haveChanges=true;

				// abstract or interface method => check for blocking candidates
				} else if (!m.hasCode() && mTable != null) {
					for (IMMethod candidate : mTable.getCandidates()) {
						if(candidate.canBlockIn == blocking_undecided)
							hasUndecided = true;
						if (candidate.decidedBlocking()) {
							m.decideBlocking( candidate.getAlias() );
							haveChanges=true;
							break;
						}
					}

				// no calls => cannot block
				} else if (m.calledMethods==null) {
					m.canBlockIn = none_blocking;
					haveChanges=true;

				// check callees for blocking methods
				} else for (MethodEntry e : m.calledMethods) {

					// if one of our callees can block, we can, too
					IMMethod callee = findMethod(e.className, e.nameAndType);
					if (callee==null) throw new CompileException("Can't find method "+e.className+" "+e.nameAndType);

					if (callee.canBlockIn == blocking_undecided)
						hasUndecided = true;

					// determine pure and const attributes
					m.analyseCalledMethod(callee);

					// in the meantime the state of this method has been determined, but
					// we continue the loop to call analyseCalledMethod() on all callees
					if(m.canBlockIn != blocking_undecided)
						continue;

					MethodTable calleeTable = callee.getMethodTable();

					if (callee.decidedBlocking()) {
						m.decideBlocking( callee.getAlias() );
						haveChanges=true;

					// if there are multiple candidates check if any of them is blocking
					} else if(calleeTable != null) {
						for (IMMethod candidate : calleeTable.getCandidates()) {
							if(candidate.canBlockIn == blocking_undecided)
								hasUndecided = true;
							if (candidate.decidedBlocking()) {
								m.decideBlocking( candidate.getAlias() );
								haveChanges=true;
								break;
							}
						}
					}
				}

				// if none of the callees is a blocking method this method is also not blocking
				if(hasUndecided==false && m.canBlockIn == blocking_undecided) {
					m.decideBlocking( none_blocking );
					haveChanges=true;
				}
			}
		}

		// due to cycles in the call graph there may still be methods where the
		// blocking status has not yet been determined; this is only possible if
		// none of the methods participating in the cycle is a blocking method,
		// since otherwise all these methods would already have been marked as
		// blocking. We thus mark all the still undecided methods as non blocking
		// here.
		for(IMMethod m: methods) {
			if(m.canBlockIn == blocking_undecided) {
				m.canBlockIn = none_blocking;
			}
		}
		analyseCalleesDone = true;
	}

	private boolean decidedBlocking() {
		return (canBlockIn != none_blocking
				&& canBlockIn != blocking_undecided);
	}

	private void decideBlocking(String state) throws CompileException {
		if(canBlockIn != blocking_undecided)
			throw new CompileException("Attempt to reset Blocking state for "
					+ getClassName() + "." + getMethodName() + " (" + canBlockIn
					+ " => " + state + ")");
		canBlockIn = state;
	}

	/**
	 * Set of domains where this method is live.
	 *
	 * This is the union of domains in that the method's basic blocks are live.
	 */
	private final Set<IMDomain> domains = new HashSet<IMDomain>();

	@Override
	public int compareTo(IMMethod o) {
		return getAlias().compareTo(o.getAlias());
	}
}
