/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import java.util.*;

import keso.classfile.datatypes.*;

import keso.compiler.*;
import keso.compiler.backend.*;

import keso.util.*;

final public class IMMethodFrame {

	private IMMethod method;
	private BuilderOptions opts;

	private int new_count;
	private BasicTypeDescriptor[] arg_types;

	private final IntegerHashtable<IMSlot> stack_index = new IntegerHashtable<IMSlot>();

	/**
	 * Maps local variable indices to the corresponding IMSlot objects.
	 *
	 * Local variables are enumerated in 32-bit slots. A single slot can store
	 * any boolean, byte, char, short, int, float, reference, or returnAddress.
	 * Two consecutive slots (and therefore indices) are occupied by a variable
	 * of type long or double.
	 */
	private IntegerHashtable<IMSlot> all = new IntegerHashtable<IMSlot>(256);

	private SplitKeyHashMap<Integer, Integer, IMSlot> foreignSlots;
	private IntegerHashtable<Integer> referenceSlots;
	private int referenceCount=0;

	private IMCallerEntry callerInfo = null;

	private final static BasicTypeDescriptor[] NOARGS = new BasicTypeDescriptor[] {};

	public IMMethodFrame(IMMethod method, int start_counter) {
		this(method,NOARGS,start_counter);
	}

	void setNewCount(int new_count) { this.new_count = new_count; }

	public IMMethodFrame(IMMethod method, BasicTypeDescriptor[] arg_types, int start_counter) {
		this.method=method;
		this.opts=method.getOptions();
		this.arg_types=arg_types;
		this.new_count=start_counter;
		int locals=0;

		for (int i=0;i<arg_types.length;i++) {
			int t = arg_types[i].toBasicDatatype();
			getLocalVariable(t,locals);
			locals+=BCBasicDatatype.sizeInWords(t);
		}
	}

	public int argRange() {
		int locals=0;
		for (int i=0;i<arg_types.length;i++) {
			locals+=BCBasicDatatype.sizeInWords(arg_types[i].toBasicDatatype());
		}
		return locals;
	}

	public int registerSlot(int alias) {
		if (referenceSlots == null)
			referenceSlots = new IntegerHashtable<Integer>(128);
		Integer slot=referenceSlots.get(alias);
		if (slot==null) {
			slot = new Integer(referenceCount++);
			referenceSlots.put(alias,slot);
		}
		return slot.intValue();
	}

	public IMMethod getMethod() {
		return method;
	}

	public BuilderOptions getOptions() {
		return opts;
	}

	public IMSlot getLocalVariable(int type, int src_alias) {

		type = normalizeType(type);

		// locals is the first local variable slot used by a local
		// variable (lower ones are occupied by the method parameters)
		int locals=0;
		for (int i=0;i<arg_types.length;i++) {
			locals+=BCBasicDatatype.sizeInWords(arg_types[i].toBasicDatatype());
		}

		int typed_alias = src_alias;
		if (src_alias >= locals) {
			// not a method argument, code the type of the local in the upper
			// bits of the alias
			typed_alias += (type<<16);
		}

		IMSlot ret = all.get(typed_alias);
		if (ret==null) {
			ret = new IMSlot(this, type, src_alias);
			all.put(typed_alias,ret);
		}

		return ret;
	}

	public IMSlot getStackVariable(int type, int stack_alias) {
		type = normalizeType(type);
		int real_alias = (type<<16) + stack_alias;
		IMSlot ret = stack_index.get(real_alias);
		if (ret==null) {
			ret = createNewVariable(type);
			ret.setIsStackVar(stack_alias);
			stack_index.put(real_alias,ret);
		}
		return ret;
	}

	public IMSlot createNewVariableSSA(IMSlot old) {
		IMSlot newSlot =  createNewVariable(old.getType());
		newSlot.setName(old.incName());
		return newSlot;
	}

	public IMSlot createNewVariable(int type) {
		type = normalizeType(type);
		int alias = new_count++;
		IMSlot ret = new IMSlot(this, type, alias);
		all.put((type<<16)+alias,ret);
		return ret;
	}

	public String getArgumentClass(int nr) throws CompileException {
		BasicTypeDescriptor desc = arg_types[nr];
		if (desc.toBasicDatatype()!=BCBasicDatatype.REFERENCE) return null;
		return desc.getClassName();
	}

	public IMNode readArgument(int type, int nr) throws CompileException {
		type = normalizeType(type);
		IMSlot[] args = getMethodArguments();
		if (args[nr].getType()!=type)
			throw new CompileException("wrong argument type!");
		return method.readLocal(0, args[nr]);
	}

	/**
	 * parameter optimization.
	 */
	public IMNode constant_parameter_propagation() throws CompileException {

		if (!opts.hasOption("const_arg_prop")) return null;

		callerInfo = method.getFoldedCallerInfo();
		if (callerInfo==null) return null;

		IMListHead list = new IMListHead();

		int t=0;
		for (int i=0;i<arg_types.length;i++) {
			int type = arg_types[i].toBasicDatatype();
			IMSlot slot = all.get(t);
			if (slot==null) slot = getLocalVariable(type,t);

			IMDataFlow df = callerInfo.elementAt(i);
			slot.setDataFlowInfo(df);

			if (df!=null && df.isConstant()) {
				try {
					IMNode constant = df.getConstValue(method,0);
					IMNode store = method.createStoreVariable(slot, constant, 0);
					slot.define(method.iterator().next(), store);
					opts.verbose("++ propagate method argument "+store.toReadableString());
					list.insert(store);
				} catch (CompileException ex) {
					System.err.println(method.getMethodNameAndType()+": "+df.getClass().getName());
					throw ex;
				}
			}
			t=t+BCBasicDatatype.sizeInWords(type);
		}

		resetArgCache();
		return list;
	}

	public IMSlot[] getOriginalArguments() throws CompileException {
		IMSlot[] jargs = new IMSlot[arg_types.length];
		int t=0;
		for (int i=0;i<arg_types.length;i++) {
			int type = arg_types[i].toBasicDatatype();
			IMSlot slot = all.get(t);
			if (slot==null) slot = getLocalVariable(type,t);
			jargs[i]=slot;
			t=t+BCBasicDatatype.sizeInWords(type);
		}
		return jargs;
	}

	public boolean isArgument(IMSlot slot) {
		if (slot.sameFrame(this) == false)
			return false;
		return (slot.getVarNumber() < argRange());
	}

	public boolean isArgUnused(IMSlot arg) {
		if (isArgument(arg) == false)
			return true;
		IMDataFlow df = arg.getDataFlowInfo();
		if (df == null)
			return false;
		return df.isConstant();
	}

	private IMSlot[] arg_slots_cache = null;

	public void resetArgCache() {
		arg_slots_cache = null;
	}

	public IMSlot[] getMethodArguments() throws CompileException {

		// always remember caches are evil!
		if (arg_slots_cache!=null && !opts.hasOption("_X_no_arg_cache"))
			return arg_slots_cache;

		IMSlot[] arg_slots;
		int l=0;

		boolean blockmethod_args = method.canBlock();
	 	if(	method.getMethodTable() != null ) {
			blockmethod_args |= method.getMethodTable().getFirstCandidate().canBlock();
			if(method.getMethodTable().getTopAbstractMethod() != null) {
				blockmethod_args |= method.getMethodTable().getTopAbstractMethod().canBlock();
			}
		}

		// blocking method get a pointer to the current tail of the linked list
		// of references as the first parameter
		if(opts.hasLinkedListOfLocalReferences() && blockmethod_args) {
			arg_slots = new IMSlot[arg_types.length+1];
			arg_slots[0]=new IMSlotPref(this);
			l = 1;
		} else {
			arg_slots = new IMSlot[arg_types.length];
		}

		for (int i=0,lv_index=0; i<arg_types.length; i++) {
			int type = arg_types[i].toBasicDatatype();
			IMSlot slot = all.get(lv_index);
			if (slot==null)
				slot = getLocalVariable(type,lv_index);

			if (callerInfo != null) {
				if (isConstantArgInAllCandidates(i) == false)
					arg_slots[l++]=slot;
			} else {
				arg_slots[l++]=slot;
			}
			lv_index += BCBasicDatatype.sizeInWords(type);
		}

		if (l < arg_slots.length)
			arg_slots = Arrays.copyOf(arg_slots, l);

		arg_slots_cache = arg_slots;
		return arg_slots;
	}

	/**
	 * arguments for method definition
	 */
	public String getArgString() throws CompileException {
		StringBuffer args = new StringBuffer();

		IMSlot[] arguments = getMethodArguments();

		if (arguments.length == 0) {    // void
			args.append("void");
		} else {
			for (int i=0;i<arguments.length;i++) {
				IMSlot slot = arguments[i];
				if (i>0) args.append(',');
				args.append(slot.definition());
			}
		}

		return args.toString();
	}

	public String getArgReCallString() throws CompileException {
		StringBuffer args = new StringBuffer();

		IMSlot[] arguments = getMethodArguments();

		args.append('(');

		for (int i=0;i<arguments.length;i++) {
			IMSlot slot = arguments[i];
			if (i>0) args.append(',');
			args.append(slot.toString());
		}

		args.append(')');
		return args.toString();
	}

	public String[] getArgNames() throws CompileException {
		IMSlot[] arguments = getMethodArguments();
		String[] args = new String[arguments.length];
		for (int i=0;i<arguments.length;i++) { args[i]=arguments[i].toString(); }
		return args;
	}

	public int numberOfRefs() {
		return referenceSlots.size();
	}

	public int[] getArgTypes() throws CompileException {
		IMSlot[] arguments = getMethodArguments();
		int[] types = new int[arguments.length];
		for (int i=0;i<arguments.length;i++) { types[i] = arguments[i].getType(); }
		return types;
	}

	public boolean hasReferenceArguments() throws CompileException {
		for(BasicTypeDescriptor btd: arg_types)
			if(!btd.isPrimitive())
				return true;

		return false;
	}

	public void beginInlineForeignFrame() {
		foreignSlots = new SplitKeyHashMap<Integer, Integer, IMSlot>();
	}

	public boolean hasUpdate(IMSlot foreignSlot) {
		return foreignSlots.containsKey(foreignSlot.getVarNumber(),
				foreignSlot.getType());
	}

	public IMSlot adjustSlot(IMSlot foreignSlot) throws CompileException {

		if (foreignSlot.sameFrame(this)) return foreignSlot;

		int varNo = foreignSlot.getVarNumber();
		int type  = foreignSlot.getType();
		IMSlot nslot = foreignSlots.get(varNo, type);
		if (nslot==null) {
			nslot = createNewVariable(type);
			foreignSlots.put(varNo, type, nslot);
		}

		return nslot;
	}

	public void endInlineForeignFrame() {
		foreignSlots=null;
	}

	public int referenceSlotSize() {
		if (referenceSlots==null) return 0;
		return referenceSlots.size();
	}

	/**
	 * emit the arguments for caller side.
	 */
	public void emitArguments(Coder coder, IMMethodFrame callerFrame,
			IMNode self, IMNode[] args) throws CompileException {

		boolean sep  = false;
		coder.add('(');

		// Add precursor link
		if (method.canBlock() && opts.hasLinkedListOfLocalReferences()) {
			int refSize = callerFrame.referenceSlotSize();
			if (refSize == 0) {
				coder.add("pref");
			} else {
				coder.add("(object_t **) &obj[" + refSize + "]");
			}
			sep = true;
		}

		// Emit object reference (if any)
		int j = 0;
		if (self != null) {
			if (sep == true) {
				if(opts.hasOption("use_coded_refs") && this.method.isVirtual()) {
					coder.add("), NATIVE_ADDR(");
				} else {
					coder.add(", ");
				}
			}
			if(opts.hasOption("use_coded_refs") && this.method.isVirtual()) {
				coder.add("NATIVE_ADDR(");
			}
			self.translate(coder);
			++j;
			sep = true;
			if(opts.hasOption("use_coded_refs") && this.method.isVirtual()) {
				coder.add(")");
			}
		}
		// Emit the actual arguments
		for (IMNode arg : args) {

			// Skip the current argument if it has been eliminated
			if (isConstantArgInAllCandidates(j)) {
				++j;
				continue;
			}

			// Emit the current argument
			if (sep == true) {
				if(opts.hasOption("use_coded_refs") && this.method.isVirtual()) {
					if(arg.isReference()) {
						coder.add(", NATIVE_ADDR(");
					} else {
						coder.add(", ");
					}
				} else {
					coder.add(", ");
				}
			}
			arg.translate(coder);
			if(sep == true) {
				if(opts.hasOption("use_coded_refs") && this.method.isVirtual()) {
					if(arg.isReference()) {
						coder.add(")");
					}
				}
			}
			sep = true;
			++j;
		}

		coder.add(')');
	}

	public void codeEOLL(Coder coder) throws CompileException {
		if (opts.hasLinkedListOfLocalReferences()) {
			if (!method.canBlock())
				throw new CompileException("Unexpected block operation! (try workaround -X:_X_all_CanBlock and report the bug)");
			if (referenceSlots==null) {
				coder.add_before("\t*pref=KESO_EOLL; /* we have no local references. */\n");
			} else {
				coder.add_before("\tobj["+referenceSlots.size()+"]=KESO_EOLL;\n");
			}
		}
	}

	public int resetUses() throws CompileException {
		int sum = 0;
		for (IMSlot slot : all.getElements()) {
			sum += slot.numberOfUses();
			slot.resetUses();
		}
		return sum;
	}

	private boolean hasValidUsageInformation = false;
	public void opt_vars() { hasValidUsageInformation = true; }

	/**
	 * Remove unused local variables and map object references
	 * to a list of linked references (llrefs) if the task may block.
	 */
	private IntegerHashtable<IMSlot> finalize_local_vars() throws CompileException {
		boolean llrefs = opts.hasLinkedListOfLocalReferences() && method.canBlock();
		int keys[] = all.sortedKeys();

		if (hasValidUsageInformation) {
			IntegerHashtable<IMSlot> nall= new IntegerHashtable<IMSlot>(keys.length);

			for (int i=0;i<keys.length;i++) {
				IMSlot slot = all.get(keys[i]);

				if (!slot.hasNoUses() || i<arg_types.length) {
					if (llrefs && slot.isReference()) {
						slot.registerRef();
					}
					nall.put(keys[i], slot);
				}
			}

			all = nall;
		} else {
			for (int i=0;i<keys.length;i++) {
				IMSlot slot = all.get(keys[i]);
				if (llrefs && slot.isReference()) {
					slot.registerRef();
				}
			}
		}
 		/*
		IntegerHashtable<IMSlot> nall=null;
		if (hasValidUsageInformation)
			nall= new IntegerHashtable<IMSlot>();

		for (int i=0;i<keys.length;i++) {
			IMSlot slot = all.get(keys[i]);

			if (llrefs && slot.getType()==BCBasicDatatype.REFERENCE) {
				if (hasValidUsageInformation) {
					if (!slot.hasNoUses() || i<arg_types.length) {
						slot.registerRef();
					}
				} else {
					slot.registerRef();
				}
			}

			if (hasValidUsageInformation) {
				if (!slot.hasNoUses() || i<arg_types.length) {
					nall.put(keys[i], slot);
				}
			}
		}

		if (hasValidUsageInformation)
			all = nall;
*/
		return all;
	}

	public void translate(Coder coder) throws CompileException {
		boolean llrefs = opts.hasLinkedListOfLocalReferences() && method.canBlock();
		int keys[] = finalize_local_vars().sortedKeys();

		int locals=0;
		for (int i=0;i<keys.length;i++) {
			IMSlot slot = all.get(keys[i]);
			locals+=BCBasicDatatype.sizeInWords(slot.getType());
		}

		int args=0;
		for (int i=0;i<arg_types.length;i++) {
			IMSlot slot = all.get(keys[i]);
			IMDataFlow df = slot.getDataFlowInfo();
			if (df != null && df.isConstant() && (slot.hasNoUses() == false
					|| slot.getDefiningStatement() != null)) {
				coder.add(slot.definition());
				coder.addln(";");
			}
			args+=BCBasicDatatype.sizeInWords(arg_types[i].toBasicDatatype());
		}

		if (args<locals) {
			for (int i=0;i<keys.length;i++) {
				IMSlot slot = all.get(keys[i]);
				if (slot.getVarNumber()<args) continue;
				if (llrefs && slot.getType()==BCBasicDatatype.REFERENCE) continue;
				if (opts.hasOption("_X_dbg_uses")) {
					coder.addln();
					for (IMNode un : slot.getUses()) {
						coder.add("/* ");
						coder.add(un.getMethod().getMethodName());
						coder.add(" ");
						coder.add(un.lineInfo());
						coder.add(":");
						coder.add(un.dumpBC());
						coder.add(',');
						coder.addln(" */");
					}
				}
				coder.add(slot.definition());
				coder.addln(";");
			}

			if (llrefs && referenceSlots!=null) {
				int frame_size = referenceSlots.size();
				/* extra space for end frame list marker */
				if (method.canBlock()) frame_size++;
				/* extra space for end frame marker */
				if (opts.useLLRefMarker2()) frame_size++;
				coder.add("object_t* volatile obj[");
				coder.add(frame_size);
				coder.addln("];");
				if (opts.useLLRefMarker2()) {
					coder.add("obj[");
					coder.add(frame_size-1);
					coder.addln("]=KESO_FRAME_END_MARK;");
				}
				if (method.canBlock()) {
					if (referenceSlots.size()>2) {
						coder.addln("/* initial new frame (big) */");
						coder.add("{int c; for (c=0;c<");
						coder.add(referenceSlots.size());
						coder.addln(";c++) obj[c]=(object_t*)0;}");
					} else {
						coder.addln("/* initial new frame (small) */");
						if (referenceSlots.size()>1)
							coder.addln("obj[1]=0;");
						coder.addln("obj[0]=0;");
					}

					/*
					 * It is adequate to terminate the list just before
					 * we block. (see blocking Systemcalls for details)
					 */
					if (!opts.useLLRefMarker2()) {
						coder.addln("*pref=(object_t*)((int)obj|0x1); /* register new frame */");
					} else {
						coder.addln("*pref=(object_t*)obj; /* register new frame */");
					}
				}
			}

			if(opts.hasOption("debug")) {
				for (int i=0;i<arg_types.length;i++) {
					if (!arg_types[i].isPrimitive()) {
						IMClass arg_class = coder.getClassStore().getClass(arg_types[i].getClassName());
						IMSlot slot = all.get(keys[i]);
						IMDataFlow df = slot.getDataFlowInfo();
						if (df!=null && df.isConstant()) continue;
						if(arg_class == null) continue;

						coder.add_class(arg_class);
						if (arg_class.isInterface()) {
							coder.add("CHK_IFACE_ID");
							coder.add(arg_class.getClassTypeID());
							coder.add('(');
							coder.add(slot.toString());
							coder.addln(");");
						} else {
							coder.add("KESO_CHECK_CAST(");
							coder.add(arg_class.getClassID());
							coder.add(',');
							coder.add(arg_class.getAlias().toUpperCase());
							coder.add("_RANGE,");
							coder.add(slot.toString());
							coder.addln(");");
						}
					}
				}
			}
		}

		/* inform the coder of the end of local definition */
		coder.endOfLocals();
	}

	public Iterable<IMSlot> getAllSlots() {
		return all.getElements();
	}

	/**
	 * Normalizes a type.
	 * This method converts every type that is smaller than an integer slot to
	 * INT. This is needed because there are no stack slots smaller than the
	 * size of an integer.
	 * @param type Original type.
	 * @return Normalized type.
	 */
	private static int normalizeType(int type) {
		switch (type) {
		case BCBasicDatatype.BOOLEAN:
		case BCBasicDatatype.BYTE:
		case BCBasicDatatype.CHAR:
		case BCBasicDatatype.SHORT:
			return BCBasicDatatype.INT;
		default:
			return type;
		}
	}

	private boolean isConstantArgInAllCandidates(int index) {
		MethodTable methodTable = method.getMethodTable();
		if (methodTable == null) // Table not computed yet
			return false;
		for (IMMethod candidate : methodTable.getCandidates()) {
			if (isConstantArg(candidate.getMethodFrame().all, index) == false)
				return false;
		}
		return true;
	}

	private static boolean isConstantArg(IntegerHashtable<IMSlot> slots,
			int index) {
		int[] keys = slots.sortedKeys();
		if (index >= keys.length)
			return false;
		IMDataFlow df = slots.get(keys[index]).getDataFlowInfo();
		return (df != null && df.isConstant());
	}

}
