/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.datatypes.*;

import keso.compiler.*;
import keso.compiler.backend.*;

public class IMNode implements Comparable<IMNode>, TranslationCallback {

	protected final static IMNode NONE = new IMNode();

	private static int nodeCount;

	protected IMMethod method;
	protected IMMethod src_method;
	protected BuilderOptions opts;

	protected int bytecode;
	protected int datatype;
	protected int bcPosition;
	protected int src_bcPosition;

	protected IMNode next_node;

	/**
	 * Virtual register ID.
	 * This is a temporary ID that is assigned to the node when dumping the
	 * intermediate code. It represents a virtual register that holds the result
	 * of the instruction.
	 * We could use the byte code position to identify nodes,
	 * but this is no longer necessarily unique after applying some
	 * transformations to the intermediate code - either because instructions
	 * are split/moved or because new instruction nodes are inserted (at the
	 * virtual position -1).
	 */
	private String regId;

	/**
	 * Node ID.
	 * In order to make things (more) deterministic, some passes use TreeSets
	 * and TreeMaps instead of HashSets and HashMaps. This requires that two
	 * {@link IMNode}s can be compared with each other.
	 */
	private int nodeId = nodeCount++;

	protected IMNode() {
		this.method     = null;
		this.opts       = null;
		this.bytecode   = -1;
		this.datatype   = -1;
		this.src_method = null;
		this.bcPosition = -1;
		this.src_bcPosition = -1;
	}

	public IMNode(IMMethod method, int bc, int type, int bcpos) {
		this.method     = method;
		this.opts       = method.getOptions();
		this.bytecode   = bc;
		this.datatype   = type;
		this.src_method = method;
		this.bcPosition = bcpos;
		this.src_bcPosition = bcpos;
	}

	final public BuilderOptions getOptions() {
		return opts;
	}

	final public void copyNodeValues(IMCopyVisitor visitor, IMNode old) {
		opts       = old.opts;
		bytecode   = old.bytecode;
		datatype   = old.datatype;
		src_method     = old.src_method;
		src_bcPosition = old.src_bcPosition;

		// this should be overridden in visitor.visit()
		method     = old.method;
		bcPosition = old.bcPosition;
	}

	public IMNode copy(IMCopyVisitor visitor) throws CompileException {
		ONI();
		return null;
	}

	public void  visitNodes(IMVisitor visitor) throws CompileException {
		ONI();
	}

	public void accept(IMTypedVisitor visitor) throws CompileException {
		ONI();
	}

	/**
	 * This method is called by the IMCallGraphVisitor.
	 */
	public void analyseCall() throws CompileException {
		return;
	}

	public IMNode ssa_optimize() throws CompileException {
		return this;
	}

	/**
	 * isConstObject() returns true if the object content is not changeable!
	 */
	public boolean isConstObject() throws CompileException {
		return false;
	}

	/**
	 * determineType() returns the runtime type of the node
	 * or null if the type is not predictable at compile time.
	 */
	public IMClass determineType() {
		return null;
	}

	public boolean isTypeOf(IMClass type) throws CompileException {
		return false;
	}

	public void escapeAnalyses(int mode) throws CompileException {
	}

	public void setEscapePath(IMNode node) throws CompileException {
	}

	public IMNode getEscapePath() throws CompileException {
		return null;
	}

	public boolean doesEscape(IMNode node, boolean isArgument) throws CompileException {
		return false;
	}

	public IMNode whyEscape(IMNode node, boolean isArgument) throws CompileException {
		return null;
	}

	public int getArrayLength() throws CompileException {
		return -1;
	}

	public int costs() throws CompileException {
		ONI();
		return 0;
	}

	final public boolean isChecked(Coder coder) throws CompileException {
		return isChecked(coder.getCurrentBasicBlock());
	}

	public boolean isChecked(IMBasicBlock bb) throws CompileException { return false; }

	final public void setChecked(Coder coder) throws CompileException {
		setChecked(coder.getCurrentBasicBlock());
	}

	public void setChecked(IMBasicBlock bb) throws CompileException { }

	public boolean emptyList() { return false; }

	public boolean isBasicBlock() { return false; }

	public boolean isConstant()   { return false; }

	public boolean isNullValue() { return false; }

	public boolean isFinalStatic() throws CompileException { return false; }

	public boolean isBranch()     { return false; }

	public boolean isVariable()   { return false; }

	public boolean usesSlot() {
		return false;
	}

	public boolean isStackVariable(int stack_pos) { return false; }

	public boolean hasConstant()  { return false; }

	/**
	 * To call this method should be faster than using a
	 * cast operation.
	 */
	public IMConstant nodeToConstant() throws CompileException { throw new CompileException("not constant!"); }

	public boolean hasSideEffect() { return true; }

	/**
	 * nodes who only read but not change the global state
	 * are pure.
	 */
	public boolean isPureNode() {
		if (hasSideEffect()) {
			return false;
		}
		// a node without side effects is always pure
		return true;
	}

	/**
	 * The byte code will be removed if the result is
	 * unused.
	 */
	public boolean removeIfUnUsed() {
		if (isPureNode()) return true;
		return false;
	}

	public boolean isVolatile() throws CompileException {
		return false;
	}

	public boolean isAdd() {
		return false;
	}

	public boolean isSub() {
		return false;
	}

	public boolean isMul() {
		return false;
	}

	public boolean isBitOpr() {
		return false;
	}

	public boolean isNew() {
		return false;
	}

	public boolean isInvoke() {
		return false;
	}

	final public boolean isDouble() throws CompileException {
		if (datatype==-1) throw new CompileException("wrong or unkown datatype on stack");
		return BCBasicDatatype.isDouble(datatype);
	}

	final public int getBCPosition() { return bcPosition; }

	public int getDatatype() { return datatype; }

	final public boolean isByte() { return datatype==BCBasicDatatype.BYTE; }

	final public boolean isShort() { return datatype==BCBasicDatatype.SHORT; }

	final public boolean isReference() { return datatype==BCBasicDatatype.REFERENCE; }

	final public IMMethod getMethod() { return method; }

	public IMNode getLeftNode() throws CompileException { throw new CompileException("node has no left operand"); }

	public IMNode getRightNode() throws CompileException { throw new CompileException("node has no right operand"); }

	public IMSlot getIMSlot() { return null; }

	public void setIMSlot(IMSlot slot) { return; }

	public boolean checkReference() { return true; }

	public boolean checkArrayRange(int index) { return true; }

	public boolean checkArrayRange(IMOperand index) { return true; }

	final protected String nodeInfo() {
		try {
			return "("+method.getClassName()+":"+method.getMethodName()+" bcpos: "+bcPosition+")";
		} catch (RuntimeException ex) {
			return "("+getClass().getName()+")";
		}
	}

	public IMNode processStack(VirtualOperandStack stack, IMNode extra_ops) throws CompileException {
		ONI();
		return null;
	}

	/**
	 * @return basic block after insertion
	 */
	public IMBasicBlock inlineMethodCalls(IMBasicBlock current, IMNode prev) throws CompileException {
		return null;
	}

	public boolean isMemoryType() throws CompileException {
		return false;
	}

	/** Determines whether this node reads a class/object field. */
	public boolean isFieldAccess() throws CompileException {
		return false;
	}

	/**
	 * emit() is similar to translate but only needed by chk_ref and
	 * therefore only implemented by IMAReadLocalVariable and IMGetField.
	 */
	public String emit() throws CompileException {
		ONI();
		return null;
	}

	public boolean translate_MT_Ref(Coder coder, String reason) throws CompileException {
		return false;
	}

	public boolean needSemicolon() throws CompileException {
		return true;
	}

	/**
	 * This method is the right place to add a piece of code to global.c if needed.
	 * The imnode must be registered with
	 * repository.registerGlobalTranslationCallback(alias, callback);
	 */
	public void translate_global(Coder coder) throws CompileException {
		System.err.println(dumpBC());
		ONI();
		return;
	}

	public void translate(Coder coder) throws CompileException {
		System.err.println(dumpBC());
		ONI();
		return;
	}

	final public boolean isEndOfBasicBlock() {
		return next_node==null;
	}

	final public void setEndOfBasicBlock() {
		next_node = null;
	}

	final public IMNode next() {
		return next_node;
	}

	final public IMNode clear_link() {
		next_node = null;
		return this;
	}

	/**
	 * Append a list of IMNode objects. The list must start with
	 * a IMListHead object which is skipped, otherwise the behaviour
	 * is the same as append(). The end of the appended list
	 * is returned.
	 */
	final public IMNode append_list(IMNode node) {
		if (node==null) return this;

		if (node instanceof IMListHead) {
			if (node.next_node==null) return this;

			this.next_node = node.next_node;
			while (node.next_node!=null) { node=node.next_node; }

			return node;
		}

		this.next_node = node;

		return node;
	}

	/**
	 * Inserts an IMNode object after this node.
	 * @param node Node to be inserted.
	 * @return The inserted node.
	 */
	final public IMNode insert(IMNode node) {
		if (node.next_node != null)
			throw new Error("Node lost: " + node.toReadableString()
					+ ", successor is " + node.next_node.toReadableString());
		node.next_node = this.next_node;
		this.next_node = node;
		return node;
	}

	/**
	 * Append an IMNode object.
	 */
	final public IMNode append(IMNode node) {
		this.next_node = node;
		return node;
	}

	public String dumpOperands() {
		return null;
	}

	public String dumpBC() {
		String ret      = getClass().getSimpleName().substring(2);
		String operands = dumpOperands();
		if (operands != null)
			ret += ' ' + operands;
		return ret;
	}

	public String lineInfo() {
		if (src_method != null) {
			return src_method.lineInfo(src_bcPosition);
		}
		return "unknown (bc:"+src_bcPosition+")";
	}

	public String toReadableString() {
		return bcPosition+": "+bytecode;
	}

	public String toMemAlias() {
		return toReadableString();
	}

	public boolean hasToBeChecked() {
		return false;
	}

	public void emitCheck(IMNode checkList, IMBasicBlock successor)
	throws CompileException {
		ONI();
	}

	public IMConstant evaluate(IMConstant... args) {
		return null;
	}

	public Integer evaluateMin(Integer... limits) {
		return null;
	}

	public Integer evaluateMax(Integer... limits) {
		return null;
	}

	@Override
	public int compareTo(IMNode o) {
		return nodeId - o.nodeId;
	}

	/* debugging stuff */

	final protected void assertIsReference(IMNode node) throws CompileException {
		if (node.getDatatype()!=BCBasicDatatype.REFERENCE)
			throw new CompileException("bcpos:"+Integer.toString(bcPosition)+" wrong datatype on stack!");
	}

	final protected void assertIsIntValue(IMNode node) throws CompileException {
		int t = node.getDatatype();
		if (!(t==BCBasicDatatype.INT||t==BCBasicDatatype.SHORT||t==BCBasicDatatype.BYTE))
			throw new CompileException("bcpos:"+Integer.toString(bcPosition)+" wrong datatype on stack!");
	}

	final protected boolean isCategory2(IMNode node) {
		return (node.getDatatype()==BCBasicDatatype.LONG)||(node.getDatatype()==BCBasicDatatype.DOUBLE);
	}

	final protected void assertIsLongOrDouble(IMNode node) throws CompileException {
		int t = node.getDatatype();
		if (!(t==BCBasicDatatype.LONG||t==BCBasicDatatype.DOUBLE))
			throw new CompileException("bcpos:"+Integer.toString(bcPosition)+" wrong datatype on stack!");
	}

	final protected void ONI() throws CompileException {
		throw new Error(getClass().getName()+": Operation not implemented! "+nodeInfo());
	}

	final public void setRegId(int id) {
		this.regId = "%" + id;
	}

	final public String getRegId() {
		return regId;
	}

	protected static String getRegId(IMNode node) {
		if (node == null)
			return null;
		return node.getRegId();
	}

	protected static String getVarName(IMSlot slot) {
		if (slot == null)
			return null;
		return slot.toString();
	}

}
