/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.CompileException;
import keso.compiler.BuilderOptions;
import keso.compiler.ClassStore;
import keso.compiler.backend.Coder;
import java.io.IOException;

final public class IMObjectArrayClass extends IMArrayClass {

	private boolean require_multi_alloc = false;
	private IMClass elem_class;

	public IMObjectArrayClass(BuilderOptions opts, ClassStore src_repository, String classname,
			String alias, String element_type, boolean isFinal) {
		super(opts,src_repository,classname,alias,element_type,isFinal);
	}

	public IMObjectArrayClass(BuilderOptions opts, ClassStore src_repository, String classname,
			IMClass elem_class, String element_type, boolean isFinal) {
		super(opts,src_repository,classname,"",element_type,isFinal);
		this.elem_class=elem_class;
		require_alloc=true;
	}

	@Override
	public void requireMultiAlloc() {
		require_alloc=true;
		require_multi_alloc=true;
	}

	@Override
	public IMClass cloneClass(ClassStore src_repository) {
		if (elem_class!=null) {
			return new IMObjectArrayClass(opts,src_repository,name,elem_class,element_type,isFinal);
		}
		return new IMObjectArrayClass(opts,src_repository,name,alias,element_type,isFinal);
	}

	@Override
	public void global_header(Coder coder) throws CompileException {
	}

	@Override
	public String getAlias() {
		if (alias.equals("") && elem_class!=null) alias=elem_class.getAlias()+"_array";
		return alias;
	}

	@Override
	public void translate(Coder coder) throws CompileException {
		opts.verbose("###### create "+getClassName());

		getAlias();

		coder.beginHeader(this);

		coder.header_add("#include \"keso_types.h\"\n");

		coder.header_add("\n#define ");
		coder.header_add(alias.toUpperCase());
		coder.header_add("_ID ((class_id_t)");
		coder.header_add(type_info.getClassTypeID());
		coder.header_add(")\n");

		coder.header_add("typedef struct {\n");
		coder.header_add("ARRAY_HEADER\n\t");
		coder.header_add("class_id_t type;\n\t");
		coder.header_add(element_type);
		coder.header_add(" data[1];\n");
		coder.header_add("} ");
		coder.header_add(alias);
		coder.header_add("_t;\n");

		if (require_alloc || require_multi_alloc) coder.beginClassMethod(this);

		if (!alias.equals("object_array")) {
			coder.add_class("object_array");
		}

		/* keso_alloc_object_array */
		if (require_alloc) {
			coder.header_add("\nobject_t	*keso_alloc_");
			coder.header_add(alias);
			coder.header_add("(class_id_t class_id, array_size_t size);\n");

			coder.local_add("#include \"");
			coder.local_add(alias);
			coder.local_add(".h\"\n\n");

			coder.local_add("\nobject_t	*keso_alloc_");
			coder.local_add(alias);
			coder.local_add("(class_id_t class_id, array_size_t size) {\n");
			coder.local_add("	object_array_t* array;\n");
			coder.local_add("	array = (object_array_t*)KESO_ALLOC(");
			coder.local_add(alias.toUpperCase());
			coder.local_add("_ID,sizeof(object_array_t)+(size*");
			coder.local_add(emitObjectSize());
			coder.local_add("),0);\n");
			coder.local_add("	array->type=class_id;\n");
			coder.local_add("	array->size=size;\n");
			coder.local_add("	return (object_t*)array;\n");
			coder.local_add("}\n");
		}

		/* keso_allocMultiArray */

		if (require_multi_alloc) {
			coder.header_add("#include <stdarg.h>\n");
			coder.header_add("object_t 	*keso_allocMultiArray(class_id_t class_id, array_size_t dim, ... );\n");

			coder.local_add("\nstatic object_t *malloc_rec(class_id_t class_id, array_size_t dim, array_size_t size, va_list arg ) {\n");
			coder.local_add("\tif(dim == 1) {\n");
			coder.local_add("\t\tif(keso_isObjectArrayClass(class_id)) {\n");
			coder.local_add("\t\t\treturn keso_alloc_object_array(class_id, size);\n");
			coder.local_add("\t\t} else { // primitive data array\n");
			coder.local_add("\t\t\tarray_t* array = (array_t*)\n");
			coder.local_add("\t\t\tKESO_ALLOC(class_id,sizeof(array_t)+(size * CLASS(class_id).size ),0);\n");
			coder.local_add("\t\t\tarray->size=size;\n");
			coder.local_add("\t\t\treturn (object_t*)array;\n");
			coder.local_add("\t\t}\n");
			coder.local_add("\n");
			coder.local_add("\t} else {\n");
			coder.local_add("\t\tobject_array_t* array = (object_array_t*)\n");
			coder.local_add("\t\t\tkeso_alloc_object_array(OBJECT_ARRAY_ID, size);\n");
			coder.local_add("\t\tarray_size_t next_size = va_arg(arg,array_size_t);\n\n");

			coder.local_add("\t\tif(next_size > 0)	while(size --> 0)\n");
			coder.local_add("\t\t\tarray->data[size] = (object_t*) malloc_rec(class_id, dim-1, next_size, arg);\n\n");

			coder.local_add("\t\treturn (object_t*) array;\n");
			coder.local_add("\t}\n");
			coder.local_add("}\n\n");

			coder.local_add("object_t 	*keso_allocMultiArray(class_id_t class_id, array_size_t dim, ... ) {\n");
			coder.local_add("\tva_list arg;\n");
			coder.local_add("\tobject_t *array;\n");
			coder.local_add("\tarray_size_t size ;\n\n");

			coder.local_add("\tva_start(arg,dim);\n");
			coder.local_add("\tsize = va_arg(arg,array_size_t);\n");
			coder.local_add("\tarray = malloc_rec(class_id, dim, size, arg);\n");
			coder.local_add("\tva_end(arg);\n\n");

			coder.local_add("\treturn array;\n");
			coder.local_add("}\n");
		}

		if (require_alloc || require_multi_alloc) coder.endClassMethod();

		coder.endHeader();
	}

	@Override
	public void dumpBC() throws IOException { return ; }

	@Override
	protected void parseMethodData() { return; }

}
