/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.*;
import keso.classfile.constantpool.*;
import keso.classfile.datatypes.*;

import keso.compiler.CompileException;
import keso.compiler.BuilderOptions;
import keso.compiler.ClassStore;
import keso.compiler.backend.Coder;
import keso.compiler.config.*;
import keso.util.DecoratedNames;

class ClassInitMethodData extends MethodData {
	public String getMethodName() { return "<clinit>"; }
	public String getMethodType() { return "()V"; }
	public String getMethodNameAndType() { return "<clinit>()V"; }
	public int getBasicReturnType() { return BCBasicDatatype.VOID; }
	public BasicTypeDescriptor getReturnType() { return null; }
}

public class IMPortalClass extends IMClass {

	private ImportDefinition impDef;
	private String[]	 ifaceNames;

	private String              protocolClassName = null;
	private MethodData[] impl = null;

	public IMPortalClass(BuilderOptions opts, ClassStore repository, ImportDefinition impDef) {
		super(opts, repository, createClassName(impDef), null, true);

		this.impDef = impDef;
		this.ifaceNames = new String[] {impDef.getService().getServiceInterfaceName()};

		if (!impDef.isLocal()) {
			this.protocolClassName = impDef.getNetwork().getProtocolName();
		}
	}

	public static String createClassName(ImportDefinition impDef) {
		return "keso/core/" + impDef.getServiceName() + "Proxy";
	}

	/**
	 * copy class constructor. Used by the prune class tree algorithm.
	 */
	@Override
	public IMClass cloneClass(ClassStore new_repository) {
		IMPortalClass nclass = new IMPortalClass(opts, new_repository, impDef);
		return nclass;
	}

	/**
	 * This callback informs the class that it is required.
	 *
	 * This method is a callback from the ClassStore!
	 */
	@Override
	public void require(int domainID) {
		// require super class and interfaces first
		super.require(domainID);

		if (!impDef.isLocal()) {

			if (domainID!=0) {
				if (domainID!=impDef.parent.domainid)
					throw new Error("domain mix up");
			}

			repository.requireClass(domainID, protocolClassName);

			IMClass packetStream = repository.getClass("keso/core/io/PacketStream");
			MethodData[] _methods = packetStream.getMethodData();

			// constructor
			String cName = "<init>(L" + impDef.getNetwork().getDriverInterfaceName() + ";IIIBB)V";
			repository.requireMethod(domainID, protocolClassName, cName);

			for (int i = 0; i < _methods.length; ++i) {
				// require all methods found in the interface for the implementation
				repository.requireMethod(domainID, protocolClassName , _methods[i].getMethodNameAndType());
			}

			String driverClassName = impDef.getDriver();
			repository.requireClass(domainID, driverClassName);
			IMClass driverIface = repository.getClass(impDef.getNetwork().getDriverInterfaceName());
			_methods = driverIface.getMethodData();

			// constructor
			repository.requireMethod(domainID, driverClassName , "<init>()V");

			for (int i = 0; i < _methods.length; ++i) {
				// require all methods found in the interface for the implementation
				repository.requireMethod(domainID, driverClassName , _methods[i].getMethodNameAndType());
			}
		}
	}

	@Override
	public boolean isInterface() {
		return false;
	}

	@Override
	public String getSuperClassName() {
		return "java/lang/Object";
	}

	@Override
	public int getObjRefFieldCount() {
		return 0;
	}

	/**
	 * overrides getInterfaceNames() in IMClass
	 */
	public String[] getInterfaceNames() {
		return ifaceNames;
	}

	@Override
	public String getSourceFile() throws CompileException {
		return impDef.getService().getServiceClassName() + ".java";
	}

	@Override
	public ConstantPool getConstantPool() {
		return null;
	}

	@Override
	public MethodData[] getMethodData() {

		if (impl==null) {
			if (impDef.isLocal() || !opts.hasOption("new_portal_init")) {
				impl=impDef.getService().getMethods();
			} else {
				MethodData[] d = impDef.getService().getMethods();
				impl = new MethodData[d.length+1];
				impl[0] = new ClassInitMethodData();
				for (int i=0;i<d.length;i++) { impl[i+1] = d[i]; }
			}
		}

		return impl;
	}

	@Override
	protected IMMethod createIMMethod(MethodData source) throws CompileException {
		if (impDef.isLocal()) {
			return new IMLocalPortalMethod(this, source, repository, impDef);
		} else {
			if (source.getMethodName().equals("<clinit>")) {
				return new IMPortalConstructor(this, repository, impDef);
			}
			return new IMRemotePortalMethod(this, source, repository, impDef);
		}
	}

	public String getReceiveBufferSize() throws CompileException {
		return IMPortalService.getSizeOfReturnValues(getMethodData());
	}

	public String getTransmitBufferSize() throws CompileException {

		MethodData[] meths = getMethodData();
		int max_size = 0;

		if (meths==null) return "0";

		for (int i=0;i<meths.length;i++) {
			IMMethod meth = getMethod(meths[i].getMethodNameAndType());
			if (meth==null) continue;

			max_size = Math.max(max_size, meth.getSizeOfArguments());
		}

		return Integer.toString(max_size);
	}

	public String getDriverObjectName() {
		return DecoratedNames.toEscString(impDef.getServiceName()) + "_" + DecoratedNames.toEscString(impDef.getNetwork().getNetworkName()) + "_driver_obj";
	}

	public String getTransmitBufferObjectName() {
		return DecoratedNames.toEscString(impDef.getServiceName()) + "_" + DecoratedNames.toEscString(impDef.getNetwork().getNetworkName()) + "_transmit_buffer_obj";
	}

	public String getProtocolObjectName() {
		return DecoratedNames.toEscString(impDef.getServiceName()) + "_" + DecoratedNames.toEscString(impDef.getNetwork().getNetworkName()) + "_protocol_obj";
	}

	@Override
	public void translate(Coder coder) throws CompileException {

		if (methods==null) parseMethodData();

		// create proxy object
		coder.global_add_allocConstObject(this, coder.getResObjSymbol(impDef));
		coder.global_add_init_end();
		coder.global_addln("\n");

		// proxy objects for remote services need to be initialized
		if (!opts.hasOption("new_portal_init") && !impDef.isLocal()) {
			new IMPortalClassInit(this, repository, impDef).translate(coder);
		}

		super.translate(coder);
	}
}
