/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.Coder;
import keso.compiler.config.*;

import keso.util.*;

import java.util.*;


/**
 * This class implements the class constructor of a remote class.
 */
final public class IMPortalClassInit implements IMPortalInit {

    private ImportDefinition importDefinition;
    private ClassStore repository;
    private IMPortalClass thisClass;

    public IMPortalClassInit(IMPortalClass clazz, ClassStore src_repository, ImportDefinition importDef) throws CompileException {
        thisClass = clazz;
        importDefinition = importDef;
        repository = src_repository;
    }

    final public void translate(Coder coder) throws CompileException {
        String name = DecoratedNames.toEscString(importDefinition.getServiceName()) + "_proxy_init";

        // create protocol object
        IMClass protocolClass = repository.getClass(importDefinition.getNetwork().getProtocolName());
        coder.global_add_allocConstObject(protocolClass, thisClass.getProtocolObjectName());
        coder.global_add_init_end();
        coder.global_addln("\n");

        // create driver object
        IMClass driverClass = repository.getClass(importDefinition.getDriver());
        coder.global_add_allocConstObject(driverClass, thisClass.getDriverObjectName());
        coder.global_add_init_end();
        coder.global_addln("\n");

        // transmit buffer
        coder.global_header_add("extern object_t *");
        coder.global_header_add(thisClass.getTransmitBufferObjectName());
        coder.global_header_add(";\n");

        coder.global_add("object_t *");
        coder.global_add(thisClass.getTransmitBufferObjectName());
        coder.global_addln(";\n");


        coder.global_header_add("void ");
        coder.global_header_add(name);
        coder.global_header_add("(void);\n");

        coder.global_add("void ");
        coder.global_add(name);
        coder.global_addln("() {\n");


        // call the default constructor
        IMMethod constructor = driverClass.getMethod("<init>()V");
        coder.global_add("\t");
        coder.global_add(constructor.getAlias());
        coder.global_add("(");
        coder.global_add(coder.add_accessStaticObject(driverClass, thisClass.getDriverObjectName()));
        coder.global_addln(");");

        // call all specified bean methods
        List<String> methods
        = importDefinition.getNetwork().getDriverInitMethods();
        List<Integer> params
        = importDefinition.getNetwork().getDriverInitParameters();

        for (int i = 0; i < methods.size(); ++i) {
            IMMethod meth = driverClass.getMethod(methods.get(i));
            coder.global_add("\t");
            coder.global_add(meth.getAlias());
            coder.global_add("(");
            coder.global_add(coder.add_accessStaticObject(driverClass, thisClass.getDriverObjectName()));
            coder.global_add(", ");
            coder.global_add(params.get(i));
            coder.global_addln(");");
        }

        String cName = "<init>(L" + importDefinition.getNetwork().getDriverInterfaceName() + ";IIIBB)V";
        constructor = protocolClass.getMethod(cName);

        coder.global_add("\t");
        coder.global_add(constructor.getAlias());
        coder.global_add("(");
        coder.global_add(coder.add_accessStaticObject(protocolClass, thisClass.getProtocolObjectName()));
        coder.global_add(", ");
        coder.global_add(coder.add_accessStaticObject(driverClass, thisClass.getDriverObjectName()));
        // rxId
        coder.global_add(", 0x");
        coder.global_add(Integer.toHexString(ServiceManager.instance.getImportIdentifier(importDefinition)));
        // rxMask
        coder.global_add(", 0x");
        coder.global_add(Integer.toHexString(ServiceManager.instance.getImportMask()));
        // bufferMask
        coder.global_add(", 0x00000000");
        // idBits
        coder.global_add(", ");
        coder.global_add(ServiceManager.instance.getIDWidth());
        // numRxBuffers
        coder.global_addln(", 1);");


        // allocate a buffer for sending and receiving
        IMMethod allocReceiveBuffer = protocolClass.getMethod("allocReceiveBuffer(ISI)Z");
        IMMethod allocTransmitBuffer = protocolClass.getMethod("allocTransmitBuffer(SI)Lkeso/core/Memory;");

        coder.global_add("\tif (!");
        coder.global_add(allocReceiveBuffer.getAlias());
        coder.global_add("(");
        coder.global_add(coder.add_accessStaticObject(protocolClass, thisClass.getProtocolObjectName()));
        coder.global_add(", 0x");
        coder.global_add(Integer.toHexString(ServiceManager.instance.getImportIdentifier(importDefinition)));
        coder.global_add(", ");
        coder.global_add(thisClass.getReceiveBufferSize());
        coder.global_add(", ");
        coder.global_add(importDefinition.getAllocTimeout());
        coder.global_addln(")) {");
        coder.global_addln("\t\t/* Allocation of the rx buffer failed */");
        coder.global_addln("\t}");

        coder.global_add("\t");
        coder.global_add(thisClass.getTransmitBufferObjectName());
        coder.global_add(" = (object_t *) ");
        coder.global_add(allocTransmitBuffer.getAlias());
        coder.global_add("(");
        coder.global_add(coder.add_accessStaticObject(protocolClass, thisClass.getProtocolObjectName()));
        coder.global_add(", ");
        coder.global_add(thisClass.getTransmitBufferSize());
        coder.global_add(", ");
        coder.global_add(importDefinition.getAllocTimeout());
        coder.global_addln(");");
        coder.global_addln("}\n");

    }
}
