/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.Coder;
import keso.compiler.config.*;

import java.util.*;

/**
 * This class implements the portal call to a service.
 */
final public class IMPortalConstructor extends IMSystemMethod {

	private ImportDefinition impDefinition;
	private ClassStore repository;
	private IMPortalClass thisClass;

	public IMPortalConstructor(IMPortalClass clazz, ClassStore src_repository, ImportDefinition impDef) throws CompileException {
		super(clazz);
		hasNoCode = false;
		this.frame  = new IMMethodFrame(this, 0);
		thisClass = clazz;
		impDefinition = impDef;
		repository = src_repository;
	}

	@Override
	public String getMethodName() {
		return "<clinit>";
	}

	@Override
	public String getMethodNameAndType() {
		return "<clinit>()V";
	}

	@Override
	public boolean isConstructor() {
		return true;
	}

	@Override
	public boolean isStatic() {
		return true;
	}

	@Override
	public boolean isFinal() {
		return true;
	}

	@Override
	public boolean canBlock() throws CompileException {
		// FIXME: Im Startup-Hook duerfen wir uns nicht blockieren.
		// Daher brauchen wir auch keine llrefs! Im Moment tun wir
		// so, als ob wir immer blockieren.
		return true;
	}

	@Override
	public String getReturnType() {
		return "void";
	}

	@Override
	final public void translate(Coder coder) throws CompileException {

		getAlias();

		boolean llrefs = opts.hasLinkedListOfLocalReferences();

		// transmit buffer
		coder.header_add("extern object_t *");
		coder.header_add(thisClass.getTransmitBufferObjectName());
		coder.header_add(";\n");

		coder.local_add("object_t *");
		coder.local_add(thisClass.getTransmitBufferObjectName());
		coder.local_add(";\n");

		coder.beginMethod(this);
		coder.add_class(clazz.getAlias());

		coder.addln("object_t* protocol_obj;");
		coder.addln("object_t* driver_obj;");
		coder.addln();

		// create protocol object
		IMClass protocolClass = repository.getClass(impDefinition.getNetwork().getProtocolName());
		coder.add_class(protocolClass.getAlias());
		coder.add("protocol_obj = ");
		coder.add_allocConstObject(true, protocolClass, thisClass.getProtocolObjectName(),null);
		coder.add_init_end();
		coder.addln(";");

		// create driver object
		IMClass driverClass = repository.getClass(impDefinition.getDriver());
		coder.add_class(driverClass.getAlias());
		coder.add("driver_obj = ");
		coder.add_allocConstObject(true, driverClass, thisClass.getDriverObjectName(),null);
		coder.add_init_end();
		coder.addln(";");
		coder.addln();

		// call the default constructor
		IMMethod constructor = driverClass.getMethod("<init>()V");
		coder.add(constructor.getAlias());
		coder.add("(");
		if (llrefs && constructor.canBlock()) {
			coder.add("pref,");
		}
		coder.addln("driver_obj);");
		coder.addln();

		// call all specified bean methods
		coder.addln("/* call specified bean methods */");
		List<String> methods
		= impDefinition.getNetwork().getDriverInitMethods();
		List<Integer> params
		= impDefinition.getNetwork().getDriverInitParameters();
		for (int i = 0; i < methods.size(); ++i) {
			IMMethod meth = driverClass.getMethod(methods.get(i));
			coder.add(meth.getAlias());
			coder.add("(");
			if (llrefs && meth.canBlock()) {
				coder.add("pref,");
			}
			coder.add(coder.add_accessStaticObject(driverClass, thisClass.getDriverObjectName()));
			coder.add(", ");
			coder.add(params.get(i));
			coder.addln(");");
		}
		coder.addln();

		coder.addln("/* call protocol constructor */");
		String cName = "<init>(L" + impDefinition.getNetwork().getDriverInterfaceName() + ";IIIBB)V";
		constructor = protocolClass.getMethod(cName);

		coder.add(constructor.getAlias());
		coder.add("(");
		if (llrefs && constructor.canBlock()) {
			coder.add("pref,");
		}
		coder.add("protocol_obj,driver_obj");
		// rxId
		coder.add(",0x");
		coder.add(Integer.toHexString(ServiceManager.instance.getImportIdentifier(impDefinition)));
		// rxMask
		coder.add(",0x");
		coder.add(Integer.toHexString(ServiceManager.instance.getImportMask()));
		// bufferMask
		coder.add(",0x00000000");
		// idBits
		coder.add(",");
		coder.add(ServiceManager.instance.getIDWidth());
		// numRxBuffers
		coder.addln(",1);");


		// allocate a buffer for sending and receiving
		coder.addln("/* allocate receive buffer */");
		IMMethod allocReceiveBuffer = protocolClass.getMethod("allocReceiveBuffer(ISI)Z");
		IMMethod allocTransmitBuffer = protocolClass.getMethod("allocTransmitBuffer(SI)Lkeso/core/Memory;");

		coder.add("if (!");
		coder.add(allocReceiveBuffer.getAlias());
		coder.add("(");
		if (llrefs && allocReceiveBuffer.canBlock()) {
			coder.add("pref,");
		}
		coder.add("protocol_obj,0x");
		coder.add(Integer.toHexString(ServiceManager.instance.getImportIdentifier(impDefinition)));
		coder.add(",");
		coder.add(thisClass.getReceiveBufferSize());
		coder.add(",");
		coder.add(impDefinition.getAllocTimeout());
		coder.addln(")) {");
		coder.addln("\t/* Allocation of the rx buffer failed */");
		coder.addln("}");
		coder.addln();

		coder.addln("/* allocate transmit buffer */");
		coder.add(thisClass.getTransmitBufferObjectName());
		coder.add(" = (object_t *) ");
		coder.add(allocTransmitBuffer.getAlias());
		coder.add("(");
		if (llrefs && allocTransmitBuffer.canBlock()) {
			coder.add("pref,");
		}
		coder.add("protocol_obj,");
		coder.add(thisClass.getTransmitBufferSize());
		coder.add(",");
		coder.add(impDefinition.getAllocTimeout());
		coder.addln(");");

		coder.endMethod();
	}

	public ImportDefinition getImportDefinition() {
		return impDefinition;
	}

	@Override
	public <R> R accept(IMMethodVisitor<R> visitor) throws CompileException {
		return visitor.visit(this);
	}

}
