/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.classfile.datatypes.*;

import keso.compiler.*;

import java.util.*;

public class IMSlot implements Comparable<IMSlot> {

	private int stackAlias=-1;
	private int type;
	private int srcAlias;
	private IMBasicBlock checked;
	private IMBasicBlock constBb;
	private IMConstant constNode;
	private String alias;

	private IMNode aliasExpr = null;
	private IMDataFlow dataflowinfo = null;

	private IMNode definition = null;
	private IMBasicBlock definingBB = null;
	private Map<IMNode, IMBasicBlock> uses = null;
	private int ssaCount=0;

	BuilderOptions opts;
	IMMethodFrame frame;

	public IMSlot(IMMethodFrame frame, String alias) {
		this.frame = frame;
		this.opts = frame.getOptions();
		this.srcAlias = -1;
		this.type=BCBasicDatatype.VOID;
		this.alias=alias;
	}

	public IMSlot(IMMethodFrame frame, int type, int alias) {
		this.frame = frame;
		this.opts = frame.getOptions();
		this.srcAlias = alias;
		this.type=type;

		if(this.alias == null) {
			switch (type) {
				case BCBasicDatatype.INT:
					this.alias = "i";
					break;
				case BCBasicDatatype.LONG:
					this.alias = "l";
					break;
				case BCBasicDatatype.FLOAT:
					this.alias = "f";
					break;
				case BCBasicDatatype.DOUBLE:
					this.alias = "d";
					break;
				case BCBasicDatatype.REFERENCE:
					this.alias = "obj";
					break;
				case BCBasicDatatype.BYTE:
					this.alias = "b";
					break;
				case BCBasicDatatype.CHAR:
					this.alias = "c";
					break;
				case BCBasicDatatype.SHORT:
					this.alias = "s";
					break;
				case BCBasicDatatype.BOOLEAN:
					this.alias = "flag";
					break;
				case BCBasicDatatype.RETURN_ADDRESS:
					this.alias = "addr";
					break;
				default:
					this.alias = "unknown";
			}
			this.alias += alias;
			alias++;
		}
	}

	public void registerRef() {
		if (!opts.hasLinkedListOfLocalReferences()) return;
		if (type!=BCBasicDatatype.REFERENCE) return;
		if (srcAlias>=0 && srcAlias<frame.argRange()) return;
		this.alias = "obj["+frame.registerSlot(srcAlias)+"]";
	}

	public String incName() {
		int pos = alias.indexOf("[");

		StringBuffer nname = new StringBuffer();
		if (pos != -1) {
			nname.append(alias.substring(0,pos));
			nname.append("_");
			nname.append(ssaCount);
			nname.append(alias.substring(pos,alias.length()));
		} else {
			nname.append(alias);
			nname.append("_");
			nname.append(ssaCount);
		}

		ssaCount++;

		return nname.toString();
	}

	public void setName(String nalias) {
		alias=nalias;
	}

	public void setDataFlowInfo(IMDataFlow df) {
		dataflowinfo = df;
	}

	public IMDataFlow getDataFlowInfo() {
		return dataflowinfo;
	}

	public void setAlias(IMNode node) {
		aliasExpr = node;
	}

	public IMNode getAlias() {
		return aliasExpr;
	}

	public int getType() {
		return type;
	}

	public boolean isReference() {
		return (type == BCBasicDatatype.REFERENCE);
	}

	// define-use-chain ============================================================

	public void clearDefUseChain() {
		definition=null;
		definingBB=null;
		uses=null;
	}

	public boolean define(IMBasicBlock bb, IMNode statement) throws CompileException {
		if (definition != null) {
			opts.warn("statement " + statement.toReadableString() + " in "
					+ bb);
			opts.warn("already defined " + definition.toReadableString()
					+ " in " + definingBB);
			//throw new CompileException("multiply defined statement!");
			return false;
		}
		redefine(bb, statement);
		return true;
	}

	public void redefine(IMBasicBlock bb, IMNode statement) {
		definition = statement;
		definingBB = bb;
	}

	public IMNode getDefiningStatement() {
		return definition;
	}

	public IMBasicBlock getDefiningBB() {
		return definingBB;
	}

	public Collection<IMNode> getUses() {
		if (uses == null)
			return Collections.emptySet();
		return uses.keySet();
	}

	/**
	 * Die lokale Variable verlaesst den Stack, wenn eine Verwendung
	 * zum Verlassen fuehrt.
	 */
	public boolean doesEscape(boolean isArgument) throws CompileException {
		Collection<IMNode> u = getUses();
		if (u.isEmpty()) {
			opts.vvwarn(frame.getMethod().getAlias()+": doEscape slot "+alias+" without uses.");
			return false;
		}
		for (IMNode nu : u) {
			if (nu.doesEscape(nu, isArgument))
				return true;
		}
		return false;
	}

	public IMNode whyEscape(boolean isArgument) throws CompileException {
		Collection<IMNode> u = getUses();
		if (u.isEmpty())
			return null;
		for (IMNode nu : u) {
			IMNode roe = nu.whyEscape(nu, isArgument);
			if (roe != null)
				return roe;
		}
		opts.warn(frame.getMethod().getAlias()+": slot "+alias+" without escape hint.");
		return null;
	}

	public int numberOfUses() {
		if (uses==null) return 0;
		return uses.size();
	}

	public IMBasicBlock getUseBB(IMNode statement) {
		if (uses==null) return null;
		return uses.get(statement);
	}

	public void use(IMBasicBlock bb, IMNode statement) {
		if (uses == null)
			uses = new HashMap<IMNode, IMBasicBlock>();
		uses.put(statement, bb);
	}

	public void deleteUse(IMNode statement) {
		if (uses==null) {
			opts.warn("var has no uses! "+statement.toReadableString()
					+" in "+frame.getMethod().getAlias());
			return;
		}
		uses.remove(statement);
		if (uses.isEmpty()) {
			frame.getMethod().doAnotherSSARun();
			uses=null;
		}
	}

	public void resetUses() {
		uses=null;
	}

	public boolean hasNoUses() {
		return uses==null;
	}

	// define-use-chain ============================================================

	public boolean sameFrame(IMMethodFrame frame) {
		return this.frame==frame;
	}

	public int getVarNumber() {
		return srcAlias;
	}

	public void setChecked(IMBasicBlock checked) throws CompileException {
		this.checked = checked;
	}

	public boolean isChecked(IMBasicBlock current)  {
		// omit null checks on the this reference
		if(getVarNumber()==0 && !frame.getMethod().isStatic()) {
			return true;
		}

		// uses is not null if ssa is true.
		if (uses!=null) {
			try {
/* REMOVE: I think the dataflowinfo is more powerfull.
				if (opts.hasOption("ssa_buggy_safechk")) {
					// FIXME: Bug: not all uses are checks!
					DominatorTree domtree = frame.getMethod().getDomTree();
					for (IMBasicBlock ub : uses.values()) {
						if (domtree.dom(ub,current)) return true;
					}
				}
*/
				if (definition!=null && definition instanceof IMAStoreLocalVariable) {
					IMAStoreLocalVariable store = (IMAStoreLocalVariable)definition;
					if (store.getOperand().isChecked(definingBB)) return true;
				}
				if (dataflowinfo!=null && dataflowinfo.isValid()) {
					if (checked==null) checked = current;
					return true;
				}
				if (checked==null) return false;
				if (checked==current) return true;
				DominatorTree domtree = frame.getMethod().getDomTree();
				if (domtree.dom(checked, current)) return true;
				return false;
			} catch (CompileException ex) {
				throw new RuntimeException();
			}
		}
		if (checked==null) return false;
		return (checked==current);
	}

	/*
	 * simple constant propagation.
	 */
	public void propagateConstant(IMBasicBlock current, IMConstant const_node) {
		this.constNode = const_node;
		this.constBb = current;
	}

	public IMConstant varToConstant(IMBasicBlock current) {
		if (current==constBb) return constNode;
		return null;
	}

	public void setIsStackVar(int alias) { stackAlias = alias; }

	public boolean isStackVariable(int stack_pos) { return (stackAlias == stack_pos); }

	public String definition() {
		return BuilderOptions.typeToString(type)+" "+this.alias;
	}

	public String cType() {
		return BuilderOptions.typeToString(type);
	}

	@Override
	public String toString() {
		return alias;
	}

	@Override
	public int compareTo(IMSlot o) {
		if (frame != o.frame) {
			IMMethod method  = frame.getMethod();
			IMMethod oMethod = o.frame.getMethod();
			return method.getAlias().compareTo(oMethod.getAlias());
		}
		if (srcAlias != o.srcAlias)
			return srcAlias - o.srcAlias;
		return alias.compareTo(o.alias);
	}

}
