/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.imcode;

import keso.compiler.*;
import keso.compiler.backend.*;

import java.util.*;
import keso.util.IntegerHashtable;

final public class MethodTable {

	final private static Integer JOKER = new Integer(-1);

	private IMMethod top_abstract_method;

	private IMClass	top_class;
	private IMMethod top_method;
	private BuilderOptions opts;
	private IntegerHashtable<IMMethod> candidates;
	private IntegerHashtable<IMMethod> entries;

	private int top_classid;
	private int mt_offset;
	private int table_size = -1;

	/* for index_table only */
	private List<Integer> index_seq;
	private int addr_offset;

	//private final static int max_lift_hight = 64;
	//private final static int max_lift_tab   = 16;
	//private final static int max_lift_hight = 128;
	//private final static int max_lift_tab   = 32;
	private final static int max_lift_hight = 127;
	private final static int max_lift_tab   = 128;

	public MethodTable(IMClass tcls, IMMethod top_abstract_method) {
		this(tcls);
		this.top_abstract_method = top_abstract_method;
	}

	public MethodTable(IMClass tcls) {
		this.top_class=tcls;
		this.top_classid=tcls.getClassTypeID();
		opts=tcls.getOptions();
		candidates = new IntegerHashtable<IMMethod>(64);
		entries    = new IntegerHashtable<IMMethod>(256);
	}

	public void joinTables(MethodTable old) throws CompileException {
		//old.entries=null;
		for (IMMethod method : old.candidates.getElements()) {
			addMethod(method.getIMClass().getClassTypeID(), method);
			method.setMethodTable(this);
		}
		//old.candidates=null;
		//top_class=null;
	}

	/**
	 * add a method to the interface table
	 */
	public void addMethod(int class_id, IMMethod method) throws CompileException {
		if (method==null) throw new CompileException();
		IMClass clazz = method.getIMClass();
		if (class_id<top_classid) {
			top_method=method;
			top_class=clazz;
			top_classid = class_id;
		}
		if (top_method==null) { top_method=method; }
		candidates.put(clazz.getClassTypeID(), method);
		entries.put(class_id, method);
	}

	public void tableIsUsedBy(IMMethod method) {
		//if (top_method==null) top_method=method;
	}

	public int get_mt_offset() {
		return mt_offset;
	}

	public Iterable<IMMethod> getCandidates() {
		return candidates.getElements();
	}

	public boolean hasDispatchTable() {
		if (opts.hasOption("vt_full")) return true;
		return (candidates.size()>1);
	}

	public void emit_invoke_macro(IMMethod current_method, Coder coder) throws CompileException {
		if (!hasDispatchTable()) return;
		IMMethod method = top_method;

		if (top_method!=current_method) {
			coder.header_add("#include \"");
			coder.header_add(top_method.getIMClass().getAlias());
			coder.header_addln(".h\"");
			coder.header_add("/* import ");
			coder.header_add(method.getInvokeAlias());
			coder.header_addln(" */");
			coder.header_add("#define ");
			coder.header_add(current_method.getInvokeAlias());
			coder.header_add(" ");
			coder.header_addln(top_method.getInvokeAlias());
		} else {
			coder.header_add("#define ");
			coder.header_add(method.getInvokeAlias());
			if (opts.hasOption("vt_switch")) {
				coder.header_add("(_self_) ");
				coder.header_add(method.getIdentifier());
				coder.header_addln("_switch");
			} else if (opts.hasOption("vt_comp") && isCompressed()) {
				coder.header_add("(_self_) ((");
				coder.header_add(method.getReturnType());
				coder.header_add(" (*) (");
				coder.header_add(method.getArgString());
				if (is16BitIndex()) {
					coder.header_add("))addr16_table[");
				} else {
					coder.header_add("))addr_table[");
				}
				coder.header_add(addr_offset);
				if (is16BitIndex()) {
					coder.header_add("+index16_table[");
				} else {
					coder.header_add("+index_table[");
				}
				coder.header_add(mt_offset);
				// parity check
				if(opts.hasOption("use_coded_refs")) {
					coder.header_addln("+(NATIVE_ADDR(_self_))->class_id]])");
				} else {
					coder.header_addln("+(_self_)->class_id]])");
				}
				// parity check end
				} else {
				coder.header_add("(_self_) ((");
				coder.header_add(method.getReturnType());
				coder.header_add(" (*) (");
				coder.header_add(method.getArgString());
				coder.header_add("))dispatch_table[");
				coder.header_add(mt_offset);
				// parity check
				if(opts.hasOption("use_coded_refs")) {
					coder.header_addln("+(NATIVE_ADDR(_self_))->class_id])");
				} else {
					coder.header_addln("+(_self_)->class_id])");
				}
				// parity check end
			}
		}
	}

	public int tableSize() {
		if (table_size<0) {
			int[] keys = entries.sortedKeys();
			if (keys.length==0) {
				table_size=0;
			} else {
				table_size=keys[keys.length-1]-top_classid+1;
			}
		}
		return table_size;
	}

	public boolean is16BitIndex() {
		int nc = candidates.size();
		if (nc>254) return true;
		return false;
	}

	public boolean isCompressed() {
		if (opts.hasOption("vt_comp_cond_all")) return true;

		/* more than 254 candidates therefore we need
		 * two bytes per index entry. The compression will only work
		 * if we need more than two bytes for the address!
		 */
		int nc = candidates.size();
		if (nc>254 && opts.targetAddrSize()<4) return false;

		/* Prefer small tables with an abstract top class!
		 * They are more likely to match an existing index table.
		 */
		int ts = tableSize();
		int delta=0;
		if (opts.hasOption("vt_comp_cond_more") && ts < 6) {
			/* Kleine Methodentabellen (<6) bevorzugen */
			if (top_class.isAbstract()) delta=1;
			delta+=(ts>>1);
		}

		/* default:
		 * only compress if the saving is higher than the size
		 * of the index table
		 */
		int iesize = (nc>254?2:1);
		int saving = ((ts-nc)+delta) * opts.targetAddrSize();
		int isize = (iesize*ts)+1;

		if (opts.hasOption("vt_comp_cond_less4")) {
			return (saving > isize+4);
		}

		if (opts.hasOption("vt_comp_cond_less8")) {
			return (saving > isize+8);
		}

		return (saving > isize);
	}

	public boolean emit_call(Coder coder, IMNode self) throws CompileException {
		if (top_method==null) {
			if (candidates.size()==0 && top_class.isInterface()) {
				coder.add("keso_throw_method_not_implemented(\"");
				coder.add(self.getMethod().getAlias());
				coder.add("\",");
				coder.add(self.getBCPosition());
				coder.add(")");
				return false;
			}
			if (candidates.size()==0 && top_class.isAbstract()) {
				coder.add("keso_throw_method_not_implemented(\"");
				coder.add(self.getMethod().getAlias());
				coder.add("\",");
				coder.add(self.getBCPosition());
				coder.add(")");
				return false;
			}
			/* never reached */
			for (IMMethod method : candidates.getElements()) {
				System.err.println(method.getAlias());
			}
			throw new CompileException("top_method == null");
		}
		if (!hasDispatchTable()) {
			coder.add_class(top_method.getIMClass());
			coder.add(top_method.getIdentifier());
			return true;
		} else {
			if (top_class!=null) coder.add_class(top_class.getAlias());
			coder.add(top_method.getInvokeAlias());
			coder.add('(');
			self.translate(coder);
			coder.add(')');
		}
		return true;
	}

	public IMMethod getFirstCandidate() throws CompileException {
		return top_method;
	}

	public IMMethod getTopAbstractMethod() throws CompileException {
		return top_abstract_method;
	}

	public void emit_switch(Coder coder) throws CompileException {
		if (!hasDispatchTable()) return;

		IMMethod m = top_method;
		boolean has_return_value = m.hasReturnValue();

		if (m!=null && m.getMethodNameAndType().equals("finalize()V")) {
			coder.global_header_add("#define KESO_NEED_FINALIZE 1\n");
			coder.global_header_add("#define KESO_INVOKE_FINALIZE(_self_) ");
			coder.global_header_add(m.getIdentifier());
			coder.global_header_add("_switch\n");
		}

		coder.global_header_add(m.getReturnType());
		coder.global_header_add(" ");
		coder.global_header_add(m.getIdentifier());
		coder.global_header_add("_switch(");
		coder.global_header_add(m.getArgString());
		coder.global_header_add(");\n");

		coder.global_add(m.getReturnType());
		coder.global_add(" ");
		coder.global_add(m.getIdentifier());
		coder.global_add("_switch(");
		coder.global_add(m.getArgString());
		coder.global_add(") {\n");
		coder.global_add("\tswitch (obj0->class_id) {\n");

		int[] keys = entries.sortedKeys();
		for (int i=0;i<keys.length;i++) {
			IMMethod method = entries.get(keys[i]);
			coder.global_add("\t\tcase ");
			coder.global_add(keys[i]);
			coder.global_add(":\n\t\t\t");
			if (has_return_value) coder.global_add("return ");
			coder.global_add(method.getIdentifier());
			coder.global_add(method.getArgReCallString());
			coder.global_add(";\n");
			if (!has_return_value) coder.global_add("\t\t\treturn;\n");
		}
		coder.global_add("\t}\n");
		//coder.global_add("\tthrow_error(\"\");\n");
		coder.global_add("}\n\n");
	}

	public int emit_addr_table(Coder coder, int offset) throws CompileException {
		int nc = candidates.size();
		if (!hasDispatchTable()) return offset;

		this.addr_offset = offset;

		IMMethod m = top_method;
		coder.global_add("/* ");
		coder.global_add(m.getMethodNameAndType());
		coder.global_add(" candidates ");
		coder.global_add(nc);
		coder.global_add(" size ");
		int ts = tableSize();
		coder.global_add(ts);
		coder.global_add(" save ");
		int addr_size = opts.targetAddrSize();
		coder.global_add((ts*addr_size)-(nc*addr_size+ts));
		coder.global_add(" */\n");
		int[] keys = candidates.sortedKeys();
		for (int i=0;i<keys.length;i++) {
			IMMethod method = candidates.get(keys[i]);

			if (method.hasCode()) {
				coder.global_add("/* ");
				coder.global_add(offset);
				coder.global_add(" */ \t((code_t)");
				if (method.clearUnused()) {
					coder.global_add("0), /* unused */\n");
				} else {
					coder.global_add(method.getIdentifier());
					coder.global_add("),\n");
				}
				method.setAddrSlot(offset-this.addr_offset);
				offset++;
			}
		}

		return offset;
	}

	private void createIndexSeq() throws CompileException {
		index_seq = new ArrayList<Integer>();

		int[] keys = entries.sortedKeys();
		int top_id = top_class.getClassTypeID();
		for (int i=0;i<keys.length;i++) {
			IMMethod method = entries.get(keys[i]);
			int ccid = keys[i];

			while (top_id!=ccid) {
				/* we have a hole in the method table if the top_id
				 * is not equal the current id. This may happen because
				 * of a interface. */
				index_seq.add(JOKER);
				top_id++;
			}

			if (method.hasCode()) {
				/* we must match the same index number here! */
				index_seq.add(new Integer(method.getAddrSlot()));
			} else {
				/* the value are not of interest! */
				index_seq.add(JOKER);
			}
			top_id++;
		}
	}

	/**
	 * Test if sequence t starting form position pos matches the sequence p
	 */
	private boolean matchSeqFussy(int pos, List<Integer> t, List<Integer> p) {
		for (int i=0;i<p.size();i++) {
			if (pos+i>=t.size()) return true;
			int ti = t.get(pos+i).intValue();
			// values less then 0 are wildcards
			if (ti<0) continue;
			int pi = p.get(i).intValue();
			// values less then 0 are wildcards
			if (pi<0) continue;

			if (ti!=pi) return false;
		}
		return true;
	}

	/**
	 * Test if sequence t starting from position pos matches the sequence p
	 * and compute a quota for every consumed hole.
	 */
	private int matchSeqFussyQuota(int pos, List<Integer> t, List<Integer> p,
			SeqMetaData meta, boolean vt_lift) {
		int quota = 0;
		int i=0;
		int ps = p.size();
		int ts = t.size();
		for (;i<ps;i++) {
			if (pos+i>=ts) {
				/* we don't prefer to extend the table */
				quota += ((pos+(ps-1))-ts)*2;
				break;
			}

			int ti = t.get(pos+i).intValue();
			int pi = p.get(i).intValue();
			if (pi>=0) pi+=meta.lift;

			// values less then 0 are wildcards
			if (ti<0 || pi<0) {
				/* we consuming a hole */
				/* I don't know why but this makes things worse.
				if (ti<0 && pi!=ti) {
					quota++;
					if (true) {
						if (i==0 && pos>0) {
							// prefer to fit the top element
							if (t.elementAt(pos-1).intValue()<0)
								quota++;
						} else if (i+1>=ps && (pos+i+1)<ts) {
							// prefer to fit the last element
							if (t.elementAt(pos+i+1).intValue()<0)
								quota++;
						}
					}
				}
				*/
				quota++;
				continue;
			}

			// fail
			if (ti!=pi) {
				if (vt_lift && (pos+ps)<ts) {
					if (i>0 || pi>ti || ti>max_lift_hight || ps>max_lift_tab) return -1;
					/*  pi <= ti <= 64 */
					/* prefer not to lift the ids */
					quota+=1;
					meta.lift = ti-pi;
				} else {
					/* no reuse */
					return -1;
				}
			}
		}

		return quota;
	}

	/**
	 * Patch the sequence p into t.
	 */
	private void patchSeq(int pos, List<Integer> t, List<Integer> p) {
		for (int i=0;i<p.size();i++) {
			if (pos+i>=t.size()) {
				t.add(p.get(i));
			} else {
				Integer pi = p.get(i);
				if (pi==JOKER) continue;
				t.set(pos+i, pi);
			}
		}
	}

	/**
	 * Patch the sequence p into t.
	 */
	private void patchSeq(int pos, List<Integer> t, List<Integer> p,
			SeqMetaData meta) {
		for (int i=0;i<p.size();i++) {
			Integer pi = p.get(i);
			int piv=pi.intValue();
			if (piv>=0) piv+=meta.lift;
			if (pos+i>=t.size()) {
				t.add(new Integer(piv));
			} else {
				if (pi==JOKER) continue;
				t.set(pos+i, new Integer(piv));
			}
		}
	}

	/**
	 * Include the sequence seq at the first possible position
	 */
	private List<Integer> includeSeq(List<Integer> global, List<Integer> seq)
	throws CompileException {

		for (int i=0;i<global.size();i++) {
			if (matchSeqFussy(i, global, seq)) {
				patchSeq(i, global, seq);
				return global;
			}
		}

		for (int j=0;j<seq.size();j++) {
			global.add(seq.get(j));
		}

		return global;
	}

	class SeqMetaData {
		int lift=0;
		public String toString() {
			if (lift>0)
				return " lift "+Integer.toString(lift);
			return "";
		}
	}

	/**
	 * Include the sequence seq at the position which consumes the minimum of
	 * wildcards (holes).
	 */
	private List<Integer> includeSeqBest(List<Integer> global,
			List<Integer> seq) throws CompileException {

		int best=-1;
		int best_quota=seq.size()+1;
		boolean vt_comp_X = opts.hasOption("vt_comp_X");
		SeqMetaData meta = new SeqMetaData();
		SeqMetaData best_meta = null;

		for (int i=0;i<global.size();i++) {

			meta.lift=0;
			int quota = matchSeqFussyQuota(i, global, seq, meta, vt_comp_X);
			if (quota<0) continue;

			int end_pos = i + seq.size();
			if (end_pos>=global.size()) {
				if (best>=0) break;
			}

			if (quota<best_quota) {
/*
				if (best>0) {
					opts.warn("vt_comp_best: position "+best+"->"+i+" size "+Integer.toString(seq.size())+" old quota "+best_quota+" new quota "+quota+" "+meta);
				}
*/
				best = i;
				best_quota = quota;
				best_meta = meta;
			}
		}

		if (best>=0) {
			patchSeq(best, global, seq, best_meta);
			return global;
		}

		for (int j=0;j<seq.size();j++) {
			global.add(seq.get(j));
		}

		return global;
	}


	public List<Integer> updateIndexSeq(List<Integer> current) throws CompileException {
		/* nothing todo */
		if (!hasDispatchTable()) return current;

		if (index_seq==null) createIndexSeq();

		if (current==null) return index_seq;

		if (opts.hasOption("vt_comp_fast")) {
			/*
			 * don't include, just append the new sequence
			 */
			for (int j=0;j<index_seq.size();j++) {
				current.add(index_seq.get(j));
			}
			return current;
		}

		if (index_seq.size()>current.size()) {
			/*
			 * include smaller sequence in larger one.
			 */
			if (opts.hasOption("vt_comp_best")) {
				current=includeSeqBest(index_seq, current);
			} else {
				current=includeSeq(index_seq, current);
			}
			index_seq=null;
			return current;
		}

		if (opts.hasOption("vt_comp_best")) {
			return includeSeqBest(current, index_seq);
		}

		return includeSeq(current, index_seq);
	}

	static private int findSeq(List<Integer> t, List<Integer> p, boolean vt_lift) {

		/*
		 * TODO: Such-Algortihmus optimieren.
		 *
		 * z.B.
		 * Knuth-Morris-Pratt-Algorithmus
		 * http://www.iti.fh-flensburg.de/lang/algorithmen/pattern/kmp.htm
		 *
		 * Boyer-Moore-Algorithmus
		 * http://www.iti.fh-flensburg.de/lang/algorithmen/pattern/bm.htm
		 * Wohl eher ungeeignet, da der Symbol-Vorrat sehr klein ist.
		 *
		 */

		int n = t.size() - (p.size()-1);
		int m = p.size();
		for (int j=0;j<n;j++) {
			boolean found = true;
			int lift = 0;
			for (int i=0;i<m;i++) {
				int ti = t.get(j+i).intValue();
				int pi = p.get(i).intValue();
				if (pi<0) continue;
				pi+=lift;
				if (ti!=pi) {
					if (vt_lift) {
						if (i>0 || pi>ti || ti>max_lift_hight || p.size()>max_lift_tab) {
							/*  pi <= ti <= 64 */
							lift = ti-pi;
							continue;
						}
					}
					found = false;
					break;
				}
			}
			if (found) {
				return j;
			}
		}
		System.err.println("global:");
		dump_seq(t);
		dump_seq(p);
		throw new RuntimeException("seq not found!");
	}

	static private void dump_seq(List<Integer> t) {
		for (Integer i : t) {
			if (i==JOKER) {
				System.err.print("?");
			} else {
				System.err.print(i);
			}
			System.err.print(",");
		}
		System.err.print("\n");
	}

	public void updateOffset(List<Integer> current) throws CompileException {
		/* nothing todo */
		if (!hasDispatchTable()) return;
		if (index_seq==null) createIndexSeq();

		int start_off = findSeq(current, index_seq, opts.hasOption("vt_comp_X"));
		this.mt_offset = start_off-top_class.getClassTypeID();

		return;
	}

	public int emit_table(Coder coder, int offset) throws CompileException {
		int nc = candidates.size();

		if (!hasDispatchTable()) return offset;

		IMMethod m = top_method;
		coder.global_add("/* ");
		if (m==null) {
			coder.global_add("null");
		} else {
			coder.global_add(m.getMethodNameAndType());
			coder.global_add(" \t off:");
			coder.global_add(offset);
			coder.global_add(" cls:");
			coder.global_add(top_class.getClassTypeID());
			coder.global_add(" len:");
			int ts = tableSize();
			coder.global_add(ts);
			coder.global_add(" candidates:");
			coder.global_add(nc);
		}
		coder.global_add(" */\n");

		int top_id = top_class.getClassTypeID();
		this.mt_offset = offset - top_id;

		if (m!=null && m.getMethodNameAndType().equals("finalize()V")) {
			coder.global_header_add("#define KESO_NEED_FINALIZE 1\n");
			coder.global_header_add("#define KESO_INVOKE_FINALIZE(_self_) ((");
			coder.global_header_add(m.getReturnType());
			coder.global_header_add(" (*) (");
			coder.global_header_add(m.getArgString());
			coder.global_header_add("))dispatch_table[");
			coder.global_header_add(mt_offset);
			if(opts.hasOption("use_coded_refs")) {
				coder.global_header_add("+(NATIVE_ADDR(_self_))->class_id])\n");
			} else {
			coder.global_header_add("+(_self_)->class_id])\n");
			}
		}

		int[] keys = entries.sortedKeys();
		for (int i=0;i<keys.length;i++) {
			IMMethod method = entries.get(keys[i]);
			int ccid = keys[i];

			while (top_id!=ccid) {
				coder.global_add("/* ");
				coder.global_add(offset);
				coder.global_add(" c");
				coder.global_add(top_id);
				if (top_id<10) coder.global_add(" ");
				coder.global_add(" */ \t((code_t)0), /* hole */\n");
				top_id++;
				offset++;
			}

			coder.global_add("/* ");
			coder.global_add(offset);
			coder.global_add(" c");
			coder.global_add(top_id);
			if (top_id<10) coder.global_add(" ");
			coder.global_add(" */ \t((code_t)");

			if (method.hasCode() && !method.clearUnused()) {
				coder.global_add(method.getIdentifier());
				coder.global_add("),\n");
			} else {
				coder.global_add("0),\n");
			}
			top_id++;
			offset++;
		}

		return offset;
	}

	public String toString() {
		StringBuffer msg = new StringBuffer("top ");
		msg.append(top_class);
		msg.append(" id ");
		msg.append(top_classid);
		msg.append(" offset ");
		msg.append(mt_offset);
		msg.append(" method ");
		if (top_method!=null) { msg.append(top_method.getMethodNameAndType()); }
		else { msg.append("null"); }
		return msg.toString();
	}
}
