/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.kni;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.backend.*;

import java.util.*;

final public class JoinPointChecker extends Weavelet {
	private Collection<Weavelet> weavelets;

	// previously computed weavelets affecting a particular joinpoint
	private Map<String, Weavelet[]> jp_cache
	        = new HashMap<String, Weavelet[]>();

	private Weavelet[] NOT_AFFECTED = new Weavelet[0];

	public JoinPointChecker(BuilderOptions opts, ClassStore repository) {
		super(opts,repository);
	}

	final public void reset() {
		weavelets = new ArrayList<Weavelet>();
		jp_cache = new HashMap<String, Weavelet[]>();
	}

	final public void registerWeavelet(Weavelet weavelet) {
		String[] pointcuts = weavelet.joinPoints;
		if (pointcuts==null || pointcuts.length<1) {
			opts.verbose("warning: Weavelet has no pointcuts! "+weavelet.toString());
			return;
		}

		if (jp_cache.size()>0) {
			opts.verbose("warning: register weavelet after usage!");
			jp_cache = new HashMap<String, Weavelet[]>();
		}

		weavelets.add(weavelet);
	}

	private Weavelet[] getJoinPoints(String className, String methodName) {
		String key = className+"."+methodName;

		// try cache first
		Weavelet[] weavelet = jp_cache.get(key);
		if (weavelet!=null) return weavelet;

		Collection<Weavelet> matches = new ArrayList<Weavelet>();

		for(Weavelet wv: weavelets) {
			if(wv.affects(key))
				matches.add(wv);
		}
		if(matches.size() == 0) {
			weavelet = NOT_AFFECTED;
		} else {
			weavelet = matches.toArray(NOT_AFFECTED);
		}

		if (weavelet!=NOT_AFFECTED) {
			opts.vverbose("jp hits: "+key+" "+weavelet.length);
		}

		jp_cache.put(key, weavelet);
		return weavelet;
	}

	public boolean ignoreMethodBody(IMClass clazz, IMMethod method) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(clazz.getClassName(), method.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.ignoreMethodBody(clazz,method)) flag=true;
		}

		return flag;
	}

	public void require(int domainID, String className, String methodName)
	{
		Weavelet[] weavelet = getJoinPoints(className, methodName);
		if (weavelet==NOT_AFFECTED) return;

		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			a.require(domainID, className, methodName);
		}
	}

	public java.util.Set<IMMethod> keepMethods() {
		final HashSet<IMMethod> km = new HashSet<IMMethod>();

		for(Weavelet a: weavelets) {
			km.addAll(a.keepMethods());
		}
		return km;
	}


	public int checkAttribute(IMMethod self, int attr) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(self.getClassName(), self.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return NOTDEF;

		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			opts.vvverbose("checkAttribute: "+a.toString()+" "+attr);
			if (a.checkAttribute(self, attr)==TRUE) return TRUE;
			opts.vvverbose("checkAttribute: FALSE");
		}

		return FALSE;
	}

	public boolean addFields(IMClass clazz, Coder coder, StringBuffer raw_fields) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(clazz.getClassName(), "<CLASS>");
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.addFields(clazz,coder,raw_fields)) flag=true;
		}

		return flag;
	}

	public int checkFieldAttribute(String classname, String fieldname, int attr)
	{
		Weavelet[] weavelet = getJoinPoints(classname, "<CLASS>");
		if (weavelet==NOT_AFFECTED) return NOTDEF;

		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			opts.vvverbose("checkAttribute: "+a.toString()+" "+attr);
			if (a.checkFieldAttribute(classname,fieldname, attr)==TRUE) return TRUE;
			opts.vvverbose("checkAttribute: FALSE");
		}

		return FALSE;
	}

	public IMNode affectIMInvoke(IMInvoke self, IMMethod method, IMMethod callee,
			IMNode obj, IMNode args[]) throws CompileException {

		Weavelet[] weavelet = getJoinPoints(callee.getClassName(), callee.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return self;

		IMNode nself = self;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			nself = a.affectIMInvoke((IMInvoke)self, method, callee, obj, args);
			if (nself==null) {
				nself = self;
				continue;
			}
			if (!(nself instanceof IMInvoke)) break;
		}

		return nself;
	}

	public boolean affectMethod(IMClass clazz, IMMethod method, Coder coder) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(clazz.getClassName(), method.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.affectMethod(clazz,method,coder)) flag=true;
		}

		return flag;
	}


	public boolean affectInvokeStatic(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode args[], Coder coder) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(callee.getClassName(), callee.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.affectInvokeStatic(node,caller,callee,args,coder)) flag=true;
		}

		return flag;
	}

	public boolean affectInvokeSpecial(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode obj, IMNode args[], Coder coder) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(callee.getClassName(), callee.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.affectInvokeSpecial(node,caller,callee,obj,args,coder)) flag=true;
		}

		return flag;
	}

	public boolean affectInvokeVirtual(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode obj, IMNode args[], Coder coder) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(callee.getClassName(), callee.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) return false;

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.affectInvokeVirtual(node,caller,callee,obj,args,coder)) flag=true;
		}

		return flag;
	}

	public boolean affectInvokeInterface(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode obj, IMNode args[], Coder coder) throws CompileException
	{
		Weavelet[] weavelet = getJoinPoints(callee.getClassName(), callee.getMethodNameAndType());
		if (weavelet==NOT_AFFECTED) {
			return false;
		}

		boolean flag = false;
		for (int i=0;i<weavelet.length;i++) {
			Weavelet a = weavelet[i];
			if (a.affectInvokeInterface(node,caller,callee,obj,args,coder)) flag=true;
		}

		return flag;
	}
}
