/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.kni;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.backend.*;

import keso.compiler.config.*;

import java.util.*;

import java.lang.Integer;

public abstract class KesoIndexedResource<RT extends SystemObject<?>> extends Weavelet implements TranslationCallback {

	protected IMInvoke node;
	protected IMClass clazz;
	protected IMMethod caller;
	protected IMMethod callee;
	protected IMNode args[];
	protected Coder coder;

	/**
	 * Maps system object names to the position in the lookup index.
	 */
	private Map<String, Integer> idmap;

	/**
	 * This set contains the name of all existing resources.
	 */
	private java.util.Set<String> existingResources;

	protected KesoIndexedResource(BuilderOptions opts, ClassStore repository) {
		super(opts, repository);

		idmap = new HashMap<String, Integer>();
		repository.registerGlobalTranslationCallback("KesoIndexedResource_" + resCategory(), this);
	}

	/**
	 * Generates the code to lookup an indexed resource from the domain descriptor.
	 *
	 * @param accessibleResources   array that contains, for each domain, the reachable resources
	 */
	protected void getIndexedResourceByName() throws CompileException {
		if(existingResources == null) {
			existingResources = new HashSet<String>();
			for(Map<String, ? extends RT> ht : accessibleResources().values()) {
				for(String key: ht.keySet()) {
					existingResources.add(key);
				}
			}
		}

		// determine the name of the resource that is queried
		final String resName = assert_string(args[0], "get"+resCategory()+"ByName(): name");
		if (resName == null)
			throw new CompileException("get"+resCategory()+"ByName(): first argument must not be null!");

		// check if the resource exists
		if (!existingResources.contains(resName)) {
			opts.warn("get"+resCategory()+"ByName(): " +resName+ " is not a known "+resCategory()+". "+node.lineInfo());
			coder.add("NULL");
			return;
		}

		// check if the lookup can be resolved statically
		boolean staticSuccess = true, staticFail = true;
		RT so = null;

		for(IMDomain dom: coder.getCurrentDomains()) {
			Map<String,? extends RT> lr = accessibleResources().get(dom);

			if(lr.get(resName) == null) {
				staticSuccess = false;
			} else {
				staticFail = false;
				so = lr.get(resName);
			}
		}

		if(staticFail) {
			// the lookup will always fail; also catches the case that the lookup
			// is not reachable from any domain

			// warn if the lookup fails only if the code is reachable from somewhere
			if(coder.getCurrentDomains().size() > 0) {
				opts.warn("static failure of " + resCategory() + " lookup: " + resName + " in " + node.lineInfo());
				for(IMDomain dom: coder.getCurrentDomains()) {
					opts.vverbose("\treachable from " + dom.getDef());
				}
			}
			coder.add("NULL");

		} else if(staticSuccess) {
			// the lookup will always succeed, since the lookup() is only reachable
			// from domains that may access the resource
			coder.add(coder.add_accessStaticObject(resClass(so), coder.getResObjSymbol(so)));
			opts.vverbose("static success of " + resCategory() + " lookup: " + resName + " in " + node.lineInfo());

		} else {
			opts.vverbose("dynamic lookup of " + resCategory() + " lookup: " + resName + " in " + node.lineInfo());
			// we really need a dynamic lookup
			Integer resRow = idmap.get(resName);

			if (resRow == null) {
				resRow = new Integer(idmap.size());
				// assign a new index in the lookup field
				idmap.put(resName, resRow);
			}
			coder.add("KESO_CURRENT_DOMAIN->");
			coder.add(resCategory().toLowerCase());
			coder.add("_lookup[");
			coder.add( resRow );
			coder.add("]");
		}
	}

	/**
	 * Add code for a function that returns StatusType and gets one reference to
	 * a Resource object that is mapped to the corresponding OSEK resource which
	 * is then used as the parameter to the OSEK call <fnname>.
	 */
	protected void retIgetResRef(String fnname, String comment, String resClass,
			String osekIDField, IMMethod method) throws CompileException {
		IMClass res_class = repository.getClass(resClass);
		coder.add_class(res_class);

		coder.add("/* ");
		coder.add(comment);
		coder.addln("*/");

		coder.add("return (jint) ");
		coder.add(fnname);
		coder.add("(");
		coder.add_getField(res_class,method, method.getMethodArgument(0).toString(),osekIDField, "BGETFIELD");
		coder.addln(");");
	}

	public boolean ignoreMethodBody(IMClass clazz, IMMethod method) throws CompileException {
		return true;
	}

	protected abstract String resCategory();
	protected abstract IMClass resClass(RT so);
	protected abstract Map<IMDomain, ? extends Map<String,? extends RT>> accessibleResources() throws CompileException;

	public void translate_global(Coder coder) throws CompileException {
		if(idmap.size() ==0) return; // no dynamic lookups needed

		String[] ids = new String[idmap.size()];
		for(String id: idmap.keySet()) {
			ids[ idmap.get(id).intValue() ] = id;
		}

		for(IMDomain imd: opts.getDomains()) {
			StringBuffer value = new StringBuffer("{ ");

			for(String resName: ids) {
				value.append("/* ");
				value.append(resName);
				value.append(" */ ");

				RT so = accessibleResources().get(imd).get(resName);

				if(so==null) { // not accessible
					value.append("NULL");
				} else { // accessible
					value.append(coder.add_accessStaticObject(resClass(so), coder.getResObjSymbol(so)));
					imd.addUsedModule(resClass(so));
				}
				value.append(",");
			}
			value.setCharAt(value.length()-1, '}');
			imd.addDescriptorField("object_t* KESO_CONST ",
					resCategory().toLowerCase() + "_lookup[" + ids.length + "]",
					value.toString(), false);
		}
	}
}
