/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.kni;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.backend.*;

public class Memory extends Weavelet {

	private Coder coder;
	private IMMethod callee;
	private IMNode[] args;
	private IMNode obj;

	private StringBuffer fn;
	private String op, ptrType;
	private int asize;

	public Memory(BuilderOptions opts, ClassStore repository) {
		super(opts, repository);
		joinPoints = new String[] { "keso/core/Memory\\..+","keso/core/Memory\\.<CLASS>" };
	}

	public void require(int domainID, String className, String methodName) {
		repository.requireClass(domainID, "keso/core/Memory");
	}

	/**
	 * Use this method to add fields to the affected class.
	 * See the Task weavelet on an example on how to use this. The added fields are
	 * for internal use only, they are not visible to the application.
	 *
	 * The fields will be appended to the regular object fields, no matter if they
	 * are reference fields or not. The class store will contain the size
	 * including the added fields, roff will not be affected since the fields are
	 * put behind the object header.
	 * This especially implicates that reference fields added with this method
	 * will _not_ be scanned by the garbage collector. Therefore you should not
	 * keep references to mortal objects in added fields, as they may be garbage
	 * collected in case they are not visible by the application anymore.
	 *
	 * @param clazz The intermediate representation of the affected class
	 * @param coder Coder object that can be used to generate code for the class
	 * @param raw_fields raw access  for adding C code into the object layout.
	 *
	 */
	public boolean addFields(IMClass clazz, Coder coder, StringBuffer raw_fields) throws CompileException {
		clazz.addRawField("addr", opts.targetAddrType());
		raw_fields.append("\t");
		raw_fields.append(opts.targetAddrType());
		raw_fields.append(" _addr;\n\t");

		clazz.addRawField("size", "jint");
		raw_fields.append("jint _size;\n");

		return true;
	}

	/**
	 * The checkAttribute method changes the attributes of a method.
	 * The default value for all attributes of all affected methods is FALSE.
	 */
	public int checkAttribute(IMMethod method, int attr) throws CompileException {
		if (attr==ATTR_OMIT_METHOD) {
			fn = new StringBuffer(method.getMethodNameAndType());
			getOperator();
			getAccSize();
			if(op!=null && asize!=-1)
				return TRUE;
		} else if (attr==ATTR_PURE) {
			if(method.termed("getSize()I")) {
				return TRUE;
			}
		}
		return FALSE;
	}

	private void getOperator() {
		op = fn.toString();
		int len=3;
		if (op.startsWith("get")) {
			op = "";
		} else if (op.startsWith("set")) {
			op = "=";
		} else if (op.startsWith("and")) {
			op = "&=";
		} else if (op.startsWith("or")) {
			op = "|="; len=2;
		} else if (op.startsWith("xor")) {
			op = "^=";
		} else {
			op=null;
			return;
		}
		fn.delete(0, len);
	}

	private void getAccSize() {
		int len=2;
		asize=-1;
		String size = fn.toString();

		if(size.startsWith("8")) {
			len=1; asize=1; ptrType="jchar";
		} else if(size.startsWith("16")) {
			asize=2; ptrType = "jshort";
		} else if(size.startsWith("32")) {
			asize=4; ptrType = "jint";
		} else if(size.startsWith("Float")) {
			asize=4; ptrType = "jfloat";
		} else return;

		fn.delete(0, len);
	}

	public boolean affectMethod(IMClass clazz, IMMethod method, Coder coder) throws CompileException {

		this.coder = coder;

		if (method.getMethodNameAndType().equals("getPart(II)Lkeso/core/Memory;")) {
			getPart(method);
			return true;
		}

		return false;
	}


	public boolean affectInvokeVirtual(IMInvoke node, IMMethod caller,
			IMMethod callee, IMNode obj, IMNode args[], Coder coder)
		throws CompileException {

			this.coder = coder;
			this.callee = callee;
			this.args = args;
			this.obj = obj;

			if (callee.getMethodNameAndType().equals("getSize()I")) {
				getSize();
				return true;
			}

			fn = new StringBuffer(callee.getMethodNameAndType());


			getOperator();
			getAccSize();

			if(op==null || asize==-1)
				return false;

			coder.add_before("\t/* MEMORY ACCESS */\n");
			String type = "MEMLOAD";
			if(op.length() > 0) type = "MEMSTORE";
			coder.chk_range(obj, args[0], asize, caller, node.getBCPosition(), type);
			accessMemlocation();
			if(!op.equals("")) {
				coder.add(op);
				args[1].translate(coder);
			}

			return true;
		}


	private void getSize() throws CompileException {
		opts.verbose("####Memory.getSize()");

		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);
		coder.add_getField(memory_class, obj, "size");
	}

	private void getPart(IMMethod method) throws CompileException {
		opts.verbose("####Memory.getPart()");

		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		// check memory parameter
		coder.addln("if ("+ method.getMethodArgument(0).toString() +"== NULL) return NULL;");

		// check offset
		coder.add("if (0 > "+method.getMethodArgument(1).toString() + " || "+ method.getMethodArgument(1).toString() + " >= ");
		coder.add_getField_fast(memory_class, method.getMethodArgument(0).toString(), "size", "IGETFIELD");
		coder.addln(") return NULL;");

		// check size
		coder.add("if (0 > "+method.getMethodArgument(2).toString() + " || " + method.getMethodArgument(2).toString() + " > (");
		coder.add_getField_fast(memory_class,method.getMethodArgument(0).toString(), "size", "LGETFIELD");
		coder.addln(" - "+method.getMethodArgument(1).toString() +")) return NULL;");


		coder.add("object_t* newMem = ");
		coder.add_allocObject(memory_class);
		coder.addln(";");
		coder.add_getField(memory_class, method, "newMem", "addr", "LGETFIELD");
		coder.add(" = (");
		coder.add_getField_fast(memory_class,method.getMethodArgument(0).toString(), "addr", "LGETFIELD");
		coder.addln(" + " + method.getMethodArgument(1).toString() + ");");

		// omit 2nd null reference check
		coder.add_getField_fast(memory_class, "newMem", "size", "IGETFIELD");
		coder.addln(" = "+method.getMethodArgument(2).toString()+";");

		coder.addln("return newMem;");
	}


	private void accessMemlocation() throws CompileException {
		IMClass mem_class = callee.getIMClass();

		coder.add("*(");
		coder.add("volatile ");
		coder.add(ptrType);
		coder.add(" *) ( (char *) ");
		if (obj instanceof IMAMemConstant) {
			IMAMemConstant const_mem = (IMAMemConstant)obj;
			coder.add("0x");
			coder.add(Integer.toHexString(const_mem.getAddr()));
		} else {
			coder.add_getField(mem_class, obj, "addr");
		}
		coder.add(" + ");

		if (args[0] instanceof IMIConstant) {
			coder.add(((IMIConstant)args[0]).getIntValue());
		} else {
			args[0].translate(coder);
		}
		coder.add(")");
	}

	public boolean ignoreMethodBody(IMClass clazz, IMMethod method)
		throws CompileException {
			return true;
		}
}
