/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.kni;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.backend.*;

public class MemoryService extends Weavelet {
	private Coder coder;
	private IMMethod callee;
	static private int identifierCounter=0;

	public MemoryService(BuilderOptions opts, ClassStore repository) {
		super(opts, repository);
		joinPoints = new String[] { "keso/core/MemoryService\\..+" };
	}

	public void require(int domainID, String className, String methodName) {
		repository.requireClass(domainID, "keso/core/Memory");
	}

	public int checkAttribute(IMMethod method, int attr) {
		if (attr==ATTR_NOTNULL) {
			if (method.termed("allocStaticDeviceMemory(II)Lkeso/core/Memory;"))
				return TRUE;
			if (method.termed("allocStaticMemoryHandle()Lkeso/core/Memory;"))
				return TRUE;
			if (method.termed("allocStaticMemory(I)Lkeso/core/Memory;"))
				return TRUE;
			if (method.termed("mapStaticDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;"))
				return TRUE;

			// FIXME many dynamic variants are NOTNULL as well, since allocation failure
			//       will result in an OutOfMemoryException, not a returned null reference
		} else if(attr == ATTR_PURE) {
			if (method.termed("allocStaticDeviceMemory(II)Lkeso/core/Memory;")
					|| method.termed("allocStaticMemoryHandle()Lkeso/core/Memory;")
					|| method.termed("allocStaticMemory(I)Lkeso/core/Memory;")
					|| method.termed("mapStaticDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")
				) {
				return TRUE;
			}
		}

		return FALSE;
	}

	/**
	 * This method contains optimized versions of some MemoryService methods if
	 * the participating memory objects are statically known.
	 *
	 * In that case, we will directly insert constant objects in the intermediate
	 * code, affectInvokeStatic will thus _NOT_ be invoked in these cases anymore.
	 * If the prerequisites for the optimized version are not given, we do not touch
	 * the node and create a dynamic version in affectInvokeStatic.
	 */
	public IMNode affectIMInvoke(IMInvoke self, IMMethod method, IMMethod callee,
			IMNode obj, IMNode args[]) throws CompileException {

		if (callee.termed("allocMemory(I)Lkeso/core/Memory;")) {
			if( !args[0].isConstant() || ( (IMIConstant) args[0]).getValue() != 0 ) {
				repository.requireSharedMemory();
			}
			return self;
		} else if (callee.termed("allocStaticDeviceMemory(II)Lkeso/core/Memory;")) {
			IMIConstant addr = assert_iconst(args[0], "Parameters to mapStaticDevMem");
			IMIConstant size = assert_iconst(args[1], "Parameters to mapStaticDevMem");
			return method.createMemConstant(self, addr.getValue(), size.getValue());

		}	else if (callee.termed("allocStaticMemory(I)Lkeso/core/Memory;")) {
			if (!(args[0] instanceof IMIConstant)) return self;
			// TODO: implement
			//IMIConstant size = (IMIConstant)args[1];
			//return method.createMemConstant(self, addr.getValue(), size.getValue());
			return self;

		} else if (callee.termed("mapStaticDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")
				|| callee.termed("mapDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")) {

			final IMIConstant addr = assert_iconst(args[0], "Parameters to map*DeviceMemory");
			final String type = assert_string(args[1], "Parameters to map*DeviceMemory");

			final IMClass memory_type_class = repository.getClass(type);
			if (memory_type_class==null)
				throw new CompileException("Class \""+type+"\" not found! ("+method.getIMClass().getSourceFile()+")");
			final IMMethod constructor = memory_type_class.getMethod("<init>()V");
			if(constructor == null)
				throw new CompileException("the memory-mapped class " + memory_type_class + " does not provide a default constructor <init>()V");
			keep_methods.add(constructor);

			if(callee.termed("mapStaticDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;"))
				return method.createMTStaticConstant(self, memory_type_class, addr.getValue());
			return method.createMTHeapConstant(self, memory_type_class, addr.getValue());

		} else if (callee.termed("mapMemoryToStaticObject(Lkeso/core/Memory;ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")
				|| callee.termed("mapMemoryToObject(Lkeso/core/Memory;ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")) {

			if (args[0] instanceof IMNullConstant) {
				opts.warn("mapMemoryTo*Object memory object is null pointer!" + self.lineInfo());
				return args[0];
			}

			final String type = assert_string(args[2],"Parameter type to mapMemoryToStaticObject");

			// keep default constructor of mapped class
			final IMClass memory_type_class = repository.getClass(type);
			if (memory_type_class==null)
				throw new CompileException("Class \""+type+"\" not found! ("+method.getIMClass().getSourceFile()+")");
			final IMMethod constructor = memory_type_class.getMethod("<init>()V");
			if(constructor == null)
				throw new CompileException("the memory-mapped class " + memory_type_class + " does not provide a default constructor <init>()V");
			keep_methods.add(constructor);

			/*
			 * If we have constant arguments, we replace the method invocation by a MTConstant
			 * in the intermediate representation. Otherwise, we will set the range dynamically
			 * in the code.
			 */
			if (!(args[0] instanceof IMAMemConstant)) { return self; }
			if (!(args[1] instanceof IMIConstant)) { return self; }

			final IMAMemConstant memory = (IMAMemConstant) args[0];
			final int addr   = memory.getAddr();
			final int offset = assert_iconst(args[1],"Parameter offset to mapMemoryToStaticObject").getIntValue();

			// for constant methods, check if the destination region is large enough for the mapping
			if(memory_type_class.mappedSize() > (memory.getSize()-offset))
				throw new CompileException("target memory region is too small to create mapping ("+method.getIMClass().getSourceFile()+")");

			if(callee.termed("mapMemoryToStaticObject(Lkeso/core/Memory;ILjava/lang/String;)Lkeso/core/MemoryMappedObject;"))
				return method.createMTStaticConstant(self, memory_type_class, addr+offset);
			return method.createMTHeapConstant(self, memory_type_class, addr+offset);
		}

		return self;
	}

	public boolean affectMethod(IMClass clazz, IMMethod method, Coder coder) throws CompileException {
		this.coder = coder;
		this.callee = method;

		if (method.termed("allocDynamicDeviceMemory(II)Lkeso/core/Memory;")) {
			allocDynDevMem(method);
		} else if (method.termed("allocMemory(I)Lkeso/core/Memory;")) {
			allocDynMem(method);
		} else if (callee.termed("adjustMemory")) {
			adjustMem(method);
		} else if (callee.termed("releaseMemory(Lkeso/core/Memory;)V")) {
			releaseMem();
		} else if (callee.termed("copy(Lkeso/core/Memory;ILkeso/core/Memory;II)V")) {
			copyMem(method);
		} else return false;
		return true;
	}

	public boolean affectInvokeStatic(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode args[], Coder coder) throws CompileException {
		this.callee = callee;
		this.coder = coder;

		if (callee.termed("allocStaticDeviceMemory(II)Lkeso/core/Memory;")) {
			allocStaticDevMem(args);
		} else if (callee.termed("mapStaticDeviceMemory(ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")) {
			mapStaticDevMem(args);
		} else if (callee.termed("mapMemoryToStaticObject(Lkeso/core/Memory;ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")) {
			mapMemToStaticObject(node,args,caller);
		} else if (callee.termed("mapMemoryToObject(Lkeso/core/Memory;ILjava/lang/String;)Lkeso/core/MemoryMappedObject;")) {
			mapMemToObject(node,args,caller);
		} else if (callee.termed("allocStaticMemoryHandle()Lkeso/core/Memory;")) {
			coder.add_allocStaticMem(0);
		} else if (callee.termed("allocStaticMemory(I)Lkeso/core/Memory;")) {
			int size = assert_iconst(args[0],"Parameter to allocStaticMem").getValue();
			if (!opts.is8BitController()) {
				// align by 4 bytes
				size = (size+3)&~3;
			}
			coder.add_allocStaticMem(size);

		} else if (callee.termed("adjustMemory(Lkeso/core/Memory;ILkeso/core/Memory;I)V")) {
			if (coder.getCodeInfo().isConstant(args[0])) {
				// FIXME
				opts.warn("adjustMemory() is broken if the adjusted Memory is a constant. The generated code will access the Memory before the adjustment." + node.lineInfo());
			}
		} else return false;

		return true;
	}

	private void allocStaticDevMem(IMNode[] args) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		IMIConstant addr = assert_iconst(args[0], "Parameters to allocStaticDeviceMemory");
		IMIConstant size = assert_iconst(args[1], "Parameters to allocStaticDeviceMemory");

		coder.add_allocConstObject(memory_class, "staticDevMem"+(identifierCounter++),null);
		coder.add_init_field_hex(memory_class,"addr", "", addr.getValue());
		coder.add_init_field(memory_class,"size", size.toReadableString());
		coder.add_init_end();
	}

	private void mapStaticDevMem(IMNode[] args) throws CompileException {

		String type = assert_string(args[1],"Parameters to mapStaticDevMem");
		IMClass memory_type_class = repository.getClass(type);

		if (memory_type_class==null)
			throw new CompileException("Class \""+type+"\" not found!");

		IMIConstant addr = assert_iconst(args[0],"Parameters to mapStaticDevMem");

		coder.add_class(memory_type_class);

		coder.add_allocConstObject(memory_type_class, "staticDevObj"+(identifierCounter++),null);
		coder.add_init_field_hex(memory_type_class, "base", "(void *)", addr.getValue());
		coder.add_init_end();
	}

	private void mapMemToStaticObject(IMInvoke node, IMNode[] args, IMMethod caller) throws CompileException {
		final String linfo = caller.getClassName() +"." + caller.getMethodName() + ", " + node.lineInfo();
		final String type = assert_string(args[2],"Parameters to mapMemoryToStaticObject");

		final IMClass memory_type_class = repository.getClass(type);
		if (memory_type_class==null)
			throw new CompileException("Class \""+type+"\" not found!");
		coder.add_class(memory_type_class);

		// we need to choose the right symbol prefix since the memory mapped object
		// is manipulated at runtime and needs to be write accessible from the
		// corresponding domain.
		//
		// currently, we require that the domain is unique for each mapping; if it
		// were not unique, we would need to create on object for each domain and
		// dynamically decide which one to use from the current domain
		String mobj = "staticMemMappedObj"+(identifierCounter++);

		IMDomain dom = coder.getCurrentDomain();
		if(dom == null) {
			opts.warn("map memory unreachable: " + linfo);
		} else {
			mobj = coder.getDomainSymbolPrefix(dom.getDef()) + mobj;
		}

		coder.add_allocConstObject(memory_type_class, mobj,null);
		coder.add_init_end();

		final IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		if (args[0] instanceof IMAMemConstant) {
			final int base = ((IMAMemConstant)args[0]).getAddr();
			final int size = ((IMAMemConstant)args[0]).getSize();
			coder.add_before("\n");

			// memory object is known so we don't need a null check here
			caller.boundsCheckKnownIndex++;
			coder.add_before("\tKESO_CHECK_ARRAY_KNOWNBOUNDS(");
			coder.add_before(size);
			coder.add_before(",(jint) (");
			coder.add_before(args[1].toReadableString()); // offset
			coder.add_before("+sizeof(");
			coder.add_before(memory_type_class.getAlias());
			coder.add_before("_mt) - 1");
			coder.add_before("),\"");
			coder.add_before(caller.getAlias());
			coder.add_before("\",");
			coder.add_before("-1, ");
			coder.add_before(Coder.getCheckID());
			coder.add_before(");\n");

			coder.add_before("\t");
			coder.add_before(mobj);
			coder.add_before(".base=(void*)(");
			coder.add_before("0x");
			coder.add_before(Integer.toHexString(base));
			coder.add_before("+");
			coder.add_before(args[1].toReadableString()); // offset
			coder.add_before(");\n");
		} else {
			final String shmName  = coder.getCodeInfo().getSHMName(args[0]);
			final String portName = coder.getCodeInfo().getPortName(args[0]);

			coder.add_before("\n");
			coder.chk_ref(args[0], callee, node.getBCPosition(), "MEMBOUNDS");

			coder.add_before("\t");
			if (shmName != null) {
				coder.add_before(coder.openSHMCheck(shmName));
				caller.boundsCheckConstant++;
			} else if (portName != null) {
				coder.add_before(coder.openMessageCheck(portName));
				caller.boundsCheckConstant++;
			} else {
				coder.add_before("KESO_CHECK_BOUNDS(");
				coder.add_before(memory_class.getClassTypeString());
				coder.add_before(",");
				coder.add_before(args[0].toReadableString());
				caller.nullCheckEmitted++;
				caller.boundsCheckFull++;
			}
			coder.add_before(",(jint) (");
			coder.add_before(args[1].toReadableString()); // offset
			coder.add_before("+sizeof(");
			coder.add_before(memory_type_class.getAlias());
			coder.add_before("_mt) - 1");
			coder.add_before("),\"");
			coder.add_before(caller.getAlias());
			coder.add_before("\",");
			coder.add_before("-1, ");
			coder.add_before(Coder.getCheckID());
			coder.add_before(");\n");

			coder.add_before("\t");
			coder.add_before(mobj);
			coder.add_before(".base= ((char*)");
			coder.add_before(coder.getField(memory_class, args[0].toReadableString(), "addr"));
			coder.add_before("+");
			coder.add_before(args[1].toReadableString());
			coder.add_before(");\n");
			if (!opts.hasOption("mt_no_save_memobj")) {
				coder.add_before("\t");
				coder.add_before(mobj);
				coder.add_before(".memobj=");
				coder.add_before(args[0].toReadableString());
				coder.add_before(";\n");
			}
		}
	}

	private void mapMemToObject(IMInvoke node, IMNode[] args, IMMethod caller) throws CompileException {
		final String type = assert_string(args[2],"Parameters to mapMemoryToObject");

		final IMClass memory_type_class = repository.getClass(type);
		if (memory_type_class==null)
			throw new CompileException("Class \""+type+"\" not found!");
		coder.add_class(memory_type_class);

		final IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		if (args[0] instanceof IMAMemConstant) {
			final int base = ((IMAMemConstant)args[0]).getAddr();
			final int size = ((IMAMemConstant)args[0]).getSize();
			coder.add_before("\n");

			// memory object is known so we don't need a null check here
			caller.boundsCheckKnownIndex++;
			coder.add_before("\tKESO_CHECK_ARRAY_KNOWNBOUNDS(");
			coder.add_before(size);
			coder.add_before(",(jint) (");
			coder.add_before(args[1].toReadableString()); // offset
			coder.add_before("+sizeof(");
			coder.add_before(memory_type_class.getAlias());
			coder.add_before("_mt) - 1");
			coder.add_before("),\"");
			coder.add_before(caller.getAlias());
			coder.add_before("\",");
			coder.add_before("-1, ");
			coder.add_before(Coder.getCheckID());
			coder.add_before(");\n");

			// alloc object and call default constructor
			IMMethod constructor = memory_type_class.getMethod("<init>()V");
			if(constructor == null)
				throw new CompileException("the memory-mapped class " + memory_type_class + " needs to provide a default constructor!");

			coder.global_add_fkt("mt_init.c");
			coder.add("keso_mt_init(");
			coder.add(memory_type_class.getClassID());
			coder.add(", NULL, (void*) ");
			coder.add(base);
			coder.add(", ");
			if(constructor.getCosts()==0)
				coder.add("NULL");
			else
				coder.add(constructor.getIdentifier());
			coder.add(")");

		} else {
			coder.add_before("\n");

			// null check memory object
			coder.chk_ref(args[0], callee, node.getBCPosition(), "MEMBOUNDS");
			coder.add_before("\t");

			// check if dest region is of sufficient size
			final String shmName  = coder.getCodeInfo().getSHMName(args[0]);
			final String portName = coder.getCodeInfo().getPortName(args[0]);
			if (shmName != null) {
				coder.add_before(coder.openSHMCheck(shmName));
				caller.boundsCheckConstant++;
			} else if (portName != null) {
				coder.add_before(coder.openMessageCheck(portName));
				caller.boundsCheckConstant++;
			} else {
				coder.add_before("KESO_CHECK_BOUNDS(");
				coder.add_before(memory_class.getClassTypeString());
				coder.add_before(",");
				coder.add_before(args[0].toReadableString());
				caller.nullCheckEmitted++;
				caller.boundsCheckFull++;
			}
			coder.add_before(",(jint) (");
			coder.add_before(args[1].toReadableString()); // offset
			coder.add_before("+sizeof(");
			coder.add_before(memory_type_class.getAlias());
			coder.add_before("_mt) - 1");
			coder.add_before("),\"");
			coder.add_before(caller.getAlias());
			coder.add_before("\",");
			coder.add_before("-1, ");
			coder.add_before(Coder.getCheckID());
			coder.add_before(");\n");

			// alloc object and call default constructor
			IMMethod constructor = memory_type_class.getMethod("<init>()V");
			if(constructor == null)
				throw new CompileException("the memory-mapped class " + memory_type_class + " needs to provide a default constructor!");

			coder.global_add_fkt("mt_init.c");
			coder.add("keso_mt_init(");
			coder.add(memory_type_class.getClassID());
			coder.add(", (char*) ");
			coder.add(coder.getField(memory_class, args[0].toReadableString(), "addr"));
			coder.add("+");
			coder.add(args[1].toReadableString());
			coder.add(", ");
			if (!opts.hasOption("mt_no_save_memobj")) {
				coder.add(args[0].toReadableString());
			} else {
				coder.add("NULL");
			}
			coder.add(", ");
			if(constructor.getCosts()==0)
				coder.add("NULL");
			else
				coder.add(constructor.getIdentifier());
			coder.add(")");
		}
	}

	private void adjustMem(IMMethod method) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		coder.addln("char *src;");
		coder.add("src = ((char*)");
		coder.add_getField(memory_class, method,method.getMethodArgument(0).toString(), "addr", "LGETFIELD");
		coder.addln(") + "+ method.getMethodArgument(1).toString()+";");

		/* check source */
		coder.add("KESO_CHECK_BOUNDS(");
		coder.add(memory_class.getClassTypeString());
		coder.add(","+ method.getMethodArgument(0).toString() +","+method.getMethodArgument(1).toString()+"+"+method.getMethodArgument(3).toString()+"-1,");
		if (opts.useExceptionStrings()) {
			coder.add('"');
			coder.add(callee.getAlias());
			coder.add("\",");
		} else {
			coder.add("(char *) 0,");
		}
		coder.add("-1, ");
		coder.add(Coder.getCheckID());
		coder.addln(");");

		coder.add_getField(memory_class, method, method.getMethodArgument(2).toString(), "addr", "LGETFIELD");
		coder.addln(" = src;");
		coder.add_getField_fast(memory_class,  method.getMethodArgument(2).toString(), "size", "IGETFIELD");
		coder.addln(" = "+method.getMethodArgument(3).toString()+";");
	}

	private void releaseMem() throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);
		// TODO releaseMemory: add support for dynamic memory!
		coder.add_getField_fast(memory_class,callee.getMethodArgument(0).toString(), "size", "IGETFIELD");
		coder.addln(" = 0;");
		coder.addln("return;");
	}
	private void copyMem(IMMethod method) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		coder.addln("char *src,*dst;");
		coder.add("src = ((char*)");
		coder.add_getField(memory_class, method,method.getMethodArgument(0).toString(), "addr", "LGETFIELD");
		coder.addln(") + "+method.getMethodArgument(1).toString()+";");
		coder.add("dst = ((char*)");
		coder.add_getField(memory_class, method,method.getMethodArgument(0).toString(), "addr", "LGETFIELD");
		coder.addln(") + "+method.getMethodArgument(3).toString()+";");

		coder.add("KESO_CHECK_BOUNDS(");
		coder.add(memory_class.getClassTypeString());
		coder.add(","+ method.getMethodArgument(0).toString() + ","+method.getMethodArgument(1).toString()+"+"+method.getMethodArgument(4).toString()+"-1,");
		if (opts.useExceptionStrings()) {
			coder.add('"');
			coder.add(callee.getAlias());
			coder.add("\",");
		} else {
			coder.add("(char *) 0,");
		}
		coder.add("-1, ");
		coder.add(Coder.getCheckID());
		coder.addln(");");

		coder.add("KESO_CHECK_BOUNDS(");
		coder.add(memory_class.getClassTypeString());
		coder.add(","+ method.getMethodArgument(2).toString()+","+method.getMethodArgument(3).toString()+"+"+method.getMethodArgument(4).toString()+"-1,");
		if (opts.useExceptionStrings()) {
			coder.add('"');
			coder.add(callee.getAlias());
			coder.add("\",");
		} else {
			coder.add("(char *) 0,");
		}
		coder.add("-1, ");
		coder.add(Coder.getCheckID());
		coder.addln(");");

		coder.addln("for (;"+method.getMethodArgument(4).toString()+">=0;"+method.getMethodArgument(4).toString()+"--) { *dst=*src; dst++; src++; };");
		coder.addln("return;");
	}

	private void allocDynDevMem(IMMethod method) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		coder.add("object_t* newMem = ");
		coder.add_allocObject(memory_class);
		coder.addln(";");
		coder.add_getField(memory_class, method, "newMem", "addr", "LGETFIELD");
		coder.addln(" = (void*) "+method.getMethodArgument(0).toString()+";");
		// omit 2nd null reference check
		coder.add_getField_fast(memory_class, "newMem", "size", "IGETFIELD");
		coder.addln(" = "+method.getMethodArgument(1).toString()+";");

		coder.addln("return newMem;");
	}

	private void allocDynMem(IMMethod method) throws CompileException {
		IMClass memory_class = repository.getClass("keso/core/Memory");
		coder.add_class(memory_class);

		if(repository.getSharedMemorySize() > 0) {
			coder.addln("void *addr;");
			coder.addln("jint size = " + method.getMethodArgument(0).toString() + ";");
			coder.addln("if (size>0) {");
			coder.addln("\tif( (addr=keso_alloc_shared_memory(size))==NULL) {");
			coder.addln("\t\treturn NULL;");
			coder.addln("\t} else {");
			coder.add("\t\tobject_t *newMem = ");
			coder.add_allocObject(memory_class);
			coder.addln(";");

			// omit null checks, allocObject would throw an OutOfMemoryException on allocation failure
			coder.add("\t\t");
			coder.add_getField_fast(memory_class, "newMem", "addr", "LGETFIELD");
			coder.addln(" = addr;\t");

			coder.add("\t\t");
			coder.add_getField_fast(memory_class, "newMem", "size", "IGETFIELD");
			coder.addln(" = ((size+3)&~3);");
			coder.addln("\t\treturn newMem;");
			coder.addln("\t}");
			coder.addln("}");
		}

		// the new Memory object is already zero-initialized, which is fine for memory objects of size 0
		coder.add("return ");
		coder.add_allocObject(memory_class);
		coder.addln(";");
	}

	public boolean ignoreMethodBody(IMClass clazz, IMMethod method)
		throws CompileException {
		return true;
	}
}
