/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.kni;

import keso.compiler.*;
import keso.compiler.imcode.*;
import keso.compiler.backend.*;
import keso.compiler.config.domain.*;
import keso.compiler.config.task.*;

import java.util.*;

public class TaskService extends KesoIndexedResource<TaskDefinition> {

	private IMClass task_clazz;

	// Why can the gc run while schedule is called?
	private boolean schedule_gc_run = false;
	private boolean more_inline = true;

	public TaskService(BuilderOptions opts, ClassStore repository) {
		super(opts, repository);
		//
		// Hier alle Methoden aufzaehlen, die von dem "Weavelet" beeinflusst werden
		// sollen.
		//
		joinPoints = new String[] { "keso/core/TaskService\\..+" };
	}

	public void require(int domainID, String className, String methodName) {
		repository.requireClass(domainID, "java/lang/Thread");
	}

	public int checkAttribute(IMMethod self, int attr) {
		if (schedule_gc_run && attr==ATTR_BLOCK) {
			if (self.termed("schedule()V")) return TRUE;
		} else if (attr==ATTR_PURE) {
			if(self.termed("getTaskByName(Ljava/lang/String;)Ljava/lang/Thread;")
				|| self.termed("getTaskID()Ljava/lang/Thread;")
				|| self.termed("getTaskState(Ljava/lang/Thread;)I")
			) {
					return TRUE;
			}
		}
		return FALSE;
	}

	public boolean ignoreMethodBody(IMClass clazz, IMMethod method) throws CompileException {
		return true;
	}

	//
	// Mit affectMethod wird der Methoden-Rumpf ersetzt. Der Aufruf bleibt ansonsten
	// erhalten. (siehe keso/compiler/weavelet/Object.)
	//

	public boolean affectMethod(IMClass clazz, IMMethod method, Coder coder) throws CompileException
	{
		this.coder = coder;
		this.callee = method;
		this.clazz = clazz;

		if (method.termed("getTaskState(Ljava/lang/Thread;)I")) {
			getTaskState(method);

		} else if (!more_inline && method.termed("chain(Ljava/lang/Thread;)I")) {
			if(opts.hasLinkedListOfLocalReferences()) {
				coder.add("KESO_CURRENT_TASK->_stack = NULL;");
			}
			retIgetResRef("ChainTask", "calls OSEK ChainTask", "java/lang/Thread", "task_id",method);

		} else if (!more_inline && callee.termed("activate(Ljava/lang/Thread;)I")) {
			retIgetResRef("ActivateTask", "calls OSEK ActivateTask", "java/lang/Thread", "task_id",method);

		} else return false;

		return true;
	}

	//
	// Mit affectInvokeXXX wird der Aufruf ersetzt. In unserem Fall der Aufruf an dem
	// Interface keso/core/TaskService.
	//
	// node   Ist das IMNode-Objekt von dem wir aufgerufen werden.
	//
	// caller Ist die aufrufende Methode.
	//
	// callee Ist die Methode, die aufgerufen werden soll.
	//
	// obj    Ist das IMNode-Objekt welches den this-Pointer representiert.
	//
	// args[] Sind die weiteren Parameter.
	//
	// coder  Ist das aktuelle C-Backend.
	//
	// Rueckgabewert. Bei false wird die translate-Methode von node weiter ausgefuehrt
	//                als waere nichts passiert. Bei true wird diese Methode abgebrochen.
	//

	public boolean affectInvokeStatic(IMInvoke node, IMMethod caller, IMMethod callee,
			IMNode args[], Coder coder) throws CompileException
	{
		this.node   = node;
		this.caller = caller;
		this.callee = callee;
		this.args   = args;
		this.coder  = coder;

		if (callee.termed("terminate()I")) {
			TerminateTask();
		} else if (callee.termed("schedule()V")) {
			if (schedule_gc_run)
				caller.getMethodFrame().codeEOLL(coder);
			Schedule();
		} else if (more_inline && callee.termed("activate(Ljava/lang/Thread;)I")) {
			ActivateTask(args[0]);
		} else if (more_inline && callee.termed("chain(Ljava/lang/Thread;)I")) {
			ChainTask(args[0]);
		} else if (callee.termed("getTaskID()Ljava/lang/Thread;")) {
			if (task_clazz==null) task_clazz = repository.getClass("java/lang/Thread");
			coder.add("KESO_GET_CURRENT_TASK()");
		} else if (callee.termed("getTaskByName(Ljava/lang/String;)Ljava/lang/Thread;")) {
			getIndexedResourceByName();
			return true;
		} else return false;

		// return value true tells the builder that this invoke has been handled
		return true;
	}

	private void ChainTask(IMNode taskID) throws CompileException {
		if (task_clazz==null) task_clazz = repository.getClass("java/lang/Thread");
		coder.add_class(task_clazz);

		if(opts.hasLinkedListOfLocalReferences()) {
			coder.add("KESO_CURRENT_TASK->_stack = NULL, ");
		}
		coder.add("(jint) ChainTask(");
		coder.add_getField(task_clazz, taskID, "task_id");
		coder.add(")");
	}

	private void ActivateTask(IMNode taskID) throws CompileException {
		if (task_clazz==null) task_clazz = repository.getClass("java/lang/Thread");
		coder.add_class(task_clazz);
		coder.add("((jint) ActivateTask(");
		coder.add_getField(task_clazz, taskID, "task_id");
		coder.add("))");
	}

	private void TerminateTask() throws CompileException {
		if (task_clazz==null) task_clazz = repository.getClass("java/lang/Thread");

		if(opts.hasLinkedListOfLocalReferences()) {
			coder.add("KESO_CURRENT_TASK->_stack = NULL, ");
		}
		coder.add("TerminateTask()");
	}

	private void Schedule() throws CompileException {
		coder.add("Schedule()");
	}

	private void getTaskState(IMMethod method) throws CompileException {
		IMClass task_class = repository.getClass("java/lang/Thread");
		coder.add_class(task_class);

		coder.addln("/* Calls OSEK GetTaskState and passes the result to the application. */\n");
		coder.addln("TaskStateType state;\n");
		coder.add("GetTaskState(");
		coder.add_getField(task_class,method,method.getMethodArgument(0).toString(),"task_id", "BGETFIELD");

		if (opts.getTarget() == BuilderOptions.CIAO_TARGET) {
			coder.addln(", state);"); // use C++ references
		} else {
			coder.addln(", &state);");
		}
		coder.addln("return (jint) state;");
	}

	private Map<IMDomain, Map<String, TaskDefinition>> localResources;

	protected Map<IMDomain, ? extends Map<String,? extends TaskDefinition>> accessibleResources() throws CompileException {
		if(localResources == null) {
			localResources = new HashMap<IMDomain, Map<String, TaskDefinition>>();

			for(IMDomain imd: opts.getDomains()) {
				JavaDomain dom = imd.getDef();
				Map<String,TaskDefinition> lr = new HashMap<String, TaskDefinition>();

				for(TaskDefinition rd: dom.getTasks())
					lr.put(rd.ident, rd);
				for(String resName: dom.getTaskImports()) {
					lr.put(resName, opts.getSysDef().getSysTask(resName));
				}
				localResources.put(imd, lr);
			}
		}
		return localResources;
	}

	protected String resCategory() { return "task"; }
	protected IMClass resClass(TaskDefinition so) {
		return repository.getClass("java/lang/Thread");
	}
}
