/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.transform;

import java.util.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;

public final class CFGSimplification extends MethodPass {

	private static final class BlockRemover extends IMBasicVisitor {

		@Override
		public void visit(IMNode node) throws CompileException {
			if (node.usesSlot()) {
				IMSlot slot = node.getIMSlot();
				slot.deleteUse(node);
				if (slot.hasNoUses())
					node.getOptions().todo("Remove unused variable "
							+ node.getIMSlot());
			}
		}

	}

	private static final class BranchUpdater extends IMBasicVisitor {

		private IMBasicBlock oldTarget;
		private IMBasicBlock newTarget;

		public BranchUpdater(IMBasicBlock oldTarget, IMBasicBlock newTarget) {
			this.oldTarget = oldTarget;
			this.newTarget = newTarget;
		}

		@Override
		public boolean skipNode(IMNode node) throws CompileException {
			return (node.isBranch() == false);
		}

		@Override
		public void visit(IMNode node) throws CompileException {
			IMBasicBlock[] targets = ((IMBranch) node).getTargets();
			for (int i = 0; i < targets.length; ++i) {
				if (targets[i].equals(oldTarget))
					targets[i] = newTarget;
			}
		}

	}

	@Override
	public String getName() {
		return "CFG simplification";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.hasOption("simplify_cfg");
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(SSADeconstruction.class);
		info.addDependency(DomTreeComputation.class, true);
		info.addInvalidated(LivenessAnalysis.class);
		info.setDoDump(opts.hasOption("dump_cfgsimp"));
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method.isEligibleForOptimization() == false)
			return false;

		// Initialize work list
		SortedSet<IMBasicBlock> workList = new TreeSet<IMBasicBlock>();
		for (IMBasicBlock bb : method) {
			workList.add(bb);
		}

		boolean      modified     = false;
		BlockRemover blockRemover = new BlockRemover();

		// Process all blocks in the work list
		while (workList.isEmpty() == false) {

			IMBasicBlock bb = workList.first();

			if (canRemove(bb)) {

				method.getOptions().verbose("++ " + method
						+ ": Removing redundant basic block " + bb.getShortLabel());

				// Update successor list and branch instructions of all
				// predecessor blocks and the predecessor list of the successor
				// block
				IMBasicBlock succ = bb.getSucc().at(0);
				for (IMBasicBlock pred : bb.getPred()) {
					pred.visitNodes(new BranchUpdater(bb, succ));
					pred.updateSucc(bb, succ);
					succ.addPred(pred);
				}
				succ.unlinkPred(bb, false);
				for (IMBasicBlock succPred : succ.getPred()) {
					workList.add(succPred);
				}

				// Remove the basic block
				bb.visitNodes(blockRemover);
				method.removeBasicBlock(bb);
				modified = true;
			}

			workList.remove(bb);
		}

		return modified;
	}

	/**
	 * Determines whether a basic block can be removed safely.
	 * @param bb Basic block to be examined.
	 * @return true if the block can be removed safely, false otherwise.
	 */
	private static boolean canRemove(IMBasicBlock bb) {

		// If the block has more than one unique successor, keep it
		IMBasicBlockList succs = bb.getSucc();
		IMBasicBlock     succ  = null;
		if (succs.length() > 0) {
			succ = succs.at(0);
			for (IMBasicBlock otherSucc : succs) {
				if (otherSucc.equals(succ) == false)
					return false;
			}
		}

		// If the block contains any other nodes than a branch, keep it
		for (IMNode node : bb) {
			if (node.isBranch() == false)
				return false;
		}

		// If the block is an infinite loop, keep it
		if (succ.equals(bb))
			return false;

		// If the block falls through to its only successor, remove it
		if (bb.isFallingThrough(succ))
			return true;

		// If the block does not fall through, only remove it if each of its
		// predecessors has an explicit branch instruction that can be
		// adjusted
		for (IMBasicBlock pred : bb.getPred()) {
			if (pred.isFallingThrough(bb)
					&& hasExplicitBranchTo(pred, bb) == false)
				return false;
		}
		return true;
	}

	private static boolean hasExplicitBranchTo(IMBasicBlock bb, IMBasicBlock to) {
		for (IMNode node : bb) {
			if (node.isBranch() == false)
				continue;
			IMBasicBlock[] targets = ((IMBranch) node).getTargets();
			if (targets.length == 1 && targets[0].equals(to))
				return true;
		}
		return false;
	}

}
