/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.transform;

import java.util.*;

import keso.classfile.constantpool.*;
import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;

/**
 * Whole-program optimization pass.
 * This pass evaluates the results of the flow analysis. It performs the
 * following optimizations:
 * - Transformation of predictable conditional branches
 * - Removal of dead basic blocks
 * - Removal of unused methods (triggering a VTable re-computation)
 * - Conversion of predictable virtual method invocations
 */
public final class FlowBasedOptimization extends MethodPass {

	private FlowLatticeInfoStore flowInfoStore;

	/**
	 * Set of methods that are actually unused, but whose abstract declaration
	 * has to stick around because they are directly referenced in a virtual
	 * invocation.
	 */
	private final Set<IMMethod> stickyMethods = new HashSet<IMMethod>();

	@Override
	public String getName() {
		return "Whole-program optimization";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.hasOption("whole_program_opt");
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(FlowAnalysis.class);
		//info.addAntiDependency(DeadInstructionRemoval.class);
		info.addAntiDependency(EscapeAnalysis.class);
		info.addAntiDependency(ExtendedEscapeAnalysis.class);
		info.addAntiDependency(SSADeconstruction.class);
		info.addInvalidated(FieldAccessReset.class);
		info.addInvalidated(LivenessAnalysis.class);
		info.setDoDump(opts.hasOption("dump_whole_program_opt"));
	}

	@Override
	public boolean init() throws CompileException {
		FlowAnalysis flowAnalysis = pm.getPass(FlowAnalysis.class);
		flowInfoStore = flowAnalysis.getFlowLatticeInfoStore();
		return true;
	}

	@Override
	public boolean runOnMethod(final IMMethod method) throws CompileException {

		// If the method is unused, remove it
		final BuilderOptions opts = method.getOptions();
		if (method.getDomains().isEmpty() && stickyMethods.contains(method) == false) {
			opts.verbose("Marking unused method " + method + " for removal");
			for (IMMethod caller : method.getCallers()) {
				caller.removeCalledMethod(method);
			}
			method.resetCallerAnalysis();
			return true;
		}

		// Otherwise make sure the method will not be removed
		if (method.getCallers().isEmpty())
			method.calledFrom(null, null);

		// If the method should be ignored, skip it
		if (method.isEligibleForOptimization() == false)
			return false;

		for (final IMBasicBlock bb : method) {

			// If the basic block is dead, skip it
			if (bb.getDomains().isEmpty()) {

				// ... but remove the uses and definitions first
				bb.visitNodes(new IMBasicVisitor() {
					@Override
					public void visit(IMNode node) throws CompileException {
						IMSlot slot = node.getIMSlot();
						if (node.usesSlot()) {
							if (slot.hasNoUses() == false)
								slot.deleteUse(node);
						} else if (node instanceof IMStoreLocalVariable) {
							slot.clearDefUseChain();
						}
					}
				});
				continue;
			}
			
			final FlowLatticeInfo flowInfo
			= flowInfoStore.getForDomains(bb.getDomains());

			bb.copy(new IMCopyVisitor() {

				private boolean visited = false;

				@Override
				public IMNode visit(IMNode nnode, IMNode onode)
						throws CompileException {

					IMDataFlow cell = flowInfo.getCell(onode);
					
					if (canConstifyNode(onode, bb, flowInfo, cell)) {
						
						opts.verbose("Folding constant node "
						+ nnode.toReadableString() + " in " + bb);

						// Update uses
						nnode.visitNodes(new IMBasicVisitor() {
							@Override
							public void visit(IMNode node)
							throws CompileException {
								if (node.usesSlot())
									node.getIMSlot().deleteUse(node);
							}
						});
						nnode = cell.getValue();

						// FIXME: We should also remove all stores of now-unused
						//        stack slots provided they are pure.

					} else if (nnode instanceof IMInvokeVirtual) {

						Collection<IMMethod> candidates
						= flowInfo.getCandidates((IMInvokeVirtual) onode);
						nnode = visitInvokeVirtual((IMInvokeVirtual) nnode, bb,
								candidates);

					} else if (cell != null && cell.isConstant()
							&& nnode.isBranch()) {

						// Replace conditional branch with unconditional branch
						IMBranch branch = (IMBranch) nnode;
						int      index  = cell.getValue().getIntValue();
						IMBasicBlock target = branch.getTargets()[index];
						int    bcPos = branch.getBCPosition();
						IMNode goTo  = method.createGoto(target, bcPos);
						opts.verbose("Converting predictable branch "
								+ branch.toReadableString() + " into "
								+ goTo.toReadableString());
						bb.replaceNode(branch, goTo);
						if (nnode instanceof IMNullCheck)
							++method.nullCheckAOTSuccess;

						// Update successor links
						int i = 0;
						IMBasicBlockList succs = bb.getSucc();
						while (i < succs.length()) {
							IMBasicBlock succ = succs.at(i);
							if (succ != target)
								removeLinks(bb, succ);
							else
								++i;
						}
						return flowInfoStore.replaceNode(onode, goTo);
					}

					// Update uses and definitions
					IMSlot slot = nnode.getIMSlot();
					if (nnode.usesSlot()) {
						if (slot.hasNoUses() == false)
							slot.deleteUse(onode);
						slot.use(bb, nnode);
					} else if (nnode instanceof IMStoreLocalVariable) {
						slot.redefine(bb, nnode);
					}

					return flowInfoStore.replaceNode(onode, nnode);
				}

				@Override
				public IMBasicBlock updateBlock(IMBasicBlock label)
						throws CompileException {

					// Only process the current basic block
					if (label != bb || visited == true)
						return label;
					visited = true;

					// FIXME: This is slow (O(n^2)) - better use or write a
					// method of IMBasicBlock
					for (IMNode root : label) {
						label.replaceNode(root, root.copy(this));
					}
					return label;
				}

				@Override
				public IMSlot adjustSlot(IMSlot foreignSlot)
						throws CompileException {
					return foreignSlot;
				}
			});
		}

		IMBasicBlock firstBB = method.iterator().next();
		IMBasicBlock prevBB = null;

		// Remove all dead basic blocks
		for (IMBasicBlock bb = firstBB; bb != null;
				bb = (IMBasicBlock) prevBB.next()) {

			// If the basic block is not dead, keep it
			// Since the reachability analysis is pessimistic, we must not keep
			// blocks that are considered live but have no predecessors (unless
			// the block in question is the entry block of its method).
			IMBasicBlockList preds = bb.getPred();
			if (bb.getDomains().isEmpty() == false
					&& (preds.length() > 0 || bb == firstBB)) {
				prevBB = bb;
				continue;
			}

			// Remove links from/to all predecessor and successor blocks and
			// erase the block
			opts.verbose("Removing dead basic block " + bb);
			while (preds.length() > 0) {
				removeLinks(preds.at(0), bb);
			}
			IMBasicBlockList succs = bb.getSucc();
			while (succs.length() > 0) {
				removeLinks(bb, succs.at(0));
			}
			prevBB = bb.removeDeadBasicBlocks(prevBB);
		}

		return true;
	}

	private IMInvoke visitInvokeVirtual(IMInvokeVirtual invoke,
			IMBasicBlock bb, Collection<IMMethod> candidates) throws CompileException {

		// If the direct callee is unused, make it abstract but keep it around
		// because it is still referenced in the vtable
		IMMethod callee = invoke.getCallee();
		if (callee.getDomains().isEmpty()) {
			callee.calledFrom(null, null);
			callee.makeAbstract();
			stickyMethods.add(callee);
		}

		// If there is exactly one candidate, convert the virtual invocation
		// into an InvokeSpecial
		if (candidates.size() == 1) {
			IMMethod candidate = candidates.iterator().next();
			if (candidate.isEligibleForOptimization() == false) // Can't convert
				return invoke;
			IMInvokeSpecial invokeSpecial
			= invoke.convertToSpecial(candidate.getIMClass());
			invokeSpecial.setCPEntry(getCPEntry(candidate));
			bb.getOptions().verbose("Simplifying " + invoke.toReadableString()
					+ " in " + bb.getMethod());
			return invokeSpecial;
		}

		return invoke;
	}
	
	static boolean areAllCandidatesPure(IMInvoke invoke,
			FlowLatticeInfo flowInfo) throws CompileException {
		Collection<IMMethod> candidates = flowInfo.getCandidates(invoke);
		for (IMMethod candidate : candidates) {
			if (candidate.isPure() == false
					|| candidate.isEligibleForOptimization() == false)
				return false;
		}
		return true;
	}
	
	private static boolean canConstifyNode(IMNode node, IMBasicBlock bb,
			FlowLatticeInfo flowInfo, IMDataFlow cell) throws CompileException {

		// Do nothing if we have no information or the node is already constant
		if (cell == null || cell.isConstant() == false || node.isConstant())
			return false;

		// If the node is an invocation, remove it if all candidates are pure
		if (node instanceof IMInvoke)
			return areAllCandidatesPure((IMInvoke) node, flowInfo);
		
		// If the node is not an invocation, ask it if it wants to be removed
		return node.removeIfUnUsed();
	}

	private static MethodRefCPEntry getCPEntry(IMMethod method) {

		// Try to find the matching entry in the constant pool
		ConstantPool cp = method.getIMClass().getConstantPool();
		String className   = method.getClassName();
		String nameAndType = method.getMethodNameAndType();
		String description = className + '.' + nameAndType;
		for (int i = 0; i < cp.getNumberOfEntries(); ++i) {
			ConstantPoolEntry entry = cp.entryAt(i);
			if (entry instanceof MethodRefCPEntry
					&& entry.getDescription().equals(description))
				return (MethodRefCPEntry) entry;
		}

		// If none has been found, create a fake entry
		int    index = nameAndType.indexOf('(');
		String name  = nameAndType.substring(0, index);
		String type  = nameAndType.substring(index);
		return new FakeMethodRefCPEntry(className, name, type);
	}

	private static void removeLinks(final IMBasicBlock from,
			final IMBasicBlock to) throws CompileException {

		final IMBasicBlockList toPreds = to.getPred();

		// Only process the successor block if it is not going to be removed
		// anyway
		if (to.getDomains().isEmpty() == false) {

			// Update Phi-functions in the successor block
			to.visitNodes(new IMBasicVisitor() {

				@Override
				public void visit(IMNode node) throws CompileException {
					if (node instanceof IMPhi == false)
						return;
					for (int i = 0; i < toPreds.length(); ++i) {
						if (toPreds.at(i) == from) {
							((IMPhi) node).removeSource(i);
							return;
						}
					}
					throw new CompileException(from
							+ " is not in the predecessor list of " + to);
				}
			});
		}

		// Remove the links
		from.getSucc().remove(to);
		toPreds.remove(from);
	}

}
