/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.transform;

import java.util.*;
import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;

public final class MethodInlining extends MethodPass {

	private Set<IMMethod> inlined = new HashSet<IMMethod>();

	@Override
	public String getName() {
		return "Method inlining";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return (opts.doInlineCalls() == true);
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(CallGraphAnalysis.class, true);
		info.addAntiDependency(ConstantFolding.class);
		info.addAntiDependency(SSAConstruction.class);
		info.addInvalidated(DomTreeComputation.class);
		info.addInvalidated(FieldAccessReset.class);
		info.addInvalidated(FlowAnalysis.class);
		info.addInvalidated(LivenessAnalysis.class);
		info.addInvalidated(SimpleReachabilityAnalysis.class);
		info.setDoDump(opts.hasOption("dump_inlining"));
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method.isEligibleForOptimization() == false
				|| inlined.contains(method))
			return false;

		BuilderOptions opts = method.getOptions();
		opts.vverbose("&& inlineMethodCalls in " + method.getAlias());
		inlined.add(method);

		IMBasicBlock bb = method.iterator().next();
		while (bb != null) {
			bb = bb.inlineMethodCalls();
		}

		if (opts.hasOption("dont_merge_basic_blocks") == false) {
			bb = method.iterator().next();
			while (bb != null) {
				bb = bb.melt();
			}
		}

		for (IMBasicBlock block : method) {

			// Work around a bug in method inlining that sets wrong targets
			// for Goto instructions
			// FIXME: Remove this workaround and find & fix the actual bug!
			for (IMNode node : block) {
				if (node.isBranch()) {
					IMBasicBlock[] targets = ((IMBranch) node).getTargets();
					IMBasicBlock   succ    = block.getSucc().at(0);
					if (targets.length == 1 && targets[0].equals(succ) == false)
						targets[0] = succ;
				}
			}

			// Fix uses and definitions
			block.visitNodes(new IMDefUseUpdater(block));
		}
		return true;
	}

}
