/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.transform;

import java.util.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class SSAConstruction extends MethodPass {

	@Override
	public String getName() {
		return "SSA construction";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		return opts.hasOption("ssa");
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(DomTreeComputation.class);
		info.addDependency(LivenessAnalysis.class, true);
		info.setDoDump(opts.hasOption("dump_ssa"));
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method.isEligibleForOptimization() == false)
			return false;

		/*
		 * Dominance frontiers capture the precise places at which we need phi functions:
		 * if the node A defines a certain variable, then that definition and that definition
		 * alone will reach every node A dominates. Only when we leave these nodes and enter
		 * the dominance frontier must we account for other flows bringing in other definitions
		 * of the same variable. Moreover, no other phi functions are needed in the control flow
		 * graph to deal with A's definitions, and we can do with no less.
		 * (Wikipedia)
		 */

		IntegerHashtable<Set<IMSlot>> A_orig
		= new IntegerHashtable<Set<IMSlot>>();
		SortedMap<IMSlot, IntegerHashtable<IMBasicBlock>> defSites
		= new TreeMap<IMSlot, IntegerHashtable<IMBasicBlock>>();
		IntegerHashtable<Collection<IMBasicBlock>> childrenOf
		= new IntegerHashtable<Collection<IMBasicBlock>>();

		IMDefVisitor  visitor = new IMDefVisitor(method, defSites, A_orig);
		DominatorTree domTree = method.getDomTree();
		for (IMBasicBlock bb : method) {
			bb.visitNodes(visitor);
			IMBasicBlock dbb = domTree.getDominatorOf(bb);
			if (dbb != null) {
				int n = dbb.getDFNum();
				Collection<IMBasicBlock> children = childrenOf.get(n);
				if (children == null) {
					children = new ArrayList<IMBasicBlock>();
					childrenOf.put(n, children);
				}
				children.add(bb);
			}
		}

		// First place Phi-functions
		placePhiFunctions(method, defSites, A_orig);

		// Now rename variables
		Map<IMSlot, MarkableStack<IMSlot>> stacks
		= new HashMap<IMSlot, MarkableStack<IMSlot>>();
		for (IMSlot var : defSites.keySet()) {
			MarkableStack<IMSlot> stack = new MarkableStack<IMSlot>(16);
			// undef. stack.push();
			var.clearDefUseChain();
			stacks.put(var, stack);
		}

		// Define all method arguments
		IMSlot args[] = method.getMethodFrame().getMethodArguments();
		for (IMSlot arg : args) {
			MarkableStack<IMSlot> stack = new MarkableStack<IMSlot>(16);
			arg.clearDefUseChain();
			IMBasicBlock firstBb = method.iterator().next();
			arg.define(firstBb, null);
			stack.push(arg);
			stacks.put(arg, stack);
		}

		renameVars(method, 1, childrenOf, stacks);

		// We have to recompute the return value because the slot has changed
		method.recomputeReturnValue();

		// Done - now we can do all the optimization stuff.
		return true;
	}

	/**
	 * defsites{var}{n} hash of hashes with all definition sites of variable var
	 * A_orig{n}{var}   hash of hashes with all variables defined in node n
	 */
	private void placePhiFunctions(IMMethod method,
			SortedMap<IMSlot, IntegerHashtable<IMBasicBlock>> defSites,
			IntegerHashtable<Set<IMSlot>> A_orig) throws CompileException {

		DominatorTree domTree = method.getDomTree();
		Map<IMSlot, IntegerHashtable<IMBasicBlock>> A_phi
		= new HashMap<IMSlot, IntegerHashtable<IMBasicBlock>>();

		// For each variable var do
		for (IMSlot var : defSites.keySet()) {

			IntegerHashtable<IMBasicBlock> defs = defSites.get(var);

			while (defs.isEmpty() == false) {

				// Work list = definition sites of var
				for (int n : defs.keys()) {

					defs.remove(n);

					for (int y : domTree.getDominanceFrontierIDs(n)) {

						IntegerHashtable<IMBasicBlock> Avar_phi
						= A_phi.get(var);
						if (Avar_phi == null) {
							Avar_phi = new IntegerHashtable<IMBasicBlock>();
							A_phi.put(var, Avar_phi);
						}

						if (Avar_phi.get(y) != null)
							continue;

						// Do not place a Phi-function if the variable is not in
						// the LiveIn set of the dominated basic block
						IMBasicBlock dfBb = domTree.getIMBasicBlock(y);
						if (dfBb.getLiveInSlots().contains(var) == false)
							continue;

						// Insert Phi-function in basic block y (dominance
						// frontier)
						IMPhi phi = new IMPhi(method, dfBb, var, var.getType());
						dfBb.placePhi(method.createStoreVariable(var, phi, -1));
						Avar_phi.put(y, dfBb);

						// Add changed basic block to the work list
						Set<IMSlot> A_orig_y = A_orig.get(y);
						if (A_orig_y == null
								|| A_orig_y.contains(var) == false)
							defs.put(y, dfBb);
					}
				}
			}
		}
	}

	private void renameVars(IMMethod method, int dfnum,
			IntegerHashtable<Collection<IMBasicBlock>> childrenOf,
			Map<IMSlot, MarkableStack<IMSlot>> stacks) throws CompileException {

		for(MarkableStack<IMSlot> s : stacks.values()) {
			s.mark();
		}

		DominatorTree domTree = method.getDomTree();
		IMBasicBlock  bb      = domTree.getIMBasicBlock(dfnum);
		bb.visitNodes(new IMRenameVisitor(method, bb, stacks));

		for (IMBasicBlock succ : bb.getSucc()) {
			succ.visitNodes(new IMPhiUpdateVisitor(method, succ, bb, stacks));
		}

		Collection<IMBasicBlock> children = childrenOf.get(dfnum);
		if (children != null) {
			for (IMBasicBlock nb : children) {
				renameVars(method, nb.getDFNum(), childrenOf, stacks);
			}
		}

		for (MarkableStack<IMSlot> s : stacks.values()) {
			s.reset();
		}
	}

}
