/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.compiler.transform;

import java.util.*;
import java.util.Map.*;

import keso.compiler.*;
import keso.compiler.analysis.*;
import keso.compiler.imcode.*;
import keso.util.*;

public final class SSADeconstruction extends MethodPass {

	@Override
	public String getName() {
		return "SSA deconstruction";
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		try {
			return pm.getPass(SSAConstruction.class).isEnabled(opts);
		} catch (CompileException e) {
			return false;
		}
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(LivenessAnalysis.class);
		info.addDependency(SSAConstruction.class, true);
		info.setDoDump(opts.hasOption("dump_dessa"));
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {

		if (method.isEligibleForOptimization() == false)
			return false;

		BuilderOptions opts = method.getOptions();
		opts.verbose("###### Destroying SSA form of " + method);

		UndirectedGraph<IMSlot> interferences = new UndirectedGraph<IMSlot>();
		IMMethodFrame           frame         = method.getMethodFrame();
		Multimap<IMSlot, IMSlot> phiCongruenceClasses
		= new HashUniqueMultimap<IMSlot, IMSlot>();
		SortedMap<IMStoreLocalVariable, IMBasicBlock> phis
		= new TreeMap<IMStoreLocalVariable, IMBasicBlock>();
		Multimap<IMBasicBlock, IMSlot> extendedLiveOuts
		= new HashMultimap<IMBasicBlock, IMSlot>();

		// Iterate over all basic blocks
		for (IMBasicBlock bb : method) {

			// All LiveOut variables interfere with each other
			Set<IMSlot> liveOut = bb.getLiveOutSlots();
			extendedLiveOuts.addAll(bb, liveOut);
			Set<IMSlot> live = new HashSet<IMSlot>(liveOut);
			updateGraph(interferences, live);

			// Iterate over all nodes in reverse order
			for (IMNode root : reverse(bb)) {

				if (root instanceof IMStoreLocalVariable) {

					// Definition
					IMSlot destSlot = root.getIMSlot();
					live.remove(destSlot);
					updateGraph(interferences, live);

					// Special handling for Phi-functions
					IMStoreLocalVariable store   = (IMStoreLocalVariable) root;
					IMNode               operand = store.getOperand();

					if (operand instanceof IMPhi) {

						IMBasicBlockList preds = bb.getPred();
						phiCongruenceClasses.add(destSlot, destSlot);

						for (int i = 0; i < preds.length(); ++i) {

							IMPhi  phiInst = (IMPhi) store.getOperand();
							IMNode source  = phiInst.getSource(i);
							IMSlot srcSlot = source.getIMSlot();

							// If the Phi-function has an operand that is not a
							// read from a stack slot, push the operand to the
							// end of the predecessor block and add a copy
							if (source.isVariable() == false) {
								srcSlot = frame.createNewVariableSSA(destSlot);
								opts.vverbose("++ Normalizing folded operand "
										+ i + " of the Phi-function defining "
										+ destSlot);
								IMStoreLocalVariable nStore
								= method.createStoreVariable(srcSlot, source,
										phiInst.getBCPosition());
								IMBasicBlock pred = preds.at(i);
								if (destSlot.hasNoUses() == false) {
									pred.appendBeforeBranch(nStore);
									srcSlot.define(pred, nStore);
								}
								IMReadLocalVariable nRead
								= nStore.getReadOperation();
								phiInst.setSource(i, nRead);
								srcSlot.use(bb, nRead);
								pred.addLiveOut(srcSlot);
								bb.addLiveIn(srcSlot);
								for (IMSlot oSlot : pred.getLiveOutSlots()) {
									interferences.addEdge(srcSlot, oSlot);
								}
							}

							phiCongruenceClasses.add(srcSlot, srcSlot);
						}

						phis.put(store, bb);
						continue;
					}
				}

				List<IMNode> flattened = IMCodeFlattener.getFlattenedCode(root);
				ListIterator<IMNode> it
				= flattened.listIterator(flattened.size());

				// Visit all instructions of the current node (in reverse order
				// as well)
				while (it.hasPrevious()) {

					IMNode node = it.previous();

					// Use
					if (node.usesSlot()) {

						IMSlot slot = node.getIMSlot();
						live.add(slot);
						updateGraph(interferences, live);

						// If the root is a conditional branch, add its slot
						// operands to the extended LiveOut set
						if (root instanceof IMConditionalBranch
								|| root instanceof IMLookupSwitch
								|| root instanceof IMTableSwitch)
							extendedLiveOuts.add(bb, slot);
					}
				}
			}
		}

		final Multimap<IMSlot, IMSlot> unresolvedNeighbourMap
		= new HashMultimap<IMSlot, IMSlot>();

		for (Entry<IMStoreLocalVariable, IMBasicBlock> entry
				: phis.entrySet()) {

			SortedSet<IMSlot> candidateResourceSet = new TreeSet<IMSlot>();
			List<IMSlot>         resources = new ArrayList<IMSlot>();
			IMStoreLocalVariable store     = entry.getKey();
			IMSlot               x0        = store.getIMSlot();
			IMBasicBlock         l0        = entry.getValue();
			IMBasicBlockList     preds     = l0.getPred();
			int                  predCount = preds.length();
			IMPhi                phiInst   = (IMPhi) store.getOperand();
			resources.add(x0);
			for (int i = 0; i < predCount; ++i) {
				resources.add(phiInst.getSourceVar(i));
			}
			for (IMSlot xi : resources) {
				unresolvedNeighbourMap.removeAll(xi);
			}
			int n = resources.size();

			for (int i = 0; i < n; ++i) {

				IMSlot xi = resources.get(i);
				IMBasicBlock li = getBasicBlock(l0, i);
				Collection<IMSlot> livei = (i == 0) ? li.getLiveInSlots()
						: extendedLiveOuts.get(li);

				for (int j = i + 1; j < n; ++j) {

					IMSlot xj = resources.get(j);
					if (xi == xj)
						continue;

					IMBasicBlock       lj    = getBasicBlock(l0, j);
					Collection<IMSlot> livej = extendedLiveOuts.get(lj);

					// According to Boissinot ("Revisiting Out-of-SSA
					// Translation for Correctness, Code Quality, and
					// Effciency"), we also have to consider all slots used in
					// the conditional branch at the end of Li and Lj in the
					// intersection, hence the use of extendedLiveOuts.
					processInterferences(xi, xj, interferences,
							phiCongruenceClasses, livei, livej,
							candidateResourceSet, unresolvedNeighbourMap);
				}
			}

			// Process the unresolved resources (Case 4) as described in Section
			// 4.3
			SortedSet<IMSlot> sorted
			= new TreeSet<IMSlot>(new Comparator<IMSlot>() {
				@Override
				public int compare(IMSlot o1, IMSlot o2) {
					int neighbourCount1 = unresolvedNeighbourMap.get(o1).size();
					int neighbourCount2 = unresolvedNeighbourMap.get(o2).size();
					if (neighbourCount1 != neighbourCount2)
						return neighbourCount2 - neighbourCount1;
					return o1.compareTo(o2);
				}
			});
			sorted.addAll(unresolvedNeighbourMap.keySet());
			for (IMSlot slot : sorted) {
				if (candidateResourceSet.contains(slot))
					continue;
				for (IMSlot neighbour : unresolvedNeighbourMap.get(slot)) {
					if (candidateResourceSet.contains(neighbour) == false) {
						candidateResourceSet.add(slot);
						break;
					}
				}
			}

			for (IMSlot xi : candidateResourceSet) {

				int phiPos = phiInst.getBCPosition();

				if (xi != x0) {

					// Note that "k" in the following code is equivalent to
					// "k - 1" in the paper.
					for (int k = 0; k < predCount; ++k) {

						if (phiInst.getSourceVar(k) != xi)
							continue;

						// Insert a copy inst: xnew_i = xi at the end of Lk
						IMSlot xnew_i = frame.createNewVariableSSA(xi);
						IMReadLocalVariable nRead = method.readLocal(phiPos, xi);
						IMStoreLocalVariable nStore
						= method.createStoreVariable(xnew_i, nRead, phiPos);
						IMBasicBlock lk = preds.at(k);
						opts.vverbose("++ Inserting copy " + xnew_i + " = " + xi
								+ " at end of " + lk
								+ " for Phi-function defining " + x0);
						lk.appendBeforeBranch(nStore);
						xi.use(lk, nRead);
						xnew_i.define(lk, nStore);

						// Replace xi with xnew_i in phiInst
						xi.deleteUse(phiInst.getSource(k));
						phiInst.setSourceVar(k, xnew_i);
						resources.set(k + 1, xnew_i);
						updatePhiInSlots(xi, xnew_i, lk, phis);

						// Add xnew_i in phiCongruenceClass[xnew_i]
						phiCongruenceClasses.add(xnew_i, xnew_i);

						// LiveOut[Lk] += xnew_i
						lk.addLiveOut(xnew_i);
						extendedLiveOuts.add(lk, xnew_i);

						// if (for Lj an immediate successor of Lk, xi not in
						//         LiveIn[Lj] and not used in a phi instruction
						//         associated with Lk in Lj)
						//    LiveOut[Lk] -= xi
						boolean xiDead = true;
						for (IMBasicBlock lj : lk.getSucc()) {
							if (lj.getLiveInSlots().contains(xi)
									|| lj.getPhiInSlots().contains(xi)) {
								xiDead = false;
								break;
							}
						}
						if (xiDead == true) {
							lk.removeLiveOut(xi);
							extendedLiveOuts.remove(lk, xi);
						}

						// Build interference edges between xnew_i and
						// LiveOut[Lk]
						for (IMSlot slot : extendedLiveOuts.get(lk)) {
							interferences.addEdge(xnew_i, slot);
						}
					}

				} else {

					// Insert a copy inst: x0 = xnew_0 at the beginning of L0
					IMReadLocalVariable nRead = store.getReadOperation();
					IMStoreLocalVariable nStore
					= method.createStoreVariable(x0, nRead, phiPos);
					insertAfter(store, nStore);

					// Replace x0 with xnew_0 as the target in phiInst
					IMSlot xnew_0 = frame.createNewVariableSSA(x0);
					opts.vverbose("++ Inserting copy " + x0 + " = " + xnew_0
							+ " at beginning of " + l0);
					store.setIMSlot(xnew_0);
					resources.set(0, xnew_0);
					nRead.setIMSlot(xnew_0);
					xnew_0.define(l0, store);
					xnew_0.use(l0, nRead);
					x0.redefine(l0, nStore);

					// Add xnew_0 in phiCongruenceClass[xnew_0]
					phiCongruenceClasses.add(xnew_0, xnew_0);

					// LiveIn[L0] -= x0
					l0.removeLiveIn(x0);

					// LiveIn[L0] += xnew_0
					l0.addLiveIn(xnew_0);

					// Build interference edges between xnew_0 and LiveIn[L0]
					for (IMSlot slot : l0.getLiveInSlots()) {
						interferences.addEdge(xnew_0, slot);
					}
				}
			}

			// Merge phiCongruenceClasses for all resources in phiInst
			Set<IMSlot> currentPhiCongruenceClass = new HashSet<IMSlot>();
			for (IMSlot xi : resources) {
				currentPhiCongruenceClass.addAll(phiCongruenceClasses.get(xi));
			}
			for (IMSlot xi : resources) {
				phiCongruenceClasses.addAll(xi, currentPhiCongruenceClass);
			}
		}

		// Nullify phi congruence classes that contain only singleton resources
		for (IMSlot slot
				: new ArrayList<IMSlot>(phiCongruenceClasses.keySet())) {
			if (phiCongruenceClasses.get(slot).size() == 1)
				phiCongruenceClasses.removeAll(slot);
		}

		// Eliminate redundant copies
		coalesce(method, interferences, phiCongruenceClasses);

		Multimap<IMSlot, IMNode> otherDefs = new HashMultimap<IMSlot, IMNode>();

		// Process all Phi-functions
		for (Entry<IMStoreLocalVariable, IMBasicBlock> entry
				: phis.entrySet()) {

			// Find a suitable replacement slot
			// In most cases, this is the destination slot of the Phi-function.
			// However, if one of the operands is a function argument, it is
			// essential to use it as the replacement slot in order to avoid
			// breaking formal arguments.
			IMStoreLocalVariable store       = entry.getKey();
			IMBasicBlock         bb          = entry.getValue();
			int                  predCount   = bb.getPred().length();
			IMPhi                phiInst     = (IMPhi) store.getOperand();
			IMSlot               replacement = null;
			for (int i = 0; i < predCount; ++i) {
				IMSlot slot = phiInst.getSourceVar(i);
				if (frame.isArgument(slot) == false)
					continue;
				if (replacement != null && slot != replacement) // Impossible?
					throw new CompileException("Phi-function with interfering "
							+ "operands detected");
				replacement = slot;
			}
			IMSlot dest = store.getIMSlot();
			if (replacement == null)
				replacement = dest;

			opts.verbose("++ Replacing all slots with " + replacement
					+ " in Phi-function defining " + dest);

			// Replace all operands with the replacement slot and remove the
			// Phi-function
			for (int i = 0; i < predCount; ++i) {
				replaceSlot(phiInst.getSourceVar(i), replacement, otherDefs);
				if (replacement.hasNoUses() == false)
					replacement.deleteUse(phiInst.getSource(i));
			}
			replaceSlot(dest, replacement, otherDefs);
			bb.removeNode(store);
		}

		// Remove stores whose destination slot is unused
		for (IMBasicBlock bb : method) {
			for (IMNode root : bb) {
				if (root instanceof IMStoreLocalVariable == false
						|| root.getIMSlot().hasNoUses() == false)
					continue;
				IMNode operand = ((IMStoreLocalVariable) root).getOperand();
				if (operand.removeIfUnUsed() == false)
					root.insert(operand);
				bb.removeNode(root);
			}
		}

		frame.opt_vars();
		return true;
	}

	private static void updateGraph(UndirectedGraph<IMSlot> interferences,
			Set<IMSlot> live) {
		List<IMSlot> copy = new ArrayList<IMSlot>(live);
		int count = copy.size();
		for (int i = 0; i < count; ++i) {
			for (int j = i + 1; j < count; ++j) {
				IMSlot xi = copy.get(i);
				IMSlot xj = copy.get(j);
				if (interferences.isEdge(xi, xj) == false)
					interferences.addEdge(xi, xj);
			}
		}
	}

	private static Collection<IMNode> reverse(Iterable<IMNode> list) {
		Deque<IMNode> lifo = new ArrayDeque<IMNode>();
		for (IMNode node : list) {
			lifo.push(node);
		}
		return lifo;
	}

	/**
	 * Returns the basic block associated with a resource number.
	 * @param l0    The current basic block.
	 * @param index The resource number.
	 * @return
	 */
	private static IMBasicBlock getBasicBlock(IMBasicBlock l0, int index) {
		if (index == 0)
			return l0;
		return l0.getPred().at(index - 1);
	}

	private static void processInterferences(IMSlot xi, IMSlot xj,
			UndirectedGraph<IMSlot> interferences,
			Multimap<IMSlot, IMSlot> phiCongruenceClasses,
			Collection<IMSlot> livei, Collection<IMSlot> livej,
			Set<IMSlot> candidateResourceSet,
			Multimap<IMSlot, IMSlot> unresolvedNeighbourMap) {

		Collection<IMSlot> pccxi = phiCongruenceClasses.get(xi);
		Collection<IMSlot> pccxj = phiCongruenceClasses.get(xj);

		for (IMSlot yi : pccxi) {

			for (IMSlot yj : pccxj) {

				// Check for interference between yi and yj
				if (yi == yj
						|| interferences.isEdge(yi, yj) == false)
					continue;

				// Determine what copies needed to break the interference
				// between xi and xj using the four cases described in Section
				// 4.3
				if (doIntersect(pccxi, livej)) {
					if (doIntersect(pccxj, livei) == false) { // Case 1
						candidateResourceSet.add(xi);
					} else {                                  // Case 3
						candidateResourceSet.add(xi);
						candidateResourceSet.add(xj);
					}
				} else {
					if (doIntersect(pccxj, livei)) {          // Case 2
						candidateResourceSet.add(xj);
					} else {                                  // Case 4
						unresolvedNeighbourMap.add(xj, xi);
						unresolvedNeighbourMap.add(xi, xj);
					}
				}
			}
		}
	}

	private static boolean
	doIntersect(Collection<IMSlot> a, Collection<IMSlot> b) {
		Set<IMSlot> intersection = new HashSet<IMSlot>(a);
		intersection.retainAll(b);
		return !intersection.isEmpty();
	}

	private static void updatePhiInSlots(IMSlot oldSlot, IMSlot newSlot,
			IMBasicBlock bb, Map<IMStoreLocalVariable, IMBasicBlock> phis)
	throws CompileException {
		bb.addPhiIn(newSlot);
		for (Entry<IMStoreLocalVariable, IMBasicBlock> entry
				: phis.entrySet()) {
			if (entry.getValue() != bb)
				continue;
			IMNode operand =
				((IMStoreLocalVariable) entry.getKey()).getOperand();
			IMBasicBlockList preds = bb.getPred();
			for (int i = 0; i < preds.length(); ++i) {
				if (((IMPhi) operand).getSourceVar(i) == oldSlot)
					return;
			}
		}
		bb.removePhiIn(oldSlot);
	}

	private static void insertAfter(IMNode prev, IMNode node) {
		node.append(prev.next());
		prev.append(node);
	}

	private static void coalesce(IMMethod method,
			UndirectedGraph<IMSlot> interferences,
			Multimap<IMSlot, IMSlot> phiCongruenceClasses)
	throws CompileException {

		for (IMBasicBlock bb : method) {
			for (IMNode node : bb) {
				if (node instanceof IMStoreLocalVariable == false)
					continue;
				IMStoreLocalVariable store   = (IMStoreLocalVariable) node;
				IMNode               operand = store.getOperand();
				if (operand.isVariable() == false)
					continue;
				IMSlot x = store.getIMSlot();
				IMSlot y = operand.getIMSlot();
				Collection<IMSlot> pccx = phiCongruenceClasses.get(x);
				Collection<IMSlot> pccy = phiCongruenceClasses.get(y);
				if (pccx.contains(y)) {
					removeCopy(store, bb);
				} else if (pccx.isEmpty()) {
					boolean canRemove = true;
					if (pccy.isEmpty() == false) {
						for (IMSlot slot : pccy) {
							if (slot != y && interferences.isEdge(slot, x)) {
								canRemove = false;
								break;
							}
						}
					}
					if (canRemove)
						removeCopy(store, bb);
				} else {
					boolean canRemove = true;
					if (pccy.isEmpty()) {
						for (IMSlot slot : pccx) {
							if (slot != x && interferences.isEdge(slot, y)) {
								canRemove = false;
								break;
							}
						}
					} else {
						for (IMSlot csx : pccx) {
							for (IMSlot csy : pccy) {
								if ((csx != x || csy != y)
										&& interferences.isEdge(csx, csy)) {
									canRemove = false;
									break;
								}
							}
						}
					}
					if (canRemove)
						removeCopy(store, bb);
				}
			}
		}
	}

	private static void replaceSlot(IMSlot oldSlot, IMSlot newSlot,
			Multimap<IMSlot, IMNode> otherDefs) {

		if (oldSlot == newSlot)
			return;

		// Replace SSA definition
		IMNode def = oldSlot.getDefiningStatement();
		if (def != null) {
			def.setIMSlot(newSlot);
			if (otherDefs != null)
				otherDefs.add(newSlot, def);
		}

		// Replace other definitions
		// These additional definitions arise in the process of the Phi-function
		// removal.
		if (otherDefs != null) {
			for (IMNode otherDef : otherDefs.get(oldSlot)) {
				if (otherDef == def)
					continue;
				otherDef.setIMSlot(newSlot);
				otherDefs.add(newSlot, otherDef);
			}
			otherDefs.removeAll(oldSlot);
		}

		// Replace uses
		for (IMNode use : oldSlot.getUses()) {
			use.setIMSlot(newSlot);
			newSlot.use(oldSlot.getUseBB(use), use);
		}
		oldSlot.clearDefUseChain();
	}

	private static void removeCopy(IMStoreLocalVariable store, IMBasicBlock bb)
	throws CompileException {
		IMSlot         destSlot = store.getIMSlot();
		IMSlot         srcSlot  = store.getOperand().getIMSlot();
		BuilderOptions opts     = bb.getMethod().getOptions();
		opts.verbose("++ Removing redundant copy: " + destSlot + " = "
				+ srcSlot);
		bb.removeNode(store);
		replaceSlot(destSlot, srcSlot, null);
	}

}
