/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.compiler.transform;

import static keso.util.ReverseListIterator.reverse;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import keso.compiler.BuilderOptions;
import keso.compiler.ClassStore;
import keso.compiler.CompileException;
import keso.compiler.MethodPass;
import keso.compiler.PassInfo;
import keso.compiler.analysis.ExtendedEscapeAnalysis;
import keso.compiler.analysis.LivenessAnalysis;
import keso.compiler.analysis.connectiongraph.CGLocalReferenceNode;
import keso.compiler.analysis.connectiongraph.CGNode;
import keso.compiler.analysis.connectiongraph.CGObjectNode;
import keso.compiler.analysis.connectiongraph.EscapeState;
import keso.compiler.imcode.IMAStoreLocalVariable;
import keso.compiler.imcode.IMAllocation;
import keso.compiler.imcode.IMBasicBlock;
import keso.compiler.imcode.IMCodeFlattener;
import keso.compiler.imcode.IMMethod;
import keso.compiler.imcode.IMNode;
import keso.compiler.imcode.IMSlot;

public final class StackAllocation extends MethodPass {
	private int		stackAllocationCount	= 0;
	private int		allocationCount			= 0;
	private int		livenessOverlapCount	= 0;

	private boolean	debug					= false;

	@Override
	public boolean runOnRepository(ClassStore repository) throws CompileException {
		boolean result;

		result = super.runOnRepository(repository);

		System.out.printf("%d of %d (%.2f%%) allocations in this repository are eligible for stack allocation%n",
				stackAllocationCount - livenessOverlapCount, allocationCount,
				((stackAllocationCount - livenessOverlapCount) * 100.0) / allocationCount);
		System.out.printf(
				"%d allocations have overlapping liveness regions and are thus not considered for stack allocation%n",
				livenessOverlapCount);

		return result;
	}

	@Override
	public boolean runOnMethod(IMMethod method) throws CompileException {
		if (debug)
			System.out.println("MTD: " + method.getAlias());

		for (IMBasicBlock bb : method) {
			if (debug)
				System.out.println(" BB: " + bb.getShortLabel());

			Set<IMSlot> live = new HashSet<IMSlot>();
			// fill the set of alive objects with the object alive at the end of the basic block
			for (IMSlot liveOutSlot : bb.getLiveOutSlots())
				if (liveOutSlot.isReference())
					live.add(liveOutSlot);

			// walk over the root nodes in reverse order
			List<IMNode> rootNodeList = new LinkedList<IMNode>();
			for (IMNode root : bb)
				rootNodeList.add(root);

			// start walking over the node list, starting at the end
			for (IMNode root : reverse(rootNodeList)) {
				for (IMNode node : reverse(IMCodeFlattener.getFlattenedCode(root))) {
					boolean changesLiveSet = false;

					if (node.usesSlot() && node.getIMSlot().isReference()) {
						changesLiveSet = true;
						// if a node uses a slot make sure it's in the alive list
						// we're only handling references
						live.add(node.getIMSlot());
					} else if (node instanceof IMAStoreLocalVariable && node.getIMSlot().isReference()) {
						changesLiveSet = true;
						// upon assignment, remove the node from the list of alive nodes
						live.remove(node.getIMSlot());
					} else if (node.isNew()) {
						// for allocations, check whether this very allocation site is reachable in the connection graph
						// from any of the currently live objects
						IMAllocation allocation = (IMAllocation) node;
						CGObjectNode objectNode = CGObjectNode.getObjectNodeForAllocation(allocation, method);
						allocation.setEscapeState(objectNode.getEscapeState());

						if (objectNode.getEscapeState() == EscapeState.Local) {

							// check for potentially overlapping liveness periods of objects created by this allocation
							boolean reachable = false;

							for (IMSlot liveSlot : live) {
								CGLocalReferenceNode localReferenceNode = CGLocalReferenceNode
										.getLocalReferenceNodeForSlot(liveSlot);

								Stack<CGNode> workList = new Stack<CGNode>();
								Set<CGNode> visited = new HashSet<CGNode>();
								workList.push(localReferenceNode);

								while (!reachable && !workList.empty()) {
									CGNode workNode = workList.pop();

									if (workNode == objectNode) {
										reachable = true;
										break;
									}

									visited.add(workNode);

									for (CGNode childNode : workNode.getDescendants()) {
										if (!visited.contains(childNode))
											workList.push(childNode);
									}
								}

								if (reachable)
									break;
							}

							if (reachable) {
								// this allocation can't be stack-allocated!
								allocation.setLivenessOverlap(true);

								if (debug)
									System.out
											.println("   OVERLAP in " + method.getAlias() + ": " + allocation.dumpBC());
								livenessOverlapCount++;
							}
						}

						allocation.setEscapeState(objectNode.getEscapeState());

						if (objectNode.getEscapeState() == EscapeState.Local) {
							stackAllocationCount++;
						}

						allocationCount++;
					}

					if (debug) {
						System.out.println("  OL: " + node.dumpBC());
						if (changesLiveSet)
							System.out.println("   live: " + live.toString());
					}
				}
			}
		}

		return true;
	}

	@Override
	public boolean isEnabled(BuilderOptions opts) {
		debug = opts.hasOption("_X_eea_stack_alloc_dbg");
		return opts.hasOption("ext_escape_analysis");
	}

	@Override
	public void getPassInfo(PassInfo info, BuilderOptions opts) {
		info.addDependency(ExtendedEscapeAnalysis.class);
		info.addDependency(LivenessAnalysis.class);
		info.addAntiDependency(SSADeconstruction.class);
		info.setDoDump(opts.hasOption("dump_stack_alloc"));
	}

	@Override
	public String getName() {
		return "Stack allocation transformation";
	}

}
