/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/**
 * (c)
 *
 * Copyright (C) 2006-2011 Christian Wawersich, Michael Stilkerich, Christoph Erhardt
 *
 * This file is part of the KESO Java Runtime Environment.
 *
 * KESO is free software: you can redistribute it and/or modify it under the terms of the Lesser GNU General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * KESO is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with KESO. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Please contact keso@cs.fau.de for more info.
 *
 * (c)
 **/

package keso.util;

import java.io.FileNotFoundException;
import java.util.*;
import java.util.Map.Entry;

public final class DirectedGraph<N> {

	private Set<N>			nodes			= new HashSet<N>();
	private Multimap<N, N>	successors		= new HashUniqueMultimap<N, N>();
	private Multimap<N, N>	predecessors	= new HashUniqueMultimap<N, N>();

	public void addNode(N node) {
		nodes.add(node);
	}

	public void addEdge(N from, N to) {
		addNode(from);
		addNode(to);
		successors.add(from, to);
		predecessors.add(to, from);
	}

	public void removeEdge(N from, N to) {
		successors.remove(from, to);
		predecessors.remove(to, from);
	}

	public boolean isEdge(N from, N to) {
		return successors.get(from).contains(to);
	}

	public Collection<N> getNodes() {
		return Collections.unmodifiableCollection(nodes);
	}

	public Collection<N> getSuccessors(N node) {
		return successors.get(node);
	}

	public int getSuccessorCount(N node) {
		return successors.get(node).size();
	}

	public Collection<N> getPredecessors(N node) {
		return predecessors.get(node);
	}

	public int getPredecessorCount(N node) {
		return predecessors.get(node).size();
	}

	public void clear() {
		nodes.clear();
		successors.clear();
		predecessors.clear();
	}

	public DirectedGraph<N> invertedGraph() {
		DirectedGraph<N> invertedGraph = new DirectedGraph<N>();
		for (N node : nodes) {
			invertedGraph.addNode(node);
			for (N to : getSuccessors(node)) {
				invertedGraph.addEdge(to, node);
			}
		}
		return invertedGraph;
	}

	public List<N> topSort() throws GraphException {
		return topSort(null);
	}

	public List<N> topSort(Comparator<N> comparator) throws GraphException {

		// If a comparator is given, guarantee a reproducible result by using a
		// TreeMap - otherwise use a HashMap
		Map<N, Integer> predCounts;
		if (comparator == null)
			predCounts = new HashMap<N, Integer>();
		else
			predCounts = new TreeMap<N, Integer>(comparator);

		// Store each node's predecessor count in a map
		for (N node : nodes) {
			predCounts.put(node, getPredecessorCount(node));
		}
		List<N> result = new ArrayList<N>();

		while (predCounts.isEmpty() == false) {

			// Remove the first node with zero predecessors and add it to the
			// result list
			N head = null;
			for (Entry<N, Integer> entry : predCounts.entrySet()) {
				if (entry.getValue() == 0) {
					head = entry.getKey();
					predCounts.remove(head);
					result.add(head);
					break;
				}
			}
			if (head == null)
				throw new GraphException("Circular dependency detected", predCounts.keySet());

			// Update successors
			for (N successor : getSuccessors(head)) {
				predCounts.put(successor, predCounts.get(successor) - 1);
			}
		}

		return result;
	}

	/**
	 * Index used in the computation of strongly connected components in this graph using Tarjan's algorithm.
	 */
	private int	tarjanIndex;

	/**
	 * Runs Tarjan's strongly connected components algorithm on this directed graph and returns a double-ended queue
	 * containing the connected components as a set; the queue is in reverse topological order of the components (or in
	 * non-reverse if you iterate over it from the back).
	 *
	 * @return a queue of sets representing the strongly connected components
	 */
	public Deque<Set<N>> tarjan() {
		Deque<Set<N>> result = new ArrayDeque<Set<N>>(nodes.size() / 2);
		Map<N, Integer> index = new HashMap<N, Integer>(nodes.size());
		Map<N, Integer> low = new HashMap<N, Integer>(nodes.size());
		HashStack<N> stack = new HashStack<N>(nodes.size());
		tarjanIndex = 0;

		for (N node : nodes) {
			if (!index.containsKey(node)) {
				tarjan(node, result, stack, index, low);
			}
		}

		return result;
	}

	private void tarjan(N node, Deque<Set<N>> result, HashStack<N> stack, Map<N, Integer> index, Map<N, Integer> low) {
		index.put(node, tarjanIndex);
		low.put(node, tarjanIndex);
		tarjanIndex++;

		stack.push(node);

		for (N successor : successors.get(node)) {
			if (!index.containsKey(successor)) {
				tarjan(successor, result, stack, index, low);
				low.put(node, Math.min(low.get(node), low.get(successor)));
			} else if (stack.contains(successor)) {
				low.put(node, Math.min(low.get(node), index.get(successor)));
			}
		}

		if (index.get(node) == low.get(node)) {
			Set<N> component = new HashSet<N>();
			N current;
			do {
				current = stack.pop();
				component.add(current);
			} while (current != node);
			result.addLast(component);
		}
	}

	public void dump(String fileName) throws DotWriterException, FileNotFoundException {
		DotWriter<N> dw = new DotWriter<N>(fileName);
		dw.initGraph("DirectedGraph");
		for (N node : getNodes()) {
			dw.addNode(node, (node != null) ? node.toString() : "<null>");
			for (N successor : getSuccessors(node)) {
				dw.addEdge(node, successor);
			}
		}
		dw.finalizeGraph();
	}

}
