/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.util;

import java.io.*;
import java.util.*;

/** Directed graph (.dot file) writer. */
public final class DotWriter<T> {

	private final PrintStream out;
	private int nodeCount;
	private Map<T, Integer> nodeIds;
	private int subLevel;

	/**
	 * Constructor.
	 * @param fileName Name of the .dot output file.
	 * @throws FileNotFoundException
	 */
	public DotWriter(String fileName) throws FileNotFoundException {
		out = new PrintStream(fileName);
	}

	/**
	 * Initializes a graph.
	 * This method must be called before any nodes or edges can be added.
	 * @param name Name of the graph.
	 * @throws DotWriterException
	 */
	public void initGraph(String name) throws DotWriterException {
		if (nodeIds != null)
			throw new DotWriterException("Graph already initialized");
		if (name == null || name.length() == 0)
			throw new DotWriterException("Graph must have a name");
		writeLine("digraph " + name + " {");
		nodeIds = new HashMap<T, Integer>();
	}

	/**
	 * Finalizes the current graph.
	 * This method must be called after all nodes and edges have been added.
	 * @throws DotWriterException
	 */
	public void finalizeGraph() throws DotWriterException {
		checkInitialization();
		while (subLevel-- > 0)
			finalizeSubGraph();
		nodeIds = null;
		writeLine("}");
		out.close();
	}

	/**
	 * Initializes a sub-graph.
	 * @param name Name of the sub-graph.
	 * @throws DotWriterException
	 */
	public void initSubGraph(String name) throws DotWriterException {
		checkInitialization();
		if (name == null || name.length() == 0)
			throw new DotWriterException("Sub-graph must have a name");
		writeLine("subgraph " + name + " {");
		++subLevel;
	}

	/**
	 * Finalizes the current sub-graph.
	 * @throws DotWriterException
	 */
	public void finalizeSubGraph() throws DotWriterException {
		checkInitialization();
		if (subLevel <= 0)
			throw new DotWriterException("Not within a sub-graph");
		--subLevel;
		writeLine("}");
	}

	/**
	 * Sets the style for all nodes within the current graph or sub-graph.
	 * @param style Node style string.
	 * @throws DotWriterException
	 */
	public void setNodeStyle(String style) throws DotWriterException {
		writeLine("node [" + style + "];");
	}

	/**
	 * Adds a node to the current graph.
	 * @param node  Node object.
	 * @param label Node label.
	 * @throws DotWriterException
	 */
	public void addNode(T node, String label) throws DotWriterException {
		addNode(node, label, null);
	}

	/**
	 * Adds a node to the current graph.
	 * @param node       Node object.
	 * @param label      Node label.
	 * @param attributes Additional node attributes.
	 * @throws DotWriterException
	 */
	public void addNode(T node, String label, String attributes)
	throws DotWriterException {
		checkInitialization();
		String line = "n" + getNodeID(node) + " [label=\"" + label + '"';
		if (attributes != null)
			line += " " + attributes;
		line += "];";
		writeLine(line);
	}

	/**
	 * Adds an edge to the current graph.
	 * @param from Node object from which the edge comes.
	 * @param to   Node object to which the edge leads.
	 * @throws DotWriterException
	 */
	public void addEdge(T from, T to) throws DotWriterException {
		addEdge(from, to, null);
	}

	/**
	 * Adds an edge to the current graph.
	 * @param from       Node object from which the edge comes.
	 * @param to         Node object to which the edge leads.
	 * @param attributes Additional node attributes.
	 * @throws DotWriterException
	 */
	public void addEdge(T from, T to, String attributes)
	throws DotWriterException {
		checkInitialization();
		String line = "n" + getNodeID(from) + " -> n" + getNodeID(to);
		if (attributes != null)
			line += " [" + attributes + ']';
		line += ';';
		writeLine(line);
	}

	/**
	 * Writes a line with correct indentation to the .dot file.
	 * @param line Line to be written.
	 * @throws DotWriterException
	 */
	public void writeLine(String line) throws DotWriterException {
		if (nodeIds != null) {
			out.print('\t');
			for (int i = 0; i < subLevel; ++i) {
				out.print('\t');
			}
		}
		out.println(line);
	}

	/**
	 * Throws an exception if no graph has been initialized.
	 * @throws DotWriterException
	 */
	private void checkInitialization() throws DotWriterException {
		if (nodeIds == null)
			throw new DotWriterException("Graph not initialized");
	}

	/**
	 * Returns a unique ID for a node.
	 * @param node Node object.
	 * @return Unique ID for the node.
	 * @throws DotWriterException
	 */
	private int getNodeID(T node) throws DotWriterException {
		checkInitialization();
		Integer id = nodeIds.get(node);
		if (id == null) {
			id = nodeCount++;
			nodeIds.put(node, id);
		}
		return id;
	}

}
