/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.util;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.Stack;

/**
 * Stack implementation that provides O(1) {@code contains}-operations by keeping track of the elements on the stack in
 * a hash set. Duplicate values cannot be inserted.
 */
public class HashStack<E> {
	private final Stack<E>	stack;
	private final Set<E>	set;

	/**
	 * Creates a new HashStack
	 */
	public HashStack() {
		this.stack = new Stack<E>();
		this.set = new HashSet<E>();
	}

	/**
	 * Creates a new HashStack with a given initial capacity
	 *
	 * @param capacity
	 *            The initial capacity of the hash set
	 */
	public HashStack(int capacity) {
		this.stack = new Stack<E>();
		this.set = new HashSet<E>(capacity);
	}

	/**
	 * Tries to push a new element to the top of the stack and returns {@code true}, if the element was inserted.
	 * Returns {@code false}, if this element was already on the stack.
	 *
	 * @param item
	 *            the element to push to the stack
	 * @return true, if successful, false otherwise
	 */
	public boolean push(E item) {
		if (!set.add(item)) {
			return false;
		}
		stack.push(item);
		return true;
	}

	/**
	 * Retrieves and removes the top item from the stack. Returns {@code null}, if the stack is empty.
	 */
	public E pop() {
		E item = stack.pop();
		set.remove(item);
		return item;
	}

	/**
	 * Retrieves, but does not remove the top item from the stack. Returns {@code null}, if the stack is empty.
	 */
	public E peek() {
		return stack.peek();
	}

	/**
	 * Removes all elements from the stack.
	 */
	public void clear() {
		stack.clear();
		set.clear();
	}

	/**
	 * Returns true, if the given item is on the stack, false otherwise.
	 */
	public boolean contains(E item) {
		return set.contains(item);
	}

	/**
	 * Returns true, if all given items are on the stack, false otherwise
	 */
	public boolean containsAll(Collection<?> items) {
		return set.containsAll(items);
	}
}
