/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package keso.util;

import java.util.*;

public final class IntegerHashtable<T> {

	private final class IntHashtableEnumeration implements Iterable<T> {

		private final class Itr implements Iterator<T> {

			private int index = 0;
			private T   next;

			public Itr() {
				step();
			}

			@Override
			public boolean hasNext() {
				return (next != null);
			}

			@Override
			public T next() {
				if (hasNext() == false)
					throw new NoSuchElementException();
				T obj = next;
				step();
				return obj;
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException();
			}

			private void step() {
				while (index < values.length) {
					T obj = valueAt(index++);
					if (obj == null)
						continue;
					next = obj;
					return;
				}
				next = null;
			}

		}

		@Override
		public Iterator<T> iterator() {
			return new Itr();
		}

	}

	int[] keys;
	Object[] values;
	int capacity;

	private int load; // "real load" * 100
	private int mask;
	private int size;

	private static Object tombstone = new Object();

	public IntegerHashtable(int initialCapacity, int loadFactor) {
		this.load = loadFactor;
		size = 0;
		for (capacity = 32; capacity < initialCapacity; capacity <<= 1) ;
		keys = new int[capacity];
		values = new Object[capacity];
		mask = (capacity - 1);
	}

	public IntegerHashtable(int initialCapacity) {
		this(initialCapacity, 75);
	}

	public IntegerHashtable() {
		this(32);
	}

	public T get(int key) {
		int hash = key & mask;

		for (int i = 0; i < capacity; i++) {
			int index = (hash + i) & mask;

			if (values[index] != null) {
				if (keys[index]==key && values[index]!=tombstone)
					return valueAt(index);
				// Continue looking through collisions.
				continue;
			}

			return null;
		}

		return null;
	}

	public int[] keys() {
		int[] ret = new int[size];

		int r=0;
		for (int i=0;i<keys.length && r<size;i++) {
			if (values[i]==null || values[i]==tombstone) continue;
			ret[r]=keys[i];
			r++;
		}

		return ret;
	}

	public int[] sortedKeys() {
		int[] arr = keys();
		Arrays.sort(arr);
		return arr;
	}

	public Iterable<T> getElements() {
		return new IntHashtableEnumeration();
	}

	public T put(int key, T value) {

		if (size * 100 > capacity * load) rehash();

		if (value == null)
			throw new NullPointerException();

		int hash = key & mask;

		for (int i = 0; i < capacity; i++) {
			int n = (hash + i) & mask;

			if (values[n] != null && values[n] != tombstone) {
				if (keys[n]==key) {
					// update entry

					T obj = valueAt(n);
					keys[n] = key;
					values[n] = value;

					return obj;
				}

				// collision
				continue;
			}

			// add entry
			keys[n] = key;
			values[n] = value;
			size++;

			return null;
		}

		// Never reached if load < 100 !!

		return null;
	}

	public T remove(int key) {
		int hash = key & mask;

		for (int i = 0; i < capacity; i++) {
			int n = (hash + i) & mask;

			if (values[n] == null)
				return null;

			if (values[n] == tombstone || (!(keys[n]==key)))
				continue;

			T value = valueAt(n);
			values[n] = tombstone;
			size--;

			return value;
		}

		return null;
	}

	public int size() { return size; }

	public boolean isEmpty() { return (size == 0); }

	public boolean containsKey(int key) {

		int hash = key & mask;

		for (int i = 0; i < capacity; i++)
		{
			int n = (hash + i) & mask;

			if (values[n] == null)
				return false;

			if ((values[n]!=tombstone) && (keys[n]==key))
				return true;
		}
		return false;
	}

	public boolean contains(T value) {
		if (value == null)
			return false;

		for (int i = 0; i < capacity; i++)
			if ((values[i] != null && values[i]!=tombstone)
			    && (values[i].equals(value)))
				return true;
		return false;
	}

	@Override
	public String toString() {
		StringBuilder buff = new StringBuilder("{");
		boolean first = true;

		for (int i = 0; i < values.length; i++)
			if (values[i] != null && values[i]!=tombstone) {
				if (first)
					first = false;
				else
					buff.append(", ");

				buff.append(String.valueOf(keys[i]));
				buff.append("=");
				buff.append(String.valueOf(values[i]));
			}
		buff.append("}");

		return buff.toString();
	}

	@SuppressWarnings("unchecked")
	private T valueAt(int index) {
		if (values[index]==tombstone) return null;
		return (T) values[index];
	}

	@SuppressWarnings("unchecked")
	private void rehash() {
		while (size*100 > capacity * load)
			capacity <<= 1;
		mask = (capacity - 1);

		int[] oldkeys = keys;
		Object[] oldvalues = values;
		keys = new int[capacity];
		values = new Object[capacity];
		size = 0;

		for (int i = 0; i < oldkeys.length; i++) {
			if (oldvalues[i] != null && oldvalues[i]!=tombstone) {
				put(oldkeys[i], (T) oldvalues[i]);
			}
		}
	}

}
