/*KESO--HEADER--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
#include "global.h"

object_t* keso_coffee_alloc(class_id_t class_id, obj_size_t size, obj_size_t roff GC_LOG_INFO_PARAMETER);

typedef struct _coffee_listel_s {
	 keso_gc_slotcnt_t size;        /* size in slots */
	 struct _coffee_listel_s *next; /* Points to the next listelement */
} coffee_listel_t;

/*KESO--CFILE--KESO*/

#define GC_LOCK() SuspendOSInterrupts()
#define GC_UNLOCK() ResumeOSInterrupts()
#define GC_LOCK_STACK()
#define GC_UNLOCK_STACK()
#define GC_LOCK_MUTATOR()
#define GC_UNLOCK_MUTATOR()


/*
 * Allocator function of the COFFEE Heap.
 *
 * Returns a pointer to a free memory area on the
 * heap of the current domain with @size bytes.
 *
 * Throws an error if no suitable memory block can
 * be found.
 */
object_t* keso_coffee_alloc(class_id_t class_id, obj_size_t size, obj_size_t roff GC_LOG_INFO_PARAMETER) {
	struct domain_coffee_t *domain;
	coffee_listel_t **prev_ptr;
	coffee_listel_t *slot;
	object_t* obj;

	domain = (struct domain_coffee_t *) KESO_CURRENT_DOMAIN;
	size = BYTES2SLOTS(size, domain->hd.slotSize);

#ifdef KESO_GCMODE_ENFORCEONDEMAND
	// FIXME
	do {
		GC_LOCK();

		GC_LOG_ALLOC_EVENT(KESO_CURRENT_TASK, GC_LOG_INFO, class_id, domain->heap_free, size);

		for (prev_ptr = &(domain->hd.freemem);*prev_ptr!=NULL;prev_ptr=&((*prev_ptr)->next)) {
			if ((*prev_ptr)->size>=size) break;
		}

		if (unlikely(*prev_ptr==NULL)) {
			GC_UNLOCK();
		        GC_LOG_OUT_OF_MEMORY(domain, domain->heap_free);
			WaitEvent(KESO_GCRun);
			ClearEvent(KESO_GCRun);
			if (domain->hd.freemem==NULL)
				KESO_THROW_ERROR("out_of_memory\n");
			continue;
		}
	} while (0);
#else
	GC_LOCK();

	GC_LOG_ALLOC_EVENT(KESO_CURRENT_TASK, GC_LOG_INFO, class_id, domain->heap_free, size);

	for (prev_ptr = &(domain->hd.freemem);*prev_ptr!=NULL;prev_ptr=&((*prev_ptr)->next)) {
		if ((*prev_ptr)->size >= size) break;
	}

	if (*prev_ptr==NULL) {
		 GC_UNLOCK();
		 GC_LOG_OUT_OF_MEMORY(domain, domain->heap_free);
		 KESO_THROW_ERROR("out_of_memory");
	}
#endif
	slot = *prev_ptr;

	if (slot->size > size) {
		unsigned int rest;
		slot->size -= size;
		rest = slot->size*domain->hd.slotSize;

#ifdef KESO_HEAP_HAS_TINY_SLOTS
		/* Don't split the block if the rest is smaller than
		   the free list header. Just give it away. */
		if (rest >= sizeof(coffee_listel_t)) {
			/* split block */
			slot = (coffee_listel_t*)((char*)slot+(slot->size * domain->hd.slotSize));
		} else {
			/* unlink */
			*prev_ptr = slot->next;

			/* rest of block already cleared */
			slot->size = 0;
			slot->next = NULL;
		}
#else
		/* split block */
		slot = (coffee_listel_t*)((char*)slot+(slot->size * domain->hd.slotSize));
#endif
	} else {
		/* unlink */
		*prev_ptr = slot->next;

		/* rest of block already cleared */
		slot->size = 0;
		slot->next = NULL;
	}

#ifdef DEBUG
	{ /* normally cleared during sweep */
	int sizeints,*tmp = (int *) slot;
	for(sizeints = size*domain->hd.slotSize/sizeof(int); sizeints>0; sizeints--)
		*(tmp++) = 0;
	}
#endif

	obj = (object_t*) ((object_t **) slot + roff);
	obj->class_id=class_id;
	/* assumes that bit 0 in colorbit is always set!! */
	obj->gcinfo = domain->colorbit;

	domain->hd.sasls += size * domain->hd.slotSize;
	domain->heap_free -= size * domain->hd.slotSize;

	GC_UNLOCK();

#ifdef KESO_GCMODE_LAZY
	if(0<keso_gctpaused) {
		keso_gctpaused=0;
		ActivateTask(keso_coffee_gc);
	}
#endif

	return obj;
}
