/*KESO--HEADER--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
#define KESO_GC_RESERVE (DOMAINID()->heap_size/4)
/*KESO--CFILE--KESO*/
#include "keso_types.h"
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
#include "object_array.h"
#endif

/* Index of the next free slot on the stack (one above
 * the top element of the stack) */
static keso_gc_slotcnt_t keso_coffee_sp;

/* very conservative estimation of the maximum number of objects that might be allocated */
static object_t *keso_coffee_stack[KESO_GC_WORKINGSTACK_SIZE];

/* internal functions */
static void keso_coffee_sweep(void);

static void keso_coffee_scanObject(object_t*);
static void keso_coffee_pushObject(object_t*);
static void chooseDomain(void);

/* This is the main GC function of the COFFEE Heap.
 * The task is chaining itself and run periodically,
 * collection garbage on the heap of one domain at
 * each run.
 */
#ifdef TARGET_TRICORE_CIAO
void KESODomZero::functionTaskkeso_coffee_gc() {
#else
TASK(keso_coffee_gc) {
#endif

	GC_LOG_ACTIVATION();

	while (1) {
		unsigned int i;
		unsigned int heap_size;

		/* choose the domain that will be scanned */
		chooseDomain();

		/* we stop the whole world for the mark phase */
		GC_LOCK();

		GC_LOG_START_EVENT(DOMAINID(), DOMAINID()->heap_free);

#ifdef KESO_NEED_FINALIZE
#if KESO_NUM_JAVADOMAINS > 1
		KESO_CURRENT_DOMAIN = (domain_t*) DOMAINID();
#endif
#endif

		/* heap begin and end are e.g. used to determine if an object reference is
		 * on the domain heap or not */
		heap_size = DOMAINID()->heap_size;
		heap_begin = (object_t*) HEAPDESC.heap_top;
		heap_end = (object_t*) ((char*) heap_begin + heap_size);

		keso_coffee_sp=0;

		/* toggle the color bit; all objects allocated up to know have been colored
		 * with the previous value; all discovered objects will be recolored with the
		 * new value
		 */
		DOMAINID()->colorbit ^= COLORBIT;

		/* record new allocations from this point */
		HEAPDESC.sasls = 0;

		/* Scan statics */
#if KESO_NUM_STATIC_REFS > 0
		for(i=0; i<KESO_NUM_STATIC_REFS; i++)
			keso_coffee_scanObject((object_t*)STATICREF(DOMAINID())[i]);
#endif

		/* push root set */
#if KESO_GC_ROOTSETSIZE > 0
		for(i=0; i<KESO_GC_ROOTSETSIZE; i++)
			keso_coffee_scanObject(HEAPDESC.rootset[i]);
#endif

		/* scan stacks and tasks */
#if NUM_BLOCKING_TASKS > 0
		for(i=0; i < NUM_BLOCKING_TASKS; i++) {
			KESO_CONST keso_stack_t *stack = blocking_tasks[i]->_stack;

			/* look through all stack partitions */
			for ( ; stack!=NULL; stack=stack->next) {
				/* scan llref if the current partition belongs to the gc'ed domain */
				if (stack->domain == (domain_t*) DOMAINID()) {
					keso_gc_walk_stack(stack->llrefs, keso_coffee_scanObject);
				}
			}
		}
#endif

		/* The root set is on the working stack now, scan and mark
		 * the objects until the working stack is empty */
		while (keso_coffee_sp>0) {
			keso_coffee_scanObject(keso_coffee_stack[--keso_coffee_sp]);
		}

		GC_LOG_MARK_END_EVENT(DOMAINID(), DOMAINID()->heap_free);

		/* keso_coffee_sweep heap */
		keso_coffee_sweep();

		GC_LOG_SWEEP_END_EVENT(DOMAINID(), DOMAINID()->heap_free);

		/* we wake up the whole world */

		GC_UNLOCK();

		KESO_SEND_GCRUN_EVENTS();

		/* clear bitmap for next run */
		keso_gc_clearBitmap();
	}
}

/* Choose the domain that needs garbage collection
 * the most of all managed domains, determined by
 * a metric called the "need" here.
 *
 * The need of a domain for a GC is determined by
 * the slots allocated since the last garbage collection (sasls)
 * compared to the remaining space. The remaining
 * space should not be less than the sasls plus a grace
 * area of 25% of the heap size. Thus, the "need" is calculated
 * freeslots - sasls - (heapsize>>2).
 * The function goes RoundRobin through the domain array and chooses
 * the first found domain that violates the above criterium (need<0). By
 * doing this instead of simply choosing the domain with the most severe
 * need, we can assure that all domains that have a need actually receive
 * a garbage collection (fairness to some degree).
 * If every domain has need>=0, the domain with the lowest need is taken.
 */
static void chooseDomain(void) {
	signed int need;
	const keso_umachineword_t prevDom=keso_curdom;

	while(1) {
		if(++keso_curdom >= MANAGEDDOMAINS) keso_curdom=0;

		need = DOMAINID()->heap_free - HEAPDESC.sasls - KESO_GC_RESERVE;

		/* The need of every domain is >0, sleep */
		if(prevDom == keso_curdom && (HEAPDESC.sasls==0 || need>=0)) {
				keso_gctpaused=1;
				HEAPDESC.sasls = 0;
				TerminateTask();
		}

		/* if need<0, this domain receives a GC treatment */
		if(need<0) return;

	}
}

static void keso_coffee_scanObject(object_t *obj) {
	signed int curref;

	if (obj==NULL) return;
	curref = refCnt(obj);

	while (curref<0) keso_coffee_pushObject( ((object_t**)obj)[curref++] );
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
	while (curref>0) keso_coffee_pushObject( ((object_array_t*)obj)->data[--curref] );
#endif
	/* Finally, mark the object's bits in the bitmap */
	keso_gc_markObj(obj);
}


/* Push a white object on the stack.
 * If the object is not white, this function
 * will do nothing.
 *
 * THIS IS THE ONLY PLACE WHERE OBJECTS ARE PUSHED
 * ON THE WORKING STACK.
 */
static void keso_coffee_pushObject(object_t *obj) {
	if(obj==NULL) return;
	ASSERTOBJ(obj);

	if(! OBJCOLORED(obj,DOMAINID()->colorbit)) {
		COLOROBJ(obj,DOMAINID()->colorbit);
		keso_coffee_stack[keso_coffee_sp++] = obj;
	}
}

/**
 * keso_coffee_sweep()
 *
 * - scan bitmap for free slots
 * - create a new free list
 */
static void keso_coffee_sweep() {
	const keso_gc_slotcnt_t numslots = DOMAINID()->heap_size / HEAPDESC.slotSize;
	coffee_listel_t **prev_free;
	keso_gc_slotcnt_t cur_slot = 0;


	/* reset free list */
	HEAPDESC.freemem = NULL;
	DOMAINID()->heap_free=0;

	prev_free = &HEAPDESC.freemem;

	while (cur_slot < numslots) {
		unsigned int sizeints;
		unsigned int blocksize;
		int *tmp;
		keso_gc_slotcnt_t size = keso_gc_findFreeBlock(&cur_slot, numslots - cur_slot);

		blocksize = size * HEAPDESC.slotSize;

		if (size > 0) {

			coffee_listel_t *block = (coffee_listel_t *) SLOT2ADDR(cur_slot);
			tmp = (int *) block;

			for(sizeints = blocksize/sizeof(int); sizeints>0; sizeints--)
				*(tmp++) = 0;

#ifdef KESO_HEAP_HAS_TINY_SLOTS
			/* Add the free block to the free list if the block size is great
			   enough to store the list header, otherwise give it away. */
			if (blocksize>=sizeof(coffee_listel_t)) {
				*prev_free = block;
				(*prev_free)->size = size;
				(*prev_free)->next = NULL;
				prev_free = &((*prev_free)->next);
				DOMAINID()->heap_free += blocksize;
			}
#else
			*prev_free = block;
			(*prev_free)->size = size;
			(*prev_free)->next = NULL;
			prev_free = &((*prev_free)->next);
			DOMAINID()->heap_free += blocksize;
#endif
			cur_slot += size;
		}
	}
}
