/*KESO--CFILE--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

typedef void (*callback_fkt_t)(object_t*);

#define LOCKBIT  1
#define REFONHEAP(obj) (heap_begin<=(obj) && (obj)<heap_end)

#define COLORBIT 2
#define OBJCOLORED(obj,cbit) ((obj)->gcinfo == (cbit))
#define COLOROBJ(obj,cbit) ((obj)->gcinfo = (cbit))

#define HEAPDESC (DOMAINID()->hd)
#define BYTES2SLOTS(bytes, slotsize) (((slotsize)-1+(bytes))/(slotsize))
#define ADDR2SLOT(_addr_) ((((unsigned int)(_addr_))-(unsigned int)heap_begin)/HEAPDESC.slotSize)
#define SLOT2ADDR(_slot_) ((char*)heap_begin + (_slot_) * HEAPDESC.slotSize)

/* internal variables */
static object_t *heap_begin, *heap_end;

#ifdef KESO_GCMODE_LAZY
/* The GC Task is autostarted; we init the gctpaused field with
 * 0 to prevent task activations during allocations in the startup
 * phase (not allowed by OSEK spec)
 */
static unsigned char keso_gctpaused;
#endif

/* index in domain[] to currently processed domainid,
 * this is NOT the domainid itself */
static keso_umachineword_t keso_curdom = 0;

#if NUM_BLOCKING_TASKS > 0
static void keso_gc_walk_stack(object_t** stack, callback_fkt_t handleObj) {

	if (stack==(object_t**) KESO_EOLL)
		return; /* empty stack partition */

	/* remove the marker bit */
	stack = keso_unpackStackRef(stack);

	/* Scan linked list of references (llref) */
	while (*stack != KESO_EOLL) {
		/* skip null references */
		if (*stack==NULL) {
			stack++;

		/* follow up next frame */
		} else if (keso_isStackRef(*stack)) {
			stack = keso_unpackStackRef(*stack);

		/* process object reference */
		} else {
			ASSERTCLASSID((*stack)->class_id);
			handleObj(*stack);
			stack++;
		}
	}
}
#endif

/* Scan a block that is about to be freed for objects and
 * call the finalize() method of the found objects
 */
#ifdef KESO_NEED_FINALIZE
static void callFinalize(keso_gc_slotcnt_t bstart, keso_gc_slotcnt_t size) {
	object_t **blockbegin = (object_t**) ((char*)heap_begin +
			( (unsigned int) bstart * (unsigned int) HEAPDESC.slotSize));
	keso_gc_slotcnt_t refsize = (keso_gc_slotcnt_t)
		(((unsigned int) size * (unsigned int) HEAPDESC.slotSize)/(unsigned int) sizeof(object_t*));
	keso_gc_slotcnt_t index = 0;
	keso_gc_slotcnt_t objbegin=index;
	object_t *obj;

	while(index < refsize) {
		if( (((unsigned int) blockbegin[index])&1) == 0 ) { index++; continue; }

		obj = (object_t*)(blockbegin+index);

		ASSERTCLASSID(obj->class_id);

		KESO_ASSERT(CLS_ROFF(obj->class_id)==(index-objbegin));

		/* call finalizer */
		KESO_INVOKE_FINALIZE(obj)(obj);

		/* position behind found object */
		objbegin += keso_objSize(HEAPDESC.slotSize,obj) * (HEAPDESC.slotSize/sizeof(object_t*));
		index = objbegin;
	}

	KESO_ASSERT(index == refsize && index == objbegin);
}
#endif

/* Mark a sequence of slots beginning at <firstslot> (start counting with 0)
 * with a length of <slotcount> slots in the bitmap.
 */
static void keso_gc_markSlots(keso_gc_slotcnt_t firstslot, keso_gc_slotcnt_t slotcount) {
	keso_gc_slotcnt_t bmint=firstslot / (sizeof(unsigned int)*8);
	keso_umachineword_t bmbit =firstslot % (sizeof(unsigned int)*8);
	keso_umachineword_t i;

	/* until next full int */
	if(bmbit>0) {
		i = bmbit = sizeof(unsigned int)*8-bmbit;
		if(i>slotcount) i=slotcount;
		slotcount-=i;

		while(i-->0) {
			unsigned int mask = 1<<--bmbit;
			keso_gc_bitmap[bmint] |= mask;
		}
		bmint++;
	}

	/* full ints */
	while(slotcount >= sizeof(unsigned int)*8) {
		slotcount-=sizeof(unsigned int)*8;
		keso_gc_bitmap[bmint++] = ~0;
	}

	/* beginning of last partially filled int */
	for(i=sizeof(unsigned int)*8; slotcount>0; slotcount--) {
		keso_gc_bitmap[bmint] |= 1<<--i;
	}
}

/* Marks the bits of the slots occupied by @obj.
 * If obj is an object outside the domain heap,
 * nothing is done (immortal objects).
 */
static int keso_gc_markObj(object_t *obj) {
	if(REFONHEAP(obj)) {
		keso_gc_slotcnt_t firstslot, size;

		/* calculate size in slots (i.e. bits used in the bitmap) */;
		size = keso_objSize(HEAPDESC.slotSize,obj);
		firstslot= ADDR2SLOT((object_t**)obj-CLS_ROFF(obj->class_id));

		keso_gc_markSlots(firstslot, size);
		return 1;
	}
	return 0;
}

/* Return reference count of an object
 *
 *  x: obj is object array with length x
 *  0: obj has no internal references
 * -x: obj is regular object w/ internal references
 */
static KESO_INLINE int refCnt(object_t *obj) {
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
	if(keso_isObjectArrayClass(obj->class_id))
		return ((array_t*)obj)->size;
#endif
	return -(signed int) CLS_ROFF(obj->class_id);
}

/* Search an int for the first bit with the specified value, starting with
 * bitposition bmbit. Maxsize may limit how much bits are searched, but it
 * is never searched more than to the end of the int.
 *
 * If a bit is found, its offset from bmbit is returned. If not, the negative value
 * of the number of bits searched is returned.
 */
static keso_machineword_t searchInt(keso_gc_slotcnt_t bmint, keso_umachineword_t bmbit, keso_umachineword_t value, keso_umachineword_t maxsize) {
	keso_umachineword_t i;
	keso_machineword_t bitssearched=0;

	/* bmbit used as shiftindex,
	 * i contains number of bits to process */
	bmbit = i = (sizeof(unsigned int)*8)-bmbit;
	if(i>maxsize) i=maxsize;

	while(i-->0) {
		if( ((keso_gc_bitmap[bmint]>>--bmbit)&1) == value) return bitssearched;
		else bitssearched++;
	}
	return -bitssearched; /* Nothing found */
}

/* Find the first bit that is set to value in the bitmap range starting at
 * bmbegin (counting from 0) with a length of maxsize.
 * The function returns the bitposition, or bmbegin+maxsize if
 * an occurance could not be found in the specified range.
 */
static keso_gc_slotcnt_t keso_gc_findBit(keso_gc_slotcnt_t bmbegin, keso_gc_slotcnt_t maxsize, keso_umachineword_t value) {
	unsigned int negvalue = ((value==0)?0xffffffff:0);
	keso_gc_slotcnt_t bmbit, bmint;
	keso_machineword_t pos;

	bmint = bmbegin / (keso_gc_slotcnt_t) (sizeof(unsigned int)*8);
	bmbit = bmbegin % (keso_gc_slotcnt_t) (sizeof(unsigned int)*8);

	/* the begin is not on an integer boundary */
	if(bmbit>0) {
		pos = searchInt(bmint, bmbit, value, maxsize);
		if(pos>=0) return (bmbegin+(keso_gc_slotcnt_t) pos);
		pos = -pos;
		bmbegin += (keso_gc_slotcnt_t) pos;
		maxsize -= (keso_gc_slotcnt_t) pos;
		bmint++;
	}

	/* process full ints */
	while(maxsize >= (keso_gc_slotcnt_t) (sizeof(unsigned int)*8)) {
		if(keso_gc_bitmap[bmint]!=negvalue) break; /* This int contains the searched bit */
		bmint++;
		maxsize -= (keso_gc_slotcnt_t) (sizeof(unsigned int)*8);
		bmbegin += (keso_gc_slotcnt_t) (sizeof(unsigned int)*8);
	}

	if(maxsize > 0) {
		pos = searchInt(bmint, 0, value, maxsize);
		if(pos>=0) return (bmbegin+(keso_gc_slotcnt_t) pos);
		pos = -pos;
		bmbegin += (keso_gc_slotcnt_t) pos;
	}
	return bmbegin;
}

/* Find a block of consecutive unset bits in the bitmap,
 * beginning at position *bmbegin. From bmbegin a maximum
 * of maxsize slots are searched.
 * The function returns the size of the found free block,
 * that may be 0. If there are multiple free blocks in the
 * specified area, the size of the first block is returned.
 * *bmbegin is adjusted to contain the first slot of the
 * free block whose size is returned. If no free block is
 * found, 0 will be returned and *bmbegin will contains its
 * original value + maxsize.
 */
static keso_gc_slotcnt_t keso_gc_findFreeBlock(keso_gc_slotcnt_t *bmbegin, keso_gc_slotcnt_t maxsize) {
	keso_gc_slotcnt_t blockstart, blockend;
	if(maxsize==0) return 0;
	blockstart = keso_gc_findBit(*bmbegin, maxsize, 0); /* Find a free slot */
	maxsize -= (blockstart-*bmbegin);
	*bmbegin = blockstart;
	if(maxsize==0) return 0;                    /* No free block in the specified range */
	blockend = keso_gc_findBit(blockstart, maxsize, 1); /* Find end of free block */
	return blockend-blockstart;                 /* return size of free block */
}

static void keso_gc_clearBitmap() {
	unsigned int i;
	for(i=0; i < (sizeof(keso_gc_bitmap)/sizeof(keso_gc_bitmap[0])); i++) {
		keso_gc_bitmap[i] = 0;
	}
}

