/*KESO--HEADER--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
#include "global.h"

object_t* keso_irr_alloc(class_id_t class_id, obj_size_t size, obj_size_t roff);

typedef struct _irr_listel_s {
	volatile keso_gc_slotcnt_t size;     /* size in slots, used by the classid in objects */
	volatile keso_gc_slotcnt_t flags;    /* flags, Bit0->Locking Mode, Bit1->AllocColor*/
	volatile struct _irr_listel_s *next; /* Points to the next listelement */
} irr_listel_t;

/*KESO--CFILE--KESO*/
#define GC_LOCK()
#define GC_UNLOCK()
#define GC_LOCK_STACK()     GetResource(GCRESOURCE())
#define GC_UNLOCK_STACK()   ReleaseResource(GCRESOURCE())
#define GC_LOCK_MUTATOR()   DisableAllInterrupts()
#define GC_UNLOCK_MUTATOR() EnableAllInterrupts()

#define KESO_EOFML ((irr_listel_t*) -1)
#define RESTR_MODE(__el__, __pmode__) if(__pmode__==0) { __el__->flags &= ~LOCKBIT; }

static void keso_irr_listwalker(domain_irr_t*, int (*) (irr_listel_t*, unsigned char, irr_listel_t volatile **, domain_irr_t*, void*), void *);
static int keso_irr_alloccbfkt(irr_listel_t *, unsigned char, irr_listel_t volatile **, domain_irr_t*, void **);


/*
 * Allocator function of the IRR Heap.
 *
 * This will return a pointer to a free memory area on the
 * heap of the current domain with @size bytes.
 *
 * This will throw an error if no suitable memory block can
 * be found.
 */
object_t* keso_irr_alloc(class_id_t class_id, obj_size_t size, obj_size_t roff) {
	/* Used to pass the required size to the allocator in the one direction,
	 * and to pass back the pointer to the allocated memory in the other
	 * direction. If no memory block is found, KESO_EOFML is passed back.
	 */
	void *retval;
	object_t* obj;
	domain_irr_t *domain;

	domain = (domain_irr_t *) KESO_CURRENT_DOMAIN;
	size = BYTES2SLOTS(size,domain->hd.slotSize);
	retval = (void *) (unsigned int) size; /* Pass size to call back function */

#ifdef IRRGC_DEBUG
	KESO_PRINTF("PREALLOC dom 0x%x, type %d, size %d slots, roff %d\n", KESO_CURRENT_DOMAIN, class_id, size,roff);
#endif

	keso_irr_listwalker(domain, (int (*) (irr_listel_t*, unsigned char, irr_listel_t volatile **, domain_irr_t*, void*)) keso_irr_alloccbfkt, (void *) &retval);

	/* TODO maybe we should return NULL here (and in all other heap specific
	 * allocs, too) and throw the out of memory error in allocObject */
	if (retval == KESO_EOFML) KESO_THROW_ERROR("out_of_memory\n");

	GC_LOCK_MUTATOR();
	domain->hd.sasls += size * domain->hd.slotSize;
	domain->heap_free -= size * domain->hd.slotSize;
	GC_UNLOCK_MUTATOR();

#ifdef KESO_GCMODE_LAZY
	if(0<keso_gctpaused) {
		keso_gctpaused=0;
		ActivateTask(keso_irr_gc);
	}
#endif
	{ /* Init object header */
		unsigned char colorbit = ((irr_listel_t*) retval)->flags | 1;

#ifdef DEBUG
	/* int-wise clear allocated memory */
	{
		unsigned int ssize = size * (domain->hd.slotSize/sizeof(int));
		while(ssize-->0) ((int*)retval)[ssize]=0;
	}
#else
	((irr_listel_t*) retval)->flags = 0;
#endif

		obj = (object_t*)retval;
		obj = (object_t*) ((object_t**)obj+roff);
		obj->class_id=class_id;
		obj->gcinfo = colorbit;
	}

#ifdef IRRGC_DEBUG
	KESO_PRINTF( "ALLOC type %d, size %d, roff %d @0x%x\n", obj->class_id, size,roff, (unsigned int)obj);
#endif
	return obj;
}

/*
 * Allocator callback function for the listwalker.
 * The parameter *retval initially contains the required size in slots
 * for the memory block that needs to be found.
 * When a block large enough for the request is found, it is either split
 * or removed from the list, and a pointer to the allocated area is returned
 * in *retval.
 */
static int keso_irr_alloccbfkt(irr_listel_t *element, unsigned char mode,
		irr_listel_t volatile **prevNextPointer, domain_irr_t *domain, void **retval) {
	keso_gc_slotcnt_t size;

	if(element == KESO_EOFML) { *retval=KESO_EOFML; return 0; }

	size = (keso_gc_slotcnt_t) (unsigned int) *retval;

	/***************************************************************************
	SPLITTING AN ELEMENT
	If the element's size is larger than the requested size, we split it by
	simply reducing its size.
	This subtraction must, however, happen atomically. There are the following
	scenarios that might happen:
	 (1) The interrupt might happen before writing back the substracted value,
	  but after having read the original value of size. In this case, we would
	  overwrite the modified value and hand out the same piece of memory twice.
	 (2) The interrupt might happen after writing back the modified value. In
	  this case, we would notice that somebody else performed a split, too, and
	  would not have a problem.
	Atomic substraction  would be possible with an compare-and-swap instruction,
	but the tricore architecture does not offer any instruction like this. We
	therefore have to disable the interrupts for a short period here.
	***************************************************************************/
	if(element->size > size) {
		keso_gc_slotcnt_t localsize;

		GC_LOCK_MUTATOR();
		localsize = element->size;
		if(localsize>size) {       /* enough memory to satisfy the request */
			localsize -= size;
			element->size = localsize;
			GC_UNLOCK_MUTATOR();
			/* Allocated memory is directly at element+new size */
			*retval =  (char*)element + (localsize * domain->hd.slotSize);
			/* Pass allocation color in first byte of allocated memory.
			 * Always set bit0 so the object header looks different from
			 * references
			 */
			((irr_listel_t*) *retval)->flags = element->flags;
			return 0;
		}
		/* interrupting alloc stole our memory */
		GC_UNLOCK_MUTATOR();
	}

	/***************************************************************************
	REMOVAL OF AN EXACTLY FITTING ELEMENT
	If we encounter an element that exactly fits, we have to remove it. In order
	to do this, we have to do the following:
	 (1) Check that the element had no locks at all set by an interrupted alloc
	 (2) Remove the element in a critical section.
	While setting the next pointer of the previous element to the next pointer
	of this element, the next element may be removed. The next pointer of the
	prevElement would point to an object not in the queue anymore, and yet another
	interrupting alloc could run over the inconsistent queue before we can fix that.
	Therefore we need to disable Interrupts here, again.
	***************************************************************************/
	if(element->size==size && mode==0) {
		GC_LOCK_MUTATOR();
		if(element->size==size) {   /* make sure the element wasn't split */
			*prevNextPointer = element->next;
			GC_UNLOCK_MUTATOR();
			element->size = 0;
			element->next = NULL;
			*retval = element;
			return 0;
		}
		GC_UNLOCK_MUTATOR();
	}
	return 1;
}

/* Interrupt-friendly walk over the memory list, applying
 * a callback function to each list-element.
 *
 * The callback function is called for each element. When reaching
 * the end of the queue, it is called with KESO_EOFML as element.
 * The listwalker will terminate after that.
 *
 * The callback function shall return 1 if it wishes further processing
 * of the queue, or 0 if it is done and the run should be aborted.
 * An optional pointer parameter is passed through to the callback function and
 * may be used for parameters to the callback or return values from the
 * callback. When aborting, the mode of the previous element will be restored,
 * and so will the mode of the current element unless the next pointer of the
 * previous Element was changed by the callback or curElement was KESO_EOFML.
 * If the callback function updates *prevNextPointer but curElement remains
 * within the queue, the callback function must restore the original mode,
 * passed to the callback function as the mode parameter. After restoring the
 * original mode, the callback function cannot assume that the element will not
 * be removed and therefore should not perform any actions on the element after
 * restoring the original mode.
 *
 * Interrupts are enabled when calling the callback, however locking bits will
 * prevent the previous and the current Element from being removed by interrupting
 * instances.
 *
 * The callback function may remove the current element if the old mode does not
 * forbid it, but must ensure that there are no problems by overlapping listwalkers,
 * i.e. the removal must happen atomically.
 * If the callback function removes an element and requests
 * further walking over the queue, it will be called with the same prevElement
 * and the new element it is pointing to as curElement.
 *
 * A callback function may insert new elements, but only between the prevElement
 * and the curElement. The next pointer must always point to an element at a
 * higher address to keep the list sorted. When the
 * callback function inserts one or more elements and requests further queue
 * walking, it will be called with the same prevElement and the new element that
 * it now points to the next time and from thereon proceed normally, i.e. all
 * inserted elements will be passed to the callback function unless removed by
 * interrupting allocators in the meantime.
 *
 * In any case, if the callback function changes the next pointer of the
 * previous element, it MUST RESTORE THE ORIGINAL MODE OF CURELEMENT IF
 * CURELEMENT IS STILL A MEMBER OF THE LIST, since this will not be done
 * by the listwalker in order to not modify the memory of removed elements.
 * I.e. when inserting a new element and having prevNextPointer pointing to
 * the new element, the mode of curElement must be restored in the callback
 * function or the element will stay locked forever.
 */
static void keso_irr_listwalker(domain_irr_t *domain,
		int (*elementCallback) (irr_listel_t*, unsigned char, irr_listel_t volatile **, domain_irr_t*, void*),
		void *cbparam) {
	irr_listel_t volatile **prevNextPointer;
	irr_listel_t *curElement, *prevElement;
	irr_listel_t scratch;
	unsigned char pmode=0, mode;
	int cbreturn=1;

	prevNextPointer = &domain->hd.freemem;
	prevElement = &scratch;           /* hack to make mode restoration work on first element */

	while(cbreturn) {
		/*****************************************************************************
		ADVANCING TO THE NEXT ELEMENT
		At this point, we have prevNextPointer pointing to the memory slot with the pointer
		to the next element in the list, which may also be KESO_EOFML.
		The memory slot of prevNextElement itself is protected and cannot be removed.
		Initially, this is due to the fact that the head pointer of the list is in the
		domain_desc and cannot be removed. The task is now to advance to the next
		element, and aquire a simple lock on it.
		The locking mode is saved in the mode member of the listelement. The object
		could be removed while aquiring the lock and writing mode to the already
		handed out object could cause inconsistencies. To avoid this, we must disable
		Interrupts here for a short portion of code.
		*****************************************************************************/
		GC_LOCK_MUTATOR();
		curElement = (irr_listel_t *) *prevNextPointer;
		if(curElement == KESO_EOFML) {
			GC_UNLOCK_MUTATOR();
			elementCallback(curElement, 1, prevNextPointer, domain, cbparam);
			break;
		}
		mode = curElement->flags;     /* save mode for later restoration */
		curElement->flags = mode | 1;    /* aquire a lock */
		GC_UNLOCK_MUTATOR();

		/* curElement is now locked and can be processed */
		mode &= 1; /* kill color bit */
		cbreturn = elementCallback(curElement, mode, prevNextPointer, domain, cbparam);

		/* If prevElement still points to curElement
		 *   move previousElement to the current element and unlock previousElement
		 * else
		 *   keep prevElement and recall callback with the new element it points to
		 *
		 * curElement and prevElement are protected from being removed, however, the
		 * callback function is allowed to remove curElement if the previous mode
		 * allows it. The callback function is also allowed to insert new elements
		 * between prevElement and curElement. Both cases change the content of
		 * prevElement->next, and we call the callback again with the same
		 * prevElement and the new element that it points to.
		 * We do not need to care about restoring the mode of curElement here, since
		 * this is the task of the callback function.
		 */
		if(*prevNextPointer==curElement) {
			RESTR_MODE(prevElement, pmode); /* restore original mode of previous element */
			prevElement = curElement;       /* advance one element */
			prevNextPointer = &curElement->next;
			pmode = mode;                   /* backup mode of new previous element */
		}
		/* in case we keep prevElement we also need to keep it locked */
	}
	/* If processing is aborted without changed prevNextPointer, prevElement
	 * was updated with curElement and we restore its mode here. */
	RESTR_MODE(prevElement, pmode);
}
