/*KESO--HEADER--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
/* Inlining Level for Writebarriers. The following levels are defined
 * 0: no inlining. keso_irr_writebarrier is always called.
 * 1: inlining of the check that looks if the GC is running. If it is, pushObject is called.
 */
#define KESO_WRBR_INLINE_LEVEL 1
/* forward declaration needed because of domain.h<->irr.h include cycle */
typedef struct domain_irr_t domain_irr_t;
extern domain_irr_t *keso_irr_gc_active_domain;

void keso_irr_pushObject(object_t *obj);
#if KESO_WRBR_INLINE_LEVEL==0
void keso_irr_writebarrier(object_t** addr, object_t* value);
#endif

/*KESO--CFILE--KESO*/
#include "keso_types.h"
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
#include "object_array.h"
#endif

/*
 * NOTES
 *
 * Calculating slot sizes is expensive (division). slotsizes that are power of 2
 * could be handled much more efficiently.
 *
 * Maybe we should automatically adjust the heapsize to have a slotcount that is
 * a multiple of 32, so we only have full ints in the bitmaps, none that are
 * only partially occupied
 *
 * chooseDomain might loop until it finds a domain with a need instead of
 * collecting the domain with the lowest need value. This will reduce the
 * interrupt jitter and the time spans with enabled write barriers.
 * As an alternative, the garbage collector could wait for an event when it
 * finds there are no domains that need garbage collection. The Event could be
 * sent by the irr-alloc function. This would additionally decrease the energy
 * consumption on certain microcontrollers.
 */

/* Parameter structure for the sweep callback function */
struct sweepparam {
	irr_listel_t *lastel; /* Pointer to the last processed element, NULL on first element */
	keso_gc_slotcnt_t bmpos;     /* Current bitposition in the bitmap, initially 0, later the first
	                         bit beyond the last scanned element. */
};

/* Indicator on which domain irr-gc is currently
 * active, intended for use by barriers.
 * DOMAIN_ZERO means gc not active
 */
domain_irr_t *keso_irr_gc_active_domain=DOMAIN_ZERO;

/* Contains the full gcbyte for an already colored object.
 * This allows checking/setting an objects color without
 * masking any bits. The initial value does not matter because
 * the GC will correctly initialize it every time before enabling
 * writebarriers. Currently, this value be always either 1 or 3.
 */
static unsigned char keso_irr_coloredobj= LOCKBIT | COLORBIT;

/* Index of the next free slot on the stack (one above
 * the top element of the stack) */
static keso_gc_slotcnt_t keso_irr_sp=0;

/* very simple estimation of the maximum number of objects that might be allocated */
static object_t *keso_irr_stack[KESO_GC_WORKINGSTACK_SIZE];

/* internal functions */
static int  keso_irr_sweep(irr_listel_t *, unsigned char, irr_listel_t volatile **, domain_irr_t *, struct sweepparam *);
static int  keso_irr_fspmarker(irr_listel_t *, unsigned char, irr_listel_t volatile **, domain_irr_t*, void *);
static void keso_irr_scanStack(void);
static void keso_irr_chooseDomain(void);

#if KESO_WRBR_INLINE_LEVEL==0
/* Write barrier function for IRR Heaptype. This will
 * also assign the new value. addr is a pointer to the
 * reference field, and value is new value that will be
 * assigned to the field.
 */
void keso_irr_writebarrier(object_t** addr, object_t* value) {
	if (keso_irr_gc_active_domain==KESO_CURRENT_DOMAIN)
		keso_irr_pushObject(*addr);
	*addr = value;
}
#endif

/* This is the main GC function of the IRR Heap.
 * The task is chaining itself and run periodically,
 * collection garbage on the heap of one domain at
 * each run.
 */
#ifdef TARGET_TRICORE_CIAO
void KESODomZero::functionTaskkeso_irr_gc() {
#else
TASK(keso_irr_gc) {
#endif
	struct sweepparam sp;
	keso_gc_slotcnt_t i;

	while (1) {

		/* choose the domain that will be scanned */
		keso_irr_chooseDomain();

#ifdef IRRGC_DEBUG
	KESO_PRINTF("domain %d, %ldf,%ldu\n", keso_curdom, DOMAINID()->heap_free, HEAPDESC.sasls);
#endif

#ifdef KESO_NEED_FINALIZE
#if KESO_NUM_JAVADOMAINS > 1
		/* TODO if the GC is preempted, we'd need to restore this domain ID
		 * when being rescheduled... */
		KESO_CURRENT_DOMAIN = (domain_t*) DOMAINID();
#endif
#endif

		/* heap begin and end are e.g. used to determine if an object reference is
		 * on the domain heap or not */
		heap_begin = (object_t*) HEAPDESC.heap_top;
		heap_end = (object_t*) ((char*) heap_begin + DOMAINID()->heap_size);
#ifdef IRRGC_DEBUG
		KESO_PRINTF("HEAPAREA 0x%x - 0x%x\n", (unsigned int)heap_begin,(unsigned int) heap_end);
#endif

		/* ENABLE WRITEBARRIERS
		 *
		 * Writebarriers are active on
		 *  - static fields (PutStatic)
		 *  - object array's fields
		 *  - reference fields in objects (PutField)
		 *
		 * We don't need writebarriers on
		 *  - immortal (constant) objects
		 */
		keso_irr_sp=0;
		DOMAINID()->colorbit ^= COLORBIT;
		keso_irr_coloredobj = DOMAINID()->colorbit;
		keso_irr_gc_active_domain = DOMAINID();

		/* Mark all slots of current freemem list */
		keso_irr_listwalker(DOMAINID(), keso_irr_fspmarker, NULL);
#ifdef IRRGC_DEBUG
		{ int i; KESO_PRINTF("BMFSP "); for(i=0; i<sizeof(keso_gc_bitmap)/sizeof(unsigned int); i++) { KESO_PRINTF("%x ", keso_gc_bitmap[i]); } KESO_PRINTF("\n"); }
#endif

		/* record new allocations from this point */
		HEAPDESC.sasls = 0;

		/*  Scan Order
		 *
		 * Stacks of waiting tasks need to be scanned before scanning the remaining
		 * parts of the root set. Otherwise, the only reference to a white object W
		 * on a task stack could, for instance, be written to a black static
		 * reference and removed from the stack. This would create a black -> white
		 * path and at the same time remove the last discoverable path to W without
		 * marking W, since write barriers are not active on stack slot updates.
		 */

		/* scan stacks */
#if NUM_BLOCKING_TASKS > 0
		for(i=0; i < NUM_BLOCKING_TASKS; i++) {
			GC_LOCK_STACK();

			KESO_CONST keso_stack_t *stack = blocking_tasks[i]->_stack;
			/* look through all stack partitions */
			for ( ; stack!=NULL; stack=stack->next) {
				/* scan llref if the current partition belongs to the gc'ed domain */
				if (stack->domain == (domain_t*) DOMAINID()) {
					keso_gc_walk_stack(stack->llrefs, keso_irr_pushObject);
				}
			}
			GC_UNLOCK_STACK();
		}
#endif

		/* push root set (immutable) */
#if KESO_GC_ROOTSETSIZE > 0
		for(i=0; i<KESO_GC_ROOTSETSIZE; i++) {
			keso_irr_pushObject(HEAPDESC.rootset[i]);
		}
#endif

		/* if Alarm/Resource-objects contained references, we
		 * would need to scan them, too
		 */

		/* Scan statics */
#if KESO_NUM_STATIC_REFS>0
		for(i=0; i<KESO_NUM_STATIC_REFS; i++)
			keso_irr_pushObject((object_t*) STATICREF(keso_irr_gc_active_domain)[i]);
#endif


		/* The root set is on the working stack now, scan and mark
		 * until the stack is empty */
		keso_irr_scanStack();
#ifdef IRRGC_DEBUG
		{ int i; KESO_PRINTF("BMSCAN "); for(i=0; i<sizeof(keso_gc_bitmap)/sizeof(unsigned int); i++) { KESO_PRINTF("%x ", keso_gc_bitmap[i]); } KESO_PRINTF("\n"); }
#endif

		/* Recover unused memory */
		sp.lastel = NULL;
		sp.bmpos  = 0;
		keso_irr_listwalker(DOMAINID(), (int (*) (irr_listel_t *, unsigned char, irr_listel_t volatile **, domain_irr_t*, void *))keso_irr_sweep, &sp);

		/* Clear the bitmap for next gc cycle */
		keso_gc_clearBitmap();
	}
}

/* Choose the domain that needs garbage collection
 * the most of all managed domains, determined by
 * a metric called the "need" here.
 *
 * The need of a domain for a GC is determined by
 * the slots allocated since the last garbage collection (sasls)
 * compared to the remaining space. The remaining
 * space should not be less than the sasls plus a grace
 * area of 25% of the heap size. Thus, the "need" is calculated
 * freeslots - sasls - (heapsize>>2).
 * The function goes RoundRobin through the domain array and chooses
 * the first found domain that violates the above criterium (need<0). By
 * doing this instead of simply choosing the domain with the most severe
 * need, we can assure that all domains that have a need actually receive
 * a garbage collection (fairness to some degree).
 * If every domain has need>=0, the domain with the lowest need is taken.
 */

static void keso_irr_chooseDomain(void) {
#ifdef KESO_GCMODE_WORKAHOLIC
	signed int minneed=65535;
	keso_umachineword_t mindom=keso_curdom;
#endif
	signed int need;
	const keso_umachineword_t prevDom=keso_curdom;

	while(1) {
		if(++keso_curdom >= MANAGEDDOMAINS) keso_curdom=0;

#ifdef KESO_GCMODE_LAZY
		if(prevDom==keso_curdom && HEAPDESC.sasls == 0) {
			keso_gctpaused=1;
			TerminateTask();
		}
#endif

		need = DOMAINID()->heap_free - HEAPDESC.sasls- (DOMAINID()->heap_size>>2);
		/* if need<0, this domain receives a GC treatment */
		if(need<0) return;

#ifdef KESO_GCMODE_WORKAHOLIC
		if(need<minneed) { mindom=keso_curdom; minneed=need; }
#endif

		/* The need of every domain is >0, choose
		 * the one with the lowest need */
#ifdef KESO_GCMODE_WORKAHOLIC
		if(prevDom == keso_curdom) {
			keso_curdom=mindom;
			return;
		}
#endif
	}
}

/* This function processes the working stack, initially containing
 * the root object set when run. Elements are popped and marked when
 * completly processed, and may be pushed upon discovery of white
 * inner references or concurrently by write barriers.
 *
 * The same reference will never be pushed onto the stack twice, which
 * ensures that the stack is always large enough. This is accomplished
 * by marking already scanned elements by setting their colorbit before
 * pushing them to the stack. Already colored elements are not pushed
 * again.
 *
 * The bits of an object in the bitmap are not marked until popping the
 * element from the stack. This enables writebarriers to push objects
 * without the need to mark them on the stack (and do it in the low
 * priority gc task). On the other hand, it introduces the need for
 * the color bit in the objects header.
 */

static void keso_irr_scanObject(object_t * curobj, int curref) {

		while(curref<0) keso_irr_pushObject( ((object_t**)curobj)[curref++] );
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
		while(curref>0) keso_irr_pushObject( ((object_array_t*)curobj)->data[--curref] );
#endif
}

static void keso_irr_scanStack(void) {
	object_t *curobj;
	signed int curref;

	while(SuspendAllInterrupts(), keso_irr_sp>0) {
		curobj = keso_irr_stack[--keso_irr_sp];
		ResumeAllInterrupts();

#ifdef IRRGC_DEBUG
		KESO_PRINTF("Obj 0x%x, type %d\n", (unsigned int)curobj, curobj->class_id);
#endif

		curref = refCnt(curobj);
		keso_irr_scanObject(curobj, curref);
	}
	/* scanning complete, disable write barriers */
	keso_irr_gc_active_domain = DOMAIN_ZERO;
	ResumeAllInterrupts();
}

/* Callback function to the listwalker, that recovers unused memory by
 * freeing unreachable objects. The memory slots that can be freed are
 * marked by having their corresponding bit in the bitmap still set to 0.
 *
 * If possible, the recovered memory is merged with preceding or succeeding
 * blocks to fight fragmentation.
 */
static int keso_irr_sweep(irr_listel_t *element, unsigned char mode,
		irr_listel_t volatile **prevNextPointer, domain_irr_t *domain,
		struct sweepparam *sp) {
	/*
	 * Previous element: sp->lastel (NULL before first element)
	 * First bit for next processing: sp->bmpos (initially 0)
	 * Current element:  element    (KESO_EOFML behind last element)
	 */

	/* Pointer to the head of the partial list */
	irr_listel_t *start;

	/* Pointer to the next pointer linking to the tail
	 * of the partial list. Initially, this points to
	 * the start pointer, but after finding 2 or more
	 * elements it will point to the next pointer of
	 * the element before the tail element in the
	 * partial list */
	irr_listel_t **pnext;

	/* Return value to the listwalker */
	unsigned int retval;

	/* Maximum position in the bitmap to scan, i.e.
	 * this will contain the first slot of element */
	keso_gc_slotcnt_t maxbmpos;

	/* Size of the currently reclaimed block */
	keso_gc_slotcnt_t size;

	/* Record the number of reclaimed slot, and add them
	 * in a single critical section before leaving the
	 * sweep function */
	keso_gc_slotcnt_t addedslots;

	if(element == KESO_EOFML) {
		/* maxbmpos = last slot of bitmap */
		maxbmpos = domain->heap_size / domain->hd.slotSize;
		/* done after this run */
		retval=0;
	} else {
		/* maxbmpos = first slot of element */
		maxbmpos=(keso_gc_slotcnt_t)(
				((unsigned int) element - (unsigned int) heap_begin) /
				(unsigned int)domain->hd.slotSize);
		/* demand further invocations */
		retval=1;
	}

	if(sp->bmpos > maxbmpos) {
		/* recalled with inserted elements from prior calls */
		sp->lastel = element;
		return retval;
	}

#ifdef IRRGC_DEBUG
	if((sp->lastel!=NULL) && (prevNextPointer != &(sp->lastel->next))) {
		KESO_PRINTF("SWEEP ERR pNP!=sp->lastel\n");
	}
#endif

	start = NULL;
	pnext = &start;
	addedslots=0;

	/** (1) Build partial list not connected to the list (uncritical) **/
	size = keso_gc_findFreeBlock(&sp->bmpos, maxbmpos-sp->bmpos);
	if(size>0){
	while(1) {
		irr_listel_t *block = (irr_listel_t*) SLOT2ADDR(sp->bmpos);


#ifdef IRRGC_DEBUG
		KESO_PRINTF( "RECLAIM 0x%x (%d slots)\n",(unsigned int) block, size);
#endif

		/* Call destructors of objects that will be freed */
#ifdef KESO_NEED_FINALIZE
		callFinalize(sp->bmpos, size);
#endif

		// clear memory
		{
			unsigned int pos = sizeof(irr_listel_t)/sizeof(int);
			for(; pos < size * (domain->hd.slotSize/sizeof(int)); pos++)
				((int*)block)[pos] = 0;
		}

		/* Init new element, except the next pointer */
		*pnext = block;
		block->size = size;
		block->flags = DOMAINID()->colorbit & ~LOCKBIT;

		/* Skip scanning of this block */
		sp->bmpos += size;
		addedslots+= size;

		size = keso_gc_findFreeBlock(&sp->bmpos, maxbmpos-sp->bmpos);
		if(size>0) pnext = (irr_listel_t **) &((*pnext)->next);
		else break;
	}}

	/* Next Pointer of the tail element of the partial list is not initialized */

#ifdef IRRGC_DEBUG
	if(sp->bmpos != maxbmpos)
		KESO_PRINTF("SWEEP ERR BP!=MBP\n");
#endif

	if(element != KESO_EOFML)   /* Skip element on next search */
		sp->bmpos += element->size;

	if(0==addedslots) { /* No free space found */
		sp->lastel = element;
		return retval;
	}

	/* (2) Insert the partial list into the free memory list
	 *
	 * *pnext: Next Pointer of last element of partial list, possibly start->next
	 * start : first element of partial list, not NULL
	 * element: element following the scanned heap area, possibly KESO_EOFML
	 * *prevNextPointer: next pointer of the previous element
	 * sp->lastel: previous element
	 */
	if((char*)element == ((char*)*pnext + (*pnext)->size*domain->hd.slotSize)) {
		/* Tail mergable (this can't be changed by interruptions) */
		SuspendAllInterrupts();
		(*pnext)->next  = element->next;
		(*pnext)->size += element->size;
		/*-----------------------------*/
		element->next = NULL;
		element->size = 0;
		element->flags= 0;
		/*-----------------------------*/
		if(sp->lastel!=NULL &&
				(char*)start == ((char*)sp->lastel + sp->lastel->size*domain->hd.slotSize)) {
			/* Head and Tail mergable */
			sp->lastel->size += start->size;
			sp->lastel->next  = start->next;
			/*-----------------------------*/
			start->next = NULL;
			start->size = 0;
			start->flags= 0;
			/*-----------------------------*/
			ResumeAllInterrupts();
		} else {
			/* head not mergeable */
			*prevNextPointer = start;
			ResumeAllInterrupts();
		}
		/* No need to unlock element on this path */
	} else {
		/* Tail not mergable */

		/* Set Next Pointer of last element */
		(*pnext)->next = element;

		if(sp->lastel!=NULL &&
				(char*)start == ((char*)sp->lastel + sp->lastel->size*domain->hd.slotSize)) {
			/* Local Copy of the next pointer enables us to put
			 * the lastel->next = start->next outside the critical section */
			register irr_listel_t *headnext = (irr_listel_t *) start->next;

			SuspendAllInterrupts();
			if((char*)start == ((char*)sp->lastel + sp->lastel->size*domain->hd.slotSize)) {
				sp->lastel->size += start->size;
				/*-----------------------------*/
				start->next = NULL;
				start->size = 0;
				start->flags= 0;
				/*-----------------------------*/
				ResumeAllInterrupts();
				sp->lastel->next = headnext;

				if((sp->lastel->next != element) && (element != KESO_EOFML))
					element->flags &= ~(LOCKBIT);
				else sp->lastel = element;
			} else {
				ResumeAllInterrupts();
				*prevNextPointer = start;
				if(element != KESO_EOFML) element->flags &= ~(LOCKBIT);
			}
		} else {
			/* Neither head nor tail mergable (atomic insertion possible) */
			*prevNextPointer = start;
			/* unlock element */
			if(element != KESO_EOFML) element->flags &= ~(LOCKBIT);
			/* sp->lastel will be recovered in subsequent listwalker invocations */
		}
	}

	/* Update freeslots in the domain descriptor */
	SuspendAllInterrupts();
	domain->heap_free += addedslots * domain->hd.slotSize;
	ResumeAllInterrupts();

	return retval;
}

/* Callback function that marks the slots of provided free memory blocks.
 * For each element it is called with, it marks the corresponding bits in
 * the bitmap. This makes objects allocated during the gccycle survive the
 * this cycle.
 *
 * It furthermore changes the allocation color of the element, so that all
 * objects allocated beyond that are allocated colored and will never be pushed
 * on the stack during this GC cycle.
 */
static int keso_irr_fspmarker(irr_listel_t *element, unsigned char mode, irr_listel_t volatile **prevNextPointer, domain_irr_t *domain, void *param) {
	keso_gc_slotcnt_t firstslot;
	keso_gc_slotcnt_t size;
	if(element == KESO_EOFML) return 0;

	firstslot= (keso_gc_slotcnt_t) ((unsigned int)element - (unsigned int)heap_begin) / (keso_gc_slotcnt_t) domain->hd.slotSize;

	/* Atomically saving current size of the element and changing
	 * its allocation color to black assures that all elements are
	 * allocated black from the area marked for survival.
	 * The critical section here would not be strictly necessary,
	 * but can save the work of marking the slots twice in case
	 * an object is allocated white before changing the allocation
	 * color but after reading the size of the element.
	 *
	 * ADDITIONAL NOTE
	 * In order to reduce the expected jitter the interrupt blockade
	 * has been removed, at the expense of the (very unlikely) case that
	 * the slots of an object will be marked twice.
	 */
	/*SuspendAllInterrupts();*/
	size = element->size;
	element->flags = DOMAINID()->colorbit; /* LOCKBIT included in colorbit value */
	/*ResumeAllInterrupts();*/

	keso_gc_markSlots(firstslot, size);
	return 1;
}

/* Push a white object on the stack.
 * If the object is not white, this function
 * will do nothing.
 *
 * THIS IS THE ONLY PLACE WHERE OBJECTS ARE PUSHED
 * ON THE STACK.
 */
void keso_irr_pushObject(object_t *obj) {
	register keso_gc_slotcnt_t sp;
	register keso_machineword_t refcnt;
	int heapobj;

	if(obj==NULL) return;
	ASSERTOBJ(obj);

	SuspendAllInterrupts();
	if(!OBJCOLORED(obj,keso_irr_coloredobj)) {
		COLOROBJ(obj,keso_irr_coloredobj);
		ResumeAllInterrupts();

		refcnt = refCnt(obj);
		heapobj = keso_gc_markObj(obj);

		if(refcnt != 0) { /* no references -> no need to scan */
			if(heapobj) {
				SuspendAllInterrupts();
				sp = keso_irr_sp;
				keso_irr_sp = sp+1;
				ResumeAllInterrupts();

				/* Uncritical, because a "popObject" will never overlap
				 * a pushObject (only the other way round) */
				keso_irr_stack[sp] = obj;
			} else {
				keso_irr_scanObject(obj, refcnt);
			}
		}
	} else ResumeAllInterrupts();
}

