/*KESO--HEADER--KESO*/
/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

object_t* keso_dup_object(object_t* obj);

/*KESO--CFILE--KESO*/

/*
 * keso_dup_object creates a deep copy of an object.
 * The function works iterative with a work list which is allocated
 * on the stack and twice KESO_DUP_OBJ_MAX large. It is therefore
 * recommended that KESO_DUP_OBJ_MAX is as small as possible.
 */
object_t* keso_dup_object(object_t* obj) {
	object_t* new_obj=(object_t*)0;
	unsigned int i, j, top, end;

	object_t* todo_list_from[KESO_DUP_OBJ_MAX];
	object_t** todo_list_to[KESO_DUP_OBJ_MAX];

	top = 0;
	end = 0;

	todo_list_from[end] = obj;
	todo_list_to[end] = &new_obj;
	end++;

	while (top<end) {
		class_id_t cls;
		object_t* obj = todo_list_from[top];
		if (obj==NULL) {
			top++;
			continue;
		}
		cls = obj->class_id;

		if (keso_isNonCopyable(cls)) {
			*todo_list_to[top] = (object_t*)0;
			top++;
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
		} else if (keso_isObjectArrayClass(cls)) {
			/* TODO */
			*todo_list_to[top] = (object_t*)0;
			top++;
#endif
		} else if (keso_isArrayClass(cls)) {
			array_t *aobj = (array_t *) obj;
			unsigned int size = aobj->size*CLASS(cls).size;

			/* alloc the new array object and init size */
			array_t *nobj = (array_t *) KESO_ALLOC(cls, sizeof(array_t)+size, 0);
			nobj->size = aobj->size;

			*todo_list_to[top] = (object_t *) nobj;
			top++;

			/* copy array data */
			keso_memcpy((jbyte*)(nobj+1), (jbyte*)(aobj+1), size);

		} else if (keso_isMemoryClass(cls)) {
			*todo_list_to[top] = keso_dup_mem(obj);
			top++;
		} else {
			/* copy object */
			unsigned int roff = CLS_ROFF(cls);
			object_t*  nobj;
			object_t** refs;

			nobj = *todo_list_to[top] = keso_allocObject(cls);
			top++;

			{ /* copy content of the object */
				int i;
				for(i = -roff; i<0; i++) { /* copy references */
					((object_t**) nobj)[i] = ((object_t**) obj)[i];
				}

				// psize = size of primitive fields + object header
				unsigned int psize = CLASS(cls).size - sizeof(object_t*)*roff;
				for(i = sizeof(object_t); i < psize; i++) { /* copy primitive fields */
					((char*) nobj)[i] = ((char*) obj)[i];
				}
			}
			refs=KESO_REF_PTR(roff,nobj);
			for (i=0;i<roff;i++) {

				if (refs[i]==NULL) continue;

				/* check if we have seen that reference before */
				for (j=0;j<top && todo_list_from[j]!=refs[i] ;j++) ;

				if (j<top) {
					/* object was already copied */
					refs[i]=*todo_list_to[j];
				} else {
					/* insert reference into the todo list */
					todo_list_from[end] = refs[i];
					todo_list_to[end] = &refs[i];
					end++;

					if (end==KESO_DUP_OBJ_MAX) {
						/* TODO: overrun */
						/* FIXME: overrun */
						return (object_t*)0;
					}
				}
			}
		}
	}

	return new_obj;
}

/*
 * keso_deep_copy is similar to keso_dup_obj: it traverses all object
 * fields from src and transfers them to dest.
 * The function works iterative with a work list which is allocated
 * on the stack and twice KESO_DUP_OBJ_MAX large. It is therefore
 * recommended that KESO_DUP_OBJ_MAX is as small as possible.
 */
void keso_deep_copy(object_t* dest, object_t* src) {
	unsigned int i, j, top, end;

	object_t* todo_list_from[KESO_DUP_OBJ_MAX];
	object_t** todo_list_to[KESO_DUP_OBJ_MAX];

	top = 0;
	end = 0;

	todo_list_from[end] = src;
	todo_list_to[end] = &dest;
	end++;

	while (top<end) {
		class_id_t cls;
		object_t* obj = todo_list_from[top];
		if (obj==NULL) {
			top++;
			continue;
		}
		cls = obj->class_id;

		if (keso_isNonCopyable(cls)) {
			*todo_list_to[top] = (object_t*)0;
			top++;
#ifdef KESO_OBJECTARRAYCLASS_AVAILABLE
		} else if (keso_isObjectArrayClass(cls)) {
			/* TODO */
			*todo_list_to[top] = (object_t*)0;
			top++;
#endif
		} else if (keso_isArrayClass(cls)) {
			array_t *aobj = (array_t *) obj;
			unsigned int size = aobj->size*CLASS(cls).size;

			/* alloc the new array object and init size */
			array_t *nobj = (array_t *) KESO_ALLOC(cls, sizeof(array_t)+size, 0);
			nobj->size = aobj->size;

			*todo_list_to[top] = (object_t *) nobj;
			top++;

			/* copy array data */
			keso_memcpy((jbyte*)(nobj+1), (jbyte*)(aobj+1), size);

		} else if (keso_isMemoryClass(cls)) {
			*todo_list_to[top] = keso_dup_mem(obj);
			top++;
		} else {
			/* copy object */
			unsigned int roff = CLS_ROFF(cls);
			object_t*  nobj;
			object_t** refs;

			nobj = *todo_list_to[top] = keso_allocObject(cls);
			top++;

			{ /* copy content of the object */
				int i;
				for(i = -roff; i<0; i++) { /* copy references */
					((object_t**) nobj)[i] = ((object_t**) obj)[i];
				}

				// psize = size of primitive fields + object header
				unsigned int psize = CLASS(cls).size - sizeof(object_t*)*roff;
				for(i = sizeof(object_t); i < psize; i++) { /* copy primitive fields */
					((char*) nobj)[i] = ((char*) obj)[i];
				}
			}
			refs=KESO_REF_PTR(roff,nobj);
			for (i=0;i<roff;i++) {

				if (refs[i]==NULL) continue;

				/* check if we have seen that reference before */
				for (j=0;j<top && todo_list_from[j]!=refs[i] ;j++) ;

				if (j<top) {
					/* object was already copied */
					refs[i]=*todo_list_to[j];
				} else {
					/* insert reference into the todo list */
					todo_list_from[end] = refs[i];
					todo_list_to[end] = &refs[i];
					end++;

					if (end==KESO_DUP_OBJ_MAX) {
						/* TODO: overrun */
						/* FIXME: overrun */
						return;
					}
				}
			}
		}
	}

}

