/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/
package keso.driver.avr.atmega128;

import keso.driver.serialport.*;
import keso.core.*;

/**
 * @author Ralf Ellner
 */
public final class Usart implements SerialPort {
	private final static int  DefaultBaudRate = 38400;
	private final static byte DefaultFrameSize = SerialPort.DATABITS_8;
	private final static byte DefaultParityMode = SerialPort.PARITY_NONE;
	private final static byte DefaultStopBits = SerialPort.STOPBITS_ONE;
	private final static byte DefaultTransferMode = SerialPort.TRANSFERMODE_ASYNC;
	private final static byte DefaultSynchronousClockMode = SerialPort.SYNC_CLOCK_MODE_RISING_EDGE;

	// NOTE: UCSR0C and UBRR0H share the same memory location!
	private static byte ucsrc;

	// Flags used in interrupts - must be made volatile
	private static boolean receiveComplete;
	private static Memory receiveBuffer;
	private static short receiveLength;
	private static short receiveIndex;
	private static boolean receivePacketStart;

	private static boolean transmitComplete;
	private static Memory transmitBuffer;
	private static short transmitLength;
	private static short transmitIndex;

	public Usart() {
		this(DefaultBaudRate, DefaultFrameSize, DefaultParityMode, DefaultStopBits, DefaultTransferMode, DefaultSynchronousClockMode);
	}

	public Usart(int baudRate) {
		this(baudRate, DefaultFrameSize, DefaultParityMode, DefaultStopBits, DefaultTransferMode, DefaultSynchronousClockMode);
	}

	public Usart(int baudRate, byte dataSize, byte parityMode, byte stopBits) {
		this(baudRate, dataSize, parityMode, stopBits, DefaultTransferMode, DefaultSynchronousClockMode);
	}

	public Usart(int baudRate, byte dataSize, byte parityMode, byte stopBits, byte transferMode, byte clockMode) {
		this.setBaudrate(baudRate);
		this.setFrameSize(dataSize);
		this.setParityMode(parityMode);
		this.setStopBits(stopBits);
		this.setSynchronousClockMode(clockMode);
		this.setTransferMode(transferMode);

		receiveComplete = false;
		transmitComplete = true;

		// set receiver and transmitter active, enable interrupts
		ATMega128.registers.UCSR0B.set((1 << ATMega128.RXEN) | (1 << ATMega128.TXEN));
	}

	public void setRXInterruptEnable(boolean state) {
		if(state) {
			ATMega128.registers.UCSR0B.setBit(ATMega128.RXCIE);
		} else {
			ATMega128.registers.UCSR0B.clearBit(ATMega128.RXCIE);
		}
	}

	public void setBaudrate(int baudRate) {
		int ubrr = Config.getInt("Frequency", 0) / 16 / baudRate - 1;
		ATMega128.registers.UBRR0H.set((ubrr >> 8) & 0x0f);
		ATMega128.registers.UBRR0L.set(ubrr & 0xff);
	}

	public void setTransferMode(byte mode) {
		if (mode == SerialPort.TRANSFERMODE_SYNC) {
			ucsrc |= (1 << ATMega128.UMSEL);
		} else {    // assume ASYNC
			ucsrc &= ~(1 << ATMega128.UMSEL);
		}

		ATMega128.registers.UBRR0H.set(ucsrc);
	}

	public void setParityMode(byte mode) {
		ucsrc &= ~((1 << ATMega128.UPM1) | (1 << ATMega128.UPM0));  // Clear bits (no parity)

		if (mode == SerialPort.PARITY_EVEN) {
			ucsrc |= (1 << ATMega128.UPM1);
		} else if (mode == SerialPort.PARITY_ODD) {
			ucsrc |= ((1 << ATMega128.UPM1) | (1 << ATMega128.UPM0));
		} else {
			// other modes unsupported assume PARITY_NONE
		}

		ATMega128.registers.UBRR0H.set(ucsrc);
	}

	public void setStopBits(byte stopBits) {

		if (stopBits == SerialPort.STOPBITS_TWO) {
			ucsrc |= (1 << ATMega128.USBS);
		} else {    // assume _ONE - _ONE_POINT_FIVE not supported
			ucsrc &= ~(1 << ATMega128.USBS);
		}

		ATMega128.registers.UBRR0H.set(ucsrc);
	}

	public void setFrameSize(byte size) {

		// Clear Bits
		ucsrc &= ~((1 << ATMega128.UCSZ1) | (1 << ATMega128.UCSZ0));
		ATMega128.registers.UCSR0B.clearBit(ATMega128.UCSZ2);

		switch (size) {
			case SerialPort.DATABITS_5:
				break;
			case SerialPort.DATABITS_6:
				ucsrc |= (1 << ATMega128.UCSZ0);
				break;
			case SerialPort.DATABITS_7:
				ucsrc |= (1 << ATMega128.UCSZ1);
				break;
			case SerialPort.DATABITS_9:
				ATMega128.registers.UCSR0B.setBit(ATMega128.UCSZ2);
				ucsrc |= ((1 << ATMega128.UCSZ0) | (1 << ATMega128.UCSZ1));
				break;
			default:    // DATABITS_8
				ucsrc |= ((1 << ATMega128.UCSZ0) | (1 << ATMega128.UCSZ1));
				break;
		}

		ATMega128.registers.UBRR0H.set(ucsrc);
	}

	public void setSynchronousClockMode(byte mode) {

		if (mode == SerialPort.SYNC_CLOCK_MODE_FALLING_EDGE) {
			ucsrc |= (1 << ATMega128.CPOL);
		} else {    // assume RISING_EDGE
			ucsrc &= ~(1 << ATMega128.CPOL);
		}

		ATMega128.registers.UBRR0H.set(ucsrc);
	}

	public boolean hasFrameError() {
		return ATMega128.registers.UCSR0A.isBitSet(ATMega128.FE);
	}

	public boolean hasParityError() {
		return ATMega128.registers.UCSR0A.isBitSet(ATMega128.UPE);
	}

	public boolean hasDataOverRun() {
		return ATMega128.registers.UCSR0A.isBitSet(ATMega128.DOR);
	}

	public void enablePacketMode() {
		setFrameSize(DATABITS_9);
	}

	public byte read() {
		while (ATMega128.registers.UCSR0A.isBitClear(ATMega128.RXC));
		return (byte) ATMega128.registers.UDR0.get();
	}

	public void readPacket(Memory data, short length, boolean needStart) {

		if (length <= 0) {
			return;
		}


		InterruptService.disableAll();
		receiveComplete = false;
		receiveBuffer = data;
		receiveLength = length;
		receiveIndex = (short) 0;
		receivePacketStart = needStart;
		InterruptService.enableAll();

		ATMega128.registers.UCSR0B.setBit(ATMega128.RXCIE);

		while (!receiveComplete);
	}

	private static void receiveCompleteInterrupt() {
		if (receiveBuffer == null) {    // The data byte gets lost!
			ATMega128.registers.UCSR0A.clearBit(ATMega128.RXC); // ISR will be called again after return if bit is still set!
			return;
		}

		if (receivePacketStart) {
			// First byte of the packet must have the ninth bit set
			if (receiveIndex == 0 && ATMega128.registers.UCSR0B.isBitClear(ATMega128.RXB8)) {
				ATMega128.registers.UCSR0A.clearBit(ATMega128.RXC); // ISR will be called again after return if bit is still set!
				return;
			} else {
				receivePacketStart = false;
			}
		}

		receiveBuffer.set8(receiveIndex++, (byte) ATMega128.registers.UDR0.get());

		if (receiveIndex >= receiveLength) {
			ATMega128.registers.UCSR0B.clearBit(ATMega128.RXCIE);
			receiveComplete = true;
			receiveBuffer = null;
		}
	}

	public void writePacket(Memory data, short length) {

		if (length <= 0) {
			return;
		}

		while (!transmitComplete);  // wait until it is safe to write

		transmitComplete = false;
		transmitBuffer = data;
		transmitLength = length;
		transmitIndex = 0;

		// enable the interrupt
		ATMega128.registers.UCSR0B.setBit(ATMega128.UDRIE);
	}

	private static void transmitCompleteInterrupt() {
		if (transmitIndex >= transmitLength) {
			// disable the interrupt and prevent a infinite interrupt loop
			ATMega128.registers.UCSR0B.clearBit(ATMega128.UDRIE);
			transmitComplete = true;
			return;
		}

		// First data byte must have the ninth bit set in the first data byte
		if (transmitIndex == 0) {
			ATMega128.registers.UCSR0B.setBit(ATMega128.TXB8);
		} else {
			ATMega128.registers.UCSR0B.clearBit(ATMega128.TXB8);
		}

		byte data = transmitBuffer.get8(transmitIndex);
		++transmitIndex;
		ATMega128.registers.UDR0.set(data);
	}

	public void write(byte data) {
		while (ATMega128.registers.UCSR0A.isBitClear(ATMega128.UDRE));
		ATMega128.registers.UDR0.set(data);
	}

	public void puts(String s){
		for(int i=0; i<s.length(); i++) {
			write((byte)s.charAt(i));
		}
	}

	public void putcarr(char[] arr){
		for(int i = 0; i<arr.length; i++){
			write((byte)arr[i]);
		}
	}

	public Memory allocTransmitBuffer(short size, int timeout) {
		//FIXME
		//return MemoryService.allocMemory(size, timeout);
		return MemoryService.allocStaticMemory(32);
	}

	public Memory allocReceiveBuffer(short size, int timeout) {
		//FIXME
		//return MemoryService.allocMemory(size, timeout);
		return MemoryService.allocStaticMemory(32);
	}
}
