/**(c)

  Copyright (C) 2006-2012 Christian Wawersich, Michael Stilkerich, Christoph Erhardt

  This file is part of the KESO Java Runtime Environment.

  KESO is free software: you can redistribute it and/or modify it under the
  terms of the Lesser GNU General Public License as published by the Free
  Software Foundation, either version 3 of the License, or (at your option)
  any later version.

  KESO is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
  more details. You should have received a copy of the GNU Lesser General
  Public License along with KESO. If not, see <http://www.gnu.org/licenses/>.

  Please contact keso@cs.fau.de for more info.

  (c)**/

package trafficlight;

import keso.core.*;

import keso.driver.avr.atmega128.ATMega128;

/**
 * Implementation of the actual traffic light control logic.
 */
public class TrafficLightTask implements Runnable {
	private static TrafficLightTask _inst;

	// bitmasks for the required LED state combinations
	private static final byte PINCONF_GREEN     = 5;
	private static final byte PINCONF_YELLOW    = 6;
	private static final byte PINCONF_YELLOWRED = 2;
	private static final byte PINCONF_RED       = 3;

	// ids for the four possible states the traffic light can enter
	// usable as indices to the states array
	private static final byte STATE_GREEN     = 1;
	private static final byte STATE_YELLOW    = 2;
	private static final byte STATE_YELLOWRED = 3;
	private static final byte STATE_RED       = 0;

	// class initializer:
	// configure the LED pins as output on startup
	static {
		// leds are active low (0=on, 1=off)
		ATMega128.registers.DDRA.setBit(ATMega128.DDA0); // yellow
		ATMega128.registers.DDRA.setBit(ATMega128.DDA1); // green
		ATMega128.registers.DDRA.setBit(ATMega128.DDA2); // red
	}

	// the traffic light is red initially
	private byte curState = STATE_RED;

	// description of the states that the traffic light can enter
	final static private byte[][] states = {
		//  LED state        seconds to next state     Next State
		{ PINCONF_RED,              (byte) 0,          STATE_YELLOWRED },
		{ PINCONF_GREEN,            (byte) 5,          STATE_YELLOW    },
		{ PINCONF_YELLOW,           (byte) 1,          STATE_RED       },
		{ PINCONF_YELLOWRED,        (byte) 1,          STATE_GREEN     },
	};

	// constructor of the task system object
	// sets the traffic light to its initial state, i.e. red
	public TrafficLightTask() {
		_inst = this;
		setLights(curState);
	}

	// change the LED state to reflect a new traffic light state
	// if the state is temporary (i.e., autotransition after
	// a certain amount of time), windup an alarm to trigger
	// the tlTask after the delay has passed
	private void setLights(byte state) {
		// clear led bits...
		ATMega128.registers.PORTA.and(0xf8);
		// ...apply new setting
		ATMega128.registers.PORTA.or(states[state][0]);

		// windup alarm if necessary
		if(states[state][1] > 0) {
			// retrieve system object of tlAlarm
			Alarm alrm = AlarmService.getAlarmByName("tlAlarm");

			// windup the alarm
			AlarmService.setRelAlarm(alrm, states[state][1], 0);
		}
	}

	/**
	 * Entry function of the tlTask task.
	 *
	 * This task is activated whenever a temporary state needs to be left.
	 */
	public void run() {
		// set curState to the next state
		curState = states[curState][2];

		// redisplay lights according to the new state
		setLights(curState);

		// terminate the task
		TaskService.terminate();
	}

	/**
	 * Handle a switch request.
	 *
	 * The function will trigger a switch sequence to green and back to red,
	 * but only, if the light is currently in the red state. Otherwise the
	 * request will be rejected.
	 */
	public static boolean signalSwitch() {
		if(_inst.curState != STATE_RED)
			return false;
		_inst.setLights(states[STATE_RED][2]);
		return true;
	}
}

